#!/usr/bin/env python

import sys

from migration.objects import version
from poaupdater import uConfig
from poaupdater import uLogging, uSysDB, uConfig, openapi

uLogging.do_time = True

config = None


def doInit(options):
    global config

    def __useOption(key, attr, default=None):
        if key in options:
            value = options[key]
            if value == "":
                value = True
            setattr(config, attr, value)
        else:
            setattr(config, attr, default)

    config = uConfig.Config(
        verbose='--verbose' in options
    )

    __useOption("--subscription-id", "subscriptionId")
    __useOption("--rt-id", "rtId")
    __useOption("--merge-to", "dstResourceId")

    config.log_file = '/var/log/merge-resources.log'

    uLogging.init(config)
    uSysDB.init(config)
    openapi.initFromEnv(config)


def printUsage():
    print 'Usage: %s --subscription-id <subscription-id> --rt-id <rt-id> --merge-to <dst-rt-id>' % sys.argv[0]
    print ' to merge subscription <subscription-id> resources of resource type <rt-id> to <dst-rt-id>'
    for opt in __long_options:
        print "    --%s - %s" % (opt[0].rstrip('='), opt[1])

__long_options = (
    ('help', 'prints this message and exits (all other arguments are ignored)'),
    ('verbose', 'if specified, will cause detailed trace output to log'),
    ('subscription-id=', 'subscription id'),
    ('rt-id=', 'resource type id'),
    ('merge-to=', 'destination resource type id')
)


def get_rti_id(con, sub_id, rt_id):
    cur = con.cursor()
    cur.execute("SELECT rt_instance_id FROM subs_resources WHERE sub_id = %s and rt_id = %s", sub_id, rt_id)
    rti_id = cur.fetchone()[0]
    cur.close()
    return rti_id


def db_get_provider_ids(con, sub_id, db_rti_id):
    cur = con.cursor()
    cur.execute("SELECT provide_id FROM dbm_provide_info WHERE sub_id = %s and db_rti_id = %s", sub_id, db_rti_id)
    data = cur.fetchall()
    cur.close()
    provider_ids = [row[0] for row in data]
    return provider_ids


def db_switch_resource(con, provider_id, new_db_rti_id):
    cur = con.cursor()
    cur.execute("SELECT db_rti_id FROM dbm_provide_info WHERE provide_id = %s", provider_id)
    db_rti_id = cur.fetchone()[0]
    cur.close()
    uLogging.info("Switching DB resource type instance for provider id %d from %d to %d",
                  provider_id, db_rti_id, new_db_rti_id)
    cur = con.cursor()
    cur.execute("""UPDATE dbm_provide_info
SET db_rti_id = %s
WHERE provide_id = %s""", new_db_rti_id, provider_id)
    cur.close()


def merge_db_resources(con, sub_id, rt_id, dst_rt_id):
    src_rti_id = get_rti_id(con, sub_id, rt_id)
    dst_rti_id = get_rti_id(con, sub_id, dst_rt_id)

    provider_ids = db_get_provider_ids(con, sub_id, src_rti_id)
    for provider_id in provider_ids:
        db_switch_resource(con, provider_id, dst_rti_id)


def merge_mysql_resources(con, sub_id, rt_id, dst_rt_id):
    merge_db_resources(con, sub_id, rt_id, dst_rt_id)

resource_classes = {
    'mysql': merge_mysql_resources,
}


def get_resource(con, sub_id, rt_id):
    cur = con.cursor()
    cur.execute("""SELECT rt.rt_id, rt.restype_name, rc.name, sr.sub_id, sr.sub_limit, sr.own_usage
FROM subs_resources sr
JOIN resource_types rt ON rt.rt_id = sr.rt_id
JOIN resource_classes rc ON rt.class_id = rc.class_id
WHERE sr.rt_id = %s and sr.sub_id = %s""", rt_id, sub_id)
    data = cur.fetchone()
    cur.close()
    if not data:
        raise Exception("Subscription %d have no resource type id %d" % (sub_id, rt_id))

    return data


def adjust_resources(con, sub_id, rt_id, dst_rt_id):
    src_resource = get_resource(con, sub_id, rt_id)
    dst_resource = get_resource(con, sub_id, dst_rt_id)

    src_usage = src_resource[5]
    dst_usage = dst_resource[5]

    uLogging.info("Adjusting resource usage for subscription %d resource type %d: increase usage by %d , previous value %d",
                  sub_id, dst_rt_id, src_usage, dst_usage)
    cur = con.cursor()
    cur.execute("""UPDATE subs_resources
SET curr_usage = curr_usage + %s, own_usage = own_usage + %s
WHERE sub_id=%s AND rt_id=%s""", src_usage, src_usage, sub_id, dst_rt_id)
    cur.close()


def remove_resource_from_subscription(con, sub_id, rt_id):
    cur = con.cursor()
    cur.execute("SELECT * FROM subs_resources WHERE sub_id = %s AND rt_id = %s", sub_id, rt_id)
    row = cur.fetchone()
    cur.close()

    uLogging.info("Removing subscription resource %s", str(row))

    cur = con.cursor()
    cur.execute("DELETE FROM subs_resources WHERE sub_id=%s and rt_id=%s", sub_id, rt_id)
    cur.close()


def checkResources(sub_id, rt_id, dst_rt_id):
    con = uSysDB.connect()

    try:
        src_resource = get_resource(con, sub_id, rt_id)
        src_rc_name = src_resource[2]
        if src_rc_name not in resource_classes:
            raise Exception("Cannot merge resources based on resource class %s", src_rc_name)
        dst_resource = get_resource(con, sub_id, dst_rt_id)
        dst_rc_name = dst_resource[2]
        if src_rc_name != dst_rc_name:
            raise Exception("Resources have different base class")

        src_usage = src_resource[5]

        dst_limit = dst_resource[4]
        dst_usage = dst_resource[5]

        if dst_limit != -1 and dst_limit < (src_usage + dst_usage):
            raise Exception("Cannot merge resources: resource %d have not enough limit %d required %d" %
                            (dst_rt_id, dst_limit, src_usage + dst_usage))
    finally:
        con.rollback()
        closeConnection(con)


def mergeResources(sub_id, rt_id, dst_rt_id):
    try:
        con = uSysDB.connect()

        src_resource = get_resource(con, sub_id, rt_id)
        src_rc_name = src_resource[2]
        merge_func = resource_classes[src_rc_name]

        merge_func(con, sub_id, rt_id, dst_rt_id)

        adjust_resources(con, sub_id, rt_id, dst_rt_id)
        remove_resource_from_subscription(con, sub_id, rt_id)

        con.commit()
    finally:
        con.rollback()
        closeConnection(con)


def main():
    try:
        sub_id = int(config.subscriptionId)
        rt_id = int(config.rtId)
        dst_rt_id = int(config.dstResourceId)
    except (AttributeError, ValueError), e:
        printUsage()
        sys.exit(1)

    checkResources(sub_id, rt_id, dst_rt_id)
    mergeResources(sub_id, rt_id, dst_rt_id)


def closeConnection(con):
    if version[0] == 6:
        con.close()
    else:
        uSysDB.close(con)

if __name__ == "__main__":
    import sys
    reload(sys)
    sys.setdefaultencoding('utf-8')

    import getopt
    try:
        opts, args = getopt.getopt(sys.argv[1:], '', dict(__long_options).keys())
        opts = dict(opts)
    except getopt.GetoptError, err:
        print str(err)
        printUsage()
        sys.exit(2)

    if opts and '--help' in opts:
        printUsage()
        sys.exit(0)

    try:
        doInit(opts)
        main()
    except Exception, e:
        uLogging.err("%s", str(e))
        sys.exit(1)
