#!/usr/bin/python
import os
import re

from poaupdater import uSysDB
from migutils import *

# global config instance
config = None

# version of RawConfigParser from python 2.4
# only difference is that it requires unique keys in config
DEFAULTSECT = "DEFAULT"

MAX_INTERPOLATION_DEPTH = 10

#
#	Errors
#


class PrerequisiteError(uSysDB.Error):

    def __init__(self, *args):
        uSysDB.Error.__init__(self, *args)


class MigrationError(uSysDB.Error):

    def __init__(self, *args):
        uSysDB.Error.__init__(self, *args)


# exception classes
class Error(Exception):

    """Base class for ConfigParser exceptions."""

    def __init__(self, msg=''):
        self.message = msg
        Exception.__init__(self, msg)

    def __repr__(self):
        return self.message

    __str__ = __repr__


class NoSectionError(Error):

    """Raised when no section matches a requested option."""

    def __init__(self, section):
        Error.__init__(self, 'No section: %r' % (section,))
        self.section = section


class DuplicateSectionError(Error):

    """Raised when a section is multiply-created."""

    def __init__(self, section):
        Error.__init__(self, "Section %r already exists" % section)
        self.section = section


class DuplicateOptionError(Error):

    """Multiple identical options in section"""

    def __init__(self, option):
        Error.__init__(self, "Duplicated option %s" % option)
        self.option = option


class NoOptionError(Error):

    """A requested option was not found."""

    def __init__(self, option, section):
        Error.__init__(self, "No option %r in section: %r" %
                       (option, section))
        self.option = option
        self.section = section


class ParsingError(Error):

    """Raised when a configuration file does not follow legal syntax."""

    def __init__(self, filename):
        Error.__init__(self, 'File contains parsing errors: %s' % filename)
        self.filename = filename
        self.errors = []

    def append(self, lineno, line):
        self.errors.append((lineno, line))
        self.message += '\n\t[line %2d]: %s' % (lineno, line)


class MissingSectionHeaderError(ParsingError):

    """Raised when a key-value pair is found before any section header."""

    def __init__(self, filename, lineno, line):
        Error.__init__(
            self,
            'File contains no section headers.\nfile: %s, line: %d\n%r' %
            (filename, lineno, line))
        self.filename = filename
        self.lineno = lineno
        self.line = line


class RawConfigParser:

    def __init__(self, defaults=None):
        self._sections = {}
        self._defaults = {}
        if defaults:
            for key, value in defaults.items():
                self._defaults[self.optionxform(key)] = value

    def defaults(self):
        return self._defaults

    def sections(self):
        """Return a list of section names, excluding [DEFAULT]"""
        # self._sections will never have [DEFAULT] in it
        return self._sections.keys()

    def add_section(self, section):
        """Create a new section in the configuration.

        Raise DuplicateSectionError if a section by the specified name
        already exists.
        """
        if section in self._sections:
            raise DuplicateSectionError(section)
        self._sections[section] = {}

    def has_section(self, section):
        """Indicate whether the named section is present in the configuration.

        The DEFAULT section is not acknowledged.
        """
        return section in self._sections

    def options(self, section):
        """Return a list of option names for the given section name."""
        try:
            opts = self._sections[section].copy()
        except KeyError:
            raise NoSectionError(section)
        opts.update(self._defaults)
        if '__name__' in opts:
            del opts['__name__']
        return opts.keys()

    def read(self, filenames):
        """Read and parse a filename or a list of filenames.

        Files that cannot be opened are silently ignored; this is
        designed so that you can specify a list of potential
        configuration file locations (e.g. current directory, user's
        home directory, systemwide directory), and all existing
        configuration files in the list will be read.  A single
        filename may also be given.

        Return list of successfully read files.
        """
        if isinstance(filenames, basestring):
            filenames = [filenames]
        read_ok = []
        for filename in filenames:
            try:
                fp = open(filename)
            except IOError:
                continue
            self._read(fp, filename)
            fp.close()
            read_ok.append(filename)
        return read_ok

    def readfp(self, fp, filename=None):
        """Like read() but the argument must be a file-like object.

        The `fp' argument must have a `readline' method.  Optional
        second argument is the `filename', which if not given, is
        taken from fp.name.  If fp has no `name' attribute, `<???>' is
        used.

        """
        if filename is None:
            try:
                filename = fp.name
            except AttributeError:
                filename = '<???>'
        self._read(fp, filename)

    def get(self, section, option):
        opt = self.optionxform(option)
        if section not in self._sections:
            if section != DEFAULTSECT:
                raise NoSectionError(section)
            if opt in self._defaults:
                return self._defaults[opt]
            else:
                raise NoOptionError(option, section)
        elif opt in self._sections[section]:
            return self._sections[section][opt]
        elif opt in self._defaults:
            return self._defaults[opt]
        else:
            raise NoOptionError(option, section)

    def items(self, section):
        try:
            d2 = self._sections[section]
        except KeyError:
            if section != DEFAULTSECT:
                raise NoSectionError(section)
            d2 = {}
        d = self._defaults.copy()
        d.update(d2)
        if "__name__" in d:
            del d["__name__"]
        return d.items()

    def _get(self, section, conv, option):
        return conv(self.get(section, option))

    def getint(self, section, option):
        return self._get(section, int, option)

    def getfloat(self, section, option):
        return self._get(section, float, option)

    _boolean_states = {'1': True, 'yes': True, 'true': True, 'on': True,
                       '0': False, 'no': False, 'false': False, 'off': False}

    def getboolean(self, section, option):
        v = self.get(section, option)
        if v.lower() not in self._boolean_states:
            raise ValueError, 'Not a boolean: %s' % v
        return self._boolean_states[v.lower()]

    def optionxform(self, optionstr):
        return optionstr.lower()

    def has_option(self, section, option):
        """Check for the existence of a given option in a given section."""
        if not section or section == DEFAULTSECT:
            option = self.optionxform(option)
            return option in self._defaults
        elif section not in self._sections:
            return False
        else:
            option = self.optionxform(option)
            return (option in self._sections[section]
                    or option in self._defaults)

    def set(self, section, option, value):
        """Set an option."""
        if not section or section == DEFAULTSECT:
            sectdict = self._defaults
        else:
            try:
                sectdict = self._sections[section]
            except KeyError:
                raise NoSectionError(section)
        sectdict[self.optionxform(option)] = value

    def write(self, fp):
        """Write an .ini-format representation of the configuration state."""
        if self._defaults:
            fp.write("[%s]\n" % DEFAULTSECT)
            for (key, value) in self._defaults.items():
                fp.write("%s = %s\n" % (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")
        for section in self._sections:
            fp.write("[%s]\n" % section)
            for (key, value) in self._sections[section].items():
                if key != "__name__":
                    fp.write("%s = %s\n" %
                             (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")

    def remove_option(self, section, option):
        """Remove an option."""
        if not section or section == DEFAULTSECT:
            sectdict = self._defaults
        else:
            try:
                sectdict = self._sections[section]
            except KeyError:
                raise NoSectionError(section)
        option = self.optionxform(option)
        existed = option in sectdict
        if existed:
            del sectdict[option]
        return existed

    def remove_section(self, section):
        """Remove a file section."""
        existed = section in self._sections
        if existed:
            del self._sections[section]
        return existed

    #
    # Regular expressions for parsing section headers and options.
    #
    SECTCRE = re.compile(
        r'\['                                 # [
        r'(?P<header>[^]]+)'                  # very permissive!
        r'\]'                                 # ]
    )
    OPTCRE = re.compile(
        r'(?P<option>[^:=\s][^:=]*)'          # very permissive!
        r'\s*(?P<vi>[:=])\s*'                 # any number of space/tab,
                                              # followed by separator
                                              # (either : or =), followed
                                              # by any # space/tab
        r'(?P<value>.*)$'                     # everything up to eol
    )

    def _read(self, fp, fpname):
        """Parse a sectioned setup file.

        The sections in setup file contains a title line at the top,
        indicated by a name in square brackets (`[]'), plus key/value
        options lines, indicated by `name: value' format lines.
        Continuations are represented by an embedded newline then
        leading whitespace.  Blank lines, lines beginning with a '#',
        and just about everything else are ignored.
        """
        cursect = None                            # None, or a dictionary
        optname = None
        lineno = 0
        e = None                                  # None, or an exception
        while True:
            line = fp.readline()
            if not line:
                break
            lineno = lineno + 1
            # comment or blank line?
            if line.strip() == '' or line[0] in '#;':
                continue
            if line.split(None, 1)[0].lower() == 'rem' and line[0] in "rR":
                # no leading whitespace
                continue
            # continuation line?
            if line[0].isspace() and cursect is not None and optname:
                value = line.strip()
                if value:
                    cursect[optname] = "%s\n%s" % (cursect[optname], value)
            # a section header or option header?
            else:
                # is it a section header?
                mo = self.SECTCRE.match(line)
                if mo:
                    sectname = mo.group('header')
                    if sectname in self._sections:
                        cursect = self._sections[sectname]
                    elif sectname == DEFAULTSECT:
                        cursect = self._defaults
                    else:
                        cursect = {'__name__': sectname}
                        self._sections[sectname] = cursect
                    # So sections can't start with a continuation line
                    optname = None
                # no section header in the file?
                elif cursect is None:
                    raise MissingSectionHeaderError(fpname, lineno, line)
                # an option line?
                else:
                    mo = self.OPTCRE.match(line)
                    if mo:
                        optname, vi, optval = mo.group('option', 'vi', 'value')
                        if vi in ('=', ':') and ';' in optval:
                            # ';' is a comment delimiter only if it follows
                            # a spacing character
                            pos = optval.find(';')
                            if pos != -1 and optval[pos - 1].isspace():
                                optval = optval[:pos]
                        optval = optval.strip()
                        # allow empty values
                        if optval == '""':
                            optval = ''
                        optname = self.optionxform(optname.rstrip())
                        if not optname in cursect:
                            cursect[optname] = optval
                        else:
                            raise DuplicateOptionError(optname)
                    else:
                        # a non-fatal parsing error occurred.  set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain a
                        # list of all bogus lines
                        if not e:
                            e = ParsingError(fpname)
                        e.append(lineno, repr(line))
        # if any parsing errors occurred, raise an exception
        if e:
            raise e


class MigrationConfig:

    def __init__(self):
        self.templateMapping = {}
        self.planMapping = {}

    def __str__(self):
        return "MigrationConfig:(templates:%s;plans:%s)" % (self.templateMapping, self.planMapping)

    def __repr__(self):
        return self.__str__()

    def values(self):
        if self.templateMapping:
            mapping = self.templateMapping
        else:
            mapping = self.planMapping

        result = []
        for val in mapping.values():
            if isinstance(val, tuple):
                result.append(val[0])
            else:
                result.append(val)

        return result

    def items(self):
        if not config.withpba:
            return self.templateMapping.items()
        else:
            return self.planMapping.items()

    def read(self, filename):
        if not os.path.exists(filename) or not os.path.isfile(filename):
            raise MigrationError("Cannot open the migration configuration file '%s'" % filename)

        parser = RawConfigParser()
        try:
            parser.read(filename)
        except DuplicateOptionError, e:
            raise PrerequisiteError(
                "The migration configuration file '%s' contains multiple source service templates or plans with id '%s'." %
                (filename, e.option)
            )
        except MissingSectionHeaderError:
            raise PrerequisiteError(
                "The migration configuration file '%s' must start with a section ('[service templates]' or '[service plans]')." %
                (filename)
            )

        try:
            if not config.withpba:
                configMapping = parser.items("service templates")
            else:
                configMapping = parser.items("service plans")
        except NoSectionError, e:
            raise PrerequisiteError(
                "The migration configuration file '%s' doesn't contain section '%s'." % (filename, e.section)
            )

        # in case of PBA integration we use service plans mapping
        # in case of POA standalone migration we use service template mapping
        if not config.withpba:
            for item in configMapping:
                try:
                    src, dst = int(item[0]), int(item[1])
                except ValueError, e:
                    raise PrerequisiteError("Service template ID must be an integer value: %s" % str(e))

                if src in self.templateMapping:
                    raise PrerequisiteError(
                        "Invalid config file '%s'. Multiple destination service templates specified for one source %d." % (filename, src))

                self.templateMapping[src] = dst
        else:
            for item in configMapping:
                try:
                    src, dst = item[0], item[1]
                    # src could be in form <SourceServicePlanID>[,<SourceSubscriptionPeriodID>]
                    if -1 != src.find(","):
                        parts = src.split(",")
                        src, srcPeriod = int(parts[0]), int(parts[1])
                    else:
                        src = int(src)
                        srcPeriod = None

                    # dst could be in form <DestinationServicePlanID>[,<DestinationSubscriptionPeriodID>]
                    if -1 != dst.find(","):
                        parts = dst.split(",")
                        dst, dstPeriod = int(parts[0]), int(parts[1])
                    else:
                        dst = int(dst)
                        dstPeriod = None
                except ValueError, e:
                    raise PrerequisiteError("Service plan ID must be an integer value: %s" % str(e))

                self.planMapping[(src, srcPeriod)] = (dst, dstPeriod)


class RevertConfig:

    def __init__(self, subs, migrationConfig):
        self.mapping = {}
        for sub in subs:
            if config.withpba:
                servicePlanID, _ = getTargetServicePlanInfo(
                    sub.initialServicePlanID, sub.initialSubscriptionPeriodID, migrationConfig
                )
                self.mapping[(sub.subID, servicePlanID)] = (sub.initialServicePlanID, sub.initialSubscriptionPeriodID)
            else:
                self.mapping[
                    (sub.subID, migrationConfig.templateMapping[sub.initialServiceTemplateID])
                ] = sub.initialServiceTemplateID

    def items(self):
        result = []
        for key, value in self.mapping.items():
            result.append((key[1], value))

        return result

    def values(self):
        result = []
        for val in self.mapping.values():
            if isinstance(val, tuple):
                result.append(val[0])
            else:
                result.append(val)

        return result


def getTargetServicePlanInfo(servicePlanID, subscriptionPeriodID, cfg):
    if (servicePlanID, subscriptionPeriodID) in cfg.planMapping:
        return cfg.planMapping[(servicePlanID, subscriptionPeriodID)]
    elif (servicePlanID, None) in cfg.planMapping:
        return cfg.planMapping[(servicePlanID, None)]
    else:
        raise PrerequisiteError(
            "The mapping for the source service plan with ID '%s' and the subscription period with ID '%s' is not specified in the migration configuration file." % (
                servicePlanID, subscriptionPeriodID
            )
        )
