#!/usr/bin/python

import time

from poaupdater import uTasks, uSysDB, uLogging, uUtil, uHCL, uDialog, openapi, pbaapi, PEMVersion
try:
    import subprocess as sp
except:
    from poaupdater import pasubprocess as sp
import sets

from migutils import *

import config
from config import PrerequisiteError, MigrationError, getTargetServicePlanInfo

HOST_PROVISIONING_NOT_READY = 2

IP_TYPE_SHARED = 0
IP_TYPE_RENEW = 5

PHP_MODULE_HANDLER = 'modules/libphp5.so'

CONFIGURED_IP_TYPE_EXCLUSIVE = 'E'
CONFIGURED_IP_TYPE_SHARED = 'S'

HOST_TYPE_WEB_SERVER = 'g'
HOST_TYPE_WEB_CLUSTER = 'e'
HOST_TYPE_WEB_CLUSTER_SERVER = 'f'
HOST_TYPE_NORMAL = 'n'

# The maximum number of attempts to execute the function and the interval between the attempts.
DEFAULT_MAX_ATTEMPTS = 120
DEFAULT_INTERVAL_BETWEEN_ATTEMPTS = 5

version = PEMVersion.getPEMVersion().version


def prepareSelectArguments(args):
    if version[0] == 7:
        return args

    res = []
    for arg in args:
        if isinstance(arg, tuple):
            res.append([x for x in arg])
        else:
            res.append(arg)

    return res


class HostFactory:
    hosts = {}

    def createHost(cls, hostID):
        if hostID not in cls.hosts:

            con = uSysDB.connect()
            cur = con.cursor()

            cur.execute("SELECT htype FROM hosts WHERE host_id = %s", hostID)
            row = cur.fetchone()

            if not row:
                raise MigrationError("Failed to find host with ID %s" % hostID)

            if HOST_TYPE_WEB_CLUSTER == row[0]:
                cls.hosts[hostID] = WebCluster(hostID)
            else:
                cls.hosts[hostID] = Host(hostID)

        return cls.hosts[hostID]

    createHost = classmethod(createHost)


class Host:

    class NetInterface:

        def __init__(self, db_row):
            self.ifID = db_row[0]
            self.hostID = db_row[1]
            self.ifName = db_row[2]
            self.ifDescription = db_row[3]

    def __init__(self, hostID):
        self.hostID = int(hostID)
        self.id = self.hostID

        con = uSysDB.connect()
        cursor = con.cursor()

        cursor.execute(
            "SELECT primary_name, htype, opsys FROM hosts h, platforms p WHERE h.platform_id=p.platform_id and host_id = %s",
            self.hostID
        )
        rv = cursor.fetchone()
        if not rv:
            raise MigrationError("Failed to get info for host with id %s. Probably it doesn't exist." % self.hostID)

        self.name = rv[0]
        self.type = rv[1]
        self.is_win = rv[2].lower().startswith('win')

        self.groups = {}
        cursor.execute("SELECT name, sys_id FROM host_users_groups WHERE user_group = 'G' and host_id = %s", self.hostID)
        for row in cursor.fetchall():
            self.groups[row[0]] = row[1]

    def __str__(self):
        return "host '%s'(%s)" % (self.name, self.hostID)

    def __eq__(self, other):
        if not other:
            return False

        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def getType(self):
        return self.type

    def isVE(self):
        return 'v' == self.getType()

    # TODO consider moving VE related functions to separate class
    def getVEAttribute(self, attribute):
        con = uSysDB.connect()
        cursor = con.cursor()

        cursor.execute("SELECT " + attribute + " FROM ves WHERE ve_host_id = %s", self.hostID)
        rv = cursor.fetchone()
        if not rv:
            raise MigrationError("Failed to get %s attribute of VE with id %s" % (attribute, self.hostID))

        return rv[0]

    def getServiceID(self, serviceType, cursor=None):
        con = uSysDB.connect()
        cursor = con.cursor()

        cursor.execute("SELECT service_id FROM services WHERE name = %s AND host_id = %s", serviceType, self.hostID)
        rv = cursor.fetchone()
        if not rv:
            raise MigrationError("Failed to find service '%s' on %s" % (serviceType, self))

        service_id = int(rv[0])
        if serviceType == "domain_parking" or serviceType == "DomainParkingNG":
            # check if DomainParking service is already bound and fail if not
            # just not to reimplement binding logic here
            cursor.execute("SELECT service_id FROM dp_bound_services WHERE service_id = %s", service_id)
            rv = cursor.fetchone()
            if not rv:
                raise MigrationError("Cannot migrate parked domains to %s until at least one was provisioned" % self)

        cursor.close()

        return service_id

    def getMainProftpdVhostID(self):
        con = uSysDB.connect()
        cur = con.cursor()

        try:
            srcFTPService = self.getServiceID("Proftpd", cur)

            # get host shared ip
            ips = self.getIPs(CONFIGURED_IP_TYPE_SHARED)
            if ips:
                normalized_ips = [normalizeIP(ip) for ip in ips]
                cur.execute(
                    "SELECT pvh.vhost_id, pi.ip FROM proftpd_vhosts pvh JOIN proftpd_ips pi ON pvh.vhost_id = pi.vhost_id WHERE pvh.service_id = %s AND pvh.vhost_type = 1 AND pi.ip IN %s",
                    srcFTPService, tuple(normalized_ips)
                )
            else:
                assert False, "migration by subscription, but there is no shared IP on source host"

            rv = cur.fetchone()
            if not rv:
                raise MigrationError("Failed to find main ProFTPD vhost for service %s on %s" % (srcFTPService, self))
            else:
                return rv[0], rv[1]
        finally:
            cur.close()

    def hasBackups(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT group_id FROM backups_groups WHERE host_id = %s", self.hostID)
        rv = cur.fetchone()

        if not rv:
            return False
        else:
            return True

    def serviceCtl(self, command, service):
        return self.cmd("service %s %s" % (service.lower(), command.lower()))

    def getAgentID(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT pleskd_id FROM hosts WHERE host_id = %s", self.hostID)
        row = cur.fetchone()
        if not row:
            raise MigrationError("Failed to get %s agent id" % self)

        return row[0]

    def getBacknetIP(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute(
            "SELECT vp.value FROM components c JOIN packages p ON c.pkg_id = p.pkg_id JOIN v_props vp ON c.component_id = vp.component_id WHERE vp.name = 'communication.ip' AND p.name = 'pleskd' AND c.host_id = %s",
            self.getOperationHostID()
        )
        row = cur.fetchone()
        if not row:
            raise MigrationError("Failed to find %s backnet IP" % self)

        return row[0].strip()

    def getDomainParkingIP(self):
        service_id = self.getServiceID("domain_parking")
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT service_ip FROM dp_bound_services WHERE service_id = %s", service_id)
        row = cur.fetchone()
        if not row:
            raise MigrationError("Failed to determine domain parking service %s ip on %s" % (service_id, self))

        return row['service_ip']

    def updateCustomerServices(self, cur):
        cur.execute(
            "UPDATE hosts_extended_info SET cs_number = (SELECT count(1) FROM domain_services ds JOIN services s ON s.service_id = ds.service_id WHERE s.host_id = hosts_extended_info.host_id) WHERE host_id = %s",
            self.hostID
        )

    def getSubscriptionSet(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute(
            """
            SELECT DISTINCT ds.subscription_id, hd.hd_id, hd.host_id, dds.ds_id
            FROM home_directories AS hd
                INNER JOIN domain_services AS ds ON ds.hd_id = hd.hd_id
                LEFT JOIN dp_domain_services AS dds ON ds.ds_id = dds.ds_id
            WHERE
                hd.host_id = %s
            """,
            self.hostID
        )

        rows = cur.fetchall()
        return SubscriptionSet(rows)

    def getHostedBrands(self):
        con = uSysDB.connect()
        cur = con.cursor()
        if version[0] == 6:
            cur.execute(
                """
                SELECT b.brand_id, b.brand_name, b.domain_name
                FROM brands b
                JOIN apache_homedir_entries ahe ON (b.domain_name = ahe.domain_name)
                JOIN home_directories hd ON (ahe.homedir_id = hd.hd_id)
                WHERE host_id = %s
                """,
                self.hostID
            )
        else:
            cur.execute(
                """
                SELECT b.brand_id, b.brand_name, d.name
                FROM brands b
                JOIN domains d ON (b.domain_id = d.domain_id)
                JOIN apache_homedir_entries ahe ON (d.name = ahe.domain_name)
                JOIN home_directories hd ON (ahe.homedir_id = hd.hd_id)
                WHERE host_id = %s
                """,
                self.hostID
            )

        rv = cur.fetchall()
        cur.close()
        return rv

    def getBoundIPPools(self, cur):
        cur.execute(
            "SELECT pool_id FROM bound_ip_pools bp JOIN net_interfaces ni ON bp.if_id = ni.if_id WHERE ni.host_id = %s",
            self.hostID
        )
        rv = cur.fetchall()
        result = [row[0] for row in rv]
        return result

    def checkServicePrerequisites(self):
        con = uSysDB.connect()
        cur = con.cursor()

        requiredServices = ['Apache', 'Proftpd', 'logrotate', 'logparser']
        domainParkingRequiredServices = ['domain_parking']

        services = self.getHostServices(cur)
        # there must be at least domain parking on node
        missingServices = sets.Set(domainParkingRequiredServices) - sets.Set(services)
        if missingServices:
            missingServices = sets.Set(requiredServices) - sets.Set(services)

        if missingServices:
            raise PrerequisiteError(
                "Unable to migrate webspaces from %s. Some of the required services were not found on it: %s." %
                (self, ', '.join(missingServices))
            )

    def checkHostPrerequisites(self, toHost):
        con = uSysDB.connect()
        cur = con.cursor()

        hostState = self.getHostProvisioningState(cur)
        if HOST_PROVISIONING_NOT_READY != hostState:
            raise PrerequisiteError("%s must be marked as 'not ready to provide' before migration" % self)

        if self.isVE() and "y" == self.getVEAttribute("offline_management"):
            raise PrerequisiteError("Offline management must be disabled for source VPS %s before migration" % self)

        srcWebalizer = self.getWebalizerName(cur)
        dstWebalizer = toHost.getWebalizerName(cur)
        if srcWebalizer and srcWebalizer != dstWebalizer:
            raise PrerequisiteError("Cannot migrate %s, destination host doesn't have compatible version of webalizer. Source: %s; Destination: %s" % (
                self, srcWebalizer, dstWebalizer))

        if self.hasBackups():
            answer = uDialog.askYesNo(
                "There are customers backups stored on %s. After migration users won't be able to automatically restore them from customer control panel. Do you want to continue?" % self)
            if not answer:
                raise MigrationError("Aborted")

        srcIPPools = self.getBoundIPPools(cur)
        dstIPPools = toHost.getBoundIPPools(cur)
        leftIPPools = sets.Set(srcIPPools) - sets.Set(dstIPPools)
        if leftIPPools:
            raise PrerequisiteError("All IP addresses from %s will be assigned to %s. All IP pools that are bound to source host need to be bound to destination host too: %s" % (
                self, toHost, ", ".join([str(ippool) for ippool in leftIPPools])))

        ipsWithoutPool = self.getIPsWithoutPool(cur)
        if ipsWithoutPool:
            raise PrerequisiteError("Cannot migrate %s, it provides hosting for sites on shared IPs, that are not bound to any IP pool: %s" % (
                self, ", ".join(ipsWithoutPool)))

        brands = self.getHostedBrands()
        if brands:
            raise PrerequisiteError("Cannot migrate %s, it provides hosting for brands: %s" % (
                self, ", ".join(["%s(%s)(%s)" % (name, brandid, domain) for brandid, name, domain in brands])))

        cur.close()

    def checkConfigPrerequisites(self, subs, config):
        # check, that service template mapping config is acceptable
        pass

    def getHostProvisioningState(self, cur):
        cur.execute("SELECT provisioning FROM hosts WHERE host_id = %s", self.hostID)
        rv = cur.fetchone()
        return rv[0]

    def getHostServices(self, cur):
        cur.execute("SELECT name FROM services WHERE host_id = %s", self.hostID)
        rows = cur.fetchall()
        services = [row[0] for row in rows]
        return services

    def getWebalizerName(self, cur):
        cur.execute(
            "SELECT p.name FROM components c JOIN packages p ON c.pkg_id = p.pkg_id WHERE p.name ILIKE '%%webalizer%%' AND c.host_id = %s",
            self.hostID
        )
        rv = cur.fetchone()
        if not rv:
            return None
        else:
            return rv[0]

    def getIPsWithoutPool(self, cur):
        cur.execute(
            "select distinct(host_ip) from apache_hostlist ah join configured_ips ip on ah.host_ip = h2e_denormalize_ip(ip.ip_address) where ip.pool_id = 0 and ip.host_id = %s",
            self.hostID
        )
        rows = cur.fetchall()
        ips = [row[0] for row in rows]
        return ips

    def getIPs(self, type, subscriptionID=None):
        con = uSysDB.connect()
        cur = con.cursor()

        if not subscriptionID:
            cur.execute(
                "SELECT h2e_denormalize_ip(ip_address) FROM configured_ips WHERE ip_type = %s AND host_id = %s",
                type, self.hostID
            )
        else:
            cur.execute(
                "SELECT h2e_denormalize_ip(ip_address) FROM configured_ips WHERE ip_type = %s AND host_id = %s AND sub_id = %s",
                type, self.hostID, subscriptionID
            )

        rows = cur.fetchall()
        ips = [row[0] for row in rows]
        return ips

    def moveConfiguredIP(self, cur, fromHost, ip):
        query = """
                    UPDATE configured_ips
                    SET if_id = %s, host_id = n.host_id
                    FROM net_interfaces AS n
                    WHERE n.host_id = %s AND configured_ips.host_id = %s AND ip_address = %s
                    AND configured_ips.ip_type = 'E'
                """
        cur.execute(query, self.getPublicNetIface().ifID, self.hostID, fromHost.hostID, ip)

    def updateLogParser(self):
        uLogging.info("Updating logparser configuration for %s", self)
        openapi.OpenAPI().pem.proclog.dumpLogConfig(host_id=self.hostID)

    def syncState(self, subs):
        pass

    def deleteCrontabs(self, serviceID=None):
        if not serviceID:
            try:
                serviceID = self.getServiceID("crontabs")
            except MigrationError, e:
                uLogging.info("Cannot delete crontabs from %s: %s", self, e)
                pass

        if serviceID:
            pobjref = uTasks.SCRef("crontabs")

            task = uTasks.Task(pobjref, "Remove crontab files from %s" % self, "taskCleanupHost")
            task.setParam("service_id", self.getServiceID("crontabs"))
            task.setParam("host_id", self.hostID)

            task.schedule()

            beautyRetriable(waitTasksCompleteRef)(pobjref)

    def dumpLogrotate(self):
        serviceID = self.getServiceID("logrotate")
        pobjref = uTasks.PObjRef(uTasks.SCRef("logrotate"), "getService", serviceID)
        task = uTasks.Task(ref=pobjref,
                           name="Dump logrotate configs for service #%s" % (serviceID,),
                           method="taskDump",
                           mutex="logrotate.%s" % (serviceID,))
        task.schedule()

    def getOperationHostID(self):
        return self.hostID

    def getVhostsRoot(self):
        return "/usr/local/pem/vhosts"

    def getPublicNetIface(self):
        cur = uSysDB.connect().cursor()
        try:
            # consider network interface with shared (in terms of POA) IP address as a public one
            cur.execute(
                "SELECT ni.if_id, ni.host_id, ni.if_name, ni.if_description FROM net_interfaces ni JOIN configured_ips ci ON (ni.if_id = ci.if_id) WHERE ci.ip_type = %s AND ni.host_id = %s",
                CONFIGURED_IP_TYPE_SHARED, self.hostID
            )
            ip = cur.fetchone()
            if not ip:
                # ok get first one
                cur.execute(
                    "SELECT if_id, host_id, if_name, if_description FROM net_interfaces ni WHERE ni.host_id = %s" % self.hostID
                )
                ip = cur.fetchone()

            return Host.NetInterface(ip)
        finally:
            cur.close()

    def shutdown(self):
        uLogging.info("Shutting down %s", self)
        try:
            self.cmd("shutdown -h now")
        except Exception, e:
            # ignore COMM_FAILURE errors, that can occur
            if -1 == str(e).find("COMM_FAILURE"):
                raise

    def cmd(self, command, args=None, valid_exit_codes=None):
        uLogging.info("%s# %s", self, command)
        hcl = uHCL.Request(self.getOperationHostID(), "root", "root")
        hcl.command(command, args=args, valid_exit_codes=valid_exit_codes or [0], stdout='stdout', stderr='stderr')
        return hcl.perform()

    def uptime(self):
        return self.cmd("uptime")

    def killSSHSessions(self):
        # kill all non root sshd processes
        return self.cmd("for i in `ps -ef | grep sshd | grep -v root | tr -s ' ' | cut -d' ' -f2`; do kill -TERM $i; done")

    def killFTPSessions(self):
        return self.cmd("service proftpd restart")

    def get_free_disk_space(self, location = "/usr/local", block_size = 1024):
        return int(self.cmd("df %s --block-size=%s | awk 'END{print $(NF-2)}'" % (location, block_size))['stdout'])


class WebCluster(Host):

    def __init__(self, id):
        Host.__init__(self, id)

        uLogging.info("Obtaining general information for web cluster/server #%s", self.hostID)

        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT director_host_id FROM apache_h2e_clusters WHERE cluster_id = %s", self.hostID)
        rv = cur.fetchone()
        if not rv:
            raise MigrationError("Failed to get info for cluster with id %s" % self.hostID)

        self.director = Host(rv[0])

        cur.execute(
            "SELECT h.default_rootpath, c.active_mpoint_id FROM hosts AS h, apache_h2e_clusters AS c WHERE c.cluster_id = h.host_id AND h.host_id = %s",
            self.hostID
        )
        rv = cur.fetchone()
        if not rv:
            raise MigrationError("Failed to get info for web cluster/server with id %s" % self.hostID)

        self.activeMountPointID = rv[1]
        self.activeMountPointPath = "/var/www/vhosts/%d" % (rv[1])

        cur.execute("SELECT primary_name FROM hosts WHERE host_id = %s" % self.hostID)
        rv = cur.fetchone()
        if not rv:
            raise MigrationError("Failed to get info for web cluster/server with id %s" % self.hostID)

        self.name = rv[0]

        cur.close()

    def __str__(self):
        return "web cluster/server '%s'(%s)" % (self.name, self.hostID)

    def getMountPoint(self, cur, mountPointID):
        cur.execute(
            "SELECT mpoint_id, fnode_id, root_path FROM apache_h2e_mount_points WHERE mpoint_id = %s",
            mountPointID
        )
        return cur.fetchone()

    def getNode(self, cur, nodeID):
        cur.execute("SELECT node_id, cluster_id, ip_address, type FROM apache_h2e_nodes WHERE node_id = %s", nodeID)
        return cur.fetchone()

    def checkConfigPrerequisites(self, subs, cfg):
        uLogging.info("Check if subscriptions from host can be provisioned on web cluster/server")

        destinations = cfg.values()
        if not config.config.withpba:
            uLogging.info("Validating destination service templates")
            for sub in subs.getSubs():
                try:
                    templateID = sub.getServiceTemplateID()
                    if templateID in destinations:
                        # template is one of destinations
                        continue

                    ServiceTemplate(cfg.templateMapping[templateID]).checkPrerequisites()
                except KeyError:
                    raise PrerequisiteError(
                        "The target service template was not found for the subscription %s that uses the service template %s." %
                        (sub.subID, templateID)
                    )
        else:
            uLogging.info("Validating destination service plans")
            for sub in subs.getSubs():
                try:
                    servicePlanID, subscriptionPeriodID = sub.getServicePlanInfo()
                    if servicePlanID in destinations:
                        continue

                    targetServicePlanInfo = getTargetServicePlanInfo(servicePlanID, subscriptionPeriodID, cfg)
                    ServicePlan(targetServicePlanInfo[0]).serviceTemplate.checkPrerequisites()
                except KeyError:
                    raise PrerequisiteError(
                        "The target service plan was not found for the subscription %s that uses the service plan %s." %
                        (sub.subID, servicePlanID)
                    )

        try:
            self.getMainProftpdVhostID()
        except MigrationError:
            raise PrerequisiteError(
                "Unable to find the main ProFTPD virtual host on %s. Make sure that at least one website with FTP access is provisioned there." % self
            )

        try:
            if subs.hasSubscriptionsWithDomainParking():
                self.getServiceID("domain_parking")
        except MigrationError:
            raise PrerequisiteError(
                "Parked domains cannot be migrated to %s until the domain parking service is properly initialized. Make sure that at least one domain is already parked there." % self
            )

    def syncState(self, subs):
        webspaceIDs = [webspace.id for sub in subs for webspace in sub.webspaces]

        uLogging.info("Synchronizing web cluster/server configuration database")
        if webspaceIDs:
            index = 0
            step = 100
            while index < len(webspaceIDs):
                uLogging.info("Batch %s\%s" % (index / step + 1, len(webspaceIDs) / step + 1))

                part = webspaceIDs[index:index + step]
                openapi.OpenAPI().pem.web_cluster.syncDatabase(clusterID=self.hostID, webspaceIDs=part)

                index += step
        else:
            openapi.OpenAPI().pem.web_cluster.syncDatabase(clusterID=self.hostID, webspaceIDs=[])

    def setUpIPs(self):
        uLogging.info("Updating web cluster/server IP addresses")
        # for WebCluster NG shstg will sync IPs on webservers and loadbalancer after this call
        openapi.OpenAPI().pem.web_cluster.updateClusterIP(clusterID=self.hostID)

        # bring up IPs
        if self.getType() in (HOST_TYPE_WEB_CLUSTER, HOST_TYPE_WEB_SERVER):
            db_ips = []
            host_ips = []

            # retrieve IPs from DB
            cur = uSysDB.connect().cursor()
            try:
                cur.execute(
                    "SELECT h2e_denormalize_ip(ci.ip_address), ci.netmask, ni.if_name FROM configured_ips ci JOIN net_interfaces ni ON ci.if_id = ni.if_id WHERE ci.host_id = %s",
                    self.hostID
                )
                db_ips = cur.fetchall()
            finally:
                cur.close()

            db_ips_only = [db_ip[0] for db_ip in db_ips]

            # retrieve IPs from host
            host_ips = self.retrieveIPs()
            host_ips_only = [host_ip[0] for host_ip in host_ips]

            ips_to_setup = [db_ip for db_ip in db_ips if not db_ip[0] in host_ips_only]
            ips_to_teardown = [host_ip for host_ip in host_ips if not host_ip[0] in db_ips_only]

            for ip in ips_to_teardown:
                self.__configureIP(*(ip + [True]))

            for ip in ips_to_setup:
                self.__configureIP(*ip)

    def retrieveIPs(self):
        if self.getType() in (HOST_TYPE_WEB_CLUSTER, HOST_TYPE_WEB_CLUSTER_SERVER):
            iface = "lo"
        else:
            iface = self.getPublicNetIface().ifName

        output = self.cmd("ip addr list dev %s | grep 'inet '" % iface)
        ips = []
        for line in output['stdout'].splitlines():
            # we need to process only secondary IPs
            if -1 != line.find("secondary"):
                ip = line.split()
                # netmask should not matter for removal of IP
                ips.append([ip[1].split('/')[0], '255.255.0.0', iface])

        return ips

    def __configureIP(self, ip_address, net_mask, iface, unconfigure=False):
        # IPv6 is not supported by this function
        if -1 != ip_address.find(':'):
            return

        logline = "Bringing up IP %s/%s at %s"
        if unconfigure:
            logline = "Remove IP %s/%s from %s"

        uLogging.info(logline, ip_address, net_mask, iface)

        hcl = HCLRequest(self.hostID, "root", "root")
        if not unconfigure:
            hcl.ipadd(ip_address, net_mask, iface)
        else:
            hcl.ipdel(ip_address, net_mask, iface)

        hcl.perform()

    def getWebalizerName(self, cur):
        return self.director.getWebalizerName(cur)

    def getOperationHostID(self):
        return self.director.hostID

    def getVhostsRoot(self):
        return self.activeMountPointPath

    def killSSHSessions(self):
        uLogging.debug("Killing active SSH sessions is not implemented for Web Cluster/Server NG.")

    def killFTPSessions(self):
        uLogging.debug("Killing active FTP sessions is not implemented for Web Cluster/Server NG.")

    def getServiceID(self, serviceType, cursor=None):
        services_to_translate = ["Apache", "Proftpd"]
        if serviceType in services_to_translate:
            serviceType += "NG"
        if serviceType == "domain_parking":
            serviceType = "DomainParkingNG"

        return Host.getServiceID(self, serviceType, cursor)


class ServiceTemplate:

    def __init__(self, serviceTemplateID):
        self.serviceTemplateID = serviceTemplateID
        self.id = self.serviceTemplateID

        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute(
            "SELECT st_id, for_sale, name, owner_id FROM service_templates WHERE st_id = %s",
            self.serviceTemplateID
        )
        rv = cur.fetchone()

        if not rv:
            self.__exists = False
            self.__active = False
            self.name = "N/A"
        else:
            self.__exists = True
            self.__active = "y" == rv[1]
            self.name = rv[2]
            self.ownerID = rv[3]

    def __str__(self):
        if not self.exists():
            return "service template 'N/A'(%s)" % (self.serviceTemplateID)
        else:
            return "service template '%s'(%s)" % (self.name, self.serviceTemplateID)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def exists(self):
        return self.__exists

    def isActive(self):
        return self.__active

    def getName(self):
        return self.name

    def getPHPHandlers(self, cur):
        phpHandlers = []
        # get resource type activation parameters. ignore service template activation parameters
        cur.execute(
            """
            SELECT rtap.name, rtap.value
            FROM resource_classes rc
            JOIN resource_types rt ON rc.class_id = rt.class_id
            JOIN st_resources str ON rt.rt_id = str.rt_id
            JOIN rt_act_params rtap ON rtap.rt_id = rt.rt_id
            WHERE str.st_id = %s AND rtap.name ilike 'path.to.php.binary' AND rc.name = 'shared_hosting_php'
            """,
            self.serviceTemplateID
        )
        rv = cur.fetchall()
        for name, value in rv:
            phpHandlers.append(value)

        return phpHandlers

    def getResourceLimit(self, cur, resourceName):
        cur.execute(
            "select str.sub_limit from resource_classes rc join resource_types rt on rc.class_id = rt.class_id join st_resources str on rt.rt_id = str.rt_id where str.st_id = %s and rc.name = %s",
            self.serviceTemplateID, resourceName
        )
        rv = cur.fetchone()
        if not rv:
            return 0
        else:
            return rv[0]

    def getResourceLimitById(self, cur, resourceId):
        cur.execute(
            """SELECT str.sub_limit
            FROM st_resources str
            WHERE str.st_id = %s and str.rt_id = %s""",
            self.serviceTemplateID, resourceId
        )

        rv = cur.fetchone()
        if not rv:
            return -1
        else:
            return rv[0]

    def checkCompatibility(self, template):
        assert self.exists() and template.exists()

        if self.ownerID != template.ownerID:
            raise MigrationError("%s and %s have different owners with IDs %s and %s." %
                                 (self, template, self.ownerID, template.ownerID))

    def checkPrerequisites(self):
        con = uSysDB.connect()
        cur = con.cursor()

        if not self.exists():
            raise PrerequisiteError("The service template with ID %s does not exist." % self.serviceTemplateID)

        phpHandlers = self.getPHPHandlers(cur)

        if PHP_MODULE_HANDLER in phpHandlers:
            raise PrerequisiteError(
                "Subscriptions using service templates with PHP as a module cannot be provisioned on NG hosting. Service template %s: %s" %
                (self.serviceTemplateID, phpHandlers)
            )

        if 0 != self.getResourceLimit(cur, "proftpd.res.anon_access"):
            raise PrerequisiteError(
                "NG hosting does not support anonymous FTP. The target service template %s should not allow it." % self.serviceTemplateID
            )

        if 0 != self.getResourceLimit(cur, "apache_fp_support"):
            raise PrerequisiteError(
                "NG hosting does not support FrontPage. The target service template %s should not allow it." % self.serviceTemplateID
            )

        cur.close()

    def getResources(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute(
            "SELECT rt.rt_id, rt.restype_name, rt.class_id, rc.name FROM resource_types rt JOIN resource_classes rc ON (rt.class_id = rc.class_id) WHERE rt.rt_id IN (SELECT rt_id FROM st_resources WHERE st_id = %s) AND parent_id = -1 ORDER BY rt.rt_id",
            self.serviceTemplateID
        )
        result = cur.fetchall()

        return result


class ServicePlan:
    # Service plan from PBA

    def __init__(self, servicePlanID):
        self.servicePlanID = servicePlanID
        self.id = self.servicePlanID

        try:
            api = pbaapi.PBAAPI()
            info = api.PlanGet(self.servicePlanID)

            self.__exists = True
            self.__active = 0 == info.Published
            self.name = info.Name
            self.serviceTemplate = ServiceTemplate(info.ServiceTemplateID)
        except:
            self.__exists = False
            self.__active = False

    def __str__(self):
        if self.exists():
            return "service plan '%s'(%s)" % (self.name, self.id)
        else:
            return "service plan 'N/A'(%s)" % (self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def exists(self):
        return self.__exists

    def isActive(self):
        return self.__active

    def checkCompatibility(self, plan):
        assert self.exists() and plan.exists()

        return self.serviceTemplate.checkCompatibility(plan.serviceTemplate)


class BaseWebspaceSet:

    def __init__(self, hostID, webspaceIDs):
        self.webspaceIDs = webspaceIDs

        self.host = HostFactory.createHost(hostID)
        self.originalHost = self.host

    def countDomainParkingWebspaces(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute(
            "SELECT count(*) as c FROM dp_domain_services dds WHERE dds.homedir_id in %s",
            prepareSelectArguments((tuple(self.webspaceIDs),))
        )
        return cur.fetchone()['c']

    def migrateBackups(self, toHost):
        pass

    def syncContent(self, toHost):
        # nothing to sync for domain parking webspaces
        pass

    def switchToSharedIPType(self):
        # nothing to switch for domain parking
        pass

    def finalizeContent(self, toHost):
        # nothing to do for domain parking webspaces
        pass

    def restoreWebspaceProperties(self, cur):
        # nothing to restore for domain parking webspaces
        pass

    def dumpCrontabs(self):
        # there cannot be any crontabs on domain parking webspaces
        pass

    def checkContentForAbsPaths(self, toHost):
        # there is no content in domain parking webspaces
        pass

    def createLocalPear(self):
        # there cannot be any pear in domain parking webspace
        pass

    def syncHtAccess(self):
        # no content -> no .htaccess
        pass

    def migrateIPDBData(self, cur, toHost, revert):
        pass

    def initWebspaces(self, hostID, webspaceIDs):
        raise NotImplementedError()

    def migrateDBData(self, con, cur, host, revert):
        raise NotImplementedError()

    def restoreWebspaceIPTypes(self):
        raise NotImplementedError()


class DomainParkingWebspaceSet(BaseWebspaceSet):

    def __init__(self, hostID, webspaceIDs):
        BaseWebspaceSet.__init__(self, hostID, webspaceIDs)

        # check, that all of these webspaces are really domain parking
        assert len(self.webspaceIDs) == self.countDomainParkingWebspaces()

        self.webspaceDict = self.initWebspaces(hostID, webspaceIDs)
        self.webspaces = self.webspaceDict.values()

    def initWebspaces(self, hostID, webspaceIDs):
        if not webspaceIDs:
            return {}

        webspaces = {}
        for webspaceID in webspaceIDs:
            # all those values, a really not necessary for domain parking webspace
            webspace = Webspace.create(webspaceID, None, None, None, None, None)
            # fill in hosts
            webspace.host = self.host
            webspace.originalHost = self.originalHost

            webspaces[webspaceID] = webspace

        return webspaces

    def migrateDBData(self, con, cur, host, revert):
        uLogging.info("Binding domain parking webspaces %s to destination %s",
                      [w.webspaceID for w in self.webspaces], host)

        self.migrateDomainParkingDBData(cur, host, revert)

        # update self object in sync with db
        self.host = host
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.host = host

        uLogging.info("Finished binding webspaces to %s", host)

    def migrateIPDBData(self, cur, toHost, revert):
        fromHost = self.host
        ip = normalizeIP(fromHost.getDomainParkingIP())
        toHost.moveConfiguredIP(cur, fromHost, ip)

    def migrateDomainParkingDBData(self, cur, toHost, revert):
        cur.execute(
            "UPDATE home_directories SET host_id = %s WHERE hd_id in %s",
            toHost.hostID, tuple(self.webspaceIDs)
        )
        dest_service_id = toHost.getServiceID("domain_parking")
        cur.execute(
            "UPDATE domain_services SET service_id = %s WHERE hd_id in %s",
            prepareSelectArguments((dest_service_id, tuple(self.webspaceIDs)))
        )
        cur.execute(
            "UPDATE dp_domain_services SET service_id = %s WHERE homedir_id in %s",
            prepareSelectArguments((dest_service_id, tuple(self.webspaceIDs)))
        )

    def restoreWebspaceIPTypes(self):
        con = uSysDB.connect()
        cur = con.cursor()

        # here we update DNS records of domains migrated to another domain parking
        toHost = self.host
        uLogging.info("Scheduling DNS records updates for migrated parked domains on %s", toHost)

        cur.execute(
            "SELECT domain_name, rr_id, www_rr_id FROM dp_domain_services WHERE homedir_id in %s",
            prepareSelectArguments((tuple(self.webspaceIDs),))
        )
        rows = cur.fetchall()

        target_ip = normalizeIP(toHost.getDomainParkingIP())
        api = openapi.OpenAPI()

        def func(batch):
            request_id = api.beginRequest()
            for row in batch:
                api.pem.changeDNSRecordData(record_id=row['rr_id'], new_data=target_ip)
                api.pem.changeDNSRecordData(record_id=row['www_rr_id'], new_data=target_ip)

            api.commit()

            openapi.waitRequestComplete(
                request_id,
                "Updating DNS records for migrated parked domains on %s (Check running or unprocessed tasks in Task Manager)" % toHost,
                True
            )

        batchExecute(rows, func)


class WebspaceSet(BaseWebspaceSet):

    def __init__(self, hostID, webspaceIDs):
        BaseWebspaceSet.__init__(self, hostID, webspaceIDs)

        # check, that none of these webspaces are domain parking
        assert 0 == self.countDomainParkingWebspaces()

        self.webspaceDict = self.initWebspaces(hostID, webspaceIDs)
        self.webspaces = self.webspaceDict.values()

    def initWebspaces(self, hostID, webspaceIDs):
        if not webspaceIDs:
            return {}

        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute(
            """
                SELECT DISTINCT
                hl.homedir_id as webspace_id,
                hl.homedir_path as path,
                hl.service_id as service_id,
                hl.host_ip as ipaddr,
                s.host_id as host_id,
                ds.subscription_id as sub_id,
                hl.ip_type as apache_ip_type,
                pds.ip_type as proftpd_ip_type
                FROM
                    apache_hostlist hl
                    INNER JOIN services s on (hl.service_id = s.service_id)
                    INNER JOIN domain_services ds ON (ds.hd_id = hl.homedir_id)
                    LEFT OUTER JOIN proftpd_domain_services pds ON (ds.hd_id = pds.hd_id)
                WHERE
                    hl.homedir_id in %s ORDER BY service_id""",
            prepareSelectArguments((tuple(self.webspaceIDs),))
        )

        webspaces = {}
        for row in cur.fetchall():
            assert row["webspace_id"] not in webspaces, 'webspace %s is already in list to be migrated' % row[
                "webspace_id"]
            webspaces[row["webspace_id"]] = Webspace.create(
                row["webspace_id"], row["sub_id"], row["host_id"], row["path"], row["apache_ip_type"], row["proftpd_ip_type"])

        # save current webspace advanced properties
        cur.execute(
            "SELECT homedir_id, " + ", ".join(Webspace.PROPERTIES) +
            ", username, homedir_path FROM apache_hostlist WHERE homedir_id in %s",
            prepareSelectArguments((tuple(self.webspaceIDs),))
        )
        for rv in cur.fetchall():

            webspace = webspaces[rv["homedir_id"]]

            i = 1
            for prop in Webspace.PROPERTIES:
                setattr(webspace, prop, rv[i])
                i += 1

            webspace.username = rv[i]
            webspace.homedirPath = rv[i + 1]

        # fill in hosts
        for webspace in webspaces.values():
            webspace.host = self.host
            webspace.originalHost = self.originalHost

        return webspaces

    def migrateDBData(self, con, cur, host, revert):
        uLogging.info("Binding webspaces %s to destination %s", [w.webspaceID for w in self.webspaces], host)

        self.migrateApacheDBData(cur, host, revert)
        self.migrateAccountedDiskPaths(cur, host, revert)
        self.migrateFTPDBData(cur, host, revert)
        self.migrateLogparserDBData(con, cur, host, revert)
        self.migrateLogrotateDBData(cur, host, revert)
        self.migrateCronDBData(cur, host, revert)
        self.migrateCertDBData(cur, host, revert)

        if version[0] == 6:
            self.updateCustomerServices(cur, host)

        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            if not revert:
                cur.execute(
                    "INSERT INTO apache_h2e_ws_to_mpoint_affinity(homedir_id, mpoint_id) SELECT %s, active_mpoint_id FROM apache_h2e_clusters WHERE cluster_id = %s", webspace.webspaceID, host.hostID)
            else:
                cur.execute("DELETE FROM apache_h2e_ws_to_mpoint_affinity WHERE homedir_id = %s", webspace.webspaceID)

        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.migrateIPDBData(cur, host, revert)

        # update self object in sync with db
        self.host = host
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.host = host
            webspace.path = webspace.getDstHomeDirPath(host)

        uLogging.info("Finished binding webspaces to %s", host)

    def migrateApacheDBData(self, cur, toHost, revert):
        cur.execute(
            "UPDATE home_directories SET host_id = %s WHERE hd_id in %s",
            prepareSelectArguments((toHost.hostID, tuple(self.webspaceIDs)))
        )

        dstApacheService = toHost.getServiceID("Apache", cur)
        fromHost = self.host
        srcApacheService = fromHost.getServiceID("Apache", cur)

        cur.execute(
            "UPDATE apache_hostlist SET service_id = %s WHERE homedir_id in %s",
            prepareSelectArguments((dstApacheService, tuple(self.webspaceIDs)))
        )
        cur.execute(
            "UPDATE domain_services SET service_id = %s WHERE hd_id in %s AND service_id = %s",
            prepareSelectArguments((dstApacheService, tuple(self.webspaceIDs), srcApacheService))
        )

        # php modes
        if not revert:
            # use CGI instead of MODULE
            cur.execute(
                "UPDATE apache_hostlist SET path_to_php = '/usr/bin/php-cgi' WHERE path_to_php LIKE '%%.so' AND homedir_id in %s",
                prepareSelectArguments((tuple(self.webspaceIDs),))
            )
        else:
            # restore saved values
            for webspace in self.webspaces:
                if not webspace.isExists():
                    continue

                cur.execute(
                    "UPDATE apache_hostlist SET path_to_php = %s WHERE homedir_id = %s",
                    webspace.path_to_php, webspace.webspaceID
                )

        # need to switch shared IPs, when migration is performed by subscription
        fromHost = self.host
        target_ips = toHost.getIPs(CONFIGURED_IP_TYPE_SHARED)
        if not target_ips:
            raise MigrationError(
                "Target host must have at least one shared IP configured. Try provisioning sample subscription on shared IP."
            )

        target_ip = target_ips[0]
        shared_ips = fromHost.getIPs(CONFIGURED_IP_TYPE_SHARED)
        for ip in shared_ips:
            cur.execute(
                "UPDATE apache_hostlist SET host_ip = %s WHERE host_ip = %s AND homedir_id in %s",
                prepareSelectArguments((target_ip, ip, tuple(self.webspaceIDs)))
            )

        # fix used ips
        cur.execute("SELECT ip_id FROM configured_ips WHERE ip_address = %s", normalizeIP(target_ip))
        ip_id = cur.fetchone()['ip_id']
        cur.execute(
            "UPDATE used_ips SET ip_id = %s FROM apache_homedir_entries ahe WHERE ahe.instance_id = used_ips.obj_id AND used_ips.obj_domain = 'Apache' AND used_ips.obj_kind = 'Instance' AND ahe.homedir_id in %s",
            prepareSelectArguments((ip_id, tuple(self.webspaceIDs)))
        )

    def migrateAccountedDiskPaths(self, cur, toHost, revert):
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.migrateAccountedDiskPaths(cur, toHost, revert)

    def migrateFTPDBData(self, cur, toHost, revert):
        dstFTPService = toHost.getServiceID("Proftpd", cur)

        fromHost = self.host
        srcFTPService = fromHost.getServiceID("Proftpd", cur)
        cur.execute(
            "UPDATE domain_services SET service_id = %s WHERE hd_id in %s AND service_id = %s",
            prepareSelectArguments((dstFTPService, tuple(self.webspaceIDs), srcFTPService))
        )
        cur.execute(
            "UPDATE proftpd_domain_services SET service_id = %s WHERE hd_id in %s AND service_id = %s",
            prepareSelectArguments((dstFTPService, tuple(self.webspaceIDs), srcFTPService))
        )

        # change home directories for real ftp users
        query = """
            UPDATE proftpd_vhosts_users
            SET sys_home = '%s/' || trim(LEADING '%s' FROM sys_home)
            FROM proftpd_domain_services AS ds
            WHERE proftpd_vhosts_users.ds_id = ds.ds_id AND ds.hd_id in %%s
            """ % (toHost.getVhostsRoot(), fromHost.getVhostsRoot())

        cur.execute(query, prepareSelectArguments((tuple(self.webspaceIDs),)))

        # change home directories for lfa users
        query = """
            UPDATE proftpd_vhosts_users
            SET sys_home = '%s/' || lfa.hd_id || '/log'
            FROM proftpd_lfa_data AS lfa
            WHERE proftpd_vhosts_users.vhost_id = lfa.vhost_id
            AND proftpd_vhosts_users.login = lfa.login AND lfa.hd_id in %%s
            """ % toHost.getVhostsRoot()

        cur.execute(query, prepareSelectArguments((tuple(self.webspaceIDs),)))

        # move vhosts with exclusive IPs and mark them as h2e
        cur.execute(
            """
            UPDATE proftpd_vhosts
            SET service_id = %s, h2e = 'y'
            FROM proftpd_domain_services AS ds
            WHERE proftpd_vhosts.vhost_id = ds.vhost_id AND proftpd_vhosts.vhost_type = 0 AND ds.hd_id in %s
            """,
            prepareSelectArguments((dstFTPService, tuple(self.webspaceIDs)))
        )

        # switch vhosts with shared IPs

        # in case of by-subscription migration, we move users and domain services
        # from one vhost (on source) to another (on target)
        dstFTPMainVhost, vhostIP = toHost.getMainProftpdVhostID()
        # domain services
        cur.execute(
            """UPDATE proftpd_domain_services SET vhost_id = %s WHERE hd_id in %s""",
            prepareSelectArguments((dstFTPMainVhost, tuple(self.webspaceIDs)))
        )

        # handle name clashing of FTP users
        # rename users with clashing names on source vhost to sys_user
        cur.execute(
            """
            SELECT user_id, login, sys_user
            FROM proftpd_vhosts_users
            WHERE ds_id in (SELECT ds_id FROM proftpd_domain_services pds WHERE pds.hd_id in %s)
            AND login IN (SELECT login FROM proftpd_vhosts_users WHERE vhost_id = %s) AND vhost_id != %s
            """,
            prepareSelectArguments((tuple(self.webspaceIDs), dstFTPMainVhost, dstFTPMainVhost))
        )
        for userId, login, sysUser in cur.fetchall():
            newLogin = sysUser + login[7:]  # name like f100000.admin
            cur.execute(
                """UPDATE proftpd_vhosts_users SET login = %s WHERE user_id = %s""",
                prepareSelectArguments((newLogin, userId))
            )

        # and users
        cur.execute(
            """
            UPDATE proftpd_vhosts_users
            SET vhost_id = %s
            WHERE ds_id in (SELECT ds_id FROM proftpd_domain_services pds WHERE pds.hd_id in %s)
            """,
            prepareSelectArguments((dstFTPMainVhost, tuple(self.webspaceIDs)))
        )
        # and lfa users
        cur.execute(
            """
            UPDATE proftpd_lfa_data
            SET vhost_id = %s
            WHERE hd_id IN %s
            """,
            prepareSelectArguments((dstFTPMainVhost, tuple(self.webspaceIDs)))
        )
        cur.execute(
            """
            UPDATE proftpd_vhosts_users
            SET vhost_id = %s
            WHERE login IN (SELECT login FROM proftpd_lfa_data lfa WHERE lfa.hd_id IN %s)
            """,
            prepareSelectArguments((dstFTPMainVhost, tuple(self.webspaceIDs)))
        )

        # and fix used ips
        assert vhostIP
        cur.execute("SELECT ip_id FROM configured_ips WHERE ip_address = %s", vhostIP)
        ip_id = cur.fetchone()['ip_id']
        cur.execute(
            """
            UPDATE used_ips
            SET ip_id = %s
            FROM proftpd_domain_services pds
            WHERE used_ips.obj_id = pds.ds_id AND used_ips.obj_domain = 'ProFTPD' AND used_ips.obj_kind = 'Hosting'
            AND pds.hd_id in %s
            """,
            prepareSelectArguments((ip_id, tuple(self.webspaceIDs)))
        )

    def migrateLogparserDBData(self, con, cur, toHost, revert):
        dstLogparserService = toHost.getServiceID("logparser", cur)

        if not revert:
            logfiles = {}
            for webspace in self.webspaces:
                if not webspace.isExists():
                    continue

                if webspace.hasFTP(cur) and 1 == webspace.getProftpdVhostType():
                    service_id, dir_path, mask_type, file_mask = dstLogparserService, webspace.getDstHomeDirPath(
                        toHost) + "log", "x", "xferlog"
                    logfiles[webspace.webspaceID] = con.insertRecordWithId(
                        "logfiles", service_id=service_id, dir_path=dir_path, mask_type=mask_type, file_mask=file_mask, fmt="%h %l %u %t \"%r\" %s %b", native_format_preamble="", ascn_key="u", todo="n", todo_status="n", case_sens="y")

            for webspace in self.webspaces:
                if not webspace.isExists():
                    continue

                if webspace.hasFTP(cur) and 1 == webspace.getProftpdVhostType():
                    logfile_id = logfiles[webspace.webspaceID]
                    rtiID = webspace.getTrafficResourceInstanceID()

                    cur.execute("SELECT login FROM proftpd_vhosts_users WHERE ds_id = %s",
                                webspace.getDomainServiceID("Proftpd"))
                    rows = cur.fetchall()
                    for row in rows:
                        con.insertRecord(
                            "logfiles_associations", logfile_id=logfile_id, ascn_value=row[0], rti_id=rtiID)
                    con.insertRecord("logfiles_classifiers", logfile_id=logfile_id, fmt_spec="t")
                    con.insertRecord("logfiles_classifiers", logfile_id=logfile_id, fmt_spec="u")

            for webspace in self.webspaces:
                if not webspace.isExists():
                    continue

                if webspace.hasFTP(cur) and 1 == webspace.getProftpdVhostType():
                    logfile_id = logfiles[webspace.webspaceID]

                    con.insertRecord("logfiles_actions", logfile_id=logfile_id, action_type=1)
                    con.insertRecord("logfiles_actions", logfile_id=logfile_id, action_type=512)
                    con.insertRecord("logfiles_actions", logfile_id=logfile_id, action_type=1024)

            for webspace in self.webspaces:
                if not webspace.isExists():
                    continue

                if webspace.hasFTP(cur) and 1 == webspace.getProftpdVhostType():
                    con.insertRecord("logfiles", service_id=dstLogparserService, dir_path=webspace.getDstHomeDirPath(
                        toHost) + "log", mask_type="x", file_mask="tls.log", fmt="", native_format_preamble="", ascn_key="", todo="n", todo_status="n", case_sens="y")
        else:
            for webspace in self.webspaces:
                if not webspace.isExists():
                    continue

                # remove
                cur.execute("DELETE FROM logfiles WHERE dir_path ILIKE %s AND file_mask = %s",
                            webspace.getDstHomeDirPath(webspace.host) + "%", "xferlog")
                cur.execute("DELETE FROM logfiles WHERE dir_path ILIKE %s AND file_mask = %s",
                            webspace.getDstHomeDirPath(webspace.host) + "%", "tls.log")

        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            logfiles = ['ssl_access_log', 'xferlog', 'access_log', 'error_log', 'tls.log']
            cur.execute(
                "UPDATE logfiles SET service_id = %s, dir_path = %s WHERE file_mask in %s AND dir_path like %s",
                prepareSelectArguments((dstLogparserService, webspace.getDstHomeDirPath(toHost) + 'log', tuple(logfiles), '%/' + str(webspace.webspaceID) + "/log"))
            )

        # actions for webalizer
        fromHost = self.host
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            cur.execute(
                "UPDATE logfiles_actions SET opts = REPLACE(opts, %s, %s) WHERE action_type = %s AND opts like %s",
                fromHost.getVhostsRoot(), toHost.getVhostsRoot(), 512, '%' + str(webspace.webspaceID) + '/webspace%'
            )

        # here there were ipcount.log modifications, but were removed, because it doesn't seem to be used anywhere

    def migrateLogrotateDBData(self, cur, toHost, revert):
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.migrateLogrotateDBData(cur, toHost, revert)

    def migrateCronDBData(self, cur, toHost, revert):
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.migrateCronDBData(cur, toHost, revert)

    def migrateCertDBData(self, cur, toHost, revert):
        for webspace in self.webspaces:
            if not webspace.isExists():
                continue

            webspace.migrateCertDBData(cur, toHost, revert)

    def updateCustomerServices(self, cur, toHost):
        toHost.updateCustomerServices(cur)
        fromHost = self.host
        fromHost.updateCustomerServices(cur)

    def syncContent(self, toHost):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            beautyRetriable(ws.syncContent)(toHost)

    def migrateBackups(self, toHost):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.migrateBackups(toHost)

    def finalizeContent(self, toHost):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            beautyRetriable(ws.finalizeContent)(toHost)

    def switchToSharedIPType(self):
        api = openapi.OpenAPI()

        def func(batch):
            uLogging.info(
                "Switching IP type of webspace(s): %s", ", ".join([str(x.webspaceID) for x in batch])
            )
            requestID = api.beginRequest()
            for ws in batch:
                if not ws.isExists():
                    continue

                ws.switchToSharedIPType(api)

            api.commit()
            openapi.waitRequestComplete(
                requestID,
                "Switching IP type of webspaces on %s (Check running or unprocessed tasks in Task Manager)" % self.host,
                True
            )

        batchExecute(self.webspaces, func)

    def restoreWebspaceIPTypes(self):
        api = openapi.OpenAPI()

        def func(batch):
            uLogging.info(
                "Restore IP type of webspace(s): %s", ", ".join([str(x.webspaceID) for x in batch])
            )

            requestID = api.beginRequest()
            for ws in batch:
                if not ws.isExists():
                    continue

                ws.restoreWebspaceIPTypes(api)

            api.commit()
            openapi.waitRequestComplete(
                requestID,
                "Switching IP type of webspaces on %s (Check running or unprocessed tasks in Task Manager)" % self.host,
                True
            )

        def beautyRetriableFunc(batch):
            beautyRetriable(func)(batch)

        batchExecute(self.webspaces, beautyRetriableFunc)

    def restoreWebspaceProperties(self, cur):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.restoreWebspaceProperties(cur)

    def dumpCrontabs(self):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.dumpCrontabs()

    def checkContentForAbsPaths(self, toHost):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.checkContentForAbsPaths(toHost)

    def createLocalPear(self):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.createLocalPear()

    def syncHtAccess(self):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.syncHtAccess()


class Webspace:
    PROPERTIES = ["cgi_state", "ssi_state", "ssl_state", "path_to_php", "erd_state", "webstat_state", "wap_state"]
    PHP_PROPERTIES = ["path_to_php"]

    def create(cls, webspaceID, sub_id, host_id, path, apache_ip_type, proftpd_ip_type):
        result = cls(webspaceID)
        result.subscriptionID = sub_id
        result.hostID = host_id
        result.path = path
        if None != apache_ip_type:
            result.apache_ip_type = int(apache_ip_type)
        else:
            result.apache_ip_type = apache_ip_type

        if None != proftpd_ip_type:
            result.proftpd_ip_type = int(proftpd_ip_type)
        else:
            result.proftpd_ip_type = proftpd_ip_type

        return result
    create = classmethod(create)

    def __init__(self, webspaceID):
        self.webspaceID = webspaceID
        self.id = self.webspaceID

    def __str__(self):
        return "Webspace(%s){host: %s, path: %s}" % (self.webspaceID, self.host, self.path)

    def __repr__(self):
        return "Webspace(%s)" % self.webspaceID

    def isExists(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT count(*) FROM domain_services WHERE hd_id = %s", self.webspaceID)
        isWebsiteExists = cur.fetchone()[0] > 0
        if not isWebsiteExists:
            uLogging.info("Webspace %s doesn't exists.", self.webspaceID)

        return isWebsiteExists

    def getDstHomeDirPath(self, host):
        assert self.webspaceID > 0
        result = "%s/%d/" % (host.getVhostsRoot(), self.webspaceID)
        return result

    def getDomainServiceID(self, serviceName):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT ds_id FROM domain_services ds JOIN services s ON (ds.service_id = s.service_id) WHERE ds.hd_id = %s AND s.name ilike %s",
                    self.webspaceID, serviceName + '%')
        rv = cur.fetchone()
        if not rv:
            result = None
        else:
            result = rv[0]

        cur.close()

        return result

    def getUserName(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT username FROM apache_hostlist WHERE homedir_id = %s", self.webspaceID)
        rv = cur.fetchone()
        if not rv or not rv[0]:
            raise MigrationError("Failed to determine username for webspace %s" % self.webspaceID)

        result = rv[0]

        cur.close()

        return result

    def getProftpdVhostType(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute(
            "SELECT pv.vhost_type FROM proftpd_domain_services pds JOIN proftpd_vhosts pv ON pv.vhost_id = pds.vhost_id WHERE pds.hd_id = %s", self.webspaceID)
        rv = cur.fetchone()
        if not rv:
            raise MigrationError("Failed to get Proftpd vhost for webspace %s" % self.webspaceID)

        result = rv[0]

        cur.close()

        return result

    def setMigratingState(self, migrating):
        api = openapi.OpenAPI()
        api.pem.web_cluster.setMigratingState(webspaceID=self.webspaceID, set=migrating)

        try:
            if migrating:
                api.pem.web_cluster.disableFTP(webspaceID=self.webspaceID)
            else:
                api.pem.web_cluster.enableFTP(webspaceID=self.webspaceID)
        except openapi.OpenAPIError, e:
            # ignore FTP not enabled errors
            if -1 == e.error_message.find("service #0 does not exist"):
                raise e

        ssh = SSHAccess()
        if migrating:
            ssh.disable(self)
        else:
            ssh.enable(self)

    def checkContentForAbsPaths(self, toHost):
        uLogging.info("Checking content of webspace %s for absolute paths", self.webspaceID)

        fromHost = self.originalHost
        toPath = self.getDstHomeDirPath(toHost).rstrip("/")
        fromPath = self.getDstHomeDirPath(fromHost).rstrip("/")

        checkContentForAbsPaths(toPath, toHost, fromPath, toPath)

    def finalizeContent(self, toHost):
        """Adapt content of webspace for WebCluster 
                Remove unnecessary directories, that we moved during migration
                Or create new ones
        """
        uLogging.info("Fixing webspace %s configuration for NG", self.webspaceID)

        fromHost = self.originalHost

        toPath = self.getDstHomeDirPath(toHost).rstrip('/')
        fromPath = self.getDstHomeDirPath(fromHost).rstrip('/')

        assert toPath.startswith("/var/www/vhosts")
        assert fromPath.startswith("/usr/local/pem")
        assert str(self.webspaceID) in toPath

        # adjust permissions for webspace
        adjustPermissions(
            toHost, fromPath, toPath, self.webspaceID,
            pemsrv=toHost.groups.get('pemsrv') or '',
            pemcln=toHost.groups.get('pemcln') or ''
        )

    def syncContent(self, toHost):
        fromHost = self.originalHost

        uLogging.info("Syncing content of webspace #%d to %s", self.webspaceID, toHost)

        fromPath = self.getDstHomeDirPath(fromHost).rstrip('/')
        toPath = toHost.getVhostsRoot()
        try:
            fromHost.cmd("/usr/bin/rsync", args=["rsync", "-e", "ssh  -o StrictHostKeyChecking=no -o CheckHostIP=no -o UserKnownHostsFile=/dev/null",
                                            "-az", fromPath, "root@%s:%s" % (toHost.getBacknetIP(), toPath)])
            uLogging.info("Content synced.")
        except:
            if config.config.skipwarnings:
                uLogging.info("Unable to sync content of webspace #%d to %s", self.webspaceID, toHost)
            else:
                raise

    def getTrafficResourceInstanceID(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute(
            """
            SELECT rti.rt_instance_id
            FROM subs_resources as rti
                INNER JOIN resource_types AS rt ON rti.rt_id = rt.rt_id
                INNER JOIN resource_classes AS rc ON rt.class_id = rc.class_id
            WHERE
                rti.sub_id = %s AND rc.name = 'traffic'
            """,
            self.subscriptionID
        )
        return cur.fetchone()[0]

    def migrateAccountedDiskPaths(self, cur, toHost, revert):
        if not revert:
            cluster, host = toHost, self.host
        else:
            cluster, host = self.host, toHost

        hostPath = os.path.join(host.getVhostsRoot(), self.homedirPath, "webspace")

        mpointID, nodeID, rootPath = cluster.getMountPoint(cur, cluster.activeMountPointID)
        nodeID, clusterID, ip, nodeType = cluster.getNode(cur, nodeID)

        if HOST_TYPE_WEB_SERVER == cluster.getType():
            clusterPath = "%s@%s" % (self.username, rootPath)
        else:
            clusterPath = "%s@%s:%s" % (self.username, ip, rootPath)

        if not revert:
            oldPath, hostID, pathType, newPath = hostPath, cluster.director.hostID, 'q', clusterPath
        else:
            oldPath, hostID, pathType, newPath = clusterPath, host.hostID, 'd', hostPath

        cur.execute(
            "UPDATE disk_space_paths SET path = %s, host_id = %s, path_type = %s WHERE path = %s",
            newPath, hostID, pathType, oldPath
        )

    def migrateLogrotateDBData(self, cur, host, revert):
        dstLogrotateService = host.getServiceID("logrotate", cur)
        fromHost = self.host
        logfileMask = "%s/%s%%" % (fromHost.getVhostsRoot(), self.path)
        cur.execute(
            "UPDATE lrs SET service_id = %s, logname = REPLACE(logname, %s, %s) WHERE logname ILIKE %s",
            dstLogrotateService, self.getDstHomeDirPath(fromHost), self.getDstHomeDirPath(host), logfileMask
        )

    def migrateCronDBData(self, cur, host, revert):
        try:
            dstService = host.getServiceID("crontabs", cur)
            srcService = self.host.getServiceID("crontabs", cur)
            cur.execute(
                "UPDATE domain_services SET service_id = %s WHERE hd_id = %s AND service_id = %s",
                dstService, self.webspaceID, srcService
            )
        except MigrationError, e:
            uLogging.info("Cannot migrate crontabs db data: %s", e)

    def getApacheCertificates(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT cert_filename FROM apache_certificates WHERE homedir_id = %s", self.webspaceID)
        return cur.fetchall()

    def migrateCertDBData(self, cur, host, revert):
        try:
            for certificate in self.getApacheCertificates():
                srcCertFileName = certificate['cert_filename']
                dstCertFileName = srcCertFileName
                vhostRoot = self.host.getVhostsRoot()
                if srcCertFileName.startswith(vhostRoot):
                    dstCertFileName = host.getVhostsRoot() + srcCertFileName[len(vhostRoot):]

                cur.execute(
                    "UPDATE apache_certificates SET cert_filename = %s WHERE homedir_id = %s AND cert_filename = %s",
                    dstCertFileName, self.webspaceID, srcCertFileName
                )
        except MigrationError, e:
            uLogging.info("Cannot migrate certificates db data: %s", e)

    def migrateIPDBData(self, cur, toHost, revert):
        fromHost = self.host
        # move IPs related to Apache
        cur.execute("SELECT host_ip FROM apache_hostlist WHERE homedir_id = %s", self.webspaceID)
        rv = cur.fetchone()
        if rv:
            ip = normalizeIP(rv[0])
            toHost.moveConfiguredIP(cur, fromHost, ip)

        # move IPs related to ProFTPD
        cur.execute(
            "SELECT pi.ip FROM proftpd_domain_services pds JOIN proftpd_vhosts pvh ON (pds.vhost_id = pvh.vhost_id) JOIN proftpd_ips pi ON pvh.vhost_id = pi.vhost_id WHERE pds.hd_id = %s",
            self.webspaceID
        )
        rv = cur.fetchone()
        if rv:
            ip = rv[0]
            toHost.moveConfiguredIP(cur, fromHost, ip)

    def restoreWebspaceProperties(self, cur):
        assignments = ["%s = %s" % (prop, getattr(self, prop))
                       for prop in self.PROPERTIES if prop not in self.PHP_PROPERTIES]
        cur.execute("UPDATE apache_hostlist SET " + ", ".join(assignments) + " WHERE homedir_id = %s", self.webspaceID)

        # On Legacy Apache hosting these directories are always separated.
        cur.execute("UPDATE apache_hostlist SET separate_http_dirs = 1 WHERE homedir_id = %s", self.webspaceID)

    def switchToSharedIPType(self, api):
        # switch to shared ip type only webspaces, that are currently not of shared ip type
        fine_ip_types = [IP_TYPE_SHARED, None]

        if self.apache_ip_type not in fine_ip_types:
            apache_ip_type = IP_TYPE_SHARED
        else:
            apache_ip_type = None

        if self.proftpd_ip_type not in fine_ip_types:
            proftpd_ip_type = IP_TYPE_SHARED
        else:
            proftpd_ip_type = None

        self.__switchIPType(api, apache_ip_type, proftpd_ip_type)

    def restoreWebspaceIPTypes(self, api):
        apache_ip_type = (self.apache_ip_type, IP_TYPE_RENEW)[IP_TYPE_SHARED == self.apache_ip_type]
        if None != self.proftpd_ip_type:
            proftpd_ip_type = (self.proftpd_ip_type, IP_TYPE_RENEW)[IP_TYPE_SHARED == self.proftpd_ip_type]
        else:
            proftpd_ip_type = None

        self.__switchIPType(api, apache_ip_type, proftpd_ip_type)
        self.changeManualDNSRecords(api)

    def changeManualDNSRecords(self, api):
        con = uSysDB.connect()
        cur = con.cursor()

        # retrieve current ip
        cur.execute("SELECT host_ip FROM apache_hostlist WHERE homedir_id = %s", self.webspaceID)
        ip_normalized = normalizeIP(cur.fetchone()[0])

        cur.execute(
            "SELECT ip_address FROM configured_ips WHERE sub_id = %s OR (sub_id = 1 AND host_id = %s)",
            (self.subscriptionID, self.hostID)
        )
        original_host_ips_normalized = [x[0] for x in cur.fetchall()]

        # update for all the related objects
        cur.execute(
            """
            SELECT DISTINCT dns_resource_records.rr_id, dns_resource_records.rr_type, dns_resource_records.data
            FROM dns_record_references
            JOIN dns_resource_records ON dns_record_references.rr_id = dns_resource_records.rr_id
            JOIN domains ON domains.domain_id = dns_resource_records.domain_id
            JOIN apache_homedir_entries ON apache_homedir_entries.domain_name = domains.name
            JOIN apache_resources ON apache_homedir_entries.homedir_id = apache_resources.homedir_id
            JOIN domain_services ON domain_services.subscription_id = apache_resources.sub_id
            JOIN services ON services.service_id = domain_services.service_id
            WHERE services.name = 'ApacheNG' AND dns_record_references.obj_domain = 'manual'
            AND dns_record_references.obj_kind = 'manual' AND apache_homedir_entries.homedir_id = %s
            """,
            self.webspaceID
        )
        rows = cur.fetchall()
        for row in rows:
            if row['rr_type'] != 'A':
                continue

            ip = row['data']
            if isIpv4(ip):
                ip = normalizeIP(ip)

            if ip in original_host_ips_normalized:
                api.pem.changeDNSRecordData(record_id=row['rr_id'], new_data=ip_normalized)

    def __switchIPType(self, api, apache_ip_type, proftpd_ip_type):
        if None != apache_ip_type:
            if IP_TYPE_RENEW == apache_ip_type:
                self.updateApacheDNSRecords(api)
            else:
                api.pem.apache.switchIPType(webspace_id=self.webspaceID, ip_type=apache_ip_type)
                self.updateApacheDNSRecords(api, apache_records=False)

        if None != proftpd_ip_type:
            if IP_TYPE_RENEW == proftpd_ip_type:
                self.updateFTPDNSRecords(api)
            else:
                api.pem.ProFTPD.switchIPType(webspace_id=self.webspaceID, ip_type=proftpd_ip_type)

    def updateApacheDNSRecords(self, api, apache_records=True, ftp_records=True):
        con = uSysDB.connect()
        cur = con.cursor()

        # retrieve current ip
        cur.execute("SELECT host_ip FROM apache_hostlist WHERE homedir_id = %s", self.webspaceID)
        ip = normalizeIP(cur.fetchone()[0])

        # update for all the related objects
        cur.execute("SELECT instance_id FROM apache_homedir_entries WHERE homedir_id = %s", self.webspaceID)
        rows = cur.fetchall()
        for row in rows:
            if apache_records:
                self.updateDNSRecords(api, 'Apache', 'Instance', row['instance_id'], ip)

            # prefixed for FTP A record is referenced using apache instance id
            if ftp_records:
                self.updateDNSRecords(api, 'ProFTPD', 'FTPSite', row['instance_id'], ip)

    def updateFTPDNSRecords(self, api):
        con = uSysDB.connect()
        cur = con.cursor()

        # retrieve current ip
        cur.execute(
            """SELECT ip
            FROM proftpd_ips pi
            JOIN proftpd_domain_services pds ON pi.vhost_id = pds.vhost_id
            WHERE pds.hd_id = %s""",
            self.webspaceID
        )

        ip = None
        ips = cur.fetchall()
        for (_ip, ) in ips:
            if isIpv4(_ip):
                ip = _ip
                break

        # update for all the related objects
        cur.execute("SELECT ds_id FROM proftpd_domain_services WHERE hd_id = %s", self.webspaceID)
        rows = cur.fetchall()
        for row in rows:
            self.updateDNSRecords(api, 'ProFTPD', 'Hosting', row['ds_id'], ip)

    def updateDNSRecords(self, api, obj_domain, obj_kind, obj_id, new_ip_normalized):
        assert obj_domain in ['Apache', 'ProFTPD', 'Parking']
        assert obj_kind in ['Instance', 'Hosting', 'FTPSite']

        con = uSysDB.connect()
        cur = con.cursor()

        obj_domain2 = None
        rr_id_to_change = None
        if obj_domain == 'ProFTPD':
            obj_domain2 = 'Apache'
        elif obj_domain == 'Apache':
            obj_domain2 = 'ProFTPD'

        if obj_domain2 is not None:
            cur.execute(
                """
                SELECT rr_id
                FROM dns_record_references drr1
                WHERE obj_domain = %s AND obj_kind = %s AND obj_id = %s
                AND NOT EXISTS (
                    SELECT rr_id FROM dns_record_references drr2 WHERE drr1.rr_id=drr2.rr_id AND drr2.obj_domain = %s
                )""",
                obj_domain, obj_kind, obj_id, obj_domain2
            )
            rr_id_to_change = [x['rr_id'] for x in cur.fetchall()]

        cur.execute(
            "SELECT rr_id FROM dns_record_references WHERE obj_domain = %s AND obj_kind = %s AND obj_id = %s",
            obj_domain, obj_kind, obj_id
        )
        rows = cur.fetchall()

        for row in rows:
            if rr_id_to_change is not None and row['rr_id'] not in rr_id_to_change:
                continue

            try:
                api.pem.changeDNSRecordData(record_id=row['rr_id'], new_data=new_ip_normalized)
            except openapi.OpenAPIError, e:
                error_str = str(e)
                uLogging.debug(error_str)
                if "Resource for record" not in error_str and "APS resource for" not in error_str:
                    raise e

    def hasFTP(self, cur):
        cur.execute("SELECT count(*) FROM proftpd_domain_services WHERE hd_id = %s", self.webspaceID)
        return cur.fetchone()[0] > 0

    def dumpCrontabs(self):
        domainServiceID = self.getDomainServiceID("crontabs")
        if not domainServiceID:
            uLogging.debug("Crontabs are not enabled for webspace %s, skipping dumping of new configs", self.webspaceID)
            return

        pobjref = uTasks.PObjRef(uTasks.SCRef("crontabs"), "getDomainService", domainServiceID)

        task = uTasks.Task(pobjref, "Dump crontabs for webspace %s" % self.webspaceID, "taskDumpFile")
        task.setParam("webspace_id", self.webspaceID)

        task.schedule()

        beautyRetriable(waitTasksCompleteRef)(pobjref)

    def createLocalPear(self):
        uLogging.info("Creating local PEAR installation for webspace %s", self.webspaceID)

        domainServiceID = self.getDomainServiceID("Apache")
        if not domainServiceID:
            raise MigrationError("Failed to determine domain service ID for webspace %s", self.webspaceID)

        pobjref = uTasks.PObjRef(uTasks.SCRef("apache"), "getDomainService", domainServiceID)

        task = uTasks.Task(pobjref, "Creating local pear installation for domain service with ID %s" %
                           domainServiceID, "taskCreateLocalPear")
        task.setParam("username", self.getUserName())

        task.schedule()

        beautyRetriable(waitTasksCompleteRef)(pobjref)

        uLogging.info("Local PEAR installation successfully created for webspace %s", self.webspaceID)

    def syncHtAccess(self):
        uLogging.info("Syncing .htaccess files for webspace %s", self.webspaceID)
        domainServiceID = self.getDomainServiceID("Apache")
        if not domainServiceID:
            raise MigrationError("Failed to determine domain service ID for webspace %s", self.webspaceID)

        pobjref = uTasks.PObjRef(uTasks.SCRef("apache"), "getDomainService", domainServiceID)

        task = uTasks.Task(pobjref, "Task that synchronize .htaccess for domain service with ID %s" %
                           domainServiceID, "taskSyncHtAccess")
        task.schedule()

        beautyRetriable(waitTasksCompleteRef)(pobjref)

    def _getBackups(self):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute(
            """
            SELECT backup_id, store_dir, host_id
            FROM backups_groups bg
                JOIN backups_backups bb ON bg.group_id=bb.group_id
                JOIN domain_services ds on ds.ds_id = bb.object_id
            WHERE ds.hd_id = %s
            """, self.webspaceID
        )

        res = []
        for backup_id, store_dir, host_id in cur.fetchall():
            host = HostFactory.createHost(host_id)
            if not host.is_win:
                res.append(Backup(backup_id, store_dir, host_id, self.webspaceID))
            else:
                uLogging.warn(
                    "The backup service node {} is running Windows. The migration of backups from Windows-based servers is not supported. The backup archive #{} will not be transferred.".
                        format(host, backup_id)
                )

        return res

    def migrateBackups(self, toHost):
        fromHost = self.originalHost
        toPath = self.getDstHomeDirPath(toHost).rstrip("/")
        fromPath = self.getDstHomeDirPath(fromHost).rstrip("/")

        for b in self._getBackups():
            if b.check():
                b.unpack()
                if not config.config.skipreplaceapspath:
                    b.checkContentForAbsPaths(fromPath, toPath)
                b.fixWebspacePermissions(fromPath, toHost)
                b.saveOriginal()
                b.pack(fromPath, toPath)


class SubscriptionSet:

    def createFromListOfSubscriptionIDs(cls, subIDs):
        con = uSysDB.connect()
        cur = con.cursor()

        # check, that all subscriptions exist
        cur.execute(
            "SELECT DISTINCT sub_id FROM subscriptions WHERE sub_id in %s",
            prepareSelectArguments((tuple(subIDs),))
        )
        rows = cur.fetchall()
        if len(rows) != len(subIDs):
            foundSubIDs = [row['sub_id'] for row in rows]
            raise MigrationError(
                "Some of specified subscriptions were not found. Check if they were specified correctly: %s" % (
                    ", ".join([str(subID) for subID in subIDs if not subID in foundSubIDs])
                )
            )

        # Retrieve information about hosts, subscriptions, and webspaces, except for the parked domains that were
        # already migrated to the target host. There is a chance that a customer first migrates only the parked domains,
        # and then starts migrating other subscriptions.
        cur.execute(
            """
            SELECT DISTINCT s.sub_id, hd.hd_id, hd.host_id, dds.ds_id
            FROM subscriptions AS s
            LEFT JOIN domain_services AS ds ON s.sub_id = ds.subscription_id
            LEFT JOIN home_directories AS hd ON ds.hd_id = hd.hd_id
            LEFT JOIN dp_domain_services AS dds ON ds.ds_id = dds.ds_id
            WHERE
                s.sub_id in %s AND (
                    dds.ds_id is NULL OR 'domain_parking' IN (SELECT name FROM services where host_id = hd.host_id)
                )

            """,
            prepareSelectArguments((tuple(subIDs),))
        )

        rows = cur.fetchall()
        return cls(rows)
    createFromListOfSubscriptionIDs = classmethod(createFromListOfSubscriptionIDs)

    def __init__(self, subWebspaces):
        self.subIDs = [pair[0] for pair in subWebspaces]
        self.webspaceIDs = [pair[1] for pair in subWebspaces if pair[1]]

        subList = [(subID, Subscription(subID)) for subID in self.subIDs]
        self.subDict = dict(subList)
        self.subs = self.subDict.values()

        # fill in service template/plan IDs
        if config.config.withpba:
            for sub in self.subDict.values():
                sub.initialServicePlanID, sub.initialSubscriptionPeriodID = sub.getServicePlanInfo()
        else:
            if self.subIDs:
                con = uSysDB.connect()
                cur = con.cursor()

                cur.execute(
                    "SELECT st_id, sub_id FROM subscriptions WHERE sub_id in %s",
                    prepareSelectArguments((tuple(self.subIDs),))
                )
                for row in cur.fetchall():
                    self.subDict[row['sub_id']].initialServiceTemplateID = row['st_id']

        # init host->webspaces dictionary
        self.webspaces = {}

        # construct intermediat host->list of webspace ids dict
        webspaceHostDict = {}
        for [subID, webspaceID, hostID, dpdsID] in subWebspaces:
            if hostID:
                assert webspaceID

                key = (hostID, dpdsID > 0)

                if key in webspaceHostDict:
                    webspaceHostDict[key] += [webspaceID]
                else:
                    webspaceHostDict[key] = [webspaceID]

        # convert it to proper one
        for key, webspaceIDs in webspaceHostDict.iteritems():
            if key[1]:
                self.webspaces[key] = DomainParkingWebspaceSet(key[0], webspaceIDs)
            else:
                self.webspaces[key] = WebspaceSet(key[0], webspaceIDs)

        # fill in subscription -> webspace references
        for [subID, webspaceID, hostID, dpdsID] in subWebspaces:
            if hostID:
                assert subID in self.subDict

                key = (hostID, dpdsID > 0)

                assert key in self.webspaces
                assert webspaceID in self.webspaces[key].webspaceDict

                self.subDict[subID].webspaces.append(self.webspaces[key].webspaceDict[webspaceID])

    def migrateDBData(self, con, cur, toHost, revert):
        if not self.webspaceIDs or not self.webspaces:
            raise MigrationError("Nothing to migrate: no webspaces were found for specified host or subscriptions")

        if not revert:
            for webspaceset in self.webspaces.values():
                webspaceset.migrateDBData(con, cur, toHost, revert)
        else:
            for webspaceset in self.webspaces.values():
                webspaceset.migrateDBData(con, cur, webspaceset.originalHost, revert)

    def cleanUpDB(self, con, cur, toHost):
        # nothing to clean up
        pass

    def syncContent(self, toHost):
        uLogging.info("Copying content of subscriptions to destination %s", toHost)
        for webspaceset in self.webspaces.values():
            webspaceset.syncContent(toHost)

    def migrateBackups(self, toHost):
        uLogging.info("Repacking backups to be suitable for destination %s", toHost)
        for webspaceset in self.webspaces.values():
            webspaceset.migrateBackups(toHost)

    def finalizeContent(self, toHost):
        for webspaceset in self.webspaces.values():
            webspaceset.finalizeContent(toHost)

    def switchToSharedIPType(self):
        uLogging.info("Switching IP types to 'shared IP'")
        for webspaceset in self.webspaces.values():
            webspaceset.switchToSharedIPType()

    def restoreWebspaceIPTypes(self):
        uLogging.info("Updating IPs of migrated sites on target host")
        for webspaceset in self.webspaces.values():
            webspaceset.restoreWebspaceIPTypes()

    def restoreWebspaceProperties(self, cur):
        for webspaceset in self.webspaces.values():
            webspaceset.restoreWebspaceProperties(cur)

    def dumpCrontabs(self):
        uLogging.info("Dumping crontabs")
        for webspaceset in self.webspaces.values():
            webspaceset.dumpCrontabs()

    def updateLogParser(self):
        for host in self.getHosts():
            host.updateLogParser()

    def checkContentForAbsPaths(self, toHost):
        for webspaceset in self.webspaces.values():
            webspaceset.checkContentForAbsPaths(toHost)

    def createLocalPear(self):
        for webspaceset in self.webspaces.values():
            webspaceset.createLocalPear()

    def syncHtAccess(self):
        for webspaceset in self.webspaces.values():
            webspaceset.syncHtAccess()

    def hasSubscriptionsWithDomainParking(self):
        for webspaceset in self.webspaces.values():
            if isinstance(webspaceset, DomainParkingWebspaceSet):
                return True

        return False

    def hasProviderSubscription(self):
        return 1 in self.subIDs

    def getSubs(self):
        return self.subs

    def getHosts(self):
        hosts = sets.Set()
        for webspaceset in self.webspaces.values():
            hosts.add(webspaceset.host)

        return hosts

    def getFromHosts(self):
        fromHosts = sets.Set()
        for webspaceset in self.webspaces.values():
            fromHosts.add(webspaceset.originalHost)

        return fromHosts


class Backup:
    BACKUP_PATH = "{storedir}/backup_data-{id}.tar"
    PACK_CMD = "tar --transform 's,^\.,{toPath}/webspace,' -cf {file} -C {tmpdir}/{fromPath}/webspace ."
    UNPACK_CMD = "tar xf {file} -C {tmpdir}"
    MKTEMP_CMD = "mktemp -d --tmpdir={storedir}"
    FILE_EXISTS_CMD = "test -e {file}"

    def __init__(self, id, storeDir, hostID, hdID):
        self.id = id
        self.storeDir = storeDir
        self.host = HostFactory.createHost(hostID)
        self.hdID = hdID

        output = self.host.cmd(self.MKTEMP_CMD.format(storedir=storeDir))
        self.tmpdir = output['stdout'].strip()

    def _getPath(self):
        return self.BACKUP_PATH.format(storedir=self.storeDir, id=self.id)

    def check(self):
        uLogging.info('Backup.check %s exists' % self._getPath())
        try:
            self.host.cmd(self.FILE_EXISTS_CMD.format(file=self._getPath()))
            return True
        except:
            uLogging.warn('File %s does not exist' % self._getPath())
            return False

    def pack(self, fromPath, toPath):
        uLogging.info('Backup.pack %s to %s' % (self._getPath(), self.tmpdir))
        self.host.cmd(self.PACK_CMD.format(fromPath=fromPath, toPath=toPath, file=self._getPath(), tmpdir=self.tmpdir))

    def unpack(self):
        uLogging.info('Backup.unpack %s to %s' % (self._getPath(), self.tmpdir))
        self.host.cmd(self.UNPACK_CMD.format(file=self._getPath(), tmpdir=self.tmpdir))

    def saveOriginal(self):
        uLogging.info('Backup.saveOriginal %s' % self._getPath())
        self.host.cmd('mv {file} {file}.{unixtime}.bak'.format(file=self._getPath(), unixtime=int(time.time())))

    def checkContentForAbsPaths(self, fromPath, toPath):
        checkContentForAbsPaths(self.tmpdir + fromPath, self.host, fromPath, toPath)

    def fixWebspacePermissions(self, fromPath, toHost):
        adjustPermissions(
            self.host, fromPath, self.tmpdir + fromPath, self.hdID,
            pemsrv=toHost.groups.get('pemsrv') or '',
            pemcln=toHost.groups.get('pemcln') or ''
        )

    def __del__(self):
        uLogging.info('Backup.__del__ %s' % self.tmpdir)
        self.host.cmd('rm -rf {tmpdir}'.format(tmpdir=self.tmpdir))


class Subscription:

    def __init__(self, subID, fromHost=None):
        if type(subID) == str:
            subID = int(subID)

        self.subID = subID
        self.id = self.subID
        self.servicePlanID = None
        self.webspaces = []

    def __str__(self):
        return "subscription %s" % (self.subID)

    def __repr__(self):
        return "Subscription(%s)" % (self.subID)

    def setMigratingState(self, migrating):
        for ws in self.webspaces:
            if not ws.isExists():
                continue

            ws.setMigratingState(migrating)

    def changeServiceTemplate(self, stID, api):
        uLogging.info("Upgrading subscription %d to service template %d", self.subID, stID)

        try:
            rts = api.pem.upgradeSubscription(subscription_id=self.subID, service_template_id=stID)

            uLogging.debug("Added resource types: %s", rts['rt_added'])
            uLogging.debug("Removed resource types: %s", rts['rt_removed'])

            uLogging.debug("Mapped resource types:")
            for rt_map in rts['rt_mapped']:
                uLogging.debug("%d->%d", rt_map['src_rt_id'], rt_map['dst_rt_id'])
        except openapi.OpenAPIError, e:
            error_str = str(e)
            if "is locked" not in error_str:
                raise

    def changeServicePlan(self, servicePlanID, planPeriodID):
        uLogging.info("Upgrading subscription %d to service plan %d", self.subID, servicePlanID)

        api = pbaapi.PBAAPI()
        oldUsername = pbaapi._settings.username
        sourceServicePlanID = int(self.getServicePlanID())
        setResellerCredentialsByServicePlan(sourceServicePlanID, api)
        if not planPeriodID:
            # just choose first one
            periods = api.PlanPeriodListGet_API(servicePlanID, 1)
            planPeriodID = periods[0].PlanPeriodID

        result = api.PlacePlanPeriodSwitchOrder_API(self.subID, servicePlanID, planPeriodID, 10)
        pbaapi._settings.username = oldUsername
        return result

    def waitOrderCompleted(self, orderID):
        uLogging.info("Waiting for the order for plans switching with ID %s to complete" % orderID)

        api = pbaapi.PBAAPI()
        attempt = 0
        while attempt < DEFAULT_MAX_ATTEMPTS:
            result = api.GetOrder_API(orderID)
            orderStatusId = result[4]
            if orderStatusId == 'CP':
                return
            elif orderStatusId == 'PF':
                raise MigrationError("The order for plans switching with ID %s was not completed." % orderID)

            attempt += 1
            time.sleep(DEFAULT_INTERVAL_BETWEEN_ATTEMPTS)

        raise MigrationError("The order for plans switching with ID %s was not completed due to a time-out." % orderID)

    def isSwitchPlanAllowed(self):
        api = pbaapi.PBAAPI()
        return api.IsSwitchPlanAllowed(self.subID)[0]

    def getAllowedPlans(self):
        api = pbaapi.PBAAPI()
        oldUsername = pbaapi._settings.username
        servicePlanID = int(self.getServicePlanID())
        setResellerCredentialsByServicePlan(servicePlanID, api)
        allowedPlans = api.AllowedPlans(servicePlanID, 0)
        pbaapi._settings.username = oldUsername
        return allowedPlans

    def updateLimits(self, api):
        uLogging.info("Updating limits of subscription %d", self.subID)

        try:
            api.pem.upgradeSubscription(subscription_id=self.subID, service_template_id=self.getServiceTemplateID())
        except openapi.OpenAPIError, e:
            error_str = str(e)

            if "is locked" in error_str:
                return
            if "domain service is disabled" in error_str:
                return

            raise

    def unprovideUnsupportedServices(self):
        uLogging.info("Unproviding unsupported services from subscription %s", self.subID)

        # unprovide FrontPage
        api = openapi.OpenAPI()

        if self.getResourceUsage("apache_fp_support") > 0:
            for rtID in self.getResourceTypeIDs("apache_fp_support"):
                api.pem.setResourceTypeLimit(
                    subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)

        if self.getResourceUsage("awstats") > 0:
            for rtID in self.getResourceTypeIDs("awstats"):
                api.pem.setResourceTypeLimit(
                    subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)

        if self.getResourceUsage("urchin.website") > 0:
            for rtID in self.getResourceTypeIDs("urchin.website"):
                api.pem.setResourceTypeLimit(
                    subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)

        if self.getResourceUsage("urchin.domainServices") > 0:
            for rtID in self.getResourceTypeIDs("urchin.domainServices"):
                api.pem.setResourceTypeLimit(
                    subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)

        if self.getResourceUsage("shared_hosting_php") > 0:
            for rtID in self.getResourceTypeIDs("shared_hosting_php"):
                act_params = self.getRTActParams(rtID)
                php_binary = act_params.get("path.to.php.binary")

                if php_binary == PHP_MODULE_HANDLER:
                    api.pem.setResourceTypeLimit(
                        subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)

        waitForFTP = False
        if self.getResourceUsage("proftpd.res.anon_access") > 0:
            for rtID in self.getResourceTypeIDs("proftpd.res.anon_access"):
                api.pem.setResourceTypeLimit(
                    subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)
                waitForFTP = True

        waitForWebApps = False
        webapp_resources = self.getResourceUsageForWebApps("webapps_", "apache_physical_hosting")
        for usage, rtID, rcName in webapp_resources:
            if usage > 0:
                uLogging.info(
                    "Unproviding legacy WebApp %s from subscription %s (resource type id %s)", rcName, self.subID, rtID)
                api.pem.setResourceTypeLimit(
                    subscription_id=self.subID, resource_type_id=rtID, limit=0, autoprovide=True)
                waitForWebApps = True

        if waitForFTP:
            waitProFTPDDumpConfig()

        if waitForWebApps:
            waitWebAppDeletion()

    def getRTActParams(self, rt_id):
        con = uSysDB.connect()
        cur = con.cursor()
        cur.execute("SELECT name, value FROM rt_act_params WHERE rt_id = %s", rt_id)
        params = {}
        for name, value in cur.fetchall():
            params[name] = value

        return params

    def getServiceTemplateID(self):
        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute("SELECT st_id FROM subscriptions WHERE sub_id = %s", self.subID)
        rv = cur.fetchone()

        cur.close()

        if not rv:
            raise MigrationError("Failed to determine service template ID for subscription %s" % (self.subID))
        else:
            return rv[0]

    def getServicePlanID(self):
        api = pbaapi.PBAAPI()
        details = api.SubscriptionDetailsGet_API(self.subID)
        return details.PlanID

    def getServicePlanInfo(self):
        api = pbaapi.PBAAPI()
        details = api.SubscriptionDetailsGetEx_API(self.subID)
        planId = int(details[3])
        period = int(details[16])
        periodType = int(details[17])
        return (planId, self.getPeriodID(planId, period, periodType))

    def getBillingSubscriptionInfo(self):
        api = pbaapi.PBAAPI()
        return api.SubscriptionDetailsGetEx_API(self.subID)

    def getApsApplications(self):
        con = uSysDB.connect()
        cur = con.cursor()

        try:
            cur.execute(
                """
                SELECT distinct aps.name
                FROM aps_application_instances ai
                JOIN aps_service_instances si ON ai.app_instance_id = si.app_instance_id
                JOIN aps_applications aps ON ai.app_id = aps.app_id
                JOIN subs_resources sr ON sr.rt_instance_id =  ai.resource_id
                JOIN subscriptions sub ON sub.sub_id = sr.sub_id
                WHERE sub.sub_id = %s
                """,
                self.subID
            )
            rv = cur.fetchall()

            return rv
        finally:
            cur.close()

    def getExternalIps(self):
        con = uSysDB.connect()
        cur = con.cursor()

        try:
            cur.execute("SELECT ip_address FROM configured_ips WHERE ip_type = 'E' AND sub_id = %s", self.subID)
            rv = cur.fetchall()

            return rv
        finally:
            cur.close()

    def getPeriodID(self, planId, period, periodType):
        api = pbaapi.PBAAPI()
        planPeriods = api.PlanPeriodListGet_API(planId, 1)
        for planPeriod in planPeriods:
            if int(planPeriod.Period) == period and int(planPeriod.PeriodType) == periodType:
                return int(planPeriod.PlanPeriodID)

        return None

    def getResourceUsage(self, resource):
        con = uSysDB.connect()
        cur = con.cursor()

        try:
            cur.execute(
                "select curr_usage from subs_resources sr join subscriptions s on s.sub_id = sr.sub_id join resource_types rt on sr.rt_id = rt.rt_id join resource_classes rc on rt.class_id = rc.class_id where s.sub_id = %s and rc.name = %s",
                self.subID, resource
            )
            rv = cur.fetchone()
            if not rv:
                return 0

            return rv[0]
        finally:
            cur.close()

    def getResourceUsageForWebApps(self, resource_prefix, parent_resource):
        assert resource_prefix, "resource_prefix must be specified"
        assert parent_resource, "parent_resource must be specified"

        con = uSysDB.connect()
        cur = con.cursor()

        cur.execute(
            "select sr.curr_usage, sr.rt_id, rc.name from subs_resources sr join subscriptions s on s.sub_id = sr.sub_id join resource_types rt on sr.rt_id = rt.rt_id join resource_classes rc on rt.class_id = rc.class_id join resource_classes parent_rc on rc.parent_class_id = parent_rc.class_id where s.sub_id = %s and rc.name ilike %s and parent_rc.name = %s",
            self.subID, resource_prefix + '%', parent_resource
        )
        return cur.fetchall()

    def getResourceTypeIDs(self, resource):
        con = uSysDB.connect()
        cur = con.cursor()

        try:
            cur.execute(
                "select sr.rt_id from subs_resources sr join subscriptions s on s.sub_id = sr.sub_id join resource_types rt on sr.rt_id = rt.rt_id join resource_classes rc on rt.class_id = rc.class_id where s.sub_id = %s and rc.name = %s",
                self.subID, resource
            )
            rows = cur.fetchall()
            if rows:
                return [row[0] for row in rows]
            else:
                return []
        finally:
            cur.close()

    def getUnmigrateableSaasApplications(self):
        con = uSysDB.connect()
        cur = con.cursor()

        try:
            cur.execute(
                "SELECT saas_apps.app_instance_id, apps.app_id, apps.name, s.sub_id, s.owner_id, d.domain_id, d.name, d.owner_id FROM saas_application_instances saas_apps JOIN subscriptions s on (saas_apps.sub_id = s.sub_id) JOIN domains d on (saas_apps.domain_id = d.domain_id) JOIN aps_applications apps ON (saas_apps.app_id = apps.app_id) WHERE s.sub_id = %s AND s.owner_id != d.owner_id",
                self.subID
            )
            rv = cur.fetchall()

            return rv
        finally:
            cur.close()

    def getResourcesTypes(self, rt_ids):
        con = uSysDB.connect()
        cur = con.cursor()
        ids = tuple(rt_ids)

        try:
            cur.execute(
                """
                SELECT sr.rt_id, sr.curr_usage, rt.restype_name, rc.providable
                FROM subs_resources sr
                JOIN resource_types rt ON rt.rt_id = sr.rt_id
                JOIN resource_classes rc ON rc.class_id = rt.class_id
                WHERE sr.rt_id IN %s  and sr.sub_id = %s
                """,
                ids, self.subID
            )
            rv = cur.fetchall()
            return rv
        finally:
            cur.close()
