import re

from migutils import listmerge
from objects import Webspace
from poaupdater import uSysDB, uLogging, uTasks, uHCL, uPEM


class CheckStatus:
    SUCCESS = 0
    FAILED = 1
    WARNING = 2
    ERROR = 3
    toString = {
        SUCCESS: '\033[92mOK\033[0m',
        FAILED: '\033[91mFAILED\033[0m',
        WARNING: '\033[93mWARNING\033[0m',
        ERROR: '\033[91mERROR\033[0m'
    }


class CheckException(Exception):
    def __init__(self, msg='', code=CheckStatus.SUCCESS):
        Exception.__init__(self, msg)
        self.code = code


def connections(webspace_ids, target_host):
    hosts = set()
    for webspace_id in webspace_ids:
        webspace = Webspace(webspace_id)
        hosts.add(webspace.get_webspace_host())

    failed_source_hosts = []
    for host in hosts:
        try:
            hcl = uHCL.Request(host.director_host.id, "root", "root")
            host_backnet_ip = target_host.get_backnet_ip()

            hcl.command(
                '/usr/bin/rsync',
                args=[
                    'rsync', '-e', 'ssh -o StrictHostKeyChecking=no -o CheckHostIP=no -o UserKnownHostsFile=/dev/null',
                    '--archive', '--compress', '--dry-run',
                    '/bin/ls', 'root@%s:/tmp' % host_backnet_ip
                ]
            )
            hcl.perform()
        except Exception, e:
            uLogging.err(e)
            failed_source_hosts.append(host)

    if not failed_source_hosts:
        return CheckStatus.SUCCESS, ''
    else:
        return CheckStatus.FAILED,\
               'The following source hosts could not be connected to the target host %s. Configure access to the ' \
               'source hosts with an SSH key, and restart migration.' % \
               ['%s' % x.director_host for x in sorted(failed_source_hosts)]


def disk_space(webspace_ids, target_host):
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute(
        """
        select sum(usage)
        from disk_space_paths p
            join disk_space_elems e on e.path_id = p.path_id
            join apache_resources r on r.rt_instance_id = e.rti_id
        where r.homedir_id in (%s)""" %
        ','.join(webspace_ids))
    usage = cur.fetchone()[0]

    hcl = uHCL.Request(target_host.director_host.id, "root", "root")
    hcl.command('df --portability %s' % target_host.get_vhost_root(), stdout="output")
    result = hcl.perform()
    output = result["output"]
    lines = output.splitlines()
    m = re.match(r'(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)', lines[-1])
    if not m:
        raise Exception('Could not parse output "%s"' % lines[-1])

    available = int(m.group(4))
    if usage < available:
        return CheckStatus.SUCCESS,\
               'Disk space was checked successfully at the host %s. Required: %s KB, available: %s KB.' % \
               (target_host, usage, available)
    else:
        return CheckStatus.FAILED,\
               'There is not enough disk space at the host %s. Required: %s KB, available: %s KB. ' \
               'Free up some disk space or reduce the list of webspaces, and then restart migration.'\
               % (target_host, usage, available)


def ip_pools(webspace_ids, target_host):

    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute(
        """
        select distinct pool_id from configured_ips ip join apache_resources r on r.rt_instance_id=ip.ip_rti_id
        where homedir_id in (%s)""" %
        ','.join(webspace_ids)
    )

    source_ip_pools = set(listmerge(cur.fetchall()))

    cur.execute(
        """
        select distinct pool_id from bound_ip_pools p join net_interfaces if on if.if_id=p.if_id where if.host_id = %s
        """,
        target_host.id
    )

    target_ip_pools = set(listmerge(cur.fetchall()))

    if source_ip_pools <= target_ip_pools:
        return CheckStatus.SUCCESS, ''
    else:
        return CheckStatus.WARNING, \
               'These IP pools are used by the source hosts but are not assigned to the target host: %s. ' \
               'If you want the migrated webspaces to use the IP addresses from the same IP pools as were used ' \
               'before migration, assign these IP pools to the target hosts.' % \
               list(source_ip_pools - target_ip_pools)


def hosts_availability(webspace_ids, target_host):
    hosts = set([target_host])
    for webspace_id in webspace_ids:
        webspace = Webspace(webspace_id)
        hosts.add(webspace.get_webspace_host().director_host)

    failed_hosts = []
    for host in hosts:
        try:
            uPEM.ping(host.id)
        except Exception:
            failed_hosts.append(host)

    if not failed_hosts:
        return CheckStatus.SUCCESS,\
               'Successfully connected to the source and target systems %s' % ['%s' % x for x in hosts]
    else:
        return CheckStatus.FAILED,\
               'Could not connect to these hosts: %s. ' \
               'Make sure that these hosts are online and that the pa-agent service is running on them.' % \
               ['%s' % x for x in failed_hosts]


def host_websites_capacity(webspace_ids, target_host):
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute(
        """
        select limit_value from services s join apache_services_limits sl on sl.service_id=s.service_id 
        where limit_name='host_configuration.sites_per_host' and host_id = %s""",
        target_host.id
    )
    sites_per_host_limit = cur.fetchone()[0]

    cur.execute(
        """
        select count(*) from apache_homedir_entries e join home_directories d on d.hd_id=e.homedir_id
        where whoami='A' and d.host_id = %s""",
        target_host.id
    )
    sites_current = cur.fetchone()[0]

    cur.execute(
        "select count(*) from apache_homedir_entries where whoami='A' and homedir_id in (%s)" %
        ','.join(webspace_ids)
    )
    sites_additional = cur.fetchone()[0]

    if sites_additional + sites_current <= sites_per_host_limit or sites_per_host_limit == 0:
        return CheckStatus.SUCCESS, \
               "limit %s, sites at host %s, sites to migration %s" % \
               (sites_per_host_limit, sites_current, sites_additional)
    else:
        return CheckStatus.FAILED, \
               "The limit on the number of websites that can be hosted by %s is reached. " \
               "The limit is %s, the number of hosted websites is %s, and the number of websites to migrate is %s. " \
               "Increase the limit or reduce the list of webspaces, and then restart migration." % \
               (target_host, sites_per_host_limit, sites_current, sites_additional)


def php_versions(webspace_ids, target_host):
    """
    check necessary php versions are installed at the target host
    
    :param webspace_ids: 
    :param target_host: 
    :return: 
    """
    con = uSysDB.connect()
    cur = con.cursor()
    webspace_ids_str = ','.join(webspace_ids)
    cur.execute(
        """
        select distinct ae.name
        from apache_extensions ae
            join services s on s.service_id=ae.service_id
            join home_directories hd on hd.host_id=s.host_id
        where ae.type like '%s' and ae.name is not null and hd_id in (%s) and
        value in (select distinct path_to_php from apache_hostlist where homedir_id in (%s))
        """ % ('php%version%', webspace_ids_str, webspace_ids_str)
    )
    used_php = set(listmerge(cur.fetchall()))

    cur.execute(
        """
        SELECT ae.name
        FROM apache_extensions ae join services s on s.service_id=ae.service_id
        WHERE ae.type LIKE '%s' AND ae.name IS NOT NULL  AND s.host_id = %s""" % ('php%version%', target_host.id))
    exist_php = set(listmerge(cur.fetchall()))

    if used_php <= exist_php:
        return CheckStatus.SUCCESS,\
               'The source websites use the following PHP versions: %s. ' \
               'The PHP versions available at the server: %s.' % (list(used_php), list(exist_php))
    else:
        return CheckStatus.FAILED,\
               'The source websites use PHP, which is not installed on the target host. ' \
               'Install PHP %s on the target host and restart migration.' % list(used_php - exist_php)


def ftp_users_uniqueness(webspace_ids, target_host):
    con = uSysDB.connect()
    cur = con.cursor()
    webspace_ids_str = ','.join(webspace_ids)
    cur.execute(
        """
        with t_proftpd_users as (
            select u.user_id, u.vhost_id, u.ds_id, u.login, ds.hd_id
            from proftpd_vhosts_users u join proftpd_domain_services ds on ds.ds_id=u.ds_id
            where ds.hd_id in (%s)
            union
            select u.user_id, u.vhost_id, u.ds_id, u.login, ds.hd_id
            from proftpd_vhosts_users u join proftpd_domain_services ds on ds.ds_id=u.ds_id 
                join proftpd_vhosts vh on vh.vhost_id=u.vhost_id join services s on s.service_id=vh.service_id
            where s.host_id=%s
            )

        select login, array_agg(hd_id)
        from t_proftpd_users
        where login in (select login from t_proftpd_users group by login having count(*)>1) group by login;
        """ % (webspace_ids_str, target_host.id)
    )
    duplicates = cur.fetchall()
    if not duplicates:
        return CheckStatus.SUCCESS, \
               'FTP usernames of the webspaces that you are migrating and FTP usernames on the target server are unique.'
    else:
        return CheckStatus.FAILED,\
               'The following FTP usernames are used in several webspaces: %s. ' \
               'Rename the duplicate FTP usernames and restart migration.' % \
               ', '.join(["'%s' %s" % (login, webspace_ids) for (login, webspace_ids) in duplicates])

checks = {
    'connections': {'func': connections, 'descr': 'connection between source and target hosts'},
    'disk_space': {'func': disk_space, 'descr': 'disk space'},
    'ip_pools': {'func': ip_pools, 'descr': 'IP pools'},
    'hosts_availability': {'func': hosts_availability, 'descr': 'hosts availability'},
    'host_websites_capacity': {'func': host_websites_capacity, 'descr': 'host websites capacity'},
    'php_versions': {'func': php_versions, 'descr': 'php versions'},
    'ftp_users_uniqueness': {'func': ftp_users_uniqueness, 'descr': 'ftp users uniqueness'}
}

checks_order = [
    'hosts_availability', 'connections', 'disk_space', 'ip_pools', 'host_websites_capacity', 'php_versions',
    'ftp_users_uniqueness'
]
