#!/usr/bin/python

import datetime
import getopt
import json
import pickle
import sys
import os

import checks
from migutils import Service
from objects import Webspace, HostBase, MigrationState, init
from poaupdater import uSysDB, uLogging, uTasks, uHCL, uUtil, uConfig, openapi

LOGPARSER = '/usr/local/pem/bin/logparser.sh'


def migrate_webspace(webspace, source_host, target_host):
    if source_host == target_host:
        uLogging.info("Webspace %s is served by host %s already, nothing to migrate.", webspace, target_host)
        return

    webspace.process_step(MigrationState.LOCK_WEBSPACE, webspace.lock, source_host)
    webspace.process_step(MigrationState.COPY_CONTENT, copy_content, webspace, source_host, target_host)

    webspace.process_step(MigrationState.RELEASE_IP, webspace.release_ip)
    webspace.process_step(MigrationState.CHANGE_SETTINGS, change_webspace_settings, webspace, source_host, target_host)
    webspace.process_step(MigrationState.RESTORE_IP, webspace.restore_ip)

    webspace.process_step(MigrationState.REPLACE_ABS_PATH, replace_abs_paths, webspace, source_host, target_host)
    webspace.process_step(MigrationState.REMOVE_SOURCE_CONTENT, remove_source_content, webspace, source_host)

    webspace.process_step(MigrationState.UNLOCK, webspace.unlock, target_host)

    webspace.process_step(MigrationState.SYNC_HTACESS, webspace.sync_htacess)
    webspace.process_step(MigrationState.SYNC_WEBALIZER, sync_webalizer_config, webspace, target_host)


def copy_content(webspace, source_host, target_host):
    uLogging.info("Copying the content of the webspace %s from host %s to host %s", webspace, source_host, target_host)
    hcl = uHCL.Request(source_host.director_host.id, "root", "root")
    host_backnet_ip = target_host.get_backnet_ip()

    source_webspace_path = webspace.source_path
    target_webspace_path = target_host.get_vhost_root().rstrip('/')
    hcl.command(
        '/usr/bin/rsync',
        args=[
            'rsync', '-e', 'ssh -o StrictHostKeyChecking=no -o CheckHostIP=no -o UserKnownHostsFile=/dev/null',
            '--archive', '--compress',
            source_webspace_path, 'root@%s:%s' % (host_backnet_ip, target_webspace_path)
        ]
    )
    hcl.perform()
    uLogging.info("The content was copied")


def replace_abs_paths(webspace, source_host, target_host):
    uLogging.info("Checking content of webspace %s for absolute paths", webspace.webspace_id)
    hcl = uHCL.Request(target_host.director_host.id, "root", "root")

    lines = []
    source_webspace_path = "%s/%s" % (source_host.get_vhost_root().rstrip('/'), webspace.webspace_id)
    target_webspace_path = "%s/%s" % (target_host.get_vhost_root().rstrip('/'), webspace.webspace_id)
    try:
        hcl.command(
            "/usr/bin/find '%s' -type f ! -iname '*log' -exec grep -lI '%s' '{}' \;" %
            (target_webspace_path, source_webspace_path), stdout="output", valid_exit_codes=[0]
        )

        result = hcl.perform()
        output = result["output"]
        lines = output.splitlines()
    except uUtil.ExecFailed, e:
        if 1 != e.status:
            raise

    if lines:
        uLogging.info("Following files seem to have absolute file paths in them:")
        for line in lines:
            uLogging.info("%s", line)

        for line in lines:
            if line.endswith(".bak"):
                continue

            hcl.command(
                "/bin/sed",
                ["/bin/sed", "-i.bak", "-e", "s|%s|%s|g" % (source_webspace_path, target_webspace_path)] + [line]
            )
            hcl.perform()


def remove_source_content(webspace, source_host):
    hcl = uHCL.Request(source_host.director_host.id, "root", "root")
    source_webspace_path = webspace.source_path
    hcl.rm(source_webspace_path)
    hcl.perform()


def change_webspace_settings(webspace, source_host, target_host):
    uLogging.info("Change webspace %s settings", webspace.webspace_id)
    con = uSysDB.connect()
    cur = con.cursor()
    change_home_directories(cur, webspace.webspace_id, source_host.id, target_host.id)
    update_apache(cur, webspace.webspace_id, source_host, target_host)
    update_domain_services(cur, webspace.webspace_id, source_host, target_host)
    update_proftpd_domain_services(cur, webspace.webspace_id, source_host, target_host)
    update_webspace_aliases(cur, webspace, source_host, target_host)
    update_disk_space_path(cur, webspace, source_host, target_host)
    update_webspace_log_name(cur, webspace, source_host, target_host)
    update_webspace_log_files(cur, webspace.webspace_id, source_host, target_host)
    update_vhosts_users(cur, webspace.webspace_id, source_host, target_host)
    update_certificate_paths(cur, webspace, source_host, target_host)
    cur.close()
    con.commit()


def sync_webalizer_config(webspace, target_host):
    uLogging.info("Updating log files for the webspace %s", webspace.webspace_id)
    target_logparser_id = target_host.get_service_id(Service.LOGPARSER)
    pobjref = uTasks.PObjRef(uTasks.SCRef("proclog"), "getService", target_logparser_id)
    task = uTasks.Task(pobjref, 'Add logfile', 'updateLogfile')
    target_webspace_log_file = '%s/%s/log' % (target_host.get_vhost_root(), webspace.webspace_id)
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute(
        "select logfile_id, file_mask FROM logfiles WHERE service_id = %s AND dir_path = %s",
        target_logparser_id, target_webspace_log_file
    )
    task_ids = []
    for (logfile_id, file_mask) in cur.fetchall():
        uLogging.info("Schedule task for log file %s (%s)", file_mask, logfile_id)
        task.setParam('logfile_id', logfile_id)
        task_ids.append(task.schedule())

    if task_ids:
        uTasks.waitTasks(con, task_ids, 30, 2)


def change_home_directories(cur, webspace_id, source_host_id, target_host_id):
    cur.execute(
        "UPDATE home_directories SET host_id = %s WHERE hd_id = %s AND host_id = %s",
        target_host_id, webspace_id, source_host_id
    )


def update_apache(cur, webspace_id, source_host, target_host):
    service_id = target_host.get_service_id(Service.APACHENG)
    host_ip = target_host.get_host_share_ip()
    cur.execute(
        """
        UPDATE apache_hostlist
        SET service_id = %s, host_ip = %s, config = replace(config, %s, %s)
        WHERE homedir_id = %s""",
        service_id, host_ip, source_host.get_vhost_root(), target_host.get_vhost_root(), webspace_id
    )

    mpoint_id = target_host.mpoint_id
    cur.execute(
        "UPDATE apache_h2e_ws_to_mpoint_affinity SET mpoint_id = %s WHERE homedir_id = %s",
        mpoint_id, webspace_id
    )


def update_domain_services(cur, webspace_id, source_host, target_host):
    services = [Service.APACHENG, Service.PROFTPDNG, Service.CRONTABS]
    for service in services:
        source_service_id = source_host.get_service_id(service)
        target_service_id = target_host.get_service_id(service)
        cur.execute(
            "UPDATE domain_services SET service_id = %s WHERE hd_id = %s AND service_id = %s",
            target_service_id, webspace_id, source_service_id
        )


def update_proftpd_domain_services(cur, webspace_id, source_host, target_host):
    source_service_id = source_host.get_service_id(Service.PROFTPDNG)
    target_service_id = target_host.get_service_id(Service.PROFTPDNG)
    cur.execute(
        "UPDATE proftpd_domain_services SET service_id = %s WHERE hd_id = %s AND service_id = %s",
        target_service_id, webspace_id, source_service_id
    )


def update_webspace_aliases(cur, webspace, source_host, target_host):
    cur.execute(
        """
        UPDATE apache_homedir_entries
        SET aliases = cast(replace(cast(aliases as text), %s, %s) as bytea)
        WHERE homedir_id = %s""",
        source_host.get_vhost_root(), target_host.get_vhost_root(), webspace.webspace_id
    )


def update_disk_space_path(cur, webspace, source_host, target_host):
    source_path = get_disk_space_path(webspace, source_host)
    target_path = get_disk_space_path(webspace, target_host)
    cur.execute(
        "UPDATE disk_space_paths SET path = %s, host_id = %s WHERE path = %s AND host_id = %s",
        target_path, target_host.director_host.id, source_path, source_host.director_host.id
    )


def update_webspace_log_name(cur, webspace, source_host, target_host):
    source_logrotate_id = source_host.get_service_id(Service.LOGROTATE)
    target_logrotate_id = target_host.get_service_id(Service.LOGROTATE)
    log_name = webspace.get_log_name(source_logrotate_id)
    cur.execute(
        "UPDATE lrs SET service_id = %s, logname = replace(logname, %s, %s) WHERE service_id = %s AND logname = %s",
        target_logrotate_id, source_host.get_vhost_root(), target_host.get_vhost_root(), source_logrotate_id, log_name
    )
    cur.execute(
        "UPDATE proftpd_lfa_data set vhost_id = %s where hd_id = %s",
        target_host.proftpd_vhost_id, webspace.webspace_id
    )


def update_webspace_log_files(cur, webspace_id, source_host, target_host):
    source_logparser_id = source_host.get_service_id(Service.LOGPARSER)
    target_logparser_id = target_host.get_service_id(Service.LOGPARSER)
    source_webspace_log_file = '%s/%s/log' % (source_host.get_vhost_root(), webspace_id)

    cur.execute(
        """
        UPDATE logfiles_actions la set opts = replace(opts, %s, %s)
        FROM logfiles l WHERE l.logfile_id = la.logfile_id and service_id = %s AND dir_path = %s
        """,
        source_host.get_vhost_root(), target_host.get_vhost_root(), source_logparser_id, source_webspace_log_file
    )

    cur.execute(
        """
        UPDATE logfiles SET service_id = %s, dir_path = replace(dir_path, %s, %s) WHERE service_id = %s AND dir_path = %s
        """,
        target_logparser_id, source_host.get_vhost_root(), target_host.get_vhost_root(),
        source_logparser_id, source_webspace_log_file
    )


def update_vhosts_users(cur, webspace_id, source_host, target_host):
    vhosts_users = get_vhosts_users(source_host, webspace_id)
    for user_id, sys_home in vhosts_users:
        cur.execute(
            "UPDATE proftpd_vhosts_users SET sys_home = replace(sys_home, %s, %s) WHERE user_id = %s",
            source_host.get_vhost_root(), target_host.get_vhost_root(), user_id
        )
        cur.execute(
            "UPDATE proftpd_vhosts_users SET vhost_id = %s WHERE ds_id is NULL and user_id = %s",
            target_host.proftpd_vhost_id, user_id
        )


def update_certificate_paths(cur, webspace, source_host, target_host):

    for certificate in webspace.get_apache_certificates():
        source_filename = certificate['cert_filename']

        source_root = source_host.get_vhost_root()
        target_root = target_host.get_vhost_root()
        target_filename = source_filename
        if source_filename.startswith(source_root):
            target_filename = target_root + source_filename[len(source_root):]

        cur.execute(
            "UPDATE apache_certificates SET cert_filename = %s WHERE homedir_id = %s AND cert_filename = %s",
            target_filename, webspace.webspace_id, source_filename
        )


def get_disk_space_path(webspace, host):
    sysuser_name = webspace.get_sysuser_name()
    root_path = host.get_root_path()
    return "%s@%s" % (sysuser_name, root_path)


def get_vhosts_users(host, webspace_id):
    con = uSysDB.connect()
    cur = con.cursor()
    cur.execute(
        "SELECT user_id, sys_home FROM proftpd_vhosts_users WHERE sys_home like %s",
        '%s/%s%%' % (host.get_vhost_root(), webspace_id)
    )
    return cur.fetchall()


def refresh_webalizer_config(target_host_id):
    hcl = uHCL.Request(target_host_id, "root", "root")
    hcl.command(LOGPARSER)
    hcl.perform()


def move_webspaces(opts):
    if '--checkpoint' in opts:
        checkpoint = opts['--checkpoint']
        f = open(checkpoint, 'r')
        webspaces, target_host = pickle.load(f)
        f.close()
    else:
        checkpoint = None
        webspace_ids = opts['--webspace-ids']
        target_host = opts['--target-host-id']

        webspaces = []
        for webspace_id in webspace_ids.split(','):
            webspace = Webspace(webspace_id)
            webspaces.append(webspace)

    failed_webspaces = {}
    for webspace in webspaces:
        try:
            uLogging.info("Start to migrate webspace with ID %s." % str(webspace.webspace_id))
            migrate_webspace(webspace, webspace.source_host, HostBase.create_host(target_host))
            uLogging.info("The webspace with ID %s was successfully migrated." % str(webspace.webspace_id))
        except Exception as e:
            uLogging.err("Unable to migrate webspace with ID %s: %s." % (str(webspace.webspace_id), e.message))
            failed_webspaces[webspace.webspace_id] = e.message

    general_errors = []
    if '--skip-webalizer' not in opts:
        try:
            refresh_webalizer_config(target_host)
        except Exception as e:
            general_errors.append(e)

    if failed_webspaces or general_errors:
        if checkpoint is None:
            checkpoint = 'checkpoint_%s' % datetime.datetime.now().strftime('%Y_%m_%d_%H_%M_%S')

        f = open(checkpoint, 'w+')
        pickle.dump((webspaces, target_host), f)
        f.close()

        if failed_webspaces:
            uLogging.err('Failed to migrate following webspaces:')
            for webspace_id in failed_webspaces:
                uLogging.err('\tWebspace #%s - %s' % (webspace_id, failed_webspaces[webspace_id]))

        if general_errors:
            uLogging.err('During the migration, the following common errors were detected:')
            for error in general_errors:
                uLogging.err('\t%s' % error)

        uLogging.err(
            "After fixing the issues, restart the migration with the command './migrate.py --checkpoint %s "
            "move-webspaces'." % checkpoint
        )
    else:
        uLogging.info('Migration completed successfully')


def check(opts):
    if '--webspace-ids' not in opts:
        raise Exception("At least one webspace id is expected, use --webspace-ids parameter")

    if '--target-host-id' not in opts:
        raise Exception("Target host id is expected, use --target-host-id")

    webspace_ids = opts['--webspace-ids'].split(',')
    target_host_id = opts['--target-host-id']
    if '--aspect' in opts:
        aspects = [x for x in checks.checks_order if x in opts['--aspect'].split(',')]
        if not aspects:
            raise Exception("Option --aspect must have arguments")

        failed_aspects = [aspect for aspect in opts['--aspect'].split(',') if aspect not in checks.checks]
        if failed_aspects:
            raise Exception("Check(s) {aspects} were not found".format(aspects=failed_aspects))
    else:
        aspects = checks.checks_order

    result = {}
    target_host = HostBase.create_host(target_host_id)

    for aspect in aspects:
        try:
            res = checks.checks[aspect]['func'](webspace_ids, target_host)
            result[aspect] = {'code': res[0], 'msg': res[1]}
        except Exception, e:
            result[aspect] = {'code': checks.CheckStatus.ERROR, 'msg': str(e)}

    text = []

    max_len = max([len(checks.checks[aspect]['descr']) for aspect in result])
    for aspect in aspects:
        descr = checks.checks[aspect]['descr']
        code = checks.CheckStatus.toString[result[aspect]['code']]
        msg = result[aspect]['msg']
        text.append("Precheck '%s'" % descr + ': ' + ' ' * (max_len - len(descr)) + code + ' ' + msg)

    uLogging.info('Precheck results:\n' + '\n'.join(text))
    code = checks.CheckStatus.SUCCESS
    text = []
    for aspect in result:
        if result[aspect]['code'] != checks.CheckStatus.SUCCESS:
            text.append(result[aspect]['msg'])

        if result[aspect]['code'] in [checks.CheckStatus.FAILED, checks.CheckStatus.ERROR]:
            code = checks.CheckStatus.FAILED
        elif code != checks.CheckStatus.FAILED and result[aspect]['code'] == checks.CheckStatus.WARNING:
            code = checks.CheckStatus.WARNING

    if code != checks.CheckStatus.SUCCESS:
        raise checks.CheckException('\n'.join(text), code)


def save_migration_result_to_file(opts, exit_code, message=''):
    if opts and '--results-file' in opts:
        filename = opts['--results-file']
    else:
        return

    data = {
        'exit_code': exit_code,
        'message': message,
    }
    f = open(filename, 'w')
    json.dump(data, f)


migrator_options = {
    'webspace-ids=': 'Webspace ids',
    'target-host-id=': 'Target host id',
    'results-file=': 'Path to migration report',
    'aspect=': 'Run a series of checks that check if the system is ready for migration',
    'checkpoint=': 'Path to the file with saved checkpoint',
    'skip-webalizer': 'Skip refreshing config of webalizer',
    'verbose': 'Detailed trace output to log'
}
command = sys.argv[-1]
options, _ = getopt.getopt(sys.argv[1:], '', dict(migrator_options).keys())
options = dict(options)
try:
    config = uConfig.Config(verbose='--verbose' in options)
    path = os.path.abspath(__file__)
    config.log_file = path[:path.rfind('.')] + '.log'
    uLogging.init(config)
    uSysDB.init(config)
    openapi.initFromEnv(config)
    init()

    if command == 'move-webspaces':
        move_webspaces(options)
    elif command == 'check':
        check(options)

    save_migration_result_to_file(options, 0)
except checks.CheckException, ex:
    uLogging.debug("%s", ex)
    save_migration_result_to_file(options, ex.code, "%s" % ex)
    exit(1)
except Exception, ex:
    uLogging.debug("%s", ex)
    save_migration_result_to_file(options, 1, "%s" % ex)
    exit(1)
