from migutils import SSHAccess, Service
from poaupdater import uHCL, uLogging, uSysDB, openapi, uTasks

api = None
con = None


def init():
    global api, con
    api = openapi.OpenAPI()
    con = uSysDB.connect()

class IPType(object):
    SHARED = 0
    EXCLUSIVE = 1
    RENEW = 5


class MigrationState(object):
    INITIAL = 0
    LOCK_WEBSPACE = 1
    COPY_CONTENT = 2
    RELEASE_IP = 3
    CHANGE_SETTINGS = 4
    RESTORE_IP = 5
    REPLACE_ABS_PATH = 6
    REMOVE_SOURCE_CONTENT = 7
    UNLOCK = 8
    SYNC_HTACESS = 9
    SYNC_WEBALIZER = 10


class HostBase:

    TYPE_WEB_SERVER = 'g'
    TYPE_WEB_CLUSTER = 'e'
    TYPE_WEB_CLUSTER_SERVER = 'f'
    TYPE_NORMAL = 'n'

    mpoint_id = None
    ip_address = None
    root_path = None
    proftpd_vhost_id = None

    @classmethod
    def create_host(cls, host_id):
        cur = con.cursor()

        host_id = int(host_id)

        cur.execute("SELECT htype FROM hosts WHERE host_id = %s", host_id)
        row = cur.fetchone()

        if not row:
            raise Exception("Failed to find host with ID %s" % host_id)

        if Host.TYPE_WEB_CLUSTER == row[0]:
            return Cluster(host_id)
        else:
            return Host(host_id)

    def __init__(self, host_id):
        self.id = int(host_id)

        cur = con.cursor()

        cur.execute("select primary_name from hosts where host_id = %s", self.id)
        self.primary_name, = cur.fetchone()

        cur.execute("select vhost_id from proftpd_vhosts where service_id = %s", self.get_service_id(Service.PROFTPDNG))
        row = cur.fetchone()
        if row:
            self.proftpd_vhost_id, = row

        cur.execute(
            """
            SELECT mpoint_id, ip_address, root_path FROM apache_h2e_mount_points p, apache_h2e_nodes n
            WHERE p.fnode_id = n.node_id and n.cluster_id = %s
            """, self.id
        )
        row = cur.fetchone()
        if row:
            self.mpoint_id, self.ip_address, self.root_path = row

    def __str__(self):
        return '#%s %s' % (self.id, self.primary_name)

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    @property
    def director_host(self):
        cur = con.cursor()
        cur.execute("SELECT director_host_id FROM apache_h2e_clusters WHERE cluster_id = %s", self.id)
        return Host(cur.fetchone()[0])

    def get_service_id(self, service_type):
        cur = con.cursor()
        cur.execute("SELECT service_id FROM services WHERE name = %s AND host_id = %s", service_type, self.id)
        row = cur.fetchone()
        if row:
            return row[0]

        return None

    def get_vhost_root(self):
        return "/var/www/vhosts/%s" % self.mpoint_id

    def get_host_share_ip(self):
        cur = con.cursor()
        cur.execute(
            "SELECT h2e_denormalize_ip(ip_address) FROM configured_ips WHERE host_id = %s AND ip_type = 'S'",
            self.id
        )
        return cur.fetchone()[0]

    def cmd(self, command):
        uLogging.info("%s# %s", self, command)
        hcl = uHCL.Request(self.id, "root", "root")
        hcl.command(command, stdout="result")
        return hcl.perform()

    def kill_ssh_sessions(self):
        # kill all non root sshd processes
        return self.cmd(
            "for i in `ps -ef | grep sshd | grep -v root | tr -s ' ' | cut -d' ' -f2`; do kill -TERM $i; done"
        )

    def kill_ftp_sessions(self):
        return self.cmd("service proftpd restart")

    def get_backnet_ip(self):
        cur = con.cursor()
        cur.execute(
            """
            SELECT vp.value
            FROM components c
            JOIN packages p ON c.pkg_id = p.pkg_id
            JOIN v_props vp ON c.component_id = vp.component_id
            WHERE vp.name = 'communication.ip' AND p.name = 'pleskd' AND c.host_id = %s
            """,
            self.id
        )
        return cur.fetchone()[0]

    def get_root_path(self):
        raise NotImplementedError()


class Host(HostBase):
    def get_root_path(self):
        return self.root_path


class Cluster(HostBase):
    def get_root_path(self):
        return '%s:%s' % (self.ip_address, self.root_path)


class Webspace:

    _fine_ip_types = [IPType.SHARED, None]

    def __init__(self, webspace_id):
        self.webspace_id = int(webspace_id)
        self.apache_ip_type = self.get_apache_ip_type()
        self.proftpd_ip_type = self.get_proftpd_ip_type()
        self._locked_host = None
        self.migration_state = MigrationState.INITIAL
        self.source_host = self.get_webspace_host()
        self.source_path = self.get_home_dir(self.source_host)

    def __str__(self):
        return '#%s' % self.webspace_id

    def get_domain_service_id(self, service_name):
        cur = con.cursor()
        cur.execute(
            """
            SELECT ds_id
            FROM domain_services ds JOIN services s ON ds.service_id = s.service_id
            WHERE ds.hd_id = %s AND s.name = %s
            """,
            self.webspace_id, service_name
        )
        return cur.fetchone()[0]

    def set_migrating_state(self, migrating):
        api.pem.web_cluster.setMigratingState(webspaceID=self.webspace_id, set=migrating)

        try:
            if migrating:
                api.pem.web_cluster.disableFTP(webspaceID=self.webspace_id)
            else:
                api.pem.web_cluster.enableFTP(webspaceID=self.webspace_id)
        except openapi.OpenAPIError, e:
            # ignore FTP not enabled errors
            if -1 == e.error_message.find("service #0 does not exist"):
                raise e

    def lock(self, host):
        self.set_migrating_state(True)
        self._locked_host = host
        self.sync_database(host)

        ssh = SSHAccess()
        ssh.disable(self, host)

        host.kill_ssh_sessions()
        host.kill_ftp_sessions()

    def unlock(self, host):
        self.set_migrating_state(False)
        if self._locked_host:
            self.sync_database(self._locked_host)
            self._locked_host = None

        self.sync_database(host)

        ssh = SSHAccess()
        ssh.enable(self, host)

    def sync_database(self, host):
        uLogging.info("Syncing redis database for the webspace %s", self)
        api.pem.web_cluster.syncDatabase(clusterID=host.id, webspaceIDs=[self.webspace_id])

    def sync_htacess(self):
        uLogging.info("Syncing .htaccess files for the webspace %s", self)
        apache_service_id = self.get_domain_service_id(Service.APACHENG)
        pobjref = uTasks.PObjRef(uTasks.SCRef("apache"), "getDomainService", apache_service_id)
        task = uTasks.Task(
            pobjref,
            "A task that synchronize .htaccess files for the domain service with ID %s" % apache_service_id,
            "taskSyncHtAccess"
        )
        task.schedule()

    def get_apache_certificates(self):
        cur = con.cursor()
        cur.execute("SELECT cert_filename FROM apache_certificates WHERE homedir_id = %s", self.webspace_id)
        return cur.fetchall()

    def get_apache_ip_type(self):
        return self._get_ip_type("SELECT ip_type FROM apache_hostlist WHERE homedir_id = %s", self.webspace_id)

    def get_proftpd_ip_type(self):
        return self._get_ip_type("SELECT ip_type FROM proftpd_domain_services WHERE hd_id = %s", self.webspace_id)

    def set_apache_ip_type(self, ip_type):
        api.pem.apache.switchIPType(webspace_id=self.webspace_id, ip_type=ip_type)

    def set_proftpd_ip_type(self, ip_type):
        try:
            api.pem.ProFTPD.switchIPType(webspace_id=self.webspace_id, ip_type=ip_type)
        except openapi.OpenAPIError, e:
            # ignore FTP not enabled errors
            if -1 == e.error_message.find('Could not find entry for home directory'):
                raise e

    def get_webspace_aliases(self):
        cur = con.cursor()
        cur.execute("SELECT aliases FROM apache_homedir_entries WHERE homedir_id = %s", self.webspace_id)
        return cur.fetchone()[0]

    def get_sysuser_name(self):
        cur = con.cursor()
        cur.execute("SELECT username FROM apache_hostlist WHERE homedir_id = %s" % self.webspace_id)
        return cur.fetchone()[0]

    def get_webspace_host(self):
        cur = con.cursor()
        cur.execute("SELECT host_id FROM home_directories WHERE hd_id = %s", self.webspace_id)
        return HostBase.create_host(cur.fetchone()[0])

    def get_home_dir(self, host):
        return "%s/%s" % (host.get_vhost_root(), self.webspace_id)

    def get_log_name(self, service_id):
        cur = con.cursor()
        cur.execute(
            "SELECT logname FROM lrs WHERE service_id = %s AND logname like %s",
            service_id, '%%%s%%' % self.webspace_id
        )
        return cur.fetchone()[0]

    def release_ip(self):
        if self.apache_ip_type not in self._fine_ip_types:
            self.set_apache_ip_type(IPType.SHARED)

        if self.proftpd_ip_type not in self._fine_ip_types:
            self.set_proftpd_ip_type(IPType.SHARED)

    def restore_ip(self):
        apache_ip_type = [self.apache_ip_type, IPType.RENEW][self.apache_ip_type == IPType.SHARED]
        if apache_ip_type not in self._fine_ip_types:
            self.set_apache_ip_type(apache_ip_type)

        proftpd_ip_type = [self.proftpd_ip_type, IPType.RENEW][self.proftpd_ip_type == IPType.SHARED]
        if proftpd_ip_type not in self._fine_ip_types:
            self.set_proftpd_ip_type(proftpd_ip_type)

    @staticmethod
    def _get_ip_type(sql, *args):
        cur = con.cursor()
        cur.execute(sql, *args)
        row = cur.fetchone()
        ip_type = None
        if row:
            ip_type = int(row[0])

        return ip_type

    def process_step(self, step, func, *arg):
        if self.migration_state < step:
            func(*arg)
            self.migration_state = step
