# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import os
import subprocess
import sys
from pipes import quote
import threading

from migrator.core.language import _T

logger = logging.getLogger(__name__)


class MigrationError(Exception):
    def __init__(self, message, code=-1):
        super(MigrationError, self).__init__(message)
        self.code = code


class MigrationNotice(Exception):
    def __init__(self, message, code=-1):
        super(MigrationNotice, self).__init__(message)
        self.code = code


class InternalError(Exception):
    pass


class LshNgMigratorError(Exception):
    pass


class NgNgMigratorError(Exception):
    pass


class ParallelExecutionError(MigrationError):
    def __init__(self):
        super(MigrationError, self).__init__(_T("EXCEPTION_PARALLEL_EXECUTION"))


class CommandExecutionError(MigrationError):
    """Exception occurred when executing some command on some host"""

    def __init__(self, message, stdout='', stderr='', host_description='', cause='', command_str=None,
                 full_command_str=None, exit_code=None):
        super(MigrationError, self).__init__(message)
        self._stdout = stdout
        self._stderr = stderr
        self._host_description = host_description
        self._cause = cause
        self._command_str = command_str
        self._full_command_str = full_command_str
        self._exit_code = exit_code

    @property
    def stdout(self):
        """stdout of the command"""
        return self._stdout

    @property
    def stderr(self):
        """stderr of the command"""
        return self._stderr

    @property
    def host_description(self):
        """Description of the host where the command was executed"""
        return self._host_description

    @property
    def cause(self):
        """Root cause of the error - exception caused command execution to fail"""
        return self._cause

    @property
    def command_str(self):
        return self._command_str

    @property
    def full_command_str(self):
        return self._full_command_str

    @property
    def exit_code(self):
        return self._exit_code


class APIError(MigrationError):
    """Exception occurred when executing some request through external API"""

    @property
    def how_to_reproduce(self):
        """User-friendly description on how to reproduce API request"""
        raise NotImplementedError()


class MigrationConfigurationFileError(Exception):
    pass


class MigrationNoContextError(Exception):
    pass


class MigrationNoRepeatError(Exception):
    """Exception explicitly stating that there is no sense to repeat failed operation for the object

    For example, if subscription creation failed because such subscription already exists,
    there is no sense to try to create subscription any more times.
    """
    pass


# raised if the subscription cannot be imported because of previous account import failure
class SubscriptionAccountImportException(MigrationError):
    def __init__(self, account, message="", code=-1):
        super(SubscriptionAccountImportException, self).__init__(message, code)
        self.account = account


# raised if import of the account has failed
class AccountImportException(SubscriptionAccountImportException):
    def __init__(self, account, message="", code=-1):
        super(AccountImportException, self).__init__(account, message, code)


class SubscriptionImportException(MigrationError):
    pass


class ChildSubscriptionImportException(MigrationError):
    def __init__(self, subscription, message="", code=-1):
        super(ChildSubscriptionImportException, self).__init__(message, code)
        self.subscription = subscription


class ParentSubscriptionImportException(MigrationError):
    def __init__(self, subscription, message="", code=-1):
        super(ParentSubscriptionImportException, self).__init__(message, code)
        self.subscription = subscription


class MappingParentSubscriptionImportException(MigrationError):
    def __init__(self, subscription, message="", code=-1):
        super(MappingParentSubscriptionImportException, self).__init__(message, code)
        self.subscription = subscription


class PopenExt(subprocess.Popen):
    def __init__(self, *args, **kwargs):
        super(PopenExt, self).__init__(*args, **kwargs)
        self._stdout = ''
        self._stderr = ''
        self._exception = None
        self._timeout = None

    def communicate(self, input=None):
        def target():
            try:
                self._stdout, self._stderr = super(PopenExt, self).communicate(input)
                if self._stdout is None:
                    self._stdout = ''
                if self._stderr is None:
                    self._stderr = ''

            except Exception, e:
                self._exception = e

        thread = threading.Thread(target=target)
        thread.start()

        thread.join(self._timeout)
        if thread.is_alive():
            self.terminate()
            thread.join()
            raise MigrationError('Timeout exceeded')

        if self._exception:
            raise self._exception

        return self._stdout, self._stderr

    def set_timeout(self, timeout):
        self._timeout = timeout


def run_and_check_local_command(command, stdin_content=None, output_codepage='utf-8', timeout=None):
    exit_code, stdout_str, stderr_str = local_command_exec(command, stdin_content, output_codepage, timeout=timeout)
    if exit_code != 0:
        raise MigrationError(
            _T("LOCAL_COMMAND_FAILED") % (
                u" ".join([quote(c) for c in command]),
                exit_code,
                stdout_str,
                stderr_str
            )
        )
    return stdout_str, stderr_str


def local_command_exec(
    command, stdin_content=None, output_codepage='utf-8', error_policy='strict', env=None, log_output=True, timeout=None
):
    from migrator.core.utils.loggers.secure_logger import hide_text
    if type(command) is list or type(command) is tuple:
        command_str = u' '.join([quote(c) for c in command])
        command_encoded = [c.encode('utf-8') for c in command]
    else:
        command_str = command
        command_encoded = command.encode('utf-8')
    logger.debug(_T("DEBUG_CALL_LOCAL_COMMAND") % command_str)

    command_env = os.environ.copy()
    command_env['LANG'] = 'en_US.utf-8'
    if env is not None:
        command_env.update(env)

    try:
        p = PopenExt(
            command_encoded,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=command_env
        )
        p.set_timeout(timeout)
        stdout_str, stderr_str = [s.decode(output_codepage, error_policy) for s in p.communicate(stdin_content)]
        logger.debug(_T("DEBUG_COMMAND_STDOUT"), hide_text(stdout_str.strip(), not log_output))
        logger.debug(_T("DEBUG_COMMAND_STDERR"), stderr_str.strip())
        logger.debug(_T("DEBUG_COMMAND_EXIT_CODE"), p.returncode)
        return p.returncode, stdout_str, stderr_str
    except Exception as e:
        logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
        raise Exception(
            _T("FAILED_TO_EXECUTE_LOCAL_COMMAND") % (command_str, unicode(e))
        ), None, sys.exc_info()[2]


def version_tuple(version):
    """Convert string product version to a tuple of ints"""
    return tuple(map(int, version.split('.')))
