# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.actions.utils.multithreading_properties import MultithreadingProperties
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.language import _T
from migrator.core.utils.loggers.log import trace
from migrator.settings import MIGRATION_TYPE_FULL

logger = logging.getLogger(__name__)


class CopyWebContentBase(SubscriptionAction):
    """Base class to copy web content"""

    def get_description(self):
        """
        :rtype: basestring
        """
        return _T("COPY_WEB_FILES_FROM_SERVERS")

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: parallels.core.global_context.GlobalMigrationContext
        :type subscription: parallels.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_COPY_WEB_FILES_FOR_SUBSCRIPTION") % subscription.name

    def is_critical(self):
        """If action is critical or not

        If action is critical and it failed for a subscription, migration tool
        won't run the next operations for the subscription.

        :rtype: bool
        """
        return False

    def get_multithreading_properties(self):
        """
        :rtype: parallels.core.actions.utils.multithreading_properties.MultithreadingProperties
        """
        return MultithreadingProperties(can_use_threads=True)

    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: parallels.core.global_context.GlobalMigrationContext
        :type subscription: parallels.core.migrated_subscription.MigratedSubscription
        """

        if global_context.migration_type != MIGRATION_TYPE_FULL or \
                global_context.migration_mode == MigrationMode.TEST_MIGRATION:
            logger.info(
                _T("LOG_NO_NEED_TO_COPY_WEBCONTENT_BECAUSE_DISABLED_BY_USER"),
                subscription.name
            )
            return False

        return True

    @trace
    def run(self, global_context, subscription):
        """
        :type global_context: parallels.core.global_context.GlobalMigrationContext
        :type subscription: parallels.core.migrated_subscription.MigratedSubscription
        """
        content_list_reader = self._get_content_list_reader(global_context)
        try:
            for items, content_migrator in content_list_reader.get_subscription_content(global_context, subscription):
                with content_migrator:
                    content_migrator.copy_content(items)
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(e)

    @trace
    def _get_content_list_reader(self, global_context):
        """
        :rtype: migrator.core.transfer.content_list_reader_base.ContentListReaderBase

        """
        raise NotImplementedError()

