# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
from collections import namedtuple

from migrator.core import _T
from migrator.core.actions.base.production_subscription_action import ProductionSubscriptionAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import \
    has_any_of_service_types, get_services
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils.loggers.report import Issue
from migrator.core.utils.loggers.report import Problem
from migrator.settings import \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass, SOURCE_SERVICE_DNS_NS, SOURCE_PRODUCT_OBAS, \
    SOURCE_PRODUCT_PLESK, MIGRATION_TYPE_FULL, SKIP_DNS_ZONE_FORWARDING

SourceHostZones = namedtuple('SourceHostZones', ('source_host', 'forwarder_impl', 'zones'))

logger = logging.getLogger(__name__)


class ConfigureSourceNameservers(ProductionSubscriptionAction):
    def run(self, global_context, subscription):
        dns_zone_manager = global_context.source_product.get_dns_zone_manager()
        platform_reader = PlatformReaderDB()
        dns_zones_by_source_host = {}
        subscription_dns_zones = dns_zone_manager.get_subscription_zones(subscription)
        subscription_zones = [dns_zone for dns_zone in subscription_dns_zones if not dns_zone.is_wildcard_dns_zone()]
        if not subscription_zones:
            logger.debug(_T('DEBUG_DNS_FORWARDING_CONFIGURING_SKIPPED_FOR_SUB').format(sub_name=subscription.name))
            return

        logger.debug(_T('DEBUG_CONFIGURE_DNS_FORWARDING_FOR_SUB_ZONES').format(
            sub_name=subscription.name,
            zones=u'\n\t'.join(unicode(dns_zone) for dns_zone in subscription_zones)
        ))

        subscription_zone_nameservers = {}
        # Disable DNS zones on the source
        for zone in subscription_zones:
            oa_nameservers = self._get_target_zone_nameservers(global_context, subscription, zone)
            if oa_nameservers:
                logger.debug(_T('DEBUG_FOUND_OA_NSES_FOR_SUB_ZONE').format(
                    zone_name=zone.name,
                    oa_nses=u'\n\t'.join(oa_nameservers)
                ))
                subscription_zone_nameservers[zone.name] = oa_nameservers
                logger.debug(_T('DISABLING_DNS_ZONE_ON_SOURCE_SYSTEM').format(zone_name=zone.name))
                dns_zone_manager.disable_zone_on_source(subscription, zone)
            else:
                logger.warning(_T('DEBUG_OA_NSES_FOR_SUB_ZONE_NOT_FOUND').format(zone_name=zone.name))

        subscription_dns_ns_services = get_services(subscription.source_info, SOURCE_SERVICE_DNS_NS)
        if not subscription_dns_ns_services:
            logger.warning(_T('DEBUG_SUB_DNS_NS_SERVICES_NOT_FOUND').format(sub_name=subscription.name))

        for service in subscription_dns_ns_services:
            if service.source_host_external_id not in dns_zones_by_source_host:
                source_host = platform_reader.get_source_host(service.source_host_external_id)
                dns_zones_by_source_host[source_host.external_id] = SourceHostZones(
                    source_host=source_host,
                    forwarder_impl=dns_zone_manager.get_dns_forwarding_impl(service),
                    zones={}
                )

            dns_zones_by_source_host[service.source_host_external_id].zones.update(subscription_zone_nameservers)

        logger.debug(_T('DEBUG_DNS_ZONES_TO_SOURCE_HOSTS_MAP').format(info=dns_zones_by_source_host))
        errors = {}
        # Setup DNS forwarding
        for source_host_zones in dns_zones_by_source_host.values():
            logger.info(_T('SETTING_DNS_FORWARDING_FOR_ZONES_ON_SOURCE_HOST').format(
                zones=source_host_zones.zones.keys(),
                host=source_host_zones.source_host.name
            ))
            errors.update(
                source_host_zones.forwarder_impl.add_forwarded_zones(source_host_zones.zones)
            )

        for zone, error in errors.iteritems():
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem('formatting-account', Problem.ERROR, error), None)
            )

    def filter_subscription(self, global_context, subscription):
        product = global_context.source_type
        return has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS_NS]) and \
               global_context.migration_mode == MigrationMode.PRODUCTION and \
               ((product == SOURCE_PRODUCT_PLESK and global_context.migration_type == MIGRATION_TYPE_FULL) or
               (product == SOURCE_PRODUCT_OBAS)) and \
               not SKIP_DNS_ZONE_FORWARDING

    def _get_target_zone_nameservers(self, global_context, subscription, zone):
        raise NotImplementedError()

    def get_failure_message(self, global_context, subscription):
        return _T("FAILED_TO_CONFIGURE_DNS_FORWARDING").format(name=subscription.name)

    def get_description(self):
        return _T("CONFIGURE_DNS_FORWARDING")
