# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
from contextlib import contextmanager

from migrator.core import _T, InternalError, MigrationError, SubscriptionAccountImportException
from migrator.core import \
    AccountImportException, SubscriptionImportException, ParentSubscriptionImportException, \
    MappingParentSubscriptionImportException, ChildSubscriptionImportException
from migrator.core.actions.base.common_action import CommonAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import SubscriptionMigrationStatus
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue
from migrator.core.utils.loggers.report import Problem

logger = logging.getLogger(__name__)


class MigratedObjectInfo(object):
    STATUS_SUCCESS = 'success'
    STATUS_FAIL = 'fail'
    STATUS_NOT_STARTED = 'not started'

    def __init__(self, status, **kwargs):
        """
        :param model: PlatformSubscription | AccountModel
        """
        super(MigratedObjectInfo, self).__init__(**kwargs)
        self.status = status

    def is_failed(self):
        return self.status == self.STATUS_FAIL

    def is_successful(self):
        return self.status == self.STATUS_SUCCESS

    def is_not_started(self):
        return self.status == self.STATUS_NOT_STARTED

    def set_failed(self):
        self.status = self.STATUS_FAIL

    def set_successful(self):
        self.status = self.STATUS_SUCCESS


class MigratedAccountInfo(MigratedObjectInfo):
    def __init__(self, model, status):
        super(MigratedAccountInfo, self).__init__(status)
        self.model = model
        self.account_id = None


class SourceMappingSubscriptionInfo(object):
    def __init__(self, model, mapping, **kwargs):
        super(SourceMappingSubscriptionInfo, self).__init__(**kwargs)
        self.model = model
        self.mapping = mapping

    def is_mapped_to_oa_subscription(self):
        return self.mapping.is_oa_subscription

    def is_mapped_to_other_subscription_in_migration_project(self, other_subscription=None):
        if other_subscription:
            # check that subscription is mapped to the specified subscription
            return \
                not self.is_mapped_to_oa_subscription() and \
                not self.model.external_id == other_subscription.model.external_id and \
                self.mapping.mapping_subscription_id == other_subscription.model.external_id
        else:
            # check that subscription is mapped to any other subscription
            return \
                not self.is_mapped_to_oa_subscription() and \
                self.mapping.mapping_subscription_id != self.model.external_id

    def is_mapped_to_self(self):
        return \
            not self.is_mapped_to_oa_subscription() and \
            self.mapping.mapping_subscription_id == self.model.external_id

    def is_parent_of(self, subscription):
        return subscription.model.parent_external_id == self.model.external_id

    def is_root(self):
        return self.model.parent_external_id is None and \
               not self.is_mapped_to_other_subscription_in_migration_project()


class MigratedSubscriptionInfo(MigratedObjectInfo, SourceMappingSubscriptionInfo):
    def __init__(self, model, mapping, progress, status):
        super(MigratedSubscriptionInfo, self).__init__(status=status, model=model, mapping=mapping)
        self.subscription_id = None
        self.progress = progress

    @trace
    def set_ids(self, subscription_id):
        self.subscription_id = subscription_id


class MigratedObjects(object):
    def __init__(self, subscriptions_to_migrate, progress):
        self.accounts = {}
        self.subscriptions = {}

        for platform_subscription in subscriptions_to_migrate:
            subscription_progress = progress.get_subscription(platform_subscription.mapping.id)
            source_subscription_info = platform_subscription.source_info
            self.subscriptions[source_subscription_info.external_id] = MigratedSubscriptionInfo(
                source_subscription_info, platform_subscription.mapping, subscription_progress, MigratedObjectInfo.STATUS_NOT_STARTED
            )
            if source_subscription_info.account.external_id not in self.accounts:
                self.accounts[source_subscription_info.account.external_id] = MigratedAccountInfo(
                    source_subscription_info.account, MigratedObjectInfo.STATUS_NOT_STARTED
                )

    def get_subscription(self, subscription_id):
        if subscription_id not in self.subscriptions:
            raise InternalError(_T("SUBSCRIPTION_WAS_NOT_FOUND_IN_MIGRATION_PROJECT").format(id=subscription_id))

        return self.subscriptions[subscription_id]

    def get_account(self, account_id):
        if account_id not in self.accounts:
            raise InternalError(_T("ACCOUNT_WAS_NOT_FOUND_IN_MIGRATION_PROJECT").format(id=account_id))

        return self.accounts[account_id]

    def get_finished_subscriptions_count(self):
        count = 0
        for subscription in self.subscriptions.itervalues():
            if subscription.is_successful() or subscription.is_failed():
                count += 1

        return count

    def get_total_subscriptions_count(self):
        return len(self.subscriptions)


class ImportObjects(CommonAction):
    def __init__(self):
        super(ImportObjects, self).__init__()
        self.import_api = None
        self.service_import_api = None

    def run(self, global_context):
        self.import_api = global_context.target_panel_obj.get_import_api(global_context)
        self.service_import_api = global_context.target_panel_obj.get_service_import_api(global_context)
        try:
            # subscriptions' statuses are reset in ResetMigrationProgress action
            migrated_objects = MigratedObjects(global_context.subscriptions_to_migrate, global_context.progress)
            for subscription in migrated_objects.subscriptions.itervalues():
                if subscription.is_root():
                    self._import_root_subscription(global_context, migrated_objects, subscription)
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(_T("ERROR_OCCURRED_DURING_THE_IMPORT").format(
                error=unicode(e)
            ))

    @trace
    def _import_root_subscription(self, global_context, migrated_objects, root_migrated_subscription):
        self._import_child_subscriptions_mapped_to_oa_subscription(
            global_context, migrated_objects, root_migrated_subscription
        )
        self._import_account_and_subscription(global_context, migrated_objects, root_migrated_subscription)
        self._import_subscriptions_mapped_to_subscription(global_context, migrated_objects, root_migrated_subscription)

    @trace
    def _import_child_subscriptions_mapped_to_themselves(self, global_context, migrated_objects, parent_subscription):
        for subscription in migrated_objects.subscriptions.itervalues():
            if (
                    parent_subscription.is_parent_of(subscription) and
                    subscription.is_mapped_to_self()
            ):
                self._import_account_and_subscription(global_context, migrated_objects, subscription)
                if not subscription.is_successful():
                    raise ChildSubscriptionImportException(subscription.model)

    @trace
    def _import_child_subscriptions_mapped_to_oa_subscription(
            self, global_context, migrated_objects, parent_subscription
    ):
        for subscription in migrated_objects.subscriptions.itervalues():
            if (
                    parent_subscription.is_parent_of(subscription) and
                    subscription.is_mapped_to_oa_subscription()
            ):
                self._import_account_and_subscription(global_context, migrated_objects, subscription)

    @trace
    def _import_subscriptions_mapped_to_subscription(self, global_context, migrated_objects, target_subscription):
        for subscription in migrated_objects.subscriptions.itervalues():
            if subscription.is_mapped_to_other_subscription_in_migration_project(target_subscription):
                self._import_account_and_subscription(global_context, migrated_objects, subscription)

    @trace
    def _import_account_and_subscription(self, global_context, migrated_objects, migrated_subscription):
        subscription = migrated_subscription.model
        try:
            # if the subscription has been already imported as a dependency for another subscription
            if migrated_subscription.is_failed() or migrated_subscription.is_successful():
                return

            self._import_child_subscriptions_mapped_to_themselves(
                global_context, migrated_objects, migrated_subscription
            )
            logger.info(_T("IMPORTING_THE_SUBSCRIPTION").format(
                sub_num=migrated_objects.get_finished_subscriptions_count() + 1,
                subs_total=migrated_objects.get_total_subscriptions_count(),
                sub_name=subscription.name,
                sub_id=subscription.external_id
            ))
            migrated_subscription.progress.status = SubscriptionMigrationStatus.IN_PROGRESS

            try:
                logger.info(_T("RUNNING_PRE_MIGRATION_SUBSCRIPTION_CHECKS"))
                self.import_api.pre_migration_subscription_check(migrated_subscription.model)
            except Exception as e:
                logger.info(_T("PRE_MIGRATION_SUBSCRIPTION_CHECKS_FAILED").format(reason=unicode(e)))
                logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                raise SubscriptionImportException(
                    _T('UNABLE_MIGRATE_SUBSCRIPTION_ERROR').format(
                        subscription_name=subscription.name,
                        error=unicode(e)
                    )
                )

            if migrated_subscription.is_mapped_to_oa_subscription():
                logger.debug(
                    "The subscription is mapped to the target subscription #%s" %
                    migrated_subscription.mapping.mapping_subscription_id
                )
                subscription_info = self.import_api.get_subscription(
                    subscription_id=migrated_subscription.mapping.mapping_subscription_id
                )
                migrated_subscription.set_ids(subscription_info.id)
            elif migrated_subscription.is_mapped_to_other_subscription_in_migration_project():
                logger.debug(
                    "The subscription is mapped to the subscription '%s' from the migration project" % subscription.name
                )
                target_migrated_subscription = migrated_objects.get_subscription(
                    migrated_subscription.mapping.mapping_subscription_id
                )
                # if the target subscription import has failed or was not started then skip the subscription with error
                if not target_migrated_subscription.is_successful():
                    raise MappingParentSubscriptionImportException(target_migrated_subscription.model)

                migrated_subscription.set_ids(target_migrated_subscription.subscription_id)
            else:
                # if the subscription is mapped to itself
                logger.debug("The subscription is not mapped to other subscription")
                if not subscription.in_production:
                    target_account = self._get_target_account(migrated_subscription, migrated_objects)
                    if isinstance(target_account, MigratedAccountInfo):
                        migrated_account = target_account
                        if migrated_account.is_failed():
                            raise SubscriptionAccountImportException(migrated_account.model)

                        if migrated_account.is_not_started():
                            migrated_account.account_id = self._import_account(
                                migrated_account, migrated_subscription.mapping.mapping_reseller_id
                            )
                            migrated_account.set_successful()
                            self._update_account_db_oa_id(migrated_account)

                        account_id = migrated_account.account_id
                    else:
                        account_id = target_account

                    subscription_id = self._import_subscription(migrated_subscription, account_id)
                    migrated_subscription.set_ids(subscription_id)
                else:
                    logger.info(_T("THE_SUBSCRIPTION_IS_IN_PRODUCTION_MODE_SKIP_UPDATING").format(
                        subscription_name=migrated_subscription.model.name,
                        subscription_oa_id=subscription.oa_id
                    ))
                    migrated_subscription.set_ids(subscription.oa_id)

            self._update_subscription_and_plan_db_oa_id(migrated_objects, migrated_subscription)
            if migrated_subscription.subscription_id is None:
                raise InternalError("The subscription hasn't been created in the target system.")
            else:
                self.import_api.check_target_subscription_status(migrated_subscription.subscription_id)

            if not migrated_subscription.model.in_production:
                self._import_services(global_context, migrated_subscription)
            else:
                logger.info(_T("THE_SUBSCRIPTION_IS_IN_PRODUCTION_MODE_SKIP_SERVICES").format(
                    subscription_name=migrated_subscription.model.name,
                    subscription_oa_id=migrated_subscription.subscription_id
                ))

            if global_context.migration_mode == MigrationMode.PRODUCTION:
                self._switch_to_production(global_context, migrated_subscription)

            migrated_subscription.set_successful()
        except SubscriptionAccountImportException as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            migrated_subscription.set_failed()
            error = _T("IMPORT_OF_THE_ACCOUNT_TO_WHICH_SUBSCRIPTION_BELONGS_FAILED").format(
                account_id=e.account.external_id,
                subscription_name=subscription.name
            )
            logger.error(error)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(
                    Problem('account-import', Problem.ERROR, error),
                    _T("RESOLVE_ACCOUNT_IMPORT_ISSUE_AND_RERUN").format(
                        account_id=e.account.external_id
                    )
                )
            )
            if isinstance(e, AccountImportException):
                migrated_account = migrated_objects.get_account(subscription.account.external_id)
                migrated_account.set_failed()
                global_context.execution_wrapper.register_error(
                    _T("FAILED_TO_CREATE_ACCOUNT").format(
                        external_id=e.account.external_id,
                        reason=unicode(e)
                    ),
                    None,
                    Problem.ERROR
                )

        except SubscriptionImportException as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            migrated_subscription.set_failed()
            error = _T("FAILED_TO_CREATE_SUBSCRIPTION").format(
                subscription_name=subscription.name,
                subscription_id=subscription.external_id,
                reason=unicode(e)
            )
            logger.error(error)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(
                    Problem('subscription-import', Problem.ERROR, unicode(e)),
                    None
                )
            )
        except ParentSubscriptionImportException as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            migrated_subscription.set_failed()
            error = ' '.join([
                _T("IMPORT_OF_PARENT_SUBSCRIPTION_FAILED").format(parent_subscription_name=e.subscription.name),
                _T("RESOLVE_DEPENDANT_SUBSCRIPTION_IMPORT_ISSUE_AND_RERUN").format(
                    dependant_subscription_name=e.subscription.name
                )
            ])
            logger.error(error)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(
                    Problem('subscription-import', Problem.ERROR, error),
                    ''
                )
            )
        except ChildSubscriptionImportException as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            migrated_subscription.set_failed()
            error = ' '.join([
                _T("IMPORT_OF_CHILD_SUBSCRIPTION_FAILED").format(child_subscription_name=e.subscription.name),
                _T("RESOLVE_DEPENDANT_SUBSCRIPTION_IMPORT_ISSUE_AND_RERUN").format(
                    dependant_subscription_name=e.subscription.name
                )
            ])
            logger.error(error)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(
                    Problem('subscription-import', Problem.ERROR, error),
                    ''
                )
            )
        except MappingParentSubscriptionImportException as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            migrated_subscription.set_failed()
            error = ' '.join([
                _T("IMPORT_OF_PARENT_MAPPING_SUBSCRIPTION_FAILED").format(parent_subscription_name=e.subscription.name),
                _T("RESOLVE_DEPENDANT_SUBSCRIPTION_IMPORT_ISSUE_AND_RERUN").format(
                    dependant_subscription_name=e.subscription.name
                )
            ])
            logger.error(error)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(
                    Problem('subscription-import', Problem.ERROR, error),
                    ''
                )
            )
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            migrated_subscription.set_failed()
            error = _T("UNEXPECTED_ERROR_DURING_MUGRATION_OF_SUBSCRIPTION").format(
                subscription_name=subscription.name,
                reason=unicode(e)
            )
            logger.error(error)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(
                    Problem('subscription-import', Problem.ERROR, error),
                    None
                )
            )
        finally:
            migrated_subscription.progress.set_finally_subscription_status(subscription.external_id)

    @trace
    def _get_target_account(self, migrated_subscription, migrated_objects):
        if migrated_subscription.model.parent_external_id is None:
            return migrated_objects.get_account(migrated_subscription.model.account.external_id)

        parent_subscription = migrated_objects.get_subscription(migrated_subscription.model.parent_external_id)
        if parent_subscription.is_mapped_to_oa_subscription():
            return self.import_api.get_subscription(
                subscription_id=parent_subscription.mapping.mapping_subscription_id
            ).account_id

        mapping_subscription = migrated_objects.get_subscription(parent_subscription.mapping.mapping_subscription_id)
        return migrated_objects.get_account(mapping_subscription.model.account.external_id)

    @trace
    def _import_account(self, migrated_account, mapping_reseller_id):
        try:
            logger.info(_T("CREATING_ACCOUNT").format(account_external_id=migrated_account.model.external_id))
            account_id = self.import_api.import_account(migrated_account.model, mapping_reseller_id)
            logger.info(_T("CREATED_ACCOUNT").format(account_oa_id=account_id))
            return account_id
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise AccountImportException(migrated_account.model, unicode(e))

    @trace
    def _import_subscription(self, migrated_subscription, account_id):
        try:
            logger.info(_T("CREATING_SUBSCRIPTION_UNDER_ACCOUNT_BY_SERVICE_PLAN").format(
                subscription_name=migrated_subscription.model.name,
                account_id=account_id,
                service_plan_id=migrated_subscription.mapping.plan_id
            ))
            subscription_id = self.import_api.import_subscription(
                migrated_subscription.model, account_id,
                migrated_subscription.mapping.plan_id
            )
            logger.info(_T("CREATED_SUBSCRIPTION").format(subscription_id=subscription_id))
            return subscription_id
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise SubscriptionImportException(unicode(e))

    @trace
    def _import_services(self, global_context, migrated_subscription):
        try:
            logger.info(_T("IMPORTING_SERVICES_FOR_SUBSCRIPTION").format(
                subscription_name=migrated_subscription.model.name,
                subscription_oa_id=migrated_subscription.subscription_id
            ))
            self.service_import_api.import_subscription_services(global_context, migrated_subscription.model)
            logger.info(_T("SERVICES_FOR_SUBSCRIPTION_HAVE_BEEN_IMPORTED").format(
                subscription_name=migrated_subscription.model.name,
                subscription_oa_id=migrated_subscription.subscription_id
            ))
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise SubscriptionImportException(unicode(e))

    @trace
    def _switch_to_production(self, global_context, migrated_subscription):
        try:
            logger.info(_T("SWITCHING_SUBSCRIPTION_TO_PRODUCTION_MODE").format(
                subscription_name=migrated_subscription.model.name,
                subscription_oa_id=migrated_subscription.subscription_id
            ))
            self._update_subscription_db_in_production(migrated_subscription)
            self.import_api.switch_subscription_to_production(migrated_subscription.model)
            self.service_import_api.switch_subscription_services_to_production(
                global_context, migrated_subscription.model
            )
            logger.info(_T("SUBSCRIPTION_HAS_BEEN_SWITCHED_TO_PRODUCTION").format(
                subscription_name=migrated_subscription.model.name,
                subscription_oa_id=migrated_subscription.subscription_id
            ))
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise SubscriptionImportException(unicode(e))

    @trace
    def _update_account_db_oa_id(self, migrated_account):
        platform_reader = PlatformReaderDB()
        account_db = platform_reader.get_account(migrated_account.model.external_id)
        account_db.oa_id = migrated_account.account_id
        platform_reader.session.commit()
        migrated_account.model = account_db

    @trace
    def _update_subscription_and_plan_db_oa_id(self, migrated_objects, migrated_subscription):
        # if the subscription is mapped to another subscription in migration project, use parent subscription's IDs
        if migrated_subscription.is_mapped_to_other_subscription_in_migration_project():
            sub = migrated_objects.get_subscription(migrated_subscription.mapping.mapping_subscription_id)
        else:
            sub = migrated_subscription

        if not migrated_subscription.is_mapped_to_oa_subscription():
            plan_id = sub.mapping.plan_id
        else:
            plan_id = self.import_api.get_subscription(migrated_subscription.mapping.mapping_subscription_id).service_plan_id

        with self._update_subscription_db_data(migrated_subscription) as subscription_db:
            subscription_db.oa_id = sub.subscription_id
            subscription_db.oa_plan_id = plan_id
            subscription_db.oa_plan_name = self.import_api.get_service_plan(plan_id).name

    @trace
    def _update_subscription_db_in_production(self, migrated_subscription):
        with self._update_subscription_db_data(migrated_subscription) as subscription_db:
            subscription_db.in_production = True

    @contextmanager
    def _update_subscription_db_data(self, migrated_subscription):
        platform_reader = PlatformReaderDB()
        subscription_db = platform_reader.get_subscription(migrated_subscription.model.external_id)
        yield subscription_db
        platform_reader.session.commit()
        migrated_subscription.model = subscription_db

    def get_failure_message(self, global_context):
        return _T("FAILED_TO_IMPORT_OBJECTS")

    def get_description(self):
        return _T("IMPORT_OBJECTS")
