# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator import settings
from migrator.core import InternalError
from migrator.core.language import _T
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


MAX_AMOUNT_OF_SUBSCRIPTIONS_TO_SHOW_IN_MSG = 5


class CheckAbstract(object):
    PHASE_PRE = 'pre'
    PHASE_RUNTIME = 'runtime'
    PHASE_POST = 'post'
    PHASE_NO_PHASE = 'no_phase'

    STATE_NEW = 'new'
    STATE_RUNNING = 'running'
    STATE_FAILED = 'failed'
    STATE_FINISHED = 'finished'

    STATUS_ERROR = 'error'
    STATUS_WARNING = 'warning'
    STATUS_SUCCESS = 'success'

    SUPPORTED_TARGET_SERVICE_CLASSES = []
    SUPPORTED_MIGRATION_TYPES = [settings.MIGRATION_TYPE_TAKEOVER, settings.MIGRATION_TYPE_FULL]
    SUPPORTED_SOURCE_PRODUCTS = []

    def __init__(self, name, phase=PHASE_NO_PHASE):
        self.name = name
        self.phase = phase
        self.code = 0
        self.meta = None
        self.source_type = Registry.get_instance().get_context().source_type
        logger.debug(_T("CHECK_CREATED").format(check_name=self.name))

    def set_result_code(self, code):
        self.code = code

    def get_result_code(self):
        return self.code

    def set_result_meta_data(self, meta_data):
        self.meta = meta_data

    def get_result_meta_data(self):
        return self.meta

    def get_formatted_response(self, status, message):
        if status == CheckAbstract.STATUS_SUCCESS:
            logger.info(message)
        elif status == CheckAbstract.STATUS_WARNING:
            logger.warning(message)
        elif status == CheckAbstract.STATUS_ERROR:
            logger.error(message)
        else:
            logger.debug(message)

        return {'status': status, 'message': message}

    def get_subscriptions_as_limited_string(self, subscriptions, names_amount=MAX_AMOUNT_OF_SUBSCRIPTIONS_TO_SHOW_IN_MSG):
        """
        :param subscriptions: list[string] Example: list[u'a.tld', u'b.org', u'c.com', u'd.com', u'e.com', u'f.com']
        :param names_amount: int Max amount of subscriptions names to return
        :return: string Example: '"a.tld", "b.org", "c.com", "d.com", "e.com" and 1 other', '"g.com", "j.com", "x.com"'
        """
        return get_subscriptions_as_limited_string(subscriptions, names_amount)

    def run(self):
        logger.debug(_T("RUN_CHECK").format(check_name=self.name, check_phase=self.phase))
        if not self.is_check_supported():
            error_message = 'Unable to run the check "%s" in the phase "%s": the check is not supported. Contact Odin technical support for assistance.' % (self.name, self.phase)
            logger.error(error_message)
            raise InternalError(error_message)

        status, message = self._do_check()
        return self.get_formatted_response(status, message)

    def _do_check(self):
        raise NotImplementedError()

    def is_check_supported(self):
        global_context = Registry.get_instance().get_context()
        config = global_context.config
        return \
            config.has_section('GLOBAL') and \
            config.has_option('GLOBAL', 'target-type') and \
            config.has_option('GLOBAL', 'source-type') and \
            config.has_option('GLOBAL', 'migration-type') and \
            config.get('GLOBAL', 'target-type').lower() in self.SUPPORTED_TARGET_SERVICE_CLASSES and \
            global_context.source_type in self.SUPPORTED_SOURCE_PRODUCTS and \
            config.get('GLOBAL', 'migration-type').lower() in self.SUPPORTED_MIGRATION_TYPES

    def get_full_description(self):
        raise NotImplementedError()


class CheckSourceAbstract(CheckAbstract):
    def __init__(self, name, phase=CheckAbstract.PHASE_NO_PHASE):
        super(CheckSourceAbstract, self).__init__(name, phase)

    def _do_check(self):
        if self.source_type not in settings.SUPPORTED_SOURCE_PRODUCTS:
            logger.error(_T("SOURCE_PANEL_TYPE_NOT_SUPPORTED_BY_CHECK").format(source_product_type=self.source_type))
            raise NotImplementedError()

        logger.debug(_T("DO_CHECK_FOR_KNOWN_PRODUCT").format(source_product_type=self.source_type))
        class_file, class_name = self.get_source_check_class().rsplit('.', 1)
        class_path = 'migrator.products.{0}.checks.{1}'.format(self.source_type, class_file)
        try:
            command_implementation_class = __import__(class_path, fromlist=[class_name]).__dict__[class_name]
            check_implementation = command_implementation_class()
        except Exception:
            error_message = _T("UNABLE_TO_RUN_SOURCE_PANEL_CHECK").format(check_name=self.name)
            logger.error(error_message)
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return CheckAbstract.STATUS_ERROR, error_message

        return check_implementation.run()

    def get_source_check_class(self):
        raise NotImplementedError()


def get_subscriptions_as_limited_string(subscriptions_names, names_amount=MAX_AMOUNT_OF_SUBSCRIPTIONS_TO_SHOW_IN_MSG):
    """
    :param subscriptions_names: list[string] Example: list[u'a.tld', u'b.org', u'c.com', u'd.com', u'e.com', u'f.com']
    :param names_amount: int Max amount of subscriptions names to return
    :return: string Example: '"a.tld", "b.org", "c.com", "d.com", "e.com" and 1 other', '"g.com", "j.com", "x.com"'
    """
    if len(subscriptions_names) > names_amount:
        return _T('CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_SUBS_LIMITED_LIST').format(
            subs_names=', '.join(
                "'%s'" % sub_name for sub_name in subscriptions_names[:names_amount]),
            subs_counter=len(subscriptions_names) - names_amount
        )
    else:
        return ', '.join("'%s'" % sub_name for sub_name in subscriptions_names)
