# (C)2017 Ingram Micro Inc. all Rights Reserved
"""
Common functions for command line interface of:
- migration between different panel instances
- move subscriptions between nodes within the same panel instance
"""
import logging
import os.path
import sys

from migrator.core.cli.commands import MainCommand
from migrator.core.cli.commands.command_builder import MigratorCommandBuilder
from migrator.core.language import _T
from migrator.core.utils import is_empty
from migrator.core.utils.loggers.log import Log
from migrator.core.utils.registry import Registry

logger = logging.getLogger('migrator')


def run_migrator(base_dir, var_dir, execution_path):
    """Read CLI options, read config file, run migrator and process errors."""
    # Determine panel migrator base directory and store it in the registry.
    registry = Registry.get_instance()
    registry.set_base_dir(base_dir)
    registry.set_var_dir(var_dir)
    registry.set_execution_path(execution_path)

    _set_up_encoding()
    try:
        _configure_logging()
        return MigratorCommandBuilder().build_commands(MainCommand).main(standalone_mode=False)
    except Exception as e:
        context = getattr(e, 'context', '')
        context_msg = _get_context_msg(context)
        message = _T("INTERNAL_MIGRATOR_ERROR_S_MIGRATION_IS") % e + context_msg
        if not logger.handlers:
            logging.exception(message)
        else:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            logger.error(message)
        return 1
    finally:
        logging.shutdown()


def _get_context_msg(context):
    if not is_empty(context):
        context_msg = u"\n%s" % _T("EXCEPTION_CONTEXT").format(context=context)
    else:
        context_msg = u''

    return context_msg


def _set_up_encoding():
    """Set up encoding to 'UTF-8', so python does not fail with "UnicodeEncodeError: 'ascii' codec can't encode
    characters in position ..." even in cases:
    1) if we do not use terminal (redirect output to pipe, like 'python migrator config.ini check | less')
    2) if we use some encoding, that do not support all range of symbols we output (for example, 'C' and any cyrillic
    symbol)
    Also there is another way described at
    http://stackoverflow.com/questions/1473577/writing-unicode-strings-via-sys-stdout-in-python
    but in case #2 it could make migration stop (while in the current solution we continue migration,
    but output "bad" symbols)
    """
    import codecs
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout)


class CommandResponse(dict):
    def __init__(self, exit_code=0, message='', details=None):
        self['exit_code'] = exit_code
        self['message'] = message
        self['details'] = details

    def update(self, exit_code=None, message=None, details=None):
        if exit_code is not None:
            self['exit_code'] = exit_code
        if message is not None:
            self['message'] = message
        if details is not None:
            self['details'] = details


def _configure_logging():
    """Configure logger using a user-provided or a default config file."""
    registry = Registry.get_instance()
    log = registry.set_log(Log())
    logging_config = os.path.join(registry.get_base_dir(), 'conf/oa-migrator.yaml')
    log.configure(logging_config)
