# (C)2017 Ingram Micro Inc. all Rights Reserved
from click import Group, Command, Argument, Option, Choice, get_current_context

from migrator.core.cli.commands import commands
from migrator.core.cli.commands.declaration_base import MigratorArgumentDeclaration, MigratorOptionDeclaration


def _run_command_implementation(_command_implementation_class_name, **kwargs):
    class_name_parts = _command_implementation_class_name.split('.')
    class_name = class_name_parts[-1]

    module_name = MigratorCommandBuilder.IMPLEMENTATIONS_MODULE + '.' + '.'.join(class_name_parts[:-1])
    command_implementation_class = __import__(module_name, fromlist=[class_name]).__dict__[class_name]

    _get_args_from_parent_command(get_current_context(), kwargs)
    command_implementation = command_implementation_class()
    return command_implementation.run(OptionsDictProxy(kwargs))


def _get_args_from_parent_command(context, params):
    if context.parent:
        params.update(context.parent.params)
        _get_args_from_parent_command(context.parent, params)


class OptionsDictProxy(object):
    def __init__(self, dict):
        self._dict = dict

    def __contains__(self, item):
        return item in self._dict

    def __getattr__(self, item):
        if item in self._dict:
            return self._dict[item]
        else:
            raise Exception("Command line option '{name}' does not exist".format(name=item))


class MigratorCommandBuilder(object):
    IMPLEMENTATIONS_MODULE = 'migrator.core.cli.commands.implementation'
    CONTEXT_SETTINGS = {
        'help_option_names': ['-h', '--help']
    }

    def build_commands(self, root_command_class):
        if root_command_class not in commands:
            raise Exception("Root command class is not in the list of commands: {root}".format(
                root=root_command_class.__name__
            ))

        all_commands = {}

        for command_class in commands:
            command_declaration = command_class()
            all_commands[command_class.__name__] = command_declaration

        for command_class in all_commands:
            command_declaration = all_commands[command_class]
            if command_declaration.parent_command_class is not None:
                all_commands[command_declaration.parent_command_class.__name__].commands.append(command_declaration)

        return self._build_command(all_commands[root_command_class.__name__], all_commands)

    def _build_command(self, command_declaration, all_commands):
        if len(command_declaration.commands) > 0:
            return self._build_group_command(command_declaration, all_commands)
        else:
            return self._build_single_command(command_declaration)

    def _build_group_command(self, command_declaration, all_commands):
        subcommands = {}

        for subcommand_object in command_declaration.commands:
            subcommand = self._build_command(subcommand_object, all_commands)

            if subcommand.name in subcommands:
                raise Exception("Two subcommands with identical names in command '{command}'".format(
                    command=command_declaration.__class__.__name__
                ))
            else:
                subcommands[subcommand.name] = subcommand

        return Group(
            command_declaration.name, help=command_declaration.help,
            params=self._build_command_params(command_declaration.params), commands=subcommands,
            context_settings=self.CONTEXT_SETTINGS
        )

    def _build_single_command(self, command_declaration):
        return Command(
            command_declaration.name, help=command_declaration.help,
            params=self._build_command_params(command_declaration.params),
            callback=lambda **kwargs: _run_command_implementation(
                command_declaration.implementation_class_name, **kwargs
            ),
            context_settings=self.CONTEXT_SETTINGS
        )

    def _build_command_params(self, param_classes):
        params = []
        for param_class in param_classes:
            params.append(self._build_param(param_class()))

        return params

    def _build_param(self, param_declaration):
        if isinstance(param_declaration, MigratorArgumentDeclaration):
            return Argument([param_declaration.name], type=self._build_param_type(param_declaration.type))
        elif isinstance(param_declaration, MigratorOptionDeclaration):
            return Option(
                [param_declaration.name], help=param_declaration.help,
                type=self._build_param_type(param_declaration.type), is_flag=param_declaration.is_flag,
                default=param_declaration.default, required=param_declaration.required
            )
        else:
            raise Exception("Unsupported option or argument declaration type: {class_name}".format(
                class_name=param_declaration.__class__.__name__
            ))

    def _build_param_type(self, type):
        if isinstance(type, list):
            return Choice(type)
        else:
            return type
