# (C)2017 Ingram Micro Inc. all Rights Reserved
import json
import logging

from migrator.core import MigrationError
from migrator.core.cli.commands.implementation_base import handle_exceptions, MigratorCommandImplementation
from migrator.core.language import _T
from migrator.core.main import GenericMigrator
from migrator.core.product.lib.host import get_panel_info_instance
from migrator.core.product.panel import UnknownPanelInfo
from migrator.core.utils.filetypes.config_utils import ConfigHostSection
from migrator.core.utils.networking.run_command import WinexeExecuteError
from migrator.core.utils.networking.runner import get_runner
from migrator.core.utils.operating_system.os_info import get_os_info

logger = logging.getLogger(__name__)


class GetSystemInfoImplementation(MigratorCommandImplementation):
    @handle_exceptions
    def run(self, options):
        migrator = GenericMigrator(options)
        try:
            system_info = self._get_system_info(migrator.global_context, options)
        except WinexeExecuteError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            system_info = {
                'exit_code': 1,
                'error_message': ' '.join([_T("MIGRATION_MANAGER_COULD_NOT_RETRIEVE").format(host=e.host), e.message])
            }
        except MigrationError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            system_info = {'exit_code': 1, 'error_message': e.message}
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise e

        system_info_string = json.dumps(system_info)
        logger.debug(system_info_string)
        print(system_info_string)

    def _get_system_info(self, global_context, options):
        config_section = ConfigHostSection(global_context.config, 'generic')
        runner = get_runner(config_section)
        try:
            os_info = get_os_info(runner)
            panel_info = get_panel_info_instance(runner, config_section)
        except MigrationError as e:
            return dict(exit_code=1, error_message=e.message)

        response = self._get_response_structure(os_info, panel_info, options.skip_check_services)

        if isinstance(panel_info, UnknownPanelInfo):
            # Do not report command error if no supported product was found.
            # Just return available info and set proper error_message.
            error_message = _T("UNABLE_TO_FIND_SUPPORTED_PANEL").format(
                host_description=runner.host_description, supported_panels='OBAS, Plesk'
            )
            logger.debug(error_message)
            response['exit_code'] = 2
            response['error_message'] = error_message

        return response

    @staticmethod
    def _get_response_structure(os_info, panel_info, skip_check_services):
        if skip_check_services:
            services = panel_info.get_service_list()
            # remove internal specifics from services list
            for k in services.iterkeys():
                services[k] = dict()
        else:
            services = panel_info.services

        return {
            'os': {
                'platform': os_info.platform,
                'arch': os_info.arch,
                'name': os_info.name,
                'version': os_info.version
            },
            'panel': {
                'name': panel_info.name,
                'short_name': panel_info.short_name,
                'version': panel_info.version,
                'patch': panel_info.patch,
                'url': panel_info.url,
            },
            'services': services,
            'exit_code': 0
        }
