# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationConfigurationFileError
from migrator.core.main import GenericMigrator
from migrator.core.cli.panels import get_product_name, load_source_panel_class, list_source_panels, list_target_panels
from migrator.core.language import _T
from migrator.core.utils import format_list
from migrator.core.utils.filetypes.config_utils import read_config

logger = logging.getLogger(__name__)


def create_migrator(options, migration_data_path=None):
    """Create migrator object"""
    config_path = options.config
    if config_path is None:
        migrator_class = GenericMigrator
    else:
        config = read_config(config_path)
        if migration_data_path is not None:
            config.set('GLOBAL', 'migration-data', value=migration_data_path)

        if _should_create_generic_migrator(config):
            migrator_class = GenericMigrator
        else:
            source_type = config.get('GLOBAL', 'source-type')
            _verify_source_panel_type(source_type)

            target_type = config.get('GLOBAL', 'target-type')
            product_name = get_product_name()
            _verify_target_panel_type(target_type, product_name)

            migrator_class = load_source_panel_class(source_type, 'main', 'Migrator')

    logger.debug(_T("MIGRATOR_OBJECT_WILL_INITIALIZE") % migrator_class)
    migrator_object = migrator_class(options)

    return migrator_object


def _should_create_generic_migrator(config):
    """Determine, if the given config.ini can be used for initialization of GenericMigrator.

    If the provided config.ini has not enough information for running the
    actual migration (when, for example, migrator is called for executing a
    supplementary command), we should initialize a 'generic' scaled-down
    version of migrator.

    Return 'true', if some information is not available from config.ini and
    we need to initialize a GenericMigrator object.
    """
    has_no_source_type = not config.has_option('GLOBAL', 'source-type')
    has_no_source_servers = not config.has_option('GLOBAL', 'source-servers')
    source_servers = None if has_no_source_servers else config.get('GLOBAL', 'source-servers')
    return has_no_source_type or source_servers and 'generic' in source_servers.split()


def _verify_source_panel_type(source_type):
    """
    :param source_type: string Examples: 'plesk', 'obas', 'oa.wsh'
    :return: void
    """
    source_panels = list_source_panels()
    if source_type not in source_panels:
        raise MigrationConfigurationFileError(
            _T("SOURCE_TYPE_NOT_SUPPORTED").format(source_type=source_type, supported_types=format_list(source_panels))
        )


def _verify_target_panel_type(target_type, product_name):
    target_panels = list_target_panels(product_name)
    if target_type not in target_panels:
        raise MigrationConfigurationFileError(
            _T("TARGET_TYPE_NOT_SUPPORTED").format(target_type=target_type, supported_types=format_list(target_panels))
        )
