# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.connections.database_server import ExternalDatabaseServer
from migrator.core.connections.source_server import SourceServer
from migrator.core.session.config import read_external_db_settings
from migrator.core.utils import cached, group_value_by_id
from migrator.core.utils.filetypes import config_utils


class Connections(object):
    def __init__(self, global_context, target_panel):
        self._config = global_context.config
        if global_context.migrator.can_initialize_target_server():
            self.target = target_panel.get_connections(global_context)

        self._external_db_servers = self._read_external_db_settings_from_config(
            global_context.config, 'external-db-servers'
        )

    @cached
    def get_source_server_object(self, server_id):
        """@moveto SourcePanelExporterBase"""
        return SourceServer(server_id, self.get_information_servers()[server_id])

    def get_information_servers(self):
        """Get servers which could provide info about subscriptions and customers"""
        raise NotImplementedError()

    def get_external_db_servers(self):
        """Get information about source external database servers

        Returns dictionary {server_id: server_settings} with connection
        information.

        :rtype: dict[str, migrator.core.migrator_config.ExternalDBConfigBase]
        """
        return self._external_db_servers

    def get_external_database_server(self, panel_server, db_type, host, port, user, password):
        external_server_id = self.get_external_db_server_id(db_type, host)
        if external_server_id is not None:
            settings = self.get_external_db_servers()[external_server_id]
            return ExternalDatabaseServer(
                settings=settings,
                panel_server=panel_server,
                port=port, user=user, password=password,
                physical_server=self.get_external_db_physical_server(external_server_id),
            )
        else:
            return None

    def get_external_db_server_id(self, db_type, host):
        """Get external database server ID by database type and host

        Return external database server ID - name of a section in configuration file which describes this
        external database server. If no server with such database type and host was found - return None.

        :param str db_type: type of database server, 'mysql' or 'mssql' or 'postgresql'
        :param: str host: hostname of database server, for example '192.168.1.15', or '10.52.1.55\MSSQLSERVER2008'
        :rtype: str | None
        """
        external_db_servers_by_host = group_value_by_id(
            self._external_db_servers.items(),
            key_func=lambda s: (s[1].db_type, s[1].host),
            value_func=lambda s: s[0]
        )
        return external_db_servers_by_host.get((db_type, host))

    @staticmethod
    def _read_external_db_settings_from_config(config, list_name):
        """
        :rtype: dict[str, migrator.core.migrator_config.ExternalDBConfigBase]
        """
        sections = config_utils.get_sections_list(config, 'GLOBAL', list_name)
        return {
            section_name: read_external_db_settings(config, section_name)
            for section_name in sections
        }
