# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import posixpath

from contextlib import contextmanager
from threading import Lock

import migrator.core.utils.networking.rsync
import migrator.products.plesk.lib.api.xmlrpc as xmlrpc

from migrator.core import MigrationNoRepeatError, CommandExecutionError
from migrator.core.language import _T
from migrator.core.connections.unix_server import UnixServer
from migrator.core.product.lib.jsonrpc.client import JsonRpcClient
from migrator.core.utils import cached
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.filetypes.ssl_keys import get_oam_ssl_certificate
from migrator.core.utils.operating_system import windows_utils
from migrator.core.utils.operating_system.windows_utils import get_from_registry, path_join as windows_path_join
from migrator.products.plesk.target.utils import plesk_utils
from migrator.products.plesk.lib.agent import install_agent_on_server
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.panel import get_linux_admin_password, get_windows_admin_password
from migrator.products.plesk.verification.connection import PleskServerConnectionChecker
from migrator.settings import AGENT_PORT, PLESK_PANEL_DEFAULT_PORT

logger = logging.getLogger(__name__)


class PleskServer(UnixServer):
    def __init__(self):
        # If migrating several subscriptions at once, errors might occur if you run a number of Plesk utilities
        # in parallel. For this reason, it is better to avoid this situation and prohibit the parallel execution
        # in multiple threads.
        self.websrvmng_lock = Lock()
        self._skeymng = None

    @contextmanager
    def runner(self):
        raise NotImplementedError()

    def description(self):
        raise NotImplementedError()

    def is_windows(self):
        raise NotImplementedError()

    def plesk_api(self):
        raise NotImplementedError()

    def _plesk_api(self, ip, api_version):
        """Get Plesk API client object for this server"""

        # Here we create a client with minimum supported version. We further make a request to find
        # out Plesk panel version, and update XML API version of the API client accordingly.
        username, password, secret_key = self.api_credentials(ip)

        client = xmlrpc.Client(
            'https://{ip}:{port}/enterprise/control/agent.php'.format(ip=ip, port=PLESK_PANEL_DEFAULT_PORT),
            xmlrpc.api_versions[api_version],
            username=username,
            password=password,
            secret_key=secret_key,
            pretty_print=True
        )

        PleskServerConnectionChecker.check_plesk_api(client, self)

        # Get Plesk version.
        logger.debug(_T("LOG_GET_SERVER_INFO"), ip)
        plesk_version = '.'.join(map(str, self.plesk_version))
        client.api_version = xmlrpc.get_max_supported_api_version(plesk_version)
        logger.debug(_T("LOG_PLESK_SERVER_INFO"), plesk_version, client.api_version)

        return client

    @cached
    @trace
    def _agent_api(self, ip_address):
        install_agent_on_server(self)

        url = 'https://{host}:{port}'.format(host=ip_address, port=AGENT_PORT)
        oam_cert = get_oam_ssl_certificate()
        ssl_info = {
            'cert_file': oam_cert.cert_filename,
            'key_file': oam_cert.pkey_filename,
        }

        return JsonRpcClient(url, ssl_info)

    @cached
    @trace
    def api_credentials(self, ip_address):
        with self.runner() as runner:
            if self.plesk_version >= (12, 0):
                if self._skeymng is None:
                    self._skeymng = xmlrpc.PleskAPISecretKeysManager(runner, self.is_windows(), self.plesk_dir)

                connection_info = self._agent_api(ip_address).common.connection_info()
                source_ip = connection_info['ip']
                if source_ip not in self._skeymng.ips:
                    self._skeymng.add_key(source_ip, 'OAM')

                return None, None, self._skeymng.ips[source_ip][0]
            else:
                if self.is_windows():
                    admin_password = get_windows_admin_password(self.plesk_version, self.plesk_dir, runner)
                else:
                    admin_password = get_linux_admin_password(self.plesk_version, self.plesk_dir, runner)

                return 'admin', admin_password, None

    @property
    @cached
    @trace
    def vhosts_dir(self):
        if self.is_windows():
            get_vhosts_dir_function = plesk_utils.get_windows_vhosts_dir
        else:
            get_vhosts_dir_function = plesk_utils.get_unix_vhosts_dir

        with self.runner() as runner:
            return get_vhosts_dir_function(runner)

    @trace
    def get_vhost_dir(self, domain):
        """Get path to virtual host's directory of domain"""
        vhost_name = domain.encode('idna')
        if self.is_windows():
            return windows_utils.path_join(self.vhosts_dir, vhost_name)
        else:
            return posixpath.join(self.vhosts_dir, vhost_name)

    @property
    @cached
    @trace
    def plesk_dir(self):
        if self.is_windows():
            with self.runner() as runner:
                return plesk_utils.get_windows_plesk_dir(runner)
        else:
            with self.runner() as runner:
                return plesk_utils.get_unix_product_root_dir(runner)

    @property
    @cached
    @trace
    def websrvmng_bin(self):
        """Get full path to websrvmng Plesk utility

        :rtype: unicode
        """
        if not self.is_windows():
            raise NotImplementedError()

        if self.plesk_version >= (12, 0) and self.is_64_bit_os:
            return ur'{plesk_dir}\admin\bin64\websrvmng'.format(plesk_dir=self.plesk_dir)
        else:
            return ur'{plesk_dir}\admin\bin\websrvmng'.format(plesk_dir=self.plesk_dir)

    @property
    @cached
    @trace
    def is_64_bit_os(self):
        """Check if this is 64-bit OS or 32-bit one

        :rtype: bool
        """
        if not self.is_windows():
            raise NotImplementedError()

        try:
            with self.runner() as runner:
                # Get architecture from registry.
                # Don't try to get it from environment variable directly, as migrator is running 32-bit version
                # of Python, and Windows sets environment variable accordingly, and you always get 'x86'.
                # Also, avoid using 'wmic os get osarchitecture', as it does not work on Windows 2003.
                return '64' in get_from_registry(
                    runner, [r'HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Session Manager\Environment'],
                    'PROCESSOR_ARCHITECTURE'
                )
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            logger.error(_T("FAILED_TO_DETECT_ARCH").format(server=self.description(), reason=unicode(e)))
            return True

    @property
    @cached
    @trace
    def mail_dir(self):
        if self.is_windows():
            raise NotImplementedError()
        else:
            with self.runner() as runner:
                return plesk_utils.get_unix_mailnames_dir(runner)

    @property
    @cached
    @trace
    def plesk_version(self):
        """Return a tuple with Plesk server version."""
        with self.runner() as runner:
            if self.is_windows():
                return plesk_utils.get_plesk_for_windows_version(runner, self.plesk_dir)
            else:
                return plesk_utils.get_plesk_for_unix_version(runner, self.plesk_dir)

    @property
    @cached
    @trace
    def rsync_bin(self):
        with self.runner() as runner:
            return migrator.core.utils.networking.rsync.detect_rsync_path(
                runner, self.description()
            )

    @trace
    def get_bin_util_path(self, name):
        if self.is_windows():
            return u'%s\\bin\\%s' % (self.plesk_dir, name)
        else:
            return u'%s/bin/%s' % (self.plesk_dir, name)

    @cached
    @trace
    def get_db_servers(self):
        return {
            result.data.id: result.data
            for result in self.plesk_api().send(plesk_ops.DbServerOperator.Get(plesk_ops.DbServerOperator.FilterAll()))
        }

    @trace
    def get_database_server(self, server):
        db_server_hosts = [x.lower() for x in server.alt_hosts] + [server.host.lower()]

        for database_server in self.get_db_servers().itervalues():
            if database_server.host.lower() in db_server_hosts and database_server.dbtype == server.dbtype:
                return database_server

        return None

    @trace
    def get_supported_db_types(self):
        if self.is_windows():
            return set(['mysql', 'mssql'])
        else:
            return set(['mysql', 'postgresql'])

    @cached
    @trace
    def get_path_to_mysqldump(self):
        if self.is_windows():
            return windows_path_join(self._get_path_to_mysql_bin_windows(), 'mysqldump')
        else:
            return 'mysqldump'

    @cached
    @trace
    def get_path_to_mysql(self):
        if self.is_windows():
            return windows_path_join(self._get_path_to_mysql_bin_windows(), 'mysql')
        else:
            return 'mysql'

    @trace
    def _get_path_to_mysql_bin_windows(self):
        return windows_path_join(self.plesk_dir, 'MySQL', 'bin')

    @cached
    @trace
    def get_path_to_sqlcmd(self):
        sqlcmd_file = 'sqlcmd.exe'
        sqlcmd_path = sqlcmd_file
        with self.runner() as runner:
            for value in ['ODBCToolsPath', 'Path']:
                path = get_from_registry(
                    runner, [r'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Microsoft SQL Server'], value,
                    recursively=True, bits=64 if self.is_64_bit_os else 32
                )
                if path:
                    sqlcmd_path = windows_path_join(path, sqlcmd_path)
                    break

            try:
                runner.sh("%s -?" % sqlcmd_path)
            except CommandExecutionError:
                raise MigrationNoRepeatError(_T('UNABLE_START_DB_COMMAND_LINE_CLIENT').format(
                    db_client=sqlcmd_file, server=self.description(), DBMS='Microsoft SQL Server'
                ))

        return sqlcmd_path

    @property
    @cached
    @trace
    def plesk_components(self):
        """Return components installed on Plesk

        Returns dictionary with key - component name, value - information about
        the component.

        Example of result : {
            'mailman': ServerComponent(
                name='mailman',
                version='2.1.9-6.el5_6.1'
            ),
            'psa-spamassassin': ServerComponent(
                name='psa-spamassassin',
                version='10.13.4-cos5.build1013111102.18'
            )
        }
        """
        logger.debug(u"Get installed Plesk components")
        request = plesk_ops.ServerOperator.Get([plesk_ops.ServerOperator.Dataset.COMPONENTS, ])
        response = self.plesk_api().send(request)
        server_info = response.data
        logger.debug(u"Server info: %s", server_info)
        return server_info.components

