# (C)2017 Ingram Micro Inc. all Rights Reserved
import ntpath
from contextlib import contextmanager

import migrator.core.utils.networking.rsync

from migrator.settings import AGENT_PORT
from migrator.core.connections.database_server import SourceDatabaseServer
from migrator.core.connections.unix_server import UnixServer
from migrator.core.product.lib.jsonrpc.client import JsonRpcClient
from migrator.core.language import _T
from migrator.core.session import session_dir
from migrator.core.utils import cached
from migrator.core.utils.filetypes.ssl_keys import get_oam_ssl_certificate
from migrator.core.utils.networking import run_command
from migrator.core.utils.networking.ssh.connection_pool import SSHConnectionPool
from migrator.products.plesk.lib.agent import install_agent_on_server


class SourceServer(UnixServer):
    """This is a class to interact with source server
    and get various common information about the source server.
    """

    def __init__(
        self, node_id, node_settings, migrator_server=None, winexe_runas=False
    ):
        """Object constructor.

        Args:
            node_id: server's section name specified in config.ini
            node_settings: server properties specified in config.ini
            migrator_server: server where migration tools are installed,
                do not pass if you don't need local execution of commands
                on Windows server
            winexe_runas: boolean, if True then use "run as" option when
                running commands with winexe
        """
        super(SourceServer, self).__init__()

        self.node_id = node_id
        self.node_settings = node_settings
        self._migrator_server = migrator_server
        if not node_settings.is_windows:
            self._ssh = SSHConnectionPool.get_instance().get(self.node_settings, self.description())
        self._winexe_runas = winexe_runas

        if self.node_settings.is_windows:
            self._session_dir = session_dir.WindowsSessionDir(self.runner, self.node_settings.session_dir)
        else:
            self._session_dir = session_dir.UnixSessionDir(self.runner, self.node_settings.session_dir)

    @contextmanager
    def runner(self):
        """Get runner object to execute commands on the source server"""
        if self.is_windows():
            yield self._get_windows_runner()
        else:
            with self._ssh.runner() as runner:
                yield runner

    @cached
    def _get_windows_runner(self):
        if self._migrator_server is not None:
            local_temp_dir = self._migrator_server.session_dir()
        else:
            local_temp_dir = None

        return run_command.WinexeRunner(
            self.node_settings,
            local_temp_dir=local_temp_dir,
            runas=self._winexe_runas
        )

    def is_windows(self):
        """Return True if source server is Windows server,
        False if source server is Unix server"""
        return self.node_settings.is_windows

    def description(self):
        """Return brief server description which should help end customer
        to identify server and which should be used in all messages"""
        return _T("SOURCE_SERVER_DESCRIPTION_1") % (self.node_id, self.ip())

    def ip(self):
        """Return main IP address of the server"""
        return self.node_settings.ip

    def user(self):
        """Return user we use to communicate with the server"""
        if not self.is_windows():
            return self.node_settings.ssh_auth.username
        else:
            return self.node_settings.windows_auth.username

    @property
    def mail_settings(self):
        """Return copy mail content configuration"""
        return self.node_settings.mail_settings

    def settings(self):
        return self.node_settings

    def get_database_server(self, model):
        """Get source database server

        :type model: parallels.core.dump.data_model.DatabaseServer
        """
        return SourceDatabaseServer(self.node_settings, model, self)

    def get_session_file_path(self, filename):
        """Return full path to a temporary file on the server by
        short file name. Function considers session directory settings
        and ensures that session directory exists on the server"""
        return self._session_dir.get_file_path(filename)

    def session_dir(self):
        """Return path to session directory - directory where
        we should store all temporary files"""
        return self._session_dir.get_session_dir_path()

    @property
    @cached
    def rsync_bin(self):
        with self.runner() as runner:
            return migrator.core.utils.networking.rsync.detect_rsync_path(runner, self.description())

    @property
    def python_bin(self):
        if self.is_windows():
            # use Python from panel migrator agent
            return ntpath.join(self.node_settings.agent_settings.agent_path, 'python\oa-migrator-python.exe')
        else:
            # use default OS Python
            return 'python'

    @cached
    def agent_api(self):
        install_agent_on_server(self)

        url = 'https://{host}:{port}'.format(host=self.ip(), port=AGENT_PORT)
        oam_cert = get_oam_ssl_certificate()
        ssl_info = {
            'cert_file': oam_cert.cert_filename,
            'key_file': oam_cert.pkey_filename,
        }

        return JsonRpcClient(url, ssl_info)

    def __repr__(self):
        return 'SourceServer(%r, %s)' % (self.node_id, self.node_settings.ip)

    def __hash__(self):
        """Implemented for using objects as dictionary keys,
        for the purpose of grouping"""
        return hash(self.node_id)

    def __eq__(self, another):
        """Implement for using objects as dictionary keys,
        for the purpose of grouping"""
        return isinstance(another, SourceServer) and self.node_id == another.node_id
