# (C)2017 Ingram Micro Inc. all Rights Reserved
from contextlib import contextmanager

from migrator.core.data.format import PagedReader


class PlatformReader(object):
    """Read source system information at once and construct all child objects"""

    def get_accounts_owned_by(self, owner_id):
        """

        :param owner_id: str
        :return: list[migrator.core.data.model.account.AccountModel]
        """
        raise NotImplementedError()

    def get_account(self, account_id):
        """

        :param account_id: str
        :return: migrator.core.data.model.account.AccountModel
        """
        raise NotImplementedError()

    def get_account_list(self, account_id_list):
        """

        :param account_id_list: list of AccountOwnership objects
        :return: a list of migrator.core.data.model.account.AccountModel
        """
        raise NotImplementedError()

    def get_reseller_list(self):
        """

        :return: list[migrator.core.data.model.reseller.ResellerModel]
        """
        raise NotImplementedError()

    def get_reseller(self, reseller_id):
        """

        :param reseller_id: str
        :return: migrator.core.data.model.reseller.ResellerModel
        """
        raise NotImplementedError()

    def get_subscriptions_list(self, owner_id=None):
        """

        :param owner_id: str
        :return: list[migrator.core.data.model.subscription.SubscriptionModel]
        """
        raise NotImplementedError()

    def get_subscription(self, subscription_id):
        """

        :param subscription_id: str
        :return: migrator.core.data.model.subscription.SubscriptionModel
        """
        raise NotImplementedError()

    def get_service_plans_list(self, owner_id=None):
        """

        :param owner_id: str
        :return: list[migrator.core.data.model.service_plan.ServicePlanModel]
        """
        raise NotImplementedError()

    def get_service_plan(self, service_plan_id):
        """

        :param service_plan_id: str
        :return: migrator.core.data.model.service_plan.ServicePlanModel
        """
        raise NotImplementedError()

    def get_source_hosts_list(self):
        """

        :return: list[migrator.core.data.model.source_host.SourceNodeModel]
        """
        raise NotImplementedError()

    def get_source_host(self, source_host_id):
        """

        :param source_host_id: str
        :return: migrator.core.data.model.source_host.SourceNodeModel
        """
        raise NotImplementedError()

    @contextmanager
    def filter_by_accounts(self, account_ids=None):
        yield

    @contextmanager
    def filter_by_hosts(self, host_ids=None):
        yield

    @property
    def is_partial(self):
        return False


class PagedPlatformReader(PlatformReader, PagedReader):
    """
    Enables migrator to retrieve source platform objects in pages.

    For large number of objects, paged retrieval is more memory-effective. This what we do for
    accounts using the method 'export_next_account_page()', which iterates over the full list of accounts,
    returning one page at a time.

    Internally, PagedPlatformReader implementation is based on the following functionality:
    - retrieve a list of accounts, that may own other accounts (administrator and reseller)
    - retrieve a mapping of account IDs to owner IDs
    - iteratively retrieve account objects for account ID list passed in parameters
    """

    def get_account_ownership_list(self):
        """Return a list of AccountOwnership objects that provides a parent/child relation between account IDs"""
        raise NotImplementedError()

    def export_next_account_page(self):
        """Retrieve accounts from source panel in pages; return a list of AccountModel objects"""
        accounts = self.get_account_ownership_list()
        account_pages = self._paginate(accounts)

        for account_page in account_pages:
            yield self.get_account_list(account_page)
