# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.language import _T
from migrator.core.utils import empty_string_if_none
from migrator.core.utils.lang.entity import Entity


class AccountModel(Entity):
    yaml_prefix = 'account_data'
    data_prefix = 'account'

    def __init__(
        self, external_id, info, staff_members=None, subscriptions=None, valid_since=None, valid_until=None, oa_id=None
    ):
        """
        :param external_id: str
        :param info: PersonalInfoModel
        :param staff_members: list[StaffMemberModel]
        :param subscriptions: list[SubscriptionModel]
        :param valid_since: DateTime
        :param valid_until: DateTime
        :param oa_id: int
        """
        self.external_id = external_id
        self.info = info
        self.staff_members = staff_members or []
        self.subscriptions = subscriptions or []
        self.valid_since = valid_since
        self.valid_until = valid_until
        self.parent_external_id = None
        self.oa_id = oa_id

    def get_human_readable_id(self):
        return '%s (ID #%s)' % (self.info.company_name, self.external_id)

    def get_human_readable_full_id(self):
        return 'Account %s' % (self.get_human_readable_id())


class PersonalInfoModel(Entity):
    yaml_prefix = 'personal_data'
    data_prefix = 'personal_info'

    def __init__(self, company_name, admin_contact, billing_contact, tech_contact, company_address, language):
        """
        :param company_name: str
        :param admin_contact: ContactModel
        :param billing_contact: ContactModel
        :param tech_contact: ContactModel
        :param company_address: AddressModel
        :param language: str
        """
        self._fill()


class StaffMemberModel(Entity):
    yaml_prefix = 'staff_member_data'
    data_prefix = 'staff_member'

    def __init__(self, contact, address, authentication, language):
        """
        :param contact: ContactModel
        :param address: AddressModel
        :param authentication: AuthModel
        :param language: str
        """
        self._fill()

    def get_human_readable_id(self):
        return _T('STAFF_MEMBER_HUMAN_READABLE_ID').format(
            first_name=self.contact.first_name,
            last_name=self.contact.last_name,
            email=self.contact.email,
            account=self.account.get_human_readable_id()
        )

    def get_human_readable_full_id(self):
        return 'User %s' % (self.get_human_readable_id())


class ContactModel(Entity):
    yaml_prefix = 'contact_data'
    data_prefix = 'contact_info'

    def __init__(self, title, first_name, middle_name, last_name, email, phone, fax=None):
        """
        :param title: str
        :param first_name: str
        :param middle_name: str
        :param last_name: str
        :param email: str
        :param phone: PhoneModel
        :param fax: PhoneModel
        """
        self._fill()


class AddressModel(Entity):
    yaml_prefix = 'address_data'
    data_prefix = 'address'

    def __init__(self, city, address1, address2, zipcode, country, state):
        """
        :param city: str
        :param address1: str
        :param address2: str
        :param zipcode: str
        :param country: str
        :param state: str
        """
        self.city = city
        self.address1 = address1
        self.address2 = address2 or ''
        self.zipcode = zipcode
        self.country = country.lower()
        self.state = state or ''


class PhoneModel(Entity):
    yaml_prefix = 'phone_data'
    data_prefix = 'phone'

    def __init__(self, country_code, area_code, phone_num, ext_num):
        """
        :param country_code: str
        :param area_code: str
        :param phone_num: str
        :param ext_num: str
        """
        self._fill({empty_string_if_none: ['country_code', 'area_code', 'phone_num', 'ext_num']})


class AuthModel(Entity):
    yaml_prefix = 'auth_data'
    data_prefix = 'authentication'

    def __init__(self, login, password):
        """
        :param login: str
        :param password: str
        """
        self._fill()
