# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import InternalError
from migrator.core.utils import empty_string_if_none, cast_to_int
from migrator.core.utils.lang.entity import Entity


class PaymentSystemType(object):
    CREDIT_CARD = 0
    BANK_ACCOUNT = 1


class SourceSubscriptionStatus(object):
    STATUS_ACTIVE = 'active'
    STATUS_GRACED = 'graced'
    STATUS_EXPIRED = 'expired'
    STATUS_DELETED = 'deleted'
    STATUS_ON_HOLD = 'on hold'
    STATUS_ORDERED = 'ordered'
    STATUS_TRIAL = 'trial'
    STATUS_TERMINATING = 'terminating'
    STATUS_TERMINATED = 'terminated'
    STATUS_CANCELED = 'canceled'
    STATUS_ADMINISTRATIVE_HOLD = 'administrative hold'
    STATUS_CREDIT_ADMINISTRATIVE_HOLD = 'credit and administrative hold'

billing_type_period = {'hour': 0, 'day': 1, 'month': 2, 'year': 3}
subscription_statuses_allowed_for_migration = [
    SourceSubscriptionStatus.STATUS_ACTIVE, SourceSubscriptionStatus.STATUS_GRACED
]


class SourceBillingSystemModel(Entity):
    data_prefix = 'billing_info'

    def __init__(self, resellers=None, accounts=None, subscriptions=None):
        """
        :param resellers: list[AccountBillingModel]
        :param accounts: list[AccountBillingModel]
        """
        self.resellers = resellers or []
        self.accounts = accounts or []
        self.subscriptions = subscriptions or []


class AccountBillingModel(Entity):
    yaml_prefix = 'account_data'
    data_prefix = 'account'

    def __init__(
        self, external_id, balance, tax_reg_id=None, credit_cards=None, bank_accounts=None
    ):
        """
        :param external_id: int
        :param balance: float
        :param tax_reg_id: str
        :param credit_cards: list[CreditCardModel]
        :param bank_accounts: list[BankAccountModel]
        """
        self.external_id = external_id
        self.balance = float(balance) if balance is not None else None
        self.tax_reg_id = tax_reg_id
        self.credit_cards = credit_cards or []
        self.bank_accounts = bank_accounts or []

    @classmethod
    def from_account(cls, account):
        return cls(account.external_id, account.balance, account.credit_cards, account.bank_accounts)


class ResellerBillingModel(AccountBillingModel):
    yaml_prefix = 'reseller_data'
    data_prefix = 'reseller'


class BankAccountModel(Entity):
    yaml_prefix = 'bank_account_data'
    data_prefix = 'bank_account'

    def __init__(self, account_number, payment_system, bank_number, holder_name, use_for_recurring):
        """
        :param account_number: str
        :param payment_system: str
        :param bank_number: str
        :param holder_name: str
        :param use_for_recurring: int
        :return:
        """
        self._fill()


class CreditCardModel(Entity):
    yaml_prefix = 'credit_card_data'
    data_prefix = 'credit_card'

    def __init__(self, credit_card_info, address, phone, fax, email, use_for_auto_payments):
        """
        :param credit_card_info: CreditCardInfoModel
        :param address: AddressModel
        :param phone: PhoneModel
        :param fax: PhoneModel
        :param email: str
        :param use_for_auto_payments: int
        """
        self._fill({empty_string_if_none: ['email']})


class CreditCardInfoModel(Entity):
    yaml_prefix = 'credit_card_data'
    data_prefix = 'credit_card_info'

    def __init__(self, payment_system, card_number, card_holder_name, cvv, start_date, expiration_date, issue_number):
        """
        :param payment_system: str
        :param card_number: str
        :param card_holder_name: str
        :param cvv: str
        :param start_date: str
        :param expiration_date: str
        :param issue_number: str
        """
        self._fill({empty_string_if_none: ['start_date', 'issue_number', 'cvv']})


class SubscriptionBillingModel(Entity):
    yaml_prefix = 'subscription_billing_data'
    data_prefix = 'subscription_billing'

    def __init__(
            self, external_id, status, creation_date=None, expiration_date=None, subscription_period=None,
            subscription_period_type=None, last_billing_date=None, next_billing_date=None
    ):
        """
        :param external_id: int
        :param status: int
        :param creation_date: int
        :param expiration_date: int
        :param subscription_period: int
        :param subscription_period_type: int
        :param last_billing_date: int
        :param next_billing_date: int
        :return:
        """
        self._fill(
            {cast_to_int: ['creation_date', 'expiration_date', 'subscription_period', 'subscription_period_type',
                           'last_billing_date', 'next_billing_date']}
        )


def get_period_type_string(period_type):
        # Determining the period measurement unit (hour, day, month, year) by its code.
        period_type_str = [
            k for k, v in billing_type_period.iteritems() if v == period_type
        ]
        if not period_type_str:
            raise InternalError('No such period type: %s' % period_type)

        return period_type_str[0]
