# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.utils import parse_bool
from migrator.core.utils.lang.entity import Entity

DNS_RESOURCE_RECORD_MX = 'mx'
DNS_RESOURCE_RECORD_A = 'a'
DNS_RESOURCE_RECORD_AAAA = 'aaaa'
DNS_RESOURCE_RECORD_CNAME = 'cname'
DNS_RESOURCE_RECORD_SRV = 'srv'
DNS_RESOURCE_RECORD_TXT = 'txt'


class DnsInfoModel(Entity):
    data_prefix = 'dns_info'

    def __init__(self, zones=None):
        """
        :param zones: list[DnsZoneModel]
        """
        self.zones = zones or []


class DnsZoneModel(Entity):
    data_prefix = 'zone'

    def __init__(self, external_id, name, records=None, soa=None, enabled=True, dns_managed=True):
        """
        :param external_id: str
        :param records: list[DnsRecordModel]
        :param soa: DnsSoaRecord
        :param enabled: bool Is the DNS zone active
        :param dns_managed: bool Is the DNS zone in 'master' mode and has DNS records
        """
        self.external_id = external_id
        self.name = name
        self.records = records or []
        self.soa = soa
        self.enabled = parse_bool(enabled)
        self.dns_managed = parse_bool(dns_managed)


class DnsSoaRecordModel(Entity):
    data_prefix = 'soa'

    def __init__(self, ttl, refresh, retry, expire, minimum, serial=None, email=None):
        """
        :param ttl: int
        :param refresh: int
        :param retry: int
        :param expire: int
        :param minimum: int
        :param serial: int
        :param email: str
        """
        self._fill()


class DnsRecordModel(Entity):
    data_prefix = 'record'

    def __init__(self, host, type, value, ttl=None, priority=None, is_edit_allowed=False):
        """
        :param host: str Examples: u'ns1.example.com'
        :param type: str Examples: 'a', 'mx', 'cname', 'txt'
        :param value: str Examples: '1.2.3.4', 'mail.example.com', 'example.com.', 'v=spf1 +a +mx +a:example.com -all'
        :param ttl: int|None Example: 600
        :param priority: int|None Example: 10
        """
        self.host = host
        self.type = type
        self.value = value
        self.priority = priority
        self.ttl = ttl
        self.is_edit_allowed = parse_bool(is_edit_allowed)


def get_filtered_dns_records_for_zone(zone, dns_records_to_exclude):
    """
    :param zone: DnsZoneModel The DNS zone to get DNS resource records from
    :param dns_records_to_exclude: list[] The list of DNS records in the pem.getDNSRecords format to exclude
        from the original DNS records list of DNS zone. Example:
        [{
            'rr_type': 'MX',
            'active': True,
            'host': 'example.com.',
            'ttl': 600,
            'record_id': 212,
            'data': '10 mail.example.com.',
            'domain_id': 16
        },
        {
            'rr_type': 'A',
            'active': True,
            'host': 'ipv4.example.com.',
            'ttl': 600,
            'record_id': 213,
            'data': '10.37.54.122',
            'domain_id': 16
        }]
    :return: list[DnsRecordModel] The original DNS records list of the zone without DNS records passed to
        exclude in dns_records_to_exclude. Example:
        [
            DnsRecordModel(priority='10', host='example.com', type='mx', value='mail.example.com', ttl='600'),
            DnsRecordModel(priority=None, host='ipv4.example.com', type='a', value='10.37.54.122', ttl='600')
        ]
    """
    if not dns_records_to_exclude:
        return zone.records

    result = []
    for record in zone.records:
        record_type = record.type.lower()
        record_value = record.value.lower()
        include_record = True
        for dns_record_to_exclude in dns_records_to_exclude:
            if "%s." % record.host.lower() == dns_record_to_exclude['host'].lower() and \
                    record_type == dns_record_to_exclude['rr_type'].lower():
                if record_type in ['cname', 'a'] and record_value == dns_record_to_exclude['data'].lower():
                    include_record = False
                    break

                if record_type == 'mx' and \
                        "%s %s." % (record.priority, record_value) == dns_record_to_exclude['data'].lower():
                    include_record = False
                    break

                if record_type == 'txt' and dns_record_to_exclude['data'].lower() == '"%s"' % record_value:
                    include_record = False
                    break

        if include_record:
            result.append(record)

    return result
