# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.data.model import ServiceModel
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.utils.lang.entity import Entity

import re


class SubscriptionMigrationStatus(object):
    NEW_SUBSCRIPTION = 0
    IN_PROGRESS = 1
    FINISHED_OK = 2
    FINISHED_WARNINGS = 3
    FINISHED_ERRORS = 4
    FINISHED_NOTICES = 5


class SubscriptionModel(Entity):
    yaml_prefix = 'subscription_data'
    data_prefix = 'subscription'

    def __init__(
            self, external_id, name, service_plan_external_id, parent_external_id=None, services=None,
            migration_status=SubscriptionMigrationStatus.NEW_SUBSCRIPTION,
            migration_mode=MigrationMode.TEST_MIGRATION, message='[]', oa_id=None, oa_plan_id=None, oa_plan_name=None,
            period_type=None, period=None, in_production=False
    ):
        """
        :param external_id: str
        :param name: str
        :param service_plan_external_id: str
        :param parent_external_id: str
        :param services: list[ServiceModel]
        :param migration_status: int
        :param migration_mode: int
        :param message: str
        :param oa_id: int
        :param oa_plan_id: int
        :param oa_plan_name: str
        :param period_type: str
        :param period: int
        :param in_production: bool
        """

        self.name = name
        self.external_id = external_id
        self.service_plan_external_id = service_plan_external_id
        self.parent_external_id = parent_external_id
        self.services = services or []
        self.migration_status = migration_status
        self.migration_mode = migration_mode
        self.message = message
        self.oa_id = oa_id
        self.oa_plan_id = oa_plan_id
        self.oa_plan_name = oa_plan_name
        self.period_type = period_type
        self.period = period
        self.in_production = in_production

    def __eq__(self, other):
        """Provide comparision operator for refresh-data command"""

        # Check is both subscriptions are domain subscriptions and have the same domain ID.
        # Needed for case when Paid domain becomes Free domain (domain subscription becomes inactive) and vice versa
        # external_id will change in following way: 'dns_{domainId}_{subId}' -> 'dns_{domainId}'.
        self_external_id = re.search('(dns_\d+)', str(self.external_id))
        self_external_id = self_external_id.group(1) if self_external_id else self.external_id
        other_external_id = re.search('(dns_\d+)', str(other.external_id))
        other_external_id = other_external_id.group(1) if other_external_id else other.external_id

        return self.name == other.name and str(self_external_id) == str(other_external_id)

    def __ne__(self, other):
        """Provide comparision operator for refresh-data command"""
        return not self == other

    def clone(self, existing_subscription):
        """Copy the payload from the existing object; leave the IDs/references as they are"""
        self.migration_status = existing_subscription.migration_status
        self.migration_mode = existing_subscription.migration_mode
        self.message = existing_subscription.message
        self.oa_id = existing_subscription.oa_id
        self.oa_plan_id = existing_subscription.oa_plan_id
        self.oa_plan_name = existing_subscription.oa_plan_name
        self.in_production = existing_subscription.in_production


def has_any_of_service_types(subscription, service_types):
    return subscription.services and \
           len(set(subscription.services).intersection(set([ServiceModel(x, 0) for x in service_types]))) > 0


def get_service(subscription, service_type):
    services = get_services(subscription, service_type)
    return services[0] if services else None


def get_services(subscription, service_type):
    if subscription.services is None:
        return None

    return [service for service in subscription.services if service.type == service_type]


def get_service_property(subscription, service_type, property_name):
    service = get_service(subscription, service_type)
    if service is None or service.properties is None:
        return None

    for property in service.properties:
        if property.key == property_name:
            return property.value

    return None
