# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.utils import parse_bool
from migrator.core.utils.lang.entity import Entity


class HostingModel(Entity):
    yaml_prefix = 'hosting_data'
    data_prefix = 'web_hosting_shm'

    def __init__(self, subscriptions=None, database_servers=None):
        """
        :param subscriptions: list[SubscriptionModel]
        :param database_servers: list[DatabaseServerModel]
        """
        self.subscriptions = subscriptions or []
        self.database_servers = database_servers or []


class SubscriptionModel(Entity):
    yaml_prefix = 'subscription_data'
    data_prefix = 'subscription'

    def __init__(self, external_id, webspaces=None):
        """
        :param external_id: str
        :param webspaces: list[WebspaceModel]
        """
        self.external_id = external_id
        self.webspaces = webspaces or []


class WebspaceModel(Entity):
    yaml_prefix = 'webspace_data'
    data_prefix = 'webspace'

    def __init__(
        self, name, external_id, type, system_user=None, web_users=None, ftp_users=None, ip_addresses=None,
        separate_ssl=None, sites=None, databases=None, database_users=None, plesk_guid=None, host_external_id=None,
        hosting_type=None, protected_directories=None
    ):
        """
        :param name: str
        :param external_id: str
        :param type: str
        :param plesk_guid: str
        :param system_user: SystemUserModel
        :param ftp_users: list[FtpuserModel]
        :param web_users: list[WebuserModel]
        :param ip_addresses: list[IpAddressModel]
        :param separate_ssl: boolean
        :param sites: list[SiteModel]
        :param databases: list[DatabaseModel]
        :param host_external_id: str
        :param protected_directories: list[ProtectedDirectoryModel]
        """
        self.name = name
        self.external_id = external_id
        self.type = type
        self.plesk_guid = plesk_guid
        self.system_user = system_user
        self.ftp_users = ftp_users or []
        self.web_users = web_users or []
        self.ip_addresses = ip_addresses or []
        self.sites = sites or []
        self.databases = databases or []
        self.protected_directories = protected_directories or []
        self.database_users = database_users or []
        self.host_external_id = host_external_id or ''
        self.hosting_type = hosting_type or 'virtual_hosting'

        if separate_ssl is not None:
            # TODO: Fix XML loader instead
            self.separate_ssl = parse_bool(separate_ssl)
        else:
            self.separate_ssl = None


class SystemUser(Entity):
    yaml_prefix = 'system_user_data'
    data_prefix = 'system_user'

    def __init__(self, name, password):
        """
        :param name: str
        :param password: str
        """
        self._fill()


class IpAddressModel(Entity):
    yaml_prefix = 'ip_address_data'
    data_prefix = 'ip_address'

    def __init__(self, type, ip, interface):
        """
        :param type: str
        :param ip: str
        """
        self._fill()


class SiteModel(Entity):
    yaml_prefix = 'site_data'
    data_prefix = 'site'

    def __init__(self, name, external_id, status, is_main, hosting, plesk_guid=None):
        """
        :param name: str
        :param external_id: str
        :param status: str
        :param is_main: bool
        :param hosting: PhysicalHostingModel \ ForwardingStdModel
        :param plesk_guid: str
        """
        self.name = name
        self.external_id = external_id
        self.status = status
        self.is_main = parse_bool(is_main)
        self.plesk_guid = plesk_guid
        self.hosting = hosting


class ForwardingStdModel(Entity):
    yaml_prefix = 'forwarding_std_data'
    data_prefix = 'forwarding_std'

    def __init__(self, destination, http_code):
        """
        :param destination: str
        :param http_code: str
        """
        self._fill()


class PhysicalHostingModel(Entity):
    yaml_prefix = 'physical_data'
    data_prefix = 'physical'

    def __init__(
        self, hosting_properties, scripting, php_settings, ssl_certificate=None, protected_directories=None,
        virtual_directories=None, application_pool=None, applications=None
    ):
        """
        :param hosting_properties: HostingPropertiesModel
        :param scripting: ScriptingModel
        :param php_settings: PHPSettingsModel
        :param ssl_certificate: SslCertificateModel
        :param protected_directories: list[ProtectedDirectoryModel]
        :param virtual_directories: list[VirtualDirectoryModel]
        :param application_pool: ApplicationPoolModel
        :param applications: list[ApplicationModel]
        """
        self.hosting_properties = hosting_properties
        self.scripting = scripting
        self.php_settings = php_settings
        self.ssl_certificate = ssl_certificate
        self.protected_directories = protected_directories or []
        self.virtual_directories = virtual_directories or []
        self.application_pool = application_pool
        self.applications = applications or []


class NoHostingModel(Entity):
    yaml_prefix = 'no_hosting_data'
    data_prefix = 'no_hosting'

    def __init__(self):
        pass


class DatabaseModel(Entity):
    yaml_prefix = 'database_data'
    data_prefix = 'database'

    def __init__(self, external_id, database_server_id, name):
        """
        :param external_id: str
        :param database_server_id: str
        :param name: str
        :param database_users: list[DatabaseUserModel]
        """
        self.external_id = external_id
        self.database_server_id = database_server_id
        self.name = name


class DatabaseUserModel(Entity):
    yaml_prefix = 'database_user_data'
    data_prefix = 'database_user'

    def __init__(self, login, password, database_external_id=None, database_server_id=None):
        """
        :param login: str
        :param password: str
        """
        self._fill()


class WebuserModel(Entity):
    yaml_prefix = 'web_user_data'
    data_prefix = 'web_user'

    def __init__(
        self, login, password, ftp_quota=None, asp=None, asp_dot_net=None, php=None, cgi=None, perl=None, python=None,
        fastcgi=None, write_modify=None
    ):
        """
        :param login: str
        :param password: str
        :param ftp_quota: int
        :param asp: bool
        :param asp_dot_net: bool
        :param php: bool
        :param cgi: bool
        :param perl: bool
        :param python: bool
        :param fastcgi: bool
        :param write_modify: bool
        """
        self._fill()


class FtpuserModel(Entity):
    yaml_prefix = 'ftp_user_data'
    data_prefix = 'ftp_user'

    def __init__(self, name, password, home, read, write, quota):
        """
        :param name: str
        :param password: str
        :param home: str
        :param read: bool
        :param write: bool
        :param quota: int
        """
        self._fill()


class SslCertificateModel(Entity):
    yaml_prefix = 'ssl_certificate_data'
    data_prefix = 'ssl_certificate'

    def __init__(self, name, csr, private_key, cert, ca):
        """
        :param name: str
        :param csr: str
        :param private_key: str
        :param cert: str
        :param ca: str
        """
        self._fill()


class VirtualDirectoryModel(Entity):
    yaml_prefix = 'virtual_directory_data'
    data_prefix = 'virtual_directory'

    def __init__(self, name, path=None, virtual_directories=None,
        asp_enable_client_debug_inherited=None, asp_enable_client_debug_default=None, asp_enable_client_debug=None,
        asp_enable_server_debug_inherited=None, asp_enable_server_debug_default=None, asp_enable_server_debug=None,
        asp_buffering_limit_inherited=None, asp_buffering_limit_default=None, asp_buffering_limit=None,
        asp_max_request_entity_allowed_inherited=None, asp_max_request_entity_allowed_default=None, asp_max_request_entity_allowed=None,
        editable=None, script_maps=None, require_ssl=None, auth_ntlm=None, auth_anonymous=None,
        asp_execute_in_mta=None, asp_enable_parent_paths=None,
        default_doc=None, enable_default_doc=None, app_create=None, dont_log=None, enable_dir_browsing=None,
        access_source=None, access_execute=None, access_script=None, access_write=None, access_read=None):
        self._fill()


class ProtectedDirectoryModel(Entity):
    yaml_prefix = 'protected_directory_data'
    data_prefix = 'protected_directory'

    def __init__(self, name, header, pd_users=None, properties=None):
        """
        :param name: str
        :param header: str
        :param properties: list[PropertyModel]
        """
        self.name = name
        self.header = header
        self.pd_users = pd_users or []
        self.properties = properties or []


class ProtectedDirectoryUserModel(Entity):
    yaml_prefix = 'pd_user_data'
    data_prefix = 'pd_user'

    def __init__(self, login, password, password_type=None):
        """
        :param login: str
        :param password: str
        :param password_type: str
        """
        self.login = login
        self.password = password
        self.password_type = 'plain' if password_type is None else password_type


class PropertyModel(Entity):
    yaml_prefix = 'property_data'
    data_prefix = 'property'

    def __init__(self, name, value):
        """
        :param name: str
        :param value: str
        """
        self._fill()


class HostingPropertiesModel(Entity):
    yaml_prefix = 'hosting_properties_data'
    data_prefix = 'hosting_properties'

    def __init__(
            self, iis_app_pool, ftp_quota, errdocs, wuscripts, at_domains, seo_redirect, webstat, webstat_protected,
            ssl, write_modify
    ):
        """
        :param iis_app_pool: bool
        :param ftp_quota: int
        :param errdocs: bool
        :param wuscripts: bool
        :param at_domains: bool
        :param seo_redirect: str
        :param webstat: str
        :param webstat_protected: bool
        :param ssl: bool
        :param write_modify: bool
        """
        self._fill()


class ScriptingModel(Entity):
    yaml_prefix = 'scripting_data'
    data_prefix = 'scripting'

    def __init__(
        self, shell, php, php_handler_id, cgi, perl, python, asp, asp_dot_net, asp_dot_net_version, fastcgi, cgi_mode
    ):
        """
        :param shell: str
        :param php: bool
        :param php_handler_id: str
        :param cgi: bool
        :param perl: bool
        :param python: bool
        :param asp: bool
        :param asp_dot_net: bool
        :param asp_dot_net_version: str
        :param fastcgi: bool
        :param cgi_mode:str
        """
        self._fill()


class PHPSettingsModel(Entity):
    yaml_prefix = 'php_settings_data'
    data_prefix = 'php_settings'

    oam_vs_ini_mapping = {
        'mail_force_extra_parameters': 'mail.force_extra_parameters',
    }

    def __init__(
            self, memory_limit=None, max_execution_time=None, max_input_time=None, post_max_size=None,
            upload_max_filesize=None, safe_mode=None, safe_mode_include_dir=None, safe_mode_exec_dir=None,
            include_path=None, mail_force_extra_parameters=None, register_globals=None,open_basedir=None,
            error_reporting=None, display_errors=None, log_errors=None, allow_url_fopen=None, file_uploads=None,
            short_open_tag=None, magic_quotes_gpc=None
    ):
        """
        :param memory_limit: str
        :param max_execution_time: str
        :param max_input_time: str
        :param post_max_size: str
        :param upload_max_filesize: str
        :param safe_mode: str
        :param safe_mode_include_dir: str
        :param safe_mode_exec_dir: str
        :param include_path: str
        :param mail_force_extra_parameters: str
        :param register_globals: str
        :param open_basedir: str
        :param error_reporting: str
        :param display_errors: str
        :param log_errors: str
        :param allow_url_fopen: str
        :param file_uploads: str
        :param short_open_tag: str
        :param magic_quotes_gpc: str
        """
        self._fill()


class ApplicationPoolModel(Entity):
    yaml_prefix = 'application_pool_data'
    data_prefix = 'application_pool'

    def __init__(
        self, managed_pipeline_mode=None, max_worker_processes=None, idle_timeout=None, idle_timeout_action=None,
        cpu_usage=None, cpu_usage_action=None, cpu_usage_interval=None,
        recycling_by_time=None, recycling_by_requests=None, recycling_by_virtual_memory=None,
        recycling_by_private_memory=None
    ):
        """
        :param managed_pipeline_mode: str
        :param max_worker_processes: str
        :param idle_timeout: str
        :param idle_timeout_action: str
        :param cpu_usage: str
        :param cpu_usage_action: str
        :param cpu_usage_interval: str
        :param recycling_by_time: str
        :param recycling_by_requests: str
        :param recycling_by_virtual_memory: str
        :param recycling_by_private_memory: str
        :return:
        """
        self._fill()


class DatabaseServerModel(Entity):
    yaml_prefix = 'database_server_data'
    data_prefix = 'database_server'

    def __init__(self, id, host, port, dbtype, login, password, default, local, alt_hosts=None):
        """
        :param id: str
        :param host: str
        :param port: int
        :param dbtype: str
        :param login: str
        :param password: str
        :param default: bool
        :param local: bool
        :param alt_hosts: list[string]
        """
        self._fill()
        self.alt_hosts = alt_hosts or []


class AliasModel(Entity):
    yaml_prefix = 'alias_data'
    data_prefix = 'alias'

    def __init__(self, site_id, seo_redirect):
        """
        :param site_id: str
        :param seo_redirect: bool
        """
        self.site_id = site_id
        # TODO: Fix XML loader instead
        self.seo_redirect = parse_bool(seo_redirect)


class ApplicationModel(Entity):
    data_prefix = 'application'

    def __init__(self, name, source_path='', target_path='', database_external_id=None):
        """
        :type name: string
        :type database_external_id: string
        """
        self.name = name
        self.source_path = source_path or ''
        self.target_path = target_path or ''
        self.database_external_id = database_external_id
