# (C)2017 Ingram Micro Inc. all Rights Reserved
from sqlalchemy import func
from sqlalchemy.orm import aliased

from migrator.core import InternalError
from migrator.core.data.source.db_meta import get_db_session, current_record
from migrator.core.data.model import *
from migrator.core.data.model.subscription import get_service as get_subscription_service
from migrator.core.utils.registry import Registry
from migrator.core.utils.filetypes.config_utils import get_source_type_section
from migrator.core.data.format.platform import PlatformReader


class PlatformReaderDB(PlatformReader):
    def __init__(self, expire_data=False):
        self.session = get_db_session()
        if expire_data:
            self.session.expire_all()

        global_context = Registry.get_instance().get_context()
        section = get_source_type_section(global_context.config)
        self.aps_id = section.get('aps-id')
        self.migration_project_aps_id = global_context.migration_project_aps_id
        self.migration_project_id = self.get_migration_project().id if self.migration_project_aps_id is not None else None

    def get_source_system_model(self):
        return self.session.query(SourceSystemModel).filter(SourceSystemModel.aps_id == self.aps_id).first()

    def get_account(self, account_id):
        query = self.session.query(AccountModel)
        return self._add_account_filter(query, account_id).one_or_none()

    def get_reseller(self, reseller_id):
        query = self.session.query(AccountModel)
        return self._add_account_filter(query, reseller_id, 'reseller').one_or_none()

    def get_source_host(self, source_host_id):
        query = self.session.query(SourceHostModel).filter(
            SourceHostModel.external_id == source_host_id,
            SourceHostModel.source_system_id == self.get_source_system_model().id
        )
        return current_record(SourceHostModel, query).one_or_none()

    def _add_account_filter(self, query, external_id, acc_type=None):
        """Create a filter for accounts that includes source system APS ID and returns only currently valid records"""
        account_query = query.join(SourceSystemModel).filter(AccountModel.external_id == external_id)\
            .filter(SourceSystemModel.aps_id == self.aps_id)
        if acc_type is not None:
            account_query = account_query.filter(AccountModel.type == acc_type)
        return current_record(AccountModel, account_query)

    def get_subscription_account(self, subscription_id):
        query = self.session.query(AccountModel).join(SubscriptionModel).join(SourceSystemModel). \
            filter(SubscriptionModel.external_id == subscription_id). \
            filter(SourceSystemModel.aps_id == self.aps_id)

        return current_record(AccountModel, query).one_or_none()

    def get_subscriptions_list(self, account_id=None):
        if account_id is None:
            query = self.session.query(SubscriptionModel).join(AccountModel).join(SourceSystemModel).\
                filter(SourceSystemModel.aps_id == self.aps_id)
        else:
            query = self.session.query(SubscriptionModel).join(AccountModel).join(SourceSystemModel).\
                filter(AccountModel.external_id == account_id).\
                filter(SourceSystemModel.aps_id == self.aps_id)

        return current_record(AccountModel, query).all()

    def get_subscription(self, subscription_external_id=None, primary_id=None):
        """
        :param subscription_external_id: string
        :param primary_id: int
        :return: SubscriptionModel|None
        """
        subscription_query = self.session.query(SubscriptionModel).join(AccountModel).join(SourceSystemModel).\
            filter(SourceSystemModel.aps_id == self.aps_id)
        if subscription_external_id:
            subscription_query = subscription_query.filter(SubscriptionModel.external_id == subscription_external_id)

        if primary_id:
            subscription_query = subscription_query.filter(SubscriptionModel.id == primary_id)

        return current_record(AccountModel, subscription_query).one_or_none()

    def get_subscription_service_plan(self, subscription_id):
        """
        :param subscription_id: string
        """
        service_plan_query = self.session.query(ServicePlanModel).\
            join(
                SubscriptionModel,
                SubscriptionModel.service_plan_external_id == ServicePlanModel.external_id
            ).join(SourceSystemModel).join(AccountModel, AccountModel.id == SubscriptionModel.account_id).\
            filter(SubscriptionModel.external_id == subscription_id).\
            filter(SourceSystemModel.aps_id == self.aps_id)

        return current_record(ServicePlanModel, current_record(AccountModel, service_plan_query)).one_or_none()

    def get_subscriptions_by_service_plan_type(self, service_plan_type):
        """
        :param subscription_id: string
        """
        service_plan_query = self.session.query(SubscriptionModel).\
            join(
                ServicePlanModel,
                SubscriptionModel.service_plan_external_id == ServicePlanModel.external_id
            ).\
            join(MigrationProjectItemModel).join(SourceSystemModel).\
            filter(ServicePlanModel.type == service_plan_type).\
            filter(MigrationProjectItemModel.migration_project_id == self.migration_project_id).\
            filter(SourceSystemModel.aps_id == self.aps_id)

        return current_record(ServicePlanModel, service_plan_query).distinct().all()

    def get_subscriptions_by_services(self, service_types):
        subscription_query = self.session.query(SubscriptionModel).join(AccountModel).join(SourceSystemModel).\
            join(ServiceModel).\
            filter(ServiceModel.type.in_(service_types)).\
            filter(SourceSystemModel.aps_id == self.aps_id)

        return current_record(AccountModel, subscription_query).distinct().all()

    def get_subscriptions_by_parent(self, parent_external_id):
        """
        :param parent_external_id: string
        :return: SubscriptionModel|None
        """
        subscription_query = self.session.query(SubscriptionModel).join(AccountModel).join(SourceSystemModel).\
            filter(SourceSystemModel.aps_id == self.aps_id)

        subscription_query = subscription_query.filter(SubscriptionModel.parent_external_id == parent_external_id)

        return current_record(AccountModel, subscription_query).all()

    def get_source_host_by_subscription_service(self, subscription_id, service_type):
        query = self.session.query(SourceHostModel).join(SourceSystemModel).\
            join(ServiceModel, ServiceModel.source_host_external_id == SourceHostModel.external_id).\
            join(SubscriptionModel).join(AccountModel).\
            filter(SubscriptionModel.external_id == subscription_id).\
            filter(ServiceModel.type == service_type).\
            filter(SourceSystemModel.aps_id == self.aps_id).\
            filter(AccountModel.source_system_id == SourceSystemModel.id)

        return current_record(SourceHostModel, query).one_or_none()

    def get_migration_project_items(self, service_types=None):
        """
        :param service_types: list[string]|None
        :return: list[MigrationProjectItemModel]|None
        """
        query = self.session.query(MigrationProjectItemModel).join(MigrationProjectModel).filter(
            MigrationProjectModel.aps_id == self.migration_project_aps_id
        )
        if service_types:
            query = query.join(SubscriptionModel).join(ServiceModel).filter(ServiceModel.type.in_(service_types))

        return query.all()

    def get_migration_project_subscriptions(self, created_in_oa=None):
        """
        Get all subscriptions IDs of the current migration project

        :param created_in_oa: bool|None True/False - return created/not created subscriptions in OA, None - all
        :return: list[SubscriptionModel]|None
        """
        query = self.session.query(SubscriptionModel).join(MigrationProjectItemModel).join(MigrationProjectModel).\
            filter(MigrationProjectModel.aps_id == self.migration_project_aps_id)

        if created_in_oa is not None:
            if created_in_oa:
                query = query.filter(SubscriptionModel.oa_id != None)
            else:
                query = query.filter(SubscriptionModel.oa_id == None)

        return query.all()

    def get_migration_project_item(self, subscription_id):
        """
        :param subscription_id: int
        :return: MigrationProjectItemModel|None
        """
        return self.session.query(MigrationProjectItemModel).join(MigrationProjectModel).filter(
            MigrationProjectModel.aps_id == self.migration_project_aps_id,
            MigrationProjectItemModel.subscription_id == subscription_id
        ).one_or_none()

    def get_migration_required_accounts(self):
        """
        :return: list[AccountModel]
        """
        accounts_query = self.session.query(AccountModel).join(SubscriptionModel).join(MigrationProjectItemModel).\
            filter(
                AccountModel.source_system_id == self.get_source_system_model().id,
                MigrationProjectItemModel.migration_project_id == self.migration_project_id
            )

        return current_record(AccountModel, accounts_query).all()

    def get_migration_required_staff_members(self):
        """
        :return: list[StaffMemberModel]
        """
        users_query = self.session.query(StaffMemberModel).join(AccountModel).join(SubscriptionModel).\
            join(MigrationProjectItemModel).\
            filter(
                AccountModel.source_system_id == self.get_source_system_model().id,
                MigrationProjectItemModel.migration_project_id == self.migration_project_id
            )

        return current_record(AccountModel, users_query).all()

    def get_source_hosts_for_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: list[SourceHostModel]
        """
        hosts_query = self.session.query(SourceHostModel).\
            join(ServiceModel, ServiceModel.source_host_external_id == SourceHostModel.external_id).\
            join(SubscriptionModel).filter(
                SubscriptionModel.id == subscription_id,
                SourceHostModel.source_system_id == self.get_source_system_model().id
            )

        return current_record(SourceHostModel, hosts_query).all()

    def get_migration_required_source_hosts(self, panel_type):
        hosts_query = self.session.query(SourceHostModel).distinct(SourceHostModel.ip).\
            join(ServiceModel, ServiceModel.source_host_external_id == SourceHostModel.external_id).\
            join(SubscriptionModel, SubscriptionModel.id == ServiceModel.subscription_id).\
            join(MigrationProjectItemModel, MigrationProjectItemModel.subscription_id == SubscriptionModel.id).\
            filter(
                MigrationProjectItemModel.migration_project_id == self.migration_project_id,
                SourceHostModel.source_system_id == self.get_source_system_model().id,
                SourceHostModel.panel == panel_type
            ).order_by(SourceHostModel.ip)

        return current_record(SourceHostModel, hosts_query).all()

    def get_migration_project(self):
        return self.session.query(MigrationProjectModel).\
            filter(MigrationProjectModel.aps_id == self.migration_project_aps_id).one_or_none()

    def get_accounts_by_hosts(self, source_host_external_ids):
        source_system = self.get_source_system_model()
        query = self.session.query(AccountModel).distinct(AccountModel.id).\
            join(SubscriptionModel, SubscriptionModel.account_id == AccountModel.id).\
            join(ServiceModel, ServiceModel.subscription_id == SubscriptionModel.id).\
            join(SourceHostModel, SourceHostModel.external_id == ServiceModel.source_host_external_id).\
            filter(
                AccountModel.source_system_id == source_system.id,
                SourceHostModel.source_system_id == source_system.id,
                SourceHostModel.external_id.in_(source_host_external_ids)
            )
        return current_record(SourceHostModel, current_record(AccountModel, query)).all()

    def get_account_subscription_map(self):
        source_system = self.get_source_system_model()
        query = self.session.query(AccountModel.external_id, func.array_agg(SubscriptionModel.external_id)).\
            filter(
            AccountModel.source_system_id == source_system.id,
            AccountModel.id == SubscriptionModel.account_id,
            SubscriptionModel.parent_external_id == None
        ).group_by(AccountModel.external_id)

        return {acc: {x for x in subscrs if x.isdigit()} for acc, subscrs in current_record(AccountModel, query).all()}

    def get_accounts_and_owner(self):
        source_system = self.get_source_system_model()
        acc = aliased(AccountModel)
        par = aliased(AccountModel)
        query = self.session.query(acc, par).\
            filter(
                acc.owner_id == par.id,
                acc.source_system_id == source_system.id,
                par.source_system_id == source_system.id
            )
        return current_record(acc, current_record(par, query)).all()

    def update_subscription_status(self, new_subscription):
        self.session.query(SubscriptionModel).filter_by(id=new_subscription.id).update(
            {
                'migration_status': new_subscription.migration_status,
                'migration_mode': new_subscription.migration_mode,
                'message': new_subscription.message
            })
        self.session.commit()

    def update_subscription_production_status(self, new_subscription):
        self.session.query(SubscriptionModel).filter_by(id=new_subscription.id).update(
            {
                'in_production': new_subscription.in_production
            }
        )
        self.session.commit()

    def update_migration_project_status(self, new_migration_project):
        self.session.query(MigrationProjectModel).filter_by(id=new_migration_project.id).update(
            {
                'status': new_migration_project.status,
                'message': new_migration_project.message
            })
        self.session.commit()

    def reset_subscription_oa_details(self, external_id):
        subscription = self.get_subscription(external_id)
        subscription.oa_id = None
        subscription.oa_plan_id = None
        subscription.oa_plan_name = None
        subscription.in_production = False
        self.session.commit()

    def update_subscription_service_property(self, primary_id, service_type, property_name, property_value):
        subscription = self.get_subscription(primary_id=primary_id)
        service = get_subscription_service(subscription, service_type)
        if service is None or service.properties is None:
            raise InternalError("Service type '%s' is absent for subscription '%s'" % (service_type, subscription.id))

        for prop in service.properties:
            if prop.key == property_name:
                prop.value = property_value
                self.session.commit()
                return

        service.properties.append(ServicePropertyModel(property_name, property_value))
        self.session.commit()
