# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import thread

from sqlalchemy import Table, MetaData, Column, Integer, String, ForeignKey, Boolean, create_engine, DateTime, Text
from sqlalchemy.orm import mapper, relationship, sessionmaker, backref
from sqlalchemy.sql.expression import and_

from migrator.core.data.model import *
from migrator.core.data.model.migration_project import MigrationProjectStatus, MigrationMode
from migrator.core.data.model.service_property import ServicePropertyModel
from migrator.core.data.model.subscription import SubscriptionMigrationStatus
from migrator.core.data.model.task import TaskModel
from migrator.core.utils.filetypes.config_utils import ConfigSection, read_config
from migrator.settings import OA_MIGRATOR_DB_CONFIG, PRINT_DB_QUERIES_TO_LOG

metadata = MetaData()

# --- DB Tables for Models
source_system = Table(
    SourceSystemModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    # UI/endpoint expects the unique values here
    Column('aps_id', String(36), nullable=False, unique=True),
    Column('message', String, server_default='')
)

source_host = Table(
    SourceHostModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'source_system_id', Integer, ForeignKey('%s.id' % SourceSystemModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column('external_id', String(64), nullable=False),
    Column('name', String(256), nullable=False),
    Column('ip', String(45), nullable=False),
    Column('panel', String(8), nullable=False),
    Column('communication_ip', String(45), nullable=True),
    Column('valid_since', DateTime()),
    Column('valid_until', DateTime()),
)

account = Table(
    AccountModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('type', String(10), nullable=False),
    Column('owner_id', Integer, ForeignKey('%s.id' % AccountModel.data_prefix, ondelete='CASCADE')),
    Column('source_system_id', Integer, ForeignKey('%s.id' % SourceSystemModel.data_prefix, ondelete='CASCADE')),
    Column('external_id', String(64), nullable=False),
    Column('personal_info_id', Integer, ForeignKey('%s.id' % PersonalInfoModel.data_prefix), nullable=False),
    Column('valid_since', DateTime()),
    Column('valid_until', DateTime()),
    Column('oa_id', Integer),
)

personal_info = Table(
    PersonalInfoModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('company_name', String(256)),
    Column('admin_contact_id', Integer, ForeignKey('%s.id' % ContactModel.data_prefix)),
    Column('billing_contact_id', Integer, ForeignKey('%s.id' % ContactModel.data_prefix)),
    Column('tech_contact_id', Integer, ForeignKey('%s.id' % ContactModel.data_prefix)),
    Column('company_address_id', Integer, ForeignKey('%s.id' % AddressModel.data_prefix)),
    Column('language', String(16))
)

address = Table(
    AddressModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('address1', String(256)),
    Column('address2', String(256)),
    Column('city', String(256)),
    Column('zipcode', String(64)),
    Column('country', String(64)),
    Column('state', String(80))
)

contact_info = Table(
    ContactModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('title', String(64)),
    Column('first_name', String(256)),
    Column('middle_name', String(256)),
    Column('last_name', String(256)),
    Column('email', String(256)),
    Column('phone_id', ForeignKey('%s.id' % PhoneModel.data_prefix)),
    Column('fax_id', ForeignKey('%s.id' % PhoneModel.data_prefix))
)

phone = Table(
    PhoneModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('country_code', String(24)),
    Column('area_code', String(24)),
    Column('phone_num', String(24)),
    Column('ext_num', String(24))
)

authentication = Table(
    AuthModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('login', String(256), nullable=False),
    Column('password', String(256), nullable=False)
)

service = Table(
    ServiceModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'subscription_id', Integer, ForeignKey('%s.id' % SubscriptionModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column('type', String(16), nullable=False),
    Column('source_host_external_id', String(64), nullable=False)
)

staff_member = Table(
    StaffMemberModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('account_id', Integer, ForeignKey('%s.id' % AccountModel.data_prefix, ondelete='CASCADE')),
    Column('contact_info_id', Integer, ForeignKey('%s.id' % ContactModel.data_prefix)),
    Column('address_id', Integer, ForeignKey('%s.id' % AddressModel.data_prefix)),
    Column('authentication_id', Integer, ForeignKey('%s.id' % AuthModel.data_prefix)),
    Column('language', String(16))
)

subscription = Table(
    SubscriptionModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('account_id', Integer, ForeignKey('%s.id' % AccountModel.data_prefix, ondelete='CASCADE'), nullable=False),
    Column('name', String(256)),
    Column('migration_status', Integer, server_default=str(SubscriptionMigrationStatus.NEW_SUBSCRIPTION)),
    Column('migration_mode', Integer, server_default=str(MigrationMode.TEST_MIGRATION)),
    Column('message', String, server_default='[]'),
    Column('external_id', String(64), nullable=False),
    Column('parent_external_id', String(64)),
    Column('service_plan_external_id', String(64)),
    Column('oa_id', Integer),
    Column('oa_plan_id', Integer),
    Column('oa_plan_name', String(256)),
    Column('period_type', Integer),
    Column('period', Integer),
    Column('in_production', Boolean)
)

service_plan = Table(
    ServicePlanModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('source_system_id', Integer, ForeignKey('%s.id' % SourceSystemModel.data_prefix, ondelete='CASCADE')),
    Column('reseller_id', Integer, ForeignKey('%s.id' % AccountModel.data_prefix, ondelete='CASCADE')),
    Column('external_id', String(64), nullable=False),
    Column('name', String(256)),
    Column('type', String(64)),
    Column('valid_since', DateTime()),
    Column('valid_until', DateTime()),
)

migration_project = Table(
    MigrationProjectModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    # UI/endpoint expects the unique values here
    Column('aps_id', String(36), nullable=False, unique=True),
    Column('message', String, server_default='[]'),
    Column('status', Integer, server_default=str(MigrationProjectStatus.NEW_PROJECT)),
)

migration_project_item = Table(
    MigrationProjectItemModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'migration_project_id', Integer, ForeignKey('%s.id' % MigrationProjectModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column('subscription_id', Integer, ForeignKey('%s.id' % SubscriptionModel.data_prefix), nullable=False),
    Column('target_service_plan_id', Integer),
    Column('mapping_reseller_id', String),
    Column('is_oa_account', Boolean),
    Column('mapping_subscription_id', String),
    Column('is_oa_subscription', Boolean)
)

service_plan_global_map = Table(
    ServicePlanGlobalMapModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'source_system_id', Integer, ForeignKey('%s.id' % SourceSystemModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column(
        'source_service_plan_id', Integer, ForeignKey('%s.id' % ServicePlanModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column('target_service_plan_id', Integer, nullable=False)
)

migration_check = Table(
    MigrationCheckModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'migration_project_id', Integer, ForeignKey('%s.id' % MigrationProjectModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column('name', String, nullable=False),
    Column('phase', String, nullable=False),
    Column('state', String, nullable=False),
    Column('last_exec_time', DateTime()),
    Column('last_message', String),
    Column('last_status', String)
)

service_property = Table(
    ServicePropertyModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'service_id', Integer, ForeignKey('%s.id' % ServiceModel.data_prefix, ondelete='CASCADE'),
        nullable=False
    ),
    Column('key', String(256), nullable=False),
    Column('value', Text, nullable=False)
)

task_table = Table(
    TaskModel.data_prefix, metadata,
    Column('id', Integer, primary_key=True),
    Column('pid', Integer, nullable=False),
    Column('aps_id', String(36), nullable=False, index=True),
    Column('operation_name', String(64), nullable=False, index=True),
    Column('config_path', String(256)),
    Column('response_path', String(256)),
    Column('message', String)
)

# utility methods for working with temporal data


def new_record(model, query):
    """Add a filter that selects only the newly added records - those, which are not yet valid"""
    return query.filter(model.valid_since == None)


def current_record(model, query):
    """Add a filter that selects the records which are currently valid"""
    return query.filter(model.valid_since != None).filter(model.valid_until == None)


def archived_record(model, query):
    """Add a filter that selects the records that are no longer valid"""
    return query.filter(model.valid_until != None)


def source_system_current_record_clause(timestamped_table):
    """Generate a ClauseElement object, that restricts relation to a timestamped table to currently valid records"""
    return and_(
        source_system.c.id==timestamped_table.c.source_system_id,
        timestamped_table.c.valid_since!=None, timestamped_table.c.valid_until==None
    )

# --- Models to DB Tables mapping
mapper(SourceSystemModel, source_system, properties={
    'providers': relationship(
        ProviderModel, backref='provider_{0}'.format(SourceSystemModel.data_prefix), order_by=account.c.id,
        cascade='delete,all',
        primaryjoin=source_system_current_record_clause(account)
    ),
    'resellers': relationship(
        ResellerModel, backref='reseller_{0}'.format(SourceSystemModel.data_prefix), order_by=account.c.id,
        cascade='delete,all',
        primaryjoin=source_system_current_record_clause(account)
    ),
    'accounts': relationship(
        AccountModel, backref='account_{0}'.format(SourceSystemModel.data_prefix), order_by=account.c.id,
        cascade='delete,all',
        primaryjoin=source_system_current_record_clause(account)
    ),
    'source_hosts': relationship(
        SourceHostModel, backref=SourceSystemModel.data_prefix,
        order_by=source_host.c.id, cascade='delete,all',
        primaryjoin=source_system_current_record_clause(source_host)
    ),
    'service_plans': relationship(
        ServicePlanModel, backref=SourceSystemModel.data_prefix, order_by=service_plan.c.id,
        cascade='delete,all', passive_deletes=True,
        primaryjoin=source_system_current_record_clause(service_plan)
    ),
    'service_plan_global_mapping': relationship(
        ServicePlanGlobalMapModel, backref=SourceSystemModel.data_prefix, order_by=service_plan_global_map.c.id,
        cascade='delete,all'
    )
})

mapper(SourceHostModel, source_host)
mapper(ServicePlanModel, service_plan)
mapper(MigrationCheckModel, migration_check)
mapper(PhoneModel, phone, confirm_deleted_rows=False)
mapper(AddressModel, address, confirm_deleted_rows=False)
mapper(AuthModel, authentication, confirm_deleted_rows=False)
mapper(ServicePropertyModel, service_property)
mapper(ServicePlanGlobalMapModel, service_plan_global_map)
mapper(TaskModel, task_table)

mapper(ServiceModel, service, confirm_deleted_rows=False, properties={
    'properties': relationship(
        ServicePropertyModel, backref=ServiceModel.data_prefix, order_by=service_property.c.id,
        cascade='delete,all'
    )
})

mapper(PersonalInfoModel, personal_info, confirm_deleted_rows=False, properties={
    'admin_contact': relationship(
        ContactModel, uselist=False, backref='admin_{0}'.format(PersonalInfoModel.data_prefix),
        primaryjoin=personal_info.c.admin_contact_id == contact_info.c.id, cascade='delete,all'
    ),
    'billing_contact': relationship(
        ContactModel, uselist=False, backref='billing_{0}'.format(PersonalInfoModel.data_prefix),
        primaryjoin=personal_info.c.billing_contact_id == contact_info.c.id, cascade='delete,all'
    ),
    'tech_contact': relationship(
        ContactModel, uselist=False, backref='tech_{0}'.format(PersonalInfoModel.data_prefix),
        primaryjoin=personal_info.c.tech_contact_id == contact_info.c.id, cascade='delete,all'
    ),
    'company_address': relationship(
        AddressModel, uselist=False, backref='company_{0}'.format(PersonalInfoModel.data_prefix),
        primaryjoin=personal_info.c.company_address_id == address.c.id, cascade='delete,all'
    )
})

mapper(ContactModel, contact_info, confirm_deleted_rows=False, properties={
    'phone': relationship(
        PhoneModel, uselist=False, backref='phone_{0}'.format(ContactModel.data_prefix),
        primaryjoin=contact_info.c.phone_id == phone.c.id, cascade='delete,all'
    ),
    'fax': relationship(
        PhoneModel, uselist=False, backref='fax_{0}'.format(ContactModel.data_prefix),
        primaryjoin=contact_info.c.fax_id == phone.c.id, cascade='delete,all'
    )
})

mapper(StaffMemberModel, staff_member, confirm_deleted_rows=False, properties={
    'contact': relationship(ContactModel, uselist=False, backref=StaffMemberModel.data_prefix, cascade='delete,all'),
    'address': relationship(AddressModel, uselist=False, backref=StaffMemberModel.data_prefix, cascade='delete,all'),
    'authentication': relationship(AuthModel, uselist=False, backref=StaffMemberModel.data_prefix, cascade='delete,all')
})

mapper(SubscriptionModel, subscription, confirm_deleted_rows=False, properties={
    'services': relationship(
        ServiceModel, backref=SubscriptionModel.data_prefix, order_by=service.c.id, cascade='delete,all'),
})

mapper(
    AccountModel, account, polymorphic_on=account.c.type, polymorphic_identity='account', confirm_deleted_rows=False,
    properties={
        'info': relationship(
            PersonalInfoModel, uselist=False, backref=AccountModel.data_prefix, cascade='delete,all'),
        'staff_members': relationship(
            StaffMemberModel, backref=AccountModel.data_prefix, order_by=staff_member.c.id, cascade='delete,all'),
        'subscriptions': relationship(
            SubscriptionModel, backref=AccountModel.data_prefix, order_by=subscription.c.id, cascade='delete,all'),
        'accounts': relationship(
            AccountModel, backref=backref('owner', remote_side=[account.c.id]),
            order_by=account.c.id, cascade='delete,all'
        ),
        'service_plans': relationship(ServicePlanModel, order_by=service_plan.c.id, cascade='delete,all')
    }
)

mapper(ResellerModel, account, inherits=AccountModel, polymorphic_identity='reseller')
mapper(ProviderModel, account, inherits=ResellerModel, polymorphic_identity='provider')

mapper(MigrationProjectItemModel, migration_project_item, properties={
    'subscription': relationship(
        SubscriptionModel, uselist=False, backref='subscription_{0}'.format(MigrationProjectItemModel.data_prefix),
        primaryjoin=migration_project_item.c.subscription_id == subscription.c.id, cascade='delete,all'
    ),
})
mapper(MigrationProjectModel, migration_project, properties={
    'items': relationship(
        MigrationProjectItemModel, backref=MigrationProjectModel.data_prefix,
        order_by=migration_project_item.c.id, cascade='delete,all'
    ),
})
# --- Models to DB Tables mapping

_db_engine = None
_sessions = {}


def _get_db_engine(db_config=None):
    global _db_engine

    if not _db_engine:
        config = read_config(OA_MIGRATOR_DB_CONFIG) if not db_config else read_config(db_config)
        db_params = ConfigSection(config, 'db')
        _db_engine = create_engine(
            'postgresql://{username}:{password}@{host}/{dbname}'.format(**db_params)
        )
        if PRINT_DB_QUERIES_TO_LOG:
            logger = logging.getLogger('sqlalchemy.engine')
            logger.setLevel(logging.DEBUG)

    return _db_engine


def init_db_tables():
    metadata.create_all(_get_db_engine())


def set_user_password(runner, port, username, password):
    cmd = 'su - postgres -c "psql --port={port} postgres postgres"'
    runner.sh(
        cmd.format(port=port),
        stdin_content="ALTER USER {username} PASSWORD \'{password}\';\\q\n".format(username=username, password=password),
        log_output=False
    )


def save_to_db(source_system_):
    session = get_db_session()

    session.add(source_system_)
    session.commit()

    # update resellers' accounts and set source_system_id to simplify UI APP data selections
    for reseller_ in source_system_.resellers:
        for account_ in reseller_.accounts:
            account_.source_system_id = source_system_.id

    session.commit()


def get_db_session():
    global _sessions
    thread_id = thread.get_ident()
    if thread_id not in _sessions:
        Session = sessionmaker(bind=_get_db_engine(), autoflush=False)
        _sessions[thread_id] = Session()

    return _sessions[thread_id]
