# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import inspect
import os

from lxml import etree

from migrator.core import InternalError
from migrator.core.data.format import WriterBase, ReaderBase
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.lang.entity import Entity

logger = logging.getLogger(__name__)


class EntityNotFound(Exception):
    pass


class CommonWriterXml(WriterBase):
    SCHEMA_INSTANCE_NS = "http://www.w3.org/2001/XMLSchema-instance"
    SCHEMA_LOCATION_TEMPLATE = "{xml_namespace} file://{xsd_file}"

    def __init__(self, filename):
        """

        :type filename: str
        """
        self._schema_location = None
        self._xml_namespace = None
        self._data_type = None
        self._filename = filename
        self._primitive_array_element_tag = {}
        self._choice_tags = {}

    def write(self, obj):
        attrib = {
            'xmlns:xsi': self.SCHEMA_INSTANCE_NS,
            'xsi:schemaLocation': self._schema_location,
            'xmlns': self._xml_namespace
        }

        logger.debug(_T("SAVING_UNIFIED_DATA_TO_FILENAME").format(data_type=self._data_type, filename=self._filename))
        if self._is_model_entity(obj):
            with etree.xmlfile(self._filename, encoding='UTF-8') as xmltree:
                xmltree.write_declaration()
                self._write_model_entity(obj, xmltree, attrib)
        else:
            raise Exception(_T("ATTEMPTED_TO_SERIALIZE_TYPE_AS").format(type=type(obj).__name__))

    def _write_object(self, key, obj, xmlfile):
        if obj is not None:
            if self._is_primitive(obj):
                self._write_primitive(key, obj, xmlfile)
            elif self._is_model_entity(obj):
                self._write_child_model_entity(key, obj, xmlfile)
            elif self._is_list(obj):
                self._write_list(key, obj, xmlfile)

    @staticmethod
    def _is_primitive(obj):
        return isinstance(obj, (basestring, int, float, bool))

    @staticmethod
    def _is_list(obj):
        return isinstance(obj, list)

    @staticmethod
    def _is_model_entity(obj):
        return isinstance(obj, Entity)

    @staticmethod
    def _write_primitive(key, obj, xmlfile):
        node = etree.Element(key)
        if isinstance(obj, bool):
            node.text = unicode(int(obj))
        else:
            node.text = unicode(obj)

        xmlfile.write(node)

    def _write_list(self, key, obj, xmlfile):
        with xmlfile.element(key):
            for item in obj:
                if self._is_model_entity(item):
                    self._write_model_entity(item, xmlfile)
                elif self._is_primitive(item):
                    self._write_primitive(self._primitive_array_element_tag[key], item, xmlfile)
                else:
                    raise Exception(_T("ARRAYS_MAY_CONTAIN_ONLY_MODEL_ENTITIES").format(type=type(item).__name__))

    def _write_model_entity(self, obj, xmlfile, attrib=None, tag_override=None):
        """

        Writes standalone entity. Used when writing root entity or entity contained in a list
        """
        with xmlfile.element(tag_override or obj.data_prefix, attrib=attrib):
            for constructor_arg_name in inspect.getargspec(type(obj).__init__).args:
                if constructor_arg_name == 'self':
                    continue

                self._write_object(constructor_arg_name, obj.__dict__[constructor_arg_name], xmlfile)

    def _write_child_model_entity(self, key, obj, xmlfile):
        """

        Writes entity as a field of another entity.
        :param key: Name of parent entity field containing target entity
        :param obj: Target entity to serialize
        """
        if key in self._choice_tags:
            # wrap target entity into an
            # intermediate XML element with tag name equal to parent entity field key
            with xmlfile.element(key):
                self._write_model_entity(obj, xmlfile)
        else:
            # override tag name with parent entity's field name
            self._write_model_entity(obj, xmlfile, None, key)


class CommonStreamWriterXml(CommonWriterXml):
    def __init__(self, filename):
        super(CommonStreamWriterXml, self).__init__(filename)

    class Element(object):
        def __init__(self, writer):
            self.writer = writer

        def __call__(self, key, attrib=None):
            self.key = key
            self.attrib = attrib
            return self

        def __enter__(self):
            next(self.writer)
            self.subwriter = self.writer.send((self.key, self.attrib))
            return CommonStreamWriterXml.Element(self.subwriter)

        def __exit__(self, exc_type, exc_val, exc_tb):
            next(self.subwriter)
            self.subwriter.close()

        def write_list(self, items):
            next(self.writer)
            self.writer.send(items)

    def __enter__(self):
        self.xml_stream_writer = self._get_xml_stream_writer()
        return CommonStreamWriterXml.Element(self.xml_stream_writer)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.xml_stream_writer.close()

    def _get_xml_stream_writer(self):
        def _writer(xmltree, key, attrib=None):
            try:
                with xmltree.element(key, attrib):
                    while True:
                        data = yield
                        if type(data) == tuple:
                            yield _writer(xmltree, *data)
                        elif type(data) == list:
                            for item in data:
                                self._write_model_entity(item, xmltree)
                            xmltree.flush()
                            yield
                        else:
                            raise Exception(_T("ATTEMPTED_TO_SERIALIZE_TYPE_AS").format(type=type(data).__name__))

            except GeneratorExit:
                pass

        with etree.xmlfile(self._filename, encoding='UTF-8') as xmltree:
            xmltree.write_declaration()
            base_attrib = {
                'xmlns:xsi': self.SCHEMA_INSTANCE_NS,
                'xsi:schemaLocation': self._schema_location,
                'xmlns': self._xml_namespace
            }
            try:
                while True:
                    key, attrib = yield  # attrib is unused for a while
                    yield _writer(xmltree, key, base_attrib)
            except GeneratorExit:
                pass


class CommonReaderXml(ReaderBase):

    def __init__(self, filename, validate=True):
        self._class_model_list = []
        self._additional_data_prefixes = {}
        self._primitive_arrays = {}
        self._array_tags = []
        self._choice_tags = []
        self._data_type = 'Common'
        self._xml_namespace = None
        self._schema_location = None
        self._xsd_path = "common.xsd"
        self._models = {}
        self._filename = filename

        self._init_reader()
        self._namespaces = {'ns': self._xml_namespace}

        self._load_and_validate_xml_tree(filename, validate)
        self._build_models_dict()

    @classmethod
    @cached
    def get(cls, filename, validate=True):
        return cls(filename, validate)

    def _init_reader(self):
        raise NotImplementedError()

    def read_all(self, model_entity_class):
        entities = []
        try:
            for node in self._get_child_nodes(self.xml_tree, model_entity_class.data_prefix):
                entities.append(self._read_model_entity(node, model_entity_class))

            return entities
        except Exception as e:
            raise Exception(_T("UNABLE_TO_RETRIEVE_THE_ENTITY").format(
                entity=model_entity_class.data_prefix, filename=self._filename, reason=str(e)
            ))

    def _build_models_dict(self):
        for model in self._class_model_list:
            self._models[model.data_prefix] = model

    def _load_and_validate_xml_tree(self, filename, validate):
        try:
            logger.debug(_T("READING_UNIFIED_DATA_FROM_FILENAME").format(
                data_type=self._data_type, filename=self._filename
            ))
            self.xml_tree = etree.parse(filename)
        except Exception as e:
            raise Exception(_T("FAILED_TO_LOAD_UNIFIED_DATA_XML").format(data_type=self._data_type, reason=str(e)))

        if validate:
            try:
                self._validate(self.xml_tree, self._xsd_path)
            except Exception as e:
                raise Exception(_T("FAILED_TO_VALIDATE_UNIFIED_DATA_XML").format(
                    data_type=self._data_type, reason=str(e)
                ))

    def _validate(self, tree, xsdpath):
        logger.debug(_T("VALIDATING_XML_FILE_AGAINST_SCHEMA_AT").format(xsdpath=xsdpath))
        if os.path.isfile(xsdpath):
            try:
                xmlschema_doc = etree.parse(xsdpath)
                xmlschema = etree.XMLSchema(xmlschema_doc)
            except Exception as e:
                raise Exception(_T("FAILED_TO_LOAD_XML_SCHEMA_FROM").format(file=xsdpath, reason=str(e)))

            if not xmlschema.validate(tree):
                raise Exception(_T("UNIFIED_DATA_XML_FORMAT_IS_INVALID").format(
                    data_type=self._data_type, reason=xmlschema.error_log.last_error
                ))
        else:
            raise Exception(_T("XML_SCHEMA_FILE_IS_MISSING_FILE").format(file=xsdpath))

    def _read_object(self, node):
        # get entity class by node tag
        tag = self._strip_namespace_from_tag(node.tag)
        model_entity_class = self._get_entity_class_by_data_prefix(tag)

        # if class is found, read entity model
        if model_entity_class is not None:
            return self._read_model_entity(node)

        if tag in self._additional_data_prefixes:
            return self._read_model_entity(node, self._additional_data_prefixes[tag])

        if tag in self._array_tags:
            return self._read_array(node)

        if tag in self._choice_tags:
            return self._read_choice(node)

        if tag in self._primitive_arrays:
            return self._read_array(node, True)

        if self._is_primitive(node):
            return self._read_primitive(node)

        # add more context when reading data from indexed file (line number, etc)
        raise Exception(_T("DATA_MODEL_CONTAINS_NO_INFORMATION").format(node=tag))

    @staticmethod
    def _strip_namespace_from_tag(tag):
        return tag[tag.find('}')+1:]

    def _read_model_entity(self, node, model_entity_class=None):
        tag = self._strip_namespace_from_tag(node.tag)
        model_entity_class = self._get_entity_class_by_data_prefix(tag) if model_entity_class is None \
            else model_entity_class

        if model_entity_class is None:
            raise Exception(_T("NO_DATA_MODEL_ENTITY_CORRESPONDS").format(tag=tag))

        constructor_args = {}
        for constructor_arg_name in inspect.getargspec(model_entity_class.__init__).args:
            if constructor_arg_name == 'self':
                continue

            child_node = self._get_child_node(node, constructor_arg_name)
            if child_node is not None:
                constructor_args[constructor_arg_name] = self._read_object(child_node)
            else:
                constructor_args[constructor_arg_name] = None

        return model_entity_class(**constructor_args)

    def _read_array(self, node, is_array_of_primitives=False):
        # if node tag is in arrays list, read each of child nodes as a model entity to an array
        entity_list = []
        if is_array_of_primitives:
            for node in node.getchildren():
                entity_list.append(self._read_primitive(node))
        else:
            for node in node.getchildren():
                entity_list.append(self._read_model_entity(node))

        return entity_list

    def _read_choice(self, node):
        child_nodes = node.getchildren()
        children_count = len(child_nodes)
        # if node tag is in choice elements list, it should contain 1 or 0 child nodes
        if children_count != 1:
            # if it's not so, throw exception exception
            raise Exception(_T("THE_ELEMENT_TAG_IS_EXPECTED_TO_CONTAIN").format(tag=node.tag, num=children_count))
        else:
            # if node has a child element, it as a model entity
            return self._read_model_entity(child_nodes[0])

    def _is_primitive(self, node):
        return len(node.getchildren()) == 0

    def _read_primitive(self, node):
        return node.text

    def _get_entity_class_by_data_prefix(self, tag):
        if tag in self._models:
            return self._models[tag]
        else:
            return None

    def _get_child_node(self, node, node_name):
        return node.find('ns:%s' % node_name, namespaces=self._namespaces)

    def _get_child_nodes(self, node, node_name):
        return node.findall('//ns:%s' % node_name, namespaces=self._namespaces)

    def _find_node_by_field(self, node, node_name, filed_name, field_value):
        xpath = './/ns:%s[ns:%s[text()="%s"]]' % (node_name, filed_name, field_value)
        nodes = node.xpath(xpath, namespaces=self._namespaces)
        if len(nodes) == 1:
            return nodes[0]
        elif len(nodes) == 0:
            raise EntityNotFound(_T("NO_ENTITY_WITH_ID_ID_WAS_FOUND").format(entity=node_name, id=field_value))
        else:
            raise InternalError("There are {number} models with xpath '{xpath}, the only is expected'".format(number=len(nodes), xpath=xpath))

    def read_model(self, steps):
        node = self.xml_tree
        try:
            for step in steps:
                node = self._find_node_by_field(node, step[0].data_prefix, step[1], step[2])
            return self._read_model_entity(node)
        except Exception as e:
            step=steps[-1]
            logger.error(_T("UNABLE_TO_RETRIEVE_THE_ENTITY_WITH_ID").format(
                entity=step[0].data_prefix, id=step[2], filename=os.path.abspath(self._filename), reason=str(e)
            ))
            raise
