# (C)2017 Ingram Micro Inc. all Rights Reserved
import errno
import logging
import re

from migrator.core import CommandExecutionError, MigrationError
from migrator.core.language import _T

logger = logging.getLogger(__name__)


# ----------------- public only for migrators implementing their own DNS forwarding with this interface ----------------
class DnsForwardingBase(object):
    """This class defines only the common initialization,

    And declares two public functions that should be defined separately for different services.
    """
    def __init__(self, migrator_server, runner, source_host_id):
        self._migrator_server = migrator_server
        self._runner = runner
        self._source_host_id = source_host_id

    def add_forwarded_zones(self, zones_forwarders):
        raise NotImplementedError()

    def remove_forwarded_zones(self, zones):
        raise NotImplementedError()

    def restart_nameserver(self):
        raise NotImplementedError()

# ---------------------------------------- this file's private code follows --------------------------------------------


class NoneDnsForwarding(DnsForwardingBase):
    def __init__(self):
        super(NoneDnsForwarding, self).__init__(None, None, None)

    def remove_forwarded_zones(self, zones):
        logger.debug("No DNS server present, do need to remove forwarded zones")

    def add_forwarded_zones(self, zones_forwarders):
        logger.debug("No DNS server present, do need to setup zones forwarding")

    def restart_nameserver(self):
        logger.debug("No DNS server present, do need to restart nameserver")


class MsDnsForwarding(DnsForwardingBase):
    """This class implements the 2 public functions of DnsForwardingBase, for Microsoft DNS
    """
    def __init__(self, migrator_server, runner, source_host_id, dnscmd_path):
        super(MsDnsForwarding, self).__init__(migrator_server, runner, source_host_id)
        self._dnscmd_path = dnscmd_path

    def add_forwarded_zones(self, zones_forwarders):
        def add(zone, forwarders):
            """Add zone in slave mode, with specified master server
            Return error string (None if there is no error, actual string if there is a error)
            """
            rc, stdout, stderr = self._runner.run_unchecked(
                self._dnscmd_path, ['/ZoneAdd', zone.encode('idna'), '/Forwarder'] + forwarders
            )
            if rc == 0:
                return None
            # Expected error that we're going to ignore, as it will occur
            # on when we set up DNS forwarding second time and further:
            # Command failed:  DNS_ERROR_FORWARDER_ALREADY_EXISTS     9619  (0x2593)
            # return code is 137, which may be 9609 % 256, but may not. so we check the code in stdout as well
            if rc == 147 and "9619" in stdout:
                return None

            return stdout + stderr

        err_zones = {}
        for zone, forwarders in zones_forwarders.iteritems():
            error = add(zone, forwarders)
            if error is not None:
                errmsg = _T("COULD_NOT_CONFIGURE_FORWARDING_FOR_DNS_ZONE") % (zone, error)
                logger.error(errmsg)
                err_zones[zone] = errmsg

        return err_zones

    def remove_forwarded_zones(self, zones):
        def remove(name):
            """Remove zone from DNS
            Return error string (None if there is no error, actual string if there is a error)
            """
            rc, stdout, stderr = self._runner.run_unchecked(
                self._dnscmd_path, ['/ZoneDelete', name.encode('idna'), '/f']
            )
            if rc == 0:
                return None
            # Expected error that we're going to ignore, as it will occur on when
            # we undo DNS forwarding second time and further:
            # Command failed:  DNS_ERROR_ZONE_DOES_NOT_EXIST     9601  (00002581)
            # return code is 129, which may be 9601 % 256, but may not. so we check the code in stdout as well
            if rc == 129 and "9601" in stdout:
                return None

            return stdout + stderr

        err_zones = {}
        for name in zones:
            error = remove(name)
            if error is not None:
                errmsg = _T("COULD_NOT_REMOVE_ZONE_ISSUE") % (name, error)
                logger.error(errmsg)
                err_zones[name] = errmsg

        return err_zones

    def restart_nameserver(self):
        pass


class BindIndexEditor(object):
    """This class implements editing the BIND zones index file: adding or removing zones from there.
    It either fails (with exception) or succeeds for all specified zones, so does not return the set of failed zones.
    """
    def __init__(self, index_filename, newline):
        self._index_filename = index_filename
        self._newline = newline

    def add_zones_to_index(self, zones_forwarders):
        zones_index = self._read_index()
        new_index = [
            line for line in zones_index
            if not self._line_refers_to_one_of_zones(line, zones_forwarders.keys()) and not len(line) == 0
            ]
        for zone, forwarders in zones_forwarders.iteritems():
            logger.debug(u"Add zone '{zone}' => {forwarders} to index".format(zone=zone, forwarders=forwarders))
            new_index.append(
                """zone "{zone}" {{ type forward; forwarders {{ {fwd}; }}; }};""".format(
                    zone=zone.encode('idna'), fwd=';'.join(forwarders)
                )
            )

        self._write_index(new_index)

    def remove_zones_from_index(self, zones):
        zones_index = self._read_index()
        new_index = [
            line for line in zones_index
            if not self._line_refers_to_one_of_zones(line, zones) and not len(line) == 0
        ]
        self._write_index(new_index)

    def _read_index(self):
        try:
            with open(self._index_filename, 'r') as zones_file:
                zones_index = zones_file.read().split(self._newline)
        except IOError as e:
            if e.errno == errno.ENOENT:
                zones_index = []
            else:
                raise
        return zones_index

    @staticmethod
    def _line_refers_to_one_of_zones(line, zones):
        for zone in zones:
            pattern = u"""^zone "%s" {.*};""" % zone.encode('idna')
            if re.match(pattern, line):
                return True
        return False

    def _write_index(self, zones_index):
        with open(self._index_filename, 'w') as zones_file:
            content = self._newline.join(zones_index)
            if len(content) > 0:
                content += self._newline     # add trailing newline, as .join doesn't
            zones_file.write(content)


class BindForwarding(DnsForwardingBase):
    PATH_SEPARATOR = ""

    def __init__(self, migrator_server, runner, source_host_id, named_conf):
        super(BindForwarding, self).__init__(migrator_server, runner, source_host_id)
        if self.PATH_SEPARATOR not in named_conf:
            raise Exception(_T("NAMED_CONFIG_SHOULD_BE_AN_ABSOLUTE_PATH").format(config=named_conf))

        self._named_conf = named_conf
        self._named_conf_dir = self._named_conf[:self._named_conf.rfind(self.PATH_SEPARATOR) + 1]
        self._index_filename_local = migrator_server.get_session_file_path(u"forwarded_zones.%s" % source_host_id)
        self._index_filename_remote_relative = 'oam_forwarded_zones.conf'
        self._index_filename_remote = u"%s%s" % (self._named_conf_dir, self._index_filename_remote_relative)

    """This class implements the 2 public functions of DnsForwardingBase, for BIND
    As such, it defines, what to do to add or remove zones, leaving unimplemented the HOW to do that
    """
    def add_forwarded_zones(self, zones_forwarders):
        try:
            # different remote path, different exceptions for ENOENT
            logger.debug(
                _T('DOWNLOADING_FORWARDED_DNS_ZONES_INDEX').format(
                    remote_file=self._index_filename_remote,
                    local_file=self._index_filename_local
                )
            )
            self._download_forwarded_zones_index()
            # different prefix and newline
            logger.debug("Add zones to index")
            self._add_zones_to_index(zones_forwarders)
            # different remote path
            logger.debug("Upload forwarded zones index")
            self._upload_forwarded_zones_index()
            # different remote path, include value and implementation how to add it
            logger.debug("Include forwarded zones index to named config")
            self._include_index_to_named_conf()
            # operation did not fail - no failed zones
            return {}
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            logger.error(_T("LOG_COULD_NOT_ADD_SUBDOMAIN_SLAVE_DNS_ZONES_ISSUE"), e)
            return {
                zone: _T("COULD_NOT_CONFIGURE_FORWARDING_FOR_DNS_ZONE") % (zone, e)
                for zone in zones_forwarders
            }

    def restart_nameserver(self):
        self._restart_named()

    def remove_forwarded_zones(self, zones):
        try:
            self._download_forwarded_zones_index()
            self._remove_zones_from_index(zones)
            self._upload_forwarded_zones_index()
            self._restart_named()
            # operation did not fail - no failed zones
            return {}
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            logger.error(_T("LOG_COULD_NOT_REMOVE_SUBDOMAIN_DNS_ZONES"), e)
            return {
                zone: _T("COULD_NOT_REMOVE_SUBDOMAIN_DNS_ZONE") % (zone, e)
                for zone in zones
            }

    def _download_forwarded_zones_index(self):
        raise NotImplementedError()

    def _add_zones_to_index(self, zones_forwarders):
        raise NotImplementedError()

    def _remove_zones_from_index(self, zones):
        raise NotImplementedError()

    def _upload_forwarded_zones_index(self):
        self._runner.upload_file(self._index_filename_local, self._index_filename_remote)
        if not self._runner.file_exists(self._index_filename_remote):
            raise MigrationError(_T('FAILED_TO_UPLOAD_FWD_ZONES_INDEX').format(
                local_file=self._index_filename_local,
                remote_file=self._index_filename_remote,
            ))

    def _include_index_to_named_conf(self):
        raise NotImplementedError()

    def _restart_named(self):
        raise NotImplementedError()


class LinuxBindForwarding(BindForwarding):
    PATH_SEPARATOR = "/"
    """This class implements the 6 private functions of BindForwarding, doing the actual changes, for BIND on Linux
    """
    def __init__(self, migrator_server, runner, source_host_id, named_conf):
        super(LinuxBindForwarding, self).__init__(migrator_server, runner, source_host_id, named_conf)
        self._index_filename_remote_chrooted = self._index_filename_remote[len(self._get_chroot()):]
        self._index_editor = BindIndexEditor(self._index_filename_local, newline='\n')

    def _get_chroot(self):
        rc, stdout, stderr = self._runner.run_unchecked('/bin/grep', ['^[[:space:]]*ROOTDIR=', '/etc/sysconfig/named'])
        if rc != 0:
            return ''

        match = re.match("^[\s]*ROOTDIR=['\"]?(.*?)['\"]?$", stdout)
        if not match:
            return ''

        return match.group(1)

    def _download_forwarded_zones_index(self):
        try:
            self._runner.get_file(self._index_filename_remote, self._index_filename_local)
        except IOError as e:
            if e.errno == errno.ENOENT:
                logger.debug(u'(expected) %s', _T("LOG_EXCEPTION"), exc_info=True)
            else:
                raise

    def _add_zones_to_index(self, zones_forwarders):
        self._index_editor.add_zones_to_index(zones_forwarders)

    def _remove_zones_from_index(self, zones):
        self._index_editor.remove_zones_from_index(zones)

    def _include_index_to_named_conf(self):
        named_cf_remote = self._named_conf
        named_cf_local = self._migrator_server.get_session_file_path(u"named.conf.%s" % (self._source_host_id,))
        self._runner.get_file(named_cf_remote, named_cf_local)

        with open(named_cf_local, 'r') as f:
            named_cf_content = f.read()

        include_str = u"""include "%s";""" % self._index_filename_remote_chrooted
        if include_str not in named_cf_content:
            self._runner.sh(u"""(echo;echo '%s') >> "%s" """ % (include_str, named_cf_remote))

    def _restart_named(self):
        # Shell startup scripts for DNS services
        startup_scripts = ['/etc/rc.d/named', '/etc/init.d/named', '/etc/init.d/bind9']
        for startup_script in startup_scripts:
            if self._runner.file_exists(startup_script):
                cmd = 'sh %s restart' % startup_script
                exit_code, stdout, stderr = self._runner.sh_unchecked(cmd)
                if exit_code != 0:
                    raise CommandExecutionError(
                        message=_T('FAILED_TO_RESTART_BIND'),
                        stdout=stdout,
                        stderr=stderr,
                        command_str=cmd,
                        exit_code=exit_code
                    )
                else:
                    logger.debug(_T("DEBUG_SUCCESSFULLY_RESTARTED_NAMED").format(cmd=cmd))
                    return

        # DNS Services controllable via systemctl
        dns_services = ['named-chroot', 'named', 'bind9']
        for dns_service in dns_services:
            exit_code, stdout, stderr = self._runner.sh_unchecked('systemctl list-units | grep %s.service' % dns_service)
            if exit_code == 0:
                cmd = 'systemctl restart %s' % dns_service
                exit_code, stdout, stderr = self._runner.sh_unchecked(cmd)
                if exit_code != 0:
                    raise CommandExecutionError(
                        message=_T('FAILED_TO_RESTART_BIND'),
                        stdout=stdout,
                        stderr=stderr,
                        command_str=cmd,
                        exit_code=exit_code
                    )
                else:
                    logger.debug(_T("DEBUG_SUCCESSFULLY_RESTARTED_NAMED").format(cmd=cmd))
                    return

        raise NotImplementedError(_T('UNABLE_TO_FIND_BIND_SERVICE'))


class WindowsBindForwarding(BindForwarding):
    PATH_SEPARATOR = "\\"

    """This class implements the 6 private functions of BindForwarding, doing the actual changes, for BIND on Windows
    """
    def __init__(self, migrator_server, runner, source_host_id, named_conf):
        super(WindowsBindForwarding, self).__init__(migrator_server, runner, source_host_id, named_conf)
        self._index_editor = BindIndexEditor(self._index_filename_local, newline='\n\r')

    def _download_forwarded_zones_index(self):
        try:
            self._runner.get_file(self._index_filename_remote, self._index_filename_local)
        except Exception as e:
            if u"NT_STATUS_OBJECT_NAME_NOT_FOUND opening remote file" in str(e):
                logger.debug(u'(expected) %s', _T("LOG_EXCEPTION"), exc_info=True)
            else:
                raise

    def _add_zones_to_index(self, zones_forwarders):
        self._index_editor.add_zones_to_index(zones_forwarders)

    def _remove_zones_from_index(self, zones):
        self._index_editor.remove_zones_from_index(zones)

    def _include_index_to_named_conf(self):
        named_cf_remote = self._named_conf
        named_cf_local = self._migrator_server.get_session_file_path(
            u"named.conf.%s" % (self._source_host_id,)
        )
        self._runner.get_file(named_cf_remote, named_cf_local)

        with open(named_cf_local, 'r') as f:
            named_cf_content = f.read()

        include_str = u"""include "%s";""" % self._index_filename_remote
        # reduce backslash during comparison so paths like "c:\a\\b" and "c:\a\b" are considered the same
        if self._reduce_char(include_str, '\\') not in self._reduce_char(named_cf_content, '\\'):
            # echo: puts an empty line, & is cmd's commands separator
            self._runner.sh(u"""cmd /c "(echo:& echo %s) >> "%s"" """ % (include_str, named_cf_remote))

    @staticmethod
    def _reduce_char(s, char):
        pair = '%s%s' % (char, char)
        while pair in s:
            s = s.replace(pair, char)
        return s

    def _restart_named(self):
        # service named could be stopped - ignore this error. TODO check other errors
        self._runner.sh_unchecked('net stop named')
        self._runner.sh('net start named')
