# (C)2017 Ingram Micro Inc. all Rights Reserved
from __future__ import absolute_import

import errno
import logging
import os
import os.path
import sys


from migrator.core import MigrationError
from migrator.core import MigrationNoContextError
from migrator.core.cli.panels import load_target_panel_class
from migrator.core.connections.connections import Connections
from migrator.core.connections.migrator_server import MigratorServer
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.session.config import read_multithreading_params
from migrator.core.session.global_context import GlobalMigrationContext
from migrator.core.session.migration_list import MigrationList, path_to_migration_list
from migrator.core.session.session_files import CommonSessionFiles
from migrator.core.transfer.execution_wrapper import TransferApiWrapper
from migrator.core.utils import obj, cached
from migrator.core.utils.filetypes.config_utils import ConfigSection, read_config
from migrator.core.utils.filetypes.yaml_utils import read_yaml, pretty_yaml
from migrator.core.utils.lang import profiler
from migrator.core.utils.loggers.report import Report, Problem
from migrator.core.utils.networking.rsync_pool import RsyncPool
from migrator.core.utils.networking.ssh.key_pool import SSHKeyPool
from migrator.core.utils.registry import Registry
from migrator.core.workflow.main import create_workflow, create_base_workflow
from migrator.core.workflow.runner.by_subscription import ActionRunnerBySubscription
from migrator.odin_automation.platform.lib.api.aps.project import get_migration_project
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.settings import \
    MIGRATION_MODE_PRODUCTION, MIGRATION_MODE_TEST_WITH_CONTENT, TargetServiceType, TARGET_PRODUCT_OA

logger = logging.getLogger(__name__)


class LicenseValidationError(MigrationError):
    pass


pretty_yaml(plesk_ops.UserContactInfo, prefix='plesk')
pretty_yaml(plesk_ops.UserGetInfo, prefix='plesk')
pretty_yaml(plesk_ops.UserGenInfo, prefix='plesk')
pretty_yaml(plesk_ops.role.RoleInfo, prefix='plesk')


class GenericMigrator(object):
    """ A version of migrator object for a case when there are no source servers known. """
    def __init__(self, options):

        logger.debug(_T("MIGRATOR_START") % (sys.argv[0], " ".join("'%s'" % arg for arg in sys.argv[1:])))
        logger.info(_T("LOG_INITIALIZE_MIGRATOR"))

        registry = Registry.get_instance()
        self.config = read_config(options.config)

        self.global_context = registry.set_context(self._create_global_context())
        self.global_context.config = self.config
        self.global_context.migrator = self

        if 'project_aps_id' in options:
            self.global_context.migration_project_aps_id = options.project_aps_id

        if 'type' in options:
            self.global_context.migration_type = options.type

        if 'mode' in options:
            if options.mode == MIGRATION_MODE_TEST_WITH_CONTENT:
                self.global_context.migration_mode = MigrationMode.TEST_MIGRATION_WITH_CONTENT
            elif options.mode == MIGRATION_MODE_PRODUCTION:
                self.global_context.migration_mode = MigrationMode.PRODUCTION
            else:
                self.global_context.migration_mode = MigrationMode.TEST_MIGRATION

        self.global_context.migrator_server = self._get_migrator_server()

        self.action_runner = ActionRunnerBySubscription(read_multithreading_params(self.config))
        workflow = create_base_workflow()
        self.workflow = registry.set_workflow(workflow)

        if self.can_initialize_target_server():
            self.target_panel_type = self.config.get('GLOBAL', 'target-type')
            self.global_context.target_panel_type = self.target_panel_type
            self.global_context.target_panel_obj = self._get_target_panel()
            self.conn = self._load_connections_configuration()
            self.global_context.conn = self.conn

    @staticmethod
    def _create_global_context():
        return GlobalMigrationContext()

    @cached
    def _get_migrator_server(self):
        return MigratorServer(self.config)

    def can_initialize_target_server(self):
        """Check, if the config options are suitable for initialization of target server"""
        target_type = ConfigSection(self.config, 'GLOBAL').get('target-type')
        return target_type in TargetServiceType.get_supported_target_service_classes() \
               and TARGET_PRODUCT_OA in self.config.sections()

    @cached
    def _get_target_panel(self):
        panel_class = load_target_panel_class(self.target_panel_type, 'panel', 'Panel')
        return panel_class()

    def _load_connections_configuration(self):
        return Connections(self.global_context, self._get_target_panel())


class Migrator(GenericMigrator):
    def __init__(self, options):
        super(Migrator, self).__init__(options)

        logger.debug(_T("MIGRATOR_START") % (sys.argv[0], " ".join("'%s'" % arg for arg in sys.argv[1:])))
        logger.info(_T("LOG_INITIALIZE_MIGRATOR"))

        self._load_configuration()

        self.read_migration_list = False
        self.target_model = None
        self.execution_wrapper = None

        self.source_servers = self.conn.get_information_servers()
        self.external_db_servers = self.conn.get_external_db_servers()

        self.session_files = self._create_session_files()

        self.global_context.session_files = self.session_files
        self.global_context.execution_wrapper = self._get_execution_wrapper()

        self.global_context.source_servers = self.source_servers
        self.global_context.iter_all_subscriptions = self._iter_all_subscriptions
        self.global_context.ssh_key_pool = SSHKeyPool(self.global_context.session_files.get_path_to_ssh_keys_pool())
        self.global_context.rsync_pool = RsyncPool(self.global_context.session_files.get_path_to_rsync())
        self.global_context.get_rsync = self._get_rsync
        self.global_context.migration_list_data = None
        self.global_context.pre_check_report = Report(_T("REPORT_POTENTIAL_ISSUES"), None)
        self.global_context.dns_forwarding_report = Report(_T("REPORT_DNS_FORWARDING_ISSUES"), None)

        self._create_workflow()

        if self.global_context.migration_project_aps_id:
            self.global_context.migration_project = get_migration_project(self.global_context.migration_project_aps_id)

    def _create_workflow(self):
        registry = Registry.get_instance()
        workflow = create_workflow(self.source_panel_type, self.target_panel_type, self.global_context.migration_type)
        self.workflow = registry.set_workflow(workflow)

    def _get_rsync(self, source_server, target_server, source_ip=None):
        """
        Retrive rsync connection object, that provide ability to transfer files
        between source and target Windows servers
        :type source_server: parallels.core.connections.source_server.SourceServer
        :type target_server: parallels.core.connections.target_servers.TargetServer
        :type source_ip: str
        :rtype: parallels.core.windows_rsync.RsyncControl
        """
        raise NotImplementedError()

    def set_options(self, options):
        self.global_context.options = options

    def configure_profiler(self):
        profiler.configure_report_locations(profiler.get_default_steps_report(), self._get_migrator_server())

    def _create_session_files(self):
        return CommonSessionFiles(self.conn, self._get_migrator_server())

    def _load_configuration(self):
        logger.info(_T("LOG_LOAD_CONFIGURATION"))

        self.target_panel_type = self.config.get('GLOBAL', 'target-type')
        self.source_panel_type = self.global_context.source_type
        logger.debug(_T("DEBUG_TARGET_PANEL"), self.target_panel_type)
        logger.debug(_T("DEBUG_SOURCE_PANEL"), self.source_panel_type)

        global_section = ConfigSection(self.config, 'GLOBAL')

        transfer_domains_modes = ['same', 'separate']
        transfer_domains_mode = global_section.get('transfer-domains-to-subscription', "separate").lower()
        if transfer_domains_mode not in transfer_domains_modes:
            raise MigrationError(
                _T("INVALID_VALUE_TRANSFER_DOMAIN_TO_SUBSCRIPTION") % (
                    transfer_domains_mode, ",".join(transfer_domains_modes)
                )
            )
        self.multiple_webspaces = transfer_domains_mode == 'same'

    def _get_session_file_path(self, filename):
        return self._get_migrator_server().get_session_file_path(filename)

    def read_migration_list_lazy(self, options):
        if not self.read_migration_list:
            try:
                self.global_context.migration_list_data = self._read_migration_list(options)
            except UnicodeDecodeError:
                logger.debug(_T("UNABLE_READ_MIGRATION_LIST_FILE"), exc_info=True)
                raise MigrationError(_T("THERE_ARE_NONUTF_CHARACTERS_IN_MIGRATION"))

            self.read_migration_list = True

        return self.global_context.migration_list_data

    @staticmethod
    def get_migration_list_source_data():
        """Return data from the source server for the creation or validation of the migration list"""

        data = []
        for subscription in PlatformReaderDB().get_subscriptions_list():
            data.append(
                obj(
                    subscription_id=subscription.external_id,
                    subscription_name=subscription.name
                )
            )

        return data

    @staticmethod
    def _read_migration_list(options):
        migration_list = MigrationList()
        if options.migration_list_file is None:
            migration_list.load_from_db()
        else:
            filename = path_to_migration_list()
            migration_list.load_from_file(filename)
        migration_list.verify()

        return migration_list.items()

    def _get_execution_wrapper(self, create_new=False):
        if self.execution_wrapper is None:
            self.execution_wrapper = TransferApiWrapper(self._get_target_model())
        elif self.execution_wrapper.model is None:
            self.execution_wrapper.model = self._get_target_model()

        return self.execution_wrapper

    def _get_target_model(self):
        if os.path.exists(self.session_files.get_path_to_target_model()):
            return self._load_target_model(False)
        else:
            # if model does not exist - do not load it to make possible use execution_wrapper in steps before 'convert'
            return None

    def _load_target_model(self, apply_filtering=True):
        """Either get target model from memory, or load it from file.

        See _write_target_model function comment for more details.
        Also apply filtering if corresponding parameter is set to True for repeatability feature:
        objects (subscriptions, clients, resellers, plans) that have some problems should be
        skipped on all the futher steps.
        """
        if self.target_model is None:
            try:
                self.target_model = read_yaml(self.session_files.get_path_to_target_model())
            except IOError as err:
                if err.errno == errno.ENOENT:
                    raise MigrationError(_T("TARGET_MODEL_FILE_NOT_FOUND"))
                else:
                    raise

        if apply_filtering:
            wrapper = self._get_execution_wrapper()
            return wrapper.get_filtering_model()
        else:
            return self.target_model

    def _iter_all_subscriptions(self):
        global_context = Registry.get_instance().get_context()
        for subscription in global_context.subscriptions_to_migrate:
            yield subscription

    def transfer_wpb_sites(self, options, finalize=False):
        raise NotImplementedError()

    def _finalize(self, finalize, options):
        if finalize:
            self._print_summary(options)
            self._exit_failed_objects()

        wrapper = self._get_execution_wrapper()

        failed_subscriptions_num = 0
        for infos in wrapper.failed_objects.subscriptions.itervalues():
            if any([info.is_critical for info in infos]):
                failed_subscriptions_num += 1

        if (
            failed_subscriptions_num == len(self.global_context.migration_list_data) and
            failed_subscriptions_num > 0
        ):
            self._print_summary(options)
            raise MigrationError(_T("ALL_SUBSCRIPTIONS_ARE_FAILED_MIGRATE_STOPPING"))

    def is_migration_successful(self):
        wrapper = self._get_execution_wrapper()
        failed_objects_lists = [
            wrapper.failed_objects.subscriptions,
            wrapper.failed_objects.resellers,
            wrapper.failed_objects.clients,
            wrapper.failed_objects.plans,
            wrapper.failed_objects.auxiliary_users,
            wrapper.failed_objects.auxiliary_user_roles,
        ]

        objects = [x.values() for x in failed_objects_lists]
        for obj_types in objects:
            for obj_type in obj_types:
                for info in obj_type:
                    if info.severity == Problem.ERROR:
                        return False

        issues_objects_lists = [
            wrapper.issues.subscriptions,
            wrapper.issues.clients,
            wrapper.issues.resellers,
            wrapper.issues.plans,
            wrapper.issues.auxiliary_users,
            wrapper.issues.auxiliary_user_roles
        ]

        objects = [x.values() for x in issues_objects_lists]
        for obj_types in objects:
            for obj_type in obj_types:
                for info in obj_type:
                    if info.problem.severity == Problem.ERROR:
                        return False

        return True

    def _exit_failed_objects(self):
        wrapper = self._get_execution_wrapper()
        failed_objects_lists = [
            wrapper.failed_objects.subscriptions,
            wrapper.failed_objects.resellers,
            wrapper.failed_objects.clients,
            wrapper.failed_objects.plans,
            wrapper.failed_objects.auxiliary_users,
            wrapper.failed_objects.auxiliary_user_roles,
        ]

        objects = [x.values() for x in failed_objects_lists]
        for obj_types in objects:
            for obj_type in obj_types:
                if self._get_report_obj_by_severity(obj_type, Problem.ERROR):
                    raise MigrationNoContextError(_T("FAILED_MIGRATE_SOME_OBJECTS_SEE_LOG"))

    def _print_summary(self, options):
        wrapper = self._get_execution_wrapper()

        logger.info(_T("SUMMARY_1"))

        if self.global_context.post_migration_checks_enabled:
            run_post_migrations_msg = "\n" + _T("RUN_TEST_ALL_AFTER_MIGRATION")
        else:
            run_post_migrations_msg = ""

        logger.info(
            _T("SUBSCRIPTIONS_MIGRATION_STATUS") + run_post_migrations_msg,
            len(self.global_context.migration_list_data) - len(wrapper.failed_objects.subscriptions),
            len(self.global_context.migration_list_data)
        )

        def quote(x):
            return u"'%s'" % (x,)

        failed_objects_types = [
            (_T("FAILED_OBJECTS_SUBSCRIPTIONS"), wrapper.failed_objects.subscriptions, quote),
            (_T("FAILED_OBJECTS_RESELLERS"), wrapper.failed_objects.resellers, quote),
            (_T("FAILED_OBJECTS_CLIENTS"), wrapper.failed_objects.clients, quote),
            (_T("FAILED_OBJECTS_PLANS"), wrapper.failed_objects.plans, lambda p: (
                _T("OWNED_BY_RESELLER") % (p[1], p[0]) if p[0] is not None else _T("OWNED_BY_ADMIN") % p[1]
            )),
            (_T("FAILED_OBJECTS_AUX_USERS"), wrapper.failed_objects.auxiliary_users, lambda a: (
                _T("OWNED_BY_CUSTOMER") % (a[1], a[0],)
            )),
            (_T("FAILED_OBJECTS_AUX_USER_ROLES"), wrapper.failed_objects.auxiliary_user_roles, lambda a: (
                _T("OWNED_BY_CUSTOMER") % (a[1], a[0],)
            ))
        ]

        for name, objects, format_func in failed_objects_types:
            error_description = []
            for obj_type, obj_info_list in objects.iteritems():
                errors = self._get_report_obj_by_severity(obj_info_list, Problem.ERROR)
                if len(errors) > 0:
                    error_description.append(format_func(obj_type))

            if len(error_description) > 0:
                logger.error(
                    _T("FAILED_TO_PERFORM_OPERATION"),
                    len(error_description), name, u', '.join(error_description)
                )

    @staticmethod
    def _get_report_obj_by_severity(obj_list, severity):
        return [info for info in obj_list if info.severity == severity]

    @classmethod
    def _add_report_issues(cls, report, object_info_list):
        for info in object_info_list:
            cls._add_report_item(report, info)

    @classmethod
    def _add_report_item(cls, report, info):
        problem = Problem(None, info.severity, cls._format_object_info(info))
        solution = info.solution if info.solution is not None else ''
        report.add_issue(problem, solution)

    @classmethod
    def _format_object_info(cls, info):
        if info.exception is not None:
            if info.error_message is None:
                return info.exception.message
            elif not isinstance(info.exception, MigrationError):
                return u"%s\nException: %s" % (info.error_message, info.exception)
            else:
                return u"%s\n%s" % (info.error_message, info.exception)
        else:
            return info.error_message

    def get_source_database_server(self, source_id, model):
        """
        :type model: parallels.core.dump.data_model.DatabaseServer
        :type source_id: str
        @moveto SourcePanelExporterBase
        """
        source_db_node = self._get_source_db_node(source_id)
        if source_db_node is None:
            return None

        external_db_server = self.conn.get_external_database_server(
            source_db_node, model.dbtype, model.host, model.port, model.login, model.password
        )
        if external_db_server is not None:
            return external_db_server
        else:
            return source_db_node.get_database_server(model)

    def _get_source_db_node(self, source_id):
        """@moveto SourcePanelExporterBase"""
        return self.conn.get_source_server_object(source_id)

    # ======================== utility functions ==============================

    def _get_source_mail_node(self, subscription_id):
        raise NotImplementedError()

    def _get_source_web_node(self, host_external_id, subscription_id):
        raise NotImplementedError()


