# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import time
import psutil

from lib.core import utils
from lib.core.pidfile import TimeoutPIDLockFile

AGENT_START_STOP_TIMEOUT = 30

logger = logging.getLogger(__name__)


class OperationsMessages(object):
    # 'status' command result messages
    AGENT_RUNNING = 'agent is running (pid {pid})'
    AGENT_STOPPED = 'agent is not running'

    AGENT_PIDFILE_STALE = 'Agent is not running and pid file exists {pid_file}'

    # 'start' command message
    AGENT_START = 'Starting agent: {status}'

    # 'stop' command message
    AGENT_STOP = 'Stopping agent: {status}'

    # 'start' or 'stop' operation status
    STATUS_OK = '[  OK  ]'
    STATUS_FAILED = '[  FAILED  ]'


class AgentCtl(object):
    def __init__(self, config):
        self._config = config

        self._command = self._config['agent_cmd']
        self._pidfile_path = self._config['pid_file']
        self._pidfile = TimeoutPIDLockFile(self._pidfile_path)

    def start(self, force=False):
        if self._pidfile.is_locked():
            if self._is_pidfile_stale():
                if not force:
                    # pid file is stale but no -f specified
                    return 1, OperationsMessages.AGENT_PIDFILE_STALE.format(pid_file=self._pidfile_path)
                else:
                    self._pidfile.break_lock()
            else:
                # agent already running
                return 0, OperationsMessages.AGENT_START.format(status=OperationsMessages.STATUS_OK)

        exit_code = self._run_agent()
        if not exit_code:
            # wait for pid file to appear
            count = 0
            while count < AGENT_START_STOP_TIMEOUT and not self._pidfile.is_locked():
                time.sleep(1)
                count += 1

            exit_code = 0 if self._pidfile.is_locked() else 1

        return exit_code, OperationsMessages.AGENT_START.format(
            status=OperationsMessages.STATUS_FAILED if exit_code else OperationsMessages.STATUS_OK
        )

    def stop(self, force=False):
        if not self._pidfile.is_locked() or self._is_pidfile_stale():
            return 0, OperationsMessages.AGENT_STOPPED

        exit_code = 0
        try:
            process = psutil.Process(self._pidfile.read_pid())
            process.terminate()
            process.wait(AGENT_START_STOP_TIMEOUT)
        except psutil.NoSuchProcess:
            exit_code = 0
        except psutil.TimeoutExpired:
            exit_code = 1
        else:
            self._pidfile.break_lock()

        return exit_code, OperationsMessages.AGENT_STOP.format(
            status=OperationsMessages.STATUS_FAILED if exit_code else OperationsMessages.STATUS_OK
        )

    def status(self):
        if self._pidfile.is_locked():
            if self._is_pidfile_stale():
                return 1, OperationsMessages.AGENT_PIDFILE_STALE.format(pid_file=self._pidfile_path)
            else:
                return 0, OperationsMessages.AGENT_RUNNING.format(pid=self._pidfile.read_pid())

        return 3, OperationsMessages.AGENT_STOPPED

    def _run_agent(self):
        raise NotImplementedError()

    def _is_pidfile_stale(self):
        result = False

        pidfile_pid = self._pidfile.read_pid()
        if pidfile_pid is not None:
            try:
                psutil.Process(pidfile_pid)
            except psutil.NoSuchProcess as e:
                logger.debug(e)
                result = True
            else:
                # TODO: add more paranoia and check that running process is really our daemon
                pass

        return result


def get_agent_ctl_class():
    if utils.is_windows():
        from .platform.win32.ctl_impl import AgentCtlWindows
        return AgentCtlWindows
    else:
        from .platform.linux.ctl_impl import AgentCtlLinux
        return AgentCtlLinux
