# (C)2017 Ingram Micro Inc. all Rights Reserved
import errno
import os
import sys
import resource

from lib.core import AgentError
from lib.core.agent import Agent, redirect_stream

MAXFD = 1024


class AgentLinux(Agent):
    def _daemonize(self):
        # Change umask to 0
        change_file_creation_mask(0)

        # Change working dir
        change_working_directory(self._config['agent_dir'])

        # Detach process from parent
        # TODO: add option to not detach for debugging purpose
        detach_process()

        # TODO: set signal handlers to catch SIGTERM and terminate gracefully

        # Close all open file descriptors excluding stdin, stdout, stderr
        exclude_fds = self._get_exclude_fds()
        close_all_open_files(exclude_fds)

        # Redirect std streams to /dev/null
        redirect_stream(sys.stdin, None)
        redirect_stream(sys.stdout, None)
        redirect_stream(sys.stderr, None)

    def _get_exclude_fds(self):
        files = [item for item in [sys.stdin, sys.stdout, sys.stderr] if hasattr(item, 'fileno')]

        exclude_fds = set()
        for item in files:
            if not item:
                continue

            fd = get_fd(item)
            if fd is None:
                exclude_fds.add(item)
            else:
                exclude_fds.add(fd)

        return exclude_fds


def change_working_directory(directory):
    try:
        os.chdir(directory)
    except Exception as e:
        raise AgentError('Unable to change working directory: {exc}'.format(exc=e))


def change_file_creation_mask(mask):
    try:
        os.umask(mask)
    except Exception as e:
        raise AgentError('Unable to change file creation mask: {exc}'.format(exc=e))


def detach_process():
    def fork_then_exit_parent(error_message):
        try:
            pid = os.fork()
            if pid > 0:
                os._exit(0)
        except OSError as e:
            raise AgentError('{msg}: {exc}'.format(msg=error_message, exc=e))

    fork_then_exit_parent(error_message='Failed to fork first time')
    os.setsid()
    fork_then_exit_parent(error_message='Failed to fork second time')


def get_max_fd():
    result = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    return result if result != resource.RLIM_INFINITY else MAXFD


def close_fd(fd):
    try:
        os.close(fd)
    except EnvironmentError as e:
        if e.errno == errno.EBADF:
            # File descriptor was not open
            pass
        else:
            raise AgentError('Failed to close file descriptor {fd}: {exc}'.format(fd=fd, exc=e))


def close_all_open_files(exclude=set()):
    maxfd = get_max_fd()
    for fd in reversed(range(maxfd)):
        if fd not in exclude:
            close_fd(fd)


def get_fd(obj):
    fd = None
    if hasattr(obj, 'fileno'):
        try:
            fd = obj.fileno()
        except ValueError:
            pass

    return fd
