# (C)2017 Ingram Micro Inc. all Rights Reserved
import os
import logging
import ntpath

from collections import namedtuple

import migrator.core.product

from migrator import settings
from migrator.core.language import _T
from migrator.core.utils.filetypes.ssl_keys import generate_agent_ssl_certificate, get_oam_ssl_certificate
from migrator.core.utils.filetypes.yaml_utils import pretty_yaml
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.operating_system import unix_thirdparty, windows_thirdparty
from migrator.core.utils.operating_system import windows_utils
from migrator.core.utils.operating_system.unix_utils import join_package_path

logger = logging.getLogger(__name__)


class ProductAgentInterface(object):
    def install(self):
        raise NotImplementedError()

    @property
    def is_installed(self):
        return False

    @property
    def agent_dir(self):
        raise NotImplementedError()

    def uninstall(self):
        raise NotImplementedError()


class ProductAgentServiceInterface(object):
    STATUS_STOPPED = 0
    STATUS_RUNNING = 1
    STATUS_DEAD = 2

    def start(self):
        raise NotImplementedError()

    def stop(self, force=False):
        raise NotImplementedError()

    def status(self):
        raise NotImplementedError()


def is_agent_service(obj):
    return isinstance(obj, ProductAgentServiceInterface)


class ProductAgentService(ProductAgentInterface, ProductAgentServiceInterface):
    DST_PREFIX_LINUX = '/opt/oam'
    DST_PREFIX_WINDOWS = '\\oam'

    def __init__(self, server):
        self._server = server
        self._runner_ctx = server.runner
        self._host_ip = server.ip()

        self._dst_prefix = self.DST_PREFIX_WINDOWS if self._server.is_windows() else self.DST_PREFIX_LINUX
        self._is_installed_marker = os.path.join(self._dst_prefix, '.installed')

        _dst_agent_dir = os.path.join(self._dst_prefix, 'agent')
        _dst_agent_libdir = os.path.join(_dst_agent_dir, 'lib')
        _dst_agent_confdir = os.path.join(_dst_agent_dir, 'conf')

        self._python = dict(src_file='', dst_file='', distfile='')
        self._agent_core = dict(
            src_dir=get_agent_file_path(migrator.core.product, ''),
            dst_dir=_dst_agent_dir,
            dst_conf_dir=_dst_agent_confdir,
            src_config=get_agent_file_path(migrator.core.product, 'conf/agent.config.sample'),
            dst_config=os.path.join(_dst_agent_confdir, 'agent.config'),
            agentctl='',
        )
        self._agent_product_lib = dict(src_dir='', dst_dir=_dst_agent_libdir)

        if self._server.is_windows():
            self._set_win_agent_params()
        else:
            self._set_lin_agent_params()

    def _set_lin_agent_params(self):
        self._python['distfile'] = os.path.basename(unix_thirdparty.get_python_dist())
        self._python['src_file'] = unix_thirdparty.get_python_dist()
        self._python['dst_file'] = os.path.join(self._dst_prefix, self._python['distfile'])

        self._agent_core['agentctl'] = os.path.join(self._agent_core['dst_dir'], 'agentctl')

    def _set_win_agent_params(self):
        self._python['distfile'] = os.path.basename(windows_thirdparty.get_python_dist())
        self._python['src_file'] = windows_thirdparty.get_python_dist()
        self._python['dst_file'] = ntpath.join(self._dst_prefix, self._python['distfile'])

        self._agent_core['agentctl'] = os.path.join(self._agent_core['dst_dir'], 'agentctl.bat')

    @trace
    def install(self):
        self.uninstall()

        logger.debug('Installing agent...')
        # TODO: check free disk space before upload?

        self._create_prefix_dir()
        self._deploy_python()
        self._upload_agent_core()
        self._upload_agent_product_lib()
        self._configure_agent()
        self._mark_installation_complete()

        logger.debug('Agent installation complete')

    @property
    def is_installed(self):
        logger.debug('Checking whether agent is already installed and configured...')
        with self._runner_ctx() as runner:
            res = runner.file_exists(self._is_installed_marker)
            logger.debug('Result: %s', 'TRUE' if res else 'FALSE')

        return res

    @property
    def agent_dir(self):
        return self._agent_core['dst_dir']

    @trace
    def uninstall(self):
        logger.debug('Uninstalling agent...')

        try:
            self.stop(True)
        except:
            pass

        self._clean_directory(self._dst_prefix)

        logger.debug('Agent removal complete')

    @trace
    def start(self, force=False):
        logger.debug('Starting migration agent service...')
        exit_code, _, stderr = self._agent_ctl('start', force)
        if exit_code != 0:
            raise Exception(
                'Migration agent service start operation failed on host {host}. {error_message}'.format(
                    host=self._host_ip, error_message=stderr.splitlines()[0] if stderr else '(Unknown)'
                )
            )
        logger.debug('Agent started')

    @trace
    def stop(self, force=False):
        logger.debug('Stopping migration agent service...')
        exit_code, _, stderr = self._agent_ctl('stop', force)
        if exit_code != 0:
            raise Exception(
                'Migration agent service stop operation failed on host {host}. {error_message}'.format(
                    host=self._host_ip, error_message=stderr.splitlines()[0] if stderr else '(Unknown)'
                )
            )
        logger.debug('Agent stopped')

    @trace
    def status(self):
        logger.debug('Getting migration agent service status...')
        exit_code_to_status_map = {
            0: self.STATUS_RUNNING,
            1: self.STATUS_DEAD,
            3: self.STATUS_STOPPED,
        }
        exit_code, _, _ = self._agent_ctl('status')
        if exit_code not in exit_code_to_status_map:
            raise Exception("Unexpected exit code returned as migration agent status '%s'" % exit_code)

        return exit_code_to_status_map.get(exit_code)

    @trace
    def _agent_ctl(self, command, force=False):
        if not force:
            self._check_installed()

        cmd = '{ctl} %s{command}' % ('-f ' if force else '')
        args = dict(
            ctl=self._agent_core['agentctl'],
            command=command,
        )
        with self._runner_ctx() as runner:
            try:
                return runner.sh_unchecked(cmd, args)
            except Exception as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                raise Exception("Migration agent control utility invocation failed with unexpected error '%s'" % str(e))

    def _check_installed(self):
        if not self.is_installed:
            raise Exception(
                'Unexpected error. Migration agent is not installed on host {host}'.format(host=self._host_ip)
            )

    @trace
    def _create_prefix_dir(self):
        logger.debug('Creating base dir on remote host: %s', self._dst_prefix)
        with self._runner_ctx() as runner:
            runner.mkdir(self._dst_prefix)

            # Secure _dst_prefix dir
            if not self._server.is_windows():
                runner.chown(self._dst_prefix, 'root', 'root')
                runner.chmod(self._dst_prefix, 0750)
            else:
                remote_spath_bin = self._server.get_session_file_path('spath.exe')
                runner.run(remote_spath_bin, [self._dst_prefix])

    @trace
    def _deploy_python(self):
        logger.debug('Deploying portable Python...')

        self._upload_python()
        self._unpack_python()

        logger.debug('Done')

    @trace
    def _upload_agent_core(self):
        logger.debug('Deploying agent core...')

        self._upload_directory(self._agent_core['src_dir'], self._agent_core['dst_dir'])

        logger.debug('Done')

    @trace
    def _upload_agent_product_lib(self):
        if not self._agent_product_lib['src_dir']:
            # Do nothing if no product specific library is configured for the agent
            return

        logger.debug('Deploying agent product library...')

        self._upload_directory(self._agent_product_lib['src_dir'], self._agent_product_lib['dst_dir'])

        logger.debug('Done')

    @trace
    def _configure_agent(self):
        logger.debug('Configuring agent...')

        self._set_files_permissions()
        self._install_ssl_certificates()
        self._install_config_file()

        logger.debug('Done')

    @trace
    def _mark_installation_complete(self):
        logger.debug('Finalizing agent installation...')

        self._upload_file_content(self._is_installed_marker, '', '0640')

        logger.debug('Done')

    @trace
    def _set_files_permissions(self):
        logger.debug('Setting correct permissions on the agent files...')

        with self._runner_ctx() as runner:
            if not self._server.is_windows():
                runner.sh('chmod -R o-rwx {agent_dir}'.format(agent_dir=self._agent_core['dst_dir']))
                runner.sh('chmod u+x {agent_ctl}'.format(agent_ctl=self._agent_core['agentctl']))
            else:
                # Do nothing in case of windows since all the agent and portable python files
                # will inherit secure permissions from self._dst_prefix dir
                pass

    @trace
    def _install_ssl_certificates(self):
        logger.debug('Installing agent SSL certificates...')

        agent_cert = None
        try:
            oam_cert = get_oam_ssl_certificate()
            agent_cert = generate_agent_ssl_certificate()

            certs_to_upload = (
                ('agent_cert.crt', agent_cert.cert_filename),
                ('agent_cert.key', agent_cert.pkey_filename),
                ('client.crt', oam_cert.cert_filename),
            )

            for cert_data in certs_to_upload:
                self._upload_file_content(
                    os.path.join(self._agent_core['dst_conf_dir'], cert_data[0]),
                    self._get_file_contents(cert_data[1]),
                    '0600'
                )
        finally:
            if agent_cert:
                agent_cert.remove()

        logger.debug('Done')

    @trace
    def _install_config_file(self):
        logger.debug('Creating agent configuration file: %s', self._agent_core['dst_config'])

        configuration = self._get_file_contents(self._agent_core['src_config'])
        configuration = self._config_set_agent_command(configuration)
        configuration = self._config_set_agent_port(configuration)

        self._upload_file_content(self._agent_core['dst_config'], configuration, '0600')

        logger.debug('Done')

    @trace
    def _upload_python(self):
        logger.debug('Uploading portable Python distribution...')
        with self._runner_ctx() as runner:
            runner.upload_file(self._python['src_file'], self._python['dst_file'])

        logger.debug('Done')

    @trace
    def _unpack_python(self):
        logger.debug('Unpacking portable Python distribution...')

        if self._server.is_windows():
            self._unpack_win_python()
        else:
            self._unpack_lin_python()

        logger.debug('Done')

    @trace
    def _unpack_lin_python(self):
        with self._runner_ctx() as runner:
            runner.sh('cd {basedir} && umask 007 && tar --no-same-permissions -xjf {distfile}'.format(
                basedir=self._dst_prefix, distfile=self._python['distfile']
            ))
            runner.remove_file(self._python['dst_file'])

    @trace
    def _unpack_win_python(self):
        with self._runner_ctx() as runner:
            # 1. Upload unzip.exe
            unzip_bin = windows_thirdparty.get_unzip_bin()
            dst_unzip_bin = ntpath.join(self._dst_prefix, os.path.basename(unzip_bin))
            runner.upload_file(unzip_bin, dst_unzip_bin)

            # 2. Unzip portable python distribution
            runner.sh('cmd /C "cd {basedir} && unzip.exe {distfile}"'.format(
                basedir=self._dst_prefix, distfile=self._python['distfile']
            ), log_output=False)

            # 3. Remove python dist & unzip.exe
            runner.remove_file(self._python['dst_file'])
            runner.remove_file(dst_unzip_bin)

    @trace
    def _clean_directory(self, remote_dir, remove_root=False):
        logger.debug('Cleaning remote dir: %s', remote_dir)

        if self._server.is_windows():
            self._clean_win_directory(remote_dir, remove_root)
        else:
            self._clean_lin_directory(remote_dir, remove_root)

        logger.debug('Done')

    @trace
    def _clean_lin_directory(self, remote_dir, remove_root=False):
        with self._runner_ctx() as runner:
            if not runner.file_exists(remote_dir):
                return

            # 1. Remove all symlinks first to avoid remote_dir escaping
            runner.sh('find {remote_dir} -type l -exec unlink {ref} \\;'.format(remote_dir=remote_dir, ref='{}'))

            # 2. Remove all files
            runner.sh('find {remote_dir} -type f -exec rm -f {ref} \\;'.format(remote_dir=remote_dir, ref='{}'))

            # 3. Remove all dirs
            runner.sh('find {remote_dir} -type d -depth -exec rmdir {ref} \\;'.format(remote_dir=remote_dir, ref='{}'))

            if remove_root:
                runner.sh('rmdir {remote_dir}'.format(remote_dir=remote_dir))

    @trace
    def _clean_win_directory(self, remote_dir, remove_root=False):
        with self._runner_ctx() as runner:
            if not runner.file_exists(remote_dir):
                return

            runner.remove_directory(remote_dir, remove_root, log_output=False)

    @trace
    def _upload_directory(self, local_dir, remote_dir):
        """Upload the given directory to the source server."""
        local_dir = local_dir.rstrip('/')

        with self._runner_ctx() as runner:
            logger.debug('Creating remote dir: %s', remote_dir)
            runner.mkdir(remote_dir)

            for src_path, dst_path in self._get_upload_directory_files_list(local_dir, remote_dir):
                if src_path.endswith('/'):
                    logger.debug('Creating remote dir: %s', dst_path)
                    runner.mkdir(dst_path)
                else:
                    logger.debug('Uploading file: %s -> %s', src_path, dst_path)
                    runner.upload_file(src_path, dst_path)

    @trace
    def _get_upload_directory_files_list(self, src_dir, remote_dir):
        if self._server.is_windows():
            return self._get_win_upload_directory_files_list(src_dir, remote_dir)
        else:
            return self._get_lin_upload_directory_files_list(src_dir, remote_dir)

    @trace
    def _get_lin_upload_directory_files_list(self, src_dir, remote_dir):
        src_dir = src_dir.rstrip('/')
        remote_dir = remote_dir.rstrip('/')

        res = []
        local_paths = self._list_local_dir(src_dir)
        for src_path in local_paths:
            res.append((src_path, os.path.join(remote_dir, src_path[len(src_dir)+1:])))

        return res

    @trace
    def _get_win_upload_directory_files_list(self, src_dir, remote_dir):
        src_dir = src_dir.rstrip('/')
        remote_dir = remote_dir.rstrip('\\/')

        res = []
        local_paths = self._list_local_dir(src_dir)
        for src_path in local_paths:
            dst_path = windows_utils.fix_path(ntpath.join(remote_dir, src_path[len(src_dir)+1:]))
            res.append((src_path, dst_path))

        return res

    @trace
    def _upload_file_content(self, filename, content, perms='0644'):
        with self._runner_ctx() as runner:
            if not self._server.is_windows():
                runner.upload_file_content(filename, '')
                runner.sh('chmod {perms} {filename}'.format(perms=perms, filename=filename))

            runner.upload_file_content(filename, content)

    @staticmethod
    @trace
    def _list_local_dir(dirname):
        dirname = dirname.rstrip('/')

        res = []
        for root, _, files in os.walk(dirname):
            res.append('%s/' % root)
            for filename in files:
                res.append(os.path.join(root, filename))

        return res

    @staticmethod
    @trace
    def _config_set_agent_port(configuration):
        return configuration.replace(';port', 'port = {port}\n;port'.format(port=settings.AGENT_PORT))

    def _config_set_agent_command(self, configuration):
        if self._server.is_windows():
            anchor_substring = ';agent_win_cmd'
        else:
            anchor_substring = ';agent_lin_cmd'

        return configuration.replace(anchor_substring, 'agent_cmd')

    @staticmethod
    @trace
    def _get_file_contents(fname):
        with open(fname, 'rb') as f:
            return f.read()


class DumpAll(namedtuple('DumpAll', [])):
    def covers(self, selection):
        """Check if this selection covers another selection

        If true, that means that backup dump created by current
        selection contains all objects from another selection, and so
        could be used.
        """
        return True

    def __repr__(self):
        return 'DumpAll()'

DumpAll = pretty_yaml(DumpAll)


class DumpSelected(namedtuple('DumpSelected', ('resellers', 'clients', 'domains'))):
    def covers(self, selection):
        """Check if this selection covers another selection

        If true, that means that backup dump created by current
        selection contains all objects from another selection, and so
        could be used.
        """
        if isinstance(selection, DumpSelected):
            return (
                set(self.resellers) >= set(selection.resellers) and
                set(self.clients) >= set(selection.clients) and
                set(self.domains) >= set(selection.domains)
            )
        else:
            return False

    def __repr__(self):
        return "DumpSelected(%r, %r, %r)" % (
            self.resellers, self.clients, self.domains
        )

DumpSelected = pretty_yaml(DumpSelected)


def get_agent_file_path(package, relative_path):
    return join_package_path(package, 'agent', relative_path)

