# (C)2017 Ingram Micro Inc. all Rights Reserved
"""
This module contains classes that retrieve high-level information about
the hosting panel installed on a given host.
"""
import logging
from collections import OrderedDict

from migrator.core.language import _T
from migrator.core.product.export import get_product_exporter, get_product_content_exporter
from migrator.settings import SERVICE_EXPORT_NOT_SUPPORTED

logger = logging.getLogger(__name__)


class ProductBase(object):
    """Common API functions for hosting panel product"""

    @classmethod
    def get_agent_instance(cls):
        raise NotImplementedError()

    @classmethod
    def get_exporter_instance(cls, service):
        """
        :param str service: Examples: 'billing', 'shm'
        :return:
        """
        product_exporter_class = get_product_exporter(service, cls.get_supported_services_exporters_map())
        return product_exporter_class if product_exporter_class is SERVICE_EXPORT_NOT_SUPPORTED else product_exporter_class()

    @classmethod
    def get_content_exporter_instance(cls, service):
        product_content_exporter_class = get_product_content_exporter(service, cls.get_supported_services_exporters_map())
        return product_content_exporter_class if product_content_exporter_class is SERVICE_EXPORT_NOT_SUPPORTED else product_content_exporter_class()

    @classmethod
    def get_dns_zone_manager(cls):
        raise NotImplementedError()

    @classmethod
    def suspend_subscription(cls, subscription):
        raise NotImplementedError()

    @classmethod
    def get_id(cls):
        raise NotImplementedError()

    @classmethod
    def get_supported_services_exporters_map(cls):
        raise NotImplementedError()


class PanelInfoBase(object):
    def __init__(self, runner, config_section=None):
        self.runner = runner
        self._config_section = config_section

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def short_name(self):
        raise NotImplementedError()

    @property
    def version(self):
        """Return product version"""
        raise NotImplementedError()

    @property
    def patch(self):
        """Return product patch"""
        raise NotImplementedError()

    @property
    def url(self):
        """Return url to the login to Control Panel"""
        raise NotImplementedError()

    @property
    def services(self):
        """Return a list of services"""
        return self._get_service_info()

    @staticmethod
    def is_installed(runner):
        """Determine, if the panel is installed on the server pointed by runner"""
        return False

    def get_service_list(self):
        """
        Get dictionary of various critical services related to Control Panel which need to be checked

        Returns an OrderedDict:
         - key: display name of a service
         - value: {
             'name_list': list of possible system specific service names (eg. (apache2, httpd))
             'func': alternate function to use instead of _get_service_status_any() to get service status
          }

        Alternate function should have the following signature:

         def my_custom_checker_for_some_service(list_of_already_checked_services):
             pass

        """
        raise NotImplementedError()

    def get_service_status(self, service_name):
        """
        Get status of a system service.

        Returns a tuple:
         - Status info value string, one of: 'Running', 'Stopped', 'Not installed'
         - exit code: 0 for running service, non-zero for other status (a service-specific value)

        """
        raise NotImplementedError()

    @classmethod
    def factory(cls, runner, config_section):
        """Create an instance of the class"""
        return cls(runner)

    def _get_service_info(self):
        """Get status of system services that hosting panel depends upon"""
        service_status_list = OrderedDict()

        services_list = self.get_service_list()
        for key, value in services_list.iteritems():
            service_name_list = value.get('name_list', ())
            func = value.get('func')

            if callable(func):
                service, status, exit_code, error_message = func(service_status_list)
            else:
                service, status, exit_code, error_message = self._get_service_status_any(service_name_list)

            service_status_list[key] = dict(service=service, status=status, exit_code=exit_code)
            if error_message is not None:
                service_status_list[key]['error_message'] = error_message

        return service_status_list

    def _get_service_status_any(self, service_name_list):
        """
        Check, if any of the services is running.

        Function accepts a list of service aliases. If no services from the list are started,
        the function returns the name of the first available service that could be started.

        If none of the services is available, return ('Unknown', 'Stopped')
        Return a tuple
         - service name
         - status string
         - exit code returned by service status command
         - error message
        """
        exit_code = 1
        error_message = None
        found_service_name = service_name_list[0] if len(service_name_list) == 1 else 'Unknown'

        for service_name in service_name_list:
            service_status, exit_code = self.get_service_status(service_name)
            if service_status == _T("SERVICE_STATUS_RUNNING"):
                return service_name, _T("SERVICE_STATUS_RUNNING"), exit_code, error_message
            elif service_status == _T("SERVICE_STATUS_STOPPED") and found_service_name == 'Unknown':
                found_service_name = service_name

        return found_service_name, _T("SERVICE_STATUS_STOPPED"), exit_code, error_message


class UnknownPanelInfo(PanelInfoBase):
    UNKNOWN = 'Unknown'
    UNKNOWN_VERSION = '0.0.0'

    @property
    def name(self):
        return self.UNKNOWN

    @property
    def short_name(self):
        return self.UNKNOWN

    @staticmethod
    def is_installed(runner):
        return False

    @property
    def url(self):
        return self.UNKNOWN

    @property
    def version(self):
        return self.UNKNOWN_VERSION

    @property
    def patch(self):
        return None

    @property
    def services(self):
        return {}

    def get_service_list(self):
        return {}
