# (C)2017 Ingram Micro Inc. all Rights Reserved
import codecs
import json
import os
from collections import namedtuple

from migrator.core import MigrationError
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.registry import Registry


class MigrationList(object):
    """ The class for operating the migration list: writing to the migration list and reading data from the migration
    list and the database, and checking for data validity. """

    def __init__(self):
        self._items = None

    def load_from_file(self, filename):
        """
        Reading data from the migration list file.

        :param str filename: migration list file name
        """

        # 1. Check that the migration file list exists.
        if not(os.path.exists(filename) and os.path.isfile(filename)):
            raise MigrationError(_T("MIGRATION_LIST_DOES_NOT_EXISTS").format(filename=filename))

        # 2. Check if it is possible to open and parse the migration list file.
        with codecs.open(filename, 'r', 'utf-8') as f:
            try:
                data = json.load(f)
                self._items = self._get_items_from_migration_list(data)
            except ValueError as e:
                raise MigrationError(
                    _T("UNABLE_TO_PARSE_MIGRATION_LIST_FILE").format(filename=filename, error=e)
                )

    def load_from_db(self):
        """
        Reading migration data from the db.
        """

        self._items = []
        for migrated_items in PlatformReaderDB().get_migration_project_items():
            mapping_subscription_id = migrated_items.mapping_subscription_id
            if migrated_items.is_oa_subscription:
                mapping_subscription_id = int(mapping_subscription_id)

            self._items.append(
                MigrationListData(
                    id=migrated_items.subscription.external_id,
                    plan_id=migrated_items.target_service_plan_id,
                    mapping_subscription_id=mapping_subscription_id,
                    is_oa_subscription=migrated_items.is_oa_subscription,
                    mapping_reseller_id=migrated_items.mapping_reseller_id,
                    is_oa_account=migrated_items.is_oa_account
                )
            )

    def items(self):
        """
        Return the list of migrated items in the format [MigrationListData(), ...]

        :return: list[MigrationListData]
        """

        return self._items

    def write(self, filename):
        """
        Generate migration objects and write to the migration list file

        :param str filename: migration list file name
        """

        migration_list_content = json.dumps(self._items, indent=4, sort_keys=True)
        with codecs.open(filename, 'w', 'utf-8') as f:
            f.write(migration_list_content)

    def verify(self):
        """
        Checking whether all required parameters for subscriptions are provided in the migration list file. Return None
        or raise MigrationError exception.
        """

        parent_subscriptions = []
        child_subscriptions = []
        errors = []
        # Check the data specified in the migration list file.
        for subscription in self._items:
            if not subscription.is_oa_subscription:
                if subscription.id in parent_subscriptions and subscription.id != subscription.mapping_subscription_id:
                    errors.append(_T("SUBSCRIPTION_WITH_ID_CANNOT_BE_ASSOCIATED").format(
                        subscription_id=subscription.id
                    ))
                elif subscription.mapping_subscription_id in child_subscriptions and\
                    subscription.mapping_subscription_id not in parent_subscriptions:
                    errors.append(_T("SUBSCRIPTION_WITH_ID_CANNOT_BE_MAPPED").format(
                        subscription_id=subscription.id, mapping_subscription_id=subscription.mapping_subscription_id
                    ))
                else:
                    child_subscriptions.append(subscription.id)
                    parent_subscriptions.append(subscription.mapping_subscription_id)
            else:
                if subscription.id in parent_subscriptions:
                    errors.append(_T("SUBSCRIPTION_WITH_ID_CANNOT_BE_ASSOCIATED").format(
                        subscription_id=subscription.id
                    ))
                else:
                    child_subscriptions.append(subscription.id)

        if errors:
            raise MigrationError(_T("MAPPING_OF_SUBSCRIPTIONS_IS_INCORRECTLY_DEFINED").format(errors='\n'.join(errors)))

    def _get_items_from_migration_list(self, data):
        """
        Checking whether all required parameters for subscriptions are provided in the migration list file and return
        migration items. Return list[MigrationListData] or raise MigrationError exception.
        """

        items = []
        subscription_number = 0
        subscription_name_list = []
        errors = []

        # Check the data specified in the migration list file.
        for subscription in data:
            subscription_number += 1

            subscription_id = subscription.get('id', None)
            if subscription_id is None:
                errors.append(_T("SUBSCRIPTION_ID_IS_NOT_DEFINED").format(subscription_number=subscription_number))

            plan_id = subscription.get('plan_id', None)
            if plan_id is None:
                errors.append(_T("PLAN_FOR_SUBSCRIPTION_IS_NOT_DEFINED").format(
                    subscription_number=subscription_number,
                    subscription_id=subscription_id
                ))

            if subscription_id in subscription_name_list:
                errors.append(_T("SUBSCRIPTION_ALREADY_DEFINED_1").format(
                    subscription_number=subscription_number,
                    subscription_id=subscription_id
                ))

            mapping_subscription_id = subscription.get('mapping_subscription_id', None)
            if mapping_subscription_id is None:
                mapping_subscription_id = subscription_id
                is_oa_subscription = False
            else:
                is_oa_subscription = subscription.get('is_oa_subscription', None)
                if is_oa_subscription is None:
                    errors.append(_T("SUBSCRIPTION_WITH_ID_IS_MAPPED_TO_SUBSCRIPTION_1").format(
                        subscription_id=subscription_id,
                        mapping_subscription_id=mapping_subscription_id
                    ))
                else:
                    is_oa_subscription = bool(is_oa_subscription)
                    if is_oa_subscription:
                        mapping_subscription_id = int(mapping_subscription_id)

            mapping_reseller_id = subscription.get('mapping_reseller_id', None)
            if mapping_reseller_id is None:
                is_oa_account = False
            else:
                is_oa_account = subscription.get('is_oa_account', None)
                if is_oa_account is None:
                    errors.append(_T("SUBSCRIPTION_WITH_ID_IS_MAPPED_TO_ACCOUNT").format(
                        subscription_id=subscription_id,
                        mapping_account_id=mapping_reseller_id
                    ))
                else:
                    is_oa_account = bool(is_oa_account)
                    if is_oa_account:
                        mapping_reseller_id = int(mapping_reseller_id)

            subscription_name_list.append(subscription_id)
            items.append(
                MigrationListData(
                    id=subscription_id,
                    plan_id=plan_id,
                    mapping_subscription_id=mapping_subscription_id,
                    is_oa_subscription=is_oa_subscription,
                    mapping_reseller_id=mapping_reseller_id,
                    is_oa_account=is_oa_account
                )
            )

        if errors:
            raise MigrationError(
                _T("FORMAT_MIGRATION_LIST_FILE_IS_NOT_VALID").format(
                    filename=path_to_migration_list(), errors='\n'.join(errors)
                )
            )

        return items


def path_to_migration_list():
    """
    Path of migration list file according to options specified by user

    :return: str
    """
    global_context = Registry.get_instance().get_context()
    if global_context.options.migration_list_file is None:
        return global_context.session_files.get_path_to_default_migration_list_file()
    else:
        return global_context.options.migration_list_file


MigrationListData = namedtuple(
    'MigrationListData',
    (
        'id',
        'plan_id',
        'mapping_subscription_id',
        'is_oa_subscription',
        'mapping_reseller_id',
        'is_oa_account',
    )
)
