# (C)2017 Ingram Micro Inc. all Rights Reserved
import json
import logging

from migrator.core import ParallelExecutionError
from migrator.core.data.model.migration_project import MigrationProjectStatus
from migrator.core.data.model.subscription import SubscriptionMigrationStatus as SubscriptionMigrationStatus
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils import find_first
from migrator.core.utils.concurrency.file_lock import FileLock, FileLockException
from migrator.core.utils.loggers.report import Issue, Problem, Report
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


class MigrationProgress(object):
    def __init__(self):
        self._subscriptions = []
        self._execution_lock = FileLock(Registry.get_instance().get_context().session_files.get_path_to_migrate_lock())
        self._migration_project = PlatformReaderDB().get_migration_project()

    def get_subscription(self, id):
        """
        :type id: basestring
        :rtype: migrator.core.session.migration_progress.SubscriptionMigrationProgress
        """

        s = find_first(self._subscriptions, lambda s: s.id == id)
        if s is None:
            s = SubscriptionMigrationProgress(id)
            self._subscriptions.append(s)

        return s

    def start(self):
        if self._migration_project is None:
            return

        try:
            self._execution_lock.acquire()
            if self._migration_project.status == MigrationProjectStatus.IN_PROGRESS:
                raise ParallelExecutionError()

            self._migration_project.description = ''
            self._migration_project.status = MigrationProjectStatus.IN_PROGRESS
            PlatformReaderDB().update_migration_project_status(self._migration_project)
            self._execution_lock.release()
        except FileLockException:
            raise ParallelExecutionError()

    def stop(self):
        if self._migration_project is None:
            return

        global_context = Registry.get_instance().get_context()
        execution_wrapper = global_context.execution_wrapper
        global_report = Report("general_issues", None)
        global_context.migrator._add_report_issues(global_report, execution_wrapper.failed_objects.general)
        if global_report.has_errors():
            migration_project_status = MigrationProjectStatus.FINISHED_ERRORS
        elif global_report.has_warnings():
            migration_project_status = MigrationProjectStatus.FINISHED_WARNINGS
        elif global_report.has_notices():
            migration_project_status = MigrationProjectStatus.FINISHED_NOTICES
        else:
            migration_project_status = MigrationProjectStatus.FINISHED_OK


        messages = []
        for issue in global_report.issues:
            messages.append({
                'type': issue.problem.severity,
                'message': issue.problem.description,
            })

        for subscription in self._subscriptions:
            subscription_status = subscription.status
            if subscription_status == SubscriptionMigrationStatus.FINISHED_NOTICES:
                if migration_project_status not in [
                    MigrationProjectStatus.FINISHED_ERRORS,
                    MigrationProjectStatus.FINISHED_WARNINGS
                ]:
                    migration_project_status = MigrationProjectStatus.FINISHED_NOTICES
            elif subscription_status == SubscriptionMigrationStatus.FINISHED_WARNINGS:
                if migration_project_status != MigrationProjectStatus.FINISHED_ERRORS:
                    migration_project_status = MigrationProjectStatus.FINISHED_WARNINGS
            elif subscription_status == SubscriptionMigrationStatus.FINISHED_ERRORS:
                migration_project_status = MigrationProjectStatus.FINISHED_ERRORS
            elif subscription_status == SubscriptionMigrationStatus.IN_PROGRESS:
                subscription.status = SubscriptionMigrationStatus.FINISHED_ERRORS
                migration_project_status = MigrationProjectStatus.FINISHED_ERRORS

        self._migration_project.message = json.dumps(messages)
        self._migration_project.status = migration_project_status
        PlatformReaderDB().update_migration_project_status(self._migration_project)


class SubscriptionMigrationProgress(object):
    def __init__(self, id):
        """
        :type id: basestring
        """

        self._id = id

    @property
    def id(self):
        """
        :rtype: basestring
        """
        return self._id

    @property
    def status(self):
        """
        :rtype: basestring
        """
        return PlatformReaderDB(expire_data=True).get_subscription(self._id).migration_status

    @status.setter
    def status(self, new_status):
        """
        :type new_status: basestring
        """

        subscription = PlatformReaderDB().get_subscription(self._id)
        global_context = Registry.get_instance().get_context()

        try:
            import_api = global_context.target_panel_obj.get_import_api(global_context)
            import_api.update_subscription_migration_status(subscription, new_status)
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.external_id,
                Issue(Problem(
                    'import_accounts_and_subscriptions', Problem.ERROR,
                    _T("SUBSCRIPTION_STATUS_WAS_NOT_UPDATED").format(subscription=subscription.name, reason=unicode(e))
                ), None)
            )
            new_status = SubscriptionMigrationStatus.FINISHED_ERRORS

        if new_status != subscription.migration_status:
            subscription.migration_status = new_status
            subscription.message = json.dumps([])

            if new_status == SubscriptionMigrationStatus.IN_PROGRESS:
                subscription.migration_mode = Registry.get_instance().get_context().migration_mode

            PlatformReaderDB().update_subscription_status(subscription)

    def set_finally_subscription_status(self, subscription_external_id):
        global_context = Registry.get_instance().get_context()
        execution_wrapper = global_context.execution_wrapper
        subscription_report = Report('Subscription', subscription_external_id)

        if subscription_external_id in execution_wrapper.failed_objects.subscriptions:
            global_context.migrator._add_report_issues(
                subscription_report, execution_wrapper.failed_objects.subscriptions[subscription_external_id]
            )

        for issue in execution_wrapper.issues.subscriptions[subscription_external_id]:
            subscription_report.add_issue_obj(issue)

        messages = []
        for issue in subscription_report.issues:
            messages.append({
                'type': issue.problem.severity,
                'message': issue.problem.description,
            })

        if subscription_report.has_errors():
            self.status = SubscriptionMigrationStatus.FINISHED_ERRORS
        elif subscription_report.has_warnings():
            self.status = SubscriptionMigrationStatus.FINISHED_WARNINGS
        elif subscription_report.has_notices():
            self.status = SubscriptionMigrationStatus.FINISHED_NOTICES
        else:
            self.status = SubscriptionMigrationStatus.FINISHED_OK

        subscription = PlatformReaderDB().get_subscription(self._id)
        subscription.message = json.dumps(messages)
        PlatformReaderDB().update_subscription_status(subscription)
