# (C)2017 Ingram Micro Inc. all Rights Reserved
import ntpath
import os
import posixpath

from migrator.core.utils import cached
from migrator.core.utils.operating_system.windows_thirdparty import get_spath_bin
from migrator.core.utils.registry import Registry


class SessionDir(object):
    def __init__(self, runner_cm, session_dir):
        """
        Constructor of migration session storage directory
        :param runner_cm: contextmanager that could create runner object
        :param str session_dir: path to the session directory
        """
        self._runner_cm = runner_cm
        self._session_dir = session_dir
        self._instantiated = False

    def get_session_dir_path(self):
        """
        Return path of a session directory
        :rtype: str
        """
        self._lazy_instantiate()
        return self._session_dir

    def get_file_path(self, filename):
        """
        Return full path to a file identified by filename
        :type filename: str
        :rtype: str
        """
        self._lazy_instantiate()
        return self._get_file_path(filename)

    def _lazy_instantiate(self):
        if self._instantiated:
            return

        with self._runner_cm() as runner:
            if not runner.file_exists(self._session_dir):
                # session directory does not exist, so create it
                runner.mkdir(self._session_dir)
                self._create_version_file(_get_migrator_version())
            else:
                # compare existing session directory version with Panel Migrator version
                version_file_name = self._get_file_path('version')
                is_backup = False
                if not runner.file_exists(version_file_name):
                    is_backup = True
                else:
                    version = runner.get_file_contents(version_file_name).strip()
                    if version != _get_migrator_version():
                        is_backup = True

                if is_backup:
                    # session directory version does not match with Panel Migrator version,
                    # so backup all sensitive data: move it out from their regular location into
                    # separate directory
                    self._backup(_get_migrator_version())
                    self._create_version_file(_get_migrator_version())

        # Set permissions even if directory already exists, to avoid issues when directory was created before
        # this object was initialized, or when if migrator was interrupted after directory was created, but
        # before we set permissions on it.
        self._set_permissions(runner)

        self._instantiated = True

    def _get_file_path(self, filename):
        return self._join_paths(self._session_dir, filename)

    def _backup(self, version):
        with self._runner_cm() as runner:
            files_to_move = []
            for file_name in runner.get_files_list(self._session_dir):
                if file_name == 'generated-passwords.yaml':
                    # do not backup generated passwords file to prevent repeat of passwords generation
                    continue
                if file_name.endswith('.yaml'):
                    # backup all rest yaml files, because format of storing data may change
                    # and become invalid after upgrade
                    files_to_move.append(file_name)
                if file_name in {'migration-node.crt', 'migration-node.key', 'source-node.crt', 'source-node.key'}:
                    # backup certificate-key pair to perform upgrade of agent on source Windows nodes
                    files_to_move.append(file_name)

            if len(files_to_move) > 0:
                backup_path = self._join_paths(self._session_dir, '_backup_%s' % version)
                runner.mkdir(backup_path)
                for file_name in files_to_move:
                    runner.move(
                        self._join_paths(self._session_dir, file_name),
                        self._join_paths(backup_path, file_name)
                    )

    def _create_version_file(self, version):
        with self._runner_cm() as runner:
            version_file_name = self._get_file_path('version')
            if runner.file_exists(version_file_name):
                runner.remove_file(version_file_name)

            runner.upload_file_content(version_file_name, version)

    @staticmethod
    def _join_paths(*paths):
        raise NotImplementedError()

    def _set_permissions(self, runner):
        raise NotImplementedError()


class WindowsSessionDir(SessionDir):
    """
    Migration session storage directory on Windows
    """
    @staticmethod
    def _join_paths(*paths):
        return ntpath.join(*paths)

    def _set_permissions(self, runner):
        # 1. Upload a special tool which we will use to secure session dir
        remote_spath_bin = self._join_paths(self._session_dir, os.path.basename(get_spath_bin()))
        if not runner.file_exists(remote_spath_bin):
            runner.upload_file(get_spath_bin(), remote_spath_bin)

        # 2. Secure session dir by changing the folder permissions to:
        #    - no permission inheritance from parent dir
        #    - full access to SYSTEM user
        #    - full access to current user
        #    - full access to Administrators group
        runner.run(remote_spath_bin, [self._session_dir])


class UnixSessionDir(SessionDir):
    """
    Migration session storage directory on Unix
    """
    @staticmethod
    def _join_paths(*paths):
        return posixpath.join(*paths)

    def _set_permissions(self, runner):
        runner.sh('chown {owner}:{group} {dir}', {'owner': 'root', 'group': 'root', 'dir': self._session_dir})
        runner.sh('chmod {mode} {dir}', {'mode': '0750', 'dir': self._session_dir})


@cached
def _get_migrator_version():
    version_file_name = os.path.join(Registry.get_instance().get_base_dir(), 'version')
    if not os.path.exists(version_file_name):
        return 'DEVELOPMENT'
    else:
        with open(version_file_name, 'r') as fp:
            return fp.read().strip()
