# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.language import _T
from migrator.core.session import config
from migrator.core.transfer.target_path_converter import WindowsRsyncPathConverter
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.rsync import create_rsync_command
from migrator.core.utils.operating_system import windows_utils
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


class ContentMigratorBase(object):
    def __init__(self, source_description, target_description, rsync_additional_args, runner_source, runner_target, test_mode=False):
        self.source_description = source_description
        self.target_description = target_description
        self.rsync_additional_args = rsync_additional_args
        self.runner_source_context_manager = runner_source
        self.runner_target_context_manager = runner_target
        self.global_context = Registry.get_instance().get_context()
        self.test_mode = test_mode
        if self.test_mode:
            if self.rsync_additional_args:
                self.rsync_additional_args.append('--dry-run')
            else:
                self.rsync_additional_args = ['--dry-run']

    def __enter__(self):
        self.runner_source = self.runner_source_context_manager.__enter__()
        self.runner_target = self.runner_target_context_manager.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.runner_source_context_manager.__exit__(exc_type, exc_val, exc_tb)
        self.runner_target_context_manager.__exit__(exc_type, exc_val, exc_tb)
        self.runner_source = None
        self.runner_target = None

    @trace
    def copy_content(self, items):
        """

        :type items: list [migrator.core.data.model.web_content_shm.ContentItemModel]
        """
        try:
            for item in items:
                try:
                    exists_on_source = self._exists_on_source(item.source_path)
                except Exception as e:
                    logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                    raise Exception(_T("FAILED_TO_CHECK_SOURCE_FILE_EXISTENCE").format(reason=str(e)))

                if exists_on_source or self.test_mode:
                    try:
                        self._copy_item(item)
                    except Exception as e:
                        logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                        raise Exception(_T("FAILED_TO_COPY_SOURCE_PATH").format(
                            source_path=item.source_path, target_path=item.target_path, reason=str(e)
                        ))
        except Exception as e:
                raise Exception(_T("FAILED_TO_COPY_CONTENT_FROM").format(
                    source=self.source_description, target=self.target_description, reason=str(e)
                ))

        for item in items:
            try:
                self._restore_permissions(item)
            except Exception as e:
                logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                raise Exception(_T("FAILED_TO_RESTORE_PERMISSIONS").format(
                    path=item.target_path, reason=str(e)
                ))

    def _exists_on_source(self, path):
        return self.runner_source.file_exists(path)

    def _copy_item(self, item):
        raise NotImplementedError()

    def _restore_permissions(self, item):
        raise NotImplementedError()


class ContentMigratorWindows(ContentMigratorBase):
    def __init__(
            self, source_description, target_description, rsync_additional_args, rsync, runner_source, runner_target, test_mode=False
    ):
        super(ContentMigratorWindows, self).__init__(
            source_description, target_description, rsync_additional_args, runner_source, runner_target, test_mode
        )
        self.rsync = rsync

    def __enter__(self):
        return super(ContentMigratorWindows, self).__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super(ContentMigratorWindows, self).__exit__(exc_type, exc_val, exc_tb)

    @trace
    def _copy_item(self, item):
        """

        :type item: migrator.core.data.model.web_content_shm.ContentItemModel
        """
        self.rsync.sync(
            source_path=windows_utils.convert_path_to_cygwin_vhosts_folder(item.source_path),
            target_path=WindowsRsyncPathConverter().expand(item.target_path),
            exclude=item.exclude,
            rsync_additional_args=self.rsync_additional_args
        )

    @trace
    def _restore_permissions(self, item):
        # TODO: restore windows permissions
        pass


class ContentMigratorUnix(ContentMigratorBase):
    def __init__(
            self, source_description, target_description, rsync_additional_args, key_pathname, runner_source,
            runner_target, source_rsync_bin, source_user, source_ip, source_ssh_port, test_mode=False
    ):
        super(ContentMigratorUnix, self).__init__(
            source_description, target_description, rsync_additional_args, runner_source, runner_target, test_mode
        )
        self.key_pathname = key_pathname
        self.source_rsync_bin = source_rsync_bin
        self.source_user = source_user
        self.source_ip = source_ip
        self.source_ssh_port = source_ssh_port

    def __enter__(self):
        return super(ContentMigratorUnix, self).__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super(ContentMigratorUnix, self).__exit__(exc_type, exc_val, exc_tb)

    @trace
    def _copy_item(self, item):
        """

        :type item: migrator.core.data.model.web_content_shm.ContentItemModel
        """
        cmd, args = create_rsync_command(
            self.key_pathname, self.runner_source, self.source_user, self.source_ip,
            windows_utils.convert_path_to_cygwin(item.source_path), item.target_path, item.exclude,
            self.rsync_additional_args, self.source_rsync_bin, self.source_ssh_port
        )
        self.runner_target.run(cmd, args)

    @trace
    def _restore_permissions(self, item):
        """

        :type item: migrator.core.data.model.web_content_shm.ContentItemModel
        """
        if item.permissions is not None:
            if item.permissions.access_rights is not None:
                # TODO: Rewrite to less dangerous algorithm which don't follow symlinks at least
                self.runner_target.run("/bin/chmod", [item.permissions.access_rights, item.target_path])

            if item.permissions.owners:
                for owner in item.permissions.owners:
                    from_user_and_group = [x for x in [owner.user_id, owner.group_id] if x]
                    target_user = owner.user if owner.user else ''
                    target_group = owner.group if owner.group else ''

                    chown_args = []
                    if len(from_user_and_group):
                        chown_args.append('--from=' + ':'.join(from_user_and_group))

                    if target_user or target_group:
                        chown_args.append(':'.join((target_user, target_group)))

                    if chown_args:
                        chown_args.insert(0, '-RPh')
                        chown_args.append(item.target_path)
                        self.runner_target.run('/bin/chown', chown_args)

                        if item.is_wwwroot:
                            self.runner_target.run('/bin/chown', [':psaserv', item.target_path])


@trace
def get_content_migrator(global_context, webspace_name, source, target, source_ip, test_mode=False):
    """

    :rtype: migrator.core.transfer.content_migrator.ContentMigratorBase
    """
    rsync_additional_args = config.read_rsync_additional_args(global_context.config)
    # TODO: apply hosting analyzer strategy
    # apply_hosting_analyser_strategy(
    # global_context, rsync_additional_args, source.node_id, webspace_name.encode('idna')
    # )
    runner_source = source.runner()
    runner_target = target.runner()

    if source.is_windows():
        rsync = global_context.get_rsync(source, target, source_ip)

        return ContentMigratorWindows(
            source.description(), target.description(), rsync_additional_args, rsync, runner_source, runner_target, test_mode
        )
    else:
        key_path = global_context.ssh_key_pool.get(source, target).key_pathname

        return ContentMigratorUnix(
            source.description(), target.description(), rsync_additional_args, key_path, runner_source,
            runner_target, source.rsync_bin, source.user(), source.ip(), source.settings().ssh_auth.port, test_mode
        )
