# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import os
from collections import namedtuple

from migrator.core import CommandExecutionError, MigrationError, MigrationNoRepeatError
from migrator.core.hosting.db.main import get_db_connection_info
from migrator.core.language import _T
from migrator.core.utils.loggers.log import trace

logger = logging.getLogger(__name__)


DatabaseTransferOptions = namedtuple('DatabaseTransferOptions', 'content_migrator')
Item = namedtuple('Item', 'source_path target_path exclude permissions')

class DatabaseTransferInfo(namedtuple('DatabaseTransferInfo', (
        'source_database_server', 'target_database_server', 'plesk_target_server', 'subscription_name', 'database_name',
        'target_database_name', 'webspace'))):

    """Information about database to copy"""
    """
    :type source_database_server: parallels.core.connections.database_server.DatabaseServer
    :type target_database_server: parallels.core.connections.database_server.DatabaseServer
    :type subscription_name: basestring
    :type database_name: basestring
    """

    def __str__(self):
        return _T("DATABASE_COPY_INFO_DESCRIPTION") % (
            self.database_name, self.subscription_name,
            self.source_database_server.description(),
            self.target_database_server.description()
        )


class DatabaseTransferConfiguration(object):
    """Configuration options and commands for database transfer."""
    DB_DUMP = 'db_dump'
    PASSWORDS_DUMP = 'passwords_dump'

    def __init__(self, database_name, target_database_name, webspace):
        self.database_name = database_name
        self.target_database_name = target_database_name
        self.webspace = webspace

    @classmethod
    def get_version_command(cls, server):
        raise NotImplementedError()

    def get_dump_command(self, server, password_file, dump_dirname, source_db_server_version, target_db_server_version):
        """Return a pair of command template and command options suitable for runner.sh()."""
        raise NotImplementedError()

    def get_restore_command(self, target_server, target_plesk_server, password_file, dump_dirname):
        """Return a pair of command template and command options suitable for runner.sh()."""
        raise NotImplementedError()

    def create_password_file(self, database_server, plesk_server):
        """Create a file with database access credentials."""
        raise NotImplementedError()

    @trace
    def _get_dump_options(self, server, existing_options):
        options = existing_options.copy()
        options['db_dump_filename'] = os.path.join(options['dump_dirname'], DatabaseTransferConfiguration.DB_DUMP)
        options['passwords_dump_filename'] = os.path.join(options['dump_dirname'], DatabaseTransferConfiguration.PASSWORDS_DUMP)

        connection_info = get_db_connection_info(server, self.database_name)
        options.update(connection_info.as_dictionary())
        return options

    @trace
    def _get_restore_options(self, server, existing_options):
        options = existing_options.copy()
        options['db_dump_filename'] = os.path.join(options['dump_dirname'], DatabaseTransferConfiguration.DB_DUMP)
        options['passwords_dump_filename'] = os.path.join(options['dump_dirname'], DatabaseTransferConfiguration.PASSWORDS_DUMP)
        connection_info = get_db_connection_info(server, self.target_database_name)
        options.update(connection_info.as_dictionary())
        return options


class DatabaseContentTransfer(object):
    """The object, that performs migration of database content."""

    def __init__(self, db_transfer_configuration, options):
        self.options = options
        self.transfer_config = db_transfer_configuration

    @trace
    def transfer(self, source_server, target_server, plesk_target_server):
        """Perform migration of database content."""
        source_version = self._get_version(source_server.panel_server)
        target_version = self._get_version(plesk_target_server)

        dump_password_file = None
        source_dump_dirname = None
        target_dump_dirname = None
        restore_password_file = None
        try:
            dump_password_file = self.transfer_config.create_password_file(source_server, source_server)
            source_dump_dirname = self._create_dump(source_server, dump_password_file, source_version, target_version)
            target_dump_dirname = self._transfer_dump(source_dump_dirname, source_server, plesk_target_server)
            restore_password_file = self.transfer_config.create_password_file(target_server, plesk_target_server)
            self._restore(target_dump_dirname, target_server, plesk_target_server, restore_password_file)
        except CommandExecutionError as e:
            if 'utf8' in e.stderr:
                raise MigrationNoRepeatError((_T('ONLY_DB_CONTENT_IN_UTF8_IS_SUPPORTED')).format(
                    database_name=self.transfer_config.database_name,
                    source_server=source_server.description(), target_server=target_server.description()
                ))
            else:
                raise
        finally:
            with source_server.runner() as runner:
                if dump_password_file:
                    runner.remove_file(dump_password_file)
                if source_dump_dirname:
                    runner.remove_directory(source_dump_dirname)

            with plesk_target_server.runner() as runner:
                if restore_password_file:
                    runner.remove_file(restore_password_file)
                if target_dump_dirname:
                    runner.remove_directory(target_dump_dirname)

    @trace
    def _get_version(self, server):
        command, version_re = self.transfer_config.get_version_command(server)
        if command is None:
            logger.debug('Skip Database Server version detecting at %s' % (server.ip()))
            return None

        with server.runner() as runner:
            out = runner.sh(command)
        m = version_re.match(out)
        if m:
            version = m.group(1)
            logger.debug('Detected Database Server version %s at %s' % (version, server.ip()))
            return version
        else:
            logger.debug('Detection Database Server version failed at %s' % server.ip())
            return None

    @trace
    def _create_dump(self, server, password_file, source_db_server_version, target_db_server_version):
        """Dump source database to disk; return dump filename."""
        with server.runner() as runner:
            dump_dirname = runner.mktempdir(server.session_dir())
            command, options = self.transfer_config.get_dump_command(
                server, password_file, dump_dirname, source_db_server_version, target_db_server_version
            )
            runner.sh(command, options)
            return dump_dirname

    @trace
    def _transfer_dump(self, source_dump_dirname, source_server, target_server):
        try:
            logger.debug(_T("COPY_DATABASE_DUMP_FROM_SOURCE_TARGET"))
            with target_server.runner() as runner:
                target_dump_dirname = runner.mktempdir(target_server.session_dir())

            item = Item(source_path=source_dump_dirname+'/*', target_path=target_dump_dirname, exclude=None, permissions=None)
            with self.options.content_migrator as content_migrator:
                content_migrator.copy_content([item])

        except CommandExecutionError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(
                _T("RSYNC_FAILED_COPY_DATABASE_DUMP_FROM") % (
                    source_server.panel_server.ip(),
                    target_server.panel_server.ip(),
                    e.stderr
                )
            )
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(
                _T("RSYNC_FAILED_COPY_DATABASE_DUMP_FROM_1") % (
                    source_server.panel_server.ip(),
                    target_server.panel_server.ip(),
                    str(e)
                )
            )

        return target_dump_dirname

    @trace
    def _restore(self, dump_dir, target_server, target_plesk_server, password_file):
        command, options = self.transfer_config.get_restore_command(target_server, target_plesk_server, password_file, dump_dir)
        with target_plesk_server.runner() as runner:
            runner.sh(command, options)
