# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.language import _T
from migrator.core import CommandExecutionError
from migrator.core.utils import is_empty
from migrator.core.utils.networking.dns import resolve_all
from migrator.core.utils.operating_system.windows_utils import get_from_registry

from common import DatabaseTransferConfiguration

logger = logging.getLogger(__name__)


class MSSqlTransferConfiguration(DatabaseTransferConfiguration):
    _restore_t_sql = """
        alter database [{db_name}] set single_user with rollback immediate;
        declare @rows varchar(256);
        declare @log varchar(256);
        set @rows = (SELECT physical_name FROM sys.master_files WHERE database_id = DB_ID('{db_name}') and type=0);
        set @log = (SELECT physical_name FROM sys.master_files WHERE database_id = DB_ID('{db_name}') and type=1);
        restore database [{db_name}] from disk='{db_dump_filename}' with replace,
            move '{db_name}' to @rows, move '{db_name}_log' to @log;
        alter database [{db_name}] set multi_user;
        
        use [{db_name}]
        declare @username varchar(128)
        declare cur cursor for select name from sys.database_principals where name in ({db_users})
        open cur
        fetch next from cur into @username
        while @@FETCH_STATUS = 0  
        begin
            exec ('alter user ' + @username + ' with login = ' +@username)
            fetch next from cur into @username;
        end
        close cur
        deallocate cur
    """

    _dump_database = '"{sqlcmd}" {connect_parameters} -Q "backup database [{db_name}] to disk=\'{db_dump_filename}\'"'
    _restore_database = '"{sqlcmd}" {connect_parameters} -Q "%s"' % _restore_t_sql

    @classmethod
    def get_version_command(cls, server):
        return None, None

    def get_dump_command(self, server, password_file, dump_dirname, source_db_server_version, target_db_server_version):
        self._set_mssql_dumps_dir_permissions(server, server, dump_dirname)

        command = self._dump_database
        options = self._get_dump_options(
            server, {
                'dump_dirname': dump_dirname,
                'sqlcmd': server.panel_server.get_path_to_sqlcmd()
            }
        )
        connect_parameters = self.get_connect_parameters(server, password_file).format(**options)
        return command.format(connect_parameters=connect_parameters, **options), {}

    def get_restore_command(self, target_server, target_plesk_server, password_file, dump_dirname):
        self._set_mssql_dumps_dir_permissions(target_server, target_plesk_server, dump_dirname)

        command = self._restore_database
        options = self._get_restore_options(
            target_server, {
                'dump_dirname': dump_dirname,
                'sqlcmd': target_plesk_server.get_path_to_sqlcmd(),
                'db_users': ','.join(["'%s'" % x.login for x in self.webspace.database_users]),
            }
        )
        connect_parameters = self.get_connect_parameters(target_server, password_file).format(**options)
        return command.format(connect_parameters=connect_parameters, **options), {}

    def create_password_file(self, database_server, plesk_server):
        pass

    @classmethod
    def get_connect_parameters(cls, server, password_file):
        return "-S {host} -E -b -m1"

    @staticmethod
    def _set_mssql_dumps_dir_permissions(db_server, plesk_server, dumps_dir):
        """
        Set permissions for MSSQL dumps directory, so MSSQL server can read/write to it.

        NOTE: Pass PhysicalServer.session_dir() or its subdir as dumps_dir for security reasons.

        :type server: parallels.core.connections.database_server.DatabaseServer
        :type dumps_dir: str | unicode
        :rtype: None
        """
        with plesk_server.runner() as runner:
            try:
                mssql_instance = _get_mssql_instance_name(db_server.host())
                if mssql_instance:
                    mssql_instance_str = '$%s' % mssql_instance
                else:
                    mssql_instance_str = ''

                user = get_from_registry(
                    runner, [
                        r'HKLM\SYSTEM\CurrentControlSet\services\MSSQL%s' % mssql_instance_str,
                        r'HKLM\SYSTEM\CurrentControlSet\services\MSSQLSERVER%s' % mssql_instance_str
                    ], 'ObjectName'
                )

                if is_empty(user):
                    # No user defined - leave default permissions
                    return
                elif user == 'LocalSystem':
                    # Don't need to set permissions for LocalSystem in explicit way
                    return
                else:
                    icacl_commands = [
                        '/grant "%s:(OI)(CI)F"' % user
                    ]
                    for icacl_command in icacl_commands:
                        runner.sh("icacls {dir} %s" % (icacl_command,), {'dir': dumps_dir})

            except CommandExecutionError as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                logger.warning(_T("FAILED_TO_SET_DBDUMP_DIR_PERMISSIONS").format(
                    reason=unicode(e), server=plesk_server.description(), directory=dumps_dir
                ))


def get_mssql_host_for_target_server(source_database_server):
    r"""Get hostname of MSSQL database server that could be used on target server
    This function replaces all references to the source server in hostname. So, for example if
    name of a server was 'localhost\MSSQLSERVER2008', it will be replaced with '10.52.1.17\MSSQLSERVER2008',
    where 10.52.1.17 is an IP address of the source database server. Named and default MSSQL instances are handled
    by this function
    :type source_database_server: parallels.core.connections.database_server.DatabaseServer
    """
    hostname, instance = _get_mssql_host_parts(source_database_server)
    if hostname in ('localhost', '.', '127.0.0.1'):
        hostname = source_database_server.ip()

    if not is_empty(instance):
        return r'%s\%s' % (hostname, instance)
    else:
        return hostname


def _get_mssql_instance_name(host):
    """Get MSSQL instance name from MSSQL host name

    For example, if you have host name:
    192.168.1.110\MSSQL2008
    This function returns "MSSQL2008"
    If that is default instance - it returns None

    :type host: str | unicode
    :rtype: str
    """
    parts = host.split('\\', 2)
    if len(parts) < 2:
        return None
    else:
        return parts[1]


def compare_mssql_servers(database_server_1, database_server_2):
    """Check if two MSSQL servers are actually the same server

    This function takes into account that the same server may be registered in different ways:
    by hostname, by IP, as 'localhost' or as '.'.

    :type database_server_1: parallels.core.connections.database_server.DatabaseServer
    :type database_server_2: parallels.core.connections.database_server.DatabaseServer
    :rtype: bool
    """
    _, instance_1 = _get_mssql_host_parts(database_server_1)
    _, instance_2 = _get_mssql_host_parts(database_server_2)
    port_1 = _normalize_mssql_port(database_server_1)
    port_2 = _normalize_mssql_port(database_server_2)

    same_host = (
        database_server_1.ip() in _resolve_mssql_hostname(database_server_2) or
        database_server_2.ip() in _resolve_mssql_hostname(database_server_1)
    )
    same_instance = instance_1 == instance_2
    same_port = port_1 == port_2

    return same_host and same_instance and same_port


def _resolve_mssql_hostname(database_server):
    """Convert MSSQL database server hostname (without instance name) to list of IP addresses

    :type database_server: parallels.core.connections.database_server.DatabaseServer
    """
    hostname, _ = _get_mssql_host_parts(database_server)
    if hostname in ('localhost', '.', '127.0.0.1'):
        return [database_server.ip()]
    else:
        return resolve_all(hostname)


def _get_mssql_host_parts(database_server):
    """Split MSSQL database servers hostname into host and instance parts

    :type database_server: parallels.core.connections.database_server.DatabaseServer
    :rtype: tuple(str, str | None)
    """
    parts = database_server.host().split('\\')
    if len(parts) == 1:
        # Plain hostname
        return parts[0], None
    elif len(parts) == 2:
        instance_name = parts[1]
        if instance_name == '':
            instance_name = None
        return parts[0], instance_name
    else:
        assert False, _T("ASSERT_INVALID_MSSQL_HOSTNAME").format(hostname=database_server.host())


def _normalize_mssql_port(database_server):
    """Get valid integer MSSQL server port, considering that all null values mean default port (1433)

    :type database_server: parallels.core.connections.database_server.DatabaseServer
    :rtype: int
    """
    if database_server.port() in (None, '', 0):
        return 1433
    else:
        return database_server.port()
