# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import re

from migrator.core.hosting.db.main import get_db_connection_info, DatabaseConnectionInfo
from migrator.core.transfer.db import escape_db_passwd
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.ip_utils import get_route_ip
from common import DatabaseTransferConfiguration

logger = logging.getLogger(__name__)


class MysqlTransferConfiguration(DatabaseTransferConfiguration):
    """Configuration, commands and options for transferring a MySQL database."""

    _restore_database_lin = '{mysql} %s --default-character-set=utf8 < {db_dump_filename}'
    _restore_database_win = '{mysql} %s --default-character-set=utf8 --execute "source {db_dump_filename}"'

    _dump_database = "{mysqldump} %s " \
        "--quick --quote-names --add-drop-table --default-character-set=utf8 --set-charset --compatible={compatible} " \
        "--result-file {db_dump_filename}"

    _get_version = '{mysql} --version'
    _version_re = re.compile('.*Distrib ([\d|\.]*)')

    @classmethod
    def get_version_command(cls, server):
        command = cls._get_version
        return command.format(mysql=server.get_path_to_mysql()), cls._version_re

    def get_dump_command(
            self, server, password_file, dump_dirname, source_db_server_version_str, target_db_server_version_str
    ):
        """Return a pair of command template and command options suitable for runner.sh()."""
        utf8mb4_version = [5, 5, 3]
        compatible = ''
        if source_db_server_version_str and target_db_server_version_str:
            source_db_server_version = [int(x) for x in source_db_server_version_str.split('.')]
            target_db_server_version = [int(x) for x in target_db_server_version_str.split('.')]

            if source_db_server_version >= utf8mb4_version and target_db_server_version < utf8mb4_version:
                compatible = 'mysql40'

        command = MysqlTransferConfiguration._dump_database
        options = self._get_dump_options(
            server, {
                'dump_dirname': dump_dirname,
                'mysqldump': server.panel_server.get_path_to_mysqldump(),
                'mysql': server.panel_server.get_path_to_mysql(),
                'compatible': compatible
            }
        )
        connect_parameters = self.get_connect_parameters(server, password_file).format(**options)
        return command % connect_parameters, options

    def get_restore_command(self, target_server, target_plesk_server, password_file, dump_dirname):
        """Return a pair of command template and command options suitable for runner.sh()."""
        command = MysqlTransferConfiguration._restore_database_win if target_plesk_server.is_windows() else \
                  MysqlTransferConfiguration._restore_database_lin
        options = self._get_restore_options(
            target_server, {
                'dump_dirname': dump_dirname,
                'mysql': target_plesk_server.get_path_to_mysql()
            }
        )
        connect_parameters = self.get_connect_parameters(target_server, password_file).format(**options)
        return command % connect_parameters, options

    def create_password_file(self, database_server, plesk_server):
        """Create a file with database access credentials."""
        connection_info = get_db_connection_info(database_server, self.database_name)
        password_file = _create_mysql_options_file(plesk_server, connection_info)
        return password_file

    @classmethod
    def get_connect_parameters(cls, server, option_file):
        cmd = "%s -h {host} -P {port} -u{login} {db_name}"
        if option_file:
            credentials = '--defaults-file="' + option_file + '"'
        elif server.host == 'localhost' and server.port == 3306:
            credentials = u" --no-defaults -p\"`cat /etc/psa/.psa.shadow`\""  # a workaround for Plesk
        else:
            credentials = u" --no-defaults -p{password}"
        return cmd % (credentials,)


@trace
def _create_mysql_options_file(remote_server, connection_info):
    """Create a mysql options file and upload it to the remote server.

    :type connection_info: parallels.core.utils.database_utils.DatabaseConnectionInfo
    """
    filename = 'my_{host}_{db_name}.cnf'.format(**connection_info.as_dictionary())
    connection_str = u"[client]\nuser={login}\npassword='{password}'".format(
        **connection_info.as_dictionary()
    )
    with remote_server.runner() as runner:
        remote_filename = runner.mktempfile(remote_server.get_session_file_path(filename))
        runner.upload_file_content(
            remote_filename,
            connection_str if not remote_server.is_windows() else connection_str.replace('\n', '\r\n')
        )

    return remote_filename


@trace
def check_admin_access(dbserver, client_host):
    with client_host.runner() as runner:
        connection_info = DatabaseConnectionInfo(
            host=dbserver.host, port=dbserver.port, login=dbserver.login, password=dbserver.password, db_name='mysql'
        )

        params = {
            'host': dbserver.host,
            'port': dbserver.port,
            'mysql': client_host.get_path_to_mysql(),
            'password_file': _create_mysql_options_file(client_host, connection_info)
        }

        # try to connect
        exit_code, stdout, stderr = runner.sh_unchecked(
            "{mysql} --defaults-file={password_file} --host={host} --port={port} --execute \"select 1\"",
            params
        )
        return exit_code == 0


@trace
def allow_admin_access(dbserver, client_host, server_host):
    select_user_query = "SELECT 'user_exists' from mysql.user where Host='{login_ip}' and User='{login}'"
    create_user_query = "CREATE USER '{login}'@'{login_ip}'"
    alter_user_query = """
SET PASSWORD FOR '{login}'@'{login_ip}' = PASSWORD('{password}');
GRANT ALL PRIVILEGES ON *.* TO '{login}'@'{login_ip}' WITH GRANT OPTION;
FLUSH PRIVILEGES;
"""
    mysql_cmd = "{mysql} --defaults-file={password_file} --port={port} --execute \"%s\""

    connection_info = DatabaseConnectionInfo(
        host=dbserver.host, port=dbserver.port, login=dbserver.login, password=dbserver.password, db_name='mysql'
    )

    cmd_params = {
        'port': dbserver.port,
        'mysql': 'mysql',
        'password_file': _create_mysql_options_file(server_host, connection_info)
    }

    sql_params = {
        'host': dbserver.host,
        'login': dbserver.login,
        'password': escape_db_passwd(dbserver.password),
    }

    with client_host.runner() as runner:
        sql_params['login_ip'] = get_route_ip(dbserver.host, runner)

    with server_host.runner() as runner:
        val = runner.sh(mysql_cmd % select_user_query.format(**sql_params), cmd_params)
        if 'user_exists' not in val:
            runner.sh(mysql_cmd % create_user_query.format(**sql_params), cmd_params)

        runner.sh(mysql_cmd % alter_user_query.format(**sql_params), cmd_params)


@trace
def allow_user_access(dbserver, dbserver_login, dbserver_password, client_host, database, user):
    replace_user_query = """
replace into mysql.db
(Host, Db, User, Select_priv, Insert_priv, Update_priv, Delete_priv, Create_priv, Drop_priv, Grant_priv, References_priv, Index_priv, Alter_priv, Create_tmp_table_priv, Lock_tables_priv, Create_view_priv, Show_view_priv, Create_routine_priv, Alter_routine_priv, Execute_priv)
values('{login_ip}', '{client_database}', '{client_login}', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y', 'Y');
FLUSH PRIVILEGES;
"""

    update_priveleges = """
update mysql.db set Event_priv='Y', Trigger_priv='Y' where Host='{login_ip}' and Db='{client_database}' and User='{client_login}';
FLUSH PRIVILEGES;
"""

    with client_host.runner() as runner:

        connection_info = DatabaseConnectionInfo(
            host=dbserver.host, port=dbserver.port, login=user.login, password=user.password, db_name=database.name
        )

        params = {
            'host': dbserver.host,
            'port': dbserver.port,
            'client_database': database.name,
            'mysql': client_host.get_path_to_mysql(),
            'password_file': _create_mysql_options_file(client_host, connection_info),
        }

        # try to connect
        exit_code, stdout, stderr = runner.sh_unchecked(
            "{mysql} --defaults-file={password_file} --host={host} --port={port} --database={client_database} --execute \"select 1\"",
            params
        )

        if exit_code != 0:
            connection_info = DatabaseConnectionInfo(
                host=dbserver.host, port=dbserver.port, login=dbserver_login, password=dbserver_password,
                db_name='mysql'
            )

            params['password_file'] = _create_mysql_options_file(client_host, connection_info)
            params['login_ip'] = get_route_ip(dbserver.host, runner)
            params['client_login'] = user.login

            with client_host.runner() as runner:
                runner.sh(
                    "{mysql} --defaults-file={password_file} --host={host} --port={port} --execute \"%s\""
                    % replace_user_query, params
                )
                try:
                    runner.sh(
                        "{mysql} --defaults-file={password_file} --host={host} --port={port} --execute \"%s\""
                        % update_priveleges, params
                    )
                except:
                    logger.debug(
                        'It is look like Database server %s (%s) does not support Event_priv and Trigger_priv' %
                        (dbserver.host, dbserver.dbtype)
                    )


