# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import MigrationNotice
from migrator.core.hosting.db.main import get_db_connection_info, DatabaseConnectionInfo
from migrator.core.language import _T
from migrator.core.utils.loggers.log import trace
from common import DatabaseTransferConfiguration


class PostgresqlTransferConfiguration(DatabaseTransferConfiguration):
    """Configuration options and commands for transferring PostgreSQL databases."""

    _dump_database = "PGPASSFILE={pgpass} pg_dump {connect_parameters} " \
                     "--format=custom --blobs --no-owner --ignore-version {db_name} > {db_dump_filename}"

    _restore_database = "PGPASSFILE={pgpass} pg_restore {connect_parameters} --verbose -d {db_name} {db_dump_filename}"

    @classmethod
    def get_version_command(cls, server):
        return None, None

    def get_dump_command(self, server, password_file, dump_dirname, source_db_server_version, target_db_server_version):
        command = self._dump_database
        options = self._get_dump_options(server, {'dump_dirname': dump_dirname, 'pgpass': password_file})
        connect_parameters = self.get_connect_parameters(server, password_file).format(**options)
        return command.format(connect_parameters=connect_parameters, **options), {}

    def get_restore_command(self, target_server, target_plesk_server, password_file, dump_dirname):
        command = self._restore_database
        options = self._get_restore_options(target_server, {'dump_dirname': dump_dirname})
        connect_parameters = self.get_connect_parameters(target_server, password_file).format(**options)
        return command.format(connect_parameters=connect_parameters, pgpass=password_file, **options), {}

    def create_password_file(self, database_server, plesk_server):
        connection_info = get_db_connection_info(database_server, self.database_name)
        if connection_info.password is not None:
            return _create_pgpass_file(plesk_server, connection_info)
        else:
            return None

    @classmethod
    def get_connect_parameters(cls, server, password_file):
        restore_command = "-U {login}"
        if server.host() != 'localhost':
            restore_command += " -h {host} -p {port}"
        return restore_command


def _create_pgpass_file(server, connection_info):
    """Create a PostgreSQL password file and upload it to the remote server.

    :type connection_info: parallels.core.utils.database_utils.DatabaseConnectionInfo
    :rtype: basestring
    """
    filename = 'pgpass_{host}_{db_name}'.format(**connection_info.as_dictionary())
    connection_str = '*:*:*:{login}:{password}'.format(**connection_info.as_dictionary())
    with server.runner() as runner:
        remote_filename = runner.mktempfile(server.get_session_file_path(filename))
        runner.upload_file_content(remote_filename, connection_str)
        runner.sh('chmod 600 %s' % remote_filename)

    return remote_filename


@trace
def allow_user_access(dbserver, dbserver_login, dbserver_password, client_host, database, user):
    psql = "PGPASSFILE={password_file} psql --host={host} --port={port} --user={user} --command='%s'"
    add_role_query = "create role {client_database}"
    grant_role_to_postgres = "grant {client_database} to postgres"
    grant_role_to_user = "grant {client_database} to {client_login}"

    connection_info = DatabaseConnectionInfo(
        host=dbserver.host, port=dbserver.port, login=dbserver_login, password=dbserver_password, db_name='postgres'
    )

    params = {
        'host': dbserver.host,
        'port': dbserver.port,
        'user': dbserver_login,
        'client_database': database.name,
        'client_login': user.login,
        'password_file': _create_pgpass_file(client_host, connection_info),
    }

    with client_host.runner() as runner:
        exit_code, stdout, stderr = runner.sh_unchecked(psql % add_role_query, params)
        if not (exit_code == 0 or 'already exists' in stderr):
            raise MigrationNotice(
                _T('UNABLE_TO_ALLOW_PHPPGADMIN_TO_ACCESS_THE_DATABASE').format(database=database.name, dbserver=dbserver.host)
            )

        runner.sh(psql % grant_role_to_postgres, params)
        runner.sh(psql % grant_role_to_user, params)
