# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
from collections import namedtuple, defaultdict
from contextlib import contextmanager

from migrator.core import MigrationError, MigrationNoRepeatError
from migrator.core.language import _T
from migrator.core.utils.lang.profiler import sleep
from migrator.core.utils.loggers.logging_context import subscription_context
from migrator.core.utils.loggers.report import Problem

TargetModelStructure = namedtuple('TargetModelStructure', (
    'subscriptions', 'clients', 'resellers', 'plans', 'auxiliary_users', 'auxiliary_user_roles', 
    'general',  # field for general errors, not related to any objects in target model
))

# 'is_critical' means, that migrator should stop migrating this object:
FailedObjectInfo = namedtuple(
    'FailedObjectInfo', (
        'error_message', 'solution', 'exception', 'is_critical', 'severity'))


class TransferApiWrapper:
    """An object that provides wrappers and error handling for transfer API methods.

    Wrappers provide the following functionality:
    - execute methods, optionally retrying failed operations
    - handle exceptions
    - register errors.
    """
    logger = logging.getLogger(__name__)

    def __init__(self, model):
        self.model = model
        self.failed_objects = self._create_target_model_structure()
        self.issues = self._create_target_model_structure()

    @staticmethod
    def _create_target_model_structure():
        return TargetModelStructure(
            subscriptions=defaultdict(list), clients=defaultdict(list), 
            resellers=defaultdict(list), plans=defaultdict(list),
            auxiliary_users=defaultdict(list),
            auxiliary_user_roles=defaultdict(list),
            general=[]
        )

    @contextmanager
    def wrap_subscription_actions(
        self, subscription_external_id, error_message, solution=None, is_critical=True, use_log_context=True
    ):
        try:
            if use_log_context:
                with subscription_context(subscription_external_id):
                    yield
            else:
                yield
        except Exception as e:
            info = FailedObjectInfo(
                error_message, solution, e, is_critical, Problem.ERROR
            )
            self.failed_objects.subscriptions[subscription_external_id].append(info)
            self.logger.error(
                _T("FAILED_TO_PERFORM_ACTION_ON_SUBSCRIPTION"), subscription_external_id, e
            )
            self.logger.debug(_T("LOG_EXCEPTION"), exc_info=True)

    def retry_subscription_action(
        self, func, subscription_external_id, error_message, solution=None, is_critical=True, repeat_error=None,
        repeat_count=3, sleep_time=10, use_log_context=True
    ):
        exceptions = []

        with self.wrap_subscription_actions(
            subscription_external_id, error_message, solution, is_critical, use_log_context=use_log_context
        ):
            for i in range(1, repeat_count + 1):
                try:
                    return func()
                except Exception as e:
                    exceptions.append(e)
                    self.logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                    if i == repeat_count or isinstance(e, MigrationNoRepeatError):
                        raise MigrationError(exceptions[0].message)

                    if repeat_error is not None:
                        self.logger.error(repeat_error)

                    sleep(sleep_time, _T("SLEEP_RETRY_EXECUTING_COMMAND"))

    def is_failed_subscription(self, subscription_id):
        """Check if subscription is completely failed or not.

        For subscriptions marked as completely failed we do not continue migration.

        :param subscription_id: basestring
        :rtype: bool
        """
        return not no_critical_errors(self.failed_objects.subscriptions, subscription_id)

    @contextmanager
    def wrap(self, error_message, solution):
        try:
            yield
        except Exception as e:
            self.logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            self.logger.error(error_message)
            info = FailedObjectInfo(
                error_message, solution, e, is_critical=False,
                severity=Problem.ERROR
            )
            self.failed_objects.general.append(info)

    def register_error(self, error_message, solution, severity=Problem.ERROR):
        """Report error that will be displayed at top (not subscription/client) level in the final migration report

        :type error_message: unicode
        :type solution: unicode
        :type severity: str
        :rtype: None
        """
        info = FailedObjectInfo(error_message, solution, None, is_critical=False, severity=severity)
        self.failed_objects.general.append(info)

    def add_subscription_issue(self, subscription_external_id, issue):
        for existing_issue in self.issues.subscriptions[subscription_external_id]:
            if existing_issue == issue:
                return

        self.issues.subscriptions[subscription_external_id].append(issue)

    def get_filtering_model(self):
        return FilteringModel(self.model, self.failed_objects)


class FilteringModel(object):
    """Adapter over target model that filters out failed objects"""

    def __init__(self, model, failed_objects):
        self.model = model
        self.failed_objects = failed_objects 

    def __getattr__(self, name):
        return getattr(self.model, name)

    def iter_all_subscriptions(self):
        return [
            subscription for subscription in self.model.iter_all_subscriptions() 
            if no_critical_errors(self.failed_objects.subscriptions, subscription.name)
        ]

    def iter_all_clients(self):
        return [
            FilteringClient(client, self.failed_objects) for client in self.model.iter_all_clients() 
            if client.login not in self.failed_objects.clients
        ]

    @property
    def clients(self):
        return dict([
            (login, FilteringClient(client, self.failed_objects)) for login, client in self.model.clients.iteritems()
            if login not in self.failed_objects.clients
        ])

    @property
    def resellers(self):
        return dict([
            (login, FilteringReseller(reseller, self.failed_objects))
            for login, reseller in self.model.resellers.iteritems()
            if login not in self.failed_objects.resellers
        ])


class FilteringReseller(object):
    """Adapter over reseller target model object that filters out failed clients of reseller"""

    def __init__(self, reseller, failed_objects):
        self.reseller = reseller
        self.failed_objects = failed_objects 

    def __getattr__(self, name):
        return getattr(self.reseller, name)

    @property
    def clients(self):
        return [
            FilteringClient(client, self.failed_objects) for client in self.reseller.clients
            if client.login not in self.failed_objects.clients
        ]


class FilteringClient(object):
    """Adapter over client target model object that filters out failed subscriptions of client"""

    def __init__(self, client, failed_objects):
        self.client = client
        self.failed_objects = failed_objects

    def __getattr__(self, name):
        return getattr(self.client, name)

    def __setattr__(self, name, value):
        if name in ('client', 'failed_objects'):
            # skip internal properties of adapter
            super(FilteringClient, self).__setattr__(name, value)
        else:
            # pass set attribute request to client target model object
            return setattr(self.client, name, value)

    @property
    def subscriptions(self):
        return [
            subscription for subscription in self.client.subscriptions
            if no_critical_errors(self.failed_objects.subscriptions, subscription.name)
        ]


def no_critical_errors(failed_objects_dict, key):
    return key not in failed_objects_dict or all(
        not failed_object_info.is_critical for failed_object_info in failed_objects_dict[key]
    )
