# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import time

from migrator.core.language import _T
from migrator.core.session.config import MailImapEncryption
from migrator.core.utils import safe_string_repr
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.operating_system import unix_utils

logger = logging.getLogger(__name__)


class CopyMailImapsync(object):
    @trace
    def copy_mail(self, global_context, webspace):
        for site in webspace.sites:
            if not site.mailboxes:
                logger.debug(_T("SKIP_COPYING_MAIL_CONTENT_FOR_DOMAIN_2"), site.name)
                continue

            for mailbox in site.mailboxes:
                self._copy_single_mailbox(global_context, webspace, site, mailbox)

    @trace
    def _copy_single_mailbox(self, global_context, webspace, site, mailbox):
        source_ip = webspace.mail_source_host.ip()
        if mailbox.is_password_generated:
            global_context.execution_wrapper.add_subscription_issue(
                webspace.subscription_id,
                Issue(Problem(
                    'copy-mailbox', Problem.NOTICE,
                    _T("UNABLE_TO_RETRIEVE_PASSWORD_FOR_MAILBOX_ON_SITE").format(
                        mailbox_name=mailbox.name,
                        site_name=site.name,
                        ip_address=source_ip,
                    ),
                ), None)
            )
            return

        email_idna = mailbox.name + "@" + site.name.encode("idna")
        email = mailbox.name + "@" + site.name
        file_suffix = '%s_%s' % (site.name.encode('idna'), mailbox.name.encode('idna'))

        migrator_server = global_context.migrator_server
        source_password_file_name = migrator_server.get_session_file_path('mail-password-source_%s.txt' % file_suffix)
        target_password_file_name = migrator_server.get_session_file_path('mail-password-target_%s.txt' % file_suffix)
        mailsync_log_file_name = migrator_server.get_session_file_path('mailsync_%s.log' % file_suffix)

        logger.debug(_T("COPY_CONTENT_MAILBOX_EMAIL_S_FOR").format(
            full_mailbox_name=email,
            mailsync_log_file_name=mailsync_log_file_name,
        ))

        with migrator_server.runner() as runner:
            runner.upload_file_content(source_password_file_name, mailbox.password)
            runner.upload_file_content(target_password_file_name, mailbox.password)

        command_arguments = dict(
            source_ip=source_ip,
            target_ip=webspace.mail_target_host.ip(),
            email=email_idna,
            source_passfile=source_password_file_name,
            target_passfile=target_password_file_name
        )

        source_encryption_settings_mapping = {
            MailImapEncryption.NONE: '',
            MailImapEncryption.SSL: ' --ssl1',
            MailImapEncryption.TLS: ' --tls1'
        }
        cmd = unix_utils.format_command(
            u"imapsync --noreleasecheck --noexpunge --automap "
            u"--host1 {source_ip} --user1 {email} --passfile1 {source_passfile} "
            u"--host2 {target_ip} --user2 {email} --passfile2 {target_passfile}",
            **command_arguments
        )
        cmd += source_encryption_settings_mapping[webspace.mail_source_host.mail_settings.mail_imap_encryption]

        with migrator_server.runner() as runner:
            exit_code, stdout, stderr = runner.sh_unchecked(cmd, env={'SERVER_SOFTWARE': ''})

        output = stdout + stderr
        with open(mailsync_log_file_name, 'a') as f:
            f.write((u"""
{date} {cmd}

""".format(date=time.asctime(), cmd=cmd)
            ).encode('utf-8'))
            f.write(
                (_T("LOG_EXIT_CODE_AND_OUTPUT") % (exit_code, safe_string_repr(output),)).encode('utf-8')
            )

        with migrator_server.runner() as runner:
            for f in [source_password_file_name, target_password_file_name]:
                runner.remove_file(f)

        if exit_code != 0:
            global_context.execution_wrapper.add_subscription_issue(
                webspace.subscription_id,
                Issue(Problem(
                    'copy-mailbox', Problem.WARNING,
                    _T("UNABLE_TO_COPY_CONTENTS_OF_MAILBOX").format(
                        mailbox_name=mailbox.name,
                        site_name=site.name,
                        ip_address=source_ip,
                    ),
                ), None)
            )
