# (C)2017 Ingram Micro Inc. all Rights Reserved
# encoding: utf-8
import codecs
import errno
import logging
import os
import platform
import random
import string
import sys
import time
import urllib

from collections import defaultdict, namedtuple
from contextlib import contextmanager
from functools import wraps

import migrator.settings

from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


def if_not_none(value, func):
    return func(value) if value is not None else None


def empty_string_if_none(value):
    return value or ''


def cast_to_int(value):
    return int(value) if value else None


class Obj:
    def __init__(self, kw):
        self.__dict__.update(kw)

    def __repr__(self):
        # TODO: add unicode support, enquote string values in representation
        return "obj(" + ", ".join(["%s=%s" % (k, v) for k, v in self.__dict__.iteritems()]) + ")"


def obj(**kw):
    return Obj(kw)


utf8_tree_symbols = {'|': u'│ ', '|-': u'├ ', '`-': u'└ ', ' ': '  '}
ascii_tree_symbols = {'|': u'|  ', '|-': u'|- ', '`-': u'`- ', ' ': '   '}


class FunctionResultWaiter(object):
    """
    This class is used for subsequent invoking of the function until it returns the expected result.
    Otherwise, an error will be thrown.
    """
    def __init__(self, func, error, **kwargs):
        """
        __init__(self, func, error, expected_result, check_function, max_attempts, interval_between_attempts)
          - func: the function that is called multiple times
          - error: the error that is thrown if the function was not executed successfully at least once
          - expected_result: if this is set, then the execution is considered successful if it returns the
            expected_result.
          - check_function: if expected_result is specified, then successful execution of check_function means that the
            function was executed successfully. If expected_result and check_function are not specified, then the
            execution will be considered successful if the result is not None.
          - max_attempts: the maximum number of attempts, for unlimited attempts pass 0 or -1 here
          - interval_between_attempts: the interval between attempts
          - throw_exception_on_failure: Should an exception be thrown on failure
        """
        expected_result = kwargs.get('expected_result', None)
        if expected_result is not None:
            self._check_function = lambda x: x == expected_result
        else:
            self._check_function = kwargs.get('check_function', lambda x: x is not None)

        self._func = func
        self._error = error

        global_context = Registry.get_instance().get_context()
        if global_context.migration_type == migrator.settings.MIGRATION_TYPE_TAKEOVER:
            default_max_attempts = migrator.settings.TAKEOVER_DEFAULT_MAX_ATTEMPTS
            default_interval_between_attempts = migrator.settings.TAKEOVER_DEFAULT_INTERVAL_BETWEEN_ATTEMPTS
        else:
            default_max_attempts = migrator.settings.DEFAULT_MAX_ATTEMPTS
            default_interval_between_attempts = migrator.settings.DEFAULT_INTERVAL_BETWEEN_ATTEMPTS
        
        self._max_attempts = kwargs.get('max_attempts', default_max_attempts)
        self._interval_between_attempts = kwargs.get('interval_between_attempts', default_interval_between_attempts)
        self._throw_exception_on_failure = kwargs.get('throw_exception_on_failure', True)

    def is_unlimited_attempts(self):
        return self._max_attempts <= 0

    def run(self, *args, **kwargs):
        """
        Executing the specified function max_attempts times once in interval_between_attempts seconds.
        If the function did not return the necessary result at least once, then throw the specified error.
        The arguments args, *kwargs are the arguments for the function being executed.
        """
        attempt = 0
        while attempt < self._max_attempts or self.is_unlimited_attempts():
            result = self._func(*args, **kwargs)
            if self._check_function(result):
                return result

            attempt += 1
            time.sleep(self._interval_between_attempts)

        if self._throw_exception_on_failure:
            raise self._error
        else:
            logger.warning(self._error.message)


def generate_random_password(chars_to_include=''):
    """
    Return a string of random characters.
    Default length is 10 characters if no :chars_to_include: specified.

    Arguments:
    chars_to_include: A list of characters to include. In addition to randomly
    chosen characters, characters from the list will be included in random
    order. The password length will increase accordingly.

    :type chars_to_include: str
    :rtype: str
    """
    chars = [random.choice(string.lowercase) for _ in range(3)] + \
            [random.choice(string.uppercase) for _ in range(3)] + \
            [random.choice(string.digits) for _ in range(4)] + \
            [c for c in chars_to_include]

    random.shuffle(chars)
    return ''.join(chars)


def generate_random_username(prefix):
    """
    Return a random string prefixed with :prefix:.
    Result string length is 15 characters.

    :type prefix: str
    :rtype: str
    """

    if prefix and prefix[0].isdigit():
        # username must start with letter
        prefix = 'a%s' % prefix

    additional_chars = ''
    if len(prefix) >= 5:
        prefix = prefix[:5]
    else:
        additional_chars = ''.join([random.choice(string.lowercase) for _ in xrange(5-len(prefix))])

    username = generate_random_password(additional_chars).lower()
    return ''.join((prefix, username))


def sort_list_by_another(lst, order_lst):
    """
    Sort one list with the order taken from another list
    """
    return [value for value in order_lst if value in lst]


def ilen(iterable):
    """Get number of elements in iterable collection.
    """
    count = 0
    for _ in iterable:
        count += 1
    return count


def partition(iterable, func):
    condition_true = []
    condition_false = []

    for item in iterable:
        if func(item):
            condition_true.append(item)
        else:
            condition_false.append(item)
    return condition_true, condition_false


def poll_data(func, intervals, sleep=time.sleep):
    value = func()
    if value is not None:
        return value

    for interval in intervals:
        sleep(interval)
        value = func()
        if value is not None:
            return value

    return None


def polling_intervals(starting, max_attempt_interval, max_total_time=None):
    """Generate sequence of times for polling some potentially long-lasting process.
    """
    interval = starting
    total = 0

    while interval < max_attempt_interval and (max_total_time is None or total < max_total_time):
        yield interval
        total += interval
        interval *= 2

    while max_total_time is None or total < max_total_time:
        yield max_attempt_interval
        total += max_attempt_interval


def unused(*args):
    """This function does nothing and is used to prevent code analyzing tools from reporting
    variable or imported module as unused.

    Please describe why this unused variable is needed next to function call.
    """
    args = args


def all_equal(lst):
    return len(lst) == 0 or all(item == lst[0] for item in lst[1:])


def cached(func):
    """Cache function results.
    (cached) call the function func as usual: func(args)
    clear cache for function func as: func.clear()
    """
    cache = {}

    @wraps(func)
    def wrapper(*args, **kw):
        key = (args, frozenset(kw.items()))
        if key in cache:
            return cache[key]
        else:
            value = func(*args, **kw)
            cache[key] = value
            return value

    wrapper.clear = lambda: cache.clear()
    return wrapper


def format_list(lst):
    return ", ".join(["'%s'" % (elem,) for elem in lst])


def format_multiline_list(lst):
    return "\n".join(["- %s" % (elem,) for elem in lst])


def group_value_by_id(seq, key_func, value_func):
    """
    Example:

    group_by(
        [obj(x=1, y=2), obj(x=3, y=4), obj(x=5, y=6)],
        lambda o: o.x,
        lambda o: o.y
    )
    {
        1: 2,
        3: 4,
        5: 6
    }
    """
    d = {}
    for item in seq:
        key = key_func(item)
        assert key not in d  # assert that key is really "id" - so it is unique
        d[key] = value_func(item)
    return d


def group_value_by(seq, key_func, value_func):
    """
    Example:

    group_by(
        [obj(x=1, y=2), obj(x=3, y=4), obj(x=3, y=7), obj(x=5, y=6)],
        lambda o: o.x,
        lambda o: o.y
    )
    {
        1: [2],
        3: [4, 7],
        5: [6]
    }
    """
    d = defaultdict(list)
    for item in seq:
        key = key_func(item)
        d[key].append(value_func(item))
    return dict(d)


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise


def find_first(lst, match_func, default_value=None):
    for item in lst:
        if match_func(item):
            return item
    return default_value


def safe_string_repr(string):
    """If string can be interpreted as UTF-8, then return it as unicode string. Otherwise return its __repr__.
    Could be useful for execution_wrapper logging, when usually you want to get just plain string, but if there are some issues
    with encoding - at least get its Python representation.
    """
    try:
        return u"%s" % (string,)
    except UnicodeError:
        return u"%r" % (string,)


def strip_multiline_spaces(string):
    """Utility function that strips leading tabs and spaces from multi-line strings.
    Useful when you want to indent multi-line string in a code, but don't want
    to split it over multiple strings for correct indentation
    """
    return u"\n".join([line.strip() for line in string.strip(" \n\t").split("\n")])


def read_file(filename, reader_func, file_may_be_absent=False, default=None):
    try:
        with file(filename, 'r') as dump_file:
            return reader_func(dump_file)
    except IOError as e:
        if e.errno == errno.ENOENT and file_may_be_absent:
            return default
        else:
            raise


def unique_list(lst):
    result = []
    for item in lst:
        if item not in result:
            result.append(item)
    return result


def extend_namedtuple(base, typename, field_names=[]):
    return namedtuple(typename, list(base._fields) + list(field_names))


def get_executable_file_name():
    return os.path.basename(sys.argv[0])


def is_empty(value):
    return value is None or value == ''


@contextmanager
def open_unicode_file(filename):
    """Open file detecting if it is UTF-8 or UTF-8 with BOM

    See for more details:
    http://stackoverflow.com/questions/13590749/reading-unicode-file-data-with-bom-chars-in-python
    """
    minbytes = min(32, os.path.getsize(filename))
    with open(filename, 'rb') as fp_raw:
        raw = fp_raw.read(minbytes)
        if raw.startswith(codecs.BOM_UTF8):
            encoding = 'utf-8-sig'
        else:
            encoding = 'utf-8'

    with codecs.open(filename, 'r', encoding) as fp:
        yield fp


def merge_dicts(*dicts):
    """Merge several dictionaries, the latest has more priority

    Arguments:
    pass dictionaries as function's arguments, like: merge_dicts(d1, d2, d3)
    """
    result = {}
    for d in dicts:
        for k, v in d.iteritems():
            result[k] = v

    return result


def is_string(o):
    """Check if object is string or not

    :rtype: bool
    """
    return isinstance(o, basestring)


def is_boolean(o):
    """Check if object is boolean or not

    :rtype: bool
    """
    return o in (True, False)


def is_run_on_windows():
    return platform.system() == 'Windows'


def split_nonempty_strs(s, separator=None):
    """Split string by lines, return list of trimmed non-empty lines
    :type s: string
    :rtype: list[string]
    """
    return [d.strip() for d in s.split(separator) if d.strip != '']


def default(passed_value, default_value):
    """Shortcut for defining default function arguments

    For example if you have function:
    def f(x):

    and you want 'x' variable to be a list, which is empty by default,
    you can write:

    def f(x=None):
        x = default(x, [])

    instead of

    def f(x=None):
        if x is None:
            x = []

    Please note that you should not use:
    def f(x=[])
    as the default list will be the same object for all function calls.
    So there could be unexpected results if 'x' variable is modified inside a function.
    """
    if passed_value is None:
        return default_value
    else:
        return passed_value


def get_env_str(env, is_win=False):
    """Transform env variables dict into one line string which can be used to prefix command execution.

    Linux:
        dict(ENV_VAR_1=12, ENV_VAR_2=test) -> ENV_VAR_1=12 ENV_VAR_2=test

    Windows:
        dict(ENV_VAR_1=12, ENV_VAR_2=test) -> SET ENV_VAR_1=12 && SET ENV_VAR_2=test &&

    :type env: dict | None
    :type is_win: bool
    :rtype: unicode
    """
    if env:
        if is_win:
            return u'%s&&' % u'&&'.join(u'SET %s=%s' % (name, value) for name, value in env.iteritems())
        else:
            return u'%s; ' % u' '.join(u"%s='%s'" % (name, value) for name, value in env.iteritems())
    else:
        return u''


# TODO: remove me
def get_thirdparties_base_path():
    return migrator.settings.EXTRAS_PATH


def parse_bool(string, special_cases=None):
    boolean_dict = {
        True: set(['true', '1', 1, True]),
        False: set(['false', '0', 0, False])
    }

    if special_cases:
        for key in boolean_dict:
            if key in special_cases:
                boolean_dict[key].update(special_cases[key])

    if string in boolean_dict[True]:
        return True
    elif string in boolean_dict[False]:
        return False
    else:
        raise Exception('Invalid boolean value "%s"' % string)


def is_subdomain(sub, domain):
    return sub.rstrip('.').endswith('.' + domain.rstrip('.'))


def is_subdomain_of_any(sub, domains, key=None):
    for domain in domains:
        if is_subdomain(sub, domain if key is None else key(domain)):
            return True

    return False


def urlquote(text):
    return urllib.quote(text.encode('utf-8'))
