# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import os.path
import urlparse

from collections import namedtuple
from ConfigParser import NoOptionError,  RawConfigParser
from StringIO import StringIO

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.utils import format_list, open_unicode_file
from migrator.core.utils.lang.entity import Entity
from migrator.core.utils.registry import Registry


logger = logging.getLogger('migrator')
Auth = namedtuple('Auth', ('username', 'password'))


class HttpConnectionSettings(Entity):
    def __init__(self, url, auth):
        self._url = url
        self._auth = auth

    @property
    def url(self):
        return self._url

    @property
    def auth(self):
        return self._auth

    @property
    def port(self):
        return urlparse.urlparse(self.url).port


class ConfigSection(object):
    """Dictionary-like interface to configuration section.

    Also allows to refer part of keys starting with given prefix.
    """
    def __init__(self, config, section_name, prefix=''):
        self.config = config
        self.section_name = section_name
        self.prefix = prefix

    def __getitem__(self, name):
        return self.config.get(self.section_name, self.prefix+name)

    def get(self, name, default=None):
        return self.get_or_exec(name, lambda: default)

    def getboolean(self, name, default=None):
        return self.getboolean_or_exec(name, lambda: default)

    def get_or_exec(self, name, get_default):
        try:
            return self.config.get(self.section_name, self.prefix+str(name))
        except NoOptionError:
            return get_default()

    def getboolean_or_exec(self, name, get_default):
        try:
            return self.config.getboolean(self.section_name, self.prefix+str(name))
        except NoOptionError:
            return get_default()

    def keys(self):
        return [key for key in self.config.options(self.section_name) if key.startswith(self.prefix)]

    def items(self):
        return [(k, v) for (k, v) in self.config.items(self.section_name) if k.startswith(self.prefix)]

    def __contains__(self, name):
        return self.config.has_option(self.section_name, self.prefix+name)

    def prefixed(self, prefix):
        return ConfigSection(self.config, self.section_name, self.prefix+prefix)

    def with_defaults(self, defaults):
        return ConfigSectionWithDefaults(self, defaults)


class ConfigHostSection(ConfigSection):
    DYNAMIC_PROPERTIES = ['username', 'password']
    OS_SUPPORTED_PREFIX = {'linux': 'ssh-', 'windows': 'smb-'}

    def __init__(self, config, section_name):
        super(ConfigHostSection, self).__init__(config, section_name)

    def get(self, name, default=None):
        if name in self.DYNAMIC_PROPERTIES and self.get('os'):
            host_os = self.get('os').lower()
            return self.get(self.OS_SUPPORTED_PREFIX[host_os] + name, default)

        return super(ConfigHostSection, self).get(name, default)

    def are_connection_credentials_empty(self):
        return True if not self.get('username') or not self.get('password') else False


class ConfigSectionWithDefaults(object):
    """Configuration section with default values.
    """
    def __init__(self, section, defaults):
        self.section = section
        self.defaults = defaults

    def __getitem__(self, name):
        return self.section.get_or_exec(name, lambda: self.defaults[name])

    def get(self, name, default=None):
        return self.get_or_exec(name, lambda: default)

    def get_or_exec(self, name, get_default):
        try:
            return self.section.get_or_exec(name, lambda: self.defaults[name])
        except ValueError:
            return get_default()

    def keys(self):
        return set(self.section.keys()) | set(self.defaults.keys())

    def __contains__(self, name):
        return name in self.section or name in self.defaults

    def with_defaults(self, defaults):
        return ConfigSectionWithDefaults(self, defaults)


def get_option(section, option, default):
    context = Registry.get_instance().get_context()
    if context is not None and context.config is not None and option in context.config.options(section):
        return context.config.get(section, option)
    else:
        return default


def read_auth(section):
    return Auth(section.get('username'), section.get('password'))


def read_url(section):
    if 'url' in section:
        url = section['url']
        
    else:
        host = section['host']
        protocol = section['protocol']
        port = int(section['port'])
        path = section['path']
        
        url = urlparse.urlunparse((protocol, '%s:%d' % (host, port), path, '', '', ''))
        
    return url


def read_http_connection_settings(section):
    return HttpConnectionSettings(url=read_url(section), auth=read_auth(section))


def get_sections_list(config, section, option):
    """Parse list of sections in specified option of configuration file and check that each section actually exists

    Consider we have list of sections in option
    =============================================
    [GLOBAL]
    source-servers: source1, source2, source3
    =============================================

    If you call this function (for with arguments (config, 'GLOBAL', 'source-servers')), it will:
    1) Return list of sections: ['source1', 'source2', 'source3']
    2) Check that each of these sections exists in configuration file.
    If section does not exist - exception will be raised. For example, if there is no section like:
    =============================================
    [source3]
    options
    =============================================
    exception will be raised.

    :rtype: list[str]
    """
    if option in config.options(section):
        list_str = config.get(section, option)
        list_str = list_str.strip()
        if list_str == '':
            return []
        sections = [s.strip() for s in list_str.split(',')]

        for section in sections:
            if not config.has_section(section):
                raise MigrationError(
                    _T("SECTION_DOES_NOT_EXIST").format(section=section, option=option)
                )
        return sections
    else:
        return []


class ConfigSelectionOption(object):
    """Description of configuration file option that may have one value from allowed set"""

    @property
    def name(self):
        """Return name of the option

        :rtype: str
        """
        raise NotImplementedError()

    @property
    def allowed_values(self):
        """Return list of allowed values for this option

        :rtype: list[str]
        """
        raise NotImplementedError()

    @property
    def case_insensitive(self):
        """Return true if option is case insensitive, False otherwise

        :rtype: bool
        """
        return True

    @property
    def default(self):
        """Return default value for this option, if it was not specified in configuration file

        :rtype: str
        """
        raise NotImplementedError()


def read_selection_option(section, selection_option):
    """Read selection option (option that may have one value from allowed set) from configuration file

    :type section: parallels.core.utils.config_utils.ConfigSection
    :type selection_option: parallels.core.utils.config_utils.ConfigSelectionOption
    :rtype: str
    """
    def lowercase_if_case_insensitive(v):
        if selection_option.case_insensitive:
            return v.lower()
        else:
            return v

    value = lowercase_if_case_insensitive(section.get(selection_option.name, selection_option.default))
    allowed_values = [
        lowercase_if_case_insensitive(allowed_value) for allowed_value in selection_option.allowed_values
    ]
    if value not in allowed_values:
        raise MigrationError(
            _T("INVALID_OPTION_VALUE").format(
                option_name=selection_option.name,
                specified_value=section.get(selection_option.name),
                allowed_values=format_list(allowed_values)
            )
        )

    return value


def global_section(config):
    return ConfigSection(config, 'GLOBAL')


def get_source_type_section(config):
    """Return a config section that contains server connection information"""
    server_section_name = config.get('GLOBAL', 'source-type')
    return ConfigSection(config, server_section_name)


def read_simple_config(config_name):
    config = {}
    with open(config_name) as f:
        for line in f.readlines():
            strip_line = line.strip()
            if len(strip_line) == 0 or strip_line.startswith('#'):
                continue

            result = strip_line.split('=', 1)
            if len(result) > 1:
                config[result[0].strip()] = result[1].strip()

    return config


def read_config(raw_config_path):
    """Parse the config file."""
    if os.path.exists(raw_config_path):
        config_path = raw_config_path
    else:
        # otherwise we've probably been given a relative path
        config_path = os.path.join(Registry.get_instance().get_var_dir(), 'conf', raw_config_path)

    Registry.get_instance().set_config_path(config_path)

    # TODO: Workaround for Zend_Config_Writer compatibility, which escape values with doublequotes
    class PanelMigratorConfigParser(RawConfigParser):
        def get(self, section, option):
            """
            Retrieve option value from given section
            :type section: str
            :type option: str
            :rtype: str
            """
            return RawConfigParser.get(self, section, option).strip('"')

    config = PanelMigratorConfigParser()
    try:
        # work with encodings so config.ini with UTF-8 BOM
        # is readable on Windows (notepad saves in UTF-8 BOM by default)
        with open_unicode_file(config_path) as fp:
            contents = fp.read()

        config.readfp(StringIO(contents))
    except Exception as e:
        logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
        raise MigrationError(_T("FAILED_TO_READ_CONFIG_FILE") % (config_path, str(e)))

    return config
