# (C)2017 Ingram Micro Inc. all Rights Reserved
import os
import datetime

from cryptography import x509, utils
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa

YEAR = 365  # days in one year


class SSLCertificate(object):
    def __init__(self, cert, pkey, cert_filename='', pkey_filename=''):
        self._pkey = pkey
        self._pkey_filename = pkey_filename
        self._cert = cert
        self._cert_filename = cert_filename

    @property
    def cert(self):
        return self._cert

    @property
    def cert_filename(self):
        return self._cert_filename

    @property
    def pkey(self):
        return self._pkey

    @property
    def pkey_filename(self):
        return self._pkey_filename

    @classmethod
    def generate(cls, dn):
        pkey = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())
        subject = issuer = x509.Name([x509.NameAttribute(k, v) for k, v in dn.items()])

        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            issuer
        ).public_key(
            pkey.public_key()
        ).serial_number(
            _random_serial_number()
        ).not_valid_before(
            # Certificate validity period begins date in the past to avoid issues with date on server
            datetime.datetime.utcnow() - datetime.timedelta(days=YEAR*5)
        ).not_valid_after(
            # Our certificate will be valid for 10 years
            datetime.datetime.utcnow() + datetime.timedelta(days=YEAR*10)
        ).sign(pkey, hashes.SHA256(), default_backend())

        return SSLCertificate(cert, pkey)

    @classmethod
    def load(cls, cert_filename, pkey_filename, password=None):
        with open(pkey_filename, 'rb') as f:
            pkey = serialization.load_pem_private_key(f.read(), password=password, backend=default_backend())

        with open(cert_filename, 'rb') as f:
            cert = x509.load_pem_x509_certificate(f.read(), default_backend())

        return SSLCertificate(cert, pkey, cert_filename=cert_filename, pkey_filename=pkey_filename)

    def save(self, cert_filename, pkey_filename, password=None):
        encryption_algorithm = serialization.NoEncryption()
        if password:
            encryption_algorithm = serialization.BestAvailableEncryption(bytes(password))

        with open(pkey_filename, 'wb') as f:
            os.chmod(pkey_filename, 0600)
            f.write(self._pkey.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=encryption_algorithm,
            ))
            self._pkey_filename = pkey_filename

        with open(cert_filename, 'wb') as f:
            os.chmod(cert_filename, 0600)
            f.write(self._cert.public_bytes(serialization.Encoding.PEM))
            self._cert_filename = cert_filename

    def remove(self):
        for fname in (self.cert_filename, self.pkey_filename):
            if fname and os.path.exists(fname):
                os.unlink(fname)


# TODO: Upgrade cryptography to v1.6 and use x509.random_serial_number() instead
def _random_serial_number():
    return utils.int_from_bytes(os.urandom(20), "big") >> 1
