# (C)2017 Ingram Micro Inc. all Rights Reserved
import imp
import ipaddress
import os
import pwd
import logging

from pipes import quote
from poaupdater import uPgSQL
from ConfigParser import RawConfigParser

from migrator.core.language import _T
from migrator.core import MigrationError
from migrator.core.session.config import SSHAuthKeyDefault
from migrator.core.utils import generate_random_password, mkdir_p
from migrator.core.utils.filetypes.config_utils import read_config, read_simple_config
from migrator.core.utils.networking.ip_utils import get_linux_host_ips
from migrator.core.utils.networking.run_command import LocalUnixRunner, SSHRunner
from migrator.core.utils.networking.ssh.utils import SSHClient
from migrator.core.utils.operating_system.unix_utils import get_or_add_sysgroup, get_or_add_sysuser
from migrator.core.data.source.db_meta import init_db_tables, set_user_password
from migrator.settings import OA_MIGRATOR_CONFIG_DIR, OA_MIGRATOR_DB_CONFIG, OA_KERNEL_CONFIG_FILE, \
    OA_APS_APP_HOME_DIR, OA_APS_APP_SYSUSER, OA_APS_APP_SYSGROUP, PGHA_CONF, SSH_DEFAULT_PORT, PG_DEFAULT_PORT

logger = logging.getLogger(__name__)


def init_db():
    oa_config = read_simple_config(OA_KERNEL_CONFIG_FILE) if os.path.exists(OA_KERNEL_CONFIG_FILE) else None
    is_pgha_installed = os.path.exists(PGHA_CONF)
    if os.path.exists(OA_MIGRATOR_DB_CONFIG):
        config = read_config(OA_MIGRATOR_DB_CONFIG)
        driver = config.get('db', 'driver')
        host = config.get('db', 'host')
        username = config.get('db', 'username')
        password = config.get('db', 'password')
        dbname = config.get('db', 'dbname')
    else:
        driver = 'PDO_PGSQL'
        if oa_config:
            host = oa_config['dsn_host']
        else:
            host = 'localhost'

        username = 'oam'
        password = generate_random_password()
        dbname = 'oam'

    master_host = ''
    master_port = PG_DEFAULT_PORT
    pgha_nodes = []

    if is_pgha_installed:
        module = imp.load_source('pgha_settings', PGHA_CONF)
        master_port = getattr(module, 'PG_PGPOOL_BACKEND_PORT')
        master_host = host
        pgha_nodes.extend([getattr(module, 'A_DB_NODE'), getattr(module, 'B_DB_NODE')])

    oa_runner = LocalUnixRunner()
    db_runner = oa_runner

    if not (host == 'localhost' or ipaddress.ip_address(unicode(host)) in [x['ip'] for x in get_linux_host_ips(oa_runner)]):
        db_runner = get_ssh_runner(host)

    _run_command(
        db_runner,
        '''su - postgres -c 'createuser -S -d -R -h "{host}" -p "{port}" {username}' '''.format(
            username=username, host=master_host, port=master_port
        ),
        error_message='role "{username}" already exists'.format(username=username)
    )
    _run_command(
        db_runner,
        '''su - postgres -c 'createdb -O {username} -h "{host}" -p "{port}" {dbname}' '''.format(
            username=username, dbname=dbname, host=master_host, port=master_port
        ),
        error_message='database "{dbname}" already exists'.format(dbname=dbname)
    )

    if is_pgha_installed:
        oss_dbname = oa_config['dsn_db_name'] if oa_config else 'oss'
        tune_pg_hba(pgha_nodes, oss_dbname, dbname)

    set_user_password(db_runner, master_port, username, password)

    if is_pgha_installed:
        add_user_password_to_pg_pool_password(pgha_nodes, username, password)

    gid = get_or_add_sysgroup(group=OA_APS_APP_SYSGROUP).gr_gid
    get_or_add_sysuser(user=OA_APS_APP_SYSUSER, home_dir=OA_APS_APP_HOME_DIR, group=OA_APS_APP_SYSGROUP)
    # disabling user expire date
    _run_command(oa_runner, 'usermod --inactive -1 {user}'.format(user=OA_APS_APP_SYSUSER))
    uid = pwd.getpwnam('root').pw_uid

    save_migrator_config(uid, gid, driver, host, username, password, dbname)
    init_db_tables()


def save_migrator_config(uid, gid, driver, host, username, password, dbname):
    db_config = RawConfigParser()
    db_config.add_section('db')
    db_config.set('db', 'driver', driver)
    db_config.set('db', 'host', host)
    db_config.set('db', 'username', username)
    db_config.set('db', 'password', password)
    db_config.set('db', 'dbname', dbname)

    if not os.path.exists(OA_MIGRATOR_DB_CONFIG):
        mkdir_p(OA_MIGRATOR_CONFIG_DIR)
        open(OA_MIGRATOR_DB_CONFIG, 'a').close()

        os.chown(OA_MIGRATOR_DB_CONFIG, uid, gid)
        os.chmod(OA_MIGRATOR_DB_CONFIG, 0640)
        with open(OA_MIGRATOR_DB_CONFIG, 'w') as f:
            db_config.write(f)


def get_ssh_runner(host):
    ssh = SSHClient(host)
    SSHAuthKeyDefault(SSH_DEFAULT_PORT, 'root').connect(host, ssh)
    return SSHRunner(ssh, 'Database node %s' % host)


def tune_pg_hba(pgha_nodes, oss_dbname, oam_dbname):
    for host in pgha_nodes:
        runner = get_ssh_runner(host)
        pg_config = uPgSQL.PostgreSQLConfig(commander=lambda x: runner.sh(x))

        # add the same lines for oam database as for oss database
        # remove duplicate strings

        cmd = r"sed --in-place=.oam.`date +%s` -e 's" \
                r"/\(local\|host\|hostssl\|hostnossl\)\(.*\){oss_dbname}\(.*\)\(trust\|reject\|md5\|password\|gss\|sspi\|ident\|peer\|pam\|ldap\|radius\|cert\)" \
                r"/\1\2{oss_dbname}\3\4\n\1\2{oam_dbname}\3\4/' {pg_hba_conf} && " \
              r"sed --in-place -e '$!N; /^\(.*\)\n\1$/!P; D' {pg_hba_conf}".\
            format(oss_dbname=oss_dbname, oam_dbname=oam_dbname, pg_hba_conf=pg_config.get_pghba_conf())

        runner.sh(cmd)
        pg_config.reload()


def add_user_password_to_pg_pool_password(pgha_nodes, username, password):
    # use socat because pg_md5 wants to stdin be pty, not pipe
    cmd = "socat STDIO EXEC:'pg_md5 --md5auth --username={username} --prompt',pty".format(username=username)

    runner = LocalUnixRunner()
    runner.sh(cmd, stdin_content=password + '\n', log_output=False)

    cmd = 'pg_md5 --md5auth --username={username} --prompt'.format(username=username)
    for host in pgha_nodes:
        get_ssh_runner(host).sh(cmd, stdin_content=password + '\n', log_output=False)


def _run_command(runner, command, error_message=None):
    exit_code, stdout_str, stderr_str = runner.sh_unchecked(command)
    if exit_code != 0:
        if error_message is not None and error_message in stderr_str:
            return stdout_str, stderr_str

        raise MigrationError(
            _T("LOCAL_COMMAND_FAILED") % (
                u" ".join([quote(c) for c in command]),
                exit_code,
                stdout_str,
                stderr_str
            )
        )
    return stdout_str, stderr_str
