# (C)2017 Ingram Micro Inc. all Rights Reserved
import inspect
import yaml

from migrator.core.utils import merge_dicts


def entity_pretty_yaml(cls):
    """Set up YAML library to dump/load entity class as mapping."""
    if cls.yaml_prefix is not None:
        prefix_str = "%s." % cls.yaml_prefix
    else:
        prefix_str = ""

    tag = u'!%s%s' % (prefix_str, cls.__name__)

    def namedtuple_representer(dumper, data):
        return dumper.represent_mapping(tag, data.as_dictionary())

    def namedtuple_constructor(loader, node):
        return cls(**loader.construct_mapping(node))

    yaml.add_representer(cls, namedtuple_representer)
    yaml.add_constructor(tag, namedtuple_constructor)

    return cls


class EntityMetaclass(type):
    def __new__(mcs, class_name, class_parents, class_attr):
        c = super(EntityMetaclass, mcs).__new__(mcs, class_name, class_parents, class_attr)
        entity_pretty_yaml(c)
        return c


class Entity(object):
    yaml_prefix = None
    data_prefix = None

    __metaclass__ = EntityMetaclass

    @property
    def properties_list(self):
        """List names of properties of this entity

        By default all constructor arguments are considered as entity properties.

        :rtype: set[str]
        """
        return set(inspect.getargspec(self.__init__).args) - {'self'}

    def as_dictionary(self):
        return {
            p: getattr(self, p)
            for p in self.properties_list
        }

    @classmethod
    def create_none_filled(cls, **kwargs):
        properties = merge_dicts(
            {name: None for name in set(inspect.getargspec(cls.__init__).args) - {'self'}},
            kwargs
        )

        return cls(**properties)

    @staticmethod
    def _meta_attributes():
        return {'properties_list', 'as_dictionary', 'yaml_prefix', 'create_none_filled', 'data_prefix'}

    def __str__(self):
        properties = ', '.join([
            "%s=%s" % (p, self._format_property_value(getattr(self, p)))
            for p in self.properties_list
        ])
        return '%s(%s)' % (self.__class__.__name__, properties)

    def __repr__(self):
        return str(self)

    def _format_property_value(self, prop):
        if isinstance(prop, list):
            return '[%s]' % ', '.join([self._format_property_value(i) for i in prop])
        if isinstance(prop, Entity):
            return str(prop)
        else:
            return repr(prop)

    def __eq__(self, other):
        return type(self) == type(other) and all([
            getattr(self, prop) == getattr(other, prop) for prop in self.properties_list
        ])

    def deep_get_attr(self, attr):
        """Recurses through an attribute chain to get the ultimate value
        :param attr: string Example: 'info.company_address.address1', 'info.admin_contact.phone.country_code'
        :return: int|string|None
        """
        try:
            return reduce(getattr, attr.split('.'), self)
        except Exception:
            return None

    def _fill(self, cast_list={}):
        args, _, _, values = inspect.getargvalues(inspect.getouterframes(inspect.currentframe())[1][0])
        for arg in args:
            val = values[arg]
            for fn in cast_list:
                if arg in cast_list[fn]:
                    val = fn(val)
            setattr(self, arg, val)
