# (C)2017 Ingram Micro Inc. all Rights Reserved
import inspect
import logging
import logging.config
import os

from functools import wraps

import yaml

from migrator.core.utils import obj
from migrator.core.utils.filetypes.config_utils import get_option
from migrator.core.utils.lang.profiler import get_default_steps_profiler
from migrator.core.utils.loggers.secure_logger import CensorLogger
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


class Log(object):
    def configure(self, logging_config_file=None):
        logging.setLoggerClass(CensorLogger)
        logging.captureWarnings(True)
        if logging_config_file:
            self._read_settings_from_file(logging_config_file)
            logging.getLogger(__name__).debug(
                'Logging settings were read from ' + logging_config_file)
        else:
            self._configure_loggers()

    def _read_settings_from_file(self, logging_config):
        """Load log configuration from config, if present."""
        with open(logging_config) as f:
            config = yaml.load(f)
            if not 'logging' in config:
                raise Exception(
                    'Unable to find the section "logging" in config file "%s"' % logging_config)

            log_section = config['logging']
            info_log_path, debug_log_path = self._get_log_files()
            log_section['handlers']['info_file']['filename'] = info_log_path
            log_section['handlers']['debug_file']['filename'] = debug_log_path
            logging.config.dictConfig(log_section)

            levels = {
                logging.CRITICAL: 'CRT',
                logging.ERROR: 'ERR',
                logging.WARNING: 'WRN',
                logging.INFO: 'INF',
                logging.DEBUG: 'DBG',
                logging.NOTSET: 'NOT'
            }
            [logging.addLevelName(x, levels[x]) for x in levels]

    def _configure_loggers(self):
        """Configure logger objects programmatically."""
        registry = Registry.get_instance()

        if get_option('GLOBAL', 'use-separate-log', False):
            # write log into current session directory
            context = registry.get_context()
            info_log_path = context.session_files.get_path_to_info_log()
            debug_log_path = context.session_files.get_path_to_debug_log()
        else:
            # write log in the default location
            info_log_path, debug_log_path = self._get_log_files()

        migrator_logger = logging.getLogger('migrator')
        paramiko_logger = logging.getLogger('paramiko')

        # clean up handlers list to prevent log entries duplicating
        migrator_logger.handlers = []
        paramiko_logger.handlers = []

        migrator_logger.setLevel(logging.DEBUG)
        paramiko_logger.setLevel(logging.WARNING)

        info_file_handler = logging.FileHandler(info_log_path)
        info_file_handler.setLevel(logging.INFO)
        migrator_logger.addHandler(info_file_handler)

        debug_file_handler = logging.FileHandler(debug_log_path)
        debug_file_handler.setLevel(logging.NOTSET)
        migrator_logger.addHandler(debug_file_handler)
        paramiko_logger.addHandler(debug_file_handler)

    def _get_log_files(self):
        registry = Registry.get_instance()
        var_dir = registry.get_var_dir()
        logs_dir = os.path.join(var_dir, 'logs')
        if not os.path.exists(os.path.join(logs_dir)):
            os.mkdir(logs_dir)
        return os.path.join(logs_dir, 'info.log'), os.path.join(logs_dir, 'debug.log')


def trace(func=None, hidelist_request=None, hidelist_response=None):
    class LoggingContext(object):
        SPACE = ' '

        def __init__(self, frame_deep, func, hidelist_request=None, hidelist_response=None, *args, **kwargs):
            self.func = func
            self.args = args
            self.kwargs = kwargs
            self.elapsed_time = obj(elapsed_time=0)
            self.hidelist_request = hidelist_request
            self.hidelist_response = hidelist_response
            self.frame_deep = frame_deep
            self.template = '%s{mark}[%s] %s {info}' % (LoggingContext.SPACE * self.frame_deep, self.frame_deep, self.func.__name__)

        def __enter__(self):
            # TODO: @anikulin log args and kwargs without private data:
            # logger.debug(self.template.format(mark='+++', info='(%s, %s)' % (self.args, self.kwargs)))
            logger.debug(self.template.format(mark='+++', info='(...)'))
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            elapsed_time = '[{elapsed_time:.3f}] second(s)'.format(elapsed_time=self.elapsed_time.elapsed_time)

            if exc_type:
                logger.debug(self.template.format(mark='...', info='*** INTERRUPTED BY EXCEPTION *** %s' % elapsed_time))
            else:
                logger.debug(self.template.format(mark='---', info=elapsed_time))

        def run(self):
            with get_default_steps_profiler().measure_time('regular_function_call', self.func.__name__, self.elapsed_time):
                result = self.func(*self.args, **self.kwargs)
            # TODO @anikulin log functions's return without private data:
            # logger.debug('%s   return %s' % (LoggingContext.SPACE * self.frame_deep, result))
            return result

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            frame_deep = len([x for x in inspect.stack() if x[0].f_code == inspect.currentframe().f_code]) - 1
            with LoggingContext(frame_deep, func, hidelist_request, hidelist_response, *args, **kwargs) as o:
                return o.run()

        return wrapper

    return decorator(func) if func and inspect.isfunction(func) else decorator
