# (C)2017 Ingram Micro Inc. all Rights Reserved
from __future__ import absolute_import

import collections
import copy
import jsonpath
import itertools
import logging

from collections import namedtuple
from contextlib import contextmanager
from lxml import etree

from migrator.core.language import _T
from migrator.core.utils.filetypes.config_utils import get_option

logger = logging.getLogger(__name__)

PositionalArgCommand = collections.namedtuple('PositionalArgCommand', 'command, arguments, serializer')
NamedArgCommand = collections.namedtuple('NamedArgCommand', 'command, arguments, serializer')


# TODO anikulin@ get rid of this class as not used, actually
class BleepingLogger(object):
    """Wipe out sensitive data before log messages to a file.
    
    This object searches for log message arguments with non-primitive
    types(xml.elementtree, list, dictionary), removes sensitive data from them,
    converts the cleaned up arguments to string, then passes message with
    string arguments to logger.
    """

    def __init__(self, logger, xml_serializer=None, context_length=10):
        self.logger = logger
        self.xml_serializer = xml_serializer
        self.context_length = int(context_length)
        if self.context_length < 0:
            self.log_context_length = 0

        self.xml_request_sensitivedata = ['.//passwd', './/password']
        self.cli_sensitive_arguments = ['-passwd']
        self.dict_cli_sensitive_arguments = ['passwd', 'password', 'src_password', 'src_passwd']
        self.message_walk_maxdepth = 11
        self.dummy_text = '***hidden***'

    def info(self, message, *args):
        self.logger.info(message, *self.get_sane_args(*args))

    def debug(self, message, *args):
        self.logger.debug(message, *self.get_sane_args(*args))

    def error(self, message, *args):
        self.logger.error(message, *self.get_sane_args(*args))

    def format(self, message, *args):
        """Clean up the message and return it as a formatted string."""
        return message % tuple(self.get_sane_args(*args))

    def get_sane_args(self, *args):
        """Remove sensitive data from log message arguments."""
        sane_args = []
        for arg in args:
            if isinstance(arg, etree._ElementTree):
                sane_args.append(self.cleanup_xmltree(arg))
            elif isinstance(arg, PositionalArgCommand):
                sane_args.append(self.cleanup_cli_command(arg))
            elif isinstance(arg, NamedArgCommand):
                sane_args.append(self.cleanup_named_arg_command(arg))
            else:
                sane_args.append(arg)

        return sane_args

    def cleanup_cli_command(self, command):
        """Remove cleartext passwords from CLI arguments.
        
        Parameters:
            command: a 'PositionalArgCommand' object
        
        Returns: arguments formatted to string.
        """
        sane_args, do_bleep = [], False
        arguments = command.arguments if command.arguments is not None else []
        for arg in arguments:
            sane_arg = arg if arg == '' or not do_bleep else self.dummy_text
            sane_args.append(sane_arg)
            # for example, if argument is '-passwd', then bleep the next arg
            do_bleep = True if arg in self.cli_sensitive_arguments else False

        return command.serializer(command.command, sane_args)

    def cleanup_named_arg_command(self, command):
        """Remove cleartext passwords from CLI arguments.
        
        Parameters:
            command: a 'NamedArgCommand' object
        
        Returns: a command with arguments formatted to string.
        """
        sane_args = {}
        if command.arguments is not None:
            for arg, arg_value in command.arguments.iteritems():
                if arg_value == '' or arg not in self.dict_cli_sensitive_arguments:
                    sane_args[arg] = arg_value
                else:
                    sane_args[arg] = self.dummy_text

        return command.serializer(command.command, sane_args)

    def cleanup_xmltree(self, xmltree):
        sane_xmltree = copy.deepcopy(xmltree)
        sane_xmlstring = self.cleanup_api_message(sane_xmltree)
        short_xmlstring = self.wrap_multiline_string(sane_xmlstring)
        return short_xmlstring

    def cleanup_api_message(self, xml):
        """Remove sensitive data from API message body; serialize it to string."""
        found_elements = map(xml.findall, self.xml_request_sensitivedata)
        for text_element_to_bleep in itertools.chain(*found_elements):
            text_element_to_bleep.text = self.dummy_text

        xmlstring = unicode(self.xml_serializer(xml), encoding='utf-8')
        return xmlstring

    def wrap_multiline_string(self, message):
        """Replace long multiline message body with ellipsis."""
        if not self.context_length > 0:
            return message
        res_list = message.split('\n')
        if len(res_list) <= self.context_length * 2:
            return message
        else:
            return '\n'.join(res_list[:self.context_length] + ['...'] + res_list[-self.context_length:])

    def get_clean_xmlrpc_uri(self, url):
        """Remove password from XML RPC connection string.
        
        For example, the URL
            admin:setup@10.52.38.8:8440
        - will be changed this way:
            admin:***hidden***@10.52.38.8:8440:
        """
        cut = url.rfind('@')
        if -1 == cut:
            return url
        else:
            return self.dummy_text + url[cut:]


def hide_text(text, hide):
    """Hide text for log/output if 'hide' flag is True, return as is if flag is False

    :type text: str | unicode
    :type hide: bool
    :rtype: str | unicode
    """
    return u'******' if hide else text


def remove_sensitive_data_xml(data, hidelist=None):
    data_censored = etree.fromstring(
        etree.tostring(data) if isinstance(data, etree._ElementTree) else data,
        parser=etree.XMLParser(remove_blank_text=True)
    )
    for path in hidelist or []:
        nodes = data_censored.xpath(path)
        for node in nodes or []:
            node.text = hide_text(node.text, True)

    return etree.tostring(data_censored, pretty_print=True)


def remove_sensitive_data_json(data, hidelist=None):
    try:
        data_censored = copy.deepcopy(data)
        for path in hidelist or []:
            pos = path.rfind('.')
            base_path = path[:pos]
            # jsonpath does not treat '$' as valid path to root
            nodes = jsonpath.jsonpath(data_censored, base_path) if base_path != '$' else [data_censored]
            for node in nodes or []:
                if node is None:
                    continue

                key = path[pos+1:]
                if key in node:
                    node[key] = hide_text(node[key], True)

        return data_censored
    except Exception as e:
        logger.warning(_T('FAILED_TO_CENSOR_SOME_DATA').format(hidelist=hidelist, data=data, error=e))
        return data


# TODO anikulin@ replace BleepingLogger by CensorLogger
class CensorLogger(logging.getLoggerClass()):
    handler = namedtuple('handler', ('type', 'func'))
    _handlers = {
        etree._ElementTree: handler('xml', remove_sensitive_data_xml),
        dict: handler('json', remove_sensitive_data_json),
        list: handler('json', remove_sensitive_data_json),
    }

    def __init__(self, *args, **kwargs):
        self.hidelist = {h.type: set() for h in self._handlers.values()}
        super(CensorLogger, self).__init__(*args, **kwargs)

    def _log(self, level, msg, args, exc_info=None, extra=None):
        censored_args = []
        for arg in args:
            handler = self._handlers.get(type(arg))
            if handler:
                censored_args.append(handler.func(arg, self.hidelist[handler.type]))
            else:
                censored_args.append(arg)

        super(CensorLogger, self)._log(level, msg, tuple(censored_args), exc_info, extra)

    @contextmanager
    def censor(self, hidelist=None):
        for t in self.hidelist:
            if t in hidelist:
                self.hidelist[t] |= set(hidelist[t])

        yield

        for t in self.hidelist:
            if t in hidelist:
                self.hidelist[t] -= set(hidelist[t])


def get_bleeping_logger(logger):
    """Create and return enhanced logger object."""
    default_context_length = 10
    log_context_length = get_option('GLOBAL', 'log-message-context-length', default_context_length)
    return BleepingLogger(logger, context_length=log_context_length)
