# (C)2017 Ingram Micro Inc. all Rights Reserved
"""Utilities for working with IP addresses"""
import re
import ipaddress
import socket

from migrator.core import MigrationNoRepeatError
from migrator.core.language import _T
from migrator.core.utils import cached


class IPAddressType(object):
    """Adapter to work with IPv4 and IPv6 addresses in a single way"""

    @property
    def title(self):
        """Title of IP address type ("IPv4" or "IPv6")

        :rtype: str
        """
        raise NotImplementedError()

    @staticmethod
    def is_valid_ip(ip):
        """Check if specified string is correct address of that type

        :type ip: basestring
        :rtype: bool
        """
        raise NotImplementedError()

    @staticmethod
    def get_for_backup_subscription(subscription):
        """Get IP address of that type for subscription

        :type subscription: parallels.core.dump.data_model.Subscription
        :rtype: basestring
        """
        raise NotImplementedError()

    @staticmethod
    def get_type_for_backup_subscription(subscription):
        """Get IP address type (shared or exclusive) of that type for subscription

        :type subscription: parallels.core.dump.data_model.Subscription
        :rtype: basestring
        """
        raise NotImplementedError()

    @staticmethod
    def set_for_target_model_subscription(subscription, new_ip):
        """Set IP address of that type for subscription

        :type subscription: parallels.core.target_data_model.Subscription
        :type new_ip: basestring
        :rtype: None
        """
        raise NotImplementedError()

    @staticmethod
    def set_type_for_target_model_subscription(subscription, ip_type):
        """Set IP address type for subscription

        :type subscription: parallels.core.target_data_model.Subscription
        :type ip_type: basestring
        :rtype: None
        """
        raise NotImplementedError()

    @staticmethod
    def get_for_ip_mapping(mapping):
        """Get IP address of that type from migration list IP mapping

        :type mapping: parallels.core.migration_list.SubscriptionIPMapping
        :rtype: basestring
        """
        raise NotImplementedError()


class IPv4AddressType(IPAddressType):
    """Adapter over IPv4 address to manage IPv4 and IPv6 addresses in a single way"""

    @property
    def title(self):
        return 'IPv4'

    @staticmethod
    def is_valid_ip(ip):
        """Check if specified string is correct address of that type

        :type ip: basestring
        :rtype: bool
        """
        return is_ipv4(ip)

    @staticmethod
    def get_for_backup_subscription(subscription):
        """Get IPv4 address of subscription

        :type subscription: parallels.core.dump.data_model.Subscription
        :rtype: basestring
        """
        return subscription.ip

    @staticmethod
    def get_type_for_backup_subscription(subscription):
        """Get IP address type (shared or exclusive) of that type for subscription

        :type subscription: parallels.core.dump.data_model.Subscription
        :rtype: basestring
        """
        return subscription.ip_type

    @staticmethod
    def set_for_target_model_subscription(subscription, new_ip):
        """Set IPv4 address for subscription

        :type subscription: parallels.core.target_data_model.Subscription
        :type new_ip: basestring
        :rtype: None
        """
        subscription.web_ip = new_ip

    @staticmethod
    def set_type_for_target_model_subscription(subscription, ip_type):
        """Set IP address type for subscription

        :type subscription: parallels.core.target_data_model.Subscription
        :type ip_type: basestring
        :rtype: None
        """
        subscription.web_ip_type = ip_type

    @staticmethod
    def get_for_ip_mapping(mapping):
        """Get IPv4 address from migration list IP mapping

        :type mapping: parallels.core.migration_list.SubscriptionIPMapping
        :rtype: basestring
        """
        return mapping.v4


class IPv6AddressType(IPAddressType):
    """Adapter over IPv6 address to manage IPv4 and IPv6 addresses in a single way"""

    @property
    def title(self):
        return 'IPv6'

    @staticmethod
    def is_valid_ip(ip):
        """Check if specified string is correct address of that type

        :type ip: basestring
        :rtype: bool
        """
        return is_ipv6(ip)

    @staticmethod
    def get_for_backup_subscription(subscription):
        """Get IPv6 address of subscription

        :type subscription: parallels.core.dump.model.Subscription
        :rtype: basestring
        """
        return subscription.ipv6

    @staticmethod
    def get_type_for_backup_subscription(subscription):
        """Get IP address type (shared or exclusive) of that type for subscription

        :type subscription: parallels.core.dump.data_model.Subscription
        :rtype: basestring
        """
        return subscription.ipv6_type

    @staticmethod
    def set_for_target_model_subscription(subscription, new_ip):
        """Set IPv6 address for subscription

        :type subscription: parallels.core.target_data_model.Subscription
        :type new_ip: basestring
        :rtype: None
        """
        subscription.web_ipv6 = new_ip

    @staticmethod
    def set_type_for_target_model_subscription(subscription, ip_type):
        """Set IPv6 address for subscription

        :type subscription: parallels.core.target_data_model.Subscription
        :type ip_type: basestring
        :rtype: None
        """
        subscription.web_ipv6_type = ip_type

    @staticmethod
    def get_for_ip_mapping(mapping):
        """Get IPv6 address from migration list IP mapping

        :type mapping: parallels.core.migration_list.SubscriptionIPMapping
        :rtype: basestring
        """
        return mapping.v6


def get_ip_address_types():
    """Get IPv4 and IPv6 address types as list

    :rtype: list[parallels.core.utils.ip_utils.IPAddressType]
    """
    return [IPv4AddressType(), IPv6AddressType()]


def is_ipv4(ip_address):
    """Check if specified string is an IPv4 address"""

    # We cannot use socket.inet_pton because it is not available on Windows
    # and may be some Unix OSes
    # https://docs.python.org/2/library/socket.html#socket.inet_pton
    # Availability: Unix (maybe not all platforms).

    # Simple solution taken from:
    # http://stackoverflow.com/questions/319279/how-to-validate-ip-address-in-python
    # This solution is:
    # - quite simple and
    # - does not require any external libraries
    # - works on any platform

    pattern = re.compile(r"""
        ^
        (?:
          # Dotted variants:
          (?:
            # Decimal 1-255 (no leading 0's)
            [3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}
          |
            0x0*[0-9a-f]{1,2}  # Hexadecimal 0x0 - 0xFF (possible leading 0's)
          |
            0+[1-3]?[0-7]{0,2} # Octal 0 - 0377 (possible leading 0's)
          )
          (?:                  # Repeat 0-3 times, separated by a dot
            \.
            (?:
              [3-9]\d?|2(?:5[0-5]|[0-4]?\d)?|1\d{0,2}
            |
              0x0*[0-9a-f]{1,2}
            |
              0+[1-3]?[0-7]{0,2}
            )
          ){0,3}
        |
          0x0*[0-9a-f]{1,8}    # Hexadecimal notation, 0x0 - 0xffffffff
        |
          0+[0-3]?[0-7]{0,10}  # Octal notation, 0 - 037777777777
        |
          # Decimal notation, 1-4294967295:
          429496729[0-5]|42949672[0-8]\d|4294967[01]\d\d|429496[0-6]\d{3}|
          42949[0-5]\d{4}|4294[0-8]\d{5}|429[0-3]\d{6}|42[0-8]\d{7}|
          4[01]\d{8}|[1-3]\d{0,9}|[4-9]\d{0,8}
        )
        $
    """, re.VERBOSE | re.IGNORECASE)

    return pattern.match(ip_address) is not None


def is_ipv6(ip_address):
    """Check if specified string is an IPv6 address
    """

    # See comment inside is_ipv4 function - the same is applicable here.

    pattern = re.compile(r"""
        ^
        \s*                      # Leading whitespace
        (?!.*::.*::)             # Only a single whildcard allowed
        (?:(?!:)|:(?=:))         # Colon iff it would be part of a wildcard
        (?:                      # Repeat 6 times:
            [0-9a-f]{0,4}        #   A group of at most four hexadecimal digits
            (?:(?<=::)|(?<!::):) #   Colon unless preceeded by wildcard
        ){6}                     #
        (?:                      # Either
            [0-9a-f]{0,4}        #   Another group
            (?:(?<=::)|(?<!::):) #   Colon unless preceeded by wildcard
            [0-9a-f]{0,4}        #   Last group
            (?: (?<=::)          #   Colon iff preceeded by exacly one colon
             |  (?<!:)           #
             |  (?<=:) (?<!::) : #
             )                   # OR
         |                       #   A v4 address with NO leading zeros
            (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            (?: \.
                (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            ){3}
        )
        \s*                         # Trailing whitespace
        $
    """, re.VERBOSE | re.IGNORECASE | re.DOTALL)

    return pattern.match(ip_address) is not None


def get_object_for_ip(ip, ipv4_object, ipv6_object):
    """
    Get an object depending on the ip address type
    :param ip: str
    :param ipv4_object:
    :param ipv6_object:
    :return:
    """
    if is_ipv4(ip):
        return ipv4_object
    elif is_ipv6(ip):
        return ipv6_object
    else:
        raise MigrationNoRepeatError(_T("UNKNOWN_IP_ADDRESS_TYPE") % ip)


class IPMapping(object):
    """Class representing IP mapping"""

    def __init__(self):
        self._ip_mapping = {}

    def set(self, source_ip, target_ip):
        """Set mapping of source IP address to target IP address

        :type source_ip: basestring
        :type target_ip: basestring
        :rtype: None
        """
        self._ip_mapping[source_ip] = target_ip

    def get(self, source_ip):
        """Get target IP address for specified source IP address.
        If there is no such source IP in the mapping - return None.

        :type source_ip: basestring
        :rtype: basestring | None
        """
        return self._ip_mapping.get(source_ip)


def get_printable_ip(ip):  # 010.037.004.154 -> 10.37.4.154
    return get_object_for_ip(
        ip,
        lambda ip: '.'.join([str(int(x)) for x in ip.split('.')]),
        lambda ip: ip,
    )(ip)


def get_zerofilled_ip(ip):  # 10.37.4.154 -> 010.037.004.154
    return get_object_for_ip(
        ip,
        lambda ip: '.'.join(x.zfill(3) for x in ip.split('.')),
        lambda ip: ':'.join(x.zfill(4) for x in ip.split(':'))
    )(ip)


def _ip_masks(bits, mask_len):
    if 0 < bits <= mask_len:
        b = 1 << (mask_len - bits)
        return ((1 << mask_len) - b), (b - 1)

    raise ValueError('Invalid IP mask width: %r' % bits)


def ipv4_masks(bits):
    return _ip_masks(bits, 32)


def ipv6_masks(bits):
    return _ip_masks(bits, 128)


def _parse_ip_addr_output(out, ip_type='all'):
    ipv4re = re.compile(r'^\d+:\s*([a-zA-Z\d]+)\s*inet\s+([\d.]+)/?\s*(\d+|peer)\s*.*$')
    ipv6re = re.compile(r'^\d+:\s*([a-zA-Z\d]+)\s*inet6\s+([\da-z:]+)/?\s*(\d+|peer)\s*.*$')

    result = list()
    for line in out.splitlines():
        res4 = None
        res6 = None

        if ip_type in ('all', 'ipv4'):
            res4 = ipv4re.match(line)

        if ip_type in ('all', 'ipv6'):
            res6 = ipv6re.match(line)

        res = res4 or res6
        if not res:
            continue

        device, ip, netmask_bits = res.groups()
        if netmask_bits == 'peer':
            netmask_bits = ''
            is_peer = True
        else:
            is_peer = False

        ip_info = {
            'ip': ipaddress.ip_address(ip),
            'device': device,
            'netmaskBits': netmask_bits,
            'peer': is_peer
        }

        if res4:
            if not is_peer:
                ip_info['netmask'] = ipaddress.ip_network(ipv4_masks(int(netmask_bits))[0])
            ip_info['type'] = 'ipv4'

            # get full device name in case of network alias
            full_dev_name = line.split('\\')[0].split()[-1]
            if full_dev_name.startswith(device):
                ip_info['device'] = full_dev_name.strip()

        if res6:
            if not is_peer:
                ip_info['netmask'] = ipaddress.ip_network(ipv6_masks(int(netmask_bits))[0])
            ip_info['type'] = 'ipv6'

        result.append(ip_info)

    return result


def get_linux_host_ips(runner, scope='all', ip_type='all'):
    assert scope in ('all', 'host', 'global')
    assert ip_type in ('all', 'ipv4', 'ipv6')

    cmd = ['/sbin/ip', '-o']
    if ip_type == 'ipv4':
        cmd.extend(('-4',))
    if ip_type == 'ipv6':
        cmd.extend(('-6',))

    cmd.extend(('addr', 'show'))
    if scope != 'all':
        cmd.extend(('scope', scope))

    out = runner.sh(' '.join(cmd))
    return _parse_ip_addr_output(out, ip_type)


@cached
def get_route_ip(destination_ip, runner=None):
    from migrator.core.utils.networking.run_command import BaseRunner, LocalUnixRunner, WindowsRunner

    if not runner:
        runner = LocalUnixRunner()

    destination_ip = destination_ip.encode('idna')
    resolve_error = MigrationNoRepeatError(_T("UNABLE_TO_RESOLVE_ROUTE_TO_HOST").format(host=destination_ip))

    if isinstance(runner, WindowsRunner):
        result = runner.sh(
            'cmd /c "pathping -n -w 1 -h 1 -q 1 %s | findstr /R /C:\" 0 [0-9]*.[0-9]*.[0-9]*.[0-9]*\""' % destination_ip
        ).split()

        if len(result) > 1:
            return result[1]

        raise resolve_error
    elif isinstance(runner, BaseRunner):
        # if destination IP is a hostname in fact
        if not is_ipv4(destination_ip) and not is_ipv6(destination_ip):
            try:
                destination_ip = socket.gethostbyname(destination_ip)
            except socket.gaierror:
                raise resolve_error

        result = runner.sh(
            "`whereis ip | awk \'{print $2;exit;}\'` route get %s | awk \'{print $NF; exit}\'" % destination_ip
        ).split()

        if len(result) > 0:
            return result[0]

        raise resolve_error
