# (C)2017 Ingram Micro Inc. all Rights Reserved
"""
Single interface for running commands via different transports (for example, SSH, Windows HCL, Unix HCL).
"""
import errno
import grp
import logging
import ntpath
import os
import pipes
import pwd
import re
import shutil

from contextlib import closing, contextmanager

from migrator.core import MigrationError, CommandExecutionError, run_and_check_local_command
from migrator.core import local_command_exec
from migrator.core.language import _T
from migrator.core.utils import \
    poll_data, polling_intervals, cached, safe_string_repr, strip_multiline_spaces, \
    split_nonempty_strs, default, get_env_str, generate_random_password
from migrator.core.utils.concurrency.threading_utils import LocksByKey
from migrator.core.utils.lang.profiler import sleep, get_default_steps_profiler
from migrator.core.utils.loggers.secure_logger import \
    PositionalArgCommand, NamedArgCommand, hide_text, get_bleeping_logger
from migrator.core.utils.networking.dns import resolve
from migrator.core.utils.networking.ssh import utils as ssh_utils
from migrator.core.utils.operating_system import unix_utils, windows_utils
from migrator.core.utils.operating_system.windows_utils import check_windows_national_symbols_command
from migrator.core.utils.registry import Registry
from migrator.core.utils.unique_name_generator import get_unique_filename

logger = logging.getLogger(__name__)
profiler = get_default_steps_profiler()


class SSHRunnerException(CommandExecutionError):
    pass


class BaseRunnerException(CommandExecutionError):
    pass


class ConnectionCheckError(MigrationError):
    pass


class WinexeRunnerException(CommandExecutionError):
    pass


class WinexeExecuteError(MigrationError):
    def __init__(self, message, host):
        super(MigrationError, self).__init__(message)
        self.host = host


class BaseRunner(object):
    def run_unchecked(self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None):
        """Run command with specified args. Arguments are properly escaped."""
        raise NotImplementedError()

    def sh_unchecked(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None,
        error_policy='strict', env=None, log_output=True
    ):
        """Substitute format fields in 'cmd_str', then execute the command string.

        For Unix: whole command is executed, then if you run

            "runner.sh('ls *')"

        , '*' will be substituted by shell, while if you run

            "runner.sh('ls {file}', dict(file='*'))"

        , '*' will be escaped, much like "run('ls', ['*'])".

        For Windows: there are no general command arguments parsing rules, and
        '*' is subsitituded by each particular program, this function differs
        from run_unchecked only by a way to substitute arguments.

        If None is passed as args, no command formatting is performed, command
        run as is.

        Returns a tuple (exit_code, stdout, stderr)
        """
        raise NotImplementedError()

    def run(
        self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True, valid_codes=None
    ):
        """The same as run_unchecker(), but checks exit code and returns only stdout"""
        exit_code, stdout, stderr = self.run_unchecked(cmd, args, stdin_content, output_codepage, error_policy, env)

        valid_exit_codes = valid_codes if valid_codes is not None else [0]

        if exit_code not in valid_exit_codes:
            self._raise_runner_exception(cmd, args, exit_code, stdout, stderr, log_output)

        return stdout

    def sh(
        self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True, valid_codes=None
    ):
        """The same as sh_unchecked(), but checks exit code and returns only stdout"""
        exit_code, stdout, stderr = self.sh_unchecked(
            cmd, args, stdin_content, output_codepage, error_policy, env, log_output=log_output
        )

        valid_exit_codes = valid_codes if valid_codes is not None else [0]

        if exit_code not in valid_exit_codes:
            self._raise_runner_exception(cmd, args, exit_code, stdout, stderr, log_output)

        return stdout

    def sh_multiple_attempts(
        self, cmd_str, args=None, stdin=None, codepage=None,
        error_policy='strict', env=None, max_attempts=5,
        interval_between_attempts=10, log_output=True, valid_codes=None
    ):
        """Run command with several attempts; check exit code; return only stdout."""
        blogger = get_bleeping_logger(logger)
        command = self._get_command_object(cmd_str, args, self._get_command_str)
        for attempt in range(0, max_attempts):
            blogger.debug(_T("COMMAND_RUN_ATTEMPTS"), command)
            exit_code, stdout, stderr = self.sh_unchecked(
                cmd_str, args, stdin, codepage, error_policy, env, log_output=log_output
            )
            valid_exit_codes = valid_codes if valid_codes is not None else [0]
            if exit_code in valid_exit_codes:
                if attempt > 0:
                    blogger.info(_T("COMMAND_EXECUTED_OK_AFTER_ATTEMPTS"), command, attempt+1)
                return stdout
            elif attempt >= max_attempts - 1:
                failed_cmd = blogger.format('%s', command)
                self._raise_runner_exception(failed_cmd, args, exit_code, stdout, stderr, log_output)
            else:
                blogger.error(_T("FAILED_EXECUTE_REMOTE_COMMAND_RETRY_IN") % interval_between_attempts)
                sleep(interval_between_attempts, _T("SLEEP_RETRY_EXECUTING_COMMAND"))

    @staticmethod
    def _get_command_object(command, args, serializer):
        if isinstance(args, dict):
            command_class = NamedArgCommand
        elif isinstance(args, list):
            command_class = PositionalArgCommand
        else:
            if args is None:
                return command
            else:
                return '%s %s' % (command, args)

        return command_class(command, args, serializer)

    def _raise_runner_exception(self, cmd_str, args, exit_code, stdout, stderr, log_output=True):
        raise BaseRunnerException(
            _T("RUNNER_COMMAND_FAILED_WITH_EXIT_CODE") % (
                cmd_str, args, exit_code,
                safe_string_repr(hide_text(stdout, not log_output)),
                safe_string_repr(hide_text(stderr, not log_output))
            ),
            stdout=stdout, stderr=stderr, exit_code=exit_code, command_str=cmd_str,
        )

    def upload_file_content(self, filename, content):
        raise NotImplementedError()

    def upload_file(self, local_filename, remote_filename):
        raise NotImplementedError()

    def get_file(self, remote_filename, local_filename):
        raise NotImplementedError()

    def get_file_contents(self, remote_filename):
        raise NotImplementedError()

    def move(self, src_path, dst_path):
        raise NotImplementedError()

    def remove_file(self, filename):
        """Remove file. If file does not exist - silently do nothing.

        :type filename: basestring
        """
        raise NotImplementedError()

    def remove_directory(self, directory, is_remove_root=True, log_output=True):
        raise NotImplementedError()

    def clean_directory(self, directory):
        self.remove_directory(directory, False)

    def file_exists(self, filename):
        raise NotImplementedError()

    def mkdir(self, dirname):
        raise NotImplementedError()

    def mktempfile(self, dirname):
        raise NotImplementedError()

    def download_directory(self, remote_directory, local_directory):
        raise NotImplementedError()

    def check(self):
        """Check connection with that runner

        :rtype None
        """
        raise NotImplementedError()

    def get_files_list(self, path):
        raise NotImplementedError()

    @staticmethod
    def _get_command_str(cmd_str, args=None):
        raise NotImplementedError()

    def has_timeout_capability(self):
        """Return true, if runner is able to set timeout for command execution"""
        return False

    def stat(self, path):
        raise NotImplementedError()

    def chmod(self, path, perm, recursive=False, action="set"):
        raise NotImplementedError()

    def chown(self, path, user, group, recursive=False):
        raise NotImplementedError()


class WindowsRunner(BaseRunner):
    """Add codepage detection to migrator's Windows CMD interface classes."""
    @property
    @cached
    def codepage(self):
        """Determine code page used by Windows CMD.

        Registry key 'HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Nls\CodePage\OEMCP'
        contains current Windows console codepage.

        Python codec name is derived from the key value output by adding 'cp' to code
        page number, for example: '866' -> 'cp866'

        Returns:
            Python codec name to be used for decoding CMD output.
        """
        def _get_cp(text):
            match = re.search('^\s*%s\s+REG_(EXPAND_)?SZ\s+(.*?)\s*$' % re.escape('OEMCP'), text, re.MULTILINE)
            if match is not None:
                return match.group(2).strip()
            return None

        code, stdout, stderr = self._run_command_and_decode_output(
            'reg query "HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Nls\CodePage" /v "OEMCP"',
            'ascii', error_policy='ignore'
        )
        if code != 0 or not _get_cp(stdout):
            raise BaseRunnerException(
                _T("UNABLE_DETECT_WINDOWS_CODEPAGE") % (safe_string_repr(stdout), safe_string_repr(stderr))
            )
        else:
            codepage = u"cp%s" % _get_cp(stdout)
            logger.debug(_T("DEBUG_DETECTED_CODEPAGE") % codepage)

        return codepage

    def move(self, src_path, dst_path):
        self.sh('cmd /c move {src_path} {dst_path}', dict(
            src_path=src_path,
            dst_path=dst_path
        ))

    def get_files_list(self, path):
        output = self.sh('cmd /c dir /b {path}', dict(path=path))
        return split_nonempty_strs(output.strip(), "\n")

    def remove_file(self, filename):
        self.sh('cmd /c del {filename}', dict(filename=filename))

    def remove_directory(self, directory, is_remove_root=True, log_output=True):
        # remove all files in a directory recursively
        self.sh('cmd /c del {directory} /s /q', dict(directory=directory), log_output=log_output)
        # remove all directories recursively
        directories_to_delete = directory if is_remove_root else ntpath.join(directory, '*')
        self.sh(
            u'cmd /c for /d %p in ({directories_to_delete}) do rmdir "%p" /s /q',
            dict(directories_to_delete=directories_to_delete),
            log_output=log_output
        )

    def _run_command_and_decode_output(self, cmd_str, output_codepage=None, error_policy='strict', env=None):
        """Run a command, return its output decoded to UTF."""
        raise NotImplementedError()

    def file_exists(self, filename):
        return windows_utils.file_exists(self, filename)


class LocalUnixRunner(BaseRunner):
    """Execute commands on local server"""
    def __init__(self):
        self.host_description = 'localhost'

    @property
    def codepage(self):
        return 'utf-8'

    def run_unchecked(self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None):
        command = [cmd] + ([] if args is None else args)
        codepage = output_codepage if output_codepage else self.codepage
        return local_command_exec(command, stdin_content, codepage, error_policy, env)

    def sh_unchecked(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None,
        error_policy='strict', env=None, log_output=True
    ):
        if args is None:
            command = ['/bin/sh', '-c', cmd_str]
        else:
            formatted_cmd = unix_utils.format_command(cmd_str, **args)
            command = ['/bin/sh', '-c', formatted_cmd]
        codepage = output_codepage if output_codepage else self.codepage
        return local_command_exec(command, stdin_content, codepage, error_policy, env, log_output=log_output)

    def mkdir(self, dirname):
        return self.sh('mkdir -p {dirname}', dict(dirname=dirname))

    def upload_file_content(self, filename, content):
        with open(filename, "wb") as fp:
            fp.write(content.encode('utf-8'))

    def upload_file(self, local_filename, remote_filename):
        if local_filename != remote_filename:
            shutil.copy2(local_filename, remote_filename)

    def get_file(self, remote_filename, local_filename):
        if local_filename != remote_filename:
            shutil.copy2(remote_filename, local_filename)

    def get_file_contents(self, remote_filename):
        with open(remote_filename, "rb") as fp:
            return fp.read().decode('utf-8')

    def move(self, src_path, dst_path):
        shutil.move(src_path, dst_path)

    def get_files_list(self, path):
        return os.listdir(path)

    def remove_file(self, filename):
        try:
            os.remove(filename)
        except OSError as e:
            # if file does not exist - just skip
            if e.errno != errno.ENOENT:
                raise

    def check(self):
        # No need to check local execution, it should never fail
        pass

    def file_exists(self, filename):
        return os.path.exists(filename)

    def chmod(self, path, perm, recursive=False, action="set"):
        os.chmod(path, perm)
        if recursive:
            for root, dirs, files in os.walk(path):
                [os.chmod(os.path.join(root, d), perm) for d in dirs]
                [os.chmod(os.path.join(root, f), perm) for f in files]

    def chown(self, path, user, group, recursive=False):
        uid = pwd.getpwnam(user).pw_uid
        gid = grp.getgrnam(group).gr_gid
        os.chown(path, uid, gid)
        if recursive:
            for root, dirs, files in os.walk(path):
                [os.chown(os.path.join(root, d), uid, gid) for d in dirs]
                [os.chown(os.path.join(root, f), uid, gid) for f in files]


class LocalWindowsRunner(WindowsRunner):
    """Execute commands on local Windows server"""

    def __init__(self, disable_profiling=False):
        self._disable_profiling = disable_profiling

    @staticmethod
    @cached
    def get_instance(disable_profiling=False):
        return LocalWindowsRunner(disable_profiling)

    def _run_command_and_decode_output(self, cmd_str, output_codepage=None, error_policy='strict', env=None):
        codepage = output_codepage if output_codepage else self.codepage
        return local_command_exec(cmd_str, None, codepage, error_policy, env)

    def sh_unchecked(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None,
        error_policy='strict', env=None, log_output=True
    ):
        if args is not None:
            cmd_str = windows_utils.format_command(cmd_str, **args)
        logger.debug(_T("EXECUTE_COMMAND_ON_LOCAL_SERVER"), cmd_str)
        codepage = output_codepage if output_codepage else self.codepage

        if not self._disable_profiling:
            with profiler.measure_command_call(cmd_str, _T("LOCAL_SERVER_TITLE")):
                return local_command_exec(cmd_str, stdin_content, codepage, error_policy, env)
        else:
            return local_command_exec(cmd_str, stdin_content, codepage, error_policy, env)

    def run_unchecked(self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None):
        command = windows_utils.format_command_list(cmd, args)
        args = None
        return self.sh_unchecked(command, args, stdin_content, output_codepage, error_policy, env)

    def upload_file(self, local_filename, remote_filename):
        shutil.copy2(local_filename, remote_filename)

    def upload_file_content(self, filename, content):
        with open(filename, "wb") as fp:
            fp.write(content)

    def get_file(self, remote_filename, local_filename):
        if local_filename != remote_filename:
            logger.debug(_T("DEBUG_COPY_LOCAL_FILE"), remote_filename, local_filename)
            shutil.copy2(remote_filename, local_filename)

    def get_file_contents(self, remote_filename):
        with open(remote_filename, 'rb') as fp:
            return fp.read().decode('utf-8')

    def move(self, src_path, dst_path):
        shutil.move(src_path, dst_path)

    def get_files_list(self, path):
        return os.listdir(path)

    def mkdir(self, dirname):
        self.sh('cmd /c if not exist {dirname} ( md {dirname} )', dict(dirname=dirname))

    def remove_file(self, filename):
        try:
            os.remove(filename)
        except OSError as e:
            # if file does not exist - just skip
            if e.errno != errno.ENOENT:
                raise

    def remove_directory(self, directory, is_remove_root=True, log_output=True):
        for root, dirs, files in os.walk(directory):
            for name in files:
                self.remove_file(ntpath.join(directory, name))
            for name in dirs:
                self.remove_directory(ntpath.join(directory, name))
            if is_remove_root:
                os.rmdir(root)

    def check(self):
        # No need to check local execution, it should never fail
        pass


class SSHRunner(BaseRunner):
    def __init__(self, ssh, host_description=''):
        self.ssh = ssh
        self.host_description = host_description

    @property
    def codepage(self):
        return 'utf-8'

    def run_unchecked(
            self, cmd, args=None, stdin_content=None, codepage=None, error_policy='strict', env=None, log_output=True
    ):
        """Execute a command with a list of positional 'args'."""
        formatter = unix_utils.format_command_list
        return self._exec(
            cmd, args, formatter, stdin_content, codepage, error_policy, env, log_output=log_output
        )

    def sh_unchecked(
        self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict',
        env=None, log_output=True
    ):
        """Format and execute a command with a dict of named 'args'."""
        formatter = self._get_command_str
        return self._exec(
            cmd, args, formatter, stdin_content, output_codepage, error_policy, env, log_output=log_output
        )

    def _exec(self, cmd, args, formatter, stdin_content, output_codepage, error_policy, env, log_output=True):
        self._log_command(cmd, args, formatter)
        codepage = output_codepage if output_codepage else self.codepage
        command_str = formatter(cmd, args)
        with profiler.measure_command_call(command_str, self.host_description):
            env = default(env, {})
            # Execute all commands under 'en_US.utf-8' locale, if it is not specified,
            # so migrator works in the same way regardless of server's locale
            if 'LANG' not in env:
                env['LANG'] = 'en_US.utf-8'

            return ssh_utils.run_unchecked(
                self.ssh, command_str, stdin_content, codepage, error_policy, env, log_output=log_output
            )

    def _log_command(self, cmd, args, formatter):
        """Write command string to log while omitting sensitive data."""
        blogger = get_bleeping_logger(logger)
        if isinstance(args, dict):
            command = NamedArgCommand(cmd, args, formatter)
        else:
            command = PositionalArgCommand(cmd, args, formatter)
        blogger.debug(_T("DEBUG_EXECUTE_SSH_REMOTE_COMMAND"), self.ssh.hostname, command)

    def sh(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True, valid_codes=None
    ):
        """The same as sh_unchecked(), but checks exit code and returns only stdout"""
        return self.sh_multiple_attempts(
            cmd_str, args, stdin_content, output_codepage, error_policy, env,
            max_attempts=1, log_output=log_output, valid_codes=valid_codes
        )

    def upload_file(self, local_filename, remote_filename):
        with closing(self.ssh.open_sftp()) as sftp:
            sftp.put(local_filename, remote_filename)

    def upload_file_content(self, filename, content):
        with closing(self.ssh.open_sftp()) as sftp:
            with closing(sftp.open(filename, 'w')) as fp:
                fp.write(content)

    def get_file(self, remote_filename, local_filename):
        with closing(self.ssh.open_sftp()) as sftp:
            sftp.get(remote_filename, local_filename)

    def get_file_contents(self, remote_filename):
        return self.run('/bin/cat', [remote_filename], log_output=False)

    def move(self, src_path, dst_path):
        return self.run('/bin/mv', [src_path, dst_path])

    def get_files_list(self, path):
        return unix_utils.get_files_list(self, path)

    def remove_file(self, filename):
        return self.run('/bin/rm', ['-f', filename])

    def remove_directory(self, directory, is_remove_root=True, log_output=True):
        if is_remove_root:
            return self.run('/bin/rm', ['-rf', directory], log_output=log_output)
        else:
            return self.run(
                '/bin/find',
                [directory, '-maxdepth', '1', '-not', '-path', directory, '-exec', 'rm', '-rf', '{}', ';'],
                log_output=log_output
            )

    def mkdir(self, dirname):
        return self.run('/bin/mkdir', ['-p', dirname])

    def mktempfile(self, dirname):
        """Create a 'filename' file with unique name."""
        self.mkdir(dirname)
        tmpfilename = self.run('/bin/mktemp', ['-p', dirname])
        return tmpfilename.strip()

    def mktempdir(self, dirname):
        """Create a 'dirname' subdirectory with unique name."""
        self.mkdir(dirname)
        tmpdirname = self.run('/bin/mktemp', ['-d', '-p', dirname])
        return tmpdirname.strip()

    def _raise_runner_exception(self, command, args, exit_code, stdout, stderr, log_output=True):
        if self.host_description is not None:
            where = " " + _T("EXECUTED_AT") % (self.host_description,)
        else:
            where = ""

            command = self._get_command_str(command, args)
        raise SSHRunnerException(
            _T("COMMAND_COMMAND_STRWHERE_RETURNED_NONZERO_EXIT").format(
                command_str=command, where=where
            ),
            command_str=command,
            stdout=stdout,
            stderr=stderr,
            host_description=self.host_description,
            exit_code=exit_code,
        )

    @staticmethod
    def _get_command_str(cmd_str, args=None):
        """Format command string with named formatting arguments."""
        if args is not None and len(args) > 0:
            cmd_str = unix_utils.format_command(cmd_str, **args)
        return cmd_str

    def file_exists(self, filename):
        return unix_utils.file_exists(self, filename)

    def check(self):
        # Integrity stub: at the moment, it is impossible to create the Class instance without real connecting to host
        pass

    def stat(self, path):
        return unix_utils.path_stat(self, path)

    def chmod(self, path, perm, recursive=False, action="set"):
        return self.run('/bin/chmod', ['--recursive'] if recursive else [] + ['{:o}'.format(perm) if isinstance(perm, int) else perm, path])

    def chown(self, path, user, group, recursive=False):
        return self.run('/bin/chown', ['--recursive'] if recursive else [] + ['%s:%s' % (user, group), path])


class WinexeRunner(WindowsRunner):
    """Remotely execute Windows CMD commands on Linux using 'winexe' tool."""

    _locks = LocksByKey()

    def __init__(self, settings, runas=False, local_temp_dir=None, host_description=''):
        self.settings = settings
        self.runas = runas
        self.local_temp_dir = local_temp_dir
        self.host_description = host_description
        self.timeout = None

        server = Registry().get_instance().get_context().migrator_server
        self.auth_file = server.get_session_file_path('auth_%s_%s' % (id(self), self.windows_host_ip))
        with server.runner() as runner:
            auth = self.settings.windows_auth
            runner.upload_file_content(self.auth_file, '')
            runner.chown(self.auth_file, 'root', 'root')
            runner.chmod(self.auth_file, 0600)
            runner.upload_file_content(
                self.auth_file,
                u'username={username}\npassword={password}\n'.format(username=auth.username, password=auth.password)
            )

        self.check()

    @property
    @cached
    def windows_host_ip(self):
        """Resolve hostname to avoide NT_STATUS_DUPLICATE_NAME errors when connecting to Windows hosts."""
        addr = resolve(self.settings.ip)
        if addr is None:
            raise MigrationError(_T("DNS_RESOLUTION_ERROR").format(hostname=self.settings.ip))

        return addr

    def has_timeout_capability(self):
        with self.set_timeout(None):
            exit_code, _, _ = local_command_exec('timeout --version'.split())

        return exit_code == 0

    def run_unchecked(
        self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True
    ):
        if stdin_content is not None:
            raise NotImplementedError(_T("RUNNER_DOES_NOT_SUPPORT_STDIN_CONTENT_3"))

        command_str = windows_utils.format_command_list(cmd, args)
        with profiler.measure_command_call(command_str, self.windows_host_ip):
            return self._run_command_and_decode_output(
                command_str, output_codepage, error_policy, env, log_output=log_output
            )

    @contextmanager
    def set_timeout(self, timeout):
        """Set timeout for the commands further executed"""
        existing_timeout = self.timeout
        self.timeout = timeout
        try:
            yield
        finally:
            self.timeout = existing_timeout

    def sh_unchecked(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True
    ):
        if stdin_content is not None:
            raise NotImplementedError(_T("RUNNER_DOES_NOT_SUPPORT_STDIN_CONTENT_4"))

        if args is not None:
            cmd_str = windows_utils.format_command(cmd_str, **args)

        with profiler.measure_command_call(cmd_str, self.windows_host_ip):
            return self._run_command_and_decode_output(
                cmd_str, output_codepage, error_policy, env, log_output=log_output
            )

    def sh(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True, valid_codes=None
    ):
        """The same as sh_unchecked(), but checks exit code and returns only stdout"""
        return self.sh_multiple_attempts(
            cmd_str, args, stdin_content, output_codepage, error_policy,
            env, max_attempts=1, log_output=log_output, valid_codes=valid_codes
        )

    def _run_command_and_decode_output(
        self, cmd_str, output_codepage=None, error_policy='strict', env=None, log_output=True
    ):
        """Run a command, return its output decoded from 'codepage' to UTF."""
        if env is not None:
            cmd_str = u'%s%s' % (get_env_str(env), cmd_str)

        check_windows_national_symbols_command(cmd_str)
        formatted_command = self._construct_winexe_command(cmd_str)
        codepage = output_codepage if output_codepage else self.codepage

        errors = []

        def run_winexe_command():
            with self._locks.lock(self.windows_host_ip):
                exit_code, stdout, stderr = local_command_exec(
                    formatted_command, None, codepage, error_policy, log_output=log_output
                )

            # remove winexe warning with no impact on us
            stderr = stderr.replace(_T("DOS_CHARSET_CP_UNAVAILABLE_USING_ASCII"), '')

            if stderr.find(u"Failed to open connection - NT_STATUS_LOGON_FAILURE") != -1:
                errors.append((_T("WINEXE_NT_STATUS_LOGON_FAILURE").format(host=self.windows_host_ip), False))
                return None
            elif stderr.find(u"Failed to open connection") != -1:
                errors.append((_T("WINEXE_FAILED_TO_OPEN_CONNECTION").format(host=self.windows_host_ip), False))
                return None
            elif stderr.find(u"NT_STATUS_") != -1:
                errors.append((_T("COMMAND_S_FAILED_EXECUTE_S_ERROR_2") % (
                    u" ".join([pipes.quote(c) for c in formatted_command]),
                    stderr.strip()
                ), True))
                return None
            else:
                return exit_code, stdout, stderr

        # run winexe command once, and if it returned None, run it one more time after 10 seconds sleep
        result = poll_data(run_winexe_command, polling_intervals(
            starting=10, max_attempt_interval=10, max_total_time=10))
        if result is not None:
            return result
        else:
            logger.debug(_T("ERRORS_WHEN_EXECUTING_COMMAND") % "\n".join([error[0] for error in errors]))
            last_error = errors[len(errors)-1]
            if last_error[1]:
                raise WinexeExecuteError(last_error[0], host=self.windows_host_ip)
            else:
                raise MigrationError(last_error[0])

    def upload_file(self, local_filename, remote_filename):
        # implementation note: winexe runner utilizes smbclient to upload files
        self._smbclient_upload_file(local_filename, remote_filename)

    def upload_file_content(self, filename, content):
        if self.local_temp_dir is None:
            raise Exception(_T("UNABLE_UPLOAD_FILE_CONTENTS_S_DIRECTORY") % (filename,))

        # implementation note: winexe runner utilizes smbclient to upload files
        temp_filename = os.path.join(self.local_temp_dir, get_unique_filename(prefix='winexe_upload-'))
        try:
            with open(temp_filename, "wb") as f:
                f.write(content)

            self._smbclient_upload_file(temp_filename, filename)
        finally:
            if os.path.isfile(temp_filename):
                os.remove(temp_filename)

    def get_file(self, remote_filename, local_filename):
        # implementation note: winexe runner utilizes smbclient to download files
        self._smbclient_download_file(remote_filename, local_filename)

    def get_file_contents(self, remote_filename):
        if self.local_temp_dir is None:
            raise Exception(_T("UNABLE_DOWNLOAD_FILE_CONTENTS_S_DIRECTORY") % (remote_filename,))

        temp_filename = os.path.join(self.local_temp_dir, get_unique_filename(prefix='winexe_download-'))
        try:
            self._smbclient_download_file(remote_filename, temp_filename)
            with open(temp_filename, 'rb') as f:
                return f.read()
        finally:
            if os.path.isfile(temp_filename):
                os.remove(temp_filename)

    def download_directory(self, remote_directory, local_directory):
        with Registry().get_instance().get_context().migrator_server.runner() as local_runner:
            local_runner.mkdir(local_directory)
        self._smbclient_download_directory_contents(remote_directory, local_directory)

    def mkdir(self, dirname):
        self.sh('cmd /c if not exist {dirname} ( md {dirname} )', dict(dirname=dirname))

    def mktempfile(self, dirname):
        self.mkdir(dirname)
        tmpfilename = windows_utils.path_join(dirname, generate_random_password())
        self.sh('cmd /c echo off > "%s"' % tmpfilename)
        return tmpfilename.strip()

    def mktempdir(self, dirname):
        tmpdirname = windows_utils.path_join(dirname, generate_random_password())
        self.mkdir(tmpdirname)
        return tmpdirname.strip()

    def check(self):
        try:
            logger.debug('Try to (re)install winexe to %s' % self.host_description)
            with self._locks.lock(self.windows_host_ip):
                exit_code, stdout, stderr = local_command_exec(
                    self._construct_winexe_command('cmd /c REM') + ['--reinstall']
                )
                if exit_code:
                    raise Exception(stderr)
        except Exception as err:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise ConnectionCheckError(strip_multiline_spaces(
                _T('FAILED_TO_INSTALL_WINEXE').format(server=self.host_description, err_msg=err)
            ))

    def _construct_winexe_command(self, cmd_str):
        winexe_command = []
        if self.timeout is not None and self.has_timeout_capability():
            winexe_command.extend([u'timeout', unicode(self.timeout)])

        winexe_command.extend([u'winexe', u'-A', self.auth_file, u'//%s' % self.windows_host_ip, cmd_str])

        if self.runas:
            winexe_command.extend([u'--runas-file', self.auth_file])

        return winexe_command

    @staticmethod
    def _get_cmd_str_listargs(cmd, args):
        return " ".join([windows_utils.quote_arg(arg) for arg in [cmd] + args])

    @staticmethod
    def _get_cmd_str_formatargs(cmd_str, args):
        if args is not None:
            return windows_utils.format_command(cmd_str, **args)
        else:
            return cmd_str

    def _raise_runner_exception(self, command_str, args, exit_code, stdout, stderr, log_output=True):
        command_str = self._get_command_str(command_str, args)
        full_command = self._construct_winexe_command(command_str)
        full_cmd = full_command[0]
        full_args = full_command[1:]
        full_command_str = unix_utils.format_command_list(full_cmd, full_args)
        where = " " + _T("EXECUTED_AT") % (self.windows_host_ip,)
        raise WinexeRunnerException(
            strip_multiline_spaces(_T("COMMAND_COMMAND_STRWHERE_RETURNED_NONZERO_EXIT")).format(
                command_str=command_str, where=where
            ),
            command_str=command_str, full_command_str=full_command_str,
            stderr=stderr, stdout=stdout, host_description=self.host_description, exit_code=exit_code,
        )

    @staticmethod
    def _get_command_str(cmd_str, args=None):
        if args is not None:
            cmd_str = windows_utils.format_command(cmd_str, **args)

        return cmd_str

    def _run_smbclient_cmd(self, cmd, share, timeout=None):
        logger.debug(_T("DEBUG_EXECUTE_SMB_COMMAND"), cmd)
        return run_and_check_local_command(
            [
                '/bin/sh', '-c',
                unix_utils.format_command(
                    u"smbclient //{ip}/{share} -A {auth_file} -c {command}",
                    ip=self.windows_host_ip, share=share, auth_file=self.auth_file, command=cmd
                )
            ],
            timeout=timeout
        )

    def _smbclient_upload_file(self, local_pathname, remote_pathname):
        share, pathname_cmd = self._split_samba_pathname(remote_pathname)

        # smbclient removes duplicate slashes and backslashes from paths it reports
        # we compare the paths it reports with paths we specified, to make sure the transfer went successfully
        # so we need to remove duplicate slaesh and backslashes too
        pathname_cmd = pathname_cmd.replace('//', '/').replace('\\\\', '\\')

        stdout, stderr = self._run_smbclient_cmd(
            u'put {local} "{remote}"'.format(local=local_pathname, remote=pathname_cmd), share
        )

        if r'putting file {local} as {remote} ('.format(local=local_pathname, remote=pathname_cmd) not in stderr:
            raise Exception(_T("FAILED_TO_UPLOAD_FILE_SMB").format(
                    local=local_pathname, remote=pathname_cmd, ip=self.settings.ip, share=share,
                    stdout=safe_string_repr(stdout), stderr=safe_string_repr(stderr)
            ))

    def _smbclient_download_file(self, remote_pathname, local_pathname):
        logger.debug(_T("DEBUG_DOWNLOADING_FILE"), remote_pathname, local_pathname)

        # remove duplicate slashes and backslashes, to successfully match smbclient's output with what we expect
        remote_pathname = remote_pathname.replace('//', '/').replace('\\\\', '\\')

        share, pathname_cmd = self._split_samba_pathname(remote_pathname)

        stdout, stderr = self._run_smbclient_cmd(
            u'get "{remote}" {local}'.format(local=local_pathname, remote=pathname_cmd), share
        )

        match = re.search(ur'getting file {remote} of size [0-9]+ as {local} \('.format(
            local=re.escape(local_pathname), remote=re.escape(pathname_cmd)), stderr, re.MULTILINE
        )
        if match is None:
            raise Exception(_T("FAILED_TO_DOWNLOAD_FILE_SMB").format(
                    local=local_pathname, remote=pathname_cmd, ip=self.settings.ip, share=share,
                    stdout=safe_string_repr(stdout), stderr=safe_string_repr(stderr)
            ))

    def _smbclient_download_directory_contents(self, remote_pathname, local_pathname):
        logger.debug(_T("LOG_DOWNLOAD_DIRECTORY"), remote_pathname, local_pathname)

        share, pathname_cmd = self._split_samba_pathname(remote_pathname)

        self._run_smbclient_cmd(
            u'prompt OFF;recurse ON;lcd {local};cd "{remote}";mget *'.format(local=local_pathname, remote=pathname_cmd),
            share
        )

    @staticmethod
    def _split_samba_pathname(pathname):
        to_cmd = lambda pathname: pathname.replace('/', '\\')
        if ':' in pathname:
            disk, pathname_part = pathname.split(':')
            share = u'%s$' % disk
            pathname_cmd = to_cmd(pathname_part)
        else:
            share = u'C$'
            pathname_cmd = to_cmd(pathname)
        # guarantee leading backslash
        pathname_cmd = u"\\%s" % pathname_cmd.lstrip('\\')

        return share, pathname_cmd
