# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple

from migrator.core.language import _T
from migrator.core import MigrationError
from migrator.core.session.config import SSHAuthPassword, PhysicalUnixServerConfig, is_windows_host
from migrator.core.utils.filetypes.config_utils import Auth
from migrator.core.utils.networking.ip_port import is_network_port_open
from migrator.core.utils.networking.run_command import SSHRunner, WinexeRunner, LocalUnixRunner
from migrator.core.utils.networking.ssh.utils import SSHClient
from migrator.core.utils.registry import Registry
from migrator.settings import SSH_DEFAULT_PORT, SMB_DEFAULT_PORT, UNIX_DEFAULT_SESSION_DIR


def get_runner(config_section):
    if config_section.get('ip') in ['localhost', '127.0.0.1']:
        return LocalUnixRunner()
    elif is_windows_host(config_section):
        return _get_windows_runner(config_section)
    else:
        return _get_ssh_runner(config_section)


def _get_ssh_runner(config_section):
    ip = config_section.get('ip')
    if ip is None:
        raise MigrationError(_T("IP_ADDRESS_IS_NOT_SPECIFIED").format(section_name=config_section.section_name))

    if not is_network_port_open(ip, 22):
        raise MigrationError(_T("UNABLE_CONNECT_HOST_BY_SSH_EXCEPTION").format(host=ip))

    ssh_auth = _get_ssh_auth(config_section)
    server_config = PhysicalUnixServerConfig(
        server_id='GLOBAL',
        ip=ip,
        session_dir=UNIX_DEFAULT_SESSION_DIR,
        ssh_auth=ssh_auth
    )

    client = SSHClient(ip)
    server_config.ssh_auth.connect(server_config.ip, client)
    return SSHRunner(client, "'Source Server %s'" % ip)


def _get_ssh_auth(config_section):
    port = int(config_section.get('ssh-port', SSH_DEFAULT_PORT))
    ssh_auth = SSHAuthPassword(
        port=port, username=config_section.get('username', 'root'), password=config_section.get('password')
    )

    return ssh_auth


def _get_windows_runner(config_section):
    WindowsSettings = namedtuple('WindowsSettings', 'ip windows_auth')
    auth = Auth(
        username=config_section.get('username'), password=config_section.get('password')
    )
    ip = config_section.get('ip')
    if ip is None:
        raise MigrationError(_T("IP_ADDRESS_IS_NOT_SPECIFIED").format(section_name=config_section.section_name))

    if not is_network_port_open(ip, SMB_DEFAULT_PORT):
        raise MigrationError(_T("WINEXE_FAILED_TO_OPEN_CONNECTION").format(host=ip))

    settings = WindowsSettings(ip=ip, windows_auth=auth)

    return WinexeRunner(
        settings,
        local_temp_dir=Registry.get_instance().get_context().migrator_server.session_dir(),
        host_description="'Source Server %s'" % ip
    )
