# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import re

from lxml import etree

from migrator.core.data.model.web_hosting_shm import ApplicationPoolModel

logger = logging.getLogger(__name__)


def get_application_pool_settings(server):
    def _get_appcmd_result(runner, cmd):
        path = runner.mktempfile(server.session_dir())
        runner.remove_file(path)  # IIsCnfg.vbs cannot overwrite existing file
        runner.sh(cmd.format(path=path))
        return etree.XML(runner.get_file_contents(path))

    win_application_pools = None
    win_sites = None
    if server.os_info().version >= '6.0':
        with server.runner() as runner:
            win_application_pools = _get_appcmd_result(
                runner,
                "cmd /c %SYSTEMROOT%\\System32\\inetsrv\\appcmd list apppool /config:* /xml > {path}"
            )
            win_sites = _get_appcmd_result(
                runner,
                "cmd /c %SYSTEMROOT%\\System32\\inetsrv\\appcmd list sites /config /xml > {path}"
            )
    else:
        logger.debug(
            "Skip exporting IIS Application Pool settings from '{host}', due to Windows 2003".format(
                host=server.description()
            )
        )

    return win_application_pools, win_sites


def get_application_pool(site_name, win_application_pools, win_sites, server):
    def _cast_time(time_str):  # convert time from format "1.05:00:00" to minutes
        m = re.search('(\d)?.?(\d\d):(\d\d):(\d\d)', time_str)
        multipliers = [1440, 60, 1, 0]
        return sum(map(lambda x: x[0] * int(x[1]), zip(multipliers, m.groups(0))))

    def _cast_zero(value):
        return value if int(value) else '-1'

    def _get_application_pool_name(site_name):
        node = win_sites.xpath(u'/appcmd/SITE[starts-with(@SITE.NAME,"{site_name}")]'.format(site_name=site_name))
        if not node:
            logger.debug(u'There are no Site config for site {site}'.format(site=site_name))
            return None

        assert (len(node) == 1)
        pool_name = node[0].xpath(u'site/application[@path="/"]/@applicationPool')
        logger.debug(u'Found Application pool {pool_name} for site {site_name}'.format(pool_name=pool_name, site_name=site_name))
        assert (len(pool_name) == 1)
        return pool_name[0]

    if win_application_pools is None:
        logger.debug('There are no Application Pools config for server {server}'.format(server=server))
        return None

    pool_name = _get_application_pool_name(site_name)
    if not pool_name:
        return None

    node = win_application_pools.xpath(u'/appcmd/APPPOOL[@APPPOOL.NAME="{pool_name}"]'.format(pool_name=pool_name))
    if not node:
        logger.debug(u'There are no Application Pools config for site {site}'.format(site=site_name))
        return None

    assert (len(node) == 1)

    node = node[0]
    params = {
        'managed_pipeline_mode':       {'xpath': 'add/@managedPipelineMode'},
        'max_worker_processes':        {'xpath': 'add/processModel/@maxProcesses'},
        'idle_timeout':                {'xpath': 'add/processModel/@idleTimeout', 'cast': _cast_time},
        'idle_timeout_action':         {'xpath': 'add/processModel/@idleTimeoutAction'},
        'cpu_usage':                   {'xpath': 'add/cpu/@limit', 'cast': lambda x: (int(x) if int(x) else 100000) / 1000},
        'cpu_usage_action':            {'xpath': 'add/cpu/@action'},
        'cpu_usage_interval':          {'xpath': 'add/cpu/@resetInterval', 'cast': _cast_time},
        'recycling_by_time':           {'xpath': 'add/recycling/periodicRestart/@time', 'cast': lambda x:_cast_zero(_cast_time(x))},
        'recycling_by_requests':       {'xpath': 'add/recycling/periodicRestart/@requests', 'cast': _cast_zero},
        'recycling_by_virtual_memory': {'xpath': 'add/recycling/periodicRestart/@memory', 'cast': _cast_zero},
        'recycling_by_private_memory': {'xpath': 'add/recycling/periodicRestart/@privateMemory', 'cast': _cast_zero}
    }
    for x in params:
        value = node.xpath(params[x]['xpath'])
        if value:
            assert(len(value) == 1)
            params[x] = params[x]['cast'](value[0]) if 'cast' in params[x] else value[0]
        else:
            logger.debug(u"Parameter '{param}' was not found with '{xpath}'".format(param=x, xpath=params[x]))
            params[x] = None

    return ApplicationPoolModel(**params)
