# (C)2017 Ingram Micro Inc. all Rights Reserved
import re

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.networking.run_command import SSHRunner, WindowsRunner, LocalUnixRunner
from migrator.core.utils.operating_system.unix_utils import \
    is_debian, is_ubuntu, is_centos, is_redhat, is_suse, read_release_file, read_lsb_release, is_cloudlinux, \
    is_virtuozzolinux
from migrator.settings import WINEXE_GET_SYSTEM_INFO_COMMAND_TIMEOUT


class OSInfoBase(object):
    def __init__(self, runner):
        self.runner = runner
        self._arch = self._get_arch()
        self._version = self._get_version()

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def version(self):
        return self._version

    @property
    def arch(self):
        return self._arch

    @property
    def platform(self):
        raise NotImplementedError()

    def _get_version(self):
        raise NotImplementedError()

    def _get_arch(self):
        raise NotImplementedError()


class LinuxInfo(OSInfoBase):
    @property
    def platform(self):
        return 'Linux'

    def _get_arch(self):
        return self.runner.sh('uname -m').strip()


class CentOSInfo(LinuxInfo):
    @property
    def name(self):
        return 'CentOS'

    def _get_version(self):
        redhat_release = read_release_file('/etc/redhat-release', self.runner)
        return re.search(r'\s([0-9.]*)\s', redhat_release).group(1)


class RedHatInfo(CentOSInfo):
    @property
    def name(self):
        return 'RHEL'


class DebianInfo(LinuxInfo):
    @property
    def name(self):
        return 'Debian'

    def _get_version(self):
        return read_release_file('/etc/debian_version', self.runner).strip()


class UbuntuInfo(LinuxInfo):
    @property
    def name(self):
        return 'Ubuntu'

    def _get_version(self):
        release_info = read_lsb_release(self.runner)
        return release_info.get('DISTRIB_RELEASE', 'Unknown')


class SUSEInfo(LinuxInfo):
    @property
    def name(self):
        return 'SUSE'

    def _get_version(self):
        suse_release = self.runner.sh(u'cat /etc/SuSE-release')
        version = re.search(r'VERSION\s=\s([^\s]*)', suse_release).group(1)
        patch_level = re.search(r'PATCHLEVEL\s=\s([^\s]*)', suse_release).group(1)
        return version + '.' + patch_level


class CloudLinuxInfo(CentOSInfo):
    @property
    def name(self):
        return 'CloudLinux'


class VirtuozzoLinuxInfo(LinuxInfo):
    @property
    def name(self):
        return 'VirtuozzoLinux'

    def _get_version(self):
        virtuozzo_release = read_release_file('/etc/virtuozzo-release', self.runner)
        return re.search(r'\s([0-9.]*)\s', virtuozzo_release).group(1)


class WindowsInfo(OSInfoBase):
    _WIN_INFO_SCRIPT = r'''
Set objWMIService = GetObject("winmgmts:{impersonationLevel=impersonate}!\\.\root\cimv2")
Set colOperatingSystems = objWMIService.ExecQuery("Select * from Win32_OperatingSystem")

For Each objOperatingSystem in colOperatingSystems
  wscript.echo "Name: " & objOperatingSystem.Caption
  wscript.echo "Version: " & objOperatingSystem.Version
Next

wscript.echo "Arch: x" & GetObject("winmgmts:root\cimv2:Win32_Processor='cpu0'").AddressWidth
    '''
    _WIN_INFO_SCRIPT_NAME = 'c:\\sysinfo.vbs'

    def __init__(self, runner):
        super(WindowsInfo, self).__init__(runner)

    @property
    def name(self):
        return self._sysinfo()['name']

    @property
    def platform(self):
        return 'Windows'

    def _get_version(self):
        return self._sysinfo()['version']

    def _get_arch(self):
        return self._sysinfo()['arch']

    @cached
    def _sysinfo(self):
        self.runner.upload_file_content(self._WIN_INFO_SCRIPT_NAME, self._WIN_INFO_SCRIPT)
        try:
            with self.runner.set_timeout(WINEXE_GET_SYSTEM_INFO_COMMAND_TIMEOUT):
                sysinfo_string = self.runner.sh('cscript /nologo %s' % self._WIN_INFO_SCRIPT_NAME)
        finally:
            self.runner.remove_file(self._WIN_INFO_SCRIPT_NAME)

        res = {
            'name': 'Unknown',
            'version': 'Unknown',
            'arch': 'Unknown',
        }
        for line in sysinfo_string.splitlines():
            if line.startswith('Name:'):
                res['name'] = line.strip().split(':')[1].strip()
            elif line.startswith('Version:'):
                res['version'] = line.strip().split(':')[1].strip()
            elif line.startswith('Arch:'):
                res['arch'] = line.strip().split(':')[1].strip()

        return res


def get_os_info(runner):
    """Factory to get system info object using the specified runner"""
    osinfo_dispatch_table = [
        # <expected runner class name>, <function that detects OS release>, <OSInfo class name>
        (SSHRunner, is_centos, CentOSInfo),
        (SSHRunner, is_redhat, RedHatInfo),
        (SSHRunner, is_debian, DebianInfo),
        (SSHRunner, is_ubuntu, UbuntuInfo),
        (SSHRunner, is_suse, SUSEInfo),
        # Keep VirtuozzoLinux before CloudLinux, because Virtuozzo can be based on CloudLinux
        (SSHRunner, is_virtuozzolinux, VirtuozzoLinuxInfo),
        (SSHRunner, is_cloudlinux, CloudLinuxInfo),
        (LocalUnixRunner, is_centos, CentOSInfo),
        (LocalUnixRunner, is_redhat, RedHatInfo),
        (WindowsRunner, lambda runner_obj: True, WindowsInfo),
    ]
    for runner_class, os_release_matcher, info_class in osinfo_dispatch_table:
        if isinstance(runner, runner_class) and os_release_matcher(runner):
            return info_class(runner)

    raise MigrationError(_T("UNABLE_TO_DETERMINE_TYPE_OS").format(host_description=runner.host_description))
