# (C)2017 Ingram Micro Inc. all Rights Reserved
import grp
import os
import pwd

from collections import namedtuple
from pipes import quote

from migrator.core import run_and_check_local_command
from migrator.core.language import _T

StatInfo = namedtuple('StatInfo', ('size', 'perm', 'user_id', 'user', 'group_id', 'group'))


def is_debian(runner):
    release_info = read_lsb_release(runner)
    if release_info is None:
        if runner.file_exists('/etc/debian_version'):
            return True

        return False

    if release_info.get('DISTRIB_ID', '').lower() != 'debian':
        return False

    return True


def is_ubuntu(runner):
    release_info = read_lsb_release(runner)
    if release_info is None or release_info.get('DISTRIB_ID', '').lower() != 'ubuntu':
        return False

    return True


def is_centos(runner):
    release_info = read_release_file('/etc/redhat-release', runner)
    if release_info is None or not release_info.startswith('CentOS'):
        return False

    return True


def is_redhat(runner):
    release_info = read_release_file('/etc/redhat-release', runner)
    if release_info is None or not release_info.startswith('Red'):
        return False

    return True


def is_suse(runner):
    if not runner.file_exists('/etc/SuSE-release'):
        return False

    return True


def is_cloudlinux(runner):
    release_info = read_release_file('/etc/redhat-release', runner)
    if release_info is None or not release_info.startswith('CloudLinux'):
        return False

    return True


def is_virtuozzolinux(runner):
    release_info = read_release_file('/etc/virtuozzo-release', runner)
    if release_info is None or not release_info.startswith('Virtuozzo'):
        return False

    return True


def format_command(template, *args, **kw):
    qargs = map(quote, args)
    qkw = dict((k, quote(unicode(v))) for k, v in kw.iteritems())
    return unicode(template).format(*qargs, **qkw)


def format_command_list(cmd, args):
    return " ".join([
        quote(arg) for arg in [cmd] + ([] if args is None else args)
    ])


def file_exists(runner, filepath):
    pos = filepath.find('*')
    exit_code, _, _ = runner.run_unchecked('/usr/bin/test', ['-e', filepath[:pos if pos != -1 else None]])
    return exit_code == 0


def get_files_list(runner, path):
    files = runner.sh('find {path} -maxdepth 1 -type f -printf "%f\\0"'.format(path=path)).split("\0")
    return [item for item in files if item]


def replace_path_prefix(string, old_prefix, new_prefix):
    if string.startswith(old_prefix):
        return new_prefix + string[len(old_prefix):]
    else:
        raise Exception(_T("STRING_EXPECTED_STARTS_WITH") % (string, old_prefix))


def join_package_path(package, *path_parts):
    """Return an absolute path starting from package directory, with appended path parts."""
    dirs = [p for p in package.__path__]
    assert all(d == dirs[0] for d in dirs)
    return os.path.join(dirs[0], *path_parts)


def get_or_add_sysgroup(group):
    try:
        return grp.getgrnam(group)
    except KeyError:
        run_and_check_local_command(['groupadd', '-f', group])
        return grp.getgrnam(group)


def get_or_add_sysuser(user, home_dir, group, shell='/sbin/nologin'):
    try:
        return pwd.getpwnam(user)
    except KeyError:
        run_and_check_local_command(
            ['useradd', '--gid', group, '--create-home', '--home', home_dir, '--shell', shell, user]
        )
        return pwd.getpwnam(user)


def read_release_file(filename, runner):
    if not runner.file_exists(filename):
        return None

    return runner.get_file_contents(filename)


def read_lsb_release(runner):
    fname = '/etc/lsb-release'

    data = read_release_file(fname, runner)
    if data is None:
        return None

    release_info = {}
    for line in data.splitlines():
        tmp = line.strip().split('=', 1)
        if len(tmp) != 2:
            continue

        release_info[tmp[0].strip()] = tmp[1].strip()

    return release_info


def path_stat(runner, path):
    out = runner.run('/usr/bin/stat', ['--format=%s %a %u %U %g %G', path])
    size, perm, user_id, user, group_id, group = out.split()
    return StatInfo(size, '0'+perm, user_id, user, group_id, group)


def getent(runner, db, objects=None):
    args = [db]
    if objects:
        args += objects

    out = runner.run('/usr/bin/getent', args)
    return out.splitlines()
