# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import ntpath
import re

from migrator.core.language import _T

logger = logging.getLogger(__name__)


class WindowsNationalSymbolsCommandError(Exception):
    def __init__(self, command):
        # We should never try to run such commands. There are always problems with Windows + national symbols.
        self.command = command
        message = (
            _T("INTERNAL_ERROR_TRYING_EXECUTE_COMMAND_THAT") % (command,)
        )
        Exception.__init__(self, message)


def check_windows_national_symbols_command(command):
    try:
        command.decode('ascii')
    except UnicodeError:
        logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
        raise WindowsNationalSymbolsCommandError(command)


def quote_arg(arg):
    return '"%s"' % arg.replace('"', '"""')


def format_command(template, *args, **kw):
    qargs = map(quote_arg, args)
    qkw = dict((k, quote_arg(unicode(v))) for k, v in kw.iteritems())
    return unicode(template).format(*qargs, **qkw)


def format_command_list(cmd, args):
    return " ".join([
        quote_arg(arg) for arg in [cmd] + ([] if args is None else args)
    ])


def convert_path_to_cygwin(path):
    drive, rel_path = ntpath.splitdrive(path)
    if drive != '':
        drive = drive[0].lower()
        rel_path = force_forward_slash(rel_path[1:])
        return u"/cygdrive/%s/%s" % (drive, rel_path)
    else:
        return force_forward_slash(path)


def convert_path_to_cygwin_root_folder_of_partition(path):
    drive, rel_path = ntpath.splitdrive(path)
    return u"/cygdrive/%s" % drive[0].lower() if drive != '' else force_forward_slash(path)


def convert_path_to_cygwin_vhosts_folder(path):
    """
    :param path: string Example: 'C:\oam-session\V87YPn0x7y'
    :return: string Example: 'vhosts/oam-session/V87YPn0x7y'
    """
    drive, rel_path = ntpath.splitdrive(path)
    if drive != '':
        rel_path = force_forward_slash(rel_path[1:])
        return u"vhosts/%s" % rel_path
    else:
        return force_forward_slash(path)


def force_forward_slash(path):
    return path.replace("\\", "/")


def file_exists(runner, filepath):
    cmd = u"cmd.exe /C if exist {filepath} ( echo true ) else ( echo false )"
    args = dict(filepath=filepath)
    result = runner.sh(cmd, args).strip()
    if result == 'true':
        return True
    elif result == 'false':
        return False
    else:
        raise Exception(
            _T("FILE_EXISTS_INVALID_COMMAND_OUTPUT").format(
                command=format_command(cmd, **args),
                output=result
            )
        )


def path_join(base_path, *paths):
    """Join the paths using backslash as a separator, and return the result.
    Take care of the excess separators between the joined path chunks.

    This function behaves similar to ntpath.join
    path_join('\\\\sambashare\\dir1\\', 'file1') returns the same '\\\\sambashare\\dir1\\file1'
    as ntpath.join('\\\\sambashare\\dir1\\', 'file1')
    but does not allow a subsequent path to override all the paths before:
    path_join('\\dir1', '\\file1') returns '\\dir1\\file1',
    whereas ntpath.join('\\dir1', '\\file1') returns '\\file1'.
    It also takes more care of the excess separators than ntpath.join does:
    path_join('\\dir1\\\\dir2\\\\', '\\file1') returns '\\dir1\\\\dir2\\file1',
    whereas ntpath.join('\\dir1\\\\dir2\\\\', '\\file1') returns '\\dir1\\\\dir2\\\\file1'.
    """
    paths = filter(None, paths)
    if len(paths) == 0:
        return base_path

    sep_to_strip = r'\/'
    sep_to_add = '\\'

    joined_path = base_path.rstrip(sep_to_strip)

    for path in paths[:-1]:
        joined_path += sep_to_add + path.strip(sep_to_strip)
    joined_path += sep_to_add + paths[-1].lstrip(sep_to_strip)

    return joined_path


def cmd_command(command):
    """Format command to be executed under cmd.exe"""
    return 'cmd /c "%s"' % (command,)


def fix_path(path):
    """Fix incorrect and double slashes in Windows path"""
    return path.replace('//', '/').replace('/', '\\').replace('\\\\', '\\').lower()


def convert_path(subscription, webspace, site, old_root, new_root, source_path):
    """Convert path from source to target
    Paths are normalized before comparison and replacement.
    Return normalized result path.
    """
    def normalize_path(p):
        return p.replace('/', '\\').rstrip('\\')

    source_path = normalize_path(source_path)
    old_root = normalize_path(old_root)
    new_root = normalize_path(new_root)
    if source_path.lower().startswith(old_root.lower()):
        # if we don't strip double backslash, error documents of File type will not work after transfer.
        return path_join(new_root, source_path[len(old_root)+1:]).replace('\\\\', '\\')
    else:
        return None


def get_from_registry(runner, keys, value, recursively=False, bits=None):
    """Get value from Windows registry

    :type keys: list[str]
    :type value: str
    :rtype: str
    """
    for key in keys:
        command = 'reg query "{key}" {flags} /v "{value}"'.format(
            key=key, value=value,
            flags=' '.join(['/s' if recursively else '', '/reg:%s' % bits if bits else ''])
        )
        (code, stdout, _) = runner.sh_unchecked(command)
        if code is 0:
            match = re.search('^\s*%s\s+REG_(EXPAND_)?SZ\s+(.*?)\s*$' % re.escape(value), stdout, re.MULTILINE)
            if match is not None:
                return match.group(2).strip()
    return None


def find_in_registry(runner, keys, pattern):
    """
    Return true if key with given pattern exist below one of given keys in registry, false otherwise
    :type runner: parallels.core.run_command.BaseRunner
    :type keys: list[str]
    :type pattern: str
    :rtype: bool
    """
    for key in keys:
        (code, stdout, _) = runner.sh_unchecked('reg query "%s"' % key)
        if code is 0:
            match = re.search('^%s%s$' % (re.escape(key + '\\'), pattern), stdout, re.MULTILINE)
            if match is not None:
                return True
    return False

