# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from contextlib import contextmanager

from migrator.core import MigrationError, MigrationNoContextError, ParallelExecutionError
from migrator.core.language import _T
from migrator.core.session.config import MultithreadingParams, MultithreadingStatus
from migrator.core.utils.concurrency.deamon import demonize
from migrator.core.utils.lang.profiler import get_default_steps_profiler
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


class Runner(object):
    def __init__(self, multithreading=None):
        """Class constructor
        :type multithreading: parallels.core.migrator_config.MultithreadingParams | None
        """
        self._context = Registry.get_instance().get_context()
        if multithreading is None:
            self._multithreading = MultithreadingParams(status=MultithreadingStatus.DISABLED, num_workers=None)
        else:
            self._multithreading = multithreading

    def run_entry_point(self, entry_point_name):
        """Run specified entry point and execute shutdown actions
        :type entry_point_name: str
        :rtype: None
        """
        registry = Registry.get_instance()
        entry_point = registry.get_workflow().get_entry_point(entry_point_name)

        if self._context.options.async:
            demonize()

        parallel_execution = False
        try:
            self._context.progress.start()
            self.run(entry_point)
        except ParallelExecutionError as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            parallel_execution = True
            raise e
        except MigrationError as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            registry.get_context().execution_wrapper.register_error(e.message, '')
            raise e
        except MigrationNoContextError as e:
            raise e
        except Exception as e:
            registry.get_context().execution_wrapper.register_error(_T("INTERNAL_ERROR_OCCURRED_DURING_MIGRATION"), '')
            raise e
        finally:
            if not parallel_execution:
                self._context.progress.stop()
            self._shutdown(entry_point)

    def run(self, action, action_id=None):
        raise NotImplementedError()

    def _shutdown(self, entry_point):
        raise NotImplementedError()

    @staticmethod
    @contextmanager
    def _profile_action(action_id, action, profiling_enabled=True):
        """Measure time of specified action

        This function considers that for all parent actions this function was already called, and we're
        already in that context.

        :type action: parallels.core.actions.base.base_action.BaseAction
        :type action_id: basestring
        :type profiling_enabled: boolean
        """
        if action.get_description() is not None and profiling_enabled:
            with get_default_steps_profiler().measure_time(action_id, action.get_description()):
                yield
        else:
            # skip profiling in one of following cases:
            # - compound actions that have no descriptions are profiled by all child actions
            # - other actions that have no descriptions are measured as "Other" action
            # - profiling was explicitly disabled
            yield
