# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.actions.import_objects import ImportObjects
from migrator.core.actions.reset_migration_progress import ResetMigrationProgress
from migrator.products.plesk.source.plesk.actions.suspend_subscriptions import SuspendSubscriptions
from migrator.core.language import _T
from migrator.core.actions.base.compound_action import CompoundAction
from migrator.core.actions.base.entry_point_action import EntryPointAction
from migrator.core.actions.base.legacy_action import LegacyAction
from migrator.core.actions.close_ssh_connections import CloseSSHConnections
from migrator.core.actions.list_ip_addresses import ListIPAddresses
from migrator.core.actions.migration_list.check_migration_list_exists_for_migration import \
    CheckMigrationListExistsForMigration
from migrator.core.actions.migration_list.check_migration_list_not_exists_for_generation import \
    CheckMigrationListNotExistsForGeneration
from migrator.core.actions.migration_list.generate import GenerateMigrationList
from migrator.core.actions.quick_check_cli_options import QuickCheckCLIOptions
from migrator.core.actions.remove_ssh_keys import RemoveSSHKeys
from migrator.core.actions.shutdown_windows_rsync_servers import ShutdownWindowsRsyncServers
from migrator.core.actions.uninstall_dump_agent import UninstallDumpAgentAction
from migrator.core.actions.utils.logging_properties import LoggingProperties
from migrator.core.workflow.extension import WorkflowExtensionBase

logger = logging.getLogger(__name__)


class SharedHostingWorkflowExtension(WorkflowExtensionBase):
    """Common workflow for shared hosting migration"""

    def extend_workflow(self, workflow):
        self._configure_shared_actions(workflow)
        self._configure_entry_points(workflow)

    def _configure_shared_actions(self, workflow):
        self._configure_shared_action_fetch_source(workflow)
        self._configure_shared_action_fetch_source_shallow(workflow)
        self._configure_shared_action_fetch_target(workflow)
        self._configure_shared_action_sync_web_content_assets(workflow)
        self._configure_shared_action_copy_web_files(workflow)
        self._configure_shared_action_copy_web_content(workflow)
        self._configure_shared_action_copy_mail_content(workflow)
        self._configure_shared_action_copy_db_content(workflow)
        self._configure_shared_action_disable_source(workflow)
        self._configure_shared_action_setup_dns_forwarding(workflow)
        self._configure_shared_action_cleanup(workflow)
        self._configure_shared_action_copy_content(workflow)
        self._configure_shared_action_check_connections(workflow)
        self._configure_shared_action_check_sources(workflow)
        self._configure_shared_action_read_migration_list(workflow)
        self._configure_shared_action_verify_hosting(workflow)
        self._configure_shared_action_finalize(workflow)
        self._configure_shared_action_initial_pre_checks(workflow)
        self._configure_shared_action_post_migration_finalize(workflow)
        self._configure_shared_action_post_migration_prepare(workflow)

    def _configure_entry_points(self, workflow):
        self._configure_entry_point_generate_migration_list(workflow)
        self._configure_entry_point_transfer_accounts(workflow)
        self._configure_entry_point_transfer_all(workflow)
        self._configure_entry_point_copy_content(workflow)
        self._configure_entry_point_list_ip_addresses(workflow)

    @staticmethod
    def _configure_shared_action_fetch_source(workflow):
        workflow.add_shared_action('fetch-source', CompoundAction(
            description=_T("FETCH_INFORMATION_FROM_SOURCE_PANEL")))

    @staticmethod
    def _configure_shared_action_fetch_source_shallow(workflow):
        # By default, consider shallow backup is the same as full backup,
        # and fetch source shallow is the same as fetch source
        workflow.add_shared_action(
            'fetch-source-shallow',
            workflow.get_shared_action('fetch-source')
        )

    @staticmethod
    def _configure_shared_action_fetch_target(workflow):
        workflow.add_shared_action('fetch-target', LegacyAction(
            description=_T("FETCH_INFORMATION_FROM_TARGET_SERVERS"),
            function=lambda ctx: ctx.migrator._fetch(ctx.options)
        ))

    @staticmethod
    def _configure_shared_action_copy_content(workflow):
        workflow.add_shared_action('copy-content', CompoundAction(
            description=_T("COPY_WEB_DATABASE_AND_MAIL_CONTENT"),
            run_by_subscription=True
        ))
        workflow.get_shared_action('copy-content').insert_action(
            'web', workflow.get_shared_action_pointer('copy-web-content')
        )
        workflow.get_shared_action('copy-content').insert_action(
            'mail', workflow.get_shared_action_pointer('copy-mail-content')
        )
        workflow.get_shared_action('copy-content').insert_action(
            'db', workflow.get_shared_action_pointer('copy-db-content')
        )

    @staticmethod
    def _configure_shared_action_copy_web_files(workflow):
        workflow.add_shared_action('copy-web-files', CompoundAction(
            description=_T("ACTION_COPY_WEB_FILES")
        ))

    @staticmethod
    def _configure_shared_action_copy_web_content(workflow):
        workflow.add_shared_action('copy-web-content', CompoundAction(
            description=_T("COPY_WEB_FILES_AND_SYNC_WEB")))
        workflow.get_shared_action('copy-web-content').insert_action(
            'copy-web-files', workflow.get_shared_action('copy-web-files')
        )
        workflow.get_shared_action('copy-web-content').insert_action(
            'web-assets', workflow.get_shared_action('sync-web-content-assets')
        )

    @staticmethod
    def _configure_shared_action_sync_web_content_assets(workflow):
        workflow.add_shared_action('sync-web-content-assets', CompoundAction(
            description=_T("ACTION_SYNC_WEB_CONTENT_ASSESTS")
        ))

    @staticmethod
    def _configure_shared_action_copy_mail_content(workflow):
        workflow.add_shared_action('copy-mail-content', CompoundAction())

    @staticmethod
    def _configure_shared_action_copy_db_content(workflow):
        workflow.add_shared_action('copy-db-content', CompoundAction())

    @staticmethod
    def _configure_shared_action_disable_source(workflow):
        workflow.add_shared_action('disable-source', CompoundAction())

    @staticmethod
    def _configure_shared_action_setup_dns_forwarding(workflow):
        workflow.add_shared_action('setup-dns-forwarding', CompoundAction())

    @staticmethod
    def _configure_shared_action_cleanup(workflow):
        workflow.add_shared_action('cleanup', CompoundAction(
            description='Perform final clean up',
            logging_properties=LoggingProperties(compound=False)
        ))
        workflow.get_shared_action('cleanup').insert_action(
            'uninstall-agent', UninstallDumpAgentAction()
        )
        workflow.get_shared_action('cleanup').insert_action(
            'remove-ssh-keys', RemoveSSHKeys()
        )
        workflow.get_shared_action('cleanup').insert_action(
            'close-ssh-connections', CloseSSHConnections()
        )
        workflow.get_shared_action('cleanup').insert_action(
            'shutdown-windows-rsync-servers', ShutdownWindowsRsyncServers()
        )

    @staticmethod
    def _configure_shared_action_check_connections(workflow):
        workflow.add_shared_action('check-connections', LegacyAction(
            description=_T("ACTION_CHECK_CONNECTIONS"),
            function=lambda ctx: ctx.migrator._check_connections(ctx.options)
        ))

    @staticmethod
    def _configure_shared_action_check_sources(workflow):
        workflow.add_shared_action('check-sources', LegacyAction(
            description=_T("ACTION_CHECK_SOURCE_CONNECTIONS"),
            function=lambda ctx: ctx.migrator._check_sources()
        ))

    @staticmethod
    def _configure_shared_action_read_migration_list(workflow):
        workflow.add_shared_action('read-migration-list', LegacyAction(
            description=_T("ACTION_READ_MIGRATION_LIST"),
            function=lambda ctx: ctx.migrator.read_migration_list_lazy(ctx.options)
        ))

    @staticmethod
    def _configure_shared_action_verify_hosting(workflow):
        workflow.add_shared_action('verify-hosting', CompoundAction(
            description=_T("ACTION_VERIFY_HOSTING")
        ))

    @staticmethod
    def _configure_shared_action_finalize(workflow):
        workflow.add_shared_action('finalize', LegacyAction(
            description=_T("COUNT_MIGRATION_STATISTICS_PRINT_REPORT"),
            function=lambda ctx: ctx.migrator._finalize(True, ctx.options),
            logging_properties=LoggingProperties(info_log=False),
            is_critical=True,
        ))

    @staticmethod
    def _configure_shared_action_post_migration_prepare(workflow):
        workflow.add_shared_action('post-migration-prepare', CompoundAction())
        workflow.get_shared_action('post-migration-prepare').insert_action(
            'check-migration-list', CheckMigrationListExistsForMigration()
        )
        workflow.get_shared_action('post-migration-prepare').insert_actions(
            workflow.get_shared_action('initial-pre-checks').get_all_actions()
        )
        workflow.get_shared_action('post-migration-prepare').insert_action(
            'fetch-source-shallow',
            workflow.get_shared_action_pointer('fetch-source-shallow')
        )
        workflow.get_shared_action('post-migration-prepare').insert_action(
            'read-migration-list',
            workflow.get_shared_action_pointer('read-migration-list')
        )
        workflow.get_shared_action('post-migration-prepare').insert_action(
            'fetch-source',
            workflow.get_shared_action('fetch-source')
        )
        workflow.get_shared_action('post-migration-prepare').insert_action(
            'fetch-target',
            workflow.get_shared_action_pointer('fetch-target')
        )
        workflow.get_shared_action('post-migration-prepare').insert_action(
            'read-migration-list',
            workflow.get_shared_action('read-migration-list')
        )

    @staticmethod
    def _configure_shared_action_post_migration_finalize(workflow):
        workflow.add_shared_action('post-migration-finalize', CompoundAction())
        workflow.get_shared_action('post-migration-finalize').insert_action(
            'cleanup', workflow.get_shared_action('cleanup')
        )
        workflow.get_shared_action('post-migration-finalize').insert_action(
            'finalize',
            workflow.get_shared_action('finalize')
        )

    @staticmethod
    def _configure_shared_action_initial_pre_checks(workflow):
        workflow.add_shared_action('initial-pre-checks', CompoundAction())
        workflow.get_shared_action('initial-pre-checks').insert_action(
            'check-cli-options', QuickCheckCLIOptions()
        )
        workflow.get_shared_action('initial-pre-checks').insert_action(
            'check-connections',
            workflow.get_shared_action_pointer('check-connections')
        )

    @staticmethod
    def _configure_entry_point_generate_migration_list(workflow):
        workflow.add_entry_point('generate-migration-list', EntryPointAction())

        workflow.get_path('generate-migration-list').insert_action(
            'check-migration-list-exists', CheckMigrationListNotExistsForGeneration()
        )
        workflow.get_path('generate-migration-list').insert_action(
            'generate-migration-list', GenerateMigrationList(),
        )
        workflow.get_path('generate-migration-list').insert_action(
            'cleanup', workflow.get_shared_action('cleanup')
        )
        workflow.get_path('generate-migration-list').register_overall_shutdown('cleanup')

    @staticmethod
    def _configure_entry_point_transfer_accounts(workflow):
        workflow.add_entry_point('transfer-accounts', EntryPointAction())
        workflow.get_path('transfer-accounts').insert_action(
            'check-cli-options', QuickCheckCLIOptions()
        )
        workflow.get_path('transfer-accounts').insert_action(
            'check-migration-list', CheckMigrationListExistsForMigration()
        )
        workflow.get_path('transfer-accounts').insert_action(
            'check-connections',
            workflow.get_shared_action_pointer('check-connections')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'fetch-source-shallow',
            workflow.get_shared_action_pointer('fetch-source-shallow')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'read-migration-list',
            workflow.get_shared_action_pointer('read-migration-list')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'fetch-source', workflow.get_shared_action('fetch-source')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'fetch-target',
            workflow.get_shared_action_pointer('fetch-target')
        )
        workflow.get_path('transfer-accounts').insert_action('suspend-source-subscription', SuspendSubscriptions())
        workflow.get_path('transfer-accounts').insert_action(
            'restore', CompoundAction("Restore")
        )
        workflow.get_path('transfer-accounts').insert_action('set-windows-security-policy', LegacyAction(
            description=_T("SET_SECURITY_POLICY_WINDOWS_NODES"),
            function=lambda ctx: ctx.migrator.set_security_policy()
        ))
        workflow.get_path('transfer-accounts').insert_action(
            'restore-hosting', CompoundAction("Restore hosting settings")
        )
        workflow.get_path('transfer-accounts/restore-hosting').insert_action(
            'restore-sysuser-logins', CompoundAction(
                description=_T("RESTORE_LOGINS_SYSTEM_USERS"),
                run_by_subscription=False
            )
        )
        workflow.get_path('transfer-accounts/restore-hosting').insert_action(
            'windows-refresh-components', LegacyAction(
                description=_T("REFRESH_COMPONENTS_LIST_TARGET_WINDOWS_SERVERS"),
                function=lambda ctx: ctx.migrator._refresh_service_node_components_for_windows(),
                logging_properties=LoggingProperties(compound=False),
                is_critical=False,
            )
        )
        workflow.get_path('transfer-accounts').insert_action(
            'verify-hosting', workflow.get_shared_action_pointer('verify-hosting')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'copy-content', workflow.get_shared_action_pointer('copy-content')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'restore-windows-security-policy', LegacyAction(
                description=_T("RESTORE_SECURITY_POLICY_WINDOWS_NODES"),
                function=lambda ctx: ctx.migrator.restore_security_policy(),
                logging_properties=LoggingProperties(compound=False),
                is_critical=False,
            )
        )
        workflow.get_path('transfer-accounts').insert_action(
            'cleanup', workflow.get_shared_action('cleanup')
        )
        workflow.get_path('transfer-accounts').insert_action(
            'finalize',
            workflow.get_shared_action('finalize')
        )

        workflow.get_path('transfer-accounts').register_shutdown(
            'set-windows-security-policy',
            'restore-windows-security-policy'
        )
        workflow.get_path('transfer-accounts').register_overall_shutdown('cleanup')

    @staticmethod
    def _configure_entry_point_transfer_all(workflow):
        workflow.add_entry_point('migrate', EntryPointAction())
        workflow.get_path('migrate').insert_action(
            'check-cli-options', QuickCheckCLIOptions()
        )
        workflow.get_path('migrate').insert_action(
            'fetch-source-shallow',
            workflow.get_shared_action_pointer('fetch-source-shallow')
        )
        workflow.get_path('migrate').insert_action(
            'read-migration-list',
            workflow.get_shared_action_pointer('read-migration-list')
        )
        workflow.get_path('migrate').insert_action(
            'reset-migration-progress', ResetMigrationProgress(), after='read-migration-list'
        )
        workflow.get_path('migrate').insert_action(
            'fetch-source', workflow.get_shared_action('fetch-source')
        )
        workflow.get_path('migrate').insert_action(
            'import-accounts-and-subscriptions', ImportObjects()
        )
        workflow.get_path('migrate').insert_action(
            'restore-hosting', CompoundAction("Restore hosting settings")
        )
        workflow.get_path('migrate').insert_action(
            'restore-mail-hosting', CompoundAction("Restore mail hosting settings")
        )
        workflow.get_path('migrate').insert_action(
            'copy-content', CompoundAction("Copy content")
        )
        workflow.get_path('migrate/copy-content').insert_action(
            'copy-web-content',
            workflow.get_shared_action_pointer('copy-web-content')
        )
        workflow.get_path('migrate/copy-content').insert_action(
            'copy-mail-content',
            workflow.get_shared_action_pointer('copy-mail-content')
        )
        workflow.get_path('migrate/copy-content').insert_action(
            'copy-db-content',
            workflow.get_shared_action_pointer('copy-db-content')
        )
        workflow.get_path('migrate').insert_action(
            'switch-to-production', CompoundAction("Switch to production mode")
        )
        workflow.get_path('migrate/switch-to-production').insert_action(
            'disable-source',
            workflow.get_shared_action_pointer('disable-source')
        )
        workflow.get_path('migrate/switch-to-production').insert_action(
            'setup-dns-forwarding', workflow.get_shared_action('setup-dns-forwarding')
        )
        workflow.get_path('migrate').insert_action(
            'cleanup', workflow.get_shared_action('cleanup')
        )
        workflow.get_path('migrate').insert_action(
            'finalize',
            workflow.get_shared_action('finalize')
        )
        workflow.get_path('migrate').register_overall_shutdown('cleanup')

    @staticmethod
    def _configure_entry_point_copy_content(workflow):
        workflow.add_entry_point('copy-content', EntryPointAction())
        workflow.get_path('copy-content').insert_actions(
            workflow.get_shared_action('post-migration-prepare').get_all_actions()
        )
        workflow.get_path('copy-content').insert_action(
            'copy-content',
            workflow.get_shared_action_pointer('copy-content')
        )
        workflow.get_path('copy-content').insert_actions(
            workflow.get_shared_action('post-migration-finalize').get_all_actions()
        )
        workflow.get_path('copy-content').register_overall_shutdown('cleanup')

    @staticmethod
    def _configure_entry_point_list_ip_addresses(workflow):
        workflow.add_entry_point('list-ip-addresses', EntryPointAction())
        workflow.get_path('list-ip-addresses').insert_action(
            'list-ip-addresses', ListIPAddresses()
        )
