# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple

from migrator.settings import OA_TARGET_OWNER_ID
from migrator.odin_automation.billing.lib.api.account import Account
from migrator.odin_automation.billing.lib.api.domain import Domain
from migrator.odin_automation.billing.lib.api.order import Order
from migrator.odin_automation.billing.lib.api.subscription import Subscription
from migrator.odin_automation.billing.lib.api.bm import BMAPI


CreditCard = namedtuple('CreditCard', (
    # 1) Owner customer
    'customer_id',

    # 2) Credit card information
    'pay_system',  # Str(30), one of PaySystem value
    'card_number',  # Str(20)
    'card_holder_name',  # Str(40)
    'cvv',  # Str(4)
    'expiration_date',  # Str(5), format: "DD/YY"
    'start_date',  # Str(5), applicable for Maestro cards only, format: "DD/YY"
    'issue_number',  # Str(4), applicable for Maestro cards only

    # 3) Contact information
    'address1',  # Str(80)
    'address2',  # Str(80)
    'city',  # Str(40)
    'state',  # Str(80)
    'zip_code',  # Str(10)
    'country',  # Str(2), 2-symbol identifier, in lower case, for example 'us' or 'ru'
    'email',  # Str(100)
    'phone',  # instance of PhoneNumber
    'fax',  # instance of PhoneNumber

    # 4) Billing configuration
    'use_for_auto_payments',  # Int
))

DomainSubscriptionNameServers = namedtuple('DomainSubscriptionNameServers', (
    'primary',  # Str(40)
    'secondary',  # Str(40)
    'third',  # Str(40)
    'fourth'  # Str(40)
))


class BillingAPI(object):
    def __init__(self, url, user, password):
        self.url = url
        self.user = user
        self.password = password
        self.api = BMAPI(url, user, password)
        self.account = Account(self.api, self.get_api)
        self.order = Order(self.api)
        self.subscription = Subscription(self.api)
        self.domain = Domain(self.api)

    def get_api(self, hidelist_request=None, hidelist_response=None):
        return BMAPI(self.url, self.user, self.password, hidelist_request, hidelist_response)

    def begin_transaction(self):
        return self.api.BeginTransaction()

    def commit_transaction(self):
        return self.api.CommitTransaction()

    def rollback_transaction(self):
        return self.api.RollbackTransaction()

    def import_credit_card(self, account_id, credit_card):
        hidelist_request = [
            # card_number
            "/methodCall/params/param/value/struct/member/name[text()='Params']/../value/array/data/value[3]/string",
            # cvv
            "/methodCall/params/param/value/struct/member/name[text()='Params']/../value/array/data/value[6]/string"
        ]
        api = self.get_api(hidelist_request=hidelist_request)
        api.BM.CreditCard2AccountAdd_API(
            account_id, credit_card.credit_card_info.card_holder_name, credit_card.credit_card_info.card_number,
            credit_card.credit_card_info.payment_system, credit_card.credit_card_info.expiration_date,
            credit_card.credit_card_info.cvv, credit_card.credit_card_info.start_date,
            credit_card.credit_card_info.issue_number, credit_card.email, credit_card.address.address1,
            credit_card.address.address2, credit_card.address.city, credit_card.address.state,
            credit_card.address.zipcode, credit_card.address.country, credit_card.phone.country_code,
            credit_card.phone.area_code, credit_card.phone.phone_num, credit_card.phone.ext_num,
            credit_card.fax.country_code, credit_card.fax.area_code, credit_card.fax.phone_num, credit_card.fax.ext_num,
            credit_card.use_for_auto_payments
        )

    def import_bank_account(self, account_id, bank_account):
        self.api.BM.BankAccountAdd_API(
            account_id, bank_account.payment_system, bank_account.bank_number, bank_account.account_number,
            bank_account.holder_name, '', bank_account.use_for_recurring
        )

    def get_payment_methods(self):
        return self.api.BM.PayToolTypeListForAccountGet_API(OA_TARGET_OWNER_ID, '', 0)

    def get_pay_tools_for_account(self, account_id, filter_payment_system_type=None):
        """
        If filter_payment_system_type is None, then a complete list of payment tools is returned.
        Otherwise, only the following payment tools of the filter_payment_system_type type will be presented in the
        result:
            0 - credit card
            1 - bank account

        :param int account_id:
        :param int filter_payment_system_type: the value of class PaymentSystemType constant
        :return: list[migrator.odin_automation.billing.lib.api.bm.PayToolListForAccountResult]
        """
        result = self.api.BM.PayToolListForAccountGet_API(account_id, 1)
        if filter_payment_system_type is None:
            return result

        return [x for x in result if x.pay_tool_type == filter_payment_system_type]

    def plan_period_list_get(self, plan_id):
        return self.api.BM.PlanPeriodListGet_API(plan_id, 1)

    def plan_details_get(self, plan_id):
        return self.api.BM.PlanDetailsGet_API(plan_id)

    def is_plan_delegated(self, plan_id):
        return self.api.BM.IsPlanDelegated(plan_id)

    def get_plan_category_list(self, account_id):
        return self.api.BM.PlanCategoryListGet_API(account_id, 1)

    def get_available_plan_list(self, account_id, plan_category_id):
        return self.api.BM.PlanListAvailableGet_API(account_id, '', plan_category_id, 1)

    def get_all_plan_list(self, account_id):
        return self.api.BM.GetOwnersPlansWithCategory(account_id, 1)

    def get_plan_list_by_service_template(self, st_id):
        return self.api.BM.PlanGetServiceTemplateList(st_id, 1)

    def get_version(self):
        return self.api.BM.GetVersion_API()

    def get_cert_product(self, plan_id):
        plan = self.api.BM.PlanGet(plan_id)
        cert_st = self.api.CERTGATE.CertSTGet(plan.service_template_id)
        cert_product = self.api.CERTGATE.CertProductGet(cert_st.cert_product_id)
        return cert_product

    def get_domain_zone_id(self, st_id):
        return self.api.DOMAINGATE.GetDomainZoneID_API(st_id)

    def get_pem_ns_settings(self):
        return self.api.PEMGATE.PEMNSSettingsGet()
