# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple

from bm import BillingAPIError
from migrator.core import _T, MigrationNoRepeatError
from migrator.core.utils import obj
from migrator.core.utils.networking.dns import convert_to_national_domain_name, convert_to_national_subscription_name
from migrator.odin_automation.billing.lib.api.bm import INT_NULL
from migrator.odin_automation.billing.lib.utils import get_service_plan_tld
from migrator.odin_automation.platform.lib.data.domain_importer import DomainImporter


SubscriptionOrder = namedtuple('SubscriptionOrder', (
    'VendorAccountID',  # Int
    'PlanID',  # Str(40)
    'PlanPeriodID',  # Int 10 "Ordered", 15 "Trial", 30 "Active", 40 "Graced", 50 "Expired", 60 "Terminated",
    # 70 "Canceled", 80 "Administrative Hold", 85 "Credit Hold", 89 "Credit+Administrative Hold"
    'SubscriptionName',  # Int
    'AccountID',  # Int
))

SSLSubscription = namedtuple('SSLSubscription', (
    'SubscriptionID',  # Int
    'ParentID',  # Int
    'SubscriptionName',  # Str(40)
    'Status',  # Int 10 "Ordered", 15 "Trial", 30 "Active", 40 "Graced", 50 "Expired", 60 "Terminated", 70 "Canceled",
    # 80 "Administrative Hold", 85 "Credit Hold", 89 "Credit+Administrative Hold"
    'startDate',  # Int
    'ExpirationDate',  # Int
    'AccountID',  # Int
    'PlanID',  # Int
    'Period',  # Int
    'PeriodType',  # Int 1  "Day(s)", 2  "Month(s)", 3  "Year(s)"
    'SetupFee',  # Double
    'SubscriptionFee',  # Double
    'RenewalFee',  # Double
    'TransferFee',  # Double
    'NonRefundableAmt',  # Double
    'RefundPeriod',  # Int
    'BillingPeriodType',  # Int
    'BillingPeriod',  # Int
    'LastBillDate',  # Int
    'NextBillDate',  # Int
    'DomainName',  # Str
    'Certificate',  # Str Services -> Certificates

    'SubscrParamValue',  # Service properties, List of pairs <ParameterId, ParameterValue>
))

DomainSubscription = namedtuple('DomainSubscription', (
    'parent_id',  # Int
    'status',  # one of SubscriptionStatus
    'start_date',  # Int
    'expiration_date',  # Int
    'account_id',  # Int
    'plan_id',  # Int
    'subscription_period',  # Int in Years
    'setup_fee',  # Double
    'subscription_fee',  # Double
    'renewal_fee',  # Double
    'transfer_fee',  # Double
    'non_refundable_amt',  # Double
    'refund_period',  # Int
    'billing_period_type',  # Int
    'billing_period',  # Int
    'last_bill_date',  # Int
    'next_bill_date',  # Int
    'full_domain_name',  # Str(255)
    'login',  # Str(40) login to registrar
    'password',  # Str(32) password to registrar
    'nameservers',  # instance of DomainSubscriptionNameServers
    'users',  # instance of DomainSubscriptionUsers
))

DomainSubscriptionInfo = namedtuple('DomainSubscriptionInfo', (
    'subscription_id',
    'domain_id',
))

SubscriptionStatus = obj(
    ORDERED=10,
    TRIAL=15,
    ACTIVE=30,
    GRACED=40,
    EXPIRED=50,
    TERMINATED=60,
    CANCELED=70,
    ADMINISTRATIVE_HOLD=80,
    CREDIT_HOLD=85,
    CREDIT_AND_ADMINISTRATIVE_HOLD=89,
)

SubscriptionServiceStatus = obj(
    NOTEXIST=10,
    CREATING=20,
    STOPPED=30,
    STARTING=40,
    ACTIVE=50,
    STOPPING=60,
    DESTROYING=70,
    CHPLAN=80,
    DESTROYED=90,
)

SubscriptionResourceStatus = obj(
    SPS_PURCHASED=0,
    SPS_PROVISIONING=1,
    SPS_INSTALLED=2,
    SPS_FAILED=3,
    SPS_ORDERED=4,
    SPS_NOTPRESENT=5,
    SPS_REMOVE=6,
    SPS_DISABLED=7,
    SPS_OPTIONAL=8,
)

SubscriptionStatusName = {
    SubscriptionStatus.ORDERED: 'Ordered',
    SubscriptionStatus.TRIAL: 'Trial',
    SubscriptionStatus.ACTIVE: 'Active',
    SubscriptionStatus.GRACED: 'Graced',
    SubscriptionStatus.EXPIRED: 'Expired',
    SubscriptionStatus.TERMINATED: 'Terminated',
    SubscriptionStatus.CANCELED: 'Canceled',
    SubscriptionStatus.ADMINISTRATIVE_HOLD: 'Administrative hold',
    SubscriptionStatus.CREDIT_HOLD: 'Credit hold',
    SubscriptionStatus.CREDIT_AND_ADMINISTRATIVE_HOLD: 'Credit and administrative hold',
}

DummySubscription = namedtuple('DummySubscription', (
    'SubscriptionID',  # Int
    'ParentID',  # Int
    'SubscriptionName',  # Str(40)
    'Status',  # Int 10 "Ordered", 15 "Trial", 30 "Active", 40 "Graced", 50 "Expired", 60 "Terminated", 70 "Canceled",
    # 80 "Administrative Hold", 85 "Credit Hold", 89 "Credit+Administrative Hold"
    'startDate',  # Int
    'ExpirationDate',  # Int
    'AccountID',  # Int
    'PlanID',  # Int
    'Period',  # Int
    'PeriodType',  # Int 1  "Day(s)", 2  "Month(s)", 3  "Year(s)"
    'SetupFee',  # Double
    'SubscriptionFee',  # Double
    'RenewalFee',  # Double
    'TransferFee',  # Double
    'NonRefundableAmt',  # Double
    'RefundPeriod',  # Int
    'BillingPeriodType',  # Int
    'BillingPeriod',  # Int
    'LastBillDate',  # Int
    'NextBillDate',  # Int
    'SubscrParamValue',  # Service properties, List of pairs <ParameterId, ParameterValue>
))

PEMSubscription = namedtuple('PEMSubscription', (
    'subscription_id',  # Int
    'plan_id',  # Int
    'expiration_date',  # Int
    'billing_period',  # instance of BillingPeriod
    'last_bill_date',  # Int
    'next_bill_date',  # Int
))

CertificateStatus = obj(
    ISSUED=2,
)


class Subscription(object):
    def __init__(self, api):
        self.api = api

    def import_subscription(self, subscription_name, billing_info, account_id, plan_info, period_info,
                            status=SubscriptionStatus.ORDERED, serv_status=SubscriptionServiceStatus.NOTEXIST):
        r = self.api.BM.SubscriptionAdd_API(
            INT_NULL, convert_to_national_subscription_name(subscription_name), status, serv_status,
            billing_info.creation_date, billing_info.expiration_date, account_id, plan_info.plan_id, period_info.period,
            period_info.period_type, period_info.setup_fee, period_info.subscription_fee, period_info.renewal_fee,
            period_info.transfer_fee, period_info.non_refundable_amt, period_info.refund_period,
            plan_info.billing_period_type, plan_info.billing_period, billing_info.last_billing_date,
            billing_info.next_billing_date
        )
        return r.SubscriptionID

    def import_subscription_with_id(self, subscription,
                                    status=SubscriptionStatus.ORDERED, serv_status=SubscriptionServiceStatus.NOTEXIST):
        self.api.BM.SubscriptionWithIDAdd_API(
            INT_NULL if subscription.SubscriptionID is None else subscription.SubscriptionID,
            INT_NULL if subscription.ParentID is None else subscription.ParentID,
            convert_to_national_subscription_name(subscription.SubscriptionName), status, serv_status,
            subscription.startDate, subscription.ExpirationDate, subscription.AccountID, subscription.PlanID,
            subscription.Period, subscription.PeriodType, subscription.SetupFee, subscription.SubscriptionFee,
            subscription.RenewalFee, subscription.TransferFee, subscription.NonRefundableAmt, subscription.RefundPeriod,
            subscription.BillingPeriodType, subscription.BillingPeriod, subscription.LastBillDate,
            subscription.NextBillDate
        )
        return subscription.SubscriptionID

    def import_pem_subscription(self, subscription_name, billing_info, account_id, plan_info, period_info,
                                status=SubscriptionStatus.ORDERED, serv_status=SubscriptionServiceStatus.NOTEXIST):

        r = self.api.PEMGATE.AddPEMSubscription_API(
            status, serv_status,
            billing_info.creation_date, billing_info.expiration_date, account_id, plan_info.plan_id, period_info.period,
            period_info.period_type, convert_to_national_subscription_name(subscription_name), period_info.setup_fee, period_info.subscription_fee, period_info.renewal_fee,
            period_info.transfer_fee, period_info.non_refundable_amt, period_info.refund_period,
            plan_info.billing_period_type, plan_info.billing_period, billing_info.last_billing_date,
            billing_info.next_billing_date
        )
        return r.SubscriptionID

    def subscription_status_update(self, subscription_id, status, serv_status):
        self.api.BM.SubscriptionStatusUpdate_API(subscription_id, status, serv_status)

    def subscription_resource_status_update(
        self, subscription_id, resource_id, status, amount=0.0, setup_fee=INT_NULL, recurring_fee=INT_NULL,
        cost_for_additional=INT_NULL
    ):
        self.api.BM.AddResourceToSubscription_API(
            subscription_id, resource_id, amount, status, setup_fee, recurring_fee, cost_for_additional
        )

    def get_subscription_resources(self, subscription_id):
        return self.api.BM.SubscriptionResourcesListGet_API(subscription_id, 0)

    def add_domain_subscription(self, subscription_name, billing_info, account_id, plan_info, period_info,
                                status=SubscriptionStatus.ORDERED, serv_status=SubscriptionServiceStatus.NOTEXIST):
        """
        subscription - instance of DomainSubscription
        return subscription_info with two fields: subscription_id and domain_id
        """
        try:
            return self.api.DOMAINGATE.DomainSubscrAdd_API(
                INT_NULL, status, serv_status,
                billing_info.creation_date, billing_info.expiration_date, account_id, plan_info.plan_id, period_info.period,
                period_info.setup_fee, period_info.subscription_fee, period_info.renewal_fee,
                period_info.transfer_fee, period_info.non_refundable_amt, period_info.refund_period,
                plan_info.billing_period_type, plan_info.billing_period, billing_info.last_billing_date,
                billing_info.next_billing_date,
                INT_NULL,
                convert_to_national_domain_name(subscription_name), '', '',
                '', '', '', '',
                INT_NULL, INT_NULL, INT_NULL, INT_NULL,
            )
        except BillingAPIError, e:
            if 'TLD does not correspond to PlanID' in e.short_message:
                required_tld = DomainImporter.get_tld_billing(subscription_name)
                plan_details, service_plan_tld = get_service_plan_tld(plan_info.plan_id)
                raise MigrationNoRepeatError(_T('TLD_OF_SERVICE_PLAN_DOES_NOT_CORRESPOND_TO_DOMAIN').format(
                    assigned_tld=service_plan_tld.tld,
                    plan=plan_info.name,
                    domain=convert_to_national_domain_name(subscription_name),
                    required_tld=convert_to_national_domain_name(required_tld)
                ))
            else:
                raise

    def get_subscription_by_order(self, order_id):
        return self.api.BM.GetSubscriptionsListByOrder_API(order_id, 0)[0]

    def add_domain_ext_data(self, domain_id, domain_ext_data):
        for key, value in domain_ext_data:
            if value is not None:
                self.api.DOMAINGATE.DomainExtDataAdd_API(domain_id, key, str(value))

    def add_subscr_param_values(self, subscription_id, param_values):
        for param, value in param_values:
            if value is not None:
                self.api.BM.SubscrParamValueUpdate_API(subscription_id, param, str(value))

    def get_domain_subscription_info(self, domain_name):
        domain_info = self.api.DOMAINGATE.DomainInfoGet_API(domain_name)
        adomain_info = self.api.DOMAINGATE.ADomainGet(domain_info.domain_id)
        return DomainSubscriptionInfo(subscription_id=adomain_info.subscription_id, domain_id=domain_info.domain_id)

    def update_domain_contacts(self, domain_id, owner_id, admin_id, billing_id, tech_id):
        self.api.DOMAINGATE.UpdateDomainContacts_API(domain_id, owner_id, admin_id, billing_id, tech_id)

    def add_domain_for_subscription(self, domain, subscription_id):
        self.api.PEMGATE.AddPEMDomainForSubscription_API(convert_to_national_domain_name(domain), subscription_id)

    def subscription_details_get(self, subscription_id):
        return self.api.BM.SubscriptionDetailsGet_API(subscription_id)

    def subscription_details_get_ex(self, subscription_id):
        return self.api.BM.SubscriptionDetailsGetEx_API(subscription_id)

    def subscription_stop(self, subscription_id, comment):
        return self.api.BM.SubscriptionStop_API(subscription_id, comment)

    def synchronize_subscription_resources(self, subscription_id):
        return self.api.PEMGATE.SynchronizeSubscrResourcesAfterMigration(subscription_id)
