# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.data.format.billing import BillingReader
from migrator.core.data.model.billing import AccountBillingModel, SubscriptionBillingModel, SourceSubscriptionStatus
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils.registry import Registry
from migrator.odin_automation.billing.lib.api.subscription import SubscriptionStatus
from migrator.odin_automation.platform.lib.connections.target_connections import AutomationTargetConnections

logger = logging.getLogger(__name__)


subscription_status_mapping = {
    SubscriptionStatus.ORDERED: SourceSubscriptionStatus.STATUS_ORDERED,
    SubscriptionStatus.TRIAL: SourceSubscriptionStatus.STATUS_TRIAL,
    SubscriptionStatus.ACTIVE: SourceSubscriptionStatus.STATUS_ACTIVE,
    SubscriptionStatus.GRACED: SourceSubscriptionStatus.STATUS_GRACED,
    SubscriptionStatus.EXPIRED: SourceSubscriptionStatus.STATUS_EXPIRED,
    SubscriptionStatus.TERMINATED: SourceSubscriptionStatus.STATUS_TERMINATED,
    SubscriptionStatus.CANCELED: SourceSubscriptionStatus.STATUS_CANCELED,
    SubscriptionStatus.ADMINISTRATIVE_HOLD: SourceSubscriptionStatus.STATUS_ADMINISTRATIVE_HOLD,
    SubscriptionStatus.CREDIT_HOLD: SourceSubscriptionStatus.STATUS_ON_HOLD,
    SubscriptionStatus.CREDIT_AND_ADMINISTRATIVE_HOLD: SourceSubscriptionStatus.STATUS_CREDIT_ADMINISTRATIVE_HOLD,
}


class OaBillingExporter(BillingReader):
    def __init__(self):
        super(OaBillingExporter, self).__init__()
        self._set_page_size(self.DEFAULT_PAGE_SIZE)
        self.source_system_reader = PlatformReaderDB()
        global_context = Registry.get_instance().get_context()
        self.is_billing_module_installed = global_context.target_panel_obj.is_billing_module_installed()

        if self.is_billing_module_installed:
            self.billing_api = AutomationTargetConnections(global_context.config).billing_api()
        else:
            self.poa_api = AutomationTargetConnections(global_context.config).poa_api()

    def get_next_account_page(self, subscription_id_list):
        """Divide account ID list to pages; each call, return a next page of account billing objects"""
        account_objects = [self.source_system_reader.get_subscription(item_id).account for item_id in
                           subscription_id_list]
        account_id_list = list({account.external_id for account in account_objects})
        for id_page in self._paginate(account_id_list):
            yield [self.get_account(account_id) for account_id in id_page]

    def get_next_subscription_page(self, subscription_id_list):
        for id_list in self._paginate(list(subscription_id_list)):
            yield [self.get_subscription(s) for s in id_list]

    def _get_account_id_list(self, subscription_id_list):
        """Return account billing objects that own the subscriptions from subscription_id_list"""
        source_system_reader = PlatformReaderDB()
        account_objects = [source_system_reader.get_subscription(item_id).account for item_id in subscription_id_list]
        return list({account.external_id for account in account_objects})

    def get_subscription(self, subscription_id):
        if self.is_billing_module_installed:
            subscription_info = self.billing_api.subscription.subscription_details_get_ex(subscription_id)
            return SubscriptionBillingModel(
                external_id=subscription_info.id,
                status=subscription_status_mapping[subscription_info.status],
                creation_date=subscription_info.start_date,
                expiration_date=subscription_info.expiration_date,
                subscription_period=subscription_info.period,
                subscription_period_type=subscription_info.period_type,
                last_billing_date=subscription_info.last_bill_date,
                next_billing_date=subscription_info.next_bill_date
            )
        else:
            subscription_info = self.poa_api.subscription.get_subscription(subscription_id)
            return SubscriptionBillingModel(
                external_id=subscription_info.subscription_id,
                status=SourceSubscriptionStatus.STATUS_ACTIVE if subscription_info.is_active else
                SourceSubscriptionStatus.STATUS_ON_HOLD
            )

    def get_account(self, account_id):
        return AccountBillingModel(
            external_id=account_id,
            balance=None,
            tax_reg_id=None,
            credit_cards=None,
            bank_accounts=None
        )
