# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.language import _T
from migrator.core.transfer.target_panel_base import TargetPanelBase
from migrator.core.utils import cached
from migrator.core.utils.registry import Registry
from migrator.odin_automation.billing.lib.api.bm import BillingAPIError
from migrator.odin_automation.platform.import_api import ImportApiOABilling, ImportApiOAPlatform
from migrator.odin_automation.platform.lib import service_plans
from migrator.odin_automation.platform.lib.api.aps.subscription import get_subscriptions, get_service_template_info
from migrator.odin_automation.platform.lib.connections.target_connections import AutomationTargetConnections
from migrator.odin_automation.platform.lib.convert.openapi_converter import OpenapiConverter
from migrator.odin_automation.platform.service_import_api import ServiceImportApiPlatform
from migrator.odin_automation.platform.lib.db.subscription import get_lsh_subscriptions, get_ng_webspaces, \
    get_whl_subscriptions, get_wsh_subscriptions
from migrator.odin_automation.platform.lib.db.service_node import get_ng_source_hosts, get_lsh_source_hosts, \
    get_ng_hosts_subscriptions_list, get_service_nodes, get_wsh_hosts_subscriptions_list, get_wsh_source_hosts, \
    get_whl_hosts_subscriptions_list, get_lsh_hosts_subscriptions_list
from migrator.odin_automation.platform.lib.db.account import get_wsh_accounts, get_wsh_accounts_subscriptions_list, \
    get_whl_accounts, get_whl_accounts_subscriptions_list, get_lsh_accounts, get_lsh_accounts_subscriptions_list
from migrator.settings import SourceHostingClass

logger = logging.getLogger(__name__)


class SubscriptionInfo:
    def __init__(self, oa_subscription_id=None, oa_subscription_name=None, oa_service_plan_id=None,
                 oa_service_plan_name=None, oa_host_id=None, oa_host_name=None, oa_account_id=None,
                 oa_account_email=None, oa_account_first_name=None, oa_account_last_name=None,
                 oa_account_company_name=None, oa_subscription_status=None, oa_subscription_period=None,
                 oa_subscription_period_type=None, oa_subscription_service_status=None):
        self.oa_subscription_id = oa_subscription_id
        self.oa_subscription_name = oa_subscription_name
        self.oa_service_plan_id = oa_service_plan_id
        self.oa_service_plan_name = oa_service_plan_name
        self.oa_host_id = oa_host_id
        self.oa_host_name = oa_host_name
        self.oa_account_id = oa_account_id
        self.oa_account_email = oa_account_email
        self.oa_account_first_name = oa_account_first_name
        self.oa_account_last_name = oa_account_last_name
        self.oa_account_company_name = oa_account_company_name
        self.oa_subscription_status = oa_subscription_status
        self.oa_subscription_period = oa_subscription_period
        self.oa_subscription_period_type = oa_subscription_period_type
        self.oa_subscription_service_status = oa_subscription_service_status


class WebspaceInfo:
    def __init__(self, oa_webspace_id=None, oa_webspace_name=None, oa_service_template_name=None,
                 oa_service_template_id=None, oa_subscription_id=None, oa_subscription_name=None,
                 oa_service_plan_id=None, oa_host_id=None, oa_host_name=None, oa_account_id=None,
                 oa_account_email=None, oa_account_first_name=None, oa_account_last_name=None,
                 oa_account_company_name=None, oa_subscription_status=None, oa_subscription_service_status=None):
        self.oa_webspace_id = oa_webspace_id
        self.oa_webspace_name = oa_webspace_name
        self.oa_service_template_name = oa_service_template_name
        self.oa_service_template_id = oa_service_template_id
        self.oa_subscription_id = oa_subscription_id
        self.oa_subscription_name = oa_subscription_name
        self.oa_service_plan_id = oa_service_plan_id
        self.oa_host_id = oa_host_id
        self.oa_host_name = oa_host_name
        self.oa_account_id = oa_account_id
        self.oa_account_email = oa_account_email
        self.oa_account_first_name = oa_account_first_name
        self.oa_account_last_name = oa_account_last_name
        self.oa_account_company_name = oa_account_company_name
        self.oa_subscription_status = oa_subscription_status
        self.oa_subscription_service_status = oa_subscription_service_status


class SourceHostInfo:
    def __init__(self, oa_host_id=None, oa_host_name=None, oa_host_type=None, oa_host_platform=None):
        self.oa_host_id = oa_host_id
        self.oa_host_name = oa_host_name
        self.oa_host_type = oa_host_type
        self.oa_host_platform = oa_host_platform


class AccountInfo:
    def __init__(self, oa_account_id=None, oa_account_company_name=None, oa_account_email=None):
        self.oa_account_id = oa_account_id
        self.oa_account_company_name = oa_account_company_name
        self.oa_account_email = oa_account_email


class SourceHostSubscriptionsListInfo:
    def __init__(self, id=None, sub_ids=None):
        self.id = id
        self.sub_ids = sub_ids


class AccountSubscriptionsListInfo:
    def __init__(self, id=None, sub_ids=None):
        self.id = id
        self.sub_ids = sub_ids


class Panel(TargetPanelBase):
    def __init__(self):
        self.global_context = Registry.get_instance().get_context()

    @cached
    def is_billing_module_installed(self):
        installed_modules = self.get_connections(self.global_context).poa_api().get_installed_modules()
        return 'PPAB' in installed_modules or 'PBA' in installed_modules

    @cached
    def is_billing_essentials(self):
        # Premium and Essentials have the same set of billing modules.
        # But credit cards and bank accounts are available only in Premium.
        # It is necessary to do check on what product is installed
        poa_api = self.get_connections(self.global_context).poa_api()
        bm_xml_rpc_address = poa_api.get_bm_xml_rpc_address()
        return poa_api.host.find_host(hostname=bm_xml_rpc_address['host']) != 1 and \
               poa_api.host.find_host(ip_address=bm_xml_rpc_address['host']) != 1

    def get_import_api(self, global_context):
        if self.is_billing_module_installed():
            return ImportApiOABilling(global_context, self)
        else:
            return ImportApiOAPlatform(global_context, self)

    def get_service_import_api(self, global_context):
        return ServiceImportApiPlatform()

    def get_service_plans(self, account_id, target_service=None, include_periods=False, plan_ids=None,
                          source_hosting_type=None):
        """
        Returns service plans by hosting class

        :param account_id: int
        :param target_service: migrator.settings.TargetServiceType
        :param include_periods: bool
        :param plan_ids: collections
        :param source_hosting_type: migrator.settings.SourceHostingClass
        :return: [migrator.odin_automation.platform.lib.service_plans.ServicePlanInfo]

        Returns all service plans if service=None
        """
        oa_conn = self.get_connections(self.global_context)
        sp = service_plans.get_service_plans_manager(oa_conn, self.is_billing_module_installed())
        return sp.get_by_service(account_id, target_service, include_periods, plan_ids, source_hosting_type)

    def get_subscriptions(self, hosting_type=None, subscription_ids=None, options=None):
        """
        Returns subscriptions list by hosting class (for LSH and WSH) or subscriptions filtered by IDs

        :param hosting_type: migrator.settings.SourceHostingClass
        :param subscription_ids: collections
        :param options: collections
        :return: [migrator.odin_automation.platform.lib.subscriptions.SubscriptionInfo | namedtuple]

        Returns all subscriptions if service=None
        """

        if hosting_type in [SourceHostingClass.LSH, SourceHostingClass.WHL, SourceHostingClass.WSH]:
            # For these hosting types paginal response will be provided. Required for UI grids.
            if hosting_type == SourceHostingClass.LSH:
                subscriptions = get_lsh_subscriptions(**self._get_filtering_options(options))
            elif hosting_type == SourceHostingClass.WHL:
                subscriptions = get_whl_subscriptions(**self._get_filtering_options(options))
            elif hosting_type == SourceHostingClass.WSH:
                subscriptions = get_wsh_subscriptions(**self._get_filtering_options(options))
            else:
                subscriptions = {'items': [], 'count': 0}

            return self._get_response_with_typed_items(subscriptions, 'SubscriptionInfo')

        # Only LSH and WSH hosting classes are supported by this method.
        # Otherwise - subscriptions_ids that are used for filtering should be specified
        if not subscription_ids:
            return {'items': [], 'count': 0}

        if self.is_billing_module_installed():
            items = []
            global_context = Registry.get_instance().get_context()
            billing_api = AutomationTargetConnections(global_context.config).billing_api()

            for subscription_id in subscription_ids:
                try:
                    subscription = billing_api.subscription.subscription_details_get_ex(subscription_id)
                    items.append(SubscriptionInfo(
                        oa_subscription_id=subscription.id,
                        oa_subscription_name=subscription.name,
                        oa_service_plan_id=subscription.plan_id,
                        oa_service_plan_name=subscription.plan_name,
                        oa_subscription_status=subscription.status,
                        oa_subscription_service_status=subscription.service_status,
                        oa_subscription_period=subscription.period,
                        oa_subscription_period_type=subscription.period_type
                    ))
                except BillingAPIError as e:
                    logger.error(_T('LOG_EXCEPTION'), exc_info=True)
                    logger.error(str(e))

        else:
            subscriptions = get_subscriptions(subscription_ids)
            items = []
            for x in subscriptions:
                items.append(SubscriptionInfo(
                    oa_subscription_id=x.subscriptionId, oa_subscription_name=x.name,
                    oa_service_plan_id=x.serviceTemplateId,
                    oa_service_plan_name=get_service_template_info(x.subscriptionId).name
                ))

        return {'count': None, 'items': items}

    def get_webspaces(self, hosting_type=None, options=None):
        """
        Returns webspaces list by hosting class

        :param hosting_type: migrator.settings.SourceHostingClass
        :param options: collections
        :return: [migrator.odin_automation.platform.lib.subscriptions.SubscriptionInfo | namedtuple]

        """

        if hosting_type == SourceHostingClass.NG:
            webspaces = get_ng_webspaces(**self._get_filtering_options(options))
        else:
            webspaces = {'count': 0, 'items': []}

        return self._get_response_with_typed_items(webspaces, 'WebspaceInfo')

    def get_source_hosts(self, hosting_type=None, options=None):
        """
        Returns source hosts list by hosting class

        :param hosting_type: migrator.settings.SourceHostingClass
        :param options: collections
        :return: [migrator.odin_automation.panel.SourceHostInfo | namedtuple]

        """

        if hosting_type == SourceHostingClass.NG:
            source_hosts = get_ng_source_hosts(**self._get_filtering_options(options))
        elif hosting_type == SourceHostingClass.WHL:
            source_hosts = get_ng_source_hosts(**self._get_filtering_options(options))
        elif hosting_type == SourceHostingClass.WSH:
            source_hosts = get_wsh_source_hosts(**self._get_filtering_options(options))
        elif hosting_type == SourceHostingClass.LSH:
            source_hosts = get_lsh_source_hosts(**self._get_filtering_options(options))
        else:
            source_hosts = {'items': [], 'count': 0}

        return self._get_response_with_typed_items(source_hosts, 'SourceHostInfo')

    def get_source_hosts_subscriptions_list(self, hosting_type=None):
        """
        Returns source hosts' subscription list by hosting class. Empty list if it isn't passed.

        :param hosting_type: migrator.settings.SourceHostingClass
        :return: [namedtuple]

        """

        if hosting_type == SourceHostingClass.NG:
            source_hosts_subscriptions_list = get_ng_hosts_subscriptions_list()
        elif hosting_type == SourceHostingClass.WHL:
            source_hosts_subscriptions_list = get_whl_hosts_subscriptions_list()
        elif hosting_type == SourceHostingClass.WSH:
            source_hosts_subscriptions_list = get_wsh_hosts_subscriptions_list()
        elif hosting_type == SourceHostingClass.LSH:
            source_hosts_subscriptions_list = get_lsh_hosts_subscriptions_list()
        else:
            source_hosts_subscriptions_list = {'items': []}

        return self._get_response_with_typed_items(
            source_hosts_subscriptions_list, 'SourceHostSubscriptionsListInfo', False
        )

    def get_accounts(self, hosting_type=None, options=None):
        """
        Returns accounts list by hosting class

        :param hosting_type: migrator.settings.SourceHostingClass
        :param options: collections
        :return: [migrator.odin_automation.panel.SourceHostInfo | namedtuple]

        """

        if hosting_type == SourceHostingClass.WHL:
            accounts = get_whl_accounts(**self._get_filtering_options(options))
        elif hosting_type == SourceHostingClass.WSH:
            accounts = get_wsh_accounts(**self._get_filtering_options(options))
        elif hosting_type == SourceHostingClass.LSH:
            accounts = get_lsh_accounts(**self._get_filtering_options(options))
        else:
            accounts = {'items': [], 'count': 0}

        return self._get_response_with_typed_items(accounts, 'AccountInfo')

    def get_accounts_subscriptions_list(self, hosting_type=None):
        """
        Returns accounts' subscription list by hosting class. Empty list if it isn't passed.

        :param hosting_type: migrator.settings.SourceHostingClass
        :return: [namedtuple]

        """

        if hosting_type == SourceHostingClass.WHL:
            accounts_subscriptions_list = get_whl_accounts_subscriptions_list()
        elif hosting_type == SourceHostingClass.WSH:
            accounts_subscriptions_list = get_wsh_accounts_subscriptions_list()
        elif hosting_type == SourceHostingClass.LSH:
            accounts_subscriptions_list = get_lsh_accounts_subscriptions_list()
        else:
            accounts_subscriptions_list = {'items': []}

        return self._get_response_with_typed_items(accounts_subscriptions_list, 'AccountSubscriptionsListInfo', False)

    def get_service_nodes(self, account_id, node_service_type=None):
        return get_service_nodes(account_id, node_service_type)

    @cached
    def get_connections(self, global_context):
        """Get target panel connections"""
        return AutomationTargetConnections(global_context.config)

    @cached
    def get_api_converter(self):
        return OpenapiConverter()

    @cached
    def get_billing_api(self, global_context):
        return self.get_connections(global_context).billing_api()

    def _get_response_with_typed_items(self, db_result, type_class_name, is_count_included=True):
        """
        Returns items instantiated with class 'type_class_name' and optionally total count of elements (from db_result)

        :param db_result: namedtuple
        :param type_class_name: str
        :param is_count_included: boolean
        :return: namedtuple
        """

        data_type_class = __import__(self.__class__.__module__, fromlist=[type_class_name]).__dict__[type_class_name]

        items = []
        for x in db_result['items']:
            items.append(data_type_class(**x))

        if is_count_included:
            return {'count': db_result['count'], 'items': items}
        else:
            return {'items': items}

    @staticmethod
    def _get_filtering_options(options):
        filtering_options = {'offset': None, 'limit': None, 'sort_field': None, 'sort_order': None, 'like': None,
                             'in_values': None, 'eq_values': None}

        if options is not None:
            filtering_options.update(options)

        return filtering_options
