# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.actions.base.common_action import CommonAction
from migrator.core.data.model.dns import DnsInfoModel
from migrator.core.data.model.subscription import has_any_of_service_types, get_service_property
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.data.source.xml.billing import BillingReaderXml, subscription_statuses_allowed_for_migration
from migrator.core.data.source.xml.dns import DnsWriterXml
from migrator.core.language import _T
from migrator.settings import \
    TargetServiceType, SOURCE_SERVICE_DNS, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, MIGRATION_TYPE_TAKEOVER, \
    SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_PLESK, SourceHostingClass

logger = logging.getLogger(__name__)


class ExportDnsData(CommonAction):
    """Export DNS data from source system."""
    def run(self, global_context):
        self._platform_db_reader = PlatformReaderDB()
        dns_subscription_ids = {
            item.id for item in global_context.migration_list_data if self._is_dns_subscription_for_export(item.id, global_context)
        }
        exporter = global_context.source_product.get_content_exporter_instance(TargetServiceType.DNS)
        with DnsWriterXml(global_context.session_files.get_path_to_dns_xml_file()) as dns_writer:
            with dns_writer(DnsInfoModel.data_prefix) as dns_info_writer:
                try:
                    for zones in exporter.get_next_zone_page(dns_subscription_ids):
                        dns_info_writer.write_list(zones)
                except Exception:
                    logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                    raise

    def _is_dns_subscription_for_export(self, subscription_id, global_context):
        """
        :param subscription_id: str Examples: '223', 'dns_104'
        :param global_context: migrator.core.global_context.GlobalMigrationContext
        :return: bool
        """
        return (self._is_managed_pure_dns_subscription(subscription_id) or \
               self._is_service_dns_subscription_for_export(subscription_id, global_context)) and \
               self._is_subscription_status_allowed_for_migration(subscription_id, global_context)

    def _is_service_dns_subscription_for_export(self, subscription_id, global_context):
        """
        :param subscription_id: str Examples: '223', 'dns_104'
        :param global_context: migrator.core.global_context.GlobalMigrationContext
        :return: bool
        """
        subscription = self._platform_db_reader.get_subscription(subscription_id)
        if subscription.parent_external_id is not None:
            parent_subscription = self._platform_db_reader.get_subscription(subscription.parent_external_id)
            if has_any_of_service_types(parent_subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.VPS]):
                return True

            if has_any_of_service_types(parent_subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]) and \
                    global_context.migration_type == MIGRATION_TYPE_TAKEOVER:
                return True

        return False

    def _is_managed_pure_dns_subscription(self, subscription_id):
        subscription = self._platform_db_reader.get_subscription(subscription_id)
        return subscription.parent_external_id is None and \
               has_any_of_service_types(subscription, [SOURCE_SERVICE_DNS]) and \
               get_service_property(subscription, SOURCE_SERVICE_DNS, 'is_dns_managed') == '1'

    def _is_subscription_status_allowed_for_migration(self, subscription_id, global_context):
        billing_reader = BillingReaderXml(global_context.session_files.get_path_to_billing_xml_file())
        billing_info = billing_reader.get_subscription(subscription_id)
        if billing_info and billing_info.status in subscription_statuses_allowed_for_migration:
            return True
        else:
            return False

    def get_description(self):
        return _T("RETRIEVING_DNS_RECORDS")

    def get_failure_message(self, global_context):
        return _T("FAILED_TO_RETRIEVE_DNS_RECORDS")

    def filter_action(self, global_context):
        source_product_type = global_context.source_type
        return source_product_type == SOURCE_PRODUCT_OBAS or \
            (source_product_type == SOURCE_PRODUCT_PLESK and global_context.migration_type == MIGRATION_TYPE_TAKEOVER)
