# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import _T
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.subscription import has_any_of_service_types, get_service_property
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.data.source.xml.dns import DnsReaderXml
from migrator.core.utils.loggers.log import trace
from migrator.odin_automation.platform.lib.api.aps.domain import get_pa_domain_by_name
from migrator.odin_automation.platform.lib.data.dns_records_importer import DnsRecordsImporter
from migrator.settings import \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SOURCE_SERVICE_DNS, SOURCE_PRODUCT_OBAS, SourceHostingClass

from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


class ImportDnsRecords(SubscriptionAction):
    def __init__(self):
        global_context = Registry.get_instance().get_context()
        self.poa_api = global_context.conn.target.poa_api()
        self._dns_records_importer = DnsRecordsImporter()

    @trace
    def run(self, global_context, subscription):

        """
        :param global_context: migrator.core.session.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: void
        """
        try:
            _dns_reader = DnsReaderXml.get(global_context.session_files.get_path_to_dns_xml_file())
            domain_name = get_service_property(subscription.source_info, SOURCE_SERVICE_DNS, 'domain')
            # TODO: get_pa_domain_by_name() checks that domain is created under the correct account making additional
            # APS requests. This check may be omitted when all dependant actions are executed in single action
            domain_id = get_pa_domain_by_name(domain_name, subscription.id)['aps']['id']
            zone = _dns_reader.get_zone(subscription.source_info.external_id)
            self._dns_records_importer.import_records(domain_id, domain_name, zone, global_context, dns_subscription=subscription)

        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise e

    @trace
    def get_description(self):
        return _T("IMPORTING_DNS_RECORDS")

    @trace
    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return global_context.source_type == SOURCE_PRODUCT_OBAS and \
        (self._is_managed_pure_dns_subscription(subscription) or self._is_vps_dns_subscription(subscription))

    @trace
    def _is_managed_pure_dns_subscription(self, subscription):
        """
        :param subscription: migrator.core.migrated_subscription.MigratedSubscription
        :return: bool
        """
        return subscription.source_info.parent_external_id is None and \
               has_any_of_service_types(subscription.source_info, [SOURCE_SERVICE_DNS]) and \
               get_service_property(subscription.source_info, SOURCE_SERVICE_DNS, 'is_dns_managed') == '1'

    @trace
    def _is_vps_dns_subscription(self, subscription):
        """
        :param subscription: migrator.core.migrated_subscription.MigratedSubscription
        :return: bool
        """
        if subscription.source_info.parent_external_id is not None:
            parent_subscription = PlatformReaderDB().get_subscription(subscription.source_info.parent_external_id)
            return has_any_of_service_types(
                parent_subscription,
                SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.VPS]
            )

        return False

    @trace
    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_TO_IMPORT_DNS_RECORDS").format(
            domain_name=get_service_property(subscription.source_info, SOURCE_SERVICE_DNS, 'domain')
        )
