# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.checks.base import CheckAbstract
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.filetypes.config_utils import ConfigHostSection
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.checks.checks import Checks
from migrator.odin_automation.services.shm.lib.aps.api import get_shm_application_nodes
from migrator.settings import \
    MIGRATION_TYPE_FULL, PANEL_TYPE_PLESK, OS_LINUX, OS_WINDOWS, TargetServiceType, SOURCE_PRODUCT_PLESK, \
    SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH

logger = logging.getLogger(__name__)

PRODUCT_TO_DOC_PARTITION_MAP = {
    SOURCE_PRODUCT_PLESK: '88377.htm#o89946',
    SOURCE_PRODUCT_OBAS: '88369.htm#o89929'
}


class CheckIpPools(CheckAbstract):
    SUPPORTED_TARGET_SERVICE_CLASSES = [TargetServiceType.SHM]
    SUPPORTED_MIGRATION_TYPES = [MIGRATION_TYPE_FULL]
    SUPPORTED_SOURCE_PRODUCTS = [
        SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH
    ]

    def __init__(self):
        super(CheckIpPools, self).__init__(Checks.CHECK_IP_POOLS, CheckAbstract.PHASE_NO_PHASE)

    def get_full_description(self):
        return _T('CHECK_FULL_DESCRIPTION_IP_POOLS')

    def _do_check(self):
        try:
            config = Registry.get_instance().get_context().config
            global_context = Registry.get_instance().get_context()
            poa_api = global_context.conn.target.poa_api()
            # Detect required OS types
            if self.source_type in [SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS]:
                required_operational_systems = self._get_required_oses_from_plesks(config)
            elif self.source_type == SOURCE_PRODUCT_OA_WSH:
                required_operational_systems = self._get_required_oses_from_oa_sh(OS_WINDOWS)
            elif self.source_type in [SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH]:
                required_operational_systems = self._get_required_oses_from_oa_sh(OS_LINUX)
            else:
                required_operational_systems = dict()

            # Detect available SHM nodes for required OS types
            # Example required_operational_systems:
            # {
            #   'linux': {
            #     'node_with_ip_pool_exists': False,
            #     'shm_nodes_guids': list['c518c8c8-3d25-436f-a6d5-3843e10d7fbe','711dfbdb-ed2b-4e18-a162-3d0f1fc7629e']
            #   },
            #  'windows': {
            #     'node_with_ip_pool_exists': False,
            #     'shm_nodes_guids': list[]
            #   }
            # }
            for shm_node in get_shm_application_nodes():
                if shm_node.os.lower() == OS_LINUX and OS_LINUX in required_operational_systems:
                    required_operational_systems[OS_LINUX]['shm_nodes_guids'].append(shm_node.uid)

                if shm_node.os.lower() == OS_WINDOWS and OS_WINDOWS in required_operational_systems:
                    required_operational_systems[OS_WINDOWS]['shm_nodes_guids'].append(shm_node.uid)

            # Check SHM nodes on IP pools availability
            for required_os, required_os_status in required_operational_systems.iteritems():
                attached_ip_pools_info = poa_api.host.get_hosts_with_bound_ip_pools(
                    required_os_status['shm_nodes_guids']
                )
                for shm_node_ip_pools_info in attached_ip_pools_info:
                    if required_operational_systems[required_os]['node_with_ip_pool_exists'] is True:
                        break

                    ip_pools = shm_node_ip_pools_info['bound_ip_pools']
                    if ip_pools:
                        required_operational_systems[required_os]['node_with_ip_pool_exists'] = True

            # Determine OSes without available IP pools
            warning_oses = list()
            for required_os, required_os_status in required_operational_systems.iteritems():
                if required_os_status['node_with_ip_pool_exists'] is False:
                    warning_oses.append(required_os)

            if not warning_oses:
                return self.STATUS_SUCCESS, _T('CHECK_IP_POOLS_SUCCESS')
            else:
                return self.STATUS_WARNING, _T('CHECK_IP_POOLS_ERROR_NO_NODE_WITH_POOLS').format(
                    oses=' or '.join(warning_os.title() for warning_os in warning_oses),
                    precheck_doc_link=self.get_link_to_precheck_documentation()
                )
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return self.STATUS_ERROR, _T('CHECK_IP_POOLS_ERROR_UNEXPECTED')

    def get_link_to_precheck_documentation(self):
        return PRODUCT_TO_DOC_PARTITION_MAP[self.source_type] if self.source_type in PRODUCT_TO_DOC_PARTITION_MAP else ''

    def _get_required_oses_from_plesks(self, config):
        """
        Detect required OS types for migration from OBAS / Plesk standalone

        :param config: migrator.core.utils.filetypes.config_utils.PanelMigratorConfigParser
        :return: dict()
        """
        required_operational_systems = dict()
        plesks_to_check = PlatformReaderDB().get_migration_required_source_hosts(PANEL_TYPE_PLESK)
        for plesk_to_check in plesks_to_check:
            plesk_config_section = ConfigHostSection(config, plesk_to_check.ip)
            plesk_os = plesk_config_section.get('os').lower()
            if plesk_os not in required_operational_systems:
                required_operational_systems[plesk_os] = dict()
                required_operational_systems[plesk_os]['node_with_ip_pool_exists'] = False
                required_operational_systems[plesk_os]['shm_nodes_guids'] = list()

        return required_operational_systems

    def _get_required_oses_from_oa_sh(self, os):
        """
        Detect required OS types for migration from OA Windows Shared Hosting (WSH) and OA Web Hosting Linux (WHL)

        :return: dict()
        """
        return {os: {
            'node_with_ip_pool_exists': False,
            'shm_nodes_guids': list()
        }}
