# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging, re

from migrator.core.checks.base import CheckAbstract
from migrator.core.data.model.billing import get_period_type_string
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.registry import Registry
from migrator.odin_automation.billing.lib.api.bm import BillingAPIError
from migrator.odin_automation.platform.checks.checks import Checks
from migrator.settings import \
    HTML_EOL, TargetServiceType, SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH

logger = logging.getLogger(__name__)


class CheckServicePlans(CheckAbstract):
    SUPPORTED_TARGET_SERVICE_CLASSES = [TargetServiceType.SHM, TargetServiceType.OACI]
    SUPPORTED_SOURCE_PRODUCTS = [
        SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH
    ]

    def __init__(self):
        super(CheckServicePlans, self).__init__(Checks.CHECK_SERVICE_PLANS, CheckAbstract.PHASE_NO_PHASE)

    def is_check_supported(self):
        return super(CheckServicePlans, self).is_check_supported() and \
            Registry.get_instance().get_context().target_panel_obj.is_billing_module_installed()

    def get_full_description(self):
        return _T('CHECK_FULL_DESCRIPTION_SERVICE_PLANS')

    def _do_check(self):
        try:
            global_context = Registry.get_instance().get_context()
            platform_reader = PlatformReaderDB()
            migration_items = platform_reader.get_migration_project_items()
            unique_service_plans_to_check = {
                item.target_service_plan_id for item in migration_items if item.target_service_plan_id is not None
            }
            service_plans_info = {}
            not_found_service_plans = []
            # Collect data required for pre-check
            for service_plan_id_to_check in unique_service_plans_to_check:
                plan_info = get_service_plan_info(service_plan_id_to_check)
                if plan_info['status'] == SERVICE_PLAN_STATUS_UNEXPECTED_ERROR:
                    return self.STATUS_ERROR, _T('CHECK_SERVICE_PLANS_GENERIC_ERROR_UNEXPECTED').format(
                        plan_id=service_plan_id_to_check,
                    )

                if plan_info['status'] == SERVICE_PLAN_STATUS_SP_NOT_FOUND:
                    not_found_service_plans.append(service_plan_id_to_check)
                elif plan_info['status'] != SERVICE_PLAN_STATUS_SP_INFO_RETRIEVED:
                    raise NotImplemented
                else:
                    service_plans_info[service_plan_id_to_check] = {
                        'periods': get_service_plan_subscription_periods(service_plan_id_to_check),
                        'details': plan_info['details']
                    }

            subscriptions_to_check = [
                item for item in migration_items
                if item.target_service_plan_id is not None and item.target_service_plan_id not in not_found_service_plans
            ]
            # Detect missed subscription periods
            missing_subscription_periods = {}
            required_subscription_periods = {}
            for item in subscriptions_to_check:
                target_plan = item.target_service_plan_id
                subscription_to_check = platform_reader.get_subscription(primary_id=item.subscription_id)
                if subscription_to_check.period is None or subscription_to_check.period_type is None:
                    # subscription period is not specified for subscription, no need to check
                    continue

                if target_plan not in required_subscription_periods:
                    required_subscription_periods[target_plan] = {'periods': set(), 'subs': set()}

                required_subscription_periods[target_plan]['periods'].add(period_as_tuple(subscription_to_check))
                required_subscription_periods[target_plan]['subs'].add(subscription_to_check.name)

                service_plan_subscription_periods = service_plans_info[target_plan]['periods'] or []
                subscription_period_is_present_in_plan = False
                for period in service_plan_subscription_periods:
                    if period.period == subscription_to_check.period and \
                       period.period_type == subscription_to_check.period_type:
                        subscription_period_is_present_in_plan = True
                        break

                if not subscription_period_is_present_in_plan:
                    if target_plan not in missing_subscription_periods:
                        missing_subscription_periods[target_plan] = {'periods': set(), 'subs': set()}

                    missing_subscription_periods[target_plan]['periods'].add(period_as_string(subscription_to_check))
                    missing_subscription_periods[target_plan]['subs'].add(subscription_to_check.name)

            errors = []
            if not_found_service_plans:
                errors.append(_T('CHECK_SERVICE_PLANS_GENERIC_ERROR_NOT_FOUND_SP').format(
                    sp_ids=', '.join(str(sp_id) for sp_id in not_found_service_plans),
                ))

            if missing_subscription_periods:
                error_detalis = []
                for target_plan, missing_periods in missing_subscription_periods.iteritems():
                    error_detalis.append(_T("SUBSCRIPTION_PERIODS_WERE_MISSING").format(
                        plan_name=service_plans_info[target_plan]['details'].name,
                        plan_id=target_plan,
                        periods=u', '.join("'%s'" % period for period in missing_periods['periods']),
                        subscriptions=self.get_subscriptions_as_limited_string(missing_periods['subs'])
                    ))

                logger.debug("Error details: %s" % error_detalis)
                logger.debug("Error details formatted: %s" % HTML_EOL + HTML_EOL.join(error_detalis) + HTML_EOL)
                errors.append(_T("THE_REQUIRED_SUBSCRIPTION_PERIODS_WERE_NOT_FOUND").format(
                    missing_periods_details=HTML_EOL + HTML_EOL.join(error_detalis) + HTML_EOL
                ))

            if errors:
                return self.STATUS_ERROR, HTML_EOL.join(errors)

            if Registry.get_instance().get_context().source_type in [SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH] and \
                    global_context.target_panel_obj.is_billing_module_installed():
                # Check prices of target subscription periods
                error_details = []
                for target_plan, required_periods in required_subscription_periods.iteritems():
                    non_zero_periods = []
                    for required_subscription_period in required_periods['periods']:
                        for plan_period_to_check in service_plans_info[target_plan]['periods']:
                            if plan_period_to_check.period == required_subscription_period[0] and \
                                    plan_period_to_check.period_type == required_subscription_period[1]:
                                if int(plan_period_to_check.setup_fee) != 0 or int(plan_period_to_check.subscription_fee) != 0:
                                    non_zero_periods.append(period_as_string(plan_period_to_check))

                    if non_zero_periods:
                        error_details.append(
                            _T('SP_GENERAL_PRECHECK_NOT_EMPTY_PRICES_MSG_PART').format(
                                sp_name=service_plans_info[target_plan]['details'].name,
                                sp_id=target_plan,
                                periods=u', '.join(non_zero_periods),
                                subscriptions=u', '.join(required_periods['subs'])
                            ))

                if error_details:
                    return self.STATUS_ERROR, _T('SP_GENERAL_PRECHECK_NOT_EMPTY_PRICES_MSG_MAIN').format(
                        error_msg=HTML_EOL + HTML_EOL.join(error_details) + HTML_EOL
                    )

            return self.STATUS_SUCCESS, _T('CHECK_SERVICE_PLANS_GENERIC_SUCCESS')
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return self.STATUS_ERROR, _T('CHECK_SERVICE_PLANS_GENERIC_ERROR_COMMON_UNEXPECTED')


SERVICE_PLAN_STATUS_SP_INFO_RETRIEVED = 0
SERVICE_PLAN_STATUS_SP_NOT_FOUND = 1
SERVICE_PLAN_STATUS_UNEXPECTED_ERROR = 2


def get_service_plan_info(service_plan_id):
    # 0 - success, 1 - SP not found, 2 - unexpected error
    result = {'status': SERVICE_PLAN_STATUS_SP_INFO_RETRIEVED, 'details': None}
    global_context = Registry.get_instance().get_context()
    target_panel_billing_api = global_context.target_panel_obj.get_billing_api(global_context)
    try:
        result['details'] = target_panel_billing_api.plan_details_get(service_plan_id)
    except BillingAPIError as e:
        if re.match("^(.*)doesn't contain row(.*)", e.short_message):
            result['status'] = SERVICE_PLAN_STATUS_SP_NOT_FOUND
        else:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            result['status'] = SERVICE_PLAN_STATUS_UNEXPECTED_ERROR
    except Exception:
        logger.error(_T("LOG_EXCEPTION"), exc_info=True)
        result['status'] = SERVICE_PLAN_STATUS_UNEXPECTED_ERROR

    return result


def get_service_plan_subscription_periods(service_plan_id):
    global_context = Registry.get_instance().get_context()
    target_panel_billing_api = global_context.target_panel_obj.get_billing_api(global_context)
    plan_period_list_full = target_panel_billing_api.plan_period_list_get(service_plan_id)
    return [x for x in plan_period_list_full if x.trial == 0 and x.enabled == 1]


def period_as_string(period):
    return "%s %s(s)" % (period.period, get_period_type_string(period.period_type))


def period_as_tuple(subscription):
    """
    :param subscription: SubscriptionModel
    :return: tuple (subscription period, subscription period type)
    """
    return (subscription.period, subscription.period_type)
