# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import InternalError
from migrator.core.checks.base import CheckSourceAbstract, CheckAbstract
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.checks.checks import Checks
from migrator.odin_automation.platform.checks.lib.service_plans_general import \
    get_service_plan_info, SERVICE_PLAN_STATUS_SP_INFO_RETRIEVED, SERVICE_PLAN_STATUS_SP_NOT_FOUND
from migrator.odin_automation.platform.lib.api.aps.subscription import get_service_template_id
from migrator.odin_automation.platform.lib.api.openapi.service import \
    R_DNS, R_EMAIL_DISK, R_EMAIL_USERS, R_HTTP_REDIR, R_IP, R_MAIL_TEMPL, R_PLESK_MAIL, R_SITE, R_TRAFF, R_WEB_LIMITS, \
    R_WEB_SERV, RL_DISK, RL_FTP_USERS, RL_HOSTING_CONF, RL_MY_OR_POSTGRE_SQL_WEB, RL_PLESK_WEB, RL_WEB_TEMPL, RW_DISK, \
    RW_FTP_USERS, RW_HOSTING_CONF, RW_MS_SQL_WEB, RW_MY_OR_POSTGRE_SQL_WEB, RW_PLESK_WEB, RW_WEB_TEMPL, \
    R_EMAIL_DOMAIN, R_WEBSPACE, RES_AVAILABILITY_RESULT_SOME_RESOURCES_NOT_FOUND, STResource, \
    RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND, R_VPS_CORE_POWER
from migrator.settings import \
    HTML_EOL, OS_WINDOWS, OS_LINUX, OS_UNKNOWN, SUPPORTED_OS_TYPES, SourceHostingClass, \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SOURCE_SERVICE_DNS, SOURCE_SERVICE_MAIL, SOURCE_SERVICE_DB, \
    SOURCE_SERVICE_WEB, SOURCE_SERVICE_FTP, TargetServiceType, SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, \
    SOURCE_PRODUCT_OA_WSH, SOURCE_SERVICE_FTP_WSH, SOURCE_SERVICE_WEB_WSH, SOURCE_SERVICE_DB_WSH, \
    SOURCE_SERVICE_FTP_WHL, SOURCE_SERVICE_DB_WHL, SOURCE_SERVICE_WEB_WHL, SOURCE_PRODUCT_OA_WHL, \
    SOURCE_SERVICE_FTP_LSH, SOURCE_SERVICE_DB_LSH, SOURCE_SERVICE_WEB_LSH, SOURCE_PRODUCT_OA_LSH, \
    SHARED_HOSTING_SERVICES

logger = logging.getLogger(__name__)

ERROR_CODE_INACTIVE_ST_DETECTED = 1
ERROR_CODE_RESOURCES_NOT_FOUND = 2

REQUIRED_RESOURCES_WHP_SERVICE_MAIL = frozenset([R_EMAIL_USERS, R_EMAIL_DOMAIN, R_EMAIL_DISK, R_MAIL_TEMPL, R_PLESK_MAIL])
REQUIRED_RESOURCES_WHP_SERVICE_FTP_WINDOWS = frozenset([RW_FTP_USERS])
REQUIRED_RESOURCES_WHP_SERVICE_FTP_LINUX = frozenset([RL_FTP_USERS])
# Equivalent resources as tuples
REQUIRED_RESOURCES_WHP_SERVICE_DB_WINDOWS = frozenset([(RW_MY_OR_POSTGRE_SQL_WEB, RW_MS_SQL_WEB)])
REQUIRED_RESOURCES_WHP_SERVICE_DB_LINUX = frozenset([RL_MY_OR_POSTGRE_SQL_WEB])
REQUIRED_RESOURCES_WHP_SERVICE_WEB_WINDOWS = frozenset([
    RW_HOSTING_CONF, RW_WEB_TEMPL, R_HTTP_REDIR, RW_DISK, R_IP, RW_PLESK_WEB, R_TRAFF, R_WEB_LIMITS,
    R_WEB_SERV, R_SITE, R_WEBSPACE
])
REQUIRED_RESOURCES_WHP_SERVICE_WEB_LINUX = frozenset([
    RL_HOSTING_CONF, RL_WEB_TEMPL, R_HTTP_REDIR, RL_DISK, R_IP, RL_PLESK_WEB, R_TRAFF, R_WEB_LIMITS,
    R_WEB_SERV, R_SITE, R_WEBSPACE
])
REQUIRED_RESOURCES_WHP_SERVICE_DNS = frozenset([R_DNS])

SUPPORTED_ST_CONFIGURATIONS = {
    TargetServiceType.SHM: {
        OS_WINDOWS: {
            SOURCE_SERVICE_DNS: REQUIRED_RESOURCES_WHP_SERVICE_DNS,
            # Shared Hosting of Plesk/OBAS
            SOURCE_SERVICE_MAIL: REQUIRED_RESOURCES_WHP_SERVICE_MAIL,
            SOURCE_SERVICE_FTP: REQUIRED_RESOURCES_WHP_SERVICE_FTP_WINDOWS,
            SOURCE_SERVICE_DB: REQUIRED_RESOURCES_WHP_SERVICE_DB_WINDOWS,
            SOURCE_SERVICE_WEB: REQUIRED_RESOURCES_WHP_SERVICE_WEB_WINDOWS,
            # OA Windows Shared Hosting
            SOURCE_SERVICE_FTP_WSH: REQUIRED_RESOURCES_WHP_SERVICE_FTP_WINDOWS,
            SOURCE_SERVICE_DB_WSH: REQUIRED_RESOURCES_WHP_SERVICE_DB_WINDOWS,
            SOURCE_SERVICE_WEB_WSH: REQUIRED_RESOURCES_WHP_SERVICE_WEB_WINDOWS
        },
        OS_LINUX: {
            SOURCE_SERVICE_DNS: REQUIRED_RESOURCES_WHP_SERVICE_DNS,
            # Shared Hosting of Plesk/OBAS
            SOURCE_SERVICE_MAIL: REQUIRED_RESOURCES_WHP_SERVICE_MAIL,
            SOURCE_SERVICE_FTP: REQUIRED_RESOURCES_WHP_SERVICE_FTP_LINUX,
            SOURCE_SERVICE_DB: REQUIRED_RESOURCES_WHP_SERVICE_DB_LINUX,
            SOURCE_SERVICE_WEB: REQUIRED_RESOURCES_WHP_SERVICE_WEB_LINUX,
            # OA WebHosting Linux
            SOURCE_SERVICE_FTP_WHL: REQUIRED_RESOURCES_WHP_SERVICE_FTP_LINUX,
            SOURCE_SERVICE_DB_WHL: REQUIRED_RESOURCES_WHP_SERVICE_DB_LINUX,
            SOURCE_SERVICE_WEB_WHL: REQUIRED_RESOURCES_WHP_SERVICE_WEB_LINUX,
            # OA Legasy Linux Shared Hosting
            SOURCE_SERVICE_FTP_LSH: REQUIRED_RESOURCES_WHP_SERVICE_FTP_LINUX,
            SOURCE_SERVICE_DB_LSH: REQUIRED_RESOURCES_WHP_SERVICE_DB_LINUX,
            SOURCE_SERVICE_WEB_LSH: REQUIRED_RESOURCES_WHP_SERVICE_WEB_LINUX,
        }
    }
}

TARGET_WHP_SERVICE_NAME_FTP = 'FTP'
TARGET_WHP_SERVICE_NAME_MAIL = 'Mail'
TARGET_WHP_SERVICE_NAME_DB = 'Database'
TARGET_WHP_SERVICE_NAME_WEB = 'WEB'
TARGET_SERVICE_NAME_DNS = 'DNS'

RESOURCES_TO_AFFECTED_SERVICES_MAP = {
    TargetServiceType.SHM: {
        RL_FTP_USERS: TARGET_WHP_SERVICE_NAME_FTP,
        RW_FTP_USERS: TARGET_WHP_SERVICE_NAME_FTP,
        R_DNS: TARGET_SERVICE_NAME_DNS,
        R_EMAIL_USERS: TARGET_WHP_SERVICE_NAME_MAIL,
        R_EMAIL_DOMAIN: TARGET_WHP_SERVICE_NAME_MAIL,
        R_EMAIL_DISK: TARGET_WHP_SERVICE_NAME_MAIL,
        RL_HOSTING_CONF: TARGET_WHP_SERVICE_NAME_WEB,
        RW_HOSTING_CONF: TARGET_WHP_SERVICE_NAME_WEB,
        RL_WEB_TEMPL: TARGET_WHP_SERVICE_NAME_WEB,
        RW_WEB_TEMPL: TARGET_WHP_SERVICE_NAME_WEB,
        R_HTTP_REDIR: TARGET_WHP_SERVICE_NAME_WEB,
        RW_DISK: TARGET_WHP_SERVICE_NAME_WEB,
        RL_DISK: TARGET_WHP_SERVICE_NAME_WEB,
        R_IP: TARGET_WHP_SERVICE_NAME_WEB,
        R_MAIL_TEMPL: TARGET_WHP_SERVICE_NAME_MAIL,
        RW_MY_OR_POSTGRE_SQL_WEB: TARGET_WHP_SERVICE_NAME_DB,
        RL_MY_OR_POSTGRE_SQL_WEB: TARGET_WHP_SERVICE_NAME_DB,
        RL_PLESK_WEB: TARGET_WHP_SERVICE_NAME_WEB,
        RW_PLESK_WEB: TARGET_WHP_SERVICE_NAME_WEB,
        R_PLESK_MAIL: TARGET_WHP_SERVICE_NAME_MAIL,
        R_TRAFF: TARGET_WHP_SERVICE_NAME_WEB,
        R_WEB_LIMITS: TARGET_WHP_SERVICE_NAME_WEB,
        R_WEB_SERV: TARGET_WHP_SERVICE_NAME_WEB,
        R_SITE: TARGET_WHP_SERVICE_NAME_WEB,
        R_WEBSPACE: TARGET_WHP_SERVICE_NAME_WEB,
        RW_MS_SQL_WEB: TARGET_WHP_SERVICE_NAME_DB
    }
}


class CheckServiceTemplates(CheckSourceAbstract):
    SUPPORTED_TARGET_SERVICE_CLASSES = [TargetServiceType.SHM, TargetServiceType.OACI]
    SUPPORTED_SOURCE_PRODUCTS = [
        SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH
    ]
    SERVICE_PLAN_NOT_FOUND = '-1'
    LOST_SP_SUBS_MAX_AMOUNT_TO_SHOW = 20

    def __init__(self):
        super(CheckServiceTemplates, self).__init__(Checks.CHECK_SERVICE_TEMPLATES, CheckAbstract.PHASE_NO_PHASE)
        self.service_templates = {}
        self.poa_api = None
        self.platform_reader = None

    def get_full_description(self):
        return _T('CHECK_FULL_DESCRIPTION_SERVICE_TEMPLATES')

    def get_pair_of_equivalent_resources_as_string(self, resources_pair):
        """
        :param resources_pair: list[STResource]|tuple[STResource]
        :return: string
        """
        return _T('PAIR_OF_EQUIVALENT_RESOURCES').format(
            first_res=resources_pair[0].get_st_resource_description(),
            second_res=resources_pair[1].get_st_resource_description()
        )

    def get_service_template_id_if_billing(self, service_plan_id):
        """
        :param service_plan_id: int
        :return: int
        :raises Exception
        """
        sp_info = get_service_plan_info(service_plan_id)
        if sp_info['status'] == SERVICE_PLAN_STATUS_SP_INFO_RETRIEVED:
            return sp_info['details'].st_id
        elif sp_info['status'] == SERVICE_PLAN_STATUS_SP_NOT_FOUND:
            return self.SERVICE_PLAN_NOT_FOUND
        else:
            raise Exception(_T('CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_UNEXPECTED'))

    def _do_check(self):
        try:
            global_context = Registry.get_instance().get_context()
            platform_reader = PlatformReaderDB()
            items = platform_reader.get_migration_project_items(self._get_supported_source_services()) + \
                    platform_reader.get_migration_project_items(
                        SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]
                    )
            return self.check_migration_items(global_context, items)
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return self.STATUS_ERROR, _T('CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_UNEXPECTED')

    def _get_supported_source_services(self):
        if self.source_type in [SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS]:
            return SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]
        elif self.source_type in [SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH]:
            return SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[self.source_type.split('.')[1]]
        else:
            raise InternalError(_T('CHECK_SERVICE_TEMPLATES_GENERIC_INT_ERR_UNEXPECTED_PRODUCT').format(
                s_product=self.source_type
            ))

    def check_migration_items(self, global_context, items):
        self.poa_api = self.poa_api or global_context.conn.target.poa_api()
        self.platform_reader = self.platform_reader or PlatformReaderDB()
        # Example of service_templates_data:
        # {
        #   3: {
        #       'required_resources': set([
        #           R_EMAIL_ACCOUNTS, R_DNS_HOSTING, R_MYSQL_AND_POSTGRESQL_DATABASES_FOR_A_WEBSPACE_LINUX,
        #           [
        #               R_MYSQL_AND_POSTGRESQL_DATABASES_FOR_A_WEBSPACE_WINDOWS,
        #               R_MS_SQL_DATABASES_FOR_A_WEBSPACE_WINDOWS
        #           ], RW_FTP_USERS
        #       ]),
        #       'missed_resources': set([R_EMAIL_ACCOUNTS, RW_FTP_USERS]),
        #       'affected_services': set(['ftp', 'mail']),
        #       'subs': set(['my-windows.tld', 'my-linux.net']),
        #   }
        #   5: {
        #       'required_resources': set([
        #           R_DNS_HOSTING, R_MYSQL_AND_POSTGRESQL_DATABASES_FOR_A_WEBSPACE_LINUX, R_EMAIL_ACCOUNTS, RW_DISK
        #        ]),
        #       'missed_resources': set([RW_DISK])
        #       'affected_services': set(['web'])
        #       'subs': set(['my-linux.net'])
        #   }
        # }
        missed_plans = {}
        service_templates_data = dict()
        for migration_item in items:
            required_resources = self._get_required_resources(global_context, migration_item)
            if not required_resources:
                continue

            service_template_id_to_check = self._get_migration_item_target_oa_st_id(global_context, migration_item)
            if service_template_id_to_check is None:
                continue

            if service_template_id_to_check == self.SERVICE_PLAN_NOT_FOUND:
                if migration_item.target_service_plan_id not in missed_plans:
                    missed_plans[migration_item.target_service_plan_id] = []

                missed_plans[migration_item.target_service_plan_id].append(
                    self.platform_reader.get_subscription(primary_id=migration_item.subscription_id).name
                )
                continue

            if service_template_id_to_check not in service_templates_data:
                service_templates_data[service_template_id_to_check] = dict()
                service_templates_data[service_template_id_to_check]['required_resources'] = set()
                service_templates_data[service_template_id_to_check]['missed_resources'] = set()
                service_templates_data[service_template_id_to_check]['affected_services'] = set()
                service_templates_data[service_template_id_to_check]['subs'] = set()

            service_templates_data[service_template_id_to_check]['required_resources'].update(
                required_resources
            )
            service_templates_data[service_template_id_to_check]['subs'].add(
                migration_item.subscription.name
            )

        if missed_plans:
            return self.STATUS_ERROR, _T('CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_FOUND_NONEXISTENT_SPS').format(
                lost_sps=u', '.join(str(sp_id) for sp_id in missed_plans.keys()),
                subs_list=self.get_subscriptions_as_limited_string(
                    [sub for subs_list in missed_plans.values() for sub in subs_list],
                    self.LOST_SP_SUBS_MAX_AMOUNT_TO_SHOW
                )
            )

        if not service_templates_data:
            return self.STATUS_SUCCESS, _T('CHECK_SERVICE_TEMPLATES_GENERIC_SUCCESS')

        st_ids = service_templates_data.keys()
        logger.debug(_T("CHECK_SERVICE_TEMPLATES_GENERIC_IDS_TO_CHECK").format(
            service_templates_ids=', '.join(str(st_id) for st_id in st_ids))
        )
        service_templates = self._get_service_templates(st_ids)
        logger.debug(_T("CHECK_SERVICE_TEMPLATES_GENERIC_DEBUG_FOUND_ST").format(
            st_info=service_templates
        ))

        # Check if STs are "active"
        inactive_service_templates = []
        for service_template in service_templates:
            if not service_template.is_active:
                inactive_service_templates.append(service_template)

        if inactive_service_templates:
            affected_subs = set()
            for inactive_service_template in inactive_service_templates:
                affected_subs.update(service_templates_data[inactive_service_template.st_id]['subs'])

            self.set_result_code(ERROR_CODE_INACTIVE_ST_DETECTED)
            return self.STATUS_ERROR, _T('CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_FOUND_INACTIVE_ST').format(
                service_templates=', '.join(st.get_identificators() for st in inactive_service_templates),
                subs_list=self.get_subscriptions_as_limited_string(list(affected_subs))
            )
        else:
            logger.debug(_T("CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_FOUND_INACTIVE_ST_OK").format(
                service_templates_ids=', '.join(str(st_id) for st_id in st_ids))
            )

        # Detect missed resources
        for service_template in service_templates:
            for required_resource in service_templates_data[service_template.st_id]['required_resources']:
                missed_resource = None
                # Unique resource
                if isinstance(required_resource, STResource):
                    st_res_status = service_template.are_resources_exist(list([required_resource]))
                    if st_res_status['status'] == RES_AVAILABILITY_RESULT_SOME_RESOURCES_NOT_FOUND:
                        missed_resource = required_resource
                # Tuple with equivalent resources
                elif isinstance(required_resource, tuple):
                    at_least_one_resource_exists = False
                    for resource_of_equivalent_res_pair in required_resource:
                        st_res_status = service_template.are_resources_exist(
                            list([resource_of_equivalent_res_pair])
                        )
                        if st_res_status['status'] == RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND:
                            at_least_one_resource_exists = True
                            break

                    if at_least_one_resource_exists is False:
                        missed_resource = required_resource
                else:
                    raise InternalError(_T('CHECK_SERVICE_TEMPLATES_GENERIC_INT_ERR_UNEXPECTED_RESOURCE'))

                if missed_resource:
                    service_templates_data[service_template.st_id]['missed_resources'].add(missed_resource)
                    if isinstance(missed_resource, tuple):
                        missed_resource = missed_resource[0]
                    service_templates_data[service_template.st_id]['affected_services'].add(
                        RESOURCES_TO_AFFECTED_SERVICES_MAP[TargetServiceType.SHM][missed_resource]
                    )

        were_some_resources_missed = False
        for st_id, st_resources_status in service_templates_data.iteritems():
            if not st_resources_status['missed_resources']:
                continue
            else:
                were_some_resources_missed = True
                break

        if not were_some_resources_missed:
            return self.STATUS_SUCCESS, _T('CHECK_SERVICE_TEMPLATES_GENERIC_SUCCESS')

        # Compose an error message
        error_message_body_parts = list()
        for service_template in service_templates:
            if not service_templates_data[service_template.st_id]['missed_resources']:
                continue

            missed_resources = list()
            for missed_resource in service_templates_data[service_template.st_id]['missed_resources']:
                if isinstance(missed_resource, STResource):
                    missed_resources.append(missed_resource.get_st_resource_description())
                elif isinstance(missed_resource, tuple):
                    missed_resources.append(self.get_pair_of_equivalent_resources_as_string(missed_resource))
                else:
                    raise InternalError(
                        'Detected unexpected resource data type: an internal error occurred. Contact Odin technical support for assistance.')

            error_message_body_parts.append(
                _T('SERVICE_TEMPLATE_WITH_MISSED_RESOURCES_PHRASE').format(
                    missed_res_list=', '.join(missed_resources),
                    st_id=service_template.get_identificators(),
                    subs_list=self.get_subscriptions_as_limited_string(
                        list(service_templates_data[service_template.st_id]['subs'])
                    ),
                    affected_service_name=', '.join(
                        service_templates_data[service_template.st_id]['affected_services']
                    )
                )
            )
            logger.debug(_T('CHECK_SERVICE_TEMPLATES_GENERIC_DEBUG_GET_AVAILABLE_RESOURCES_PARAMS_FOR_ST').format(
                missed_resources=', '.join(missed_resources),
                st_with_info=service_template.get_available_resource_types()
            ))

        missed_resources_details = HTML_EOL + HTML_EOL.join(error_message_body_parts) + HTML_EOL
        self.set_result_code(ERROR_CODE_RESOURCES_NOT_FOUND)
        self.set_result_meta_data(_T('SKIPPED_CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_RESOURCES_MISSED').format(
            st_name=service_templates[0].get_identificators(),
            missed_resources_details=missed_resources_details
        ))

        return self.STATUS_ERROR, _T('CHECK_SERVICE_TEMPLATES_GENERIC_ERROR_RESOURCES_MISSED').format(
            missed_resources_details=missed_resources_details
        )

    def _get_required_resources(self, global_context, item):
        if has_any_of_service_types(item.subscription, SHARED_HOSTING_SERVICES):
            return self._get_required_resources_shm(global_context, item)
        elif has_any_of_service_types(item.subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            return self._get_required_resources_dns(item)
        elif has_any_of_service_types(item.subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.VPS]):
            # ToDo
            return set([R_VPS_CORE_POWER])
        else:
            raise InternalError("Unsupported services set: %s" % item.subscription.services)

    def _get_required_resources_shm(self, global_context, item):
        required_resources = set()
        subscription_services = item.subscription.services
        subscription_os = self._get_source_subscription_os(global_context, subscription_services)
        for subscription_service in subscription_services:
            s_type = subscription_service.type
            if s_type not in self._get_supported_source_services_shm(subscription_os):
                logger.warning(
                    _T('CHECK_SERVICE_TEMPLATES_GENERIC_FOUND_UNEXPECTED_SERVICE').format(
                        service_name=s_type,
                        sub_name=item.subscription.name
                    )
                )
                continue

            required_resources.update(SUPPORTED_ST_CONFIGURATIONS[TargetServiceType.SHM][subscription_os][s_type])

        return required_resources

    @cached
    def _get_supported_source_services_shm(self, os):
        """
        :param os: string Examples: 'linux', 'windows'
        :return: list
        """
        return SUPPORTED_ST_CONFIGURATIONS[TargetServiceType.SHM][os].keys()

    def _get_source_subscription_os(self, global_context, subscription_services):
        """
        Detect subscription OS by its services

        :param global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription_services: list Example: ['ftp', 'web', 'db', 'mail']
        :return: str Examples: 'windows', 'linux'
        """
        subscription_os = OS_UNKNOWN
        for subscription_service in subscription_services:
            if subscription_service.type == SOURCE_SERVICE_WEB:
                web_host = self.platform_reader.get_source_host(subscription_service.source_host_external_id)
                subscription_os = global_context.config.get(web_host.ip, 'os').lower()
                break
            elif subscription_service.type == SOURCE_SERVICE_WEB_WSH:
                subscription_os = OS_WINDOWS
                break
            elif subscription_service.type in [SOURCE_SERVICE_WEB_WHL, SOURCE_SERVICE_WEB_LSH]:
                subscription_os = OS_LINUX
                break

        if subscription_os not in SUPPORTED_OS_TYPES:
            raise InternalError(
                'Unable to detect the subscription source OS: an internal error occurred. Contact Odin technical support for assistance.'
            )

        return subscription_os

    def _get_required_resources_dns(self, item):
        # DNS subscriptions with None target_service_plan_id requires "DNS Hosting" resource in WEB migration item
        return set([R_DNS]) if item.target_service_plan_id is None else set()

    def _get_migration_item_target_oa_st_id(self, global_context, item):
        is_billing = global_context.target_panel_obj.is_billing_module_installed()
        if has_any_of_service_types(item.subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            if item.target_service_plan_id is not None:
                # standalone DNS subscription should not be checked, something is wrong with migration items selection
                raise InternalError()

            # free DNS subscriptions use the hosting subscription's service plan, find it
            service_template_id = self._get_target_plan_id_for_free_dns_subscription(item.subscription_id)
        else:
            service_template_id = item.target_service_plan_id
            # Merging into existing OA subscription - check service template of OA subscription
            if item.is_oa_subscription is True and service_template_id is None:
                return get_service_template_id(item.mapping_subscription_id)

        if is_billing and service_template_id is not None:
            service_template_id = self.get_service_template_id_if_billing(service_template_id)

        return service_template_id

    def _get_target_plan_id_for_free_dns_subscription(self, subscription_id):
        subscription_id_in_migration_item_to_check_dns = self.platform_reader.get_subscription(
            primary_id=subscription_id
        ).parent_external_id
        web_subscription_to_check_st = self.platform_reader.get_subscription(
            subscription_id_in_migration_item_to_check_dns
        )
        migr_item_with_target_sp_id_to_check = self.platform_reader.get_migration_project_item(
            web_subscription_to_check_st.id
        )
        return migr_item_with_target_sp_id_to_check.target_service_plan_id

    def _get_service_templates(self, st_ids):
        templates = []
        for st_id in st_ids:
            if st_id not in self.service_templates:
                self.service_templates[st_id] = self.poa_api.service.get_service_template(st_id, True)

            templates.append(self.service_templates[st_id])

        return templates
