# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import re

from migrator.core.checks.base import CheckAbstract
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.format_info import \
    MAX_ADDRESS_LENGTH, MAX_AREA_CODE_LENGTH, MAX_CITY_LENGTH, MAX_COUNTRY_CODE_LENGTH, \
    MAX_COMPANY_NAME_LENGTH, MAX_EMAIL_LENGTH, MAX_EXTENSION_NUMBER_LENGTH, MAX_NAME_LENGTH, \
    MAX_PHONE_NUMBER_LENGTH, MAX_STATE_LENGTH, MAX_ZIPCODE_LENGTH, NUMBER_REGEXP, EMAIL_REGEXP, COUNTRIES_LIST, \
    MAX_LOGIN_LENGTH, MIN_LOGIN_LENGTH, MIN_PASSWORD_LENGTH, MAX_PASSWORD_LENGTH, LOGIN_REGEXP
from migrator.odin_automation.platform.checks.checks import Checks
from migrator.settings import HTML_EOL, SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, TargetServiceType

logger = logging.getLogger(__name__)

# Generic account properties
ACCOUNT_FIELD_COMPANY_NAME = 'info.company_name'
ACCOUNT_FIELD_ADDRESS_1 = 'info.company_address.address1'
ACCOUNT_FIELD_ADDRESS_2 = 'info.company_address.address2'
ACCOUNT_FIELD_CITY = 'info.company_address.city'
ACCOUNT_FIELD_ZIP_CODE = 'info.company_address.zipcode'
ACCOUNT_FIELD_COUNTRY = 'info.company_address.country'
ACCOUNT_FIELD_STATE = 'info.company_address.state'
# Billing person of account
ACCOUNT_FIELD_BILLING_CONTACT_FIRST_NAME = 'info.billing_contact.first_name'
ACCOUNT_FIELD_BILLING_CONTACT_MIDDLE_NAME = 'info.billing_contact.middle_name'
ACCOUNT_FIELD_BILLING_CONTACT_LAST_NAME = 'info.billing_contact.last_name'
ACCOUNT_FIELD_BILLING_CONTACT_EMAIL = 'info.billing_contact.email'
ACCOUNT_FIELD_BILLING_CONTACT_PHONE_COUNTRY_CODE = 'info.billing_contact.phone.country_code'
ACCOUNT_FIELD_BILLING_CONTACT_PHONE_AREA_CODE = 'info.billing_contact.phone.area_code'
ACCOUNT_FIELD_BILLING_CONTACT_PHONE_NUM = 'info.billing_contact.phone.phone_num'
ACCOUNT_FIELD_BILLING_CONTACT_PHONE_EXT_NUM = 'info.billing_contact.phone.ext_num'
ACCOUNT_FIELD_BILLING_CONTACT_FAX_COUNTRY_CODE = 'info.billing_contact.fax.country_code'
ACCOUNT_FIELD_BILLING_CONTACT_FAX_AREA_CODE = 'info.billing_contact.fax.area_code'
ACCOUNT_FIELD_BILLING_CONTACT_FAX_NUM = 'info.billing_contact.fax.phone_num'
ACCOUNT_FIELD_BILLING_CONTACT_FAX_EXT_NUM = 'info.billing_contact.fax.ext_num'
# Admin person of account
ACCOUNT_FIELD_ADMIN_CONTACT_FIRST_NAME = 'info.admin_contact.first_name'
ACCOUNT_FIELD_ADMIN_CONTACT_MIDDLE_NAME = 'info.admin_contact.middle_name'
ACCOUNT_FIELD_ADMIN_CONTACT_LAST_NAME = 'info.admin_contact.last_name'
ACCOUNT_FIELD_ADMIN_CONTACT_EMAIL = 'info.admin_contact.email'
ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_COUNTRY_CODE = 'info.admin_contact.phone.country_code'
ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_AREA_CODE = 'info.admin_contact.phone.area_code'
ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_NUM = 'info.admin_contact.phone.phone_num'
ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_EXT_NUM = 'info.admin_contact.phone.ext_num'
ACCOUNT_FIELD_ADMIN_CONTACT_FAX_COUNTRY_CODE = 'info.admin_contact.fax.country_code'
ACCOUNT_FIELD_ADMIN_CONTACT_FAX_AREA_CODE = 'info.admin_contact.fax.area_code'
ACCOUNT_FIELD_ADMIN_CONTACT_FAX_NUM = 'info.admin_contact.fax.phone_num'
ACCOUNT_FIELD_ADMIN_CONTACT_FAX_EXT_NUM = 'info.admin_contact.fax.ext_num'
# Technical person of account
ACCOUNT_FIELD_TECH_CONTACT_FIRST_NAME = 'info.tech_contact.first_name'
ACCOUNT_FIELD_TECH_CONTACT_MIDDLE_NAME = 'info.tech_contact.middle_name'
ACCOUNT_FIELD_TECH_CONTACT_LAST_NAME = 'info.tech_contact.last_name'
ACCOUNT_FIELD_TECH_CONTACT_EMAIL = 'info.tech_contact.email'
ACCOUNT_FIELD_TECH_CONTACT_PHONE_COUNTRY_CODE = 'info.tech_contact.phone.country_code'
ACCOUNT_FIELD_TECH_CONTACT_PHONE_AREA_CODE = 'info.tech_contact.phone.area_code'
ACCOUNT_FIELD_TECH_CONTACT_PHONE_NUM = 'info.tech_contact.phone.phone_num'
ACCOUNT_FIELD_TECH_CONTACT_PHONE_EXT_NUM = 'info.tech_contact.phone.ext_num'
ACCOUNT_FIELD_TECH_CONTACT_FAX_COUNTRY_CODE = 'info.tech_contact.fax.country_code'
ACCOUNT_FIELD_TECH_CONTACT_FAX_AREA_CODE = 'info.tech_contact.fax.area_code'
ACCOUNT_FIELD_TECH_CONTACT_FAX_NUM = 'info.tech_contact.fax.phone_num'
ACCOUNT_FIELD_TECH_CONTACT_FAX_EXT_NUM = 'info.tech_contact.fax.ext_num'

# Staff member properties
USER_FIELD_LOGIN = 'authentication.login'
USER_FIELD_PASSWORD = 'authentication.password'
USER_FIELD_FIRST_NAME = 'contact.first_name'
USER_FIELD_LAST_NAME = 'contact.last_name'
USER_FIELD_MIDDLE_NAME = 'contact.middle_name'
USER_FIELD_TITLE = 'contact.title'
USER_FIELD_PHONE_AREA_CODE = 'contact.phone.area_code'
USER_FIELD_PHONE_EXT_NUM = 'contact.phone.ext_num'
USER_FIELD_PHONE_NUM = 'contact.phone.phone_num'
USER_FIELD_PHONE_COUNTRY_CODE = 'contact.phone.country_code'
USER_FIELD_FAX_AREA_CODE = 'contact.fax.area_code'
USER_FIELD_FAX_EXT_NUM = 'contact.fax.ext_num'
USER_FIELD_FAX_NUM = 'contact.fax.phone_num'
USER_FIELD_FAX_COUNTRY_CODE = 'contact.fax.country_code'
USER_FIELD_EMAIL = 'contact.email'
USER_FIELD_CITY = 'address.city'
USER_FIELD_ADDRESS1 = 'address.address1'
USER_FIELD_ADDRESS2 = 'address.address2'
USER_FIELD_ZIPCODE = 'address.zipcode'
USER_FIELD_STATE = 'address.state'
USER_FIELD_COUNTRY = 'address.country'

DESCRIPTIONS = {
    # Account
    ACCOUNT_FIELD_COMPANY_NAME: "Company name",
    ACCOUNT_FIELD_ADDRESS_1: "Company address1",
    ACCOUNT_FIELD_ADDRESS_2: "Company address2",
    ACCOUNT_FIELD_CITY: "Company city",
    ACCOUNT_FIELD_ZIP_CODE: "Company zipcode",
    ACCOUNT_FIELD_COUNTRY: "Company country",
    ACCOUNT_FIELD_STATE: "Company state",
    ACCOUNT_FIELD_BILLING_CONTACT_FIRST_NAME: "Billing contact's first name",
    ACCOUNT_FIELD_BILLING_CONTACT_MIDDLE_NAME: "Billing contact's middle name",
    ACCOUNT_FIELD_BILLING_CONTACT_LAST_NAME: "Billing contact's last name",
    ACCOUNT_FIELD_BILLING_CONTACT_EMAIL: "Billing contact's email",
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_COUNTRY_CODE: "Billing contact's phone country code",
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_AREA_CODE: "Billing contact's phone area code",
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_NUM: "Billing contact's phone number",
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_EXT_NUM: "Billing contact's phone ext. number",
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_COUNTRY_CODE: "Billing contact's fax country code",
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_AREA_CODE: "Billing contact's fax area code",
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_NUM: "Billing contact's fax number",
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_EXT_NUM: "Billing contact's fax ext. number",
    ACCOUNT_FIELD_ADMIN_CONTACT_FIRST_NAME: "Admin contact's first name",
    ACCOUNT_FIELD_ADMIN_CONTACT_MIDDLE_NAME: "Admin contact's middle name",
    ACCOUNT_FIELD_ADMIN_CONTACT_LAST_NAME: "Admin contact's last name",
    ACCOUNT_FIELD_ADMIN_CONTACT_EMAIL: "Admin contact's email",
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_COUNTRY_CODE: "Admin contact's phone country code",
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_AREA_CODE: "Admin contact's phone area code",
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_NUM: "Admin contact's phone number",
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_EXT_NUM: "Admin contact's phone ext. number",
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_COUNTRY_CODE: "Admin contact's fax country code",
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_AREA_CODE: "Admin contact's fax area code",
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_NUM: "Admin contact's fax number",
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_EXT_NUM: "Admin contact's fax ext. number",
    ACCOUNT_FIELD_TECH_CONTACT_FIRST_NAME: "Tech contact's first name",
    ACCOUNT_FIELD_TECH_CONTACT_MIDDLE_NAME: "Tech contact's middle name",
    ACCOUNT_FIELD_TECH_CONTACT_LAST_NAME: "Tech contact's last name",
    ACCOUNT_FIELD_TECH_CONTACT_EMAIL: "Tech contact's email",
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_COUNTRY_CODE: "Tech contact's phone country code",
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_AREA_CODE: "Tech contact's phone area code",
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_NUM: "Tech contact's phone number",
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_EXT_NUM: "Tech contact's phone ext. number",
    ACCOUNT_FIELD_TECH_CONTACT_FAX_COUNTRY_CODE: "Tech contact's fax country code",
    ACCOUNT_FIELD_TECH_CONTACT_FAX_AREA_CODE: "Tech contact's fax area code",
    ACCOUNT_FIELD_TECH_CONTACT_FAX_NUM: "Tech contact's fax number",
    ACCOUNT_FIELD_TECH_CONTACT_FAX_EXT_NUM: "Tech contact's fax ext. numbers",
    # Staff member (user)
    USER_FIELD_LOGIN: "User's login",
    USER_FIELD_PASSWORD: "User's password",
    USER_FIELD_FIRST_NAME: "User's first name",
    USER_FIELD_LAST_NAME: "User's last name",
    USER_FIELD_MIDDLE_NAME: "User's middle name",
    USER_FIELD_TITLE: "User's title",
    USER_FIELD_PHONE_AREA_CODE: "User's phone area code",
    USER_FIELD_PHONE_EXT_NUM: "User's phone ext. number",
    USER_FIELD_PHONE_NUM: "User's phone number",
    USER_FIELD_PHONE_COUNTRY_CODE: "User's phone country code",
    USER_FIELD_FAX_AREA_CODE: "User's fax area code",
    USER_FIELD_FAX_EXT_NUM: "User's fax ext. number",
    USER_FIELD_FAX_NUM: "User's fax number",
    USER_FIELD_FAX_COUNTRY_CODE: "User's fax country code",
    USER_FIELD_EMAIL: "User's email",
    USER_FIELD_CITY: "User's city",
    USER_FIELD_ADDRESS1: "User's address1",
    USER_FIELD_ADDRESS2: "User's address2",
    USER_FIELD_ZIPCODE: "User's zipcode",
    USER_FIELD_STATE: "User's state",
    USER_FIELD_COUNTRY: "User's country"
}

REQUIRED_ACCOUNT_FIELDS = [
    ACCOUNT_FIELD_ADDRESS_1, ACCOUNT_FIELD_CITY, ACCOUNT_FIELD_ZIP_CODE, ACCOUNT_FIELD_COUNTRY,
    ACCOUNT_FIELD_BILLING_CONTACT_FIRST_NAME, ACCOUNT_FIELD_ADMIN_CONTACT_FIRST_NAME,
    ACCOUNT_FIELD_TECH_CONTACT_FIRST_NAME, ACCOUNT_FIELD_BILLING_CONTACT_EMAIL, ACCOUNT_FIELD_ADMIN_CONTACT_EMAIL,
    ACCOUNT_FIELD_TECH_CONTACT_EMAIL, ACCOUNT_FIELD_BILLING_CONTACT_PHONE_NUM, ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_NUM,
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_NUM
]

REQUIRED_USER_FIELDS = [
    USER_FIELD_LOGIN, USER_FIELD_FIRST_NAME, USER_FIELD_PHONE_NUM, USER_FIELD_EMAIL,
    USER_FIELD_CITY, USER_FIELD_ADDRESS1, USER_FIELD_ZIPCODE, USER_FIELD_COUNTRY
]

ACCOUNT_FIELDS_VALIDATION_RULES = {
    ACCOUNT_FIELD_COMPANY_NAME: {'max': MAX_COMPANY_NAME_LENGTH},
    ACCOUNT_FIELD_ADDRESS_1: {'max': MAX_ADDRESS_LENGTH},
    ACCOUNT_FIELD_ADDRESS_2: {'max': MAX_ADDRESS_LENGTH},
    ACCOUNT_FIELD_CITY: {'max': MAX_CITY_LENGTH},
    ACCOUNT_FIELD_ZIP_CODE: {'max': MAX_ZIPCODE_LENGTH},
    ACCOUNT_FIELD_COUNTRY: {'validator': 'country_validator'},
    ACCOUNT_FIELD_STATE: {'max': MAX_STATE_LENGTH},
    ACCOUNT_FIELD_BILLING_CONTACT_FIRST_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_ADMIN_CONTACT_FIRST_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_TECH_CONTACT_FIRST_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_BILLING_CONTACT_MIDDLE_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_ADMIN_CONTACT_MIDDLE_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_TECH_CONTACT_MIDDLE_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_BILLING_CONTACT_LAST_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_ADMIN_CONTACT_LAST_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_TECH_CONTACT_LAST_NAME: {'max': MAX_NAME_LENGTH},
    ACCOUNT_FIELD_BILLING_CONTACT_EMAIL: {'max': MAX_EMAIL_LENGTH, 'validator': 'email_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_EMAIL: {'max': MAX_EMAIL_LENGTH, 'validator': 'email_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_EMAIL: {'max': MAX_EMAIL_LENGTH, 'validator': 'email_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_FAX_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_FAX_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_TECH_CONTACT_FAX_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    ACCOUNT_FIELD_BILLING_CONTACT_PHONE_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    ACCOUNT_FIELD_BILLING_CONTACT_FAX_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    ACCOUNT_FIELD_ADMIN_CONTACT_PHONE_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    ACCOUNT_FIELD_ADMIN_CONTACT_FAX_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    ACCOUNT_FIELD_TECH_CONTACT_PHONE_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    ACCOUNT_FIELD_TECH_CONTACT_FAX_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH}
}

USER_FIELDS_VALIDATION_RULES = {
    # The 'unique' rule for the USER_FIELD_LOGIN field has dynamic value and is determined within pre-check execution
    USER_FIELD_LOGIN: {'max': MAX_LOGIN_LENGTH, 'validator': 'login_validator', 'min': MIN_LOGIN_LENGTH, 'unique': set()},
    USER_FIELD_PASSWORD: {'max': MAX_PASSWORD_LENGTH, 'validator': 'password_validator'},
    USER_FIELD_FIRST_NAME: {'max': MAX_NAME_LENGTH},
    USER_FIELD_LAST_NAME: {'max': MAX_NAME_LENGTH},
    USER_FIELD_MIDDLE_NAME: {'max': MAX_NAME_LENGTH},
    USER_FIELD_PHONE_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    USER_FIELD_PHONE_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    USER_FIELD_PHONE_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    USER_FIELD_PHONE_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    USER_FIELD_FAX_AREA_CODE: {'max': MAX_AREA_CODE_LENGTH, 'validator': 'int_validator'},
    USER_FIELD_FAX_EXT_NUM: {'max': MAX_EXTENSION_NUMBER_LENGTH},
    USER_FIELD_FAX_NUM: {'max': MAX_PHONE_NUMBER_LENGTH, 'validator': 'int_validator'},
    USER_FIELD_FAX_COUNTRY_CODE: {'max': MAX_COUNTRY_CODE_LENGTH, 'validator': 'int_validator'},
    USER_FIELD_EMAIL: {'max': MAX_EMAIL_LENGTH, 'validator': 'email_validator'},
    USER_FIELD_CITY: {'max': MAX_CITY_LENGTH},
    USER_FIELD_ADDRESS1: {'max': MAX_ADDRESS_LENGTH},
    USER_FIELD_ADDRESS2: {'max': MAX_ADDRESS_LENGTH},
    USER_FIELD_ZIPCODE: {'max': MAX_ZIPCODE_LENGTH},
    USER_FIELD_STATE: {'max': MAX_STATE_LENGTH},
    USER_FIELD_COUNTRY: {'validator': 'country_validator'}
}

CRITICAL_VALIDATION_RULES = list([
    {'rule_type': 'validator', 'rule_value': 'login_validator'},
    {'rule_type': 'validator', 'rule_value': 'email_validator'},
    {'rule_type': 'validator', 'rule_value': 'country_validator'},
    {'rule_type': 'validator', 'rule_value': 'password_validator'},
    {'rule_type': 'min', 'rule_value': MIN_LOGIN_LENGTH},
    {'rule_type': 'max', 'rule_value': MAX_EMAIL_LENGTH},
    {'rule_type': 'max', 'rule_value': MAX_LOGIN_LENGTH}
])

PRODUCT_TO_DOC_PARTITION_MAP = {
    SOURCE_PRODUCT_PLESK: '88377.htm#o89946',
    SOURCE_PRODUCT_OBAS: '88369.htm#o89929'
}

MAX_AMOUNT_OF_ACCOUNTS_TO_SHOW_IN_MSG = 5
MAX_AMOUNT_OF_USERS_TO_SHOW_IN_MSG = 5


class CheckSourceData(CheckAbstract):
    SUPPORTED_TARGET_SERVICE_CLASSES = [TargetServiceType.SHM, TargetServiceType.OACI]
    SUPPORTED_SOURCE_PRODUCTS = [SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS]

    def __init__(self):
        super(CheckSourceData, self).__init__(Checks.CHECK_SOURCE_DATA, CheckAbstract.PHASE_NO_PHASE)
        self.found_critical_validation_errors = False
        self.validation_error_level = self.STATUS_WARNING

    def is_check_supported(self):
        return super(CheckSourceData, self).is_check_supported() and \
            Registry.get_instance().get_context().target_panel_obj.is_billing_module_installed()

    def get_full_description(self):
        return _T('CHECK_FULL_DESCRIPTION_SOURCE_DATA')

    def _do_check(self):
        try:
            platform_reader = PlatformReaderDB()
            source_system_aps_id = platform_reader.aps_id
            # Check required fields on emptiness: accounts & staff members (users)
            error_message_parts_to_show = list()
            accounts = platform_reader.get_migration_required_accounts()
            logger.debug('Found accounts IDs (primary): {ids}'.format(ids=','.join(str(account.id) for account in accounts)))
            accounts_with_empty_fields = self.get_entities_with_empty_fields(accounts, REQUIRED_ACCOUNT_FIELDS)
            if accounts_with_empty_fields:
                error_message_parts_to_show.append(
                    _T('CHECK_SOURCE_DATA_ERROR_ACCOUNTS_MISSED_REQUIRED_FIELDS').format(
                        missed_fields_info=self.get_missed_fields_info_as_limited_string(
                            accounts_with_empty_fields,
                            MAX_AMOUNT_OF_ACCOUNTS_TO_SHOW_IN_MSG
                        ),
                        source_system_aps_id=source_system_aps_id,
                        precheck_doc_link=self.get_link_to_precheck_documentation()
                    )
                )

            users = platform_reader.get_migration_required_staff_members()
            logger.debug('Found staff members (users) IDs (primary): {ids}'.format(ids=','.join(str(user.id) for user in users)))
            users_with_empty_fields = self.get_entities_with_empty_fields(users, REQUIRED_USER_FIELDS)
            if users_with_empty_fields:
                error_message_parts_to_show.append(
                    _T('CHECK_SOURCE_DATA_ERROR_USERS_MISSED_REQUIRED_FIELDS').format(
                        missed_fields_info=self.get_missed_fields_info_as_limited_string(
                            users_with_empty_fields,
                            MAX_AMOUNT_OF_USERS_TO_SHOW_IN_MSG
                        ),
                        source_system_aps_id=source_system_aps_id,
                        precheck_doc_link=self.get_link_to_precheck_documentation()
                    )
                )

            if error_message_parts_to_show:
                return self.STATUS_ERROR, (HTML_EOL + HTML_EOL).join(error_message_parts_to_show)

            # Determine duplicate logins for users
            # Example of user_logins_counters = dict('dpotapoaccov@odin.com': 1, 'cyen.bkems@nonexist.com': 5)
            user_logins_counters = dict()
            for user in users:
                login = user.authentication.login
                if login not in user_logins_counters:
                    user_logins_counters[login] = 0

                user_logins_counters[login] += 1

            for login, counter in user_logins_counters.iteritems():
                if counter > 1:
                    USER_FIELDS_VALIDATION_RULES[USER_FIELD_LOGIN]['unique'].add(login)

            # Check all fields validity for accounts & staff members (users)
            error_message_parts_to_show = list()
            accounts_with_incorrect_fields = self.get_entities_with_incorrect_fields(
                accounts, ACCOUNT_FIELDS_VALIDATION_RULES
            )
            if accounts_with_incorrect_fields:
                if self.found_critical_validation_errors:
                    self.validation_error_level = self.STATUS_ERROR
                    recommendation_to_skip_this_error = ''
                else:
                    recommendation_to_skip_this_error = _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_CAN_SKIP')

                error_message_parts_to_show.append(
                    _T('CHECK_SOURCE_DATA_ERROR_ACCOUNTS_INCORRECT_FIELDS').format(
                        incorrect_fields_info=self.get_entities_with_incorrect_fields_as_limited_string(
                            accounts_with_incorrect_fields, MAX_AMOUNT_OF_ACCOUNTS_TO_SHOW_IN_MSG
                        ),
                        source_system_aps_id=source_system_aps_id,
                        recommendation_to_skip_this_error=recommendation_to_skip_this_error,
                        precheck_doc_link=self.get_link_to_precheck_documentation()
                    )
                )

            self.reset_validation_error_criticality()
            users_with_incorrect_fields = self.get_entities_with_incorrect_fields(users, USER_FIELDS_VALIDATION_RULES)
            if users_with_incorrect_fields:
                if self.found_critical_validation_errors:
                    recommendation_to_skip_this_error = ''
                    self.validation_error_level = self.STATUS_ERROR
                else:
                    recommendation_to_skip_this_error = _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_CAN_SKIP')

                error_message_parts_to_show.append(
                    _T('CHECK_SOURCE_DATA_ERROR_USERS_INCORRECT_FIELDS').format(
                        incorrect_fields_info=self.get_entities_with_incorrect_fields_as_limited_string(
                            users_with_incorrect_fields, MAX_AMOUNT_OF_USERS_TO_SHOW_IN_MSG
                        ),
                        source_system_aps_id=source_system_aps_id,
                        recommendation_to_skip_this_error=recommendation_to_skip_this_error,
                        precheck_doc_link=self.get_link_to_precheck_documentation()
                    )
                )

            if error_message_parts_to_show:
                return self.validation_error_level, (HTML_EOL + HTML_EOL).join(error_message_parts_to_show)

            return self.STATUS_SUCCESS, _T('CHECK_SOURCE_DATA_SUCCESS')
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return self.STATUS_ERROR, _T('CHECK_SOURCE_DATA_ERROR_UNEXPECTED')

    def get_entities_with_empty_fields(self, entities, required_fields):
        """
        :param entities: list[AccountModel]|list[StaffMemberModel]
        :param required_fields: list[string]
        :return: list[AccountModel]|list[StaffMemberModel]
        """
        entities_with_empty_fields = list()
        for entity in entities:
            for required_field in required_fields:
                field_value = entity.deep_get_attr(required_field)
                if isinstance(field_value, basestring):
                    field_value = field_value.strip()

                if not field_value:
                    if not hasattr(entity, 'empty_fields'):
                        entity.empty_fields = list()

                    entity.empty_fields.append(required_field)

            if hasattr(entity, 'empty_fields'):
                entities_with_empty_fields.append(entity)

        return entities_with_empty_fields

    def get_entities_with_incorrect_fields(self, entities, validation_rules):
        """
        :param entities: list[AccountModel]|list[StaffMemberModel]
        :param validation_rules: dict[dict[int|string]]
        :return: list[AccountModel]|list[StaffMemberModel]
        """
        entities_with_incorrect_fields = list()
        for entity in entities:
            entity.incorrect_fields = dict()
            for field_to_check, field_validation_rules in validation_rules.iteritems():
                field_value = entity.deep_get_attr(field_to_check)
                if isinstance(field_value, basestring):
                    field_value = field_value.strip()

                field_errors = list()
                if field_value:
                    for rule_type, rule_value in field_validation_rules.iteritems():
                        is_field_value_valid, validation_message = self.is_field_value_valid(
                            rule_type, rule_value, field_value, entity, field_to_check
                        )
                        if is_field_value_valid is False:
                            field_errors.append(validation_message)

                if field_errors:
                    entity.incorrect_fields[field_to_check] = field_errors

            if entity.incorrect_fields:
                entities_with_incorrect_fields.append(entity)

        return entities_with_incorrect_fields

    def reset_validation_error_criticality(self):
        self.found_critical_validation_errors = False

    def check_if_validation_error_is_critical(self, rule_type, rule_value):
        if {'rule_type': rule_type, 'rule_value': rule_value} in CRITICAL_VALIDATION_RULES:
            self.found_critical_validation_errors = True

    def is_field_value_valid(self, rule_type, rule_value, field_value, entity, field_to_check):
        """
        :param rule_type: string Examples: 'min', 'max', 'validator', 'unique'
        :param rule_value: string|int|set() Examples: 5, 100, 'int_validator', set(['not-unique-email-n@nov.net'])
        :param field_value: string|int Examples: 'dpotapov@odin.com', 'X1p82lH2Bd', 'Dmitriy', 'Green Street, 100 - 500'
        :param entity AccountModel|StaffMemberModel The checked entity
        :param field_to_check: string Example: 'info.company_name', 'info.admin_contact.middle_name', 'contact.email'
        :return: bool, string
        """
        if rule_type == 'max' and len(field_value) > rule_value:
            self.check_if_validation_error_is_critical(rule_type, rule_value)
            self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
            return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_MAX_LEN_EXCEEDED').format(
                max_length=rule_value
            )

        if rule_type == 'min' and len(field_value) < rule_value:
            self.check_if_validation_error_is_critical(rule_type, rule_value)
            self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
            return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_MIN_LEN_NOT_REACHED').format(
                min_length=rule_value
            )

        if rule_type == 'unique' and field_value in rule_value:
            self.check_if_validation_error_is_critical(rule_type, rule_value)
            self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
            return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_NOT_UNIQUE')

        if rule_type == 'validator':
            if rule_value == 'country_validator' and field_value not in COUNTRIES_LIST:
                self.check_if_validation_error_is_critical(rule_type, rule_value)
                self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
                return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_WRONG_COUNTRY')

            if rule_value == 'email_validator' and re.match(EMAIL_REGEXP, field_value) is None:
                self.check_if_validation_error_is_critical(rule_type, rule_value)
                self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
                return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_WRONG_EMAIL')

            if rule_value == 'int_validator' and re.match(NUMBER_REGEXP, field_value) is None:
                self.check_if_validation_error_is_critical(rule_type, rule_value)
                self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
                return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_FOUND_NOT_INT_CHARS')

            if rule_value == 'login_validator' and re.match(LOGIN_REGEXP, field_value) is None:
                self.check_if_validation_error_is_critical(rule_type, rule_value)
                self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
                return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_WRONG_LOGIN')

            if rule_value == 'password_validator':
                # Validate only non-empty password.
                # Empty password is ok. It will be generated during staff member creation
                if field_value and len(field_value) < MIN_PASSWORD_LENGTH:
                    self.check_if_validation_error_is_critical(rule_type, rule_value)
                    self.log_invalid_field(rule_type, rule_value, field_value, entity, field_to_check)
                    return False, _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELDS_PHRASE_MIN_LEN_NOT_REACHED').format(
                        min_length=MIN_PASSWORD_LENGTH
                    )

        return True, ''

    def log_invalid_field(self, rule_type, rule_value, field_value, entity, field_to_check):
        """
        :param rule_type: string Examples: 'min', 'max', 'validator', 'unique'
        :param rule_value: string|int|set() Examples: 5, 100, 'int_validator', set(['not-unique-email-n@nov.net'])
        :param field_value: string|int Examples: 'dpotapov@odin.com', 'X1p82lH2Bd', 'Dmitriy', 'Green Street, 100 - 500'
        :param entity AccountModel|StaffMemberModel The checked entity
        :param field_to_check: string Example: 'info.company_name', 'info.admin_contact.middle_name', 'contact.email'
        """
        if rule_type == 'max':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_MAX_LENGTH').format(
                max_length=rule_value
            )
        elif rule_type == 'min':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_MIN_LENGTH').format(
                min_length=rule_value
            )
        elif rule_type == 'unique':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_USERNAME_IS_NOT_UNIQUE').format(
                usernames=rule_value
            )
        elif rule_type == 'validator' and rule_value == 'country_validator':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_COUNTRY_ENUM').format(
                country=field_value, countries=COUNTRIES_LIST.keys()
            )
        elif rule_type == 'validator' and rule_value == 'email_validator':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_EMAIL_REGEXP').format(
                regexp=EMAIL_REGEXP, entity_type=entity.get_human_readable_full_id()
            )
        elif rule_type == 'validator' and rule_value == 'int_validator':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_INTEGER').format(
                regexp=NUMBER_REGEXP
            )
        elif rule_type == 'validator' and rule_value == 'login_validator':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_USERNAME_REGEXP').format(
                regexp=LOGIN_REGEXP
            )
        elif rule_type == 'validator' and rule_value == 'password_validator':
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_MIN_LENGTH').format(
                min_length=MIN_PASSWORD_LENGTH
            )
        else:
            field_validation_info = _T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_DEFAULT_DESCRIPTION').format(
                rule_type=rule_type, rule_value=rule_value
            )

        logger.debug(_T('CHECK_SOURCE_DATA_DEBUG_INCORRECT_FIELD_THE_MSG_HEAD').format(
            field_value=field_value if field_to_check != USER_FIELD_PASSWORD else '*******',
            field_name=self.get_field_description(field_to_check),
            entity_human_readable_identificators=entity.get_human_readable_full_id(),
            field_validation_info=field_validation_info
        ))

    def get_missed_fields_info_as_limited_string(self, entities, max_amount_to_show):
        """
        :param entities: list[AccountModel]|list[StaffMemberModel]
        :param max_amount_to_show: int The maximum amount of entities to show in message
        :return: string
        """
        error_message_content = HTML_EOL
        for entity in entities[:max_amount_to_show]:
            error_message_content += _T('CHECK_SOURCE_DATA_ERROR_ENTITY_WITH_MISSED_REQUIRED_FIELDS').format(
                entity_identificators=entity.get_human_readable_id(),
                empty_critical_fields=', '.join(
                    self.get_field_description(empty_field) for empty_field in entity.empty_fields
                )
            ) + HTML_EOL

        if len(entities) > max_amount_to_show:
            error_message_content += _T('CHECK_SOURCE_DATA_ERROR_PHRASE_OTHER_ENTITIES').format(
                amount_of_entities=len(entities) - max_amount_to_show
            ) + HTML_EOL

        return error_message_content

    def get_entities_with_incorrect_fields_as_limited_string(self, entities, max_amount_to_show):
        """
        :param entities: list[AccountModel]|list[StaffMemberModel]
        :param max_amount_to_show: int The maximum amount of entities to show in message
        :return: string
        """
        warning_message_content = HTML_EOL
        for entity in entities[:max_amount_to_show]:
            entity_errors = list()
            for field_key, field_errors in entity.incorrect_fields.iteritems():
                entity_errors.append(
                    _T('CHECK_SOURCE_DATA_ERROR_INCORRECT_FIELD_WITH_DETAILS').format(
                        field_name=self.get_field_description(field_key),
                        field_errors='; '.join(field_errors)
                    )
                )

            warning_message_content += _T('CHECK_SOURCE_DATA_ERROR_ENTITY_WITH_INCORRECT_FIELDS').format(
                entity_identificators=entity.get_human_readable_id(),
                incorrect_fields=', '.join(entity_errors)
            ) + HTML_EOL

        if len(entities) > max_amount_to_show:
            warning_message_content += _T('CHECK_SOURCE_DATA_ERROR_PHRASE_OTHER_ENTITIES').format(
                amount_of_entities=len(entities) - max_amount_to_show
            ) + HTML_EOL

        return warning_message_content

    def get_field_description(self, field_key):
        return "'%s'" % DESCRIPTIONS[field_key]

    def get_link_to_precheck_documentation(self):
        return PRODUCT_TO_DOC_PARTITION_MAP[self.source_type]
