# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import InternalError
from migrator.core.language import _T
from migrator.odin_automation.platform.checks.checks import Checks
from migrator.odin_automation.platform.checks.check_groups import CheckGroups

logger = logging.getLogger(__name__)


class CheckManager(object):
    def run_check(self, check_name):
        logger.info(_T('TRY_TO_RUN_MIGRATION_CHECK').format(check_name=check_name))
        self._check_if_check_name_exists(check_name)
        check = self._get_check_instance_by_name(check_name)
        return check.run()

    def is_check_supported(self, check_name):
        self._check_if_check_name_exists(check_name)
        check = self._get_check_instance_by_name(check_name)
        return check.is_check_supported()

    def get_supported_pre_checks_info(self):
        result = []
        supported_check_list = self._get_supported_pre_checks_list()
        supported_checks_by_groups = self._get_filtered_check_groups(supported_check_list)
        if not supported_checks_by_groups:
            return result

        for group_id, group_checks_list in supported_checks_by_groups.iteritems():
            group_checks_info = []
            for check_name in group_checks_list:
                check_info = {}
                check_instance = self._get_check_instance_by_name(check_name)
                check_info['check_name'] = check_name
                check_info['check_description'] = check_instance.get_full_description()
                group_checks_info.append(check_info)

            result.append(
                {
                    'group_id': group_id,
                    'group_title': self._get_group_title(group_id),
                    'group_order': self._get_group_order(group_id),
                    'group_checks': group_checks_info
                }
            )

        return result

    @staticmethod
    def _get_filtered_check_groups(checks_to_get=None):
        result = {}
        if not checks_to_get:
            return result

        all_check_groups = CheckGroups.check_names_to_groups_map()
        for group_key, group_checks in all_check_groups.iteritems():
            filtered_checks = []
            for possibly_supported_check in group_checks:
                if possibly_supported_check in checks_to_get:
                    filtered_checks.append(possibly_supported_check)

            if filtered_checks:
                result[group_key] = filtered_checks

        return result

    @staticmethod
    def _get_all_possible_check_names():
        result = []
        for _, group_checks in CheckGroups.check_names_to_groups_map().iteritems():
            result.extend(group_checks)

        return result

    def _check_if_check_name_exists(self, check_name):
        if check_name not in self._get_all_possible_check_names():
            raise InternalError('Unable to find the check "%s": an internal error occurred. Contact Odin technical support for assistance.' % check_name)

    def _get_check_class_by_name(self, check_name):
        self._check_if_check_name_exists(check_name)
        return Checks.check_names_to_main_classes_map()[check_name]

    def _get_supported_pre_checks_list(self):
        all_possible_check_names = self._get_all_possible_check_names()
        supported_pre_checks = []
        for check_name in all_possible_check_names:
            check = self._get_check_instance_by_name(check_name)
            if check.is_check_supported():
                supported_pre_checks.append(check_name)

        return supported_pre_checks

    @staticmethod
    def _get_group_title(group_id):
        try:
            return CheckGroups.groups_descriptions()[group_id]
        except:
            raise InternalError('Unable to determine an info for the group "%s": an internal error occurred. Contact Odin technical support for assistance.' % group_id)

    @staticmethod
    def _get_group_order(group_id):
        try:
            return CheckGroups.groups_order()[group_id]
        except:
            raise InternalError('Unable to determine an order for the group "%s": an internal error occurred. Contact Odin technical support for assistance.' % group_id)

    def _get_check_instance_by_name(self, check_name):
        self._check_if_check_name_exists(check_name)
        check_location = self._get_check_class_by_name(check_name)
        class_path, class_name = check_location.rsplit('.', 1)
        command_implementation_class = __import__(class_path, fromlist=[class_name]).__dict__[class_name]
        return command_implementation_class()
