# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import MigrationError
from migrator.core.session.config import read_ssh_auth
from migrator.core.utils.filetypes.config_utils import ConfigSection, HttpConnectionSettings, read_url
from migrator.core.utils.networking.ip_utils import is_ipv6
from migrator.products.plesk.target.config import TargetPleskConfigBase
from migrator.core.language import _T
from migrator.settings import UNIX_DEFAULT_SESSION_DIR, WINDOWS_DEFAULT_SESSION_DIR


POA_URL_DEFAULTS = dict(
    protocol='https',
    port=8440,
    path='/RPC2'
)


class TargetAutomationConfig(TargetPleskConfigBase):
    def __init__(
        self, ip, is_local, ssh_auth, poa_api, unix_session_dir, windows_session_dir,
        cligate_ip=None, webmail_ipv4=None, apache_restart_interval=None
    ):
        super(TargetAutomationConfig, self).__init__(
            ip, is_local, None, unix_session_dir, windows_session_dir, cligate_ip, apache_restart_interval
        )
        self._ssh_auth = ssh_auth
        self._poa_api = poa_api
        self._webmail_ipv4 = webmail_ipv4

    @property
    def ssh_auth(self):
        return self._ssh_auth

    @property
    def poa_api(self):
        return self._poa_api

    @property
    def is_local(self):
        """If target server is local or remote

        Local server means that migrator works on the target node, remote means
        that migrator's node and target node are different servers
        """
        # always false, local execution on Unix target node is not supported
        # yet, we always execute commands by SSH
        return self.ssh_auth is None


def read_automation_settings(config, section_name, default_apache_restart_interval='1800'):
    """
    :rtype: parallels.shm.config.TargetAutomationConfig
    """
    section = ConfigSection(config, section_name)

    ip = section['ip']
    is_local = section.getboolean('is-local', True)
    if is_ipv6(ip):
        raise MigrationError(
            _T("OA_API_DOES_NOT_WORK_OVER") % ip
        )
    cligate_ip = section.get('cligate-ip')

    ssh_auth = read_ssh_auth(section, required=False)

    poa_api = HttpConnectionSettings(url=read_url(
        section.prefixed('poa-').with_defaults(dict(POA_URL_DEFAULTS))
    ), auth=None)

    unix_session_dir = section.get('unix-session-dir', UNIX_DEFAULT_SESSION_DIR)
    windows_session_dir = section.get('windows-session-dir', WINDOWS_DEFAULT_SESSION_DIR)

    apache_restart_interval = section.get('apache-restart-interval', default_apache_restart_interval)

    return TargetAutomationConfig(
        ip=ip, is_local=is_local, ssh_auth=ssh_auth, poa_api=poa_api,
        unix_session_dir=unix_session_dir, windows_session_dir=windows_session_dir,
        cligate_ip=cligate_ip, apache_restart_interval=apache_restart_interval
    )
