# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import _T, MigrationError, InternalError
from migrator.core.data.model.subscription import has_any_of_service_types, SubscriptionMigrationStatus
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.data.source.xml.billing import BillingReaderXml, SourceSubscriptionStatus, \
    subscription_statuses_allowed_for_migration
from migrator.core.transfer.import_api_base import ImportApiBase
from migrator.core.utils.registry import Registry
from migrator.odin_automation.billing import is_subscription_nonrestorable
from migrator.odin_automation.billing.lib.api.subscription import \
    SubscriptionStatus, SubscriptionServiceStatus, SubscriptionStatusName
from migrator.odin_automation.platform.checks.lib.service_templates_general import \
    CheckServiceTemplates, ERROR_CODE_INACTIVE_ST_DETECTED, ERROR_CODE_RESOURCES_NOT_FOUND
from migrator.odin_automation.platform.lib.data.account_importer import \
    AccountImporterPBA, AccountImporterPPAB, AccountImporter
from migrator.odin_automation.platform.lib.data.subscription_importer import DomainSubscriptionImporterPBA, \
    HostingSubscriptionImporterPBA, DomainSubscriptionImporter, HostingSubscriptionImporter
from migrator.settings import SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass

logger = logging.getLogger(__name__)


class ImportApiOABase(ImportApiBase):
    def __init__(self, global_context, panel):
        self.global_context = global_context
        self.panel = panel
        self.billing_reader = None
        self._service_templates_check = CheckServiceTemplates()
        self._platform_reader = PlatformReaderDB()
        self._init_importers()

    def _init_importers(self):
        self.domain_subscription_importer = None
        self.hosting_subscription_importer = None
        self.account_importer = None

    def get_service_plan(self, service_plan_id):
        return self.hosting_subscription_importer.get_plan_info(service_plan_id)

    def get_subscription(self, subscription_id=None, account_id=None, plan_id=None):
        return self.hosting_subscription_importer.get_subscription(account_id, subscription_id, plan_id)

    def import_account(self, account, reseller_id):
        billing_info = self._get_account_biling_info(account.external_id)
        return self.account_importer.import_account(account, billing_info)

    def import_subscription(self, subscription, account_id, plan_id):
        """
        :param subscription: migrator.core.data.model.subscription.SubscriptionModel
        :param account_id: string Example: '1000001'
        :param plan_id: int Example: 1000004
        :return:
        """
        billing_info = self._get_subscription_biling_info(subscription.external_id)
        if billing_info and billing_info.status not in self.get_subscription_statuses_allowed_for_migration():
            raise MigrationError(_T("SUBSCRIPTION_IN_SOURCE_SYSTEM_IS_IN_UNSUPPORTED_STATE").format(
                state=billing_info.status
            ))

        if has_any_of_service_types(subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            subscription_id = self.domain_subscription_importer.import_subscription(
                subscription, account_id, plan_id, billing_info
            )
        else:
            subscription_id = self.hosting_subscription_importer.import_subscription(
                subscription, account_id, plan_id, billing_info
            )

        return subscription_id

    def _get_subscription_biling_info(self, subscription_id):
        raise NotImplementedError()

    def _get_account_biling_info(self, account_id):
        raise NotImplementedError()

    def _check_service_plan_compatibility(self, subscription):
        migration_item = self._platform_reader.get_migration_project_item(subscription.id)
        status, _ = self._service_templates_check.check_migration_items(self.global_context, [migration_item])
        if status == CheckServiceTemplates.STATUS_ERROR:
            code = self._service_templates_check.get_result_code()
            if code == ERROR_CODE_INACTIVE_ST_DETECTED:
                inactive_service_template_id = self._service_templates_check._get_migration_item_target_oa_st_id(
                    self.global_context,
                    migration_item
                )
                inactive_service_template = self._service_templates_check._get_service_templates([inactive_service_template_id])[0]
                raise MigrationError(_T('INACTIVE_ST_FOR_MIGRATED_SUBSCRIPTION').format(
                    st_name=inactive_service_template.get_identificators()
                ))
            elif code == ERROR_CODE_RESOURCES_NOT_FOUND:
                raise MigrationError(self._service_templates_check.get_result_meta_data())
            else:
                raise InternalError(_T('ST_PRE_CHECK_UNEXPECTED_CODE').format(code=code))

    def _check_source_subscription_status(self, subscription):
        if not self.billing_reader:
            self.billing_reader = BillingReaderXml(self.global_context.session_files.get_path_to_billing_xml_file())

        source_subscription_status = self.billing_reader.get_subscription(subscription.external_id).status
        if source_subscription_status not in self.get_subscription_statuses_allowed_for_migration():
            raise MigrationError(_T("SUBSCRIPTION_IN_SOURCE_SYSTEM_IS_IN_UNSUPPORTED_STATE").format(
                state=source_subscription_status
            ))

    def pre_migration_subscription_check(self, subscription):
        self._check_service_plan_compatibility(subscription)
        if not subscription.in_production:
            self._check_source_subscription_status(subscription)

    def get_subscription_statuses_allowed_for_migration(self):
        return subscription_statuses_allowed_for_migration

    def check_target_subscription_status(self, subscription_id):
        pass


class ImportApiOAPlatform(ImportApiOABase):
    def _init_importers(self):
        self.domain_subscription_importer = DomainSubscriptionImporter(self.panel)
        self.hosting_subscription_importer = HostingSubscriptionImporter(self.panel)
        self.account_importer = AccountImporter(self.panel)

    def update_subscription_migration_status(self, subscription, subscription_status):
        pass

    def _get_subscription_biling_info(self, subscription_id):
        return None

    def _get_account_biling_info(self, account_id):
        return None

    def switch_subscription_to_production(self, subscription):
        pass


class ImportApiOABilling(ImportApiOAPlatform):
    TEST_SUBSCRIPTION_STATUS_MAP = {
        SubscriptionMigrationStatus.NEW_SUBSCRIPTION: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.NOTEXIST),
        SubscriptionMigrationStatus.IN_PROGRESS: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.CREATING),
        SubscriptionMigrationStatus.FINISHED_OK: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_WARNINGS: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_NOTICES: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_ERRORS: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.NOTEXIST),
    }
    PRODUCTION_SUBSCRIPTION_STATUS_MAP = {
        SubscriptionMigrationStatus.NEW_SUBSCRIPTION: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.IN_PROGRESS: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_OK: (SubscriptionStatus.ACTIVE, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_WARNINGS: (SubscriptionStatus.ACTIVE, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_NOTICES: (SubscriptionStatus.ACTIVE, SubscriptionServiceStatus.ACTIVE),
        SubscriptionMigrationStatus.FINISHED_ERRORS: (SubscriptionStatus.ORDERED, SubscriptionServiceStatus.ACTIVE),
    }

    def _init_importers(self):
        self.domain_subscription_importer = DomainSubscriptionImporterPBA(self.panel)
        self.hosting_subscription_importer = HostingSubscriptionImporterPBA(self.panel)
        if self.global_context.target_panel_obj.is_billing_essentials():
            self.account_importer = AccountImporterPBA(self.panel)
        else:
            self.account_importer = AccountImporterPPAB(self.panel)

    def update_subscription_migration_status(self, subscription, subscription_status):
        try:
            global_context = Registry.get_instance().get_context()
            if subscription.oa_id and self.panel.is_billing_module_installed():
                api = self.panel.get_billing_api(global_context)
                billing_subscription = api.subscription.subscription_details_get(subscription.oa_id)
                if billing_subscription.status == SubscriptionStatus.ORDERED:
                    self._set_billing_subscription_status(api, subscription, subscription_status)
                elif is_subscription_nonrestorable(billing_subscription):
                    self._platform_reader.reset_subscription_oa_details(subscription.external_id)

        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise Exception(_T("FAILED_SUBSCRIPTION_STATUS_UPDATE_IN_BILLING").format(
                subscription=subscription.name, reason=str(e)
            ))

    def _set_billing_subscription_status(self, api, subscription, new_status):
        is_multiwebspace = self._is_multiwebspace_subscription(subscription)
        if self._is_oa_or_child_dns_subscription(subscription) and not is_multiwebspace:
            # Only set billing status for parent or multiwebspace subscriptions
            return

        if is_multiwebspace:
            new_status = self._get_multiwebspace_subscription_status(subscription, new_status)

        if not subscription.in_production:
            taget_billing_statuses = ImportApiOABilling.TEST_SUBSCRIPTION_STATUS_MAP[new_status]
        else:
            taget_billing_statuses = ImportApiOABilling.PRODUCTION_SUBSCRIPTION_STATUS_MAP[new_status]

        # set graced status for graced subscriptions
        if subscription.in_production and not is_multiwebspace and new_status in [
            SubscriptionMigrationStatus.FINISHED_OK, SubscriptionMigrationStatus.FINISHED_WARNINGS,
            SubscriptionMigrationStatus.FINISHED_NOTICES
        ]:
            source_subscription_info = self._get_subscription_biling_info(subscription.external_id)
            if source_subscription_info.status == SourceSubscriptionStatus.STATUS_GRACED:
                taget_billing_statuses = (SubscriptionStatus.GRACED, SubscriptionServiceStatus.ACTIVE)

        api.subscription.subscription_status_update(subscription.oa_id, *taget_billing_statuses)

    def _get_subscription_biling_info(self, subscription_id):
        if not self.billing_reader:
            self.billing_reader = BillingReaderXml(self.global_context.session_files.get_path_to_billing_xml_file())

        return self.billing_reader.get_subscription(subscription_id)

    def _get_account_biling_info(self, account_id):
        if not self.billing_reader:
            self.billing_reader = BillingReaderXml(self.global_context.session_files.get_path_to_billing_xml_file())

        return self.billing_reader.get_account(account_id)

    def switch_subscription_to_production(self, subscription):
        self.global_context.source_product.suspend_subscription(subscription)

    def check_target_subscription_status(self, subscription_id):
        oa_subscription = self.get_subscription(subscription_id)
        if oa_subscription.status in (
            SubscriptionStatus.EXPIRED, SubscriptionStatus.CANCELED, SubscriptionStatus.TERMINATED,
            SubscriptionStatus.ADMINISTRATIVE_HOLD, SubscriptionStatus.CREDIT_HOLD,
            SubscriptionStatus.CREDIT_AND_ADMINISTRATIVE_HOLD, SubscriptionStatus.TRIAL
        ):
            raise MigrationError(_T("SUBSCRIPTION_IN_OA_IS_IN_UNSUPPORTED_STATE").format(
                oa_subscription_name=oa_subscription.name,
                oa_subscription_id=oa_subscription.id,
                state=SubscriptionStatusName[oa_subscription.status]
            ))

    def _is_oa_or_child_dns_subscription(self, subscription):
        for item in self.global_context.migration_list_data:
            if item.id == subscription.external_id:
                return (item.is_oa_subscription and str(item.mapping_subscription_id) != str(subscription.oa_id)) or \
                       item.id != item.mapping_subscription_id and has_any_of_service_types(
                           subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]
                       )

        raise MigrationError(_T("INTERNAL_ERROR_OCCURRED_WHILE_CREATING_SUBSCRIPTIONS"))

    def _is_multiwebspace_subscription(self, subscription):
        # A multiwebspace subscription is a subscription that has at least one non-dns subscription mapped to it or a
        # hosting subscription that is mapped to another subscription
        if has_any_of_service_types(subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            return False

        return self._get_parent_multiwebspace_subscription(subscription) is not None or \
               len(self._get_child_multiwebspace_subscriptions(subscription)) > 0

    def _get_parent_multiwebspace_subscription(self, subscription):
        platform_reader = PlatformReaderDB()
        for item in self.global_context.migration_list_data:
            if item.id == subscription.external_id and not item.is_oa_subscription and \
                            item.mapping_subscription_id != subscription.external_id:
                return platform_reader.get_subscription(item.mapping_subscription_id)

        return None

    def _get_child_multiwebspace_subscriptions(self, subscription):
        child_subscriptions = []
        platform_reader = PlatformReaderDB()
        for item in self.global_context.migration_list_data:
            if not item.is_oa_subscription and item.mapping_subscription_id == subscription.external_id and \
                            item.id != subscription.external_id:
                child_subscription = platform_reader.get_subscription(item.id)
                if not has_any_of_service_types(child_subscription, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
                    child_subscriptions.append(child_subscription)

        return child_subscriptions

    def _get_multiwebspace_subscription_status(self, subscription, new_status):
        multiwebspace_subscriptions = self._get_child_multiwebspace_subscriptions(subscription)
        parent_subscription = self._get_parent_multiwebspace_subscription(subscription)
        if parent_subscription:
            multiwebspace_subscriptions = [parent_subscription]

        progress = self.global_context.progress
        statuses = [new_status]
        statuses += [progress.get_subscription(sub.external_id).status for sub in multiwebspace_subscriptions]
        for status in statuses:
            if status in (SubscriptionMigrationStatus.IN_PROGRESS, SubscriptionMigrationStatus.NEW_SUBSCRIPTION):
                return SubscriptionMigrationStatus.IN_PROGRESS

        for status in statuses:
            if status in (
                    SubscriptionMigrationStatus.FINISHED_OK,
                    SubscriptionMigrationStatus.FINISHED_WARNINGS,
                    SubscriptionMigrationStatus.FINISHED_NOTICES
            ):
                return status

        return SubscriptionMigrationStatus.FINISHED_ERRORS
