# (C)2017 Ingram Micro Inc. all Rights Reserved
import types

from migrator.core import InternalError
from migrator.core.data.model.dns import \
    DNS_RESOURCE_RECORD_MX, DNS_RESOURCE_RECORD_A, DNS_RESOURCE_RECORD_AAAA, DNS_RESOURCE_RECORD_CNAME, \
    DNS_RESOURCE_RECORD_SRV, DNS_RESOURCE_RECORD_TXT
from migrator.core.utils import cached, urlquote
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.dns import convert_to_national_domain_name
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.aps.account import get_account_by_subscription, get_account
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import get_aps_connector, APS2
from migrator.odin_automation.platform.lib.api.aps.subscription import get_subscription
from migrator.odin_automation.platform.lib.api.aps.types import \
    PADnsARecord, PADnsAAAARecord, PADnsCNAMERecord, PADnsMXRecord, PADnsSRVRecord, PADnsTXTRecord


ACCOUNT_TYPE_ADMIN = "ADMIN"


@cached
def get_dns_record_type(record_aps_type):
    """
    :param record_aps_type: str Example: 'http://parallels.com/aps/types/pa/dns/record/aaaa/1.0'
    :return: str 'aaaa', 'mx'
    """
    DNS_RECORD_APS_TYPE_MAP = {
        PADnsARecord: DNS_RESOURCE_RECORD_A,
        PADnsAAAARecord: DNS_RESOURCE_RECORD_AAAA,
        PADnsCNAMERecord: DNS_RESOURCE_RECORD_CNAME,
        PADnsMXRecord: DNS_RESOURCE_RECORD_MX,
        PADnsSRVRecord: DNS_RESOURCE_RECORD_SRV,
        PADnsTXTRecord: DNS_RESOURCE_RECORD_TXT
    }

    if record_aps_type not in DNS_RECORD_APS_TYPE_MAP:
        raise InternalError('Unable to determine a type of the DNS resource record "%s"' % record_aps_type)

    return DNS_RECORD_APS_TYPE_MAP[record_aps_type]


@cached
def get_domain_manager_uid(subscription_id):
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(filter_str='implementing(%s)' % types.DomainManagement)['aps']['id']


@trace
def get_pa_domain_by_name(pa_domain_name, subscription_id=None, account_id=None):
    # We cannot search domains by name by means comparison of string,
    # because there are situation when national domains with different unicode string have the same punicode

    poa_api = Registry.get_instance().get_context().conn.target.poa_api()
    domain = poa_api.domain.get_domain_by_name(pa_domain_name)
    if not domain:
        return None

    if subscription_id is not None:
        account_id = get_account_by_subscription(subscription_id)['id']
        aps = APS2.get_connector_for_account_user(account_id)
    elif account_id is not None:
        aps = APS2.get_connector_for_account_user(account_id)
    else:
        aps = APS2.get_connector()

    return aps.get_first_resource(
        filter_str='implementing(%s),eq(zoneId,%s),select(hosting,registrationInfo,parentDomain)' %
        (types.PAHostedDomain, domain['domain_id'])
    )


def is_pa_domain_owned_by_provider(pa_domain_name):
    """
    :param pa_domain_name: string Example: u'domain-1.com'
    :return: bool|None None - domain not found, True/False - is domain owned by OA administrator
    """
    aps = APS2.get_connector()
    domain_owner = aps.get_first_resource(filter_str='implementing(%s),like(name,%s),select(account)' % (
        types.PAHostedDomain, urlquote(pa_domain_name.lower())
    ))

    return domain_owner['account']['type'].lower() == ACCOUNT_TYPE_ADMIN.lower() if domain_owner is not None else None


@trace
def get_parent_pa_domain(pa_domain_name, subscription_id=None, account_id=None):
    domain = get_pa_domain_by_name(pa_domain_name, subscription_id, account_id)
    return domain.get('parentDomain', None)


@trace
def get_pa_domain_records(pa_domain_name, record_type=None, record_source=None):
    """
    :param pa_domain_name: str Example: u'domain.com'
    :param record_type: str Example: 'http://parallels.com/aps/types/pa/dns/record/a/1.0'
    :param record_source: str Example: u'sub'
    :return: list[]
    """
    pa_domain = get_pa_domain_by_name(pa_domain_name=pa_domain_name)
    if not pa_domain:
        return []

    filters = []
    if record_type is not None:
        filters.append('implementing(%s)' % record_type)

    if record_source is not None:
        filters.append('eq(source,%s)' % record_source.encode('idna'))

    filter_str = None if not filters else ','.join(filters)
    aps = APS2.get_connector()
    return aps.get_resources(resource_path='%s/records/' % pa_domain['aps']['id'], filter_str=filter_str)


@trace
def create_pa_domain(account_id, hosting_subscription_id, domain_name, parent_domain_id=None):
    """
    Create a PA domain with PA DNS zone.

    If a PA domain with this name already exists, return the PA domain ID for this domain.
    If parent_domain_id is specified, create a subdomain of the parent domain.
    """
    account = get_account(account_id)
    account_uid = account['aps']['id']
    aps = APS2.get_connector_for_account_user(account_id)
    json_data = {
        'aps': {'type': types.PAHostedDomain},
        'name': convert_to_national_domain_name(domain_name),
        'account': {'aps': {'id': account_uid}}
    }

    if parent_domain_id is not None:
        json_data['parentDomain'] = {
            'aps': {'id': parent_domain_id}
        }

    pa_dns_subscription_uid = get_subscription(hosting_subscription_id)['aps']['id']
    json_data['hosting'] = {
        'aps': {'id': pa_dns_subscription_uid}
    }

    return aps.post_resource('', '', json_data)


@trace
def _create_dns_record(account_id, subscription_id, resource):
    if subscription_id is not None:
        # to create 'system' DNS record
        aps = APS2.get_connector(account_id=account_id, subscription_id=subscription_id)
    else:
        # to create 'manual' DNS record
        aps = APS2.get_connector_for_account_user(account_id)

    return aps.post_resource('', '', resource)


@trace
def _get_base_record(record_type, domain_id, prefix=None, ttl=None, active=True):
    resource = {
        'aps': {'type': record_type},
        'zone': {'aps': {'id': domain_id}},
        'source': convert_to_national_domain_name(prefix),
        'RRState': 'active' if active else 'inactive'
    }

    if ttl is not None:
        resource['TTL'] = ttl

    return resource


@trace
def update_zone_soa(account_id, domain_id, admin_email=None, refresh=None, retry=None, expire=None, minimum_ttl=None):
    zone = {}
    resource = {
        'aps': {'id': domain_id},
        'zone': zone
    }

    if admin_email:
        zone['adminEmail'] = admin_email

    if refresh:
        zone['refresh'] = refresh

    if retry:
        zone['retry'] = retry

    if expire:
        zone['expire'] = expire

    if minimum_ttl:
        zone['TTL'] = minimum_ttl

    if len(zone) > 0:
        aps = APS2.get_connector(account_id, subscription_id=0)
        aps.put_resource(domain_id, '', resource)


@trace
def create_record_a(account_id, domain_id, prefix, address, ttl=None, active=True, subscription_id=None):
    resource = _get_base_record(PADnsARecord, domain_id, prefix, ttl, active)
    resource['address'] = address
    return _create_dns_record(account_id, subscription_id, resource)


@trace
def create_record_aaaa(account_id, domain_id, prefix, address, ttl=None, active=True, subscription_id=None):
    resource = _get_base_record(PADnsAAAARecord, domain_id, prefix, ttl, active)
    resource['address'] = address
    return _create_dns_record(account_id, subscription_id, resource)


@trace
def create_record_cname(account_id, domain_id, prefix, cname, ttl=None, active=True, subscription_id=None):
    resource = _get_base_record(PADnsCNAMERecord, domain_id, prefix, ttl, active)
    resource['cname'] = convert_to_national_domain_name(cname)
    return _create_dns_record(account_id, subscription_id, resource)


@trace
def create_record_mx(account_id, domain_id, prefix, exchange, priority, ttl=None, active=True, subscription_id=None):
    resource = _get_base_record(PADnsMXRecord, domain_id, prefix, ttl, active)
    resource['exchange'] = convert_to_national_domain_name(exchange)
    resource['priority'] = priority
    return _create_dns_record(account_id, subscription_id, resource)


@trace
def create_record_srv(account_id, domain_id, prefix, priority, weight, port, target, ttl=None, active=True, subscription_id=None):
    resource = _get_base_record(PADnsSRVRecord, domain_id, prefix, ttl, active)
    resource['priority'] = priority
    resource['weight'] = weight
    resource['port'] = port
    resource['target'] = convert_to_national_domain_name(target)
    return _create_dns_record(account_id, subscription_id, resource)


@trace
def create_record_txt(account_id, domain_id, prefix, txt_data, ttl=None, active=True, subscription_id=None):
    resource = _get_base_record(PADnsTXTRecord, domain_id, prefix, ttl, active)
    resource['txt_data'] = txt_data
    return _create_dns_record(account_id, subscription_id, resource)


@trace
def enable_record(record_id):
    aps = APS2.get_connector()
    aps.put_resource(record_id, '', {'RRState': 'active'})


@trace
def disable_record(record_id):
    aps = APS2.get_connector()
    aps.put_resource(record_id, '', {'RRState': 'inactive'})


@trace
def delete_record(record_id):
    aps = APS2.get_connector()
    aps.delete_resource(record_id)
