# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.odin_automation.platform.lib.api.aps import types
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import \
    get_aps_connector_with_admin_token, get_connector_with_app_certificate
from migrator.odin_automation.platform.lib.api.aps_runtime.rest import RestException

logger = logging.getLogger(__name__)


class NotificationStatus(object):
    IN_PROGRESS = 'inProgress'
    READY = 'ready'
    ERROR = 'error'


class Notification(object):
    def __init__(self, account_id, package_id, context, message, status, details_message='', details_keys=None,
                 service='', full_details_message='', full_details_keys=None, link_more_text_message='', link_more=''):
        """
        :param account_id: str APS guid of account
        :param package_id: str APS guid of application
        :param context: str string by which event could be found and updated/deleted later
        :param message: str
        :param status: str One of constants of NotificationStatus
        :param details_message: str
        :param details_keys: dict
        :param service: str
        :param full_details_message: str
        :param full_details_keys: dict
        :param link_more_text_message: str
        """
        self.account_id = account_id
        self.package_id = package_id
        self.context = context
        self.message = message
        self.status = status
        self.details_message = details_message
        self.details_keys = details_keys if details_keys else {}
        self.service = service
        self.full_details_message = full_details_message
        self.full_details_keys = full_details_keys if full_details_keys else {}
        self.link_more_text_message = link_more_text_message
        self.link_more = link_more

    def get_data_for_aps_request(self):
        return {
            'message': {'message': self.message},
            'details': {'message': self.details_message, 'keys': self.details_keys},
            'status': self.status,
            'accountId': self.account_id,
            'packageId': self.package_id,
            'context': self.context,
            'service': self.service,
            'fullDetails': {'message': self.full_details_message, 'keys': self.full_details_keys},
            'linkMoreText': {'message': self.link_more_text_message},
            'linkMore': self.link_more
        }


def add_notification(notification, app_certificate):
    """
    :param notification: Notification
    :param app_certificate: str Path to application certificate which will be used in APS request
    """
    try:
        aps = get_connector_with_app_certificate(app_certificate)
        notification_manager_aps_id = _get_notification_manager_aps_id()
        if notification_manager_aps_id:
            return aps.post_resource(
                notification_manager_aps_id,
                'notifications',
                notification.get_data_for_aps_request()
            )
    except RestException as e:
        logger.debug("Exception:", exc_info=True)
        logger.error(_T("ERROR_OCCURRED_DURING_SENDING_NOTIFICATION").format(error=str(e)))
        return None


def update_notification(notification, notification_id, app_certificate):
    """
    :param notification: Notification
    :param notification_id: str
    :param app_certificate: str Path to application certificate which will be used in APS request
    """
    try:
        aps = get_connector_with_app_certificate(app_certificate)
        notification_manager_aps_id = _get_notification_manager_aps_id()
        if notification_manager_aps_id:
            return aps.put_resource(
                notification_manager_aps_id,
                'notifications/%s' % notification_id,
                notification.get_data_for_aps_request()
            )
    except RestException as e:
        logger.debug("Exception:", exc_info=True)
        logger.error(_T("ERROR_OCCURRED_DURING_SENDING_NOTIFICATION").format(error=str(e)))
        return None


def send_notification_if_not_exist(notification, app_certificate):
    """
    :param notification: Notification
    :param app_certificate: str Path to application certificate which will be used in APS request
    """
    stored_notification = get_notification_by_context(notification.context)
    if not stored_notification:
        return add_notification(notification, app_certificate)


def get_notifications_list_by_context(context):
    try:
        aps = get_aps_connector_with_admin_token()
        notification_manager_aps_id = _get_notification_manager_aps_id()
        if notification_manager_aps_id:
            return aps.get_resources(
                '%s/notifications/lastMessagesByContext' % notification_manager_aps_id,
                'context=%s' % context
            )
        else:
            return []
    except RestException as e:
        logger.debug("Exception:", exc_info=True)
        logger.error(_T("ERROR_OCCURRED_DURING_GETTING_NOTIFICATIONS").format(error=str(e)))
        return None


def get_notification_by_context(context):
    notifications = get_notifications_list_by_context(context)
    if notifications:
        return notifications[0]
    else:
        return None


@cached
def _get_notification_manager_aps_id():
    try:
        aps = get_aps_connector_with_admin_token()
        aps_resource = aps.get_first_resource(filter_str='implementing(%s)' % types.NotificationManager)
        return aps_resource.aps.id if aps_resource else None
    except RestException as e:
        logger.debug("Exception:", exc_info=True)
        logger.error(_T("ERROR_OCCURRED_DURING_GETTING_NOTIFICATION_MANAGER").format(error=str(e)))
        return None
