# (C)2017 Ingram Micro Inc. all Rights Reserved
import copy
import logging

from collections import namedtuple

from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import Proto, ControllerProxy
from migrator.odin_automation.platform.lib.api.aps_runtime.core.aps_resource import APSResource
from migrator.odin_automation.platform.lib.api.aps import types
from migrator.settings import MAX_APS_FILTER_STRING_LENGTH

logger = logging.getLogger(__name__)


def get_error_message(e):
    """
    Retrieve an error message from the body of an APS response.
    """
    body = e.body

    msg = ''
    if isinstance(body, dict):
        if 'message' in body and body['message']:
            msg = body['message']

        if 'details' in body and 'errorMessages' in body['details'] and body['details']['errorMessages']:
            msg = body['details']['errorMessages'][0]
    else:
        msg = body

    return msg


def get_error(e):
    """
    Retrieve an error type from the body of an APS response.
    """
    body = e.body

    error = ''
    if isinstance(body, dict):
        if 'details' in body and 'error' in body['details']:
            error = body['details']['error']
        elif 'error' in body and body['error']:
            error = body['error']
    else:
        error = body

    return error


def get_message_with_details(e):
    """
    Retrieve an error message with details from the body of an APS response.
    """
    body = e.body

    if isinstance(body, basestring):
        return body
    elif isinstance(body, dict) and 'message' in body:
        message = body['message']
        if 'details' in body and 'details' in body['details'] and 'details' in body['details']['details'] \
                and body['details']['details']['details']:
            details = body['details']['details']['details']
        elif 'details' in body and body['details']:
            details = body['details']
        else:
            return body

        for k, v in details.iteritems():
            if isinstance(k, basestring) and isinstance(v, basestring):
                message = message.replace(u'__%s__' % k, v)

        return message
    else:
        return ''


APS2TokenId = namedtuple('APS2TokenId', ('account_id', 'user_id', 'subscription_id', 'app_instance_id'))


class APS2(object):
    """
    APS 2 API client.
    For more information check http://doc.apsstandard.org/2.1/spec/rql/
    """

    def __init__(self, token_id, ssl_cert_file_path=None):
        self._apsc = ControllerProxy(token_id, Proto.getSession(), ssl_cert_file_path)

    @staticmethod
    @cached
    def get_connector(account_id=None, user_id=None, subscription_id=None, app_instance_id=None):
        key = APS2TokenId(account_id=account_id, subscription_id=subscription_id, user_id=user_id,
                          app_instance_id=app_instance_id)
        return APS2(token_id=key, ssl_cert_file_path=None)

    @staticmethod
    @cached
    def get_connector_with_app_certificate(certificate_file_path):
        return APS2(token_id=None, ssl_cert_file_path=certificate_file_path)

    @staticmethod
    def get_connector_for_account_user(account_id, subscription_id=0):
        """
        :param account_id: string|integer Examples: '1000001', 1000001
        :param subscription_id: integer
        :return:
        """
        aps = get_aps_connector_with_admin_token()
        aps_id = aps.get_first_resource(filter_str='implementing(%s),eq(id,%s)' % (types.OAAccount, account_id))
        users = []
        if aps_id:
            users = aps.get_resources(resource_path='%s/users' % aps_id['aps']['id'], filter_str='implementing(%s),eq(isAccountAdmin,true)' % types.PAAdminUser)

        for user in users:
            # Migrator expects that enabled user has no fields "disabled" and "locked" or fields set to false
            if ('locked' in user and user.locked) or ('disabled' in user and user.disabled):
                continue

            return APS2.get_connector(None, user.userId, subscription_id)

        return None

    def __getattr__(self, item):
        """
        Magic method to proxy method calls to self._apsc with RestException conversion to APSCTransportError
        Example:
            aps = APS2(poa_api, None)
            aps.getResources('/aps/2/applications')

        Note:
            Exceptions raised by aps.getIo().sendRequest() will not be handled this way!

        :type item: str | unicode
        :rtype: object
        """
        if hasattr(self._apsc, item):
            return getattr(self._apsc, item)
        else:
            raise AttributeError()

    def get_resource(self, resource_uid, hidelist_response=None):
        """
        :param resource_uid: str Resource uid
        :param hidelist_response: list of str
        """
        return self.getResource(resource_uid, hidelist_response=hidelist_response)

    def get_resources(
            self, resource_path='', filter_str=None, paged_parameter=None, paged_parameter_values=None,
            hidelist_response=None
    ):
        """
        Example of usage with paging:
        resources = aps.get_resources(
            filter_str=implementing(http://parallels.com/aps/types/pa/subscription/1.0),in({sub_ids}),
            paged_parameter='sub_ids',
            paged_parameter_values=['1', '2', '3', '4', u'1000', '50000', '900000']
        )
        The paging works transparently and the function will return all the requested items at once.

        :param resource_path: str
        :param filter_str: str
        :param paged_parameter: str
        :param paged_parameter_values: iterable
        :param hidelist_response: list of str
        """
        if resource_path:
            resource_path = self._apsc.getIo().resourcePath(resource_path)

        if paged_parameter and paged_parameter_values:
            return self._get_resources_paged(filter_str, resource_path, paged_parameter, paged_parameter_values)
        else:
            return self.getResources(filter_str, resource_path, hidelist_response)

    def _get_resources_paged(self, filter_str, resource_path, paged_parameter, paged_parameter_values):
        ret = []
        initial_length = len(filter_str.format(**{paged_parameter: ''}))
        page = []
        page_length = initial_length
        for value in paged_parameter_values:
            value = value.encode('utf-8') if isinstance(value, unicode) else value
            new_page_length = self._get_new_page_length(page, page_length, value)
            if new_page_length <= MAX_APS_FILTER_STRING_LENGTH or len(page) == 0:
                page.append(value)
                page_length = new_page_length
            else:
                ret.extend(self._get_resources_page(resource_path, filter_str, paged_parameter, page))
                page = [value]
                page_length = initial_length + len(value)

        if page:
            ret.extend(self._get_resources_page(resource_path, filter_str, paged_parameter, page))

        return ret

    def _get_resources_page(self, resource_path, filter_str, paged_parameter, page):
        paged_filter_str = filter_str.format(**{paged_parameter: ','.join(page)})
        filter_str_length = len(paged_filter_str)
        if filter_str_length > MAX_APS_FILTER_STRING_LENGTH:
            logger.warning(_T("APS_REQUEST_FILTER_STRING_EXCEEDED_MAXIMUM_LENGTH").format(
                length=filter_str_length,
                max_length=MAX_APS_FILTER_STRING_LENGTH
            ))

        return self.getResources(paged_filter_str, resource_path)

    def get_types(self, type_id=None):
        """
        :param type_id: str
        """

        return self.getTypes(rql_filter='implementing(%s)' % type_id)

    def _get_new_page_length(self, page, page_length, value):
        comma_length = 1 if len(page) > 0 else 0
        return page_length + len(value) + comma_length

    def get_resources_count(self, resource_type):
        header_key = 'CONTENT_RANGE'
        self.get_resources(filter_str='implementing(%s),limit(0,0)' % resource_type)
        # Example of expected value: "items */100500"
        content_range = self._apsc.getLastResponseHeader(header_key)
        if content_range is None:
            raise Exception(_T("APS_RESPONSE_HEADER_NOT_FOUND").format(header=header_key))

        return int(content_range.split('/')[-1])

    def get_first_resource(self, resource_path='', filter_str=None, hidelist_response=None):
        """
        :param resource_path: str Resource type of aps object
        :param filter_str: str
        :param hidelist_response: list of str
        """
        _filter_str = 'limit(0,1),%s' % filter_str if filter_str else 'limit(0,1)'
        resources = self.get_resources(resource_path, _filter_str, hidelist_response=hidelist_response)
        if resources:
            return resources[0]
        else:
            return None

    @staticmethod
    def _replace_resource_version(data):
        def _recursive(data):
            keys = []
            if isinstance(data, dict):
                keys = data.keys()

            elif isinstance(data, list):
                keys = range(0, len(data))

            for el in keys:
                if isinstance(data[el], APSResource):
                    data[el] = data[el].post
                else:
                    _recursive(data[el])

        data = copy.deepcopy(data)
        _recursive(data)
        return data

    def post_resource(self, resource, path, data=None, hidelist_request=None):
        """
        :param resource: str resource to send request
        :param path: str method of resource who should be called
        :param data: dict request data
        :param hidelist_request: list of str
        """
        url = self._apsc.getIo().resourcePath(resource, path)
        return self._apsc.getIo().sendRequest(
            Proto.POST, url, self._replace_resource_version(data), hidelist_request=hidelist_request
        )

    def put_resource(self, resource, path, data=None, hidelist_request=None, hidelist_response=None):
        """
        :param resource: str resource to send request
        :param path: str method of resource who should be called
        :param data: dict request data
        :param hidelist_request: list of str
        :param hidelist_response: list of str
        """
        url = self._apsc.getIo().resourcePath(resource, path)
        return self._apsc.getIo().sendRequest(
            Proto.PUT, url, self._replace_resource_version(data), hidelist_request=hidelist_request, hidelist_response=hidelist_response
        )

    def register_resource(self, application_instance_id, path, data, hidelist_request=None):
        """
        :type application_instance_id: str | unicode
        :param path: path to send request to
        :type path: str | unicode
        :param dict data: request data
        """
        url = '/aps/2/applications/%s/%s' % (application_instance_id, path)
        return self._apsc.getIo().sendRequest(
            Proto.POST, url, self._replace_resource_version(data), hidelist_request=hidelist_request
        )

    def get_applications(self):
        """
        Get all applications instances.

        Note:
            There is no any filtering support for applications in APSC!
        """
        url = '/aps/2/applications'
        return self._apsc.getIo().sendRequest(Proto.GET, url)

    def delete_resource(self, resource_uid):
        """
        :param string resource_uid: Resource uid
        """
        return self.unprovisionResource(resource_uid)

    def link_resource(self, resource, relation_name, link_resource):
        """
        :type resource: str | dict
        :type relation_name: str
        :type link_resource: str | dict
        """
        return self._apsc.linkResource(resource, relation_name, link_resource)

    def update_resource(self, resource, service, hidelist_request=None, hidelist_response=None):
        """
        :type resource: dict
        :type service: str
        :param hidelist_request: list of str
        :param hidelist_response: list of str
        """
        return self._apsc.updateResource(resource, service, hidelist_request=hidelist_request,
                                         hidelist_response=hidelist_response)


# backward compatibility
def get_aps_connector(subscription_id):
    return APS2.get_connector(None, None, subscription_id)


# backward compatibility
def get_aps_connector_with_admin_token():
    return APS2.get_connector(1, None, 0)


def get_connector_with_app_certificate(certificate_file_path):
    return APS2.get_connector_with_app_certificate(certificate_file_path)


@cached
def get_app_instance_uid(app_type):
    aps = get_aps_connector_with_admin_token()
    for app_info in aps.get_applications():
        if app_info['aps']['type'] == app_type:
            return app_info['aps']['id']

    return None
