# (C)2017 Ingram Micro Inc. all Rights Reserved
import cStringIO
import copy
import logging
import os
import pycurl
import re
import uuid
import json

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.utils.loggers import secure_logger
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.aps_runtime.core.resource import ResourceBase
from migrator.odin_automation.platform.lib.api.aps_runtime.core.resource_proxy import ResourceProxy
from migrator.odin_automation.platform.lib.api.aps_runtime.rest import RestException, save_json_decode, save_json_encode
from migrator.odin_automation.platform.lib.api.aps_runtime.utils import AttrDict, json_decode
from migrator import settings

logger = logging.getLogger(__name__)


class APSCTransportError(MigrationError):
    def __init__(self, error_message, status, reason, body):
        super(APSCTransportError, self).__init__(error_message.format(status=status, reason=reason))
        self.status = status
        self.reason = reason
        self.body = body


def get_new_token(token_id):
    """Request a new API token"""
    global_context = Registry.get_instance().get_context()
    poa_api = global_context.conn.target.poa_api()
    # TODO: get rid of get_subscription_aps_token() call and use APS-Subscription-ID header instead
    if token_id.subscription_id is not None and token_id.account_id is None and token_id.user_id is None:
        token = poa_api.get_subscription_aps_token(token_id.subscription_id)
    elif token_id.account_id is not None and token_id.user_id is None:
        token = poa_api.get_account_aps_token(token_id.account_id, token_id.subscription_id)
    elif token_id.account_id is None and token_id.user_id is not None:
        token = poa_api.get_user_aps_token(token_id.user_id, token_id.subscription_id)
    elif token_id.app_instance_id is not None:
        token = poa_api.get_app_instance_token(token_id.app_instance_id)
    else:
        token = poa_api.get_account_aps_token()

    logger.debug(_T("APS_TOKEN_CREATED").format(token=token_id))
    return token


class Proto(object):
    """
    Sending Request to APSC
    """
    GET = 'GET'
    PUT = 'PUT'
    POST = 'POST'
    DELETE = 'DELETE'

    # APS HTTP headers
    APS_CONTROLLER_URI = 'APS-Controller-URI'
    APS_SESSION = 'APS-Transaction-ID'
    APS_SUBSCRIPTION_ID = 'APS-Subscription-ID'
    APS_RESOURCE_ID = 'APS-Resource-ID'
    APS_REQUEST_ID = 'APS-Request-ID'
    APS_TOKEN = 'APS-Token'

    PATH_SUBSCRIPTIONS = 'aps/subscriptions'
    RESOURCE_LOG_EVENTS = 'aps/log'

    timeout = 300

    def __init__(self, apsc, token_id):
        """
        :type apsc: migrator.odin_automation.platform.lib.api.aps_runtime.apsc.ControllerProxy
        """
        self._apsc = apsc
        self._token_id = token_id

    def applicationPath(self, service_path=None):
        """
        :type service_path: basestring
        :rtype: basestring
        """
        prefix = '{base_url}/application/'.format(base_url=self._apsc.getApsBaseUrlPath())
        if service_path:
            return '{prefix}{service_path}'.format(prefix=prefix, service_path=service_path)

        return prefix

    def typesPath(self, id=None):
        """
        :type id: basestring
        :rtype: basestring
        """
        prefix = '{base_url}/types'.format(base_url=self._apsc.getApsBaseUrlPath())
        if id:
            return '{prefix}?id={id}'.format(prefix=prefix, id=id)

        return prefix

    def packagesPath(self, id=None):
        """
        :type id: basestring
        :rtype: basestring
        """
        prefix = '{base_url}/packages'.format(base_url=self._apsc.getApsBaseUrlPath())
        if id:
            return '{prefix}/{id}'.format(prefix=prefix, id=id)

        return prefix

    def resourcePath(self, id=None, suffix=None):
        """
        :type id: basestring
        :type suffix: basestring
        :rtype: basestring
        """
        prefix = '{base_url}/resources'.format(base_url=self._apsc.getApsBaseUrlPath())
        if not id:
            return prefix

        if isinstance(id, ResourceBase):
            _id = id.aps.id
        else:
            _id = id

        if suffix:
            return '{prefix}/{id}/{suffix}'.format(prefix=prefix, id=_id, suffix=suffix)

        return '{prefix}/{id}'.format(prefix=prefix, id=_id)

    @staticmethod
    def getSession():
        """
        :rtype: basestring
        """
        return os.environ.get('HTTP_APS_TRANSACTION_ID')

    def getHeaders(self, content_type):
        """
        :type content_type: basestring
        :rtype: list
        """
        h = {
            Proto.APS_CONTROLLER_URI: self._apsc.getEndPoint(),
            Proto.APS_REQUEST_ID: uuid.uuid4(),
            'Content-type': content_type,
            'Connection': 'Keep-Alive',
            'Keep-Alive': Proto.timeout,
            'Accept': '*/*'
        }

        self._setHeaderIfDefined(h, Proto.APS_SESSION, self._apsc.getSession())
        self._setHeaderIfDefined(h, Proto.APS_TOKEN, self._apsc.getToken())
        self._setHeaderIfDefined(h, Proto.APS_RESOURCE_ID, self._apsc.getResourceId())

        h.update(self._apsc.getAdditionalHeaders())

        return ['{name}: {value}'.format(name=k, value=v) for k, v in h.iteritems()]

    def renew_token(self):
        """Replace the expired APS token with a newly obtained one"""
        new_token = get_new_token(self._token_id)
        self._apsc.setToken(new_token['aps_token'])
        self._apsc.setEndPoint(new_token['controller_uri'] or Proto.getControllerURI())
        logger.debug(_T("APS_TOKEN_RENEWED").format(token=self._token_id))

    def _aps2_method_error_wrapper(func):
        """A decorator that handles RestException errors and renews expired APS tokens"""

        def wrapper(*args, **kwargs):
            class_instance = args[0]
            logger.debug(_T("WRAPPED_APS2_METHOD").format(token_id=str(class_instance._token_id)))

            if not class_instance._apsc.getToken() and class_instance._token_id:
                class_instance.renew_token()

            try:
                return func(*args, **kwargs)
            except RestException as e:
                if 'Token expired' in e.message:
                    class_instance.renew_token()
                    return func(*args, **kwargs)
                else:
                    raise APSCTransportError(_T("APS_REQUEST_FAILED"), e.code, e.message, e.details or e.message)

        return wrapper

    @_aps2_method_error_wrapper
    def sendRequest(self, verb, path, body=None, content_type='application/json', hidelist_request=None,
                    hidelist_response=None):
        """
        :type verb: basestring
        :type path: basestring
        :type body: basestring | dict | None
        :type content_type: basestring
        :param hidelist_request: list of str
        :param hidelist_response: list of str
        :rtype: basestring
        """
        if self._apsc.getInstanceId():
            os.environ['HTTP_APS_INSTANCE_ID'] = self._apsc.getInstanceId()

        uri = '{base_url}{path}'.format(base_url=self._apsc.getEndPoint(), path=path)

        get_printable_censored_json = lambda body, hidelist: \
            save_json_encode(secure_logger.remove_sensitive_data_json(body, hidelist), indent=4) if body else None

        logger_debug_rest = lambda log_data: \
            logger.debug(u'\n'.join([u'{}{}'.format(x[0], x[1]) for x in log_data if x[1]]))

        ch = pycurl.Curl()
        try:
            ch.setopt(ch.HTTP_VERSION, ch.CURL_HTTP_VERSION_1_1)
            ch.setopt(ch.CONNECTTIMEOUT, Proto.timeout)
            ch.setopt(ch.TIMEOUT, Proto.timeout)
            ch.setopt(ch.HEADER, True)

            if uri.startswith('https'):
                # FIXME: SSLCERT, CAINFO, SSL_VERIFYPEER
                ch.setopt(ch.SSL_VERIFYHOST, False)
                ch.setopt(ch.SSL_VERIFYPEER, False)

            # Force curl to switch between GET and POST strategy because setting CUSTOMREQUEST is not enough
            if verb in (Proto.GET, Proto.DELETE):
                ch.setopt(ch.HTTPGET, True)
            else:
                ch.setopt(ch.POST, True)

            headers = self.getHeaders(content_type)
            if not self._apsc.getToken() and self._apsc.getSslCertFilePath():
                ch.setopt(pycurl.SSLCERT, self._apsc.getSslCertFilePath())

            ch.setopt(ch.HTTPHEADER, headers)
            ch.setopt(ch.URL, uri)
            ch.setopt(ch.CUSTOMREQUEST, verb)
            if body is not None:
                ch.setopt(ch.POSTFIELDS, save_json_encode(body).encode('utf-8'))
            else:
                # In case of PUT request, if POSTFIELDS is not set, curl may randomly set
                # the Content-Length header to -1
                ch.setopt(ch.POSTFIELDSIZE, 0)

            logger_debug_rest([
                ('REST request: ', '{} {}'.format(verb, uri)),
                ('Body:\n', get_printable_censored_json(body, hidelist_request)),
                ('Headers: ', [x for x in headers if 'APS-Token:' not in x]),
            ])

            try:
                http_code, response_headers_string, response_body = self._curlExec(ch)

                self._apsc.setLastResponseHeaders(self._parseHeaders(response_headers_string))
            except pycurl.error as e:
                # FIXME: check SSL cert error
                raise RestException(
                    500,
                    'Connection error: "{0}".\nWas trying to connect to "{1}".'.format(e.args[1], uri)
                )

            if http_code >= 300:
                logger_debug_rest([
                    ('REST response: ', http_code),
                    ('Body:\n', response_body.decode('utf-8')),
                    ('Headers:\n', response_headers_string.decode('utf-8')),
                ])
                if response_body.startswith('{'):
                    # Try to extract error if exists (JSON only)
                    error_obj = json_decode(response_body)
                    if isinstance(error_obj, AttrDict):
                        error_code = getattr(error_obj, 'code', http_code)
                        error_message = getattr(error_obj, 'type', '')
                        if error_message:
                            error_message += ': '
                        error_message += getattr(error_obj, 'message', '')

                        raise RestException(error_code, error_message, error_obj)
                else:
                    raise RestException(http_code, response_body)

        except RestException:
            raise
        except Exception as e:
            logger.debug("Exception:", exc_info=True)
            raise Exception('Request to APS controller failed: "{0}".'.format(str(e)))
        finally:
            ch.close()

        response = save_json_decode(response_body)

        logger_debug_rest([
            ('REST response: ', http_code),
            ('Body:\n', get_printable_censored_json(response, hidelist_response)),
            ('Headers:\n', response_headers_string.decode('utf-8')),
        ])

        return response

    @staticmethod
    def _setHeaderIfDefined(headers, name, value):
        if value:
            headers[name] = value

    @staticmethod
    def _curlExec(ch):
        headers = cStringIO.StringIO()
        body = cStringIO.StringIO()

        ch.setopt(ch.WRITEFUNCTION, body.write)
        ch.setopt(ch.HEADERFUNCTION, headers.write)

        try:
            ch.perform()
            # skip headers and return actual response payload only
            body.seek(ch.getinfo(ch.HEADER_SIZE))

            return ch.getinfo(ch.RESPONSE_CODE), headers.getvalue(), body.read()
        finally:
            body.close()
            headers.close()

    @staticmethod
    def _parseHeaders(headers_string):
        headers = dict()
        for line in headers_string.splitlines(False):
            if ':' not in line:
                continue

            name, value = line.split(':', 1)
            name = name.strip().upper().replace('-', '_')
            value = value.strip()

            if name not in ('CONTENT_TYPE', 'CONTENT_LENGTH', 'CONTENT_RANGE'):
                name = 'HTTP_{0}'.format(name)

            headers[name] = value

        return headers


class ControllerProxy(object):
    """
    APSC REST API implemetation
    """
    DEFAULT_APS_VERSION = '2'
    DEFAULT_VERSION_INFO = '''
{
    "2": {
        "version": "2.0.0",
        "href": "/aps/2",
        "features": {
            "navigation": {
                "vars": [ "filter" ]
            },
            "rql": [
                "and()", "or()", "not()", "eq()", "ne()", "ge()", "le()", "lt()", "gt()", "like()", "in()", "out()",
                "true", "false", "true()", "false()", "empty()", "null()", "select()", "limit()", "sort()",
                "implementing()", "composing()"
            ]
        }
    }
}
'''

    # FIXME: various log/resource log methods
    # FIXME: Event subscription functions

    def __init__(self, token_id, session=None, ssl_cert_file_path=None):
        """
        :type end_point: str
        :type session: basestring
        :type token: basestring
        """
        self._resource_id = None
        self._headers = dict()
        self._aps_major_version = None
        self._supported_aps_versions = None
        self._last_response_headers = dict()

        self._end_point = os.environ.get('HTTP_APS_CONTROLLER_URI') or settings.HTTP_APS_CONTROLLER_URI

        self._instance_id = os.environ.get('HTTP_APS_INSTANCE_ID')
        self._session = session
        self._io = Proto(self, token_id)
        self._token = None
        self._ssl_cert_file_path = ssl_cert_file_path
        self._aps_version = os.environ.get('HTTP_APS_VERSION') or self.DEFAULT_APS_VERSION

        if not self._aps_version or float(self._getApsMajorVersion(self._aps_version)) < 2.2:
            self._aps_version = self.DEFAULT_APS_VERSION

        self._setApsMajorVersion(self._aps_version)

    def getInstanceId(self):
        """
        :rtype: basestring
        """
        return self._instance_id

    def getSupportedApsVersions(self):
        """
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        """
        if not self._supported_aps_versions:
            try:
                result = self._io.sendRequest(Proto.GET, '/aps')
                self._supported_aps_versions = result.versions
            except Exception:
                return json_decode(ControllerProxy.DEFAULT_VERSION_INFO)

        return self._supported_aps_versions

    def setApsVersion(self, version):
        """
        :type version: basestring
        """
        self._aps_version = version
        if self._aps_version:
            self._setApsMajorVersion(version)
        else:
            self._aps_major_version = ControllerProxy.DEFAULT_APS_VERSION

    def getApsVersion(self):
        """
        :rtype: basestring
        """
        return self._aps_version

    def getApsMajorVersion(self):
        """
        :rtype: basestring
        """
        return self._aps_major_version

    def getApsBaseUrlPath(self):
        """
        :rtype: basestring
        """
        # FIXME:
        #if ( (float) \Rest\RestService::VERSION < 2.2) {
        #     return "aps/".self::DEFAULT_APS_VERSION;
        #}
        return 'aps/{0}'.format(self.getApsMajorVersion())

    def resetSession(self):
        self._session = None

    def getSession(self):
        """
        :rtype: basestring
        """
        return self._session

    def setEndPoint(self, end_point):
        """
        :rtype: basestring
        """
        self._end_point = end_point
        if not self._end_point:
            raise Exception("Can't determine controller end-point.")

        if not self._end_point.endswith('/'):
            self._end_point += '/'

    def getEndPoint(self):
        """
        :rtype: basestring
        """
        return self._end_point

    def setToken(self, token):
        """
        :type token: basestring
        """
        self._token = token

    def getToken(self):
        """
        :rtype: basestring
        """
        return self._token

    def getSslCertFilePath(self):
        """
        :rtype: basestring
        """
        return self._ssl_cert_file_path

    def setResourceId(self, resource_id):
        """
        :type resource_id: basestring
        """
        self._resource_id = resource_id

    def getResourceId(self):
        """
        :rtype: basestring
        """
        return self._resource_id

    def impersonate(self, resource):
        """
        :type resource: basestring | migrator.odin_automation.platform.lib.api.aps_runtime.core.resource.ResourceBase | migrator.odin_automation.platform.lib.api.aps_runtime.core.resource_proxy.ResourceProxy
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.apsc.ControllerProxy
        """
        apsc = copy.deepcopy(self)

        resource_id = resource.aps.id if isinstance(resource, (ResourceBase, ResourceProxy)) else resource
        apsc.setResourceId(resource_id)

        return apsc

    def setAdditionalHeaders(self, headers):
        """
        :type headers: dict
        """
        self._headers = headers

    def getAdditionalHeaders(self):
        """
        :rtype: dict
        """
        return self._headers

    def setLastResponseHeaders(self, headers=None):
        """
        :type headers: dict | none
        """
        self._last_response_headers = headers if headers else dict()

    def getLastResponseHeader(self, name):
        """
        :type name: basestring
        :rtype: basestring | None
        """
        return self._last_response_headers.get(name)

    def getTypes(self, type_id=None, rql_filter=None):
        """
        :type type_id: basestring
        :type rql_filter: basestring
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        """
        path = self._io.typesPath(type_id)
        if rql_filter:
            path += '?'
            path += rql_filter

        return self._io.sendRequest(Proto.GET, path)

    def getApplicationServices(self):
        """
        :rtype: basestring
        """
        return self._io.sendRequest(Proto.GET, self._io.applicationPath())

    def getPackageInfo(self, package_id):
        """
        :rtype: basestring
        """
        return self._io.sendRequest(Proto.GET, self._io.packagesPath(package_id))

    def getResources(self, rql_filter=None, path=None, hidelist_response=None):
        """
        :type rql_filter: basestring
        :type path: basestring
        :param hidelist_response: list of str
        :rtype: list
        """
        if not path:
            path = self._io.resourcePath()

        path += '?'
        if rql_filter:
            path += rql_filter

        # FIXME: apsc.php@getResources()
        return self._io.sendRequest(Proto.GET, path, hidelist_response=hidelist_response)

    def getIo(self):
        """
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.apsc.Proto
        """
        return self._io

    def getResource(self, resource_id, hidelist_response=None):
        """
        :type resource_id: basestring
        :type hidelist_response: list of str
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        :raises Exception: if no resource_id provided
        """
        if not resource_id:
            raise Exception('No resource_id provided to getResource.')

        # FIXME: apsc.php@getResource()
        return self._io.sendRequest(Proto.GET, self._io.resourcePath(resource_id), hidelist_response=hidelist_response)

    def linkResource(self, resource, relation_name, link_resource):
        """
        :type resource: basestring | migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict | migrator.odin_automation.platform.lib.api.aps_runtime.core.resource.ResourceBase
        :type relation_name: basestring
        :type link_resource: basestring | migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        :raises Exception: if no resource or link_resource provided
        """
        if isinstance(resource, basestring) and resource:
            resource_id = resource
        elif 'id' in resource.get('aps', {}):
            resource_id = resource.aps.id
        else:
            raise Exception('No resource_id of local end provided to link resources')

        if isinstance(link_resource, basestring) and link_resource:
            resource_str = {'aps': {'id': link_resource}}
        elif 'id' in link_resource.get('aps', {}):
            resource_str = json.dumps(link_resource)
        else:
            raise Exception('No resource_id of remote end provided to link resources')

        # FIXME: apsc.php@linkResource()
        return self._io.sendRequest(Proto.POST, self._io.resourcePath(resource_id, relation_name), resource_str)

    def unlinkResource(self, resource, relation_name, link_resource):
        # FIXME: apsc.php@unlinkResource()
        pass

    def provisionResource(self, resource):
        """
        :type resource: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict | migrator.odin_automation.platform.lib.api.aps_runtime.core.resource.ResourceBase
        :rtype: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        """
        # FIXME: apsc.php@provisionResource()
        return self._io.sendRequest(Proto.POST, self._io.resourcePath(), resource)

    def configureResource(self, resource):
        # FIXME: apsc.php@configureResource()
        pass

    def updateResource(self, resource, service, hidelist_request=None, hidelist_response=None):
        """
        Updates resource on APS bus
        Should be called after impersonalization under app APS Instance of service
        
        Note: implementation differs from PHP-Runtime because type-caching is not implemented
        
        :param resource: dict
        :param service: str service name
        :param hidelist_request: list of str
        :param hidelist_response: list of str
        :return: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict
        """

        if 'id' in resource.get('aps', {}):
            resource_id = resource['aps']['id']
        else:
            raise Exception('No mandatory resource_id provided to update resource')

        return self._io.sendRequest(Proto.PUT, self._io.applicationPath('%s/%s' % (service, resource_id)),
                                    hidelist_request=hidelist_request, hidelist_response=hidelist_response,
                                    body=resource)

    def registerResource(self, resource):
        # FIXME: apsc.php@registerResource()
        pass

    def unregisterResource(self, resource):
        # FIXME: apsc.php@unregisterResource()
        pass

    def unprovisionResource(self, resource):
        """
        :type resource: basestring | migrator.odin_automation.platform.lib.api.aps_runtime.core.resource.ResourceBase | migrator.odin_automation.platform.lib.api.aps_runtime.core.resource_proxy.ResourceProxy
        """
        resource_id = resource.aps.id if isinstance(resource, (ResourceBase, ResourceProxy)) else resource
        self._io.sendRequest(Proto.DELETE, self._io.resourcePath(resource_id))

    def _getApsMajorVersion(self, version):
        m = re.match(r'(\d+\.\d+)\..*', version)
        if m is not None and m.groups():
            return m.group(1)

        return version

    def _setApsMajorVersion(self, version):
        self._aps_major_version = self._getApsMajorVersion(version)
