# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import urllib
from ssl import SSLError
import xmlrpclib
from collections import defaultdict

from migrator.core.hosting.panel import model
from migrator.core.utils import obj
from migrator.core.utils.filetypes.yaml_utils import read_yaml, write_yaml
from migrator.core.utils.loggers.secure_logger import BleepingLogger
from migrator.core.utils.networking.xmlrpc import LoggingTransport, UnverifiedSafeTransport
from migrator.core.utils.lang.profiler import get_default_steps_profiler
from migrator.core.language import _T
from migrator.odin_automation.platform.lib.api.openapi.resource import Resource
from migrator.odin_automation.platform.lib.api.openapi.host import Host
from migrator.odin_automation.platform.lib.api.openapi.service import Service
from migrator.odin_automation.platform.lib.api.openapi.subscription import Subscription
from migrator.odin_automation.platform.lib.api.openapi.account import Account
from migrator.odin_automation.platform.lib.api.openapi.domain import Domain

RequestStatus = obj(SUCCESS=0, RUNNING=1, FAILED=2)

profiler = get_default_steps_profiler()


class OpenAPI(Resource):
    logger = logging.getLogger(__name__)

    poa_to_oa_statuses = {
        RequestStatus.SUCCESS: model.WebspaceStatus.ACTIVE,
        RequestStatus.RUNNING: model.WebspaceStatus.PENDING,
        RequestStatus.FAILED: model.WebspaceStatus.FAILED
    }

    def __init__(self, url):
        self.url = url
        self.proxy = self._get_proxy()

        self.host = Host(self.proxy, self._get_proxy)
        self.service = Service(self.proxy)
        self.subscription = Subscription(self.proxy)
        self.account = Account(self.proxy, self._get_proxy)
        self.domain = Domain(self.proxy)

    def get_request_status(self, request_id):
        assert request_id is not None, _T("INTERNAL_ERROR_MISSING_REQUEST_ID")

        req_status, status_messages = self.getRequestStatus(request_id)
        return req_status, u"\n".join(status_messages)

    def txn_begin(self, request_id):
        """
        Begin OpenAPI transaction.
        Return transaction id, to be used with txnCommit.
        """
        response = self.proxy.txn.Begin({'request_id': str(request_id)})
        return self._parse_response(response)['txn_id']

    def txn_commit(self, txn_id):
        """Commit OpenAPI transaction."""
        response = self.proxy.txn.Commit({'txn_id': txn_id})
        self._parse_response(response)

    def getRequestStatus(self, request_id):
        """
        Get request status.
        Return tuple(request_status, status_messages).
        """
        response = self.proxy.pem.getRequestStatus({'request_id': str(request_id)})
        result = self._parse_response(response)
        return result['request_status'], result['status_messages']

    def get_domain_name_servers(self, domain_name):
        response = self.proxy.pem.getDomainNameServers({'domain_name': domain_name})
        return self._parse_response(response)

    def set_resource_type_limits(self, subscription_id, limits):
        """limits - list of dicts with 'resource_type_id' and 'limit' keys"""
        args = {'subscription_id': subscription_id, 'resource_limits': limits}
        return self._parse_response(self.proxy.pem.setResourceTypeLimits(args))

    def get_resource_types_by_class(self, resclass_name):
        args = {'resclass_name': resclass_name, 'show_system_rt': True}
        return self._parse_response(self.proxy.pem.getResourceTypesByClass(args))

    def get_bm_xml_rpc_address(self):
        args = {}
        response = self.proxy.pem.billing.getBMXmlRpcAddress(args)
        return self._parse_response(response)

    def get_bm_api_user(self):
        args = {}
        response = self.proxy.pem.billing.getBMAPIUser(args)
        return self._parse_response(response)

    def get_bm_common_password(self):
        args = {}
        response = self._get_proxy(hidelist_response=['$.result']).pem.billing.getBMCommonPassword(args)
        return self._parse_response(response)

    def get_installed_modules(self):
        response = self.proxy.pem.packaging.listInstalledModules()
        return self._parse_response(response)

    def get_subscription_app_instances(self, subscription_id):
        args = {
            'subscription_id': subscription_id
        }
        response = self.proxy.pem.APS.getSubscriptionApplicationInstances(args)
        result = self._parse_response(response)
        return result[0]['application_instance_id']

    def get_app_instance_settings(self, app_instance_id):
        args = {
            'application_instance_id': app_instance_id
        }
        response = self.proxy.pem.APS.getApplicationInstanceSettings(args)
        result = defaultdict(list)
        for item in self._parse_response(response):
            result[item['name']].append(item['value'])

        return result

    def get_subscription_aps_token(self, subscription_id):
        args = {
            'subscription_id': subscription_id
        }
        response = self._get_proxy(hidelist_response=['$.result.aps_token']).pem.APS.getSubscriptionToken(args)
        return self._parse_response(response)

    def get_account_aps_token(self, account_id=1, subscription_id=0):
        args = {
            'account_id': account_id,
            'subscription_id': subscription_id
        }
        response = self._get_proxy(hidelist_response=['$.result.aps_token']).pem.APS.getAccountToken(args)
        return self._parse_response(response)

    def get_user_aps_token(self, user_id, subscription_id=0):
        args = {
            'user_id': user_id,
            'subscription_id': subscription_id
        }
        response = self._get_proxy(hidelist_response=['$.result.aps_token']).pem.APS.getUserToken(args)
        return self._parse_response(response)

    def get_app_instance_token(self, app_instance_id):
        args = {
            'application_instance_id': app_instance_id
        }
        response = self._get_proxy(hidelist_response=['$.result.aps_token']).pem.APS.getApplicationInstanceToken(args)
        return self._parse_response(response)

    def account_added_to_pba(self, account_id):
        self.proxy.pem.billing.accountAddedToPBA({'account_id': account_id})

    def _get_proxy(self, hidelist_request=None, hidelist_response=None):
        proto, _ = urllib.splittype(self.url)

        if proto == 'https':
            transport = UnverifiedSafeTransport(use_datetime=0)
        elif proto == 'http':
            transport = xmlrpclib.Transport(use_datetime=0)
        else:
            assert False, _T("ONLY_HTTP_AND_HTTPS_PROTOCOLS_ARE_SUPPORTED")

        logger = BleepingLogger(self.logger)
        transport = LoggingTransport(transport, logger, hidelist_request, hidelist_response)
        transport = PoaApiSettingsExceptionTransport(transport, self.logger)

        return xmlrpclib.ServerProxy(self.url, transport=transport)


class WebspacesRequestsHolder(object):
    def __init__(self, filename):
        self.filename = filename
        self.webspaces_requests = read_yaml(self.filename, True, {})

    def get(self, webspace_id):
        return self.webspaces_requests.get(webspace_id)

    def put(self, webspace_id, request_id):
        self.webspaces_requests[webspace_id] = request_id
        write_yaml(self.filename, self.webspaces_requests)


class PoaApiSettingsExceptionTransport(object):
    """
    Transport wrapper class to handle common mistake when you have migration/moving tools running
    on another host than Odin Automation management node, but have not configured Odin Automation Public API properly.
    It just catches all SSLError exceptions, and throws an exception with comprehensive instructions what to do.
    """
    def __init__(self, transport, logger):
        self.transport = transport
        self.logger = logger

    def request(self, host, handler, request_body, verbose=0):
        try:
            # parse method name in quick way for profiling
            method_name = 'unknown'
            method_name_start = request_body.find('<methodName>')
            if method_name_start != -1:
                method_name_end = request_body.find('</methodName')
                if method_name_end != -1:
                    method_name = request_body[
                        method_name_start + len('<methodName>'):method_name_end
                    ]

            with profiler.measure_api_call(host, 'POA API', method_name):
                return self.transport.request(host, handler, request_body, verbose)
        except SSLError as e:
            self.logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            self.logger.error(_T("FAILED_TO_CALL_OA_PUBLIC_API_METHOD"), e)
            raise Exception(_T("INTERNAL_ERROR_OCCURRED_DURING_CALL_OA_API"))

    def close(self):
        self.transport.close()
