# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple

from migrator.core.hosting.panel import model
from migrator.core.language import _T
from migrator.core.utils import obj, generate_random_password
from migrator.settings import OA_ADMIN_ID, OA_TARGET_OWNER_ID
from migrator.odin_automation.platform.lib.api.openapi.resource import Resource, PoaApiError
from migrator.odin_automation.platform.lib.db import get_system_locales


AccountInfo = namedtuple('AccountInfo',
                         ('account_id', 'account_type', 'language_code', 'company_name', 'admin_contact',
                          'tech_contact', 'billing_contact', 'address', 'staff_members')
                         )
ContactInfo = namedtuple('ContactInfo', ('first_name', 'middle_name', 'last_name', 'phone', 'fax', 'email'))
AddressInfo = namedtuple('AddressInfo',
                         ('city', 'street_name', 'address2', 'zipcode', 'country', 'state')
                         )
PhoneInfo = namedtuple('PhoneInfo', ('country_code', 'area_code', 'phone_num', 'ext_num'))
FaxInfo = namedtuple('FaxInfo', ('country_code', 'area_code', 'phone_num', 'ext_num'))
MemberInfo = namedtuple('MemberInfo', ('account_id', 'member_id', 'auth', 'contact', 'address'))
AuthInfo = namedtuple('AuthInfo', ('login', 'password'))

AccountMemberIds = namedtuple('AccountMemberIds', ('account_id', 'user_id'))
AccountTypes = obj(ADMIN='A', RESELLER='R', CLIENT='C')

PersonInfo = namedtuple('PersonInfo', ('first_name', 'middle_name', 'last_name', 'company_name'))
LocaleInfo = namedtuple('LocaleInfo', ('country_code', 'language_code', 'variant'))


class Account(Resource):
    def __init__(self, proxy, get_proxy):
        self.proxy = proxy
        self._get_proxy = get_proxy

    def create_account(self, account):
        return self._create_account(AccountTypes.CLIENT, OA_TARGET_OWNER_ID, account)

    def add_staff_member(self, account_id, staff_member_info):
        person = self._get_staff_member_person_info(staff_member_info.contact)
        address = self._get_address_info(staff_member_info.address)
        phone = self._get_phone_info(staff_member_info.contact.phone)
        fax = self._get_fax_info(staff_member_info.contact.fax)
        locale = self._get_locale_info(staff_member_info.language)

        return self.add_account_member(
            account_id, staff_member_info.authentication, person, address, phone, staff_member_info.contact.email, fax,
            locale
        )

    def add_service_user(self, account_id, first_name, login, password):
        args = {
            'owner_account_id': account_id,
            'login': login,
            'password': password,
            'check_password_quality': False,
            'properties': [
                {
                    'name': 'first_name',
                    'value': first_name
                }, {
                    'name': 'email',
                    'value': login
                }
            ]
        }

        response = self.proxy.pem.addUser(args)

        result = self._parse_response(response)
        return result['user_id']

    def get_staff_member_by_login(self, login):
        try:
            return self.get_account_member_by_login(login)
        except PoaApiError as e:
            if e.module_id == 'AccountManagement' and e.extype_id == 41:
                # No such member with name
                return None

            raise e

    def create_reseller(self, reseller):
        return self._create_account(AccountTypes.RESELLER, OA_ADMIN_ID, reseller)

    def create_customer(self, owner_id, customer):
        return self._create_account(AccountTypes.CLIENT, owner_id, customer), []

    def list_resellers(self, reseller_logins):
        return self._find_accounts(AccountTypes.RESELLER, reseller_logins)

    def list_customers(self, customer_logins):
        return self._find_accounts(AccountTypes.CLIENT, customer_logins)

    def get_account_info_by_login(self, login):
        try:
            staff_member = self.get_account_member_by_login(login)
        except PoaApiError as e:
            if e.module_id == 'AccountManagement' and e.extype_id == 41:
                # No such member with name
                return None

            raise e

        account_info = self.get_account_info(staff_member.account_id)
        for member_id in self.get_account_members(staff_member.account_id):
            account_info.staff_members.append(self.get_member_full_info(member_id))
        return account_info

    def get_account_info_by_id(self, account_id):
        try:
            members = self.get_account_members(account_id)
        except PoaApiError as e:
            if e.module_id == 'AccountManagement' and e.extype_id == 10:
                # There is no such account
                return None

            raise e

        if len(members) == 0:
            return None

        account_info = self.get_account_info(account_id)
        for member_id in members:
            account_info.staff_members.append(self.get_member_full_info(member_id))
        return account_info

    def get_account_info(self, account_id):
        args = {'account_id': account_id}
        response = self.proxy.pem.getAccountInfo(args)
        result = self._parse_response(response)
        return AccountInfo(
            account_id=account_id,
            account_type=result['account_type'],
            language_code='',
            company_name=result['person']['company_name'],
            admin_contact=ContactInfo(
                first_name=result['person']['first_name'],
                middle_name=result['person']['middle_name'],
                last_name=result['person']['last_name'],
                phone=PhoneInfo(
                    country_code=result['phone']['country_code'],
                    area_code=result['phone']['area_code'],
                    phone_num=result['phone']['phone_num'],
                    ext_num=result['phone']['ext_num'],
                ),
                fax=FaxInfo(
                    country_code=result['fax']['country_code'],
                    area_code=result['fax']['area_code'],
                    phone_num=result['fax']['phone_num'],
                    ext_num=result['fax']['ext_num'],
                ),
                email=result['email'],
            ),
            tech_contact=ContactInfo(
                first_name=result['tech_contact']['first_name'],
                middle_name=result['tech_contact']['middle_name'],
                last_name=result['tech_contact']['last_name'],
                phone=PhoneInfo(
                    country_code=result['tech_contact']['phone']['country_code'],
                    area_code=result['tech_contact']['phone']['area_code'],
                    phone_num=result['tech_contact']['phone']['phone_num'],
                    ext_num=result['tech_contact']['phone']['ext_num'],
                ),
                fax=FaxInfo(
                    country_code=result['tech_contact']['fax']['country_code'],
                    area_code=result['tech_contact']['fax']['area_code'],
                    phone_num=result['tech_contact']['fax']['phone_num'],
                    ext_num=result['tech_contact']['fax']['ext_num'],
                ),
                email=result['tech_contact']['email'],
            ),
            billing_contact=ContactInfo(
                first_name=result['billing_contact']['first_name'],
                middle_name=result['billing_contact']['middle_name'],
                last_name=result['billing_contact']['last_name'],
                phone=PhoneInfo(
                    country_code=result['billing_contact']['phone']['country_code'],
                    area_code=result['billing_contact']['phone']['area_code'],
                    phone_num=result['billing_contact']['phone']['phone_num'],
                    ext_num=result['billing_contact']['phone']['ext_num'],
                ),
                fax=FaxInfo(
                    country_code=result['billing_contact']['fax']['country_code'],
                    area_code=result['billing_contact']['fax']['area_code'],
                    phone_num=result['billing_contact']['fax']['phone_num'],
                    ext_num=result['billing_contact']['fax']['ext_num'],
                ),
                email=result['billing_contact']['email'],
            ),
            address=AddressInfo(
                city=result['address']['city'],
                street_name=result['address']['street_name'],
                address2=result['address']['address2'],
                zipcode=result['address']['zipcode'],
                country=result['address']['country'],
                state=result['address']['state'],
            ),
            staff_members=[]
        )

    def get_member_full_info(self, member_id):
        """Get information about specified member (user with type 'member')
        """
        args = {'member_id': member_id}
        response = self._get_proxy(hidelist_response=['$.result.auth_info.password']).pem.getMemberFullInfo(args)
        result = self._parse_response(response)
        return MemberInfo(
            account_id=result['account_id'],
            member_id=member_id,
            auth=AuthInfo(
                login=result['auth_info']['login'],
                password=result['auth_info']['password']
            ),
            contact=ContactInfo(
                first_name=result['pers_contact']['first_name'],
                middle_name=result['pers_contact']['middle_name'],
                last_name=result['pers_contact']['last_name'],
                phone=PhoneInfo(
                    country_code=result['pers_contact']['phone']['country_code'],
                    area_code=result['pers_contact']['phone']['area_code'],
                    phone_num=result['pers_contact']['phone']['phone_num'],
                    ext_num=result['pers_contact']['phone']['ext_num'],
                ),
                fax=FaxInfo(
                    country_code=result['pers_contact']['fax']['country_code'],
                    area_code=result['pers_contact']['fax']['area_code'],
                    phone_num=result['pers_contact']['fax']['phone_num'],
                    ext_num=result['pers_contact']['fax']['ext_num'],
                ),
                email=result['pers_contact']['email'],
            ),
            address=AddressInfo(
                city=result['address']['city'],
                street_name=result['address']['street_name'],
                address2=result['address']['address2'],
                zipcode=result['address']['zipcode'],
                country=result['address']['country'],
                state=result['address']['state'],
            )
        )

    def add_account(
        self, account_type, parent_account_id, person, address, phone, email, fax=None, locale=None, txn_id=None
    ):
        args = {
            'account_type': account_type,
            'parent_account_id': parent_account_id,
        }
        args.update(self._make_account_info_args(person, address, phone, email, fax, locale))
        if txn_id is not None:
            args['txn_id'] = txn_id

        response = self.proxy.pem.addAccount(args)

        result = self._parse_response(response)
        return result['account_id']

    def update_account(self, account_id, account_info):
        person = self._get_account_person_info(account_info.info)
        address = self._get_address_info(account_info.info.company_address)
        phone = self._get_phone_info(account_info.info.admin_contact.phone)
        fax = self._get_fax_info(account_info.info.admin_contact.fax)
        locale = self._get_locale_info(account_info.info.language)

        self._update_account(account_id, person, address, phone, account_info.info.admin_contact.email, fax, locale)

    def add_account_member(self, account_id, auth, person, address, phone, email, fax=None, locale=None, txn_id=None):
        password = auth.password if auth.password else generate_random_password()
        args = {
            'account_id': account_id,
            'auth': {
                'login': auth.login,
                'password': password
            }
        }
        args.update(self._make_account_info_args(person, address, phone, email, fax, locale))
        if txn_id is not None:
            args['txn_id'] = txn_id

        response = self.proxy.pem.addAccountMember(args)

        result = self._parse_response(response)
        return result['user_id']

    def update_staff_member(self, user_id, staff_member_info):
        person = self._get_staff_member_person_info(staff_member_info.contact)
        address = self._get_address_info(staff_member_info.address)
        phone = self._get_phone_info(staff_member_info.contact.phone)
        fax = self._get_fax_info(staff_member_info.contact.fax)
        locale = self._get_locale_info(staff_member_info.language)

        self.update_staff_member_info(
            user_id, staff_member_info.authentication, person, address, phone, staff_member_info.contact.email, fax,
            locale
        )

        self.update_staff_member_password(staff_member_info.authentication)

    def update_staff_member_info(
            self, user_id, auth, person, address, phone, email, fax=None, locale=None, txn_id=None
    ):
        args = {
            'member_id': user_id,
            'login': auth.login,
        }
        args.update(self._make_account_info_args(person, address, phone, email, fax, locale))
        if txn_id is not None:
            args['txn_id'] = txn_id

        response = self.proxy.pem.setMemberInfo(args)
        self._parse_response(response)

    def update_staff_member_password(self, auth, txn_id=None):
        args = {
            'login': auth.login,
            'password': auth.password,
        }

        if txn_id is not None:
            args['txn_id'] = txn_id

        xpath = "/methodCall/params/param/value/struct/member/name[text()='password']/../value/string"
        response = self._get_proxy(hidelist_request=[xpath]).pem.setMemberPassword(args)
        self._parse_response(response)

    def get_account_member_by_login(self, login):
        """Get information (identifiers) about account member by its login
        """
        response = self.proxy.pem.getAccountMemberByLogin({'login': login})
        result = self._parse_response(response)
        return AccountMemberIds(account_id=result['account_id'], user_id=result['user_id'])

    def get_account_members(self, id):
        """Get information (identifiers) about account member by its login
        """
        response = self.proxy.pem.getAccountMembers({'account_id': id})
        return self._parse_response(response)

    def disable_account(self, account_id):
        response = self.proxy.pem.disableAccount({'account_id': account_id})
        self._parse_response(response)

    def _create_account(self, account_type, owner_id, account_info):
        person = self._get_account_person_info(account_info.info)
        address = self._get_address_info(account_info.info.company_address)
        phone = self._get_phone_info(account_info.info.admin_contact.phone)
        fax = self._get_fax_info(account_info.info.admin_contact.fax)
        locale = self._get_locale_info(account_info.info.language)
        parent_account_id = owner_id if owner_id is not None else OA_TARGET_OWNER_ID

        return self.add_account(
            account_type, parent_account_id, person, address, phone, account_info.info.admin_contact.email, fax, locale
        )

    def _update_account(self, account_id, person, address, phone, email, fax=None, locale=None, txn_id=None):
        args = {
            'account_id': account_id,
        }
        args.update(self._make_account_info_args(person, address, phone, email, fax, locale))
        if txn_id is not None:
            args['txn_id'] = txn_id

        response = self.proxy.pem.setAccountInfo(args)
        self._parse_response(response)

    def _find_accounts(self, accounts_type, logins):
        assert accounts_type in [AccountTypes.RESELLER, AccountTypes.CLIENT], \
            _T("MIGRATOR_INTERNAL_ERROR_ACCOUNT_OF_UNSUPPORTED_TYPE") % accounts_type
        accounts = []
        for login in set(logins):
            try:
                member = self.get_account_member_by_login(login)
            except PoaApiError as e:
                if e.module_id == 'AccountManagement' and e.extype_id == 41:    # NoSuchMemberWithName
                    continue
                else:
                    raise
            account_info = self.get_account_info(member.account_id)
            if account_info.account_type != accounts_type:
                continue

            member_info = self.get_member_full_info(member.user_id)
            contact = model.BaseContactInfo(
                username=member_info.auth.login, email=member_info.contact.email,
                first_name=member_info.contact.first_name, last_name=member_info.contact.last_name
            )
            if accounts_type == AccountTypes.RESELLER:
                account = model.BaseResellerInfo(id=account_info.account_id, contact=contact)
            else:    # accounts_type == poa_api.AccountTypes.CLIENT:
                account = model.BaseCustomerInfo(id=account_info.account_id, contact=contact)

            accounts.append(account)
        return accounts

    @staticmethod
    def _make_account_info_args(person, address, phone, email, fax, locale, company=None):
        args = {
            'person': dict(person._asdict()),
            'address': dict(address._asdict()),
            'phone': dict(phone._asdict()),
            'email': email
        }

        if fax is not None:
            args.update({'fax': dict(fax._asdict())})

        if locale is not None:
            args.update({'locale': dict(locale._asdict())})

        return args

    @staticmethod
    def _get_account_person_info(account_info):
        return PersonInfo(
            first_name=account_info.admin_contact.first_name, middle_name=account_info.admin_contact.middle_name,
            last_name=account_info.admin_contact.last_name, company_name=account_info.company_name
        )

    @staticmethod
    def _get_staff_member_person_info(contact):
        return PersonInfo(
            first_name=contact.first_name, middle_name=contact.middle_name, last_name=contact.last_name, company_name=''
        )

    @staticmethod
    def _get_address_info(address):
        # Note that country code in Account's address and in Member's contact should be converted to lower case
        # otherwise POA UI treats it as unknown and shows 'United States'
        return AddressInfo(
            zipcode=address.zipcode, city=address.city, country=address.country.lower(), state=address.state,
            street_name=address.address1, address2=address.address2
        )

    @staticmethod
    def _get_phone_info(phone):
        return PhoneInfo(
            country_code=phone.country_code,
            area_code=phone.area_code,
            phone_num=phone.phone_num,
            ext_num=phone.ext_num,
        )

    @staticmethod
    def _get_fax_info(fax):
        return FaxInfo(
            country_code=fax.country_code,
            area_code=fax.area_code,
            phone_num=fax.phone_num,
            ext_num=fax.ext_num,
        )

    @staticmethod
    def _get_locale_info(locale):
        # Here is a very simple locale processing: only locales supported by Odin Automation are transferred as is,
        # all the others are changed to 'en_US'.
        # Another disadvantage of the current implementation as it sets unknown locale to 'en_US' silently,
        # without any pre-migration or post-migration warnings.
        if locale not in get_system_locales():
            locale = 'en_US'

        lc, cc = locale.split('_')
        return LocaleInfo(language_code=lc, country_code=cc, variant='')
