# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import _T
from migrator.core.data.model.dns import \
    DNS_RESOURCE_RECORD_MX, DNS_RESOURCE_RECORD_A, DNS_RESOURCE_RECORD_AAAA, DNS_RESOURCE_RECORD_CNAME, \
    DNS_RESOURCE_RECORD_SRV, DNS_RESOURCE_RECORD_TXT
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.dns import convert_to_national_domain_name
from migrator.odin_automation.platform.lib.api.openapi.resource import Resource, PoaApiError


class Domain(Resource):
    def __init__(self, proxy):
        self.proxy = proxy

    @trace
    def add_domain(self, subscription_id, domain_name):
        args = {
            'subscription_id': subscription_id,
            'domain_name': convert_to_national_domain_name(domain_name),
            'registrar_status': 2,  # 2 - Domain is registered.
        }
        response = self.proxy.pem.addDomain(args)
        result = self._parse_response(response)
        return result['domain_id']

    @trace
    def disable_domain(self, domain_id=None, domain_name=None):
        self.switch_domain(False, domain_id, domain_name)

    @trace
    def enable_domain(self, domain_id=None, domain_name=None):
        self.switch_domain(True, domain_id, domain_name)

    @trace
    def switch_domain(self, enable, domain_id=None, domain_name=None):
        if domain_id is not None:
            args = {
                'domain_id': int(domain_id)
            }
        elif domain_name is not None:
            args = {
                'domain_name': domain_name
            }
        else:
            raise Exception(_T("INTERNAL_ERROR_OCCURRED_DURING_CALL_OA_API"))

        if enable:
            response = self.proxy.pem.enableDomain(args)
        else:
            response = self.proxy.pem.disableDomain(args)

        self._parse_response(response)

    @trace
    def create_record(self, domain_name, host, record_type, data, ttl=None):
        args = {
            'domain_name': domain_name,
            'host': convert_to_national_domain_name(host),
            'type': record_type.upper(),
            'data': data,
        }
        if ttl is not None:
            args['ttl'] = int(ttl)

        response = self.proxy.pem.createDNSRecord(args)
        result = self._parse_response(response)
        return result['record_id']

    @trace
    def create_record_a(self, domain_name, host, address, ttl=None):
        """
        :param domain_name: str Name of the Domain to create DNS Record for. Examples: 'example.com', 'sub.example.com'
        :param host: str Host part of the DNS Record. Example: 'my-subdomain'
        :param address: str A valid IPv4 value. Example: '1.1.1.1'
        :param ttl: int
        :return: int The ID of the created DNS record
        """
        return self.create_record(domain_name, convert_to_national_domain_name(host), DNS_RESOURCE_RECORD_A, address, ttl)

    @trace
    def create_record_aaaa(self, domain_name, host, address, ttl=None):
        return self.create_record(domain_name, convert_to_national_domain_name(host), DNS_RESOURCE_RECORD_AAAA, address, ttl)

    @trace
    def create_record_cname(self, domain_name, host, address, ttl=None):
        return self.create_record(
            domain_name, convert_to_national_domain_name(host), DNS_RESOURCE_RECORD_CNAME, convert_to_national_domain_name(address), ttl
        )

    @trace
    def create_record_mx(self, domain_name, host, address, priority, ttl=None):
        data = "%s %s" % (priority, convert_to_national_domain_name(address))
        return self.create_record(domain_name, convert_to_national_domain_name(host), DNS_RESOURCE_RECORD_MX, data, ttl)

    @trace
    def create_record_srv(self, domain_name, host, priority, weight, port, target, ttl=None):
        data = "%s %s %s %s" % (priority, weight, port, convert_to_national_domain_name(target))
        return self.create_record(domain_name, convert_to_national_domain_name(host), DNS_RESOURCE_RECORD_SRV, data, ttl)

    @trace
    def create_record_txt(self, domain_name, host, address, ttl=None):
        return self.create_record(domain_name, convert_to_national_domain_name(host), DNS_RESOURCE_RECORD_TXT, address, ttl)

    @trace
    def get_records(self, domain_name, record_type=None):
        args = {
            'name': domain_name,
            'rr_type': record_type or ''
        }

        response = self.proxy.pem.getDNSRecords(args)
        result = self._parse_response(response)
        return result

    @trace
    def delete_record(self, record_id):
        args = {
            'record_id': int(record_id)
        }
        response = self.proxy.pem.deleteDNSRecord(args)
        self._parse_response(response)

    @trace
    def get_domain_name_servers(self, domain, subscription_id=None):
        args = {
            'domain_name': domain
        }
        if subscription_id:
            args['subscription_id'] = int(subscription_id)

        response = self.proxy.pem.getDomainNameServers(args)
        return self._parse_response(response)

    @trace
    def enable_record(self, record_id):
        response = self.proxy.pem.enableDNSRecord({'record_id': int(record_id)})
        self._parse_response(response)

    @trace
    def disable_record(self, record_id):
        response = self.proxy.pem.disableDNSRecord({'record_id': int(record_id)})
        self._parse_response(response)

    @trace
    def bind_services_to_domain(self, domain_id, subscription_id):
        args = {
            'domain_id': domain_id,
            'subscription_id': subscription_id
        }
        response = self.proxy.pem.bindServicesToDomain(args)
        self._parse_response(response)

    @trace
    def unbind_services_from_domain(self, domain_id, subscription_id):
        args = {
            'domain_id': domain_id,
            'subscription_id': subscription_id
        }
        response = self.proxy.pem.unbindServicesFromDomain(args)
        self._parse_response(response)

    @trace
    def get_domain_by_name(self, domain):
        args = {
            'domain_name': domain
        }
        response = self.proxy.pem.getDomainByName(args)
        try:
            return self._parse_response(response)
        except PoaApiError as e:
            if e.error_code != 3006:  # Domain does not exist.
                raise e

        return None
