# (C)2017 Ingram Micro Inc. all Rights Reserved
import psycopg2
import psycopg2.extras

from collections import namedtuple

from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.networking.ip_utils import get_zerofilled_ip
from migrator.odin_automation.platform.lib.api.openapi.resource import PoaApiError, Resource
from migrator.odin_automation.platform.lib.db import get_platform_db_connection, execute_sql_and_get_formatted_result

HostInfo = namedtuple('HostInfo', ('attributes', 'ready_to_provide', 'hostname', 'ip_addresses', 'interfaces'))
IpAddressInfo = namedtuple('IpAddressInfo', ('type', 'address'))
InterfaceInfo = namedtuple('InterfaceInfo', ('id', 'name'))
ServiceNodeInfo = namedtuple('ServiceNodeInfo', ('host_id', 'attributes', 'is_ready_to_provide', 'license_status'))
IPPoolInfo = namedtuple('IPPoolInfo', ('id', 'name', 'ip_from', 'ip_to', 'netmask'))


class HostNotFoundException(Exception):
    pass


class Host(Resource):
    def __init__(self, proxy, get_proxy):
        self.proxy = proxy
        self._get_proxy = get_proxy

    def get_host(self, host_id, txn_id=None):
        args = {'host_id': host_id}
        if txn_id is not None:
            args['txn_id'] = txn_id

        response = self.proxy.pem.getHost(args)
        result = self._parse_response(response)
        return HostInfo(
            attributes=result['attributes'],
            ready_to_provide=result['ready_to_provide'],
            hostname=result['hostname'],
            ip_addresses=[
                IpAddressInfo(type=ip_address['ip_type'], address=ip_address['ip_address'])
                for ip_address in result['ip_addresses']
            ],
            interfaces=[InterfaceInfo(id=interface['id'], name=interface['name']) for interface in result['interfaces']]
        )

    def find_host(self, hostname=None, ip_address=None, txn_id=None):
        assert hostname is not None or ip_address is not None
        args = dict()
        if hostname is not None:
            args['hostname'] = hostname

        if ip_address is not None:
            args['ip_address'] = ip_address

        if txn_id is not None:
            args['txn_id'] = txn_id

        response = self.proxy.pem.findHost(args)
        try:
            result = self._parse_response(response)
        except PoaApiError as e:
            # POA API does not provide more reliable way to detect error type:
            # - there is no error code
            # - 'extype_id' for this error intersects with other errors
            if e.error_message == "Host not found":
                return None
            elif e.error_message.startswith("Host") and e.error_message.endswith("does not exist."):
                return None
            else:
                raise

        return result['host_id']

    def get_service_nodes_info(self):
        response = self.proxy.pem.msp_license_manager.getServiceNodesInfo()
        result = self._parse_response(response)
        return [
            ServiceNodeInfo(
                host_id=item['host_id'],
                attributes=item['attributes'],
                is_ready_to_provide=item['is_ready_to_provide'],
                license_status=item['license_status']
            ) for item in result
        ]

    @cached
    def get_host_id_by_ip(self, ip):
        host_id = self.find_host(ip_address=ip)
        if host_id is None:
            raise HostNotFoundException(_T("HOST_ID_FOR_IP_S_NOT").format(ip_address=ip))

        return host_id

    def register_linux_service_node(
            self, ip_address, login, password, role=None, role_params=None, txn_id=None, network_config=None,
            register_only_shared_ip=False
    ):
        args = {
            'host': ip_address,
            'login': login,
            'password': password,
            'register_only_shared_ip': register_only_shared_ip
        }

        if role is not None:
            args['role'] = role

        if role_params is not None:
            args['role_params'] = role_params

        if txn_id is not None:
            args['txn_id'] = txn_id

        if network_config is not None:
            args['network_config'] = network_config

        xpath = "/methodCall/params/param/value/struct/member/name[text()='password']/../value/string"
        response = self._get_proxy(hidelist_request=[xpath]).pem.registerSharedNode(args)
        return self._parse_response(response)

    def register_windows_service_node(
        self, ip_address, username, password, resource_type_id=None, role=None, role_params=None, txn_id=None,
            network_config=None, register_only_shared_ip=False
    ):
        args = {
            'host': ip_address,
            'username': username,
            'password': password,
            'register_only_shared_ip': register_only_shared_ip
        }

        if role is not None:
            args['role'] = role

        if role_params is not None:
            args['role_params'] = role_params

        if resource_type_id is not None:
            args['resource_type_id'] = resource_type_id

        if txn_id is not None:
            args['txn_id'] = txn_id

        if network_config is not None:
            args['network_config'] = network_config

        xpath = "/methodCall/params/param/value/struct/member/name[text()='password']/../value/string"
        response = self._get_proxy(hidelist_request=[xpath]).pem.registerWindowsNode(args)
        return self._parse_response(response)

    def create_ip_pool(self, name, start_ip, end_ip, netmask='0.0.0.0', sharing='a', purpose=('SHARED.HOSTING',)):
        args = {
            'name': name, 'startIP': start_ip, 'endIP': end_ip,
            'netmask': netmask, 'sharing': sharing, 'purpose': purpose
        }
        response = self.proxy.pem.createIPPool(args)
        return self._parse_response(response)['pool_id']

    def create_ipv6_pool(
            self, name, pool_prefix, pool_prefix_len=128, subnet_prefix_len=16, sharing='a', purpose=('SHARED.HOSTING',)
    ):
        args = {
            'name': name, 'poolPrefix': pool_prefix, 'poolPrefixLength': str(pool_prefix_len),
            'subnetPrefixLength': str(subnet_prefix_len), 'sharing': sharing, 'purpose': purpose
        }
        response = self.proxy.pem.createIPv6Pool(args)
        return self._parse_response(response)['pool_id']

    def bind_ip_pool(self, pool_id, interface_id):
        args = {'pool_id': pool_id, 'interface_id': interface_id}
        response = self.proxy.pem.bindIPPool(args)
        return self._parse_response(response)

    def get_ip_pools(self, ip_from, ip_to):
        connection = get_platform_db_connection(cursor_factory=psycopg2.extras.RealDictCursor)
        cursor = connection.cursor()

        sql = """
            SELECT pool_id as id, pool_name as name, ip_from, ip_to, netmask
            FROM ip_pools
            WHERE ip_from=%s AND ip_to=%s
        """
        cursor.execute(sql, (get_zerofilled_ip(ip_from), get_zerofilled_ip(ip_to)))
        return [IPPoolInfo(**row) for row in cursor.fetchall()]

    def get_hosts_with_bound_ip_pools(self, hosts_guids=None):
        """
        :param hosts_guids: list[string]
        :return: list([dict]) Example:
        [
           {'host_id': 2, 'host_aps_guid': 'c518c8c8-3d25-436f-a6d5-3843e10d7fbe', 'bound_ip_pools': [4, 1, 2, 3]},
           {'host_id': 3, 'host_aps_guid': '711dfbdb-ed2b-4e18-a162-3d0f1fc7629e', 'bound_ip_pools': []},
           {'host_id': 4, 'host_aps_guid': '41970491-1346-480c-a474-9bc18f18f1c7', 'bound_ip_pools': [2]}
        ]
        """
        if hosts_guids:
            where_clause = 'WHERE h.aps_uuid IN %s'
        else:
            where_clause = ''

        sql = """
        SELECT h.host_id, h.aps_uuid, array_remove(array_agg(ip.pool_id), NULL)
        FROM hosts AS h
        LEFT JOIN net_interfaces AS ni ON ni.host_id = h.host_id
        LEFT JOIN bound_ip_pools AS bip ON bip.if_id = ni.if_id
        LEFT JOIN ip_pools AS ip ON ip.pool_id = bip.pool_id
        %s
        GROUP BY h.host_id
        """ % where_clause

        return execute_sql_and_get_formatted_result(
            sql, ('host_id', 'host_aps_guid', 'bound_ip_pools'), [tuple(hosts_guids)] if hosts_guids else None
        )
