# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
from collections import namedtuple

from migrator.core import APIError
from migrator.core.language import _T


ResourceDetails = namedtuple('ResourceDetails', ('rt_id', 'name', 'resclass_name', 'limit', 'act_params', 'usage'))
ResourceTypeDetailsBase = namedtuple('ResourceTypeDetails', ('rt_id', 'name', 'resclass_name', 'limit', 'act_params'))
ActParamInfo = namedtuple('ActParamInfo', ('name', 'value'))


class ResourceTypeDetails(ResourceTypeDetailsBase):
    def get_aps_guid(self):
        for act_param in self.act_params:
            if act_param.name == 'resource_uid':
                return act_param.value

        return None


class PoaApiError(APIError):
    """Error occured when executing request to POA API"""
    def __init__(self, status, module_id, extype_id, error_code, error_message, properties):
        if error_code is not None:
            self.full_error_message = "[%s] %s" % (error_code, error_message)
        else:
            self.full_error_message = error_message
        super(PoaApiError, self).__init__(self.full_error_message)
        self.status = status
        self.module_id = module_id
        self.extype_id = extype_id
        self.error_code = error_code
        self.error_message = error_message
        self.properties = properties

    @property
    def how_to_reproduce(self):
        """User-friendly description on how to reproduce API request"""
        return _T("HOW_TO_REPRODUCE_FAILED_REQUEST_1").format(error_message=self.full_error_message)


class Resource:
    logger = logging.getLogger(__name__)

    @staticmethod
    def _parse_resource_types(result):
        rts = []
        for rt_item in result['resource_types']:
            act_params_dict = {}
            if 'act_params' in rt_item:
                for rtap_item in rt_item['act_params']:
                    act_params_dict[rtap_item['var_name']] = rtap_item['var_value']
            if 'st_act_params' in rt_item:
                for rtap_item in rt_item['st_act_params']:
                    act_params_dict[rtap_item['var_name']] = rtap_item['var_value']
            act_params = [ActParamInfo(name, value) for name, value in act_params_dict.iteritems()]
            rts.append(ResourceTypeDetails(
                rt_id=rt_item['resource_type_id'], name=rt_item['name'],
                resclass_name=rt_item['resclass_name'], limit=int(rt_item['limit']), act_params=act_params
            ))
        return rts

    @staticmethod
    def _parse_resources(result):
        resources = []
        for rt_item in result['resource_types']:
            act_params = []
            if 'act_params' in rt_item:
                for rtap_item in rt_item['act_params']:
                    act_params.append(ActParamInfo(name=rtap_item['var_name'], value=rtap_item['var_value']))
            resources.append(ResourceDetails(
                rt_id=rt_item['resource_type_id'], name=rt_item['name'], resclass_name=rt_item['resclass_name'],
                limit=int(rt_item['limit']), act_params=act_params, usage=int(rt_item['usage'])
            ))
        return resources

    @staticmethod
    def _parse_response(response):
        if response['status'] != 0:
            raise PoaApiError(
                error_code=response.get('error_code', None),
                **(dict((name, response.get(name)) for name in ('status', 'module_id', 'extype_id', 'error_message', 'properties')))
            )
        return response.get('result')
