# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.odin_automation.platform.lib.api.openapi.resource import Resource, ActParamInfo
from migrator.settings import OA_ADMIN_ID


ServiceTemplateInfo = namedtuple('ServiceTemplateInfo', ('st_id', 'name', 'description', 'active', 'owner_id'))
# XXX why ServiceTemplateDetails is missing 2 fields of ServiceTemplateInfo? how these fields are used?
ServiceTemplateDetailsBase = namedtuple('ServiceTemplateDetails', (
    'st_id', 'name', 'owner_id', 'is_active', 'autoprovidable', 'resource_types'
))
ServiceInfo = namedtuple('ServiceInfo', ('host_id', 'service_name'))


class STResource(object):
    def __init__(self, description, reclass_name, service_id, resource_id):
        self.description = description
        self.reclass_name = reclass_name
        self.service_id = service_id
        self.resource_id = resource_id

    def convert_to_comparable_format(self):
        result = {'reclass_name': self.reclass_name, 'act_params': set()}
        for key in ['service_id', 'resource_id']:
            value = getattr(self, key)
            if value is not None:
                result['act_params'].add(ActParamInfo(name=key, value=value))

        return result

    def get_st_resource_description(self):
        return '"%s"' % self.description

    def __repr__(self):
        return '"{description}" (r_class: "{r_class}", service_id: "{ser_id}", resource_id: "{res_id}")'.format(
            description=self.description,
            r_class=self.reclass_name,
            ser_id=self.service_id,
            res_id=self.resource_id
        )

# Resource class names
RC_RES = 'rc.saas.resource'
RC_RES_UNIT = 'rc.saas.resource.unit'
RC_RES_MHZ = 'rc.saas.resource.mhz'
RC_SER = 'rc.saas.service'
RC_SER_LINK = 'rc.saas.service.link'

# WebHosting Plesk resources
RL_FTP_USERS = STResource('Additional FTP Accounts for a Webspace (Linux Apache/Nginx)', RC_RES_UNIT, 'ServicePleskWebApache', 'additionalFtpAccounts')
RW_FTP_USERS = STResource('Additional FTP Accounts for a Webspace (Windows IIS)', RC_RES_UNIT, 'ServicePleskWebIis', 'additionalFtpAccounts')
R_DNS = STResource('DNS Hosting', 'DNS Hosting', None, None)
R_EMAIL_USERS = STResource('Email Accounts', RC_SER, 'MailBox', None)
R_EMAIL_DOMAIN = STResource('Email Domain', RC_SER, 'MailDomain', None)
R_EMAIL_DISK = STResource('Email Storage (total)', RC_RES, 'ServicePleskMail', 'mailDiskSpace')
RL_HOSTING_CONF = STResource('Generic Hosting Configuration (Linux Apache/Nginx)', RC_SER, 'WebConfigurationApache', None)
RW_HOSTING_CONF = STResource('Generic Hosting Configuration (Windows IIS)', RC_SER, 'WebConfigurationIis', None)
RL_WEB_TEMPL = STResource('Generic Apache Web Hosting Template', RC_SER_LINK, 'WebTemplateApache', None)
RW_WEB_TEMPL = STResource('Generic IIS Web Hosting Template', RC_SER_LINK, 'WebTemplateIis', None)
R_HTTP_REDIR = STResource('HTTP Domain Redirect', RC_SER, 'SiteForwardingStd', None)
RW_DISK = STResource('Hard Disk Quota (Windows IIS)', RC_RES, 'ServicePleskWebIis', 'hardDiskQuota')
RL_DISK = STResource('Hard Disk Quota (Linux Apache/Nginx)', RC_RES, 'ServicePleskWebApache', 'hardDiskQuota')
R_IP = STResource('IP addresses', 'ips', None, None)
R_MAIL_TEMPL = STResource('Mail Hosting Template', RC_SER_LINK, 'TemplatePleskMail', None)
RW_MY_OR_POSTGRE_SQL_WEB = STResource('MySQL and PostgreSQL Databases for a Webspace (Windows IIS)', RC_RES_UNIT, 'ServicePleskWebIis', 'dbLimit')
RL_MY_OR_POSTGRE_SQL_WEB = STResource('MySQL and PostgreSQL Databases for a Webspace (Linux Apache/Nginx)', RC_RES_UNIT, 'ServicePleskWebApache', 'dbLimit')
RL_PLESK_WEB = STResource('Plesk Apache Web Service', RC_SER, 'ServicePleskWebApache', None)
RW_PLESK_WEB = STResource('Plesk IIS Web Service', RC_SER, 'ServicePleskWebIis', None)
R_PLESK_MAIL = STResource('Plesk Mail Service', RC_SER, 'ServicePleskMail', None)
R_TRAFF = STResource('Traffic', 'traffic', None, None)
R_WEB_LIMITS = STResource('Web Hosting General Limits', RC_SER, 'Subscription', None)
R_WEB_SERV = STResource('Web Hosting Service', RC_SER_LINK, 'WebApplication', None)
R_SITE = STResource('Website', RC_SER, 'SiteWeb', None)
R_WEBSPACE = STResource('Webspace', RC_SER, 'WebSpace', None)
RW_MS_SQL_WEB = STResource('MS SQL Databases for a Webspace (Windows IIS)', RC_RES_UNIT, 'ServicePleskWebIis', 'dbMssqlLimit')
# OACI resources
R_VPS_VST = STResource('VPS Server Template', RC_SER_LINK, 'samples', None)
R_VPS_CORE_POWER = STResource('VPS Core Power', RC_RES_MHZ, 'vps', 'corePower')
# Mail resources
RL_QMAIL = STResource('Mail hosting (qmail with clustering support)', 'cqmail.rc.hosting', None, None)
RW_EXCHANGE = STResource('Hosted Exchange', 'exchange.domainServices', None, None)
RW_EXCHANGE_MAILBOXES = STResource('Exchange Mailboxes', 'exchange.mailboxes', None, None)
RW_EXCHANGE_RES_MAILBOXES = STResource('Exchange Resource Mailboxes', 'exchange.resource_mailboxes', None, None)
RW_EXCHANGE_MAILBOX_TEMPL = STResource('Hosted Exchange Mailbox Template', 'exchange.mailbox_plan', None, None)
R_O365_DOMAINS = STResource('Office 365 Domains', RC_SER, 'office365domains', None)

RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND = 0
RES_AVAILABILITY_RESULT_SOME_RESOURCES_NOT_FOUND = 1


class ServiceTemplateDetails(ServiceTemplateDetailsBase):
    def get_identificators(self):
        return "'{service_template_name}' (ID #{service_template_id})".format(
            service_template_name=self.name,
            service_template_id=self.st_id
        )

    def are_resources_exist(self, st_res_list):
        """
        :param st_res_list: list[STResource]
        :return: dict
        """
        result = {'status': RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND, 'missed_st_res_list': list()}
        for st_res in st_res_list:
            is_resource_found = False
            st_res_in_comparable_format = st_res.convert_to_comparable_format()
            for rt_details in self.resource_types:
                if rt_details.resclass_name == st_res_in_comparable_format['reclass_name']:
                    rt_details_in_comparable_format = set(rt_details.act_params)
                    if st_res_in_comparable_format['act_params'].issubset(rt_details_in_comparable_format):
                        is_resource_found = True
                        break

            if not is_resource_found:
                result['status'] = RES_AVAILABILITY_RESULT_SOME_RESOURCES_NOT_FOUND
                result['missed_st_res_list'].append(st_res)

        return result

    def get_resource_full_info(self, st_res):
        """
        :param st_res: STResource The identifiers of a ST resource to retrieve full info by, e.g: R_IP, R_OACI_VPS_VST
        :return: list[ResourceTypeDetails]|list[] The detailed list of ST resources matched the passed identifiers
        """
        result = list()
        st_res_in_comparable_format = st_res.convert_to_comparable_format()
        for rt_details in self.resource_types:
            if rt_details.resclass_name == st_res_in_comparable_format['reclass_name']:
                rt_details_in_comparable_format = set(rt_details.act_params)
                if st_res_in_comparable_format['act_params'].issubset(rt_details_in_comparable_format):
                    result.append(rt_details)

        return result

    def get_available_resource_types(self):
        return '{st_id}:\n{res_types_info}\n'.format(
            st_id=self.get_identificators(),
            res_types_info='\n'.join(
                '- reclass_name: "{reclass_name}" | act_params: {act_params}'.format(
                    reclass_name=rt_details.resclass_name,
                    act_params=str(rt_details.act_params)
                ) for rt_details in self.resource_types
            )
        )


class Service(Resource):
    def __init__(self, proxy):
        self.proxy = proxy

    def get_service_template_for_migration_item(self, global_context, migration_item):
        """
        :param global_context: migrator.core.session.GlobalMigrationContext
        :param migration_item: migrator.core.data.model.migration_project_item.MigrationProjectItemModel
        :return: ServiceTemplateDetails
        """
        is_billing = global_context.target_panel_obj.is_billing_module_installed()

        if not is_billing:
            st_id = migration_item.target_service_plan_id
        else:
            st_id = global_context.target_panel_obj.get_billing_api(global_context).plan_details_get(
                migration_item.target_service_plan_id
            ).st_id

        return self.get_service_template(st_id, True)

    def get_plan_by_id(self, plan_id, get_resources=False):
        return self.get_service_template(plan_id, get_resources)

    def get_plan_by_name(self, plan_name, active=True):
        for service_template in self.get_service_template_list(OA_ADMIN_ID, active):
            if service_template.name == plan_name:
                return service_template

        raise MigrationError(_T("PLAN_IS_MISSING_FROM_OA").format(plan_name=plan_name))

    @cached
    def get_service_template_list(self, owner_id, active=False):
        """Get list of service templates that belong to specified owner, with their information
        """
        response = self.proxy.pem.getServiceTemplateList(dict(owner_id=owner_id, active=active))
        result = self._parse_response(response)
        return [ServiceTemplateInfo(i['st_id'], i['name'], i['description'], i['active'], i['owner_id']) for i in result]

    @cached
    def get_service_template(self, st_id, get_resources):
        """Get details about specified service template
        """
        args = {
            'st_id': int(st_id),
            'get_resources': get_resources,
            'get_full_info': get_resources  # resources' activation parameters
        }

        response = self.proxy.pem.getServiceTemplate(args)
        result = self._parse_response(response)

        if get_resources:
            rts = self._parse_resource_types(result)
        else:
            rts = None

        return ServiceTemplateDetails(
            st_id=result['st_id'], name=result['name'], owner_id=result['owner_id'], is_active=result['is_active'],
            autoprovidable=result['autoprovidable'], resource_types=rts
        )

    @cached
    def listServices(self):
        args = {}
        response = self.proxy.pem.packaging.listServices(args)
        result = self._parse_response(response)
        return [ServiceInfo(r['host_id'], r['service_name']) for r in result]
