# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple
from migrator.core.utils.networking.dns import convert_to_national_subscription_name
from migrator.odin_automation.platform.lib.api.openapi.resource import Resource


WebspaceInfo = namedtuple('WebspaceInfo', ('webspace_id', 'domain', 'sub_id', 'owner_id', 'status', 'rt_ids'))
SubscriptionInfo = namedtuple('SubscriptionInfo', (
    'subscription_id', 'owner_id', 'st_id', 'name', 'is_active', 'resources'
))


class Subscription(Resource):
    def __init__(self, proxy):
        self.proxy = proxy

    def create_subscription(self, account_id, plan_id, subscription_name, subscription_id):
        act_params = [
            {'var_name': 'domain_name', 'var_value': subscription_name},
            {'var_name': 'plesk_disable_provisioning', 'var_value': 'false'},
        ]

        subscription_id = self.add_subscription(account_id, plan_id, act_params, subscription_name, subscription_id)

        return subscription_id

    def add_subscription(self, account_id, service_template_id, act_params=None, subscription_name=None, subscription_id=None):
        args = {
            'account_id': account_id,
            'service_template_id': service_template_id,
        }

        if act_params is not None:
            args['parameters'] = act_params
        if subscription_name is not None:
            args['subscription_name'] = convert_to_national_subscription_name(subscription_name)
        if subscription_id is not None:
            args['subscription_id'] = subscription_id

        response = self.proxy.pem.addSubscription(args)
        result = self._parse_response(response)
        if type(result) is int:
            return result
        else:
            return result['subscription_id']

    def create_webspace(self, sub_id, domain, resources, txn_id=None):
        """Create webspace with specified domain name and resources in specified subscription.
        Return webspace identifier.
        """
        # [ { } ]
        r_array = []
        for r in resources:
            r_serialized = {
                'rt_id': r.rt_id
            }
            # [ { } ]
            p_array = []
            for p in r.params:
                p_array.append({
                    'name': p.name,
                    'value': p.value
                })
            if len(r.params) > 0:
                r_serialized['params'] = p_array
            r_array.append(r_serialized)

        args = {
            'new_webspace': {
                'sub_id': sub_id,
                'domain': domain,
                'resources': r_array,
            }
        }

        if txn_id is not None:
            args['txn_id'] = txn_id
        response = self.proxy.pem.pleskintegration.createWebspace(args)
        result = self._parse_response(response)
        return result['webspace_id']

    def get_webspace(self, webspace_id):
        response = self.proxy.pem.pleskintegration.getWebspace({'webspace_id': webspace_id})
        result = self._parse_response(response)
        return WebspaceInfo(
            webspace_id=result['webspace_id'], domain=result['domain'], sub_id=result['sub_id'],
            owner_id=result['owner_id'], status=result['status'], rt_ids=result['rt_ids']
        )

    def get_account_subscriptions(self, account_id):
        """Get identifiers of subscriptions of specified account
        """
        response = self.proxy.pem.getAccountSubscriptions({'account_id': account_id})
        # result consists of array of subscription identifiers, no need in conversion, return it as is
        return self._parse_response(response)

    def get_subscription(self, subscription_id, get_resources=False):
        """Get information about specified subscription
        """
        response = self.proxy.pem.getSubscription({
            'subscription_id': int(subscription_id),
            'get_resources': get_resources
        })
        result = self._parse_response(response)

        if get_resources:
            resources = self._parse_resources(result)
        else:
            resources = None

        return SubscriptionInfo(
            subscription_id=result['subscription_id'], owner_id=result['owner_id'], st_id=result['st_id'],
            name=result['name'], is_active=result['is_active'], resources=resources
        )

    def disable_subscription(self, subscription_id):
        response = self.proxy.pem.disableSubscription({'subscription_id': subscription_id})
        self._parse_response(response)

    def get_webspace_id_by_primary_domain(self, domain_name):
        args = {'domain': domain_name}
        return self._parse_response(self.proxy.pem.pleskintegration.getWebspaceIDByPrimaryDomain(args))['webspace_id']

    def remove_webspace(self, webspace_id):
        args = {'webspace_id': webspace_id}
        response = self.proxy.pem.pleskintegration.removeWebspace(args)
        self._parse_response(response)

    def get_apache_webspaces(self, subscription_id):
        args = {'subscription_id': subscription_id}
        response = self.proxy.pem.apache.getWebspacesList(args)
        return self._parse_response(response)

    def get_subscription_webspaces(self, subscription_id):
        args = {'subscription_id': subscription_id}
        response = self.proxy.pem.getSubscriptionWebspaces(args)
        return self._parse_response(response)

    def activate_subscription(
        self, account_id, service_template_id, subscription_name, act_params=None,
        resource_limits=None, txn_id=None, subscription_id=None
    ):
        """
        :param account_id: string|integer Examples: '1000001', 1000001
        :param service_template_id: integer Example: 5
        :param subscription_name: string Example: 'Windows Shared Hosting (wsh-example.com)'
        :param act_params: list
        :param resource_limits: None
        :param txn_id: None
        :param subscription_id: integer Example: 1000005
        :return:
        """
        args = {
            'account_id': int(account_id),
            'service_template_id': service_template_id,
        }

        if act_params is not None:
            args['parameters'] = act_params

        if subscription_id is not None:
            args['subscription_id'] = subscription_id
        if subscription_name is not None:
            args['subscription_name'] = convert_to_national_subscription_name(subscription_name)
        if resource_limits:
            args['resource_limits'] = resource_limits
        if txn_id is not None:
            args['txn_id'] = txn_id
        response = self.proxy.pem.activateSubscription(args)
        result = self._parse_response(response)
        return result['subscription_id']
