# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
from contextlib import contextmanager

from migrator.core.connections.target_connections import TargetConnections
from migrator.core.session.session_dir import UnixSessionDir
from migrator.core.utils import cached
from migrator.core.utils.networking.run_command import LocalUnixRunner
from migrator.core.utils.networking.ssh.connection_pool import SSHConnectionPool
from migrator.odin_automation.platform.lib.api import openapi
from migrator.odin_automation.billing.lib import api as billing_api
from migrator.odin_automation.platform.config import read_automation_settings
from migrator.core.language import _T
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner.unix_hcl_runner import UnixHclRunner
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner.windows_hcl_runner import WindowsHclRunner
from migrator.settings import TARGET_PRODUCT_OA

logger = logging.getLogger(__name__)


class AutomationTargetConnections(TargetConnections):
    def __init__(self, config):
        self._automation_settings = read_automation_settings(
            config, TARGET_PRODUCT_OA,
        )
        if not self._automation_settings.is_local:
            self._ssh = SSHConnectionPool.get_instance().get(
                self._automation_settings, _T("OA_MN_DESCRIPTION")
            )
        self._session_dir = UnixSessionDir(
            self.main_node_runner, self._automation_settings.unix_session_dir
        )

        self._config = config
        logger.debug(_T("LOG_OA_MN"), self._automation_settings.ip)

        self._oa_panel = None

    @property
    def main_node_ip(self):
        return self._automation_settings.ip

    @contextmanager
    def main_node_runner(self):
        if self._automation_settings.is_local:
            yield LocalUnixRunner()
        else:
            with self._ssh.runner() as runner:
                yield runner

    @contextmanager
    def unix_node_runner(self, node_id):
        with self.main_node_runner() as runner:
            yield UnixHclRunner(
                runner, node_id, self.get_automation_node_description(node_id)
            )

    @contextmanager
    def windows_node_runner(self, node_id):
        with self.main_node_runner() as runner:
            yield WindowsHclRunner(
                runner, node_id, self.get_automation_node_description(node_id)
            )

    @contextmanager
    def billing_node_runner(self):
        xml_rpc_address = self.poa_api().get_bm_xml_rpc_address()
        billing_host_id = self.poa_api().host.find_host(ip_address=xml_rpc_address['host'])
        with self.main_node_runner() as runner:
            description = self.get_automation_node_description(billing_host_id)
            yield UnixHclRunner( runner, billing_host_id, description)

    @cached
    def poa_api(self):
        return openapi.OpenAPI(self._automation_settings.poa_api.url)

    @cached
    def billing_api(self):
        xml_rpc_address = self.poa_api().get_bm_xml_rpc_address()
        bm_api_user = self.poa_api().get_bm_api_user()
        bm_common_password = self.poa_api().get_bm_common_password()
        protocol = 'https' if xml_rpc_address.get('ssl', True) else 'http'
        url = '%s://%s:%s%s' % (
            protocol, xml_rpc_address['host'], xml_rpc_address.get('portnum', '5224'),
            xml_rpc_address.get('uri', '/RPC2'))

        return billing_api.BillingAPI(url, bm_api_user, bm_common_password)

    @cached
    def get_automation_node_description(self, node_id):
        """Get description of service node that should be used in log messages.
        Function is cached, so use it only in stable environment,
        when you do not expect new nodes to be registered/assimilated"""

        node_info = self.poa_api().host.get_host(node_id)

        if len(node_info.ip_addresses) > 0:
            ip_address = node_info.ip_addresses[0].address
        else:
            # it should not be possible, but if it is so due to some bug, use this just not to fail all checks
            ip_address = _T("UNKNOWN_IP")

        return _T("OA_SN_DESCRIPTION") % (node_info.hostname, node_id, ip_address)

    @property
    def unix_session_dir(self):
        return self._automation_settings.unix_session_dir

    @property
    def windows_session_dir(self):
        return self._automation_settings.windows_session_dir

    @property
    def is_windows(self):
        return False

    @property
    def is_local(self):
        """If main node of target server is local or remote

        Local server means that migrator works on the target node, remote means
        that migrator's node and target node are different servers
        """
        return self._automation_settings.is_local
