# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from contextlib import contextmanager

import migrator.core.utils.networking.rsync

from migrator.core import MigrationNoRepeatError
from migrator.core.connections.plesk_server import PleskServer
from migrator.core.connections.target_servers import TargetServer
from migrator.core.language import _T
from migrator.core.session import session_dir
from migrator.core.session.config import SSHAuthKeyDefault
from migrator.core.transfer.db.mysql import allow_admin_access, check_admin_access
from migrator.core.utils import cached, obj
from migrator.core.utils.operating_system import os_info
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import Failure
from migrator.odin_automation.platform.lib.api.aps import host
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import get_aps_connector_with_admin_token
from migrator.odin_automation.platform.lib.db.ip import get_backnet_ip_address
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner.hcl_common import hcl_is_windows
from migrator.odin_automation.services.shm.lib.aps import types
from migrator.settings import SSH_DEFAULT_PORT

logger = logging.getLogger(__name__)


class AutomationServer(TargetServer):
    host_id = None
    ip_address = None

    @classmethod
    @cached
    def get(cls, conn, id=None, ip=None):
        return cls(conn, id, ip)

    def __init__(self, conn, id=None, ip=None):
        super(AutomationServer, self).__init__()
        self.conn = conn

        if id:
            self.host_id = int(id)
        if ip:
            self.ip_address = ip

        if not self.host_id and not self.ip_address:
            raise Exception('Either host_id or host_ip must be not None')

        if not self.host_id:
            self.host_id = conn.poa_api().host.get_host_id_by_ip(ip)

        self.host_info = conn.poa_api().host.get_host(self.host_id)
        if not self.ip_address:
            self.ip_address = get_backnet_ip_address(self.host_id)

        if self.is_windows():
            self._session_dir = session_dir.WindowsSessionDir(self.runner, self.conn.windows_session_dir)
        else:
            self._session_dir = session_dir.UnixSessionDir(self.runner, self.conn.unix_session_dir)

        self._settings = obj(ssh_auth=SSHAuthKeyDefault(port=SSH_DEFAULT_PORT, username='root'))

    def user(self):
        return self._settings.ssh_auth.username

    def settings(self):
        return self._settings

    @contextmanager
    def runner(self):
        if self.is_windows():
            with self.conn.windows_node_runner(self.host_id) as runner:
                yield runner
        else:
            with self.conn.unix_node_runner(self.host_id) as runner:
                yield runner

    def is_windows(self):
        with self.conn.main_node_runner() as runner:
            return hcl_is_windows(runner, self.host_id, self.description())

    def description(self):
        return self.conn.get_automation_node_description(self.host_id)

    def ip(self):
        return self.ip_address

    def session_dir(self):
        return self._session_dir.get_session_dir_path()

    def get_session_file_path(self, filename):
        return self._session_dir.get_file_path(filename)

    def get_hostname(self):
        return self.host_info.hostname

    @property
    @cached
    def rsync_bin(self):
        with self.runner() as runner:
            return migrator.core.utils.networking.rsync.detect_rsync_path(runner, self.description())

    def __hash__(self):
        return hash(self.host_id)

    def __eq__(self, another):
        return False

    @cached
    def os_info(self):
        with self.runner() as runner:
            return os_info.get_os_info(runner)


class PleskTargetServer(AutomationServer, PleskServer):

    def __init__(self, conn, id=None, ip=None):
        super(PleskTargetServer, self).__init__(conn, id, ip)

    @cached
    def plesk_api(self):
        return self._plesk_api(self.ip_address, '12.5')

    @property
    @cached
    def capabilities(self):
        aps = get_aps_connector_with_admin_token()
        plesk_node_uid = host.get_oa_host_uid(self.host_id)
        shm_node_uid = aps.get_first_resource(
            filter_str='implementing(%s),eq(license.nodeId,%s)' % (types.SHMHost, plesk_node_uid)
        )['aps']['id']
        return aps.get_resource('%s/capabilities' % shm_node_uid)

    def add_database_server(self, dbserver):
        _allow_admin_access = {
            'mysql': allow_admin_access,
            'mssql': lambda x, y, z: None,
            'postgresql': lambda x, y, z: None,
        }
        _check_admin_access = {
            'mysql': check_admin_access,
            'mssql': lambda x, y: None,
            'postgresql': lambda x, y: None,
        }

        if not _check_admin_access[dbserver.dbtype](dbserver, self):
            try:
                oaserver = AutomationServer(self.conn, ip=dbserver.host)
                _allow_admin_access[dbserver.dbtype](dbserver, self, oaserver)
            except Exception:
                logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                raise MigrationNoRepeatError(_T('UNABLE_TO_ALLOW_ADMIN_ACCESS_TO_DBSERVER').format(
                    host=self.description(), database_server=dbserver.host
                ))

        result = self.plesk_api().send(plesk_ops.DbServerOperator.Add(
            dbserver.host, dbserver.port, dbserver.dbtype, dbserver.login or '', dbserver.password, False
        ))
        if isinstance(result, Failure):
            raise MigrationNoRepeatError(_T('FAILED_REGISTER_DB_SERVER').format(
                host=dbserver.host, port=dbserver.port, dbtype=dbserver.dbtype, login=dbserver.login,
                target_server=self.description()
            ))

        self.get_db_servers.clear()
