# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

import migrator.odin_automation.platform.lib.api.aps.user as aps_user
import migrator.odin_automation.platform.lib.api.aps.account as aps_account

from migrator.core.language import _T
from migrator.core import MigrationError
from migrator.core.utils import FunctionResultWaiter
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Problem
from migrator.core.utils.registry import Registry
from migrator.core.data.model.billing import PaymentSystemType
from migrator.odin_automation.billing.lib.api.bm import BillingAPIError
from migrator.odin_automation.platform.lib.api.aps.account import get_country_rule
from migrator.odin_automation.platform.lib.format_info import InfoFormatter
from migrator.settings import OA_SERVICE_MODULES, SKIP_ACCOUNT_BALANCE_MIGRATION

logger = logging.getLogger(__name__)


class AccountIsNotPresentError(MigrationError):
    pass


class AccountImporterBase(object):
    def __init__(self, panel):
        global_context = Registry.get_instance().get_context()

        self.api_converter = panel.get_api_converter()
        self.target_panel_api = panel.get_connections(global_context).poa_api()

    def import_account(self, account_info, account_billing_info):
        raise NotImplementedError()

    def get_acount_api(self):
        raise NotImplementedError()

    def _create_account(self, account_info, account_billing_info):
        raise NotImplementedError()

    def is_source_account_exported_from_oa(self):
        return Registry.get_instance().get_context().source_type in OA_SERVICE_MODULES

    def create_account(self, account_info, account_billing_info):
        try:
            if account_info.oa_id is not None:
                target_account_info = self.get_account_info(account_info.oa_id)
            else:
                target_account_info = self.get_account_info_by_logins(
                    [x.authentication.login for x in account_info.staff_members]
                )
        except AccountIsNotPresentError:
            target_account_info = None

        # In case we migrate between OA modules, we use an existing OA account to migrate its subscriptions.
        # OAM APS application fills the OAM DB with a minimal account info to keep the account ID only for a certain
        # subscription, so we should not check if an account was changed because it's the same object before and after
        # migration and an account model could not be created using a minimal account info
        if self.is_source_account_exported_from_oa():
            return target_account_info.oa_id

        formatted_account_info = self._format_account_info(account_info)
        if target_account_info is None:
            # The account ID on the destination server is not specified and we need to create or find an account with
            # the same name
            return self._create_account(formatted_account_info, account_billing_info)

        if account_info.oa_id is None and target_account_info.info.admin_contact.email != formatted_account_info.info.admin_contact.email:
            # Accounts with the same username have different email addresses
            raise MigrationError(_T("ACCOUNT_ALREADY_PRESENT").format(
                account_id=target_account_info.oa_id,
                source_account=formatted_account_info.info.admin_contact.email,
                target_account=target_account_info.info.admin_contact.email
            ))

        return target_account_info.oa_id

    def get_account_info(self, account_id):
        account_info = None
        if account_id is not None:
            account_info = self.target_panel_api.account.get_account_info_by_id(account_id)

        if account_info is None:
            raise AccountIsNotPresentError(_T("ACCOUNT_IS_MISSING_FROM_OA").format(account=account_id))

        return self.api_converter.convert_account_data(account_info)

    def get_account_info_by_logins(self, logins):
        account_info = None
        for login in logins:
            account_info = self.target_panel_api.account.get_account_info_by_login(login)
            if account_info is not None:
                return self.api_converter.convert_account_data(account_info)

        if account_info is None:
            raise AccountIsNotPresentError(_T("ACCOUNT_IS_MISSING_FROM_OA").format(account=logins[0]))

    def create_staff_members(self, account_id, account_info):
        for staff_member_info in account_info.staff_members:
            user_info = self.target_panel_api.account.get_staff_member_by_login(staff_member_info.authentication.login)
            if user_info is None:
                try:
                    self._create_staff_member(account_id, staff_member_info)
                except Exception as e:
                    logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                    raise MigrationError(_T("UNABLE_TO_CREATE_STAFF_MEMBER_ACCOUNT_REASON").format(
                        login=staff_member_info.authentication.login,
                        account_id=account_id,
                        reason=unicode(e)
                    ))
            elif user_info.account_id != account_id:
                raise MigrationError(_T("UNABLE_TO_CREATE_STAFF_MEMBER_ACCOUNT").format(
                    login=staff_member_info.authentication.login,
                    account1_id=account_id,
                    account2_id=user_info.account_id,
                ))

    def _create_staff_member(self, account_id, staff_member_info):
        account_api = self.get_acount_api()
        account_api.add_staff_member(account_id, staff_member_info)
        self._wait_until_staff_member_exists(staff_member_info.authentication.login)

    def _update_staff_member(self, user_id, staff_member_info):
        account_api = self.get_acount_api()
        account_api.update_staff_member(user_id, staff_member_info)

    def _update_account(self, account_id, account_info):
        account_api = self.get_acount_api()
        account_api.update_account(account_id, account_info)

    def _format_account_info(self, account_info):
        formatted_account_info, issue_message, is_critical = InfoFormatter().format_account_data(account_info)
        if issue_message is not None:
            if is_critical:
                raise MigrationError(issue_message)
            else:
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.register_error(issue_message, None, Problem.NOTICE)

        return formatted_account_info

    def _wait_until_account_exists(self, account_id):
        function = FunctionResultWaiter(
            self.target_panel_api.account.get_account_info_by_id,
            MigrationError(_T("ACCOUNT_CREATED_IN_BILLING_BUT_MISSING_IN_OA").format(account=account_id))
        )
        function.run(account_id)

    def _wait_until_staff_member_exists(self, login):
        function = FunctionResultWaiter(
            self.target_panel_api.account.get_staff_member_by_login,
            MigrationError(_T("STAFF_MEMBER_CREATED_IN_BILLING_BUT_MISSING_IN_OA").format(login=login))
        )
        function.run(login)


class AccountImporter(AccountImporterBase):
    @trace
    def import_account(self, account_info, account_billing_info):
        account_id = self.create_account(account_info, account_billing_info)
        self.create_staff_members(account_id, account_info)

        # The account appears in Odin Automation only after a staff member for it is created in Billing.
        # The account is created asynchronously; therefore, before creating subscription, make sure that the account
        # already exists in Odin Automation.
        self._wait_until_account_exists(account_id)

        return account_id

    def get_acount_api(self):
        return self.target_panel_api.account

    def _create_account(self, account_info, account_billing_info):
        return self.target_panel_api.account.create_account(account_info)

    def _update_account(self, account_id, account_info):
        try:
            super(AccountImporter, self)._update_account(account_id, account_info)
        except Exception:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.register_error(
                _T("UNABLE_TO_UPDATE_PERSONAL_INFORMATION_FOR_ACCOUNT").format(account_name=account_id),
                None,
                Problem.NOTICE
            )


class AccountImporterPPAB(AccountImporterBase):
    def __init__(self, panel):
        super(AccountImporterPPAB, self).__init__(panel)

        global_context = Registry.get_instance().get_context()
        self.target_panel_billing_api = panel.get_billing_api(global_context)

    @trace
    def import_account(self, account_info, account_billing_info):
        account_id = self.create_account(account_info, account_billing_info)
        self.create_staff_members(account_id, account_info)

        # The account appears in Odin Automation only after a staff member for it is created in Billing.
        # The account is created asynchronously; therefore, before creating subscription, make sure that the account
        # already exists in Odin Automation.
        self._wait_until_account_exists(account_id)

        # There is no OA Billing API to set account's contact language, thus updating it directly in OA Operations
        if not self.is_source_account_exported_from_oa():
            self._set_contacts_locale_aps(account_id, account_info.info.language)

        self.import_account_billing_info(account_id, account_billing_info)
        return account_id

    def get_acount_api(self):
        return self.target_panel_billing_api.account

    def import_account_billing_info(self, account_id, account_billing_info):
        if account_billing_info.balance is not None and not SKIP_ACCOUNT_BALANCE_MIGRATION:
            self.target_panel_billing_api.account.update_account_balance(account_id, account_billing_info.balance)

    def _update_staff_member(self, user_id, staff_member_info):
        super(AccountImporterPPAB, self)._update_staff_member(user_id, staff_member_info)
        # There is no OA Billing API to set account's staff member language, thus updating it directly in OA Operations
        self._set_staff_member_locale_aps(user_id, staff_member_info.language)
        self.target_panel_api.account.update_staff_member(user_id, staff_member_info)

    def _create_staff_member(self, account_id, staff_member_info):
        super(AccountImporterPPAB, self)._create_staff_member(account_id, staff_member_info)
        # There is no OA Billing API to set account's staff member language, thus updating it directly in OA Operations
        user_info = self.target_panel_api.account.get_staff_member_by_login(staff_member_info.authentication.login)
        self._set_staff_member_locale_aps(user_info.user_id, staff_member_info.language)

    def _create_account(self, account_info, account_billing_info):
        tax_reg_id = account_billing_info.tax_reg_id if account_billing_info.tax_reg_id is not None else ''
        account_id = self.target_panel_billing_api.account.add_account(account_info, tax_reg_id)

        country_rule = get_country_rule(account_info.info.company_address.country)
        if country_rule is not None and country_rule['taxRegReqPers'] != 0 and tax_reg_id == '':
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.register_error(
                _T("NEED_TO_SPECIFY_TAX_REGISTRATION_ID").format(
                    account_name=account_id
                ),
                None,
                Problem.NOTICE
            )

        return account_id

    def _update_account(self, account_id, account_info):
        try:
            super(AccountImporterPPAB, self)._update_account(account_id, account_info)
            # OA Billing does not sync new account data with OA Operations,
            # thus updating account info in OA Operations manually
            self.target_panel_api.account.update_account(account_id, account_info)
        except BillingAPIError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            global_context = Registry.get_instance().get_context()
            if 'Tax Registration ID should be specified' in e.message:
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_UPDATE_PERSONAL_INFORMATION_FOR_ACCOUNT_TAX_REG_ID").format(
                        account_name=account_id
                    ),
                    None,
                    Problem.WARNING
                )
            else:
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_UPDATE_PERSONAL_INFORMATION_FOR_ACCOUNT").format(account_name=account_id),
                    None,
                    Problem.NOTICE
                )
        except Exception:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.register_error(
                _T("UNABLE_TO_UPDATE_PERSONAL_INFORMATION_FOR_ACCOUNT").format(account_name=account_id),
                None,
                Problem.WARNING
            )

    @staticmethod
    @trace
    def _set_staff_member_locale_aps(user_id, locale):
        user_info = aps_user.get_user(user_id)
        aps_user.set_user_locale(user_info['aps']['id'], locale)

    @staticmethod
    @trace
    def _set_contacts_locale_aps(account_id, locale):
        account_info = aps_account.get_account(account_id)
        aps_account.set_account_contacts_locale(account_info['aps']['id'], locale)


class AccountImporterPBA(AccountImporterPPAB):
    def import_account_billing_info(self, account_id, account_billing_info):
        if account_billing_info.balance is not None and not SKIP_ACCOUNT_BALANCE_MIGRATION:
            self.target_panel_billing_api.account.update_account_balance(account_id, account_billing_info.balance)
        self.import_bank_accounts(account_id, account_billing_info)
        self.import_credit_cards(account_id, account_billing_info)

    def import_bank_accounts(self, account_id, account_billing_info):
        for bank_account in account_billing_info.bank_accounts:
            try:
                if self._is_bank_account_already_exists(account_id, bank_account):
                    continue

                self.target_panel_billing_api.import_bank_account(account_id, bank_account)
            except MigrationError as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_ADD_BANK_ACCOUNT_ERROR").format(
                        account_name=account_id,
                        account_number=bank_account.account_number,
                        error=e.message
                    ),
                    None,
                    Problem.WARNING
                )
            except Exception:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_ADD_BANK_ACCOUNT").format(
                        account_name=account_id,
                        account_number=bank_account.account_number
                    ),
                    None,
                    Problem.WARNING
                )

    def import_credit_cards(self, account_id, account_billing_info):
        # Business Automation does not support credit cards. For this reason, we will add information about credit cards
        # only if the Plesk Automation billing module is installed.
        for credit_card in account_billing_info.credit_cards:
            try:
                if self._is_credit_card_already_exists(account_id, credit_card):
                    continue

                self.target_panel_billing_api.import_credit_card(account_id, credit_card)
            except MigrationError as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_ADD_CREDIT_CARD_TO_ACCOUNT_ERROR").format(
                        account_name=account_id,
                        credit_card=self._get_credit_card_cut_number(credit_card.credit_card_info.card_number),
                        error=e.message
                    ),
                    None,
                    Problem.WARNING
                )
            except BillingAPIError as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_ADD_CREDIT_CARD_TO_ACCOUNT_ERROR").format(
                        account_name=account_id,
                        credit_card=self._get_credit_card_cut_number(credit_card.credit_card_info.card_number),
                        error=e.short_message
                    ),
                    None,
                    Problem.WARNING
                )
            except Exception:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.register_error(
                    _T("UNABLE_TO_ADD_CREDIT_CARD_TO_ACCOUNT").format(
                        account_name=account_id,
                        credit_card=self._get_credit_card_cut_number(credit_card.credit_card_info.card_number)
                    ),
                    None,
                    Problem.WARNING
                )

    def _is_bank_account_already_exists(self, account_id, bank_account):
        if not self._is_payment_system_supported(bank_account.payment_system):
            raise MigrationError(_T("PAYMENT_SYSTEM_WAS_NOT_SELECTED").format(
                payment_system=bank_account.payment_system
            ))

        for pay_tool in self.target_panel_billing_api.get_pay_tools_for_account(
                account_id, PaymentSystemType.BANK_ACCOUNT
        ):
            if bank_account.bank_number == pay_tool.bank_number and \
                    bank_account.payment_system.lower() == pay_tool.pay_system.lower():
                return True

        return False

    def _is_credit_card_already_exists(self, account_id, credit_card):
        credit_card_cut_number = self._get_credit_card_cut_number(credit_card.credit_card_info.card_number)

        if not self._is_payment_system_supported(credit_card.credit_card_info.payment_system):
            raise MigrationError(_T("PAYMENT_SYSTEM_WAS_NOT_SELECTED").format(
                payment_system=credit_card.credit_card_info.payment_system
            ))

        for pay_tool in self.target_panel_billing_api.get_pay_tools_for_account(
                account_id, PaymentSystemType.CREDIT_CARD
        ):
            if credit_card_cut_number == pay_tool.cut_number and \
                    pay_tool.pay_system.lower() == credit_card.credit_card_info.payment_system.lower():
                return True

        return False

    @staticmethod
    def _get_credit_card_cut_number(card_number):
        return card_number[:4] + '****' + card_number[-4:]

    def _is_payment_system_supported(self, payment_system):
        for credit_card_payment_method in self.target_panel_billing_api.get_payment_methods():
            if credit_card_payment_method.pay_system.lower() == payment_system.lower():
                return True

        return False
