import logging

from migrator.core import _T, MigrationError
from migrator.core.data.model.dns import \
    get_filtered_dns_records_for_zone, DNS_RESOURCE_RECORD_MX, DNS_RESOURCE_RECORD_A, DNS_RESOURCE_RECORD_AAAA, \
    DNS_RESOURCE_RECORD_CNAME, DNS_RESOURCE_RECORD_SRV, DNS_RESOURCE_RECORD_TXT
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.networking.dns import normalize_domain_name
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.aps.domain import update_zone_soa, \
    create_record_a, create_record_aaaa, create_record_cname, create_record_mx, create_record_srv, create_record_txt
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import APS2, get_error
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError


logger = logging.getLogger(__name__)


class DnsRecordsImporter(object):
    def __init__(self):
        global_context = Registry.get_instance().get_context()
        self.poa_api = global_context.conn.target.poa_api()

    @trace
    def import_records(
            self, domain_id, domain_name, zone, global_context,
            target_hosting_subscription=None, dns_subscription=None, link_to=None
    ):
        if not zone.dns_managed:
            logger.debug(_T('SKIP_DNS_RRS_IMPORTING_FOR_EMPTY_ZONE').format(domain_name=domain_name))
            return

        self._remove_existing_records(domain_name)
        # Exclude DNS resource records with 'System' kind from import because pem.deleteDNSRecord which was called
        # above did not delete them and returns 0 exit code. So, attempt to import an already existed records
        # raises an error
        existing_dns_records = self.poa_api.domain.get_records(domain_name)
        records_to_import = get_filtered_dns_records_for_zone(zone, existing_dns_records)
        logger.debug(_T("DEBUG_DNS_RRS_DETECTED_FOR_IMPORT").format(
            domain=domain_name,
            orig_rrs_amount=len(zone.records),
            original_rrs='\n\t'.join([str(zone_record) for zone_record in zone.records]),
            filtered_rrs_amount=len(records_to_import),
            filtered_rrs='\n\t'.join([str(zone_record) for zone_record in records_to_import])
        ))

        aps = APS2.get_connector(1, None, 0)
        for record in records_to_import:
            record_aps = self._import_record(
                domain_id, domain_name, record, global_context,
                target_hosting_subscription if target_hosting_subscription else dns_subscription
            )
            if record_aps and link_to:
                aps.link_resource(link_to, 'records', record_aps['aps']['id'])

        self._update_soa(dns_subscription.account_id, domain_id, zone.soa)
        if zone.enabled:
            self.poa_api.domain.enable_domain(domain_name=domain_name)
        else:
            self.poa_api.domain.disable_domain(domain_name=domain_name)


    @trace
    def _import_record(self, domain_id, domain_name, record, global_context, subscription):
        """
        :param domain_name: str
        :param record: migrator.core.data.model.dns.DnsRecordModel Examples:
            DnsRecordModel(priority=None, host='example.com', type='txt', value='v=spf1 +a +mx -all +a:domain.tld', ttl='600')
            DnsRecordModel(priority=None, host='ftp.example.com', type='cname', value='orange-with-dns.com.', ttl='600')
            DnsRecordModel(priority=None, host='ns1.example.com', type='a', value='10.37.54.122', ttl='600')
            DnsRecordModel(priority='10', host='example.com', type='mx', value='mail.orange-with-dns.com', ttl='600')
        :param global_context: migrator.core.session.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: None
        """
        dns_record_type = record.type.lower()
        dns_record_val = record.value.lower()
        dns_record_host = record.host.lower()
        dns_record_ttl = record.ttl
        dns_record_priority = record.priority

        if record.is_edit_allowed:
            dns_record_subscription = None
            logger.debug("Manual DNS records of domain '%s' will not be linked to any subscription" % (domain_name))
        else:
            dns_record_subscription = subscription.id
            logger.debug("System DNS records of domain '%s' will be linked to subscription #%s" % (domain_name, dns_record_subscription))

        record_aps = None
        try:
            logger.debug(_T("DEBUG_CREATING_DNS_RR").format(
                rr_type=dns_record_type,
                rr_val=dns_record_val,
                rr_host=dns_record_host,
                domain=domain_name
            ))

            record_host_prefix = self._get_record_host_prefix(domain_name, dns_record_host)

            if dns_record_type == DNS_RESOURCE_RECORD_A:
                record_aps = create_record_a(
                    subscription.account_id, domain_id, record_host_prefix,
                    address=dns_record_val, ttl=dns_record_ttl, subscription_id=dns_record_subscription
                )
            elif dns_record_type == DNS_RESOURCE_RECORD_AAAA:
                record_aps = create_record_aaaa(
                    subscription.account_id, domain_id, record_host_prefix,
                    address=dns_record_val, ttl=dns_record_ttl, subscription_id=dns_record_subscription
                )
            elif dns_record_type == DNS_RESOURCE_RECORD_CNAME:
                record_aps = create_record_cname(
                    subscription.account_id, domain_id, record_host_prefix,
                    cname=dns_record_val, ttl=dns_record_ttl, subscription_id=dns_record_subscription
                )
            elif dns_record_type == DNS_RESOURCE_RECORD_MX:
                record_aps = create_record_mx(
                    subscription.account_id, domain_id, record_host_prefix,
                    exchange=dns_record_val, priority=dns_record_priority, ttl=dns_record_ttl, subscription_id=dns_record_subscription
                )
            elif dns_record_type == DNS_RESOURCE_RECORD_SRV:
                weight, port, target = dns_record_val.split(' ')
                record_aps = create_record_srv(
                    subscription.account_id, domain_id, record_host_prefix,
                    priority=dns_record_priority, weight=weight, port=port, target=target, ttl=dns_record_ttl, subscription_id=dns_record_subscription
                )
            elif dns_record_type == DNS_RESOURCE_RECORD_TXT:
                record_aps = create_record_txt(
                    subscription.account_id, domain_id, record_host_prefix,
                    txt_data=dns_record_val, ttl=dns_record_ttl, subscription_id=dns_record_subscription
                )
            else:
                logger.warning(_T("ATTEMPT_TO_IMPORT_UNSUPPORTED_DNS_RR").format(
                    rr_type=dns_record_type,
                    rr_val=dns_record_val,
                    rr_host=dns_record_host,
                    domain=domain_name
                ))

            logger.debug(_T("DEBUG_DNS_RR_CREATED").format(
                rr_type=dns_record_type,
                rr_val=dns_record_val,
                rr_host=dns_record_host,
                domain=domain_name
            ))
        except APSCTransportError as e:
            # The record already exists
            if get_error(e) == "Plesk.ErrorHandling.dns.DuplicateDNSRecord":
                msg = _T("SKIP_DNS_RR_IMPORTING").format(
                    rr_type=dns_record_type,
                    rr_val=dns_record_val,
                    rr_host=dns_record_host,
                    domain=domain_name,
                    error_details=e
                )
                logger.warning(msg)
                global_context.execution_wrapper.add_subscription_issue(
                    subscription.source_info.external_id,
                    Issue(Problem('rr-creation-skipped', Problem.NOTICE, msg), None)
                )
            else:
                msg = _T("FAILED_TO_CREATE_DNS_RR").format(
                    rr_type=dns_record_type,
                    rr_val=dns_record_val,
                    rr_host=dns_record_host,
                    domain=domain_name,
                    error_details=e
                )
                logger.warning(msg)
                global_context.execution_wrapper.add_subscription_issue(
                    subscription.source_info.external_id,
                    Issue(Problem('rr-creation-failed', Problem.WARNING, msg), None)
                )
        except Exception as e:
            msg = _T("FAILED_TO_CREATE_DNS_RR").format(
                rr_type=dns_record_type,
                rr_val=dns_record_val,
                rr_host=dns_record_host,
                domain=domain_name,
                error_details=e
            )
            logger.warning(msg)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem('rr-creation-failed', Problem.WARNING, msg), None)
            )

        return record_aps

    @trace
    def _remove_existing_records(self, domain_name):
        existing_records = self.poa_api.domain.get_records(domain_name)
        for record in existing_records:
            self.poa_api.domain.delete_record(record['record_id'])

    @staticmethod
    @trace
    def _update_soa(account_id, domain_id, soa_record):
        update_zone_soa(
            account_id, domain_id, soa_record.email, soa_record.refresh, soa_record.retry, soa_record.expire,
            soa_record.minimum
        )

    @staticmethod
    @trace
    def _get_record_host_prefix(domain_name, record_host):
        """
        Extract prefix of DNS record host from DNS record host FQDN
        Examples:

        _get_record_prefix('domain.tld', 'sub.domain.tld') == 'sub'
        _get_record_prefix('domain.tld', 'domain.tld') == ''
        """
        normalized_domain_name = normalize_domain_name(domain_name)
        normalized_record_host = normalize_domain_name(record_host)
        if not normalized_record_host.endswith(normalized_domain_name):
            raise MigrationError(_T("DNS_RECORD_DOES_NOT_BELONG_TO_ZONE").format(
                record_host=record_host,
                zone_name=domain_name
            ))

        return normalized_record_host[:-len(normalized_domain_name)].rstrip('.')
