# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import _T, InternalError, MigrationError, MigrationNoRepeatError
from migrator.core.data.model.subscription import get_services, get_service_property
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils import cached
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.odin_automation.platform.lib.api.aps.domain import \
    create_pa_domain, get_pa_domain_by_name, get_pa_domain_records
from migrator.core.utils.networking.dns import normalize_domain_name
from migrator.odin_automation.platform.lib.api.aps.account import get_account_vendor_id
from migrator.odin_automation.platform.lib.api.aps.types import PADnsARecord, PADnsAAAARecord
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError
from migrator.odin_automation.platform.lib.db.domain import add_php_plugin_link_for_domain
from migrator.settings import \
    SOURCE_SERVICE_DNS, SOURCE_SERVICE_DNS_NS, SOURCE_SERVICE_DNS_CUSTOM_NS, SOURCE_SERVICE_DNS_NON_MANAGEABLE_NS,\
    OA_TARGET_OWNER_ID

logger = logging.getLogger(__name__)


class DomainImporterBase(object):

    @trace
    def import_domain(self, domain_name, target_hosting_subscription=None, dns_subscription=None):
        """
        Import domain domain_name. Either target_hosting_subscription or dns_subscription should be supplied, the other
        one will be found in migration project
        """
        raise NotImplementedError()


class DomainImporter(DomainImporterBase):
    def __init__(self, global_context):
        self._global_context = global_context

    @trace
    def import_domain(self, domain_name, target_hosting_subscription=None, dns_subscription=None):
        """
        Import domain domain_name. Either target_hosting_subscription or dns_subscription should be supplied, the other
        one will be found in migration project

        :param domain_name str
        :param target_hosting_subscription None|
        :param dns_subscription None|migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        """
        logger.info(_T("IMPORTING_DOMAIN").format(name=domain_name))
        if target_hosting_subscription is None and dns_subscription is None:
            raise InternalError(
                _T("NEITHER_DOMAIN_NOR_HOSTING_SUBSCRIPTION_WAS_PROVIDED").format(domain_name=domain_name)
            )

        target_subscription = dns_subscription or target_hosting_subscription

        domain = get_pa_domain_by_name(domain_name, account_id=target_subscription.account_id)
        if domain is None:
            parent_domain_id = self._get_parent_aps_domain_id(domain_name, target_subscription.account_id)

            try:
                domain = create_pa_domain(target_subscription.account_id, target_subscription.id, domain_name, parent_domain_id)
                logger.debug(_T("CREATED_DOMAIN").format(name=domain_name, id=domain['aps']['id']))
            except APSCTransportError as e:
                if self.get_domain_level(domain_name) > 2:
                    if isinstance(e.body, dict) and 'error' in e.body and isinstance(e.body['error'], basestring) and \
                            e.body['error'] == 'Plesk.ErrorHandling.dns.DomainIsChildFor' and 'details' in e.body and \
                            isinstance(e.body['details'], dict) and 'domain_id' in e.body['details']:
                        domain_id = int(e.body['details']['domain_id'])
                        conflicting_parent_domains_ids = self.get_all_parent_domains_ids(domain_name)
                        if domain_id in conflicting_parent_domains_ids:
                            raise MigrationNoRepeatError(_T('CONFLICT_BETWEEN_DOMAINS').format(
                                child_domain=domain_name,
                                parent_domain=conflicting_parent_domains_ids[domain_id]
                            ))

                raise
        else:
            logger.debug(_T("DOMAIN_ALREADY_EXISTS").format(name=domain_name, id=domain['aps']['id']))

        return domain

    @trace
    def _get_target_subscription(self, dns_subscription, hosting_subscription):
        if dns_subscription and (dns_subscription.is_mapped_to_self() or dns_subscription.mapping.is_oa_subscription):
            target_subscription = dns_subscription
        elif dns_subscription is not None:
            target_subscription = hosting_subscription or self._get_subscription(
                dns_subscription.source_info.parent_external_id
            )
        else:
            target_subscription = hosting_subscription

        if not target_subscription.is_mapped_to_self() and not target_subscription.mapping.is_oa_subscription:
            target_subscription = self._get_subscription(target_subscription.mapping.mapping_subscription_id)

        return target_subscription

    @classmethod
    @trace
    def get_dns_subscription(cls, global_context, parent_external_id, domain_name):
        """
        Find DNS subscription for domain_name or its parent domain of highest possible level linked to
        hosting_subscription in migration list
        """
        dns_subscriptions = {}
        for sub in global_context.subscriptions_to_migrate:
            if sub.source_info.parent_external_id == parent_external_id:
                # if current subscription is a child subscription of hosting_subscription
                subscription_domain = get_service_property(sub.source_info, SOURCE_SERVICE_DNS, 'domain')
                if subscription_domain is not None:
                    if subscription_domain == domain_name:
                        return sub

                    dns_subscriptions[subscription_domain] = sub

        parent_domain = cls._get_highest_level_parent_domain(domain_name, dns_subscriptions.keys())
        return dns_subscriptions[parent_domain] if parent_domain is not None else None

    @staticmethod
    def _get_highest_level_parent_domain(target_domain, domains):
        """
        From list of domains select parent domain of highest possible level for target_domain.
        """
        domain_levels = target_domain.split('.')
        possible_parent_domain = domain_levels.pop()
        while possible_parent_domain != target_domain:
            if possible_parent_domain in domains:
                return possible_parent_domain

            possible_parent_domain = domain_levels.pop() + '.' + possible_parent_domain

        return None

    @staticmethod
    @cached
    @trace
    def get_tld_billing(dns_zone):
        """
        Get top level domain for the passed dns_zone in terms of Billing. Examples:
        - for the zone 'example.com' TLD is 'com'
        - fro the zone 'a.b.c.d.com' TLD is 'b.c.d.com'

        :param dns_zone: str Examples: u'example.com', u'sub.example.ru', u'com'
        :return: str Examples: u'com', u'example.ru'
        """
        return dns_zone if DomainImporter.get_domain_level(dns_zone) == 1 else dns_zone.split(u'.', 1)[1]

    @trace
    def _get_subscription(self, external_id):
        for sub in self._global_context.subscriptions_to_migrate:
            if sub.source_info.external_id == external_id:
                return sub

        raise InternalError("Subscription ID '%s' was not found in migration project" % external_id)

    @staticmethod
    @trace
    def _get_parent_aps_domain_id(target_domain, account_id):
        domain_levels = target_domain.split('.')
        possible_parent_domain = domain_levels.pop()
        while possible_parent_domain != target_domain:
            domain = get_pa_domain_by_name(possible_parent_domain, account_id=account_id)
            if domain:
                return domain['aps']['id']

            possible_parent_domain = domain_levels.pop() + '.' + possible_parent_domain

        return None

    @staticmethod
    @cached
    @trace
    def get_domain_level(domain_name):
        """
        :param domain_name: str Examples: 'com', 'example.com', 'sub1.sub2.example.com'
        :return: int Examples: 'com' => 1, 'example.com' => 2, 'sub1.sub2.example.com' => 4
        """
        return domain_name.count(u'.') + 1

    @cached
    @trace
    def get_all_possible_parent_domains(self, domain_name):
        """
        Determine and return the domains of the all possible parent DNS zones, except top-level domain

        :param domain_name: str 'red.red.red.com'
        :return: list[str] Example: ['red.red.com', 'red.com']
        """
        result = []
        while self.get_domain_level(domain_name) > 2:
            domain_name = domain_name.split(u'.', 1)[1]
            result.append(domain_name)

        return result

    @trace
    def get_all_parent_domains_ids(self, domain_name):
        """
        Get IDs and names of all existing potential parent domains for the domain domain_name in OA

        :param domain_name: str Examples: 'sub.example.com', 'sub1.sub2.example.com'
        :return: dict Map of Domain ID to domain name. Example: {49: 'sub2.example.com', 50: 'example.com'}
        """
        result = {}
        for possible_parent_domain in self.get_all_possible_parent_domains(domain_name):
            oa_parent_domain = get_pa_domain_by_name(possible_parent_domain)
            if oa_parent_domain is not None:
                result[oa_parent_domain['zoneId']] = oa_parent_domain['name']

        return result

    @cached
    @trace
    def get_record_of_child_domain_equal_to_parent_domain_fqdn(self, child_domain, parent_domain):
        """
        :param child_domain: str Example: 'child.sub-parent.example.com'
        :param parent_domain: str Example: 'example.com'
        :return: str Example: 'child.sub-parent'
        """
        if self.get_domain_level(child_domain) <= self.get_domain_level(parent_domain):
            raise InternalError(_T('INVALID_ATTEMPT_TO_DETECT_CONFLICTS_BETWEEN_DOMAIN_AND_SUBDOMAIN').format(
                parent_domain=parent_domain,
                subdomain=child_domain
            ))

        return child_domain.rsplit(u'.%s' % parent_domain, 1)[0]

    @cached
    @trace
    def is_wildcard_subdomain(self, zone_name):
        """
        Check if a zone_name is a wildcard subdomain

        :param zone_name: str Examples: 'example.com', '*.example.com', '*.sub.example.com'
        :return: bool
        """
        return u'*' in zone_name and self.get_domain_level(zone_name) > 2

    @trace
    def is_wildcard_subdomain_exist_in_oa(self, wildcard_zone):
        """
        Try to find a wildcard A/AAAA records in all possible parent zones in OA for the passed wildcard_zone starting
        from the high-leveled parent zone.

        Example:
        For the wildcard_zone '*.sub.example.com' will be checked the following records in the following order:
        1) zone 'sub.example.com' has a DNS record with host '*' and type A or AAAA
        2) zone 'example.com' has a DNS record with host '*.sub' and type A or AAAA
        If any of the records were found we say that the wildcard subdomain wildcard_zone already exists

        :param wildcard_zone: str Example: u'*.sub.example.com'
        :return: bool True if wildcard subdomain wildcard_zone already exists in OA
        """
        logger.debug(_T('CHECK_IF_WILDCARD_SUBDOMAIN_EXISTS').format(subdomain=wildcard_zone))
        possible_parent_domains = self.get_all_possible_parent_domains(wildcard_zone)
        for possible_parent_domain in possible_parent_domains:
            possible_wildcard_record = self.get_record_of_child_domain_equal_to_parent_domain_fqdn(
                wildcard_zone, possible_parent_domain
            )
            found_wildcard_records = \
                get_pa_domain_records(possible_parent_domain, PADnsARecord, possible_wildcard_record) + \
                get_pa_domain_records(possible_parent_domain, PADnsAAAARecord, possible_wildcard_record)
            if len(found_wildcard_records) > 0:
                logger.debug(_T('WILDCARD_SUBDOMAIN_EXISTS').format(
                    subdomain=wildcard_zone,
                    domain=possible_parent_domains
                ))
                return True

        logger.debug(_T('WILDCARD_SUBDOMAIN_DOES_NOT_EXIST').format(subdomain=wildcard_zone))
        return False

    @trace
    def create_wildcard_subdomain(self, wildcard_zone, ip_address):
        """
        Create the wildcard subdomain wildcard_zone in the highest-leveled parent zone.

        Example:
        For the wildcard_zone '*.sub.example.com' the following actions will be executed in the sorted order:
        1) Create A record with value 'ip_address' and host '*' in the zone 'sub.example.com' if it exists
        2) Create A record with value 'ip_address' and host '*.sub' in the zone 'example.com' if it exists

        :param wildcard_zone: str Example: u'*.sub.example.com'
        :param ip_address: str u'8.8.8.8'
        :return: void
        """
        logger.debug(_T('CREATING_WILDCARD_SUBDOMAIN').format(subdomain=wildcard_zone, ip_v4=ip_address))
        possible_parent_domains = self.get_all_possible_parent_domains(wildcard_zone)
        # Iterate all possible parent domains from the highest leveled domain to the lowest one
        for possible_parent_domain in possible_parent_domains:
            pa_domain = get_pa_domain_by_name(possible_parent_domain)
            if pa_domain is None:
                continue

            possible_wildcard_record = self.get_record_of_child_domain_equal_to_parent_domain_fqdn(
                wildcard_zone, possible_parent_domain
            )
            logger.debug(_T('FOUND_DOMAIN_FOR_WILDCARD_RECORD_CREATION').format(
                domain=possible_parent_domain,
                subdomain=wildcard_zone,
                wildcard_record=possible_wildcard_record
            ))
            self._global_context.conn.target.poa_api().domain.create_record_a(
                possible_parent_domain,
                possible_wildcard_record,
                ip_address
            )
            logger.debug(_T('WILDCARD_SUBDOMAIN_CREATED').format(
                subdomain=wildcard_zone,
                host=possible_wildcard_record,
                ip=ip_address,
                domain=possible_parent_domain
            ))
            return

        raise MigrationError(_T('FAILED_TO_CREATE_WILDCARD_SUBDOMAIN').format(subdomain=wildcard_zone))


class DomainImporterPPAB(DomainImporter):
    def __init__(self, global_context):
        super(DomainImporterPPAB, self).__init__(global_context)
        self._target_panel_billing_api = global_context.target_panel_obj.get_billing_api(global_context)

    @trace
    def import_domain(self, domain_name, target_hosting_subscription=None, dns_subscription=None):
        domain = super(DomainImporterPPAB, self).import_domain(
            domain_name, target_hosting_subscription, dns_subscription
        )
        target_subscription = dns_subscription or target_hosting_subscription
        self._target_panel_billing_api.subscription.add_domain_for_subscription(domain_name, target_subscription.id)

        if dns_subscription:
            normalized_domain_name = normalize_domain_name(domain_name)
            domain_ext = self._target_panel_billing_api.domain.get_domain_info(domain_name).domain_zone_id

            subscription_details = self._target_panel_billing_api.subscription.subscription_details_get(target_subscription.id)
            account_id_with_configured_plugin = self.get_account_id_with_configured_php_plugin(subscription_details.plan_id)
            if self.is_php_plugin_configured(domain_ext, account_id_with_configured_plugin):
                add_php_plugin_link_for_domain(normalized_domain_name, domain_ext, account_id_with_configured_plugin)

            ns_list = []
            if get_services(dns_subscription.source_info, SOURCE_SERVICE_DNS_NS) or \
               get_services(dns_subscription.source_info, SOURCE_SERVICE_DNS_NON_MANAGEABLE_NS):
                ns_list.extend(domain.get('nameServers', []))
            else:
                for service in get_services(dns_subscription.source_info, SOURCE_SERVICE_DNS_CUSTOM_NS):
                    source_host = PlatformReaderDB().get_source_host(service.source_host_external_id)
                    ns_list.append(source_host.name)

            logger.debug(u'domain: {domain}, nameservers: {nameservers}'.format(domain=domain_name, nameservers=ns_list))
            if len(ns_list) >= 2:
                self._target_panel_billing_api.domain.update_name_servers(domain_name, *ns_list[:4])
            else:
                msg = _T('LIST_OF_NAMESERVERS_WILL_NOT_BE_UPDATED').format(domain=domain_name)

                self._global_context.execution_wrapper.add_subscription_issue(
                    dns_subscription.source_info.external_id, Issue(Problem('update-nameservers-list-notice', Problem.NOTICE, msg), None)
                )

        return domain

    def is_php_plugin_configured(self, domain_ext, account_id):
        domain_zone_list = self._target_panel_billing_api.domain.get_full_domain_zone_list(account_id)
        for domain_zone in domain_zone_list:
            if domain_zone.DomainZoneID.lower() == domain_ext.lower() and domain_zone.Container.upper() == 'PEM_REGISTRAR':
                return True

        return False

    def get_account_id_with_configured_php_plugin(self, plan_id):
        if self._target_panel_billing_api.is_plan_delegated(plan_id)[0]:
            return get_account_vendor_id(OA_TARGET_OWNER_ID)

        return OA_TARGET_OWNER_ID


class DomainImporterPBA(DomainImporterPPAB):
    pass
