# (C)2017 Ingram Micro Inc. all Rights Reserved
import datetime
import itertools
import logging

from sqlalchemy import func
from sqlalchemy.orm import aliased

from migrator.core import MigrationError
from migrator.core.data.format.platform import PagedPlatformReader
from migrator.core.data.model import \
    AccountModel, SourceHostModel, ServicePlanModel, SourceSystemModel, MigrationProjectItemModel, SubscriptionModel
from migrator.core.data.source import db_meta
from migrator.core.data.source.db_meta import get_db_session
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


@trace
def get_loader(target_service_type, source_system_aps_id):
    """Construct a loader instance depending on the type of exporter object"""
    global_context = Registry.get_instance().get_context()
    exporter = global_context.source_product.get_exporter_instance(target_service_type)

    loader_class = PagedLoader if isinstance(exporter, PagedPlatformReader) else RecursiveLoader
    logger.debug("constructing a '%s' object for '%s'", loader_class, exporter.__class__)
    return loader_class(exporter, source_system_aps_id)


class Loader(object):
    """Populate migrator database using data from an Exporter object"""

    @trace
    def __init__(self, exporter, aps_id):
        self.exporter = exporter
        self.platform_reader = PlatformReaderDB()
        self.source_system_aps_id = aps_id
        self.session = get_db_session()

    def export(self):
        raise NotImplementedError()

    def refresh(self, account_ids=None, host_ids=None):
        raise NotImplementedError()

    @trace
    def _query_existing_source_system_object(self):
        """Retrieve an existing source system object from the database by its aps_id"""
        source_system = self.platform_reader.get_source_system_model()

        if not source_system:
            raise MigrationError(
                _T("REFRESH_DATA_INVALID_SOURCE_SYSTEM_APS_ID").format(aps_id=self.source_system_aps_id, system_count=0)
            )
        else:
            return source_system

    @trace
    def _sync_subscription_data(self, source_system, new_subscriptions):
        """Match old/new subscriptions and copy old subscriptions data to the new ones"""

        current_subscriptions = []
        for account in itertools.chain(source_system.accounts, source_system.providers, source_system.resellers):
            current_subscriptions.extend(account.subscriptions)

        for existing_subscription in current_subscriptions:
            new_subscription_record = self._match_subscription(existing_subscription, new_subscriptions, self.session)
            if new_subscription_record:
                new_subscription_record.clone(existing_subscription)

    @trace
    def _sync_project_item_subscriptions(self, source_system, new_subscriptions):
        """
        Update references from existing project items to new subscriptions; delete obsolete ones

        Return: a list of project items, which subscriptions no longer exist (=deleted from source system)
        """
        existing_project_items = self._get_migration_project_items(source_system, self.session)
        for project_item in existing_project_items:
            existing_subscription = self.session.query(SubscriptionModel).filter(
                SubscriptionModel.id == project_item.subscription_id
            ).one()
            new_subscription_record = self._match_subscription(existing_subscription, new_subscriptions, self.session)
            if new_subscription_record is not None:
                project_item.subscription = new_subscription_record

    @staticmethod
    @trace
    def _get_migration_project_items(source_system, session):
        """Return all migration project items related to the given source system"""
        existing_subscriptions = []
        for account in itertools.chain(source_system.accounts, source_system.providers, source_system.resellers):
            existing_subscriptions.extend(account.subscriptions)

        if not existing_subscriptions:
            return []

        existing_subscription_ids = [s.id for s in existing_subscriptions]
        project_items = session.query(MigrationProjectItemModel).filter(
            MigrationProjectItemModel.subscription_id.in_(existing_subscription_ids))

        return project_items.all()

    @staticmethod
    def _match_subscription(existing_subscription, subscription_list, session):
        """Return a subscription from the list, that matches an existing subscription."""
        matches = [new for new in subscription_list if new == existing_subscription]
        if len(matches) == 1:
            matched_subscription = matches[0]
            if existing_subscription.external_id != matched_subscription.external_id:
                # If matched_subscription passed 'eq' comparison and external_id differs means that it is DNS
                # subscription that changed from Paid to Free or vice versa.
                Loader._repair_project_items(existing_subscription, matched_subscription, session)
                existing_subscription.external_id = matched_subscription.external_id

            return matched_subscription
        elif len(matches) > 1:
            raise MigrationError(
                _T("REFRESH_DATA_UNABLE_TO_MATCH_SUBSCRIPTION").format(
                    subscription=existing_subscription.name, found_matches=len(matches))
            )
        else:
            return None

    @staticmethod
    @trace
    def _repair_project_items(existing_sub, new_sub, session):
        """Change mapping_subscription_id for items that mapped to subscription external_id of which was changed"""
        logger.warning(
            _T("REFRESH_DATA_SUBSCRIPTION_THAT_CHANGED_EXTERNAL_ID_FOUND").format(
                name=existing_sub.name, old_external_id=existing_sub.external_id, new_external_id=new_sub.external_id)
        )

        project_items = session.query(MigrationProjectItemModel).filter(
            MigrationProjectItemModel.mapping_subscription_id == existing_sub.external_id)
        for project_item in project_items:
            project_item.mapping_subscription_id = new_sub.external_id


class RecursiveLoader(Loader):
    """Export and refresh all data at once"""
    @trace
    def export(self):
        source_system = SourceSystemModel(
            resellers=self.exporter.get_resellers_list(),
            accounts=self.exporter.get_accounts_owned_by(owner_id=None),
            source_hosts=self.exporter.get_source_hosts_list(),
            service_plans=self.exporter.get_service_plans_list()
        )

        source_system.aps_id = self.source_system_aps_id
        db_meta.save_to_db(source_system)

    @trace
    def refresh(self, account_ids=None, host_ids=None):

        # create newly exported objects in the database
        new_resellers = self.exporter.get_resellers_list()
        new_accounts = self.exporter.get_accounts_owned_by(owner_id=None)
        new_source_hosts = self.exporter.get_source_hosts_list()
        new_service_plans = self.exporter.get_service_plans_list()

        self.session.add_all(new_resellers)
        self.session.add_all(new_accounts)
        self.session.flush()

        new_subscriptions = []
        for a in itertools.chain(new_accounts, new_resellers):
            new_subscriptions.extend(a.subscriptions)

        source_system = self._query_existing_source_system_object()
        self._sync_subscription_data(source_system, new_subscriptions)
        self._sync_project_item_subscriptions(source_system, new_subscriptions)

        # drop existing objects
        models_to_delete = [AccountModel, SourceHostModel, ServicePlanModel]
        for model_object in models_to_delete:
            delete_me = self.session.query(model_object).filter(model_object.source_system_id == source_system.id).all()
            map(self.session.delete, delete_me)

        for r in new_resellers:
            r.reseller_source_system = source_system
        for a in new_accounts:
            a.account_source_system = source_system
        for a in new_source_hosts:
            a.source_system = source_system
        for a in new_service_plans:
            a.source_system = source_system

        self.session.commit()


class PagedLoader(Loader):
    def __init__(self, exporter, aps_id, page_size=50):
        super(PagedLoader, self).__init__(exporter, aps_id)
        self.exporter.page_size = page_size

    def export(self):
        """Load account data from a new source system in fixed length batches."""
        self._export_source_system()
        self.refresh()

    @trace
    def refresh(self, account_ids=None, host_ids=None):
        validity_start_period = datetime.datetime.now()
        source_system = self._query_existing_source_system_object()

        self._cleanup_migration_project_items(source_system)
        self.session.commit()

        with self.exporter.filter_by_hosts(host_ids=host_ids),\
             self.exporter.filter_by_accounts(account_ids=account_ids):

            self._export_source_hosts(validity_start_period, source_system)
            self.session.commit()

            self._export_service_plans(validity_start_period, source_system)
            self.session.commit()

            self._export_accounts(validity_start_period, source_system)
            self.session.commit()

        self._cleanup_migration_project_items(source_system)
        self.session.commit()

    @trace
    def _export_source_system(self):
        """Create a new source system object and store it in the database"""
        existing_source_system = self.platform_reader.get_source_system_model()
        if existing_source_system is None:
            source_system = SourceSystemModel()
            source_system.aps_id = self.source_system_aps_id
            self.session.add(source_system)
            self.session.commit()

    @trace
    def _export_source_hosts(self, validity_start_period, source_system):
        new_source_hosts = self.exporter.get_source_hosts_list()
        for a in new_source_hosts:
            a.source_system = source_system

        self.session.flush()

        self._shift_validity_period(
            SourceHostModel, source_system.id, validity_start_period,
            [x.external_id for x in new_source_hosts] if self.exporter.is_partial else None
        )

    @trace
    def _export_service_plans(self, validity_start_period, source_system):
        new_source_plans = self.exporter.get_service_plans_list()
        for a in new_source_plans:
            a.source_system = source_system

        self.session.flush()

        self._shift_validity_period(ServicePlanModel, source_system.id, validity_start_period)

    @trace
    def _export_accounts(self, validity_start_period, source_system):
        """Export accounts page by page, then replace current accounts with the new ones"""
        for account_batch in self.exporter.export_next_account_page():
            self._save_accounts(account_batch, source_system, self.exporter.is_partial)

        external_ids = self.exporter._accounts_and_owners.keys() if self.exporter.is_partial else None

        self._renew_accounts_with_production_subscriptions(source_system.id, external_ids)
        self._renew_production_subscriptions(source_system.id, external_ids)

        self._shift_validity_period(AccountModel, source_system.id, validity_start_period, external_ids)

    @trace
    def _update_subscriptions(self, new_accounts, source_system):
        """Replace references to the old subscriptions with new ones; copy subscription table data"""
        new_subscriptions = []
        for a in new_accounts:
            new_subscriptions.extend(a.subscriptions)

        if not new_subscriptions:
            return

        self._sync_subscription_data(source_system, new_subscriptions)
        self._sync_project_item_subscriptions(source_system, new_subscriptions)
        self.session.flush()

    @trace
    def _renew_accounts_with_production_subscriptions(self, source_system_id, external_account_ids=None):
        """
        Renew accounts which have production subscriptions.

        Since we export only accounts with active subscriptions (except provider & resellers), we need to keep alive
        accounts already registered in db which have only production subscriptions.

        Otherwise, on next refresh data, they will be purged from db and completed migration projects become invalid
        along with overall source system statistics.
        """
        accounts_to_renew = self.session.query(AccountModel).join(SubscriptionModel).\
            filter(AccountModel.source_system_id == source_system_id).\
            filter(SubscriptionModel.in_production)
        accounts_to_renew = db_meta.current_record(AccountModel, accounts_to_renew)

        if external_account_ids is not None:
            accounts_to_renew = accounts_to_renew.filter(AccountModel.external_id.in_(external_account_ids))

        # TODO: paging
        for account in accounts_to_renew:
            new_account = self.get_account(account.external_id, source_system_id, db_meta.new_record)
            if new_account:
                # account already present in db as new account, so do nothing
                continue

            # Update owner.
            # Since resellers and accounts are exported despite of existence of own subscriptions,
            # we do not check for get_account() result. It should always return valid account
            owner_external_id = account.owner.external_id
            account.owner = self.get_account(owner_external_id, source_system_id, db_meta.new_record)
            if account.owner is None:
                account.owner = self.get_account(owner_external_id, source_system_id, db_meta.current_record)
                account.owner.valid_since = None
                account.owner.valid_until = None

            account.valid_since = None

        self.session.flush()

    @trace
    def _renew_production_subscriptions(self, source_system_id, external_account_ids=None):
        """
        Renew production subscriptions which belong to partially migrated accounts

        See docstring in _renew_accounts_with_production_subscriptions() for details
        """
        subscription_curr = aliased(SubscriptionModel)
        subscription_arch = aliased(SubscriptionModel)

        account_of_subscr_curr = aliased(AccountModel)
        account_of_subscr_arch = aliased(AccountModel)

        production_subscriptions = self.session. \
            query(subscription_curr).\
            join(account_of_subscr_curr, account_of_subscr_curr.id == subscription_curr.account_id). \
            outerjoin(subscription_arch, subscription_arch.external_id == subscription_curr.external_id). \
            outerjoin(account_of_subscr_arch, account_of_subscr_arch.id == subscription_arch.account_id). \
            filter(account_of_subscr_curr.source_system_id == source_system_id). \
            filter(account_of_subscr_arch.source_system_id == source_system_id). \
            filter(subscription_curr.in_production)
        production_subscriptions = db_meta.current_record(account_of_subscr_curr, production_subscriptions)

        if external_account_ids is not None:
            production_subscriptions = production_subscriptions.filter(account_of_subscr_curr.external_id.in_(external_account_ids))

        production_subscriptions = production_subscriptions.group_by(subscription_curr).having(func.count() == 1)

        # TODO: paging
        for subscription in production_subscriptions:
            # If there are more than 1 copy of subscription don't re-link it because new exported subscription will be
            # correctly linked
            # re-link subscription with actual owner
            subscription.account = self.get_account(
                subscription.account.external_id, source_system_id, db_meta.new_record
            )

        self.session.flush()

    @trace
    def _cleanup_migration_project_items(self, source_system):
        """
        Drop migration project items which refer to outdated accounts.

        The cleanup operation drops all migration project items records (db_meta.migration_project_item)
        which refer to subscriptions (db_meta.subscription) which belongs to accounts (db_meta.account)
        which are no longer valid (db_meta.archived_records)
        """
        items_to_delete = self.session.query(MigrationProjectItemModel).\
            join(SubscriptionModel).join(AccountModel).\
            filter(AccountModel.source_system_id == source_system.id)
        items_to_delete = db_meta.archived_record(AccountModel, items_to_delete)

        logger.debug("Project items that has no referring subscriptions: '%s'", [p.id for p in items_to_delete])
        map(self.session.delete, items_to_delete)

        self.session.flush()

    @trace
    def _shift_validity_period(self, model_object, source_system_id, current_time, external_ids=None):
        """Close validity period of current records; start validity period of new records"""
        records = self.session.query(model_object).filter(model_object.source_system_id == source_system_id)
        if external_ids is not None:
            records = records.filter(model_object.external_id.in_(external_ids))

        # Remove existing objects which are no longer valid.
        delete_records = db_meta.archived_record(model_object, records)
        count = delete_records.delete(synchronize_session=False)
        logger.debug("Deleted %s '%s' records of total %s" % (count, model_object.__name__, count))

        # close validity period of the existing records
        logger.debug("Purging %s records at _shift_validity_period" % model_object.__name__)
        current_records = db_meta.current_record(model_object, records)
        count = current_records.update({model_object.valid_until: current_time}, synchronize_session=False)
        logger.debug("Ended validity period for %d '%s' objects" % (count, model_object.__name__))

        # open validity period of the newly added records
        logger.debug("Starting validity period for %s records at _shift_validity_period" % model_object.__name__)
        new_records = db_meta.new_record(model_object, records)
        count = new_records.update({model_object.valid_since: current_time}, synchronize_session=False)
        logger.debug("Started validity period for %d '%s' objects" % (count, model_object.__name__))

        self.session.flush()

    @trace
    def _save_accounts(self, accounts, source_system, is_partial):
        """
        Register accounts in the source system and commit changes.

        Parameters:
            accounts_and_owners: A list of dictionaries (account_object, owner_id)
        """
        new_db_accounts = []

        for a in accounts:
            a.account_source_system = source_system
            old_account = self.get_account(a.external_id, source_system.id)
            if old_account is not None:
                a.oa_id = old_account.oa_id

            a.owner = self.get_account(a.parent_external_id, source_system.id, db_meta.new_record)
            if not a.owner and is_partial:
                a.owner = self.get_account(a.parent_external_id, source_system.id, db_meta.current_record)

            self.session.add(a)
            new_db_accounts.append(a)

        self.session.flush()
        self._update_subscriptions(new_db_accounts, source_system)

        logger.debug("Saved %s accounts to the database" % len(new_db_accounts))

    def get_account(self, account_id, source_system_id, validity_period=db_meta.current_record):
        """Get a valid account by ID"""
        account = self.session.query(AccountModel).filter(AccountModel.source_system_id == source_system_id)
        account = validity_period(AccountModel, account)
        account = account.filter(AccountModel.external_id == account_id).one_or_none()
        return account
