# (C)2017 Ingram Micro Inc. all Rights Reserved
import time
import logging

from collections import namedtuple

from migrator.core import MigrationError, MigrationNoRepeatError
from migrator.core.data.model.billing import billing_type_period, get_period_type_string
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.transfer.import_api_base import SubscriptionInfo, ServicePlanInfo
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.registry import Registry
from migrator.odin_automation.billing import is_subscription_nonrestorable
from migrator.odin_automation.billing.lib.api.bm import BillingAPIError
from migrator.odin_automation.platform.lib import service_plans
from migrator.odin_automation.platform.lib.api.aps import subscription
from migrator.odin_automation.billing.lib.api.subscription import \
    SubscriptionStatus, SubscriptionServiceStatus, SubscriptionResourceStatus
from migrator.odin_automation.platform.lib.api.aps.account import get_account_by_subscription
from migrator.odin_automation.platform.lib.api.aps.domain import get_pa_domain_by_name
from migrator.odin_automation.platform.lib.task_manager import wait_for_event_notifications_completed_on_resource,\
    EventType
from migrator.odin_automation.platform.lib.time_tool import advance_months, advance_days, advance_years

logger = logging.getLogger(__name__)


OaSubscriptionParameters = namedtuple('OaSubscriptionParameters', (
    'service_activation_parameters', 'synchronize_resource_limits'
))


class SubscriptionInfoOA(SubscriptionInfo):
    def __init__(self, id, name, account_id, service_plan_id, status, service_status):
        super(SubscriptionInfoOA, self).__init__(id, name, account_id, service_plan_id)
        self.status = status
        self.service_status = service_status

    def __repr__(self):
        return 'SubscriptionInfoOA (id="{id}", name="{name}", account_id="{acc_id}", service_plan_id="{sp_id}", ' \
            'status="{status}", service_status="{service_status}")'.format(
            id=self.id,
            name=self.name,
            acc_id=self.account_id,
            sp_id=self.service_plan_id,
            status=self.status,
            service_status=self.service_status
        )


class SubscriptionImporterBase(object):
    def __init__(self):
        self.global_context = Registry.get_instance().get_context()

    @trace
    def create_subscription(self, subscription_info, account_id, plan_id, billing_info=None, subscription_is=None):
        raise NotImplementedError()

    @trace
    def update_subscription(self, subscription_info, account_id):
        """
        :param subscription_info: migrator.odin_automation.platform.lib.data.subscription_importer.SubscriptionInfoOA
        :param account_id: int Example: 1000001
        :return: int ID of updated subscription. Example: 1000022
        """
        raise NotImplementedError()

    @trace
    def get_subscription(self, account_id, id, plan_id):
        raise NotImplementedError()

    @trace
    def get_plan_info(self, plan):
        raise NotImplementedError()

    @trace
    def import_subscription(self, source_subscription_info, account_id, plan_id, billing_info=None):
        """
        :param source_subscription_info: migrator.core.data.model.subscription.SubscriptionModel
        :param account_id: int
        :param plan_id: int
        :param billing_info: migrator.core.data.model.billing.SubscriptionBillingModel
        :return: int The imported subscription ID
        """
        target_subscription_info = self.get_subscription(
            account_id, source_subscription_info.oa_id, plan_id
        )

        target_plan_info = self.get_plan_info(plan_id)
        try:
            subscription_act_parameters = self._get_activation_parameters_for_subscription(source_subscription_info)
        except Exception as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(_T("MIGRATION_ERROR_UNABLE_TO_DETERMINE_SUB_ACT_PARAMS").format(
                subscription_name=source_subscription_info.name,
                error=e.message
            ))

        source_subscription_info.act_params = subscription_act_parameters

        # There is subscription with this id on the destination server, but Terminated/destroyed
        if target_subscription_info is not None and is_subscription_nonrestorable(target_subscription_info):
            logger.warning(
                "Subscription {id} is in Terminated/Removed status. New subscription will be created.".format(
                    id=source_subscription_info.oa_id
                )
            )
            target_subscription_info = None

        # There is no subscription with this id on the destination server
        if target_subscription_info is None:
            return self.create_subscription(
                source_subscription_info, account_id, target_plan_info.id, billing_info, None
            )

        target_subscription_info.act_params = subscription_act_parameters

        # The subscription is associated with another plan
        subscription_plan_info = self.get_plan_info(target_subscription_info.service_plan_id)

        if target_plan_info.st_id != subscription_plan_info.st_id:
            raise MigrationError(_T("SUBSCRIPTION_ALREADY_PRESENT_PLAN").format(
                subscription=source_subscription_info.name,
                migration_plan=target_plan_info.name,
                target_plan=subscription_plan_info.name
            ))

        # The subscription is associated with another account
        if target_subscription_info.account_id != account_id:
            raise MigrationError(_T("SUBSCRIPTION_ALREADY_PRESENT_ACCOUNT").format(
                subscription=source_subscription_info.name,
                migration_list_account=account_id,
                target_account=target_subscription_info.account_id
            ))

        return self.update_subscription(target_subscription_info, account_id)

    @trace
    def _get_activation_parameters_for_subscription(self, source_subscription):
        """
        :param source_subscription: migrator.core.data.model.subscription.SubscriptionModel
        :return: list
        """
        platform_reader = PlatformReaderDB()
        migration_item = platform_reader.get_migration_project_item(source_subscription.id)
        service_import_api = self.global_context.target_panel_obj.get_service_import_api(self.global_context)
        return service_import_api.get_service_specific_activation_parameters(self.global_context, migration_item)

    @staticmethod
    @trace
    def _check_if_domain_exists(domain_name):
        domain = get_pa_domain_by_name(domain_name)
        if domain is not None:
            raise MigrationError(_T("UNABLE_TO_CREATE_DOMAIN_SUBSCRIPTION_BECAUSE_DOMAIN_ALREADY_EXISTS").format(
                domain_name=domain_name
            ))


class SubscriptionImporterPlatform(SubscriptionImporterBase):
    def __init__(self, panel):
        super(SubscriptionImporterPlatform, self).__init__()
        self.target_panel_api = panel.get_connections(self.global_context).poa_api()

    @trace
    def get_subscription(self, account_id, id, plan_id):

        target_subscription_info = subscription.get_subscription(id) if id else None

        if target_subscription_info is None:
            return None

        return SubscriptionInfoOA(
            id=target_subscription_info.subscriptionId,
            name=target_subscription_info.name,
            service_plan_id=target_subscription_info.serviceTemplateId,
            account_id=target_subscription_info.account.id,
            status=None,
            service_status=None
        )

    @trace
    def get_plan_info(self, plan_id):
        plan_info = None
        if plan_id is not None:
            plan_info = self.target_panel_api.service.get_plan_by_id(plan_id)

        if plan_info is None:
            raise MigrationError(_T("PLAN_IS_NOT_PRESENT_IN_OA").format(plan=plan_id))

        return ServicePlanInfo(
            id=plan_info.st_id,
            st_id=plan_info.st_id,
            name=plan_info.name,
        )


class SubscriptionImporterPBA(SubscriptionImporterBase):
    def __init__(self, panel):
        super(SubscriptionImporterPBA, self).__init__(panel)
        self.target_panel_billing_api = panel.get_billing_api(self.global_context)

    @trace
    def get_subscription(self, account_id, id, plan_id):
        # Check that subscription exists in Billing
        target_billing_subscription_info = None
        if account_id is None:
            account_id = get_account_by_subscription(id)['id']

        if id is not None and account_id is not None:
            target_billing_subscription_info = subscription.get_billing_subscription(account_id, id)

        if target_billing_subscription_info is None:
            return None

        return SubscriptionInfoOA(
            id=target_billing_subscription_info.subscriptionId,
            name=target_billing_subscription_info.subscriptionName,
            service_plan_id=target_billing_subscription_info.planId,
            account_id=account_id,
            status=target_billing_subscription_info.status,
            service_status=target_billing_subscription_info.serviceStatus
        )

    @trace
    def get_plan_info(self, plan_id):
        plan_info = None
        if plan_id is not None:
            plan_info = self.target_panel_billing_api.plan_details_get(plan_id)

        if plan_info is None:
            raise MigrationError(_T("BILLING_PLAN_IS_NOT_PRESENT_IN_OA").format(
                plan=plan_id,
            ))

        return ServicePlanInfo(
            id=plan_info.plan_id,
            st_id=plan_info.st_id,
            name=plan_info.name,
        )

    @trace
    def _get_period_info(self, plan_id, plan_name, subscription_billing_info):
        plan_period_list_full = self.target_panel_billing_api.plan_period_list_get(plan_id) or []
        plan_period_list = [x for x in plan_period_list_full if x.trial == 0 and x.enabled == 1 and x.period > 0]
        if subscription_billing_info.subscription_period is None:
            logger.debug(
                'There is no subscription period in billing info, using the minimal available period from service plan'
            )
            if not plan_period_list:
                raise MigrationError(_T("NO_SUBSCRIPTION_PERIODS_IN_PLAN").format(plan_name=plan_name))

            period = min(plan_period_list, key=lambda x: (x.period_type, x.period))
            logger.debug("Using period '%s %s(s)'" % (period.period, get_period_type_string(period.period_type)))
            return period

        period_type_str = get_period_type_string(subscription_billing_info.subscription_period_type)
        logger.debug("Subscription period '%s %s(s)' is specified in billing info" % (
            subscription_billing_info.subscription_period, period_type_str
        ))
        for plan_period in plan_period_list:
            if plan_period.period == subscription_billing_info.subscription_period and \
                    plan_period.period_type == subscription_billing_info.subscription_period_type:
                return plan_period

        raise MigrationError(_T("SUBSCRIPTION_PERIOD_IS_MISSING").format(
            plan_name=plan_name,
            period=subscription_billing_info.subscription_period,
            period_type=period_type_str
        ))

    @staticmethod
    @trace
    def _fill_dates(billing_info, period_info, plan_info):
        _time_adv_map = {
            'day': advance_days,
            billing_type_period['day']: advance_days,
            'month': advance_months,
            billing_type_period['month']: advance_months,
            'year': advance_years,
            billing_type_period['year']: advance_years,
        }
        logger.debug('Generating creation and expiration dates from the subscription period and the current time')
        creation_date_seconds = int(time.time())

        billing_info.creation_date = creation_date_seconds
        billing_info.expiration_date = \
            _time_adv_map[period_info.period_type](billing_info.creation_date, period_info.period)

        billing_info.last_billing_date = creation_date_seconds
        billing_info.next_billing_date = \
            _time_adv_map[plan_info.billing_period_type](billing_info.last_billing_date, plan_info.billing_period)


class HostingSubscriptionImporter(SubscriptionImporterPlatform):
    def __init__(self, panel):
        super(HostingSubscriptionImporter, self).__init__(panel)

    @trace
    def create_subscription(self, subscription_info, account_id, plan_id, billing_info=None, subscription_id=None):
        return self._activate_subscription(
            account_id,
            plan_id,
            subscription_info.name,
            subscription_id,
            subscription_info.act_params.service_activation_parameters
        )

    @trace
    def _activate_subscription(self, account_id, st_id, subscription_name, subscription_id=None, act_params=None):
        return self.target_panel_api.subscription.activate_subscription(
            account_id, st_id, subscription_name, subscription_id=subscription_id, act_params=act_params
        )

    @trace
    def update_subscription(self, subscription_info, account_id):
        """
        :param subscription_info: migrator.odin_automation.platform.lib.data.subscription_importer.SubscriptionInfoOA
        :param account_id: int Example: 1000001
        :return: int ID of updated subscription. Example: 1000022
        """
        return subscription_info.id


class HostingSubscriptionImporterPPAB(SubscriptionImporterPBA, HostingSubscriptionImporter):
    def __init__(self, panel):
        super(HostingSubscriptionImporterPPAB, self).__init__(panel)

    @trace
    def create_subscription(self, subscription_info, account_id, plan_id, billing_info=None, subscription_id=None):
        plan_info = self.target_panel_billing_api.plan_details_get(plan_id)
        period_info = self._get_period_info(plan_id, plan_info.name, billing_info)
        if billing_info.expiration_date is None or billing_info.next_billing_date is None or \
                billing_info.creation_date is None or billing_info.last_billing_date is None:
            self._fill_dates(billing_info, period_info, plan_info)

        try:
            subscription_id = self.target_panel_billing_api.subscription.import_subscription(
                subscription_info.name, billing_info, account_id, plan_info, period_info
            )
        except BillingAPIError as e:
            if 'this Plan does not belong to Vendor of Account' in e.short_message:
                raise MigrationNoRepeatError(
                    _T('ACCOUNT_CANNOT_BE_SUBSCRIBED_TO_PLAN_BECAUSE_THIS_PLAN_DOES_NOT_BELONG_TO_VENDOR_OF_ACCOUNT').format(
                        account_id=account_id, plan_id=plan_id
                    )
                )
            else:
                raise

        self.target_panel_billing_api.subscription.subscription_status_update(
            subscription_id, SubscriptionStatus.ORDERED, SubscriptionServiceStatus.CREATING
        )

        subscription_id = super(HostingSubscriptionImporterPPAB, self).create_subscription(
            subscription_info, account_id, plan_info.st_id, billing_info, subscription_id
        )

        self._set_billing_subscription_resources_status(subscription_id)
        subscription_parameters = self._get_activation_parameters_for_subscription(subscription_info)
        if subscription_parameters.synchronize_resource_limits:
            self._synchronize_subscription_billing_resource_rates_with_oa(subscription_id)

        return subscription_id

    @trace
    def update_subscription(self, subscription_info, account_id):
        """
        :param subscription_info: migrator.odin_automation.platform.lib.data.subscription_importer.SubscriptionInfoOA
        :param account_id: int Example: 1000001
        :return: int ID of updated subscription. Example: 1000022
        """
        # Check that subscription exists in OA too
        target_subscription_info = HostingSubscriptionImporter.get_subscription(
            self, account_id, subscription_info.id, None
        )
        if target_subscription_info:
            return target_subscription_info.id

        plan_info = self.target_panel_billing_api.plan_details_get(subscription_info.service_plan_id)
        subscription_id = self._activate_subscription(
            account_id, plan_info.st_id, subscription_info.name, subscription_info.id,
            subscription_info.act_params.service_activation_parameters
        )

        self._set_billing_subscription_resources_status(subscription_id)
        if subscription_info.act_params.synchronize_resource_limits:
            self._synchronize_subscription_billing_resource_rates_with_oa(subscription_id)

        return subscription_id

    @trace
    def _set_billing_subscription_resources_status(self, subscription_oa_id):
        subscription_resources = self.target_panel_billing_api.subscription.get_subscription_resources(
            subscription_oa_id
        )
        for resource in subscription_resources:
            if float(resource.IncludedAmount) >= 0:
                resource_status = SubscriptionResourceStatus.SPS_INSTALLED
            else:
                resource_status = SubscriptionResourceStatus.SPS_OPTIONAL

            self.target_panel_billing_api.subscription.subscription_resource_status_update(
                subscription_oa_id, resource.ResourceID, resource_status
            )

    @trace
    def _synchronize_subscription_billing_resource_rates_with_oa(self, subscription_oa_id):
        self.target_panel_billing_api.subscription.synchronize_subscription_resources(subscription_oa_id)
        subscription_aps_id = subscription.get_subscription(subscription_oa_id)['aps']['id']
        wait_for_event_notifications_completed_on_resource(subscription_aps_id, EventType.LIMIT_CHANGE)


class HostingSubscriptionImporterPBA(HostingSubscriptionImporterPPAB):
    pass


class DomainSubscriptionImporter(SubscriptionImporterPlatform):
    def __init__(self, panel):
        super(DomainSubscriptionImporter, self).__init__(panel)

    @trace
    def get_subscription(self, account_id, id, plan_id):
        return super(DomainSubscriptionImporter, self).get_subscription(account_id, id, plan_id)

    @trace
    def update_subscription(self, subscription_info, account_id):
        """
        :param subscription_info: migrator.odin_automation.platform.lib.data.subscription_importer.SubscriptionInfoOA
        :param account_id: int Example: 1000001
        :return: int ID of updated subscription. Example: 1000022
        """
        return subscription_info.id

    @trace
    def create_subscription(self, subscription_info, account_id, plan_id, billing_info=None, subscription_id=None):
        # pem.addSubscription
        subscription_id = self.target_panel_api.subscription.create_subscription(
            account_id, plan_id, subscription_info.name, subscription_id
        )
        return subscription_id

    @trace
    def _get_activation_parameters_for_subscription(self, source_subscription_info):
        return OaSubscriptionParameters(
            service_activation_parameters=[],
            synchronize_resource_limits=False
        )


class DomainSubscriptionImporterPPAB(SubscriptionImporterPBA, DomainSubscriptionImporter):
    def __init__(self, panel):
        super(DomainSubscriptionImporterPPAB, self).__init__(panel)

    @trace
    def get_subscription(self, account_id, id, plan_id):
        return SubscriptionImporterPBA.get_subscription(self, account_id, id, plan_id)

    @trace
    def update_subscription(self, subscription_info, account_id):
        """
        :param subscription_info: migrator.odin_automation.platform.lib.data.subscription_importer.SubscriptionInfoOA
        :param account_id: int Example: 1000001
        :return: int ID of updated subscription. Example: 1000022
        """
        # Check that subscription exists in OA too
        target_subscription_info = DomainSubscriptionImporter.get_subscription(
            self, account_id, subscription_info.id, None
        )
        if target_subscription_info:
            return target_subscription_info.id

        plan_info = self.target_panel_billing_api.plan_details_get(subscription_info.service_plan_id)
        return DomainSubscriptionImporter.create_subscription(
            self, subscription_info, account_id, self._get_dns_service_template_id(plan_info),
            subscription_id=subscription_info.id
        )

    @trace
    def create_subscription(self, subscription_info, account_id, plan_id, billing_info=None, subscription_id=None):
        plan_info = self.target_panel_billing_api.plan_details_get(plan_id)
        period_info = self._get_period_info(plan_id, plan_info.name, billing_info)
        if billing_info.expiration_date is None or billing_info.next_billing_date is None or \
                billing_info.creation_date is None or billing_info.last_billing_date is None:
            self._fill_dates(billing_info, period_info, plan_info)

        # DOMAINGATE.DomainSubscrAdd_API
        domain_subscription = self.target_panel_billing_api.subscription.add_domain_subscription(
            subscription_info.name, billing_info, account_id, plan_info, period_info
        )
        return DomainSubscriptionImporter.create_subscription(
            self, subscription_info, account_id, self._get_dns_service_template_id(plan_info), billing_info,
            domain_subscription.subscription_id
        )

    @trace
    def _get_dns_service_template_id(self, plan_info):
        global_context = Registry.get_instance().get_context()
        sp = service_plans.get_service_plans_manager(global_context.conn.target, True)
        dns_service_template_id = sp.get_dns_service_template_id(plan_info)
        if dns_service_template_id is None:
            raise Exception(_T("NO_DNS_HOSTING_SERVICE_TEMPLATES_FOUND"))

        return dns_service_template_id


class DomainSubscriptionImporterPBA(DomainSubscriptionImporterPPAB):
    pass
