# (C)2017 Ingram Micro Inc. all Rights Reserved
import psycopg2
import logging

from psycopg2.extras import LoggingConnection

from migrator.settings import PRINT_DB_QUERIES_TO_LOG
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.filetypes.config_utils import read_simple_config
from migrator.settings import OA_KERNEL_CONFIG_FILE

logger = logging.getLogger(__name__)


class OaServiceName(object):
    # LSH node service
    APACHE = ['Apache']
    # NG node service
    APACHE_NG = ['ApacheNG']
    # NG web cluster service
    WEB_CLUSTER = ['WebCluster']
    # Windows Shared Hosting service
    WSH = ['msiis', 'msiis7']


class DbConnectionError(Exception):
    def __init__(self, message):
        super(DbConnectionError, self).__init__(message)


def get_platform_db_connection(cursor_factory=None):
    db_config = read_simple_config(OA_KERNEL_CONFIG_FILE)
    try:
        conn = psycopg2.connect(
            cursor_factory=cursor_factory,
            host=db_config['dsn_host'],
            dbname=db_config['dsn_db_name'],
            user=db_config['dsn_login'],
            password=db_config['dsn_passwd'],
            connection_factory=LoggingConnection if PRINT_DB_QUERIES_TO_LOG else None
        )
        if PRINT_DB_QUERIES_TO_LOG:
            conn.initialize(logger)

        return conn
    except Exception:
        logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
        raise DbConnectionError(_T("FAILED_TO_CONNECT_TO_OA_DB"))


@cached
def get_application_instances_ids(app_uid):
    connection = None
    cursor = None
    try:
        connection = get_platform_db_connection()
        cursor = connection.cursor()

        sql = """
            SELECT aps_application_instances.app_instance_id
            FROM aps_application_instances
            JOIN aps_applications ON aps_application_instances.app_id = aps_applications.app_id
            WHERE aps_applications.app_uid=%s
        """

        cursor.execute(sql, [app_uid])
        return _fetchall(cursor)
    finally:
        if cursor is not None:
            cursor.close()

        if connection is not None:
            connection.close()


@cached
def get_system_locales():
    sql = """
        SELECT language_code || '_' || country_code as locale
        FROM system_locales
    """

    columns = ('locale', )
    return [x['locale'] for x in execute_sql_and_get_formatted_result(sql, columns)]


def _get_sort_order(sort_order):
    if sort_order not in ['ASC', 'DESC']:
        sort_order = 'ASC'

    return sort_order


def _get_sort_field(sort_field, default_sort_field):
    return sort_field if sort_field is not None else default_sort_field


def _get_limit(limit):
    return 'LIMIT %s' % limit if limit is not None else ''


def _get_offset(offset):
    return offset if offset is not None else 0


def _get_list_of_where_conditions(like, in_values, eq_values, available_condition_fields, special_rules=None):
    where_list = []
    parameters = []

    def get_key_and_value(k, v, condition_type):
        if special_rules and condition_type in special_rules and k in special_rules[condition_type]:
            k, v = special_rules[condition_type][k](v)

        return k, v

    if like is not None:
        for key, value in like.iteritems():
            if key in available_condition_fields:
                append_key, append_value = get_key_and_value(key, value, 'like')
                where_list.append('{0} LIKE %s'.format(append_key))
                parameters.append(append_value.replace('*', '%'))

    if in_values is not None:
        for key, value in in_values.iteritems():
            if key in available_condition_fields:
                append_key, append_value = get_key_and_value(key, value, 'in')
                where_list.append('{0} IN ({1})'.format(append_key, ','.join(['%s'] * len(append_value))))
                parameters.extend(append_value)

    if eq_values is not None:
        for key, value in eq_values.iteritems():
            if key in available_condition_fields:
                append_key, append_value = get_key_and_value(key, value, 'eq')
                where_list.append('{0} = %s'.format(append_key))
                parameters.append(append_value)

    return where_list, parameters


def _execute_sql_and_get_paginal_response(sql, result_columns, where_parameters, countable_field, limit, offset):
    connection = None
    cursor = None

    try:
        connection = get_platform_db_connection()
        cursor = connection.cursor()
        cursor.execute(
            'SELECT COUNT({0}) AS counter FROM ({1}) AS subs'.format(countable_field, sql),
            where_parameters
        )
        result_count = _fetchone(cursor)[0]

        cursor.execute(
            'SELECT * FROM ({0}) AS subs {1} OFFSET {2}'.format(sql, limit, offset),
            where_parameters
        )

        result_items = []
        for row in _fetchall(cursor):
            result_items.append(dict(zip(result_columns, row)))
    finally:
        if cursor is not None:
            cursor.close()

        if connection is not None:
            connection.close()

    return {
        'count': result_count,
        'items': result_items
    }


def execute_sql_and_get_formatted_result(sql, result_columns, sql_parameters=None):
    result_items = []

    def decodeutf8(row):
        return [x.decode('utf-8') if type(x) == str else x for x in row]

    for row in execute_sql_and_get_result(sql, sql_parameters):
        result_items.append(dict(zip(result_columns, decodeutf8(row))))

    return result_items


def execute_sql_and_get_result(sql, sql_parameters=None):
    connection = None
    cursor = None
    try:
        connection = get_platform_db_connection()
        cursor = connection.cursor()
        prepared_sql = u' '.join(sql.split())
        cursor.execute(prepared_sql, sql_parameters)

        return _fetchall(cursor)
    finally:
        if cursor is not None:
            cursor.close()

        if connection is not None:
            connection.close()


def _fetchall(cursor):
    result = cursor.fetchall()
    if PRINT_DB_QUERIES_TO_LOG:
        logger.debug(str(result))

    return result


def _fetchone(cursor):
    result = cursor.fetchone()
    if PRINT_DB_QUERIES_TO_LOG:
        logger.debug(str(result))

    return result


def transform_list_to_in_construction(objects):
    return '(' + ', '.join(["'%s'" % x for x in objects]) + ')'
