# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.utils.loggers.log import trace
from migrator.odin_automation.billing.lib.api.bm import BillingAPIError, INT_NULL
from migrator.odin_automation.platform.lib.api.aps_runtime.core.aps_resource import APSResource
from migrator.odin_automation.platform.lib.resources import dns_resource_class_names, dns_st_type
from migrator.odin_automation.platform.lib.api.openapi.service import \
    R_VPS_CORE_POWER, RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND
from migrator.odin_automation.platform.lib.db.service_template import get_ng_service_templates
from migrator.settings import TargetServiceType, SourceHostingClass, OA_ADMIN_ID, OS_LINUX, OS_WINDOWS

import migrator.odin_automation.services.shm.lib.aps.types as shm_aps_types

logger = logging.getLogger(__name__)


class ServicePlanInfo:
    @trace
    def __init__(self, plan_id=None, name=None, description=None, active=None, periods=None):
        self.id = plan_id
        self.name = name
        self.description = description
        self.active = active
        self.periods = periods


class ServicePlansManager(object):
    @trace
    def __init__(self, oa_conn):
        self._poa_api = oa_conn.poa_api()

    @trace
    def get_by_service(self, account_id, target_service, include_periods=False, plan_ids=None,
                       source_hosting_type=None):
        try:
            # ToDo: checkme: POA w/o Billing
            if target_service is None:
                return self.get_service_plans(account_id, plan_ids)
            elif target_service == TargetServiceType.SHM:
                filter_by_platform = None
                if source_hosting_type == SourceHostingClass.WSH:
                    filter_by_platform = OS_WINDOWS
                elif source_hosting_type in [SourceHostingClass.WHL, SourceHostingClass.LSH]:
                    filter_by_platform = OS_LINUX
                return self.get_shared_service_plans(account_id, plan_ids, platform=filter_by_platform)
            elif target_service == TargetServiceType.OACI:
                return self.get_vps_service_plans(account_id, plan_ids)
            elif target_service == TargetServiceType.DNS:
                return self.get_dns_service_plans(account_id, plan_ids)
            elif target_service == TargetServiceType.NG:
                return self.get_ng_service_plans(account_id, plan_ids)
            else:
                return []

        except BillingAPIError as e:
            logger.debug(_T('LOG_EXCEPTION'), exc_info=True)
            error_message = e.message.replace('\n', ' ')
            access_denied_error = 'Error Code: 0xb390bec5'
            possible_cause = _T('BILLING_LICENSE_VALIDITY_WARNING') if access_denied_error in error_message else ''

            raise MigrationError(_T('SERVICE_PLAN_DATA_ACCESS_DENIED').format(
                service=target_service, error_message=error_message, cause=possible_cause))

    @trace
    def get_service_plans(self, account_id, plan_ids):
        return self._filter_templates_by_ids(self._get_service_templates(account_id), plan_ids)

    @trace
    def get_shared_service_plans(self, account_id, plan_ids, platform=None):
        return self._filter_templates_by_ids(self._get_shared_service_templates(account_id, platform), plan_ids)

    @trace
    def get_vps_service_plans(self, account_id, plan_ids):
        return self._filter_templates_by_ids(self._get_vps_service_templates(account_id), plan_ids)

    @trace
    def get_dns_service_plans(self, account_id, plan_ids):
        return self._filter_templates_by_ids(self._get_dns_service_templates(account_id), plan_ids)

    @trace
    def get_ng_service_plans(self, account_id, plan_ids):
        return self._filter_templates_by_ids(self._get_ng_service_templates(account_id), plan_ids)

    @trace
    def _get_service_templates(self, account_id):
        service_tmpls = {}

        for st in self._poa_api.service.get_service_template_list(account_id):
            service_tmpls[st.st_id] = ServicePlanInfo(st.st_id, st.name, st.description, st.active)

        return service_tmpls

    @trace
    def _get_service_templates_by_resources(self, account_id, resclass_names, resource_types, contains_only=False):
        """
        The function returns a list of service templates. It works in two modes:
            1. It searches for service templates that contain at least one resource with specified name and type.
            2. It searches for service templates that contain only the specified resources and are not empty.
        """
        service_tmpls = {}

        for st in self._poa_api.service.get_service_template_list(account_id):
            st_info = self._poa_api.service.get_service_template(st.st_id, True)

            def _matched():
                for rt in st_info.resource_types:
                    if rt.resclass_name in resclass_names:
                        return True

                    for ac in rt.act_params:
                        if ac.name == 'service_type':
                            for type in resource_types:
                                if APSResource(ac.value) == type:
                                    return True

                return False

            def _contains_only():
                if not st_info.resource_types:
                    return False

                for rt in st_info.resource_types:
                    if rt.resclass_name not in resclass_names:
                        return False

                return True

            check_function = _contains_only if contains_only else _matched
            if check_function():
                service_tmpls[st.st_id] = ServicePlanInfo(st.st_id, st.name, st.description, st.active)

        return service_tmpls

    @trace
    def _get_shared_service_templates(self, account_id, platform=None):
        shared_resource_types_by_platform = {
            OS_LINUX: shm_aps_types.WebspaceApache,
            OS_WINDOWS: shm_aps_types.WebspaceIIS
        }

        resource_types = [shm_aps_types.SiteWeb] if platform is None else [shared_resource_types_by_platform[platform]]
        return self._get_service_templates_by_resources(account_id, (), resource_types)

    @trace
    def _get_vps_service_templates(self, account_id):
        service_tmpls = {}
        for st in self._poa_api.service.get_service_template_list(account_id):
            st_info = self._poa_api.service.get_service_template(st.st_id, True)
            st_res_status = st_info.are_resources_exist([R_VPS_CORE_POWER])
            if st_res_status['status'] == RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND:
                service_tmpls[st.st_id] = ServicePlanInfo(st.st_id, st.name, st.description, st.active)

        return service_tmpls

    @trace
    def _get_dns_service_templates(self, account_id):
        return self._get_service_templates_by_resources(account_id, dns_resource_class_names, (), contains_only=True)

    @staticmethod
    @trace
    def _get_ng_service_templates(account_id):
        service_tmpls_list = get_ng_service_templates(account_id)
        service_tmpls_dict = {}
        for item in service_tmpls_list:
            service_tmpls_dict[item['st_id']] = ServicePlanInfo(
                item['st_id'], item['name'], item['description'], item['active']
            )

        return service_tmpls_dict

    @staticmethod
    @trace
    def _is_in_list(plan_id, plan_ids_list):
        if plan_ids_list is not None:
            return plan_id in plan_ids_list
        else:
            return True

    @staticmethod
    @trace
    def _filter_templates_by_ids(templates_dict, plan_ids):
        if plan_ids is not None:
            return {plan_id: templates_dict[plan_id] for plan_id in plan_ids}.values()
        else:
            return templates_dict.values()


class ServicePlansManagerPPAB(ServicePlansManager):
    def __init__(self, oa_conn):
        super(ServicePlansManagerPPAB, self).__init__(oa_conn)
        self._bm_api = oa_conn.billing_api()

    @trace
    def get_by_service(self, account_id, target_service, include_periods=False, plan_ids=None,
                       source_hosting_type=None):
        service_plans = super(ServicePlansManagerPPAB, self).get_by_service(account_id, target_service, include_periods,
                                                                            plan_ids, source_hosting_type)

        if include_periods:
            for plan in service_plans:
                plan.periods = [x.__dict__ for x in self._bm_api.plan_period_list_get(plan.id)]

        return service_plans

    @trace
    def get_service_plans(self, account_id, plan_ids):
        service_plans = []
        active_sts = self._get_active_service_templates(account_id)

        for plan in self._bm_api.get_all_plan_list(account_id):
            if self._is_in_list(plan.id, plan_ids):
                plan_details = self._bm_api.plan_details_get(plan.id)
                service_plans.append(
                    ServicePlanInfo(plan.id, plan.name, plan.short_description, plan_details.st_id in active_sts)
                )

        return service_plans

    @trace
    def get_shared_service_plans(self, account_id, plan_ids, platform=None):
        service_templates = self._get_shared_service_templates(account_id, platform=platform)
        return self._get_service_plans_by_templates(account_id, service_templates, plan_ids)

    @trace
    def get_vps_service_plans(self, account_id, plan_ids):
        service_tmpls = self._get_vps_service_templates(account_id)
        return self._get_service_plans_by_templates(account_id, service_tmpls, plan_ids)

    @trace
    def get_dns_service_plans(self, account_id, plan_ids):
        service_plans = []
        active_sts = self._get_active_service_templates(account_id)

        for plan in self._bm_api.get_all_plan_list(account_id):
            plan_details = self._bm_api.plan_details_get(plan.id)
            st_type = plan_details.st_type
            if st_type == dns_st_type and self._is_in_list(plan.id, plan_ids):
                st_id = self.get_dns_service_template_id(plan_details)
                service_plans.append(
                    ServicePlanInfo(plan.id, plan.name, plan.short_description, st_id in active_sts)
                )

        return service_plans

    @trace
    def _get_active_service_templates(self, account_id):
        return [st.st_id for st in self._poa_api.service.get_service_template_list(account_id) if st.active]

    @trace
    def get_ng_service_plans(self, account_id, plan_ids):
        service_tmpls = self._get_ng_service_templates(account_id)
        return self._get_service_plans_by_templates(account_id, service_tmpls, plan_ids)

    @trace
    def _get_service_plans_by_templates(self, account_id, service_tmpls, plan_ids=None):
        service_plans = []

        for st_id, st in service_tmpls.iteritems():
            for plan in self._bm_api.get_plan_list_by_service_template(st_id):
                if self._is_in_list(plan.id, plan_ids):
                    service_plans.append(
                        ServicePlanInfo(plan.id, plan.name, plan.short_description, st.active)
                    )

        return service_plans

    @trace
    def get_dns_service_template_id(self, plan_info):
        # If service plan is provisioned through PEMGATE, return the PEM service template that it is based on
        if plan_info.st_type == 'PEMGATE':
            logger.debug(
                "Service plan #%s %s is provisioned through PEMGATE, service template id is %s" %
                (plan_info.plan_id, plan_info.name, plan_info.st_id)
            )
            return plan_info.st_id

        # Else get the default NS service template and if it is set, return it
        # TODO: default NS template is probably may be configured per-reseller. The API call below returns the settings
        # for the provider
        default_ns_st_id = self._bm_api.get_pem_ns_settings().default_ns_st_id
        if default_ns_st_id != INT_NULL:
            logger.debug(
                "Service plan #%s %s is not provisioned through PEMGATE, using default NS ST id %s" %
                (plan_info.plan_id, plan_info.name, default_ns_st_id)
            )
            return default_ns_st_id

        # Else find the first suitable service template in PEM and return it
        service_templates = super(ServicePlansManagerPPAB, self).get_dns_service_plans(OA_ADMIN_ID, None)
        if len(service_templates) < 1:
            return None

        logger.debug(
            "Service plan #%s %s is not provisioned through PEMGATE, and default NS ST is not set, using ST id %s" %
            (plan_info.plan_id, plan_info.name, service_templates[0].id)
        )
        return service_templates[0].id


@trace
def get_service_plans_manager(oa_conn, is_billing):
    if is_billing:
        return ServicePlansManagerPPAB(oa_conn)
    else:
        return ServicePlansManager(oa_conn)
