# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from poaupdater import uTasks

from migrator.core import MigrationError
from migrator.core import MigrationNoRepeatError
from migrator.core.language import _T
from migrator.core.utils import FunctionResultWaiter
from migrator.odin_automation.platform.lib.db.task_manager import get_scheduled_tasks_for_resource, get_event_notifications

logger = logging.getLogger(__name__)


class EventType(object):
    LIMIT_CHANGE = 'http://parallels.com/aps/events/pa/subscription/limits/changed'


class TmConst(object):
    DEFAULT_WAITING_ATTEMPTS = 60
    DEFAULT_WAITING_INTERVAL = 5
    EVENT_APPEARING_ATTEMPTS = 10
    EVENT_APPEARING_INTERVAL = 1
    FAILED_TASK_STATUS = 'f'


def wait_for_all_scheduled_tasks_on_resource_completed(resource_aps_id, attempts=TmConst.DEFAULT_WAITING_ATTEMPTS,
                                                       interval=TmConst.DEFAULT_WAITING_INTERVAL):
    def check_tasks_status(tasks):
        if not tasks:
            logger.debug(_T("TM_ALL_TASKS_COMPLETED").format(guid=resource_aps_id))
            return True

        for task in tasks:
            if task['status'] == TmConst.FAILED_TASK_STATUS:
                raise MigrationError(_T("TM_TASK_FAILED").format(
                    id=task['task_id'], name=task['name'], guid=resource_aps_id
                ))

        logger.debug(_T("TM_TASKS_IN_PROGRESS").format(
            guid=resource_aps_id, tasks_list=','.join(str(p) for p in tasks)
        ))
        return False

    func = FunctionResultWaiter(
        get_scheduled_tasks_for_resource,
        MigrationNoRepeatError(_T('TM_TIMEOUT_REACHED').format(timeout=attempts * interval)),
        check_function=check_tasks_status,
        max_attempts=attempts,
        interval_between_attempts=interval
    )

    return func.run(resource_aps_id)


def wait_for_event_notifications_completed_on_resource(resource_aps_id, event_type,
                                                       attempts=TmConst.DEFAULT_WAITING_ATTEMPTS,
                                                       interval=TmConst.DEFAULT_WAITING_INTERVAL,
                                                       throw_exception_on_failure=False):
    def check_events_completed(notifications):
        if not notifications:
            logger.debug(
                _T("EVENT_NOTIFICATION_COMPLETED").format(event_type=event_type, resource_aps_id=resource_aps_id))
            return True

        logger.debug(_T("EVENT_NOTIFICATION_RUNNING").format(event_type=event_type, resource_aps_id=resource_aps_id))
        return False

    def check_events_appeared(notifications):
        if notifications:
            logger.debug(
                _T("EVENT_NOTIFICATION_REGISTERED").format(event_type=event_type, resource_aps_id=resource_aps_id))
            return True

        logger.debug(
            _T("EVENT_NOTIFICATION_NOT_REGISTERED").format(event_type=event_type, resource_aps_id=resource_aps_id))
        return False

    func = FunctionResultWaiter(
        get_event_notifications,
        MigrationNoRepeatError(
            _T("EVENT_REGISTRATION_TIMED_OUT").format(event_type=event_type, resource_aps_id=resource_aps_id)),
        check_function=check_events_appeared,
        max_attempts=TmConst.EVENT_APPEARING_ATTEMPTS,
        interval_between_attempts=TmConst.EVENT_APPEARING_INTERVAL,
        throw_exception_on_failure=throw_exception_on_failure
    )
    events_appeared = func.run(resource_aps_id, event_type)

    if events_appeared:
        func = FunctionResultWaiter(
            get_event_notifications,
            MigrationNoRepeatError(
                _T("EVENT_NOTIFICATION_NOT_COMPLETED").format(event_type=event_type, resource_aps_id=resource_aps_id)),
            check_function=check_events_completed,
            max_attempts=attempts,
            interval_between_attempts=interval,
            throw_exception_on_failure=throw_exception_on_failure
        )
        func.run(resource_aps_id, event_type)


def schedule_task(service_type, name, method, mutex=None, description=None, delay=0, params=None):
    # TODO: May be we have to check if task with the same parameters is already exists ?

    scref = uTasks.SCRef(service_type)
    task = uTasks.Task(ref=scref, name=name, method=method, mutex=mutex, description=description, delay=delay)
    for key in params or {}:
        task.setParam(key, params[key])

    task.schedule()
