# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import CommandExecutionError
from migrator.core.utils import strip_multiline_spaces, default
from migrator.core.language import _T
from migrator.core.utils.lang.profiler import sleep
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner import hcl_utils
from migrator.settings import HCL_EXEC_MULTI_ATTEMPTS_COUNT, HCL_EXEC_MULTI_ATTEMPTS_INTERVAL

logger = logging.getLogger(__name__)


class HclRunnerException(CommandExecutionError):
    pass


def hcl_execute_multiple_attempts(run_function, host, command, max_attempts=HCL_EXEC_MULTI_ATTEMPTS_COUNT,
        interval_between_attempts=HCL_EXEC_MULTI_ATTEMPTS_INTERVAL):
    for attempt in range(0, max_attempts):
        try:
            result = run_function()
            if attempt > 0:
                logger.info(
                    _T("COMMAND_EXECUTED_AFTER_ATTEMPTS").format(
                        command=command, host=host, attempts=attempt+1
                    )
                )
            return result
        except HclRunnerException as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            if attempt >= max_attempts - 1:
                raise e
            else:
                logger.error(
                    _T("COMMAND_EXECUTION_RETRY").format(
                        interval_between_attempts=interval_between_attempts
                    )
                )
                sleep(interval_between_attempts, _T("SLEEP_RETRY_COMMAND_EXECUTION"))


def hcl_is_windows(runner, host_id, host_description=None):
    """Hack to detect service node platform"""

    host_description = default(host_description, _T("DEFAULT_OA_HOST_DESCRIPTION") % host_id)
    windows_pattern = "User '' does not exist"
    command = 'cmd /c REM'

    def run_func():
        # REM - command used for comments in batch file; does nothing, has no output.
        hcl = hcl_utils.Hcl(runner)
        try:
            exit_code, stdout, stderr = hcl.windows_exec_unchecked(host_id, command, 'utf-8')
            if exit_code != 0:
                # looks like Windows, but there are some problems executing the command,
                # so raise HclRunnerException which means - try again
                logger.error(
                    strip_multiline_spaces(_T("FAILED_DETECT_OS_NODE_COMMAND_COMMAND")).format(
                        stderr=stderr,
                        stdout=stdout,
                        command=command,
                        host_id=host_id,
                        host_description=host_description,
                        hcl_script=hcl.get_windows_hcl_script(command),
                        windows_pattern=windows_pattern
                    )
                )
                raise HclRunnerException(
                    _T("INTERNAL_ERROR_OCCURRED_DURING_EXECUTION_COMMAND"),
                    stdout, stderr, exit_code=exit_code, host_description=host_description
                )

            # no errors occurred, so that is Windows
            return True
        except hcl_utils.HclPleskdCtlNonZeroExitCodeError as e:
            if windows_pattern in e.stdout or windows_pattern in e.stderr:
                return False
            else:
                # looks like a problem communicating the node, or something else unexpected - try again
                logger.error(
                    strip_multiline_spaces(_T("FAILED_DETECT_OS_NODE_COMMAND_COMMAND")).format(
                        stderr=e.stderr,
                        stdout=e.stdout,
                        command=command,
                        host_id=host_id,
                        host_description=host_description,
                        hcl_script=hcl.get_windows_hcl_script(command),
                        windows_pattern=windows_pattern
                    )
                )
                raise HclRunnerException(
                    _T("INTERNAL_ERROR_OCCURRED_DURING_EXECUTION_COMMAND"),
                    e.stdout, e.stderr,
                    host_description=host_description,
                    cause=e, exit_code=e.exit_code
                )

    return hcl_execute_multiple_attempts(run_func, host_description, command)
