# (C)2017 Ingram Micro Inc. all Rights Reserved
import base64
import logging
import tempfile

from lxml import etree as et

from migrator.core import CommandExecutionError
from migrator.core.language import _T
from migrator.core.utils import default, safe_string_repr
from migrator.core.utils.filetypes import xml
from migrator.core.utils.registry import Registry

logger = logging.getLogger(__name__)


def _bool_str(val):
    if val:
        return "yes"
    else:
        return "no"


class HclPleskdCtlNonZeroExitCodeError(CommandExecutionError):
    """Exception that raised in case of non zero exit code from ssh runner"""
    pass


class HclDocument(object):
    """
    Class to represent OA HCL document as ElemetTree tree.

    Pass broken_hcl=True for different HCL syntax (OA 6.1)
    """
    def __init__(self, broken_hcl=False):
        self._xml_doc = et.ElementTree(xml.elem('HCL'))
        self._broken_hcl = broken_hcl

    def declare_vars(self, hcl_vars):
        """
        :type hcl_vars: dict
        :rtype: None
        """
        declare_elem = self._get_declare_element()
        for name, value in hcl_vars.iteritems():
            declare_elem.append(self._declare_var(name, value))

    def exec_command(self, command, outvar=None, errvar=None, retvar=None, env=None):
        """
        Execute simple command

        :type command: string | unicode
        :type outvar: string | unicode | None
        :type errvar: string | unicode | None
        :type retvar: string | unicode | None
        :type env: dict | None
        :rtype: None
        """
        env_vars = default(env, {})
        exec_elem = xml.elem('EXEC', self._declare_env_vars(env_vars), dict(command=command))

        self._set_elem_attr(exec_elem, 'outvar', outvar)
        self._set_elem_attr(exec_elem, 'errvar', errvar)
        self._set_elem_attr(exec_elem, 'retvar', retvar)

        self._get_perform_element().append(exec_elem)

    def execb64_command(self, command, args=None, stdin=None, outvar=None, errvar=None, retvar=None, env=None):
        """
        Execute command and pass some content to its stdin (Linux only)

        :type command: string | unicode
        :type args: list | None
        :type stdin: string | unicode | None
        :type outvar: string | unicode | None
        :type errvar: string | unicode | None
        :type retvar: string | unicode | None
        :type env: dict | None
        :rtype: None
        """
        children = []
        if args:
            children.extend(self._declare_args(command, args))
        if env:
            children.extend(self._declare_env_vars(env))

        exec_elem = xml.elem('EXECB64', children, dict(command=command, user='root', group='root'))
        self._set_elem_attr(exec_elem, 'outvar', outvar)
        self._set_elem_attr(exec_elem, 'errvar', errvar)
        self._set_elem_attr(exec_elem, 'retvar', retvar)

        if stdin:
            exec_elem.text = base64.encodestring(stdin)

        self._get_perform_element().append(exec_elem)

    def createfile(self, filename, contents):
        """
        Create file on disk

        :type filename: string | unicode
        :type contents: string | unicode
        :rtype: None
        """
        createfile_elem = xml.text_elem(
            'CREATEFILEB64', base64.b64encode(contents), dict(path=filename, owner='', group='', overwrite='yes')
        )

        self._get_perform_element().append(createfile_elem)

    def readfile(self, filename, contents_var):
        """
        Read file from disk and put its content to specified variable

        :type filename: string | unicode
        :type contents_var: string | unicode
        :rtype: None
        """
        self._get_perform_element().append(xml.elem('READB64', None, dict(path=filename, variable=contents_var)))

    def chmod(self, path, perm, recursive=False, action="set"):
        self._get_perform_element().append(
            xml.elem('CHMOD', None, dict(path=path, perm=perm, recursive=_bool_str(recursive), action=action))
        )

    def chown(self, path, user, group, recursive=False):
        self._get_perform_element().append(
            xml.elem('CHOWN', None, dict(path=path, user=user, group=group, recursive=_bool_str(recursive)))
        )

    def tostring(self):
        """
        Get HCL document as pretty printed string

        :rtype: string | unicode
        """

        doc_head = '''<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
<!DOCTYPE HCL PUBLIC "HCL" "HCL 1.0">
'''
        return xml.prettify_xml('%s%s' % (doc_head, xml.create_xml_stream(self._xml_doc, False).read()))

    @staticmethod
    def _set_elem_attr(elem, attr_name, attr_value):
        if attr_value is None:
            return

        elem.set(attr_name, attr_value)

    def _declare_var(self, name, value):
        var_elem = xml.elem('VAR', None, dict(name=name, transient='no'))
        if self._broken_hcl:
            var_elem.set('attrValue', value)
        else:
            var_elem.set('value', value)

        return var_elem

    @staticmethod
    def _declare_args(command, args):
        args_list = [xml.elem('ARG', None, dict(value=command))]
        for arg in args:
            args_list.append(xml.elem('ARG', None, dict(value=arg)))

        return args_list

    @staticmethod
    def _declare_env_vars(env):
        env_vars = list()
        for name, value in env.iteritems():
            env_vars.append(xml.elem('ENV_VAR', None, dict(name=name, value=value)))

        return env_vars

    def _get_root_child_element(self, name):
        root = self._xml_doc.getroot()
        if root.find(name) is None:
            root.append(et.Element(name))

        return root.find(name)

    def _get_perform_element(self):
        return self._get_root_child_element('PERFORM')

    def _get_declare_element(self):
        return self._get_root_child_element('DECLARE')


class Hcl(object):
    """
    Class to perform some POA HCL commands on POA hosts
    """
    _poa_root = '/usr/local/pem'

    def __init__(self, oa_runner):
        self._oa_runner = oa_runner

    def run_hcl(
            self, host_id, script, output_codepage='utf-8', log_script=None, error_policy='strict', log_output=True
    ):
        log_script = default(log_script, script)

        logger.debug(_T("RUN_HCL_DETAILS"), host_id, safe_string_repr(log_script))
        with tempfile.NamedTemporaryFile(dir=Registry.get_instance().get_context().migrator_server.session_dir(), delete=True) as tmp:
            tmp.write(script)
            tmp.flush()
            exit_status, stdout_content, stderr_content = self._oa_runner.sh_unchecked(
                u'{poa_root}/bin/pleskd_ctl -f {poa_root}/etc/pleskd.props processHCL {filename} {host_id}'.format(
                    poa_root=self._poa_root, filename=tmp.name, host_id=host_id
                ),
                output_codepage=output_codepage,
                error_policy=error_policy,
                log_output=log_output
            )
        if exit_status != 0:
            raise HclPleskdCtlNonZeroExitCodeError(None, stdout_content, stderr_content)

        return stdout_content, stderr_content

    def process_hcl(
            self, host_id, script, output_codepage='utf-8', log_script=None, error_policy='strict', log_output=True
    ):
        """Returns variables dictionary as returned by processHCL"""
        # Stderr contains pleskd debug messages, they are already logged by ssh_utils.run
        # Stdout contains variable values in format "name\x00value\x00name2\x00value2\0xx"
        stdout, _ = self.run_hcl(
            host_id, script, output_codepage=output_codepage, log_script=log_script, error_policy=error_policy,
            log_output=log_output
        )

        parts = stdout.split('\x00')[:-1]  # Split will return empty string after last '\x00', trim it.
        variables = dict(zip(parts[::2], parts[1::2]))

        return variables

    def get_windows_hcl_script(self, command, env=None):
        hcl = self._get_hcl_document()
        hcl.declare_vars(dict(stdout='', stderr='', exit_code=''))
        hcl.exec_command(command, outvar='stdout', errvar='stderr', retvar='exit_code', env=env)

        return hcl.tostring()

    def windows_exec_unchecked(self, host_id, command, codepage, error_policy='strict', env=None, log_output=True):
        script = self.get_windows_hcl_script(command, env)

        variables = self.process_hcl(
            host_id, script, output_codepage=codepage, error_policy=error_policy, log_output=log_output
        )
        exit_code, stdout, stderr = int(variables['exit_code']), variables['stdout'], variables['stderr']

        return exit_code, stdout, stderr

    def upload_file(self, host_id, file_name, file_contents):
        """Upload file to Windows host"""
        hcl = self._get_hcl_document()
        hcl.createfile(file_name, file_contents)
        script = hcl.tostring()

        # avoid logging contents of any file for security reasons
        temp_hcl = self._get_hcl_document()
        temp_hcl.createfile(file_name, '')
        log_script = temp_hcl.tostring()
        log_script = log_script.replace('/>', '>******</CREATEFILEB64>')

        self.process_hcl(host_id, script, log_script=log_script)

    def get_file_contents(self, host_id, file_name):
        out_var = 'contents_base64'

        hcl = self._get_hcl_document()
        hcl.declare_vars({out_var: ''})
        hcl.readfile(file_name, out_var)

        result = self.process_hcl(host_id, hcl.tostring(), log_output=False)
        return base64.decodestring(result[out_var])

    def chmod(self, host_id, path, perm, recursive, action):
        hcl = self._get_hcl_document()
        hcl.chmod(path, perm, recursive, action)
        self.process_hcl(host_id, hcl.tostring())

    def chown(self, host_id, path, user, group, recursive):
        hcl = self._get_hcl_document()
        hcl.chown(path, user, group, recursive)
        self.process_hcl(host_id, hcl.tostring())

    def get_unix_hcl_script(self, command, args, stdin_content=None, env=None):
        if args is None:
            args = []

        hcl = self._get_hcl_document()
        hcl.declare_vars(dict(stdout='', stderr='', exit_code=''))
        hcl.execb64_command(command, args, stdin_content, outvar='stdout', errvar='stderr', retvar='exit_code', env=env)

        return hcl.tostring()

    def unix_exec_unchecked(self, host_id, command, args, stdin_content=None, env=None, log_output=True):
        script = self.get_unix_hcl_script(command, args, stdin_content, env)
        variables = self.process_hcl(host_id, script, log_output=log_output)
        exit_code, stdout, stderr = int(variables['exit_code']), variables['stdout'], variables['stderr']

        return exit_code, stdout, stderr

    @staticmethod
    def _get_hcl_document():
        return HclDocument()

