# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.utils import safe_string_repr, default
from migrator.core.utils.lang.profiler import get_default_steps_profiler
from migrator.core.utils.networking.run_command import BaseRunner
from migrator.core.utils.operating_system import unix_utils
from migrator.core.language import _T
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner.hcl_common import HclRunnerException, \
    hcl_execute_multiple_attempts
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner import hcl_utils

logger = logging.getLogger(__name__)
profiler = get_default_steps_profiler()


class UnixHclRunner(BaseRunner):
    def __init__(self, runner, poa_host_id, host_description=None):
        self.poa_host_id = poa_host_id
        self.host_description = host_description
        self._hcl = hcl_utils.Hcl(runner)

    def run(
        self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True, valid_codes=None
    ):
        """The same as run_unchecker(), but checks exit code and returns only stdout"""
        def run_func():
            log_cmd_str = self._get_cmd_str_listargs(cmd, args)
            exit_code, stdout, stderr = self._run_unchecked(
                cmd, args, stdin_content, output_codepage, error_policy, env
            )
            valid_exit_codes = valid_codes if valid_codes is not None else [0]

            if exit_code not in valid_exit_codes:
                logger.debug(
                    _T("LOG_COMMAND_FAILED"),
                    log_cmd_str, exit_code, safe_string_repr(stdout), safe_string_repr(stderr)
                )
                logger.error(
                    _T("COMMAND_RETURNED_NON_ZERO_EXIT_CODE").format(
                        stdout=stdout,
                        stderr=stderr,
                        log_command=log_cmd_str,
                        host=self._get_host_description(),
                        host_id=self.poa_host_id,
                        hcl_script=self._hcl.get_unix_hcl_script(cmd, args, stdin_content, env)
                    )
                )
                raise HclRunnerException(
                    _T("INTERNAL_ERROR_OCCURRED_DURING_EXECUTION_COMMAND"),
                    stdout, stderr, exit_code=exit_code
                )

            return stdout

        return hcl_execute_multiple_attempts(
            run_func, self._get_host_description(), self._get_cmd_str_listargs(cmd, args)
        )

    def run_unchecked(self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None):
        run_func = lambda: self._run_unchecked(cmd, args, stdin_content, output_codepage, error_policy, env)

        return hcl_execute_multiple_attempts(
            run_func, self._get_host_description(), self._get_cmd_str_listargs(cmd, args)
        )

    def _run_unchecked(
        self, cmd, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None, log_output=True
    ):
        log_cmd_str = self._get_cmd_str_listargs(cmd, args)

        try:
            with profiler.measure_command_call(log_cmd_str, self._get_host_description()):
                return self._hcl.unix_exec_unchecked(self.poa_host_id, cmd, args, stdin_content, env, log_output)
        except hcl_utils.HclPleskdCtlNonZeroExitCodeError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            logger.error(
                _T("FAILED_TO_EXECUTE_COMMAND_DESCRIPTION").format(
                    stderr=e.stderr,
                    stdout=e.stdout,
                    command=log_cmd_str,
                    host=self._get_host_description(),
                    host_id=self.poa_host_id,
                    hcl_script=self._hcl.get_unix_hcl_script(cmd, args, stdin_content, env)
                )
            )
            raise HclRunnerException(
                _T("INTERNAL_ERROR_OCCURRED_DURING_EXECUTION_COMMAND"),
                '', unicode(e), cause=e, host_description=self._get_host_description(), exit_code=e.exit_code
            )

    def sh(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None, error_policy='strict', env=None,
        log_output=True, valid_codes=None
    ):
        """The same as sh_unchecked(), but checks exit code and returns only stdout"""
        log_cmd_str = self._get_cmd_str_formatargs(cmd_str, args)

        def run_func():
            exit_code, stdout, stderr = self._sh_unchecked(
                cmd_str, args, stdin_content, output_codepage, error_policy, env, log_output
            )
            valid_exit_codes = valid_codes if valid_codes is not None else [0]

            if exit_code not in valid_exit_codes:
                logger.debug(
                    _T("LOG_COMMAND_FAILED"),
                    log_cmd_str, exit_code, safe_string_repr(stdout), safe_string_repr(stderr)
                )
                logger.error(
                    _T("COMMAND_RETURNED_NON_ZERO_EXIT_CODE").format(
                        stdout=stdout,
                        stderr=stderr,
                        log_command=log_cmd_str,
                        host=self._get_host_description(),
                        host_id=self.poa_host_id,
                        hcl_script=self._hcl.get_unix_hcl_script(log_cmd_str, [], stdin_content, env)
                    )
                )
                raise HclRunnerException(
                    _T("INTERNAL_ERROR_OCCURRED_DURING_EXECUTION_COMMAND"),
                    stdout, stderr, exit_code=exit_code
                )

            return stdout

        return hcl_execute_multiple_attempts(run_func, self._get_host_description(), log_cmd_str)

    def sh_unchecked(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None,
        error_policy='strict', env=None, log_output=True
    ):
        run_func = lambda: self._sh_unchecked(cmd_str, args, stdin_content, output_codepage, error_policy, env, log_output)

        return hcl_execute_multiple_attempts(
            run_func, self._get_host_description(), self._get_cmd_str_formatargs(cmd_str, args)
        )

    def _sh_unchecked(
        self, cmd_str, args=None, stdin_content=None, output_codepage=None,
        error_policy='strict', env=None, log_output=True
    ):
        run_cmd, run_args = self._get_command_list(cmd_str, args)

        return self._run_unchecked(run_cmd, run_args, stdin_content, output_codepage, error_policy, env, log_output)

    @staticmethod
    def _get_command_list(cmd_str, args):
        if args is not None:
            cmd_str = unix_utils.format_command(cmd_str, **args)

        return '/bin/sh', ['-c', cmd_str]

    def get_file(self, remote_filename, local_filename):
        with open(local_filename, 'wb') as fp:
            fp.write(self._hcl.get_file_contents(self.poa_host_id, remote_filename))

    def get_file_contents(self, remote_filename):
        return self._hcl.get_file_contents(self.poa_host_id, remote_filename)

    def upload_file(self, local_filename, remote_filename):
        with open(local_filename, 'rb') as fp:
            return self._hcl.upload_file(self.poa_host_id, remote_filename, fp.read())

    def upload_file_content(self, filename, content):
        return self._hcl.upload_file(self.poa_host_id, filename, content)

    def move(self, src_path, dst_path):
        return self.run('/bin/mv', [src_path, dst_path])

    def get_files_list(self, path):
        return unix_utils.get_files_list(self, path)

    def remove_file(self, filename):
        return self.run('/bin/rm', ['-f', filename])

    def remove_directory(self, directory, is_remove_root=True, log_output=True):
        if is_remove_root:
            return self.run('/bin/rm', ['-rf', directory])
        else:
            return self.run(
                '/bin/find',
                [directory, '-maxdepth', '1', '-not', '-path', directory, '-exec', 'rm', '-rf', '{}', ';']
            )

    def mkdir(self, dirname):
        return self.run('/bin/mkdir', ['-p', dirname])

    def mktempfile(self, dirname):
        """Create a 'filename' file with unique name."""
        self.mkdir(dirname)
        tmpfilename = self.run('/bin/mktemp', ['-p', dirname])
        return tmpfilename.strip()

    def mktempdir(self, dirname):
        """Create a 'dirname' subdirectory with unique name."""
        self.mkdir(dirname)
        tmpdirname = self.run('/bin/mktemp', ['-d', '-p', dirname])
        return tmpdirname.strip()

    def file_exists(self, filename):
        return unix_utils.file_exists(self, filename)

    def _get_host_description(self):
        return default(self.host_description, _T("DEFAULT_OA_HOST_DESCRIPTION") % self.poa_host_id)

    @staticmethod
    def _get_cmd_str_listargs(cmd, args):
        return unix_utils.format_command_list(cmd, args)

    def _get_cmd_str_formatargs(self, cmd_str, args):
        run_cmd, run_args = self._get_command_list(cmd_str, args)
        return unix_utils.format_command_list(run_cmd, run_args)

    def stat(self, path):
        return unix_utils.path_stat(self, path)

    def chmod(self, path, perm, recursive=False, action='set'):
        self._hcl.chmod(self.poa_host_id, path, str(perm), recursive, action)

    def chown(self, path, user, group, recursive=False):
        self._hcl.chown(self.poa_host_id, path, user, group, recursive)
