# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationNoRepeatError
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.language import _T
from migrator.odin_automation.platform.lib.api.aps.subscription import get_subscription_aps_id
from migrator.odin_automation.services.oaci.lib.aps.virtual_environment import get_ve
from migrator.odin_automation.services.oaci.lib.vps import \
    get_random_active_target_pcs_server, launch_vps_import, get_human_readable_ve_ids, wait_for_vps_import, \
    get_takeoverable_pcs_server, get_ve_info
from migrator.settings import MIGRATION_TYPE_TAKEOVER, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass

logger = logging.getLogger(__name__)


class ImportVPS(SubscriptionAction):
    def run(self, global_context, subscription):
        """
        :param global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        """
        # Check if VPS already imported
        ve = get_ve(get_subscription_aps_id(subscription.id))
        if ve is not None and ve['forMigrate'] is False:
            logger.info(_T("SKIP_VPS_IMPORT").format(vps_id=get_ve_info(subscription).ve_id, sub_name=subscription.name))
            return

        if global_context.migration_type == MIGRATION_TYPE_TAKEOVER:
            target_pcs = get_takeoverable_pcs_server(global_context, subscription)
        else:
            target_pcs = get_random_active_target_pcs_server(global_context, subscription)

        try:
            launch_vps_import(subscription, target_pcs)
            wait_for_vps_import(subscription)
        except MigrationNoRepeatError as e:
            raise e
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationNoRepeatError(_T('UNEXPECTED_ERROR_OCCURRED_SEE_THE_LOGFILE_MSG'))

    def get_description(self):
        """
        :return: string
        """
        return _T("IMPORT_VE_INTO_OA_SUBSCRIPTION")

    def filter_subscription(self, global_context, subscription):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: bool False - to skip subscription processing
        """
        return global_context.migration_type == MIGRATION_TYPE_TAKEOVER and \
               global_context.migration_mode == MigrationMode.PRODUCTION and \
               has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.VPS])

    def is_critical(self):
        """
        :return: bool
        """
        return True

    def get_failure_message(self, global_context, subscription):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: string
        """
        return _T("FAILED_IMPORT_VPS_FOR_SUBSCRIPTION").format(
            vps_id=get_human_readable_ve_ids(subscription),
            sub_id=subscription.name
        )
