# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import os
import hashlib

import migrator.odin_automation.services.oaci

from migrator.core import MigrationNoRepeatError
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.language import _T
from migrator.core.utils.filetypes.json_utils import get_json
from migrator.core.utils.operating_system.unix_utils import join_package_path
from migrator.odin_automation.platform.lib.api.aps import subscription as aps_subscription_api
from migrator.odin_automation.platform.lib.api.aps.account import get_account_by_subscription
from migrator.odin_automation.services.oaci.lib.vps import \
    get_human_readable_ve_ids, get_takeoverable_pcs_server, get_ve_info
from migrator.settings import MIGRATION_TYPE_TAKEOVER, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass

logger = logging.getLogger(__name__)


class NormalizeVPS(SubscriptionAction):
    NORMALIZATION_SCRIPT_NAME = 've_normalization.py'
    NORMALIZATION_SCRIPT_SOURCE_PATH = join_package_path(migrator.odin_automation.services.oaci, 'agent')
    RUN_NORMALIZE_COMMAND = 'python "{normalization_script_path}" "{ve_id}" "{billing_subscription_json_info}"'

    def run(self, global_context, subscription):
        """
        :param global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        """
        try:
            target_pcs = get_takeoverable_pcs_server(global_context, subscription)
            account_id = get_account_by_subscription(subscription.id)['id']
            ve_normalization_script_target_path = target_pcs.get_session_file_path(self.NORMALIZATION_SCRIPT_NAME)
            ve_normalization_script_source_path = os.path.join(
                self.NORMALIZATION_SCRIPT_SOURCE_PATH,
                self.NORMALIZATION_SCRIPT_NAME
            )
            ve_id = get_ve_info(subscription).ve_id
            with target_pcs.runner() as pcs_runner:
                if not pcs_runner.file_exists(ve_normalization_script_target_path):
                    pcs_runner.upload_file(ve_normalization_script_source_path, ve_normalization_script_target_path)
                else:
                    source_script_hash = hashlib.md5(open(ve_normalization_script_source_path, 'rb').read()).hexdigest()
                    target_script_hash = hashlib.md5(
                        pcs_runner.get_file_contents(ve_normalization_script_target_path)
                    ).hexdigest()
                    if source_script_hash != target_script_hash:
                        pcs_runner.upload_file(ve_normalization_script_source_path, ve_normalization_script_target_path)

                target_billing_subscription_info = get_json(
                    aps_subscription_api.get_billing_subscription(account_id, subscription.id)
                )
                subscription_info_file_json = pcs_runner.mktempfile(target_pcs.session_dir())
                pcs_runner.upload_file_content(subscription_info_file_json, target_billing_subscription_info)
                cmd = self.get_normalization_cmd(ve_normalization_script_target_path, ve_id, subscription_info_file_json)
                exit_status, stdout_content, stderr_content = pcs_runner.sh_unchecked(cmd)
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationNoRepeatError(_T('UNEXPECTED_ERROR_OCCURRED_SEE_THE_LOGFILE_MSG'))

        if exit_status != 0:
            raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERR_NORMALIZATION_FAILED').format(
                cmd=cmd,
                pcs_ip=target_pcs.ip(),
                stderr=stdout_content + stderr_content,
                script=ve_normalization_script_source_path
            ))

    def get_normalization_cmd(self, script_path, ve_id, billing_subscription_json_info):
        """
        :param script_path: string Tha path to target normalization script, e.g: '/opt/oam-session/ve_normalization.py'
        :param ve_id: string The VPS UUID, e.g: 'aee631ff-7e1d-4c7b-a343-3151d3863202'
        :param billing_subscription_json_info: string The path to a file with subscription billing info in JSON format,
                e.g: '/opt/oam-session/tmp.UdVoivcdXU'
        :return: string
        """
        return self.RUN_NORMALIZE_COMMAND.format(
            normalization_script_path=script_path,
            ve_id=ve_id,
            billing_subscription_json_info=billing_subscription_json_info
        )

    def get_description(self):
        """
        :return: string
        """
        return _T("NORMALIZE_VE_INTO_OA_SUBSCRIPTION")

    def filter_subscription(self, global_context, subscription):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: bool False - to skip subscription processing
        """
        return global_context.migration_type == MIGRATION_TYPE_TAKEOVER and \
               global_context.migration_mode == MigrationMode.PRODUCTION and \
               has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.VPS])

    def is_critical(self):
        """
        :return: bool
        """
        return True

    def get_failure_message(self, global_context, subscription):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: string
        """
        return _T("FAILED_NORMALIZE_VPS_FOR_SUBSCRIPTION").format(
            vps_id=get_human_readable_ve_ids(subscription),
            sub_id=subscription.name
        )
