# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import re

from collections import namedtuple

from migrator.core import MigrationNoRepeatError, MigrationError
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.filetypes.config_utils import ConfigHostSection
from migrator.core.utils.networking.runner import get_runner
from migrator.odin_automation.services.oaci.lib.vps import \
    get_ve_info, get_vps_service, get_random_active_target_pcs_server
from migrator.settings import MIGRATION_TYPE_FULL, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass

logger = logging.getLogger(__name__)

ve_info = namedtuple('ve_info', ('ve_id', 've_type'))

LOCALHOST_IP = 'localhost'


class CopyVPS(SubscriptionAction):
    # Example: prlctl migrate root:secret@10.37.54.141/105 localhost/105
    _migrate_ve = 'prlctl migrate ' \
               '{source_pcs_username}:{source_pcs_password}@{source_pcs_ip}/{source_ve_id} ' \
               '{target_pcs_ip}/{target_ve_id}'
    # Example: prlctl stop 115
    _stop_ve = 'prlctl stop {ve_id}'
    # Example: prlctl start bf845b29-6553-409a-86cf-0c56900673e7
    _start_ve = 'prlctl start {ve_id}'

    def run(self, global_context, subscription):
        """
        :param global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        """
        vps_service = get_vps_service(subscription)
        source_ve_info = get_ve_info(subscription)
        source_pcs_host_model = PlatformReaderDB().get_source_host(vps_service.source_host_external_id)
        s_pcs_ip = source_pcs_host_model.ip
        s_ve_id = source_ve_info.ve_id
        s_ve_type = source_ve_info.ve_type
        source_pcs_config_section = ConfigHostSection(global_context.config, s_pcs_ip)
        target_pcs_server_object = get_random_active_target_pcs_server(global_context, subscription)
        t_pcs_ip = target_pcs_server_object.ip()
        err_unexpected = _T('UNEXPECTED_ERROR_OCCURRED_SEE_THE_LOGFILE_MSG')
        err_msg_base = _T('VPS_MIGRATION_ERR_MIGRATE_VPS_BASE_MSG').format(
            s_ve_type=s_ve_type, s_ve_id=s_ve_id, s_pcs_ip=s_pcs_ip, t_pcs_ip=t_pcs_ip, t_ve_type=s_ve_type,
            t_ve_id=s_ve_id
        )

        # ToDo: move the 1, 2 and 3 steps into migrator.odin_automation.services.oaci.lib.*
        # 1. Stop source VE
        stop_ve_cmd = self._get_stop_ve_cmd(s_ve_id)
        source_runner = get_runner(source_pcs_config_section)
        try:
            exit_status, stdout_content, stderr_content = source_runner.sh_unchecked(stop_ve_cmd)
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationNoRepeatError(
                err_msg_base + _T('VPS_MIGRATION_ERR_STOP_S_VE').format(details=err_unexpected)
            )

        if exit_status != 0:
            raise MigrationError(err_msg_base + _T('VPS_MIGRATION_ERR_STOP_S_VE').format(details=stderr_content))

        with target_pcs_server_object.runner() as target_runner:
            # 2. Move VE from source to target
            migrate_cmd = self._get_migrate_ve_cmd(
                s_pcs_pass=re.escape(source_pcs_config_section.get('password')), s_ve_id=s_ve_id, s_pcs_ip=s_pcs_ip,
                s_pcs_user=source_pcs_config_section.get('username'), t_ve_id=s_ve_id, t_pcs_ip=LOCALHOST_IP
            )
            try:
                exit_status, stdout_content, stderr_content = target_runner.sh_unchecked(migrate_cmd)
            except Exception:
                # todo: try to start VPS if the moving fails and VPS was running before migration (prlctl status VE_ID)
                logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                raise MigrationNoRepeatError(
                    err_msg_base + _T('VPS_MIGRATION_ERR_MOVE_VE').format(details=err_unexpected)
                )

            if exit_status != 0:
                # todo: try to start VPS if the moving fails and VPS was running before migration
                raise MigrationError(err_msg_base + _T('VPS_MIGRATION_ERR_MOVE_VE').format(details=stderr_content))

            # 3. Start migrated VE
            start_ve_cmd = self._get_start_ve_cmd(s_ve_id)
            try:
                exit_status, stdout_content, stderr_content = target_runner.sh_unchecked(start_ve_cmd)
            except Exception:
                logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                raise MigrationNoRepeatError(
                    err_msg_base + _T('VPS_MIGRATION_ERR_START_MOVED_VE').format(details=err_unexpected)
                )

            if exit_status != 0:
                raise MigrationError(
                    err_msg_base + _T('VPS_MIGRATION_ERR_START_MOVED_VE').format(details=stderr_content)
                )

    def _get_migrate_ve_cmd(self, s_pcs_user, s_pcs_pass, s_pcs_ip, s_ve_id, t_pcs_ip, t_ve_id):
        """
        :param s_pcs_user: string Username for the source PCS server, ex: 'root'
        :param s_pcs_pass: string Password for the source PCS server, ex: 'secret'
        :param s_pcs_ip: string IP address of the source PCS server, ex: '10.37.54.141'
        :param s_ve_id: string ID of the source VE, ex: '105'
        :param t_pcs_ip: string IP address of the target PCS server, ex: 'localhost'
        :param t_ve_id: string ID of the target VE, ex: '100105'
        :return: string. Example: 'prlctl migrate root:sec\"ret@10.37.54.141/105 localhost/105'
        """
        return self._migrate_ve.format(
            source_pcs_username=s_pcs_user,
            source_pcs_password=s_pcs_pass,
            source_pcs_ip=s_pcs_ip,
            source_ve_id=s_ve_id,
            target_pcs_ip=t_pcs_ip,
            target_ve_id=t_ve_id
        )

    def _get_stop_ve_cmd(self, ve_id):
        """
        :param ve_id: string ID of the VE to be stopped, ex: '105', '4722cff2-9af7-40a4-93af-8206ab101d43'
        :return: string. Example: 'prlctl stop 105'
        """
        return self._stop_ve.format(ve_id=ve_id)

    def _get_start_ve_cmd(self, ve_id):
        """
        :param ve_id: string ID of the VE to be started, ex: '105', '4722cff2-9af7-40a4-93af-8206ab101d43'
        :return: string. Example: 'prlctl start 4722cff2-9af7-40a4-93af-8206ab101d43'
        """
        return self._start_ve.format(ve_id=ve_id)

    def get_description(self):
        """
        :return: string
        """
        return _T("COPY_VE_FROM_SERVERS")

    def filter_subscription(self, global_context, subscription):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: bool False - to skip subscription processing
        """
        return global_context.migration_type == MIGRATION_TYPE_FULL and \
               global_context.migration_mode == MigrationMode.PRODUCTION and \
               has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.VPS])

    def is_critical(self):
        """
        :return: bool
        """
        return True

    def get_failure_message(self, global_context, subscription):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :return: string
        """
        return _T("FAILED_MOVE_VPS_FOR_SUBSCRIPTION").format(subscription_name=subscription.name)
