# (C)2017 Ingram Micro Inc. all Rights Reserved
import json
import random
import string
import subprocess
import sys
import time

VE_STATE_STOPPED = 'stopped'


def fatal(error):
    print error
    sys.exit(2)


class App:
    def __init__(self, ve_id, subscription):
        self.subscription, self.ve_id = subscription, ve_id
        self.ve = None
        self.old_state = None

    def normalize(self):
        self.get_ve()
        self.stop_ve()
        self.normalize_cpu()
        self.normalize_ram()
        self.normalize_name()
        self.normalize_net_device()
        self.start_ve()
        print "[INFO] The normalization was completed successfully."

    def stop_ve(self):
        if self.old_state == VE_STATE_STOPPED:
            print "[INFO] The VPS %s is already stopped. VPS stopping will be skipped." % self.ve_id
            return

        print "[DEBUG] Stopping the VPS ..."
        self.shell_multi("prlctl stop %s" % self.ve_id)

    def start_ve(self):
        print "[DEBUG] Starting the VPS ..."
        self.shell_multi("prlctl start %s" % self.ve_id)

    def normalize_cpu(self):
        cpus = self.get_sub_resource("cpu", "unit")
        self.shell("prlctl set %s --cpuunits 1000 --cpulimit 0 --cpus %s" % (self.ve_id, cpus))

    def normalize_ram(self):
        ram = self.get_sub_resource("ram", "MB")
        self.shell("prlctl set %s --memsize %s" % (self.ve_id, ram))

    def normalize_name(self):
        name = "%s.mm-imported.%s" % (self.subscription["subscriptionId"], self.generate_random_string())
        self.shell("prlctl set %s --name \"%s\"" % (self.ve_id, name))

    def generate_random_string(self, lenght=6):
        chars = [random.choice(string.lowercase) for _ in range(lenght)]
        random.shuffle(chars)
        return ''.join(chars)

    def normalize_net_device(self):
        devices = self.get_net_devices()
        if not devices:
            print "[WARNING] Network devices were not found in the VPS %s. The normalization of VPS network devices will be skipped." % self.ve_id
            return

        for device in devices:
            if self.is_net_device_valid(device):
                self.net_device_del(device)
                self.net_device_add(device)

    def is_net_device_valid(self, device):
        for index in ['ID', 'ips', 'gw']:
            if index not in device:
                ifname = device['ifname'] if 'ifname' in device else 'unknown interface'
                print "[WARNING] An invalid network interface '%s' was detected. The normalization of it will be skipped." % ifname
                return False

        return True

    def net_device_add(self, device):
        cmd = "prlctl set %s --device-add net" % self.ve_id
        cmd += " --ipadd \"%s\"" % (device["ips"].split()[0])
        cmd += " --gw \"%s\"" % device["gw"]
        cmd += " --network \"Bridged\""
        self.shell(cmd)

    def net_device_del(self, device):
        cmd = "prlctl set %s --device-del %s" % (self.ve_id, device["ID"])
        self.shell(cmd)

    def get_sub_resource(self, aps_prop, unit_id):
        resources = filter(lambda r: "apsProperty" in r and r["apsProperty"] == aps_prop and r["unitId"] == unit_id,
                           self.subscription["resourceRates"])
        if len(resources) == 0:
            raise Exception("Subscription resource %s not found" % aps_prop)

        return int(resources[0]["included"])

    def get_net_devices(self):
        hw = self.ve["Hardware"]
        for key in hw.keys():
            hw[key]["ID"] = key
        return filter(lambda v: "network" in v, hw.values())

    def shell_multi(self, cmd, attempts=30, retry_interval=10):
        attempt = 1
        while attempt <= attempts:
            try:
                return self.shell(cmd)
            except Exception:
                attempt += 1
                time.sleep(retry_interval)

        raise Exception("[ERROR] Failed to execute the command '%s' %d times in a row." % (cmd, attempts))

    def shell(self, cmd):
        print "[DEBUG] Executing command: " + cmd
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        r = p.communicate()
        output = r[0]
        error = r[1]
        code = p.returncode

        if code != 0 or error != "":
            error_message = "Command failed with status %s and error:\n%s" % (code, error)
            print "[ERROR] " + error_message
            raise Exception(error_message)

        return output

    def get_ve(self):
        ve_json = self.shell("prlctl list -i %s --json" % self.ve_id)
        ves = json.loads(ve_json)
        self.ve = ves[0]
        self.old_state = self.ve["State"]


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print """Usage: python ve_normalization VE_ID SUBSCRIPTION_JSON_FILE
        VE_ID - uuid of VE to be normalized
        SUBSCRIPTION_JSON_FILE - A file path with billing subscription in JSON format
        """
        sys.exit(1)

    arg_ve_id = sys.argv[1]
    arg_sub_json = sys.argv[2]

    try:
        with open(arg_sub_json) as sub_json:
            arg_sub = json.load(sub_json)
    except Exception as e:
        print e
        fatal("Unable to parse subscription JSON")

    app = App(arg_ve_id, arg_sub)
    try:
        app.normalize()
    except Exception as ex:
        print ex
        fatal("Unable to normalize VE")
