# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from collections import namedtuple

from migrator.core import MigrationNoRepeatError
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.odin_automation.platform.lib.api.aps.host import get_ip_of_first_active_oa_pcs_host
from migrator.odin_automation.platform.lib.api.aps.subscription import get_subscription_aps_id
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import get_aps_connector_with_admin_token
from migrator.odin_automation.platform.lib.task_manager import wait_for_all_scheduled_tasks_on_resource_completed
from migrator.odin_automation.services.oaci.lib.aps.virtual_environment import get_ve_aps_id
from migrator.settings import SOURCE_SERVICE_VPS

logger = logging.getLogger(__name__)

ve_info = namedtuple('ve_info', ('ve_id', 've_type', 've_name'))

OACI_MN_LOGS = ['/var/log/pa/vps.vm2vf.log', '/var/log/pa/vps.aps.log']
OACI_IM_LOGS = ['/var/log/pa/vps.log']

VPS_IMPORT_MAX_ATTEMPTS = -1
VPS_IMPORT_INTERVAL_BETWEEN_ATTEMPTS = 30


def get_vps_service(subscription):
    """
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :return: migrator.core.data.model.service.ServiceModel
    """
    for subscription_service in subscription.source_info.services:
        if subscription_service.type == SOURCE_SERVICE_VPS:
            return subscription_service

    raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERROR_VPS_SERVICE_NOT_FOUND').format(
        subscription_name=subscription.name_idn)
    )


def get_ve_info(subscription):
    """
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :return: namedtuple
    """
    vps_service = get_vps_service(subscription)
    source_ve_id = None
    source_ve_type = None
    source_ve_name = None
    for service_property in vps_service.properties:
        if service_property.key == 've_id':
            source_ve_id = service_property.value
            continue

        if service_property.key == 've_type':
            source_ve_type = service_property.value
            continue

        if service_property.key == 've_name':
            source_ve_name = service_property.value
            continue

    if not source_ve_id or not source_ve_type or not source_ve_name:
        raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERROR_NOT_ALL_REQUIRED_VPS_PARAMETERS_WERE_FOUND').format(
            subscription_name=subscription.name_idn
        ))

    return ve_info(source_ve_id, source_ve_type, source_ve_name)


def get_human_readable_ve_ids(subscription):
    """
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :return: string
    """
    ve_info = get_ve_info(subscription)
    return u'VPS "{vps_name}"'.format(vps_name=ve_info.ve_name)


def get_random_active_target_pcs_server(global_context, subscription):
    """
    :param global_context: migrator.core.global_context.GlobalMigrationContext
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :return: migrator.odin_automation.platform.lib.connections.target_server.AutomationServer
    """
    target_pcs_ip = get_ip_of_first_active_oa_pcs_host()
    if not target_pcs_ip:
        raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERROR_UNABLE_TO_FIND_AVAILABLE_OACI_PCS').format(
            subscription_name=subscription.name_idn
        ))

    oa_connections = global_context.target_panel_obj.get_connections(global_context)
    return global_context.target_panel_obj.get_target_pcs_server(oa_connections, target_pcs_ip)


def get_takeoverable_pcs_server(global_context, subscription):
    """
    :param global_context: migrator.core.global_context.GlobalMigrationContext
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :return: migrator.odin_automation.platform.lib.connections.target_server.AutomationServer
    """
    pcs_host = PlatformReaderDB().get_source_host_by_subscription_service(
        subscription_id=subscription.source_info.external_id,
        service_type=SOURCE_SERVICE_VPS
    )

    if pcs_host is None:
        raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERROR_UNABLE_TO_DETECT_TAKEOVERABLE_PCS').format(
            subscription_name=subscription.name_idn
        ))

    oa_connections = global_context.target_panel_obj.get_connections(global_context)
    return global_context.target_panel_obj.get_target_pcs_server(oa_connections, pcs_host.ip)


def launch_vps_import(subscription, target_pcs):
    """
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :param target_pcs: migrator.odin_automation.platform.lib.connections.target_server.AutomationServer
    """
    try:
        aps = get_aps_connector_with_admin_token()
        subscription_aps_id = get_subscription_aps_id(subscription.id)
        ve_aps_id = get_ve_aps_id(subscription_aps_id)
        ve_info = get_ve_info(subscription)
        ve_uuid = ve_info.ve_id
        aps.post_resource(
            ve_aps_id,
            '/import/{ve_uuid}/at/{target_pcs_hostname}'.format(
                ve_uuid=ve_uuid,
                target_pcs_hostname=target_pcs.get_hostname()
            )
        )
    except Exception:
        logger.error(_T("LOG_EXCEPTION"), exc_info=True)
        raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERR_IMPORT_NOT_RUN'))


def wait_for_vps_import(subscription):
    """
    :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
    :return: migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict The VE APS resource
    """
    def check_vps_status(ve):
        if 'status' in ve and 'state' in ve['status'] and 'transition' in ve['status'] and \
                ve['status']['state'].lower() == 'started' and ve['status']['transition'].lower() == 'started':
            return True
        else:
            return False

    subscription_aps_id = get_subscription_aps_id(subscription.id)
    ve_aps_id = get_ve_aps_id(subscription_aps_id)
    try:
        wait_for_all_scheduled_tasks_on_resource_completed(
            ve_aps_id,
            VPS_IMPORT_MAX_ATTEMPTS,
            VPS_IMPORT_INTERVAL_BETWEEN_ATTEMPTS
        )
    except Exception:
        raise MigrationNoRepeatError(_T('VPS_MIGRATION_ERR_IMPORT_VE_TIMEOUT').format(
            mn_logfiles=', '.join(OACI_MN_LOGS),
            im_logfiles=', '.join(OACI_IM_LOGS)
        ))
