# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import InternalError
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.filetypes.json_utils import get_json
from migrator.odin_automation.platform.lib.data.subscription_importer import OaSubscriptionParameters
from migrator.odin_automation.services.oaci.lib.aps.types import VirtualEnvironment
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import get_aps_connector_with_admin_token
from migrator.odin_automation.platform.lib.api.openapi.service import R_VPS_VST
from migrator.odin_automation.platform.service_import_api import ServiceImportApiPlatform
from migrator.odin_automation.services.oaci.actions.import_vps import ImportVPS
from migrator.odin_automation.services.oaci.actions.normalize_vps import NormalizeVPS
from migrator.odin_automation.services.oaci.actions.suspend_subscription import SuspendObasSubscription
from migrator.odin_automation.services.oaci.actions.transfer.copy_vps import CopyVPS

logger = logging.getLogger(__name__)


class ServiceImportApiVPS(ServiceImportApiPlatform):
    def _get_service_import_actions_full_migration(self, global_context):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :return: list[migrator.core.actions.base.subscription_action.SubscriptionAction]
        """
        return [
            CopyVPS(),
            ImportVPS()
        ]

    def _get_service_import_actions_takeover(self, global_context):
        """
        :param global_context: global_context: migrator.core.global_context.GlobalMigrationContext
        :return: list[]|list[migrator.core.actions.base.subscription_action.SubscriptionAction]
        """
        return [
            SuspendObasSubscription(),
            NormalizeVPS(),
            ImportVPS()
        ]

    def get_service_specific_activation_parameters(self, global_context, migration_item_model):
        """
        :param global_context: migrator.core.session.GlobalMigrationContext
        :param migration_item_model: migrator.core.data.model.migration_project_item.MigrationProjectItemModel
        :return: list
        """
        platform_reader = PlatformReaderDB()
        subscription_model = platform_reader.get_subscription(primary_id=migration_item_model.subscription_id)
        sub_name = subscription_model.name
        service_template = global_context.conn.target.poa_api().service.get_service_template_for_migration_item(
            global_context,
            migration_item_model
        )

        vst_list = service_template.get_resource_full_info(R_VPS_VST)
        logger.debug(_T('VPS_MIGRATION_DEBUG_FOUND_VST_LIST').format(sub_name=sub_name, vst_list=str(vst_list)))

        vst_guid = vst_list[0].get_aps_guid()
        logger.debug(_T('VPS_MIGRATION_DEBUG_SELECTED_VST_FROM_LIST').format(aps_guid=vst_guid, sub_name=sub_name))

        act_params = get_json({
            'aps': {'type': VirtualEnvironment},
            'forMigrate': True,
            'hostname': self._get_ve_hostname(subscription_model),
            'osId': self._get_os_guid(vst_guid)
        })

        return OaSubscriptionParameters(
            service_activation_parameters=[{"var_name": "aps_defaults_1", "var_value": act_params}],
            synchronize_resource_limits=False
        )

    def _get_virtualization_technology(self, subscription_model):
        """
        :param subscription_model: migrator.core.data.model.subscription.SubscriptionModel
        :return: string Examples: 'vm', 'ct'
        """
        for service in subscription_model.services:
            if service.type == 'vps':
                for service_property in service.properties:
                    if service_property.key == 've_type':
                        vps_type = service_property.value
                        logger.debug(_T('VPS_MIGRATION_DEBUG_DETECTED_VIRTUALIZATION_TYPE_FOR_SUBSCRIPTION').format(
                            v_type=vps_type,
                            sub_name=subscription_model.name
                        ))
                        return vps_type

        raise InternalError(
            'Unable to determine a virtualization type for the subscription "%s".'
            % (subscription_model.name + ' (ID: #%s)' % subscription_model.external_id)
        )

    def _get_ve_hostname(self, subscription_model):
        """
        :param subscription_model: migrator.core.data.model.subscription.SubscriptionModel
        :return: string Examples: 'vps1.domain.net', 'xn--vps-vasya--8rs9eyeye.local', '792640c9-a835-4882-a3c1-813510fa6b69'
        """
        for service in subscription_model.services:
            if service.type == 'vps':
                for service_property in service.properties:
                    if service_property.key == 've_hostname':
                        hostname = service_property.value
                        logger.debug(_T('VPS_MIGRATION_DEBUG_DETECTED_HOSTNAME_FOR_SUBSCRIPTION').format(
                            hostname=hostname,
                            sub_name=subscription_model.name
                        ))
                        return hostname

        raise InternalError(
            'Unable to determine the hostname of the VPS for the subscription "%s".'
            % (subscription_model.name + ' (ID: #%s)' % subscription_model.external_id)
        )

    def _get_os_guid(self, virtual_server_template_guid):
        """
        :param virtual_server_template_guid: string VST APS ID, example: '9dd291b7-0e1e-4e1b-9bfc-fb5f401b5dcb'
        :return: string Example: 'fc009392-9533-42d7-b562-48267abc25f7'
        """
        aps = get_aps_connector_with_admin_token()
        vst_resource = aps.get_resource(virtual_server_template_guid)
        if not vst_resource:
            raise InternalError(
                'Unable to determine the operating system for the Virtual Server Template (APS ID "%s").'
                % virtual_server_template_guid
            )

        os_guid = vst_resource['templates']['defaultOS']
        logger.debug(_T('VPS_MIGRATION_DEBUG_DETECTED_OS_FOR_VST').format(
            os_aps_guid=os_guid,
            vst_guid=virtual_server_template_guid
        ))
        return os_guid
