import logging

from migrator.core import MigrationNoRepeatError
from migrator.core.actions.base.common_action import CommonAction
from migrator.core.data.model.migration_project_item import MigrationProjectItemModel
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils import is_subdomain_of_any
from migrator.odin_automation.services.shm.lib import get_sites_list
from migrator.settings import \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SourceHostingClass

logger = logging.getLogger(__name__)


class CheckMigrationProjectItems(CommonAction):
    def run(self, global_context):
        platform_reader = PlatformReaderDB()

        migration_project_items = {}
        for item in platform_reader.get_migration_project_items(SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            migration_project_items[item.subscription_id] = item

        new_migration_project_items = []
        for item in global_context.migration_list_data:
            subscription = platform_reader.get_subscription(item.id)
            if subscription.external_id.startswith('dns_'):
                continue

            child_dns_subscriptions = [
                x for x in platform_reader.get_subscriptions_by_parent(subscription.external_id)
                    if x.external_id.startswith('dns_')
            ]
            child_dns_subscriptions_names = set([x.name.lower().encode('idna') for x in child_dns_subscriptions])

            sites_names_list = set([x.name.lower().encode('idna') for x in get_sites_list(subscription.external_id)])
            sites_names_list_wo_subdomains = set([
                x for x in sites_names_list
                if not is_subdomain_of_any(x, child_dns_subscriptions_names) and not is_subdomain_of_any(x, sites_names_list)
            ])

            new_domains = sites_names_list_wo_subdomains - child_dns_subscriptions_names
            if new_domains:
                msg = _T('THERE_ARE_NEW_DOMAINS_IN_SUBSCRIPTIONS').format(
                    domains=u', '.join(new_domains),
                    source_system_aps_id=platform_reader.aps_id
                )
                if global_context.source_type == SOURCE_PRODUCT_OBAS:
                    msg += ' ' + _T('MAKE_SURE_THAT_THESE_DOMAINS_ARE_REGISTERED_IN_OBAS')

                raise MigrationNoRepeatError(msg)

            for s in filter(lambda x: x.id not in migration_project_items, child_dns_subscriptions):
                item = MigrationProjectItemModel(platform_reader.migration_project_id, s.id, None, None, None, subscription.external_id, False)
                new_migration_project_items.append(item)

        if len(new_migration_project_items):
            platform_reader.session.add_all(new_migration_project_items)
            platform_reader.session.commit()
            logger.debug("Saved %s migration project items to the database" % len(new_migration_project_items))
            raise MigrationNoRepeatError(_T('THERE_ARE_NEW_DOMAINS_IN_MIGRATION_PROJECT').format(
                migration_project_aps_id=global_context.migration_project_aps_id
            ))

    def get_failure_message(self, global_context):
        return ''

    def get_description(self):
        return _T('CHECK_NEW_DOMAIN_AT_SOURCE_SYSTEM')

    def filter_action(self, global_context):
        product = global_context.source_type
        return product in [SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS]
