# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.actions.base.common_action import CommonAction
from migrator.core.data.model.web_hosting_shm import HostingModel
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.data.source.xml.billing import BillingReaderXml, subscription_statuses_allowed_for_migration
from migrator.core.data.source.xml.web_hosting_shm import ShmWebHostingWriterXml, SubscriptionModel
from migrator.core.language import _T
from migrator.settings import \
    TargetServiceType, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass, SERVICE_EXPORT_NOT_SUPPORTED, \
    SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH

logger = logging.getLogger(__name__)


class ExportWebHostingData(CommonAction):
    """Export SHM web hosting data from source system."""

    def get_description(self):
        return _T("EXPORT_SHM_WEB_HOSTING_DATA_ACTION")

    def get_failure_message(self, global_context):
        return _T("EXPORT_SHM_WEB_HOSTING_DATA_FAILURE")

    def run(self, global_context):
        subscriptions = {}
        exporter = global_context.source_product.get_exporter_instance(TargetServiceType.SHM)
        if exporter is SERVICE_EXPORT_NOT_SUPPORTED:
            return

        platform_reader = PlatformReaderDB()
        source_product_id = global_context.source_product.get_id()
        if source_product_id in [SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_PLESK]:
            hosting_subscriptions = platform_reader.get_subscriptions_by_services(
                SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]
            )
        elif source_product_id in [SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH]:
            hosting_subscriptions = platform_reader.get_subscriptions_by_services(
                SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[source_product_id.split('.')[1]]
            )
        else:
            hosting_subscriptions = []

        billing_reader = BillingReaderXml(global_context.session_files.get_path_to_billing_xml_file())
        hosting_subscription_ids = {sub.external_id for sub in hosting_subscriptions}
        for item in global_context.migration_list_data:
            if item.id not in subscriptions and item.id in hosting_subscription_ids:
                billing_info = billing_reader.get_subscription(item.id)
                if billing_info and billing_info.status in subscription_statuses_allowed_for_migration:
                    subscriptions[item.id] = exporter.get_subscription(item.id)
                else:
                    subscriptions[item.id] = SubscriptionModel(external_id=item.id, webspaces=[])

        database_servers = exporter.get_database_servers()
        shm_model = HostingModel(subscriptions=subscriptions.values(), database_servers=database_servers)

        xml_writer = ShmWebHostingWriterXml(global_context.session_files.get_path_to_shm_web_hosting_xml_file())
        try:
            xml_writer.write(shm_model)
        except:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(_T("FAILED_TO_EXPORT_SHM_WEB_HOSTING_INFORMATION"))
