# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.xml import EntityNotFound
from migrator.core.data.source.xml.dns import DnsReaderXml
from migrator.core.data.source.xml.mail_hosting_shm import ShmMailHostingReaderXml
from migrator.core.utils import generate_random_username, generate_random_password
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.networking.ip_utils import is_ipv4, is_ipv6, get_object_for_ip
from migrator.core.utils.registry import Registry
from migrator.odin_automation.services.shm.actions.transfer.create_webspace import CreateWebspace
from migrator.odin_automation.platform.lib.api.openapi.resource import PoaApiError
from migrator.odin_automation.platform.lib.api.aps import types as oa_types
from migrator.odin_automation.platform.lib.api.aps.ip_address import provision_ip_address, get_ip_address_uid
from migrator.odin_automation.platform.lib.api.aps.types import PADnsARecord
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import get_message_with_details
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError
from migrator.odin_automation.platform.lib.api.aps_runtime.core.aps_resource import APSResource
from migrator.odin_automation.platform.lib.api.aps.domain import get_pa_domain_records, get_dns_record_type, delete_record
from migrator.odin_automation.platform.lib.data.dns_records_importer import DnsRecordsImporter
from migrator.odin_automation.platform.lib.data.domain_importer import DomainImporter
from migrator.odin_automation.services.shm.lib import is_windows_webspace
from migrator.odin_automation.services.shm.lib.aps import types as shm_types
from migrator.odin_automation.services.shm.lib.aps.webspace import \
    register_webspace, get_webspace_by_webspace_name, get_web_configurations
from migrator.odin_automation.services.shm.lib.aps.website import \
    register_website, get_website_info_by_pa_domain_uid, register_website_forwarding, \
    register_website_forwarding_frame
from migrator.odin_automation.services.shm.lib.aps.website_alias import register_alias
from migrator.odin_automation.services.shm.lib.connections.web_server import get_shared_hosting_host_by_webspace
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.settings import TargetServiceType, SOURCE_PRODUCT_PLESK, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, \
    SourceHostingClass

logger = logging.getLogger(__name__)


class TakeoverWebspace(CreateWebspace):
    def __init__(self):
        super(TakeoverWebspace, self).__init__()
        self._dns_records_importer = DnsRecordsImporter()

    @trace
    def _create_webspace(self, subscription, global_context, webspace, subscription_external_id, keep_ip):
        poa_api = global_context.conn.target.poa_api()
        # Step 0: Check that webspace already exists
        webspace_aps = get_webspace_by_webspace_name(subscription.id, webspace.name)
        if webspace_aps:
            return webspace_aps

        # Step 1: Determine PA host id of a service node we are working with
        ipv4 = self._get_webspace_ip_address(webspace, is_ipv4)
        ipv6 = self._get_webspace_ip_address(webspace, is_ipv6)

        web_source_host = global_context.migrator._get_source_web_node(webspace.host_external_id, subscription_external_id)
        host_id = poa_api.host.get_host_id_by_ip(web_source_host.node_settings.ip)

        # Step 2: Provision IP address
        if ipv4:
            ip_pool_id = self._get_ip_pool(ipv4, poa_api, host_id)
            try:
                ipv4_address_uid = provision_ip_address(
                    subscription.id, host_id, ip=ipv4.ip, pool_id=ip_pool_id, ip_kind=ipv4.type
                )
            except APSCTransportError as e:
                ipv4_address_uid = get_ip_address_uid(ipv4.ip, subscription.id)
                if ipv4_address_uid is None:
                    raise self._get_ip_provisioning_error(e, subscription, ipv4.ip, ipv4.type)
        else:
            # For webspaces w/o ipv4 we request default shared ipv4 address for now
            ipv4_address_uid = provision_ip_address(subscription.id, host_id)

        if ipv6:
            ip_pool_id = self._get_ip_pool(ipv6, poa_api, host_id)
            try:
                ipv6_address_uid = provision_ip_address(
                    subscription.id, host_id, ip=ipv6.ip, pool_id=ip_pool_id, ip_kind=ipv6.type
                )
            except APSCTransportError as e:
                ipv6_address_uid = get_ip_address_uid(ipv6.ip, subscription.id)
                if ipv6_address_uid is None:
                    raise self._get_ip_provisioning_error(e, subscription, ipv6.ip, ipv6.type)
        else:
            ipv6_address_uid = None

        # Step 3: Register SHM webspace resource
        web_configuration_aps = get_web_configurations(subscription.id, is_windows_webspace(webspace))

        if webspace.system_user:
            _sysuser_login = webspace.system_user.name
            _sysuser_passwd = webspace.system_user.password
        else:
            _sysuser_login = generate_random_username('gen_')
            _sysuser_passwd = generate_random_password()

        return register_webspace(
            subscription.id, webspace.name, _sysuser_login, _sysuser_passwd,
            web_configuration_aps, host_id, ipv4_address_uid, ipv6_address_uid, webspace.plesk_guid,
            webspace.hosting_type
        )

    @trace
    def _create_site_common(self, site, subscription, webspace, webspace_id, migrated_websites, plesk_target_server):
        """
        :param site: migrator.core.data.model.web_hosting_shm.SiteModel
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :param webspace: migrator.core.data.model.web_hosting_shm.WebspaceModel
        :param webspace_id: str
        :param migrated_websites: dict
        :param plesk_target_server: migrator.odin_automation.platform.lib.connections.target_server.PleskTargetServer
        :return: tuple site_id, pa_domain_id
        """
        global_context = Registry.get_instance().get_context()
        site_name_utf = site.name
        site_name_idna = site.name.encode('idna')

        if not self._is_alias(site) and self._is_subdomain(site.external_id, webspace_id):
            try:
                pa_domain_id = self._domain_importer.import_domain(
                    site_name_utf,
                    target_hosting_subscription=subscription
                )['aps']['id']
            except APSCTransportError as e:
                logger.debug(_T("DEBUG_FAILED_TO_IMPORT_SUBDOMAIN").format(subdomain=site_name_utf, problem=e))
                # Unexpected error
                if e.status != 500 or self._domain_importer.get_domain_level(site_name_idna) <= 2:
                    logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                    error_message = _T("FAILED_TO_IMPORT_SUBDOMAIN").format(subdomain=site_name_utf)
                    logger.error(error_message)
                    raise MigrationError(error_message)

                # Try to solve possible conflicts in DNS records
                dns_conflict_resolve_result = self._resolve_resource_records_conflicts_for_subdomain(site_name_utf)
                pa_domain_id = self._domain_importer.import_domain(
                    site_name_utf,
                    target_hosting_subscription=subscription
                )['aps']['id']

                # Restore deleted DNS records of parent domains in the subdomain
                if bool(dns_conflict_resolve_result):
                    missed_dns_records = self._restore_deleted_conflicting_dns_records(
                        site_name_utf,
                        dns_conflict_resolve_result
                    )
                    if bool(missed_dns_records):
                        warning_message = self._compose_dns_conflicts_message(missed_dns_records, site_name_utf)
                        logger.warning(warning_message)
                        global_context.execution_wrapper.add_subscription_issue(
                            subscription.source_info.external_id,
                            Issue(Problem('lost-rrs-after-subdomain-created', Problem.NOTICE, warning_message), None)
                        )
                    else:
                        logger.debug(_T('DELETED_RRS_AFTER_CONFLICT_RESOLVING_RESTORED').format(subdomain=site_name_utf))

            # Limitation: DNS records not migrated if DNS zone in master mode
            if global_context.source_type == SOURCE_PRODUCT_PLESK:
                exporter = global_context.source_product.get_content_exporter_instance(TargetServiceType.DNS)
                zone_filter = plesk_ops.DnsOperator.FilterBySiteId([site.external_id])
                subdomain_info = exporter.get_domain_general_info(site.external_id, False, zone_filter)
                if exporter.is_dns_zone_managed(subdomain_info):
                    warning_message = _T('MASTER_SUBDOMAIN_ZONE_IN_PLESK_LIMITATION').format(subdomain=site_name_utf)
                    logger.warning(warning_message)
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription.source_info.external_id,
                        Issue(Problem('skip-rrs-restore-for-master-subdomain', Problem.NOTICE, warning_message), None)
                    )

            site_id = None

        else:
            site_id, pa_domain_id = super(TakeoverWebspace, self)._create_site_common(
                site, subscription, webspace, webspace_id, migrated_websites, plesk_target_server
            )

            dns_subscription = DomainImporter.get_dns_subscription(global_context, subscription.source_info.external_id, site_name_utf)
            if dns_subscription is None:
                dns_subscription = subscription

            if has_any_of_service_types(dns_subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
                _dns_reader = DnsReaderXml.get(global_context.session_files.get_path_to_dns_xml_file())
                zone = _dns_reader.get_zone(dns_subscription.source_info.external_id)
                self._dns_records_importer.import_records(
                    pa_domain_id, site_name_utf, zone, global_context,
                    target_hosting_subscription=subscription,
                    dns_subscription=dns_subscription,
                    link_to=site_id
                )

        return site_id, pa_domain_id

    @trace
    def _is_wildcard_subdomains_migration_supported(self):
        return True

    @trace
    def _migrate_wildcard_subdomain(self, wildcard_zone, webspace):
        """
        :param wildcard_zone: str Example: '*.example.com'
        :param webspace: migrator.core.data.model.web_hosting_shm.WebspaceModel
        :return: void
        """
        if not self._domain_importer.is_wildcard_subdomain_exist_in_oa(wildcard_zone):
            # find the nearest parent domain and create wildcard record in it
            ip_v4 = self._get_webspace_ip_address(webspace, is_ipv4).ip
            self._domain_importer.create_wildcard_subdomain(wildcard_zone, ip_v4)

    @trace
    def _restore_deleted_conflicting_dns_records(self, site_name, dns_conflict_resolve_result):
        """
        :param site_name: str Example: 'sub.example.com'
        :param dns_conflict_resolve_result: dict() data returned from _resolve_resource_records_conflicts_for_subdomain
        :return: dict() The map of parent domain to its lost DNS records
        """
        logger.debug(_T('RESTORE_DELETED_RRS_AFTER_CONFLICT_RESOLVING').format(subdomain=site_name))
        poa_api = Registry.get_instance().get_context().conn.target.poa_api()
        missed_dns_records = dict()
        for conflicting_domain, conflicting_resource_records in dns_conflict_resolve_result.iteritems():
            for conflicting_resource_record in conflicting_resource_records:
                try:
                    fqdn = conflicting_resource_record['sourceFQN']
                    dns_record_address = conflicting_resource_record['address']
                    if APSResource(conflicting_resource_record['aps']['type']) == PADnsARecord:
                        poa_api.domain.create_record_a(fqdn, '', dns_record_address)
                    else:
                        poa_api.domain.create_record_aaaa(fqdn, '', dns_record_address)
                except PoaApiError as e:
                    # If such record already exists, we're fine with it
                    if e.error_code != 3000:
                        if conflicting_domain not in missed_dns_records:
                            missed_dns_records[conflicting_domain] = []

                        missed_dns_records[conflicting_domain].append(conflicting_resource_record)

        return missed_dns_records

    @trace
    def _compose_dns_conflicts_message(self, dns_conflict_resolve_result, subdomain_name):
        """
        Compose and return a message for the subdomain which we solved DNS conflicts for

        :param dns_conflict_resolve_result: dict The map of parent domain to its deleted DNS records
        :param subdomain_name: string Example: u'sub.domain.com'
        :return: str
        """
        message_details = []
        for parent_domain, conflicting_dns_records in dns_conflict_resolve_result.iteritems():
            message_details.append(
                _T('SUBDOMAIN_CREATED_AFTER_CONFLICTS_RESOLVING_WA_DETAIL').format(
                    parent_domain=parent_domain,
                    deleted_records=u'\n\t\t'.join(
                        _T('DNS_RESOURCE_RECORD').format(
                            rr_source=conflicting_record.source,
                            rr_type=get_dns_record_type(APSResource(conflicting_record.aps.type)).upper(),
                            rr_address=conflicting_record.address
                        ) for conflicting_record in conflicting_dns_records
                    )
                )
            )

        return _T("SUBDOMAIN_CREATED_AFTER_CONFLICTS_RESOLVING_WA").format(
            subdomain=subdomain_name,
            deleted_records=u'\n\t'.join(message_details)
        )

    @trace
    def _resolve_resource_records_conflicts_for_subdomain(self, subdomain_name):
        """
        Find all possible parent DNS zones for the passed 3+ level domain and delete all the possible
        conflicting A/AAAA records. The functionality introduced by OA DNS manager's specific: it is forbidden to create
        a subdomain if there is a parent domain with A or AAAA resource record and FQDN equal to the subdomain's FQDN

        Example: for the subdomain 'x.x.x.x.com' (5 level) will be checked the following
        possible parent domains and the following DNS records will be deleted if found any:
        - 'x.x.x.com' (4 level parent): records with A/AAAA type, address 'x' and any value IPv4/IPv6 will be deleted
        - 'x.x.com' (3 level parent): records with A/AAAA type, address 'x.x' and any value IPv4/IPv6 will be deleted
        - 'x.com' (2 level parent): records with A/AAAA type, address 'x.x.x' and any value IPv4/IPv6 will be deleted

        :param subdomain_name: str Example: u'sub.example.com'
        :return: dict The map of parent domain to its deleted DNS records Example: dict(
            u'x.x.x.com' => list[migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict],
            u'x.x.com' => list[migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict],
            u'x.com' => list[migrator.odin_automation.platform.lib.api.aps_runtime.utils.AttrDict]
        )
        """
        result = dict()
        logger.debug(_T("TRYING_TO_SOLVE_DNS_RRS_CONFLICTS_FOR_SUBDOMAIN_IMPORT").format(
            subdomain=subdomain_name
        ))

        possible_parent_domains = self._domain_importer.get_all_possible_parent_domains(subdomain_name)
        logger.debug(
            _T("DEBUG_SUBDOMAIN_CREATION_WA_DETERMINED_POSSIBLE_PARENT_ZONES").format(
                domains=u','.join(possible_parent_domains)
            )
        )
        poa_api = Registry.get_instance().get_context().conn.target.poa_api()

        for possible_parent_domain in possible_parent_domains:
            possible_conflicting_record_source = self._domain_importer.get_record_of_child_domain_equal_to_parent_domain_fqdn(
                subdomain_name,
                possible_parent_domain
            )
            logger.debug(_T("DEBUG_SUBDOMAIN_CREATION_WA_DETERMINED_POSSIBLE_CONFLICTING_RRS").format(
                subdomain=subdomain_name,
                dns_records=possible_conflicting_record_source
            ))

            conflicting_records = get_pa_domain_records(
                possible_parent_domain,
                oa_types.PADnsARecord,
                possible_conflicting_record_source
            ) + get_pa_domain_records(
                possible_parent_domain,
                oa_types.PADnsAAAARecord,
                possible_conflicting_record_source
            )

            if not conflicting_records:
                logger.debug(_T('DEBUG_SUBDOMAIN_CREATION_WA_CONFLICTS_NOT_FOUND').format(
                    domain=possible_parent_domain,
                    subdomain=subdomain_name
                ))
                continue

            result[possible_parent_domain] = conflicting_records
            logger.info(_T('FOUND_CONFLICTS_FOR_SUBDOMAIN_CREATION').format(
                domain=possible_parent_domain,
                subdomain=subdomain_name,
                dns_records=u'\n\t'.join(
                    _T('DNS_RESOURCE_RECORD').format(
                        rr_source=conflicting_record.source,
                        rr_type=get_dns_record_type(APSResource(conflicting_record.aps.type)).upper(),
                        rr_address=conflicting_record.address
                    ) for conflicting_record in conflicting_records
                )
            ))
            for conflicting_record in conflicting_records:
                #  There is a bug https://jira.int.zone/browse/OA-7622.
                #  As workaround we will try to remove records with both REST and Open APIs:
                #  Manual records are not deleted completely by #1: records are deleted from APS, only,
                #  and #2 removes them from database.
                #  System records are deleted by #1 and are not deleted by #2.
                #  Thus, the combination of the both tries can delete any of manual and system records, really.

                delete_record(conflicting_record['aps']['id'])  # 1

                try:
                    poa_api.domain.delete_record(conflicting_record['recordId'])  # 2
                except Exception:
                    logger.debug("Call OpenAPI method pem.deleteDNSRecord is failed", exc_info=True)

        return result

    @trace
    def _create_website(self, subscription, webspace_id, webspace, site):
        pa_domain_uid = self._domain_importer.import_domain(site.name, target_hosting_subscription=subscription)['aps']['id']
        is_mail_domain_need = self._is_mail_domain_need(
            subscription.source_info.external_id, webspace.external_id, site.external_id
        )
        site_uid = register_website(
            subscription, pa_domain_uid, webspace_id, site.name, site.plesk_guid, is_mail_domain_need,
            is_main=site.is_main
        )

        return site_uid, pa_domain_uid

    @trace
    def _create_alias(self, subscription, site_id, alias_name, seo_redirect):
        pa_domain_id = self._domain_importer.import_domain(alias_name, target_hosting_subscription=subscription)['aps']['id']
        info = get_website_info_by_pa_domain_uid(subscription.id, pa_domain_id, shm_types.WebSiteAlias)
        if info:
            alias_id = info['aps']['id']
        else:
            alias_id = register_alias(subscription.id, pa_domain_id, site_id)

        return alias_id, pa_domain_id

    @trace
    def _create_website_forwarding(self, subscription, webspace_id, webspace, site, forwarding_destination,
                                   forwarding_http_code):
        pa_domain_uid = self._domain_importer.import_domain(site.name, target_hosting_subscription=subscription)['aps']['id']
        is_mail_domain_need = self._is_mail_domain_need(
            subscription.source_info.external_id, webspace.external_id, site.external_id
        )

        site_uid = register_website_forwarding(
            subscription, pa_domain_uid, webspace_id, site.plesk_guid, forwarding_destination, forwarding_http_code,
            is_mail_domain_need
        )

        return site_uid, pa_domain_uid

    @trace
    def _create_website_forwarding_frame(self, subscription, webspace_id, webspace, site, forwarding_destination):
        pa_domain_uid = self._domain_importer.import_domain(site.name, target_hosting_subscription=subscription)['aps']['id']
        is_mail_domain_need = self._is_mail_domain_need(
            subscription.source_info.external_id, webspace.external_id, site.external_id
        )

        site_uid = register_website_forwarding_frame(
            subscription, pa_domain_uid, webspace_id, site.plesk_guid, forwarding_destination, is_mail_domain_need
        )

        return site_uid, pa_domain_uid

    @staticmethod
    @trace
    def _get_interface_id_by_name(name, host):
        for interface in host.interfaces:
            if interface.name == name:
                return interface.id

        return host.interfaces[0].id  # temporary stub for windows
#        raise Exception("Interface {} is not found among {} at host {}".format(name, host.interfaces, host.hostname))

    @trace
    def _is_subdomain(self, site_id, webspace_uid):
        plesk_server = get_shared_hosting_host_by_webspace(webspace_uid)
        plesk_api = plesk_server.plesk_api()
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterById([site_id]),
        )
        subdomain = plesk_api.send(operator)[0]
        return not (hasattr(subdomain, 'code') and subdomain.code == 1013)

    @trace
    def _is_mail_domain_need(self, subscription_id, webspace_external_id, site_external_id):
        try:
            global_context = Registry.get_instance().get_context()
            shm_mail_hosting_reader = ShmMailHostingReaderXml(
                global_context.session_files.get_path_to_shm_mail_hosting_xml_file()
            )
            mailboxes = shm_mail_hosting_reader.get_site(
                subscription_id, webspace_external_id, site_external_id
            ).mailboxes
            return len(mailboxes) > 0
        except EntityNotFound:
            return False

    @trace
    def _get_ip_pool(self, ip, poa_api, host_id):
        ip_pool_id = None

        ip_pool = poa_api.host.get_ip_pools(ip.ip, ip.ip)
        if not ip_pool:
            host = poa_api.host.get_host(host_id)
            logger.debug(
                _T('THERE_IS_NO_APPROPRIATE_POOL_FOUND_FOR_IP_ADDRESS_IT_WILL_BE_CREATED').format(
                    ip=ip.ip, host=host.hostname
                )
            )

            try:
                pool_name = '{}, {} (created by oam)'.format(ip.ip, ip.type)
                purpose = ['COMMON.INTERNAL', 'SHARED.HOSTING']
                ip_pool_id = get_object_for_ip(
                    ip.ip,
                    lambda ip: poa_api.host.create_ip_pool(pool_name, ip.ip, ip.ip, purpose=purpose),
                    lambda ip: poa_api.host.create_ipv6_pool(pool_name, ip.ip, purpose=purpose),
                )(ip)

                poa_api.host.bind_ip_pool(
                    ip_pool_id, TakeoverWebspace._get_interface_id_by_name(ip.interface, host)
                )
            except PoaApiError as e:
                if e.extype_id not in [54, 70]:  # 54 - IntersectedIPPool, 70 - IPPoolAlreadyExists
                    raise
        else:
            ip_pool_id = ip_pool[0].id

        return ip_pool_id

    @staticmethod
    @trace
    def _get_ip_provisioning_error(e, subscription, ip, ip_type):
        if 'Cannot configure IP' in e.message:
            if 'configured as' in e.message:
                s_ip_type = e.body['details']['current_type']

                error_message = _T("CANNOT_CONFIGURE_IP_TYPE").format(
                    ip_type=ip_type.upper(), ip=ip, subscription_name=subscription.name, s_ip_type=s_ip_type.upper()
                )
            else:
                s_account = e.body['details']['current_account']
                t_account = e.body['details']['target_account']

                error_message = _T("CANNOT_CHANGE_IP_OWNERSHIP").format(
                    ip_type=ip_type.upper(), ip=ip, s_account=s_account, t_account=t_account
                )

            return MigrationError(error_message)
        elif isinstance(e.body, dict) and 'details' in e.body and 'message' in e.body:
            return MigrationError(get_message_with_details(e))
        else:
            return e
