# (C)2017 Ingram Micro Inc. all Rights Reserved
import ntpath
import os.path

from migrator.core.utils.operating_system.windows_utils import fix_path
from migrator.odin_automation.services.shm.lib import get_database_server, WebspaceMacroDefinition, SiteMacroDefinition
from migrator.odin_automation.services.shm.lib.transfer.content.web.content import WebContentListReader


class ApplicationConfig(object):
    NAME = None
    replace_map = {}

    def __init__(self, global_context, subscription, webspace, site, application):
        self.global_context = global_context
        self.subscription = subscription
        self.webspace = webspace
        self.site = site
        self.application = application

        self.path_cls = ntpath if webspace.web_target_host.is_windows() else os.path
        self.content_list_reader = WebContentListReader(global_context.session_files.get_path_to_shm_web_content_xml_file())
        self.site_content_info = self.content_list_reader.content_data_reader.get_site(
            subscription.source_info.external_id, webspace.external_id, site.external_id
        )

    @classmethod
    def _fill_std_replace_map(cls, db_host=None, db_name=None, db_user=None):
        replace_map = {
            lambda self, line: self.site_content_info.source_webroot in line: [
                lambda self, line, source_database_server, target_database_server, database:
                    line.replace(self.site_content_info.source_webroot, self.site.www_root)
            ],
            lambda self, line: self.application.source_path in line: [
                lambda self, line, source_database_server, target_database_server, database:
                line.replace(self.application.source_path, self.substitute_path(self.application.target_path))
            ]
        }

        if db_host:
            replace_map[lambda self, line: db_host in line] = [ApplicationConfig._replace_database_server]
        if db_name:
            replace_map[lambda self, line: db_name in line] = [ApplicationConfig._replace_database_name]
        if db_user:
            replace_map[lambda self, line: db_user in line] = [ApplicationConfig._replace_database_user_name]

        return replace_map

    def substitute_path(self, path):
        webspace_macro_definition = WebspaceMacroDefinition(self.webspace)
        site_macro_definition = SiteMacroDefinition(webspace_macro_definition, self.site)
        return site_macro_definition.substitute(path or '')

    def config_path(self, file):
        return self.path_cls.join(
            self.substitute_path(self.application.target_path),
            fix_path(file) if self.webspace.web_target_host.is_windows() else file
        )

    def is_config(self, contents):
        raise NotImplementedError()

    def fix_access(self):
        with self.webspace.web_target_host.runner() as runner:
            for file in self.replace_map:
                path = self.config_path(file)
                if runner.file_exists(path):
                    stat_info = runner.stat(path)
                    runner.chown(path, stat_info.user, stat_info.group)
                    runner.chmod(path, stat_info.perm)

    def fix_config(self):
        databases_by_external_id = {database.external_id: database for database in self.webspace.databases}

        if self.application.database_external_id is not None:
            database = databases_by_external_id[self.application.database_external_id]
        else:
            database = None

        self._fix_config_database(database)

    def _fix_config_database(self, database):
        if database is not None:
            source_database_server = get_database_server(database.database_server_id)
            target_database_server = self.global_context.target_panel_obj.get_target_database_server(
                self.global_context, self.webspace.web_target_host, self.webspace.name,
                self.webspace.database_map[database.name], source_database_server.dbtype
            )
        else:
            source_database_server = None
            target_database_server = None

        target_host = self.webspace.web_target_host

        with target_host.runner() as runner_target:
            for file in self.replace_map:
                config_file = self.config_path(file)
                if not runner_target.file_exists(config_file):
                    continue

                file_content = runner_target.get_file_contents(config_file)
                if not self.is_config(file_content):
                    continue

                new_file_content = []
                for line in file_content.splitlines(True):
                    for checker in self.replace_map[file]:
                        if checker(self, line):
                            for replacer in self.replace_map[file][checker]:
                                line = replacer(self, line, source_database_server, target_database_server, database)

                    new_file_content.append(line)

                bak_config_file = self.path_cls.join(
                    self.path_cls.dirname(config_file), 'bak_'+self.path_cls.basename(config_file)
                )
                if self.webspace.web_target_host.is_windows() and runner_target.file_exists(bak_config_file):
                    runner_target.sh('attrib -R %s' % bak_config_file)

                runner_target.move(config_file, bak_config_file)
                runner_target.upload_file_content(config_file, ''.join(new_file_content))

    @staticmethod
    def _replace_database_server(self, line, source_database_server, target_database_server, database):
        if not (source_database_server and target_database_server and database):
            return line

        source_host_port = '%s:%s' % (source_database_server.host, source_database_server.port)
        target_host_port = '%s:%s' % (target_database_server.host(), target_database_server.port())
        if source_host_port in line:
            return line.replace(source_host_port, target_host_port)
        else:
            return line.replace(source_database_server.host, target_database_server.host())

    @staticmethod
    def _replace_database_name(self, line, source_database_server, target_database_server, database):
        if not (source_database_server and target_database_server and database):
            return line

        return line.replace(database.name, self.webspace.database_map[database.name])

    @staticmethod
    def _replace_database_user_name(self, line, source_database_server, target_database_server, database):
        if not (source_database_server and target_database_server and database):
            return  line

        for user in self.webspace.database_users:
            line = line.replace(user.login, self.webspace.database_user_map[user.login])
        return line
