# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.model.web_hosting_shm import PhysicalHostingModel
from migrator.core.language import _T
from migrator.odin_automation.services.shm.actions.transfer.aps.drupal import DrupalApplicationConfig
from migrator.odin_automation.services.shm.actions.transfer.aps.joomla import JoomlaApplicationConfig
from migrator.odin_automation.services.shm.actions.transfer.aps.magento import MagentoApplicationConfig
from migrator.odin_automation.services.shm.actions.transfer.aps.prestashop import PrestashopApplicationConfig
from migrator.odin_automation.services.shm.actions.transfer.aps.webconfig import WindowsWebConfig
from migrator.odin_automation.services.shm.actions.transfer.aps.wordpress import WordpressApplicationConfig
from migrator.odin_automation.services.shm.actions.transfer.aps.phpbb import PHPBBApplicationConfig
from migrator.settings import MIGRATION_TYPE_FULL, SHARED_HOSTING_SERVICES

logger = logging.getLogger(__name__)

APPLICATION_MAPPING = {Config.NAME: Config for Config in [
    DrupalApplicationConfig, JoomlaApplicationConfig, MagentoApplicationConfig, PrestashopApplicationConfig,
    WordpressApplicationConfig, PHPBBApplicationConfig
]}


class FixApplicationConfig(SubscriptionAction):
    """
    The database used by the APS application might reside on an external server. For this reason, after the migration
    of content, we need to set the proper host name of the database server in the APS application configuration file.
    """

    def run(self, global_context, subscription):
        """
        :type global_context: migrator.core.session.global_context.GlobalMigrationContext
        :type subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        """

        for webspace in subscription.service.webspaces:
            for site in webspace.sites:
                if not isinstance(site.hosting, PhysicalHostingModel):
                    continue

                for application in site.hosting.applications:
                    application_class = APPLICATION_MAPPING.get(application.name.lower())
                    if application_class:
                        obj = application_class(global_context, subscription, webspace, site, application)
                        obj.fix_config()

                        # Web Application files inherits permissions from parent directory under Windows,
                        # so we should not change it
                        if not webspace.web_target_host.is_windows():
                            obj.fix_access()
                    else:
                        logger.debug('Skip configuring application {application}'.format(application=application.name))

                    # Correcting the web.config files only for Windows and only if there is a MS SQL database on the
                    # webspace.
                    if webspace.web_target_host.is_windows():
                        WindowsWebConfig(global_context, subscription, webspace, site, application).fix_config()

    def get_description(self):
        return _T("CORRECTING_DB_HOST_NAMES_OF_APS")

    def filter_subscription(self, global_context, subscription):
        return subscription.service and \
               global_context.migration_type == MIGRATION_TYPE_FULL and \
               global_context.migration_mode != MigrationMode.TEST_MIGRATION and \
               has_any_of_service_types(subscription.source_info, SHARED_HOSTING_SERVICES)

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: migrator.core.session.global_context.GlobalMigrationContext
        :type subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        """

        return _T("UNABLE_TO_CORRECT_DB_HOST_NAMES_OF_APS").format(name=subscription.name)
