# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.language import _T
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.odin_automation.services.shm.lib import is_windows_webspace
from migrator.odin_automation.services.shm.lib.aps.webspace import get_webspace_info
from migrator.settings import \
    MIGRATION_TYPE_FULL, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, \
    SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SourceHostingClass, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH

logger = logging.getLogger(__name__)


class CalculateStatistics(SubscriptionAction):
    """Calculate migrated sites statistics in SHM Plesks by running statistics calculation for each webspace"""

    def run(self, global_context, subscription):
        errors = []

        for webspace in subscription.service.webspaces:
            # Default value to make error reporting happy in case of problems retrieving real webspace name
            webspace_name = webspace.name
            try:
                webspace_name = get_webspace_info(subscription.id, webspace.id)['name']
                self._calc_webspace_statistics(webspace_name, webspace)
            except MigrationError as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                errors.append(
                    _T("ERROR_OCCURRED_WHILE_CALCULATING_WEBSPACE_STATISTICS").format(
                        webspace_name=webspace_name, error=e.message
                    )
                )

        if errors:
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem(
                    'calculate-statistics', Problem.WARNING,
                    _T("FOLLOWING_ERRORS_OCCURRED_WHILE_CALCULATING_SUBSCRIPTION_STATISTICS").format(
                        subscription_name=subscription.name,
                        errors='\n'.join(errors),
                    )
                ), None)
            )

    @staticmethod
    def _calc_webspace_statistics(webspace_name, webspace):
        target_server = webspace.web_target_host

        if is_windows_webspace(webspace):
            webspace_name = webspace_name.encode('idna')
            cmd_template = u'{plesk_dir}\\admin\\bin\\statistics --calculate-one --no-webstat --domain-name {domain}'
        else:
            cmd_template = u'{plesk_dir}/admin/bin/statistics --calculate-one --no-webstat --domain-name {domain}'

        with target_server.runner() as runner:
            runner.sh(cmd_template.format(plesk_dir=target_server.plesk_dir, domain=webspace_name))

    def filter_subscription(self, global_context, subscription):
        source_product_type = global_context.source_type
        supported_source_products = [
            SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH
        ]
        if source_product_type not in supported_source_products:
            return False

        if global_context.migration_type != MIGRATION_TYPE_FULL or \
                global_context.migration_mode == MigrationMode.TEST_MIGRATION or \
                not has_any_of_service_types(
                    subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]
                ):
            return False

        return subscription.service

    def get_description(self):
        return _T("CALCULATE_SUBSCRIPTION_STATISTICS")

    def get_failure_message(self, global_context, subscription):
        return _T("FAILED_TO_CALCULATE_SUBSCRIPTION_STATISTICS").format(subscr_name=subscription.name)

    def is_critical(self):
        return False
