# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError, MigrationNoRepeatError
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.connections.database_server import is_local_database_host
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.transfer.db.main import copy_db_content
from migrator.core.transfer.db.common import DatabaseTransferInfo
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.odin_automation.services.shm.lib import get_database_server
from migrator.odin_automation.services.shm.lib.connections.web_server import get_shared_hosting_host
from migrator.settings import SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, MIGRATION_TYPE_FULL, SourceHostingClass

logger = logging.getLogger(__name__)


class CopyDBContent(SubscriptionAction):
    @trace
    def run(self, global_context, subscription):
        subscription_id = subscription.source_info.external_id

        for webspace in subscription.service.webspaces:
            for database in webspace.databases:
                try:
                    database_server = get_database_server(database.database_server_id)
                    plesk_target_server = get_shared_hosting_host(subscription.id, webspace.name)

                    if is_local_database_host(database_server.host, [plesk_target_server.ip]):
                        source_host = PlatformReaderDB().get_source_host_by_subscription_service(subscription_id, 'db')

                        target_database_name = webspace.database_map.get(database.name)
                        if not target_database_name:
                            raise MigrationNoRepeatError("%s: %s" % (
                                self.get_failure_message(global_context, subscription),
                                _T("DATABASE_WAS_NOT_FOUND_FOR_SUBSCRIPTION_AT_TARGET_SERVER").format(
                                    database_name=database.name, subscription_name=subscription.name
                                )
                            ))

                        db_info = DatabaseTransferInfo(
                            source_database_server=global_context.migrator.get_source_database_server(source_host.ip, database_server),
                            target_database_server=global_context.target_panel_obj.get_target_database_server(
                                global_context, plesk_target_server, webspace.name, target_database_name, database_server.dbtype
                            ),
                            plesk_target_server=plesk_target_server,
                            subscription_name=subscription.name,
                            database_name=database.name,
                            target_database_name=target_database_name,
                            webspace=webspace,
                        )

                        copy_db_content(global_context, db_info)

                except (MigrationError, MigrationNoRepeatError) as e:
                    logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                    msg = _T('FAILED_COPY_DB_CONTENT_ERROR').format(
                        database_name=database.name,
                        database_type=database_server.dbtype if database_server else 'unknown type',
                        subscription_name=subscription.name,
                        error=e.message
                    )
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription_id, Issue(Problem('copy-db-content-warning', Problem.WARNING, msg), None)
                    )
                except Exception as e:
                    logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                    msg = _T('FAILED_COPY_DB_CONTENT').format(
                        database_name=database.name,
                        database_type=database_server.dbtype if database_server else 'unknown type',
                        subscription_name=subscription.name,
                    )
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription_id, Issue(Problem('copy-db-content-error', Problem.ERROR, msg), None)
                    )

    def get_description(self):
        return _T("ACTION_COPY_DB")

    def filter_subscription(self, global_context, subscription):
        return subscription.service and \
               global_context.migration_type == MIGRATION_TYPE_FULL and \
               global_context.migration_mode != MigrationMode.TEST_MIGRATION and \
               has_any_of_service_types(
                   subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]
               )

    def is_critical(self):
        return False

    def get_failure_message(self, global_context, subscription):
        return _T("FAILED_COPY_DB_CONTENT_SUBSCRIPTION_MOST").format(name=subscription.name)
