# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.actions.utils.multithreading_properties import MultithreadingProperties
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.xml import EntityNotFound
from migrator.core.data.source.xml.mail_content_shm import ShmMailContentReaderXml
from migrator.core.data.source.xml.mail_hosting_shm import ShmMailHostingReaderXml
from migrator.core.language import _T
from migrator.core.transfer.mail.imapsync import CopyMailImapsync
from migrator.core.session.config import MailContent
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.loggers.log import trace
from migrator.odin_automation.services.shm.lib.aps.mail import get_mail_domain, get_mailbox_info_by_name
from migrator.odin_automation.services.shm.lib.connections.web_server import get_shared_hosting_host
from migrator.odin_automation.services.shm.lib.data.subscription import \
    Mailbox, MailWebsite, MailWebspace, MailboxWithoutRestoringPassword
from migrator.odin_automation.services.shm.lib.transfer.content.mail.rsync import CopyMailRsync
from migrator.products.plesk.lib import get_plesk_site_id
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.settings import \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, MIGRATION_TYPE_FULL, SourceHostingClass, SKIP_MAILBOXES_CONTENT_MIGRATION

logger = logging.getLogger(__name__)


class CopyMailContent(SubscriptionAction):
    @trace
    def run(self, global_context, subscription):
        """
        :type global_context: parallels.core.global_context.GlobalMigrationContext
        :type subscription: parallels.core.migrated_subscription.MigratedSubscription
        """

        webspaces_list = subscription.service.webspaces
        if len(webspaces_list) == 0:
            return

        subscription_id = subscription.source_info.external_id
        shm_mail_content_reader = ShmMailContentReaderXml(
            global_context.session_files.get_path_to_shm_mail_content_xml_file()
        )
        shm_mail_hosting_reader = ShmMailHostingReaderXml(
            global_context.session_files.get_path_to_shm_mail_hosting_xml_file()
        )

        source_host = self._get_mail_source_host(global_context.migrator, subscription_id)
        for webspace in webspaces_list:
            target_host = self._get_mail_target_host(subscription.id, webspace.name)
            copy_content_mode = source_host.mail_settings.mode

            if target_host.ip() == source_host.ip():
                logger.info(_T("LOG_NO_NEED_TO_COPY_MAILCONTENT_BECAUSE_OF_ASSIMILATION"), webspace.name)
                return

            if copy_content_mode == MailContent.NONE:
                logger.info(_T("LOG_NO_NEED_TO_COPY_MAILCONTENT_BECAUSE_DISABLED_BY_USER"), webspace.name)
                return

            if target_host.is_windows() or source_host.is_windows():
                copy_action = CopyMailImapsync()
            else:
                copy_action = CopyMailRsync()

            sites_to_migrate = []
            for site in webspace.sites:
                mailboxes_to_migrate = []
                try:
                    mailboxes_content = shm_mail_content_reader.get_site(
                        subscription_id, webspace.external_id, site.external_id
                    ).mailboxes
                    mailboxes_hosting_data = shm_mail_hosting_reader.get_site(
                        subscription_id, webspace.external_id, site.external_id
                    ).mailboxes
                except EntityNotFound:
                    continue

                for mailbox in mailboxes_content:
                    mailbox_hosting_data = self._get_mailbox_hosting_data(mailboxes_hosting_data, mailbox.name)
                    if mailbox_hosting_data is None:
                        global_context.execution_wrapper.add_subscription_issue(
                            subscription_id,
                            Issue(Problem(
                                'copy-mailbox-content', Problem.WARNING,
                                _T("UNABLE_TO_COPY_MAILBOX_CONTENT").format(
                                    mailbox_name=mailbox.name,
                                    site_name=site.name,
                                )
                            ), None)
                        )
                        continue

                    if MailboxWithoutRestoringPassword(
                            site_name=site.name, mailbox_name=mailbox.name
                    ) in global_context.created_mailbox_without_restoring_password:
                        logger.debug(_T("MIGRATION_OF_MAIL_CONTENT_WILL_BE_SKIPPED").format(
                            mailbox_name=mailbox.name, site_name=site.name)
                        )
                        continue

                    mailboxes_to_migrate.append(
                        Mailbox(
                            name=mailbox.name,
                            password=mailbox_hosting_data.password,
                            is_password_generated=mailbox_hosting_data.is_password_generated,
                            source_path=mailbox.source_path,
                            target_path=mailbox.target_path,
                        )
                    )

                sites_to_migrate.append(
                    MailWebsite(
                        name=site.name,
                        mailboxes=mailboxes_to_migrate,
                    )
                )

            mail_webspace = MailWebspace(
                subscription_id=subscription_id,
                name=webspace.name,
                sites=sites_to_migrate,
                mail_source_host=source_host,
                mail_target_host=target_host,
            )

            copy_action.copy_mail(global_context, mail_webspace)
            if not target_host.is_windows():
                # Sync mail box disk usage by manipulating quota a bit
                self._sync_mailboxes_disk_usage_linux(subscription.id, mail_webspace)

    @trace
    def _get_mailbox_hosting_data(self, mailboxes_settings, mailbox_name):
        for mailbox in mailboxes_settings:
            if mailbox.name == mailbox_name:
                return mailbox

        return None

    @staticmethod
    @trace
    def _get_mail_source_host(migrator, subscription_id):
        return migrator._get_source_mail_node(subscription_id)

    @staticmethod
    @trace
    def _get_mail_target_host(subscription_id, webspace_name):
        return get_shared_hosting_host(subscription_id, webspace_name)

    @staticmethod
    @trace
    def _sync_mailboxes_disk_usage_linux(subscription_id, mail_webspace):
        """
        Synchronize mail boxes disk usage after content migration on SHM node.

        The idea is to manipulate mail box disk quota to force Plesk on SHM Linux node to recalculate disk usage
        which is stored in <mailbox_name>/Maildir/maildirsize file.

        :param subscription_id: existing OA subscription ID
        :param mail_webspace: SHM webspace with sites' mail boxes details for which mail content need to be copied
        :type mail_webspace: migrator.odin_automation.services.shm.lib.data.subscription.MailWebspace
        """
        _1GB = 1024 * 1024  # KB

        plesk_api = mail_webspace.mail_target_host.plesk_api()
        for site in mail_webspace.sites:
            mail_domain_uid = get_mail_domain(subscription_id, site.name)['aps']['id']
            for mailbox in site.mailboxes:
                # 1. Get mail box current disk quota
                mailbox_resource = get_mailbox_info_by_name(subscription_id, mail_domain_uid, mailbox.name)
                resource_name = {'actualDiskSpace', 'diskSpaceLimit'} & set(mailbox_resource)

                current_disk_quota = mailbox_resource[resource_name.pop()]

                # 2. Set mail box new disk quota:
                #    if quota > 1 MB: new disk quota = current disk quota - 1 MB
                #    if 0 < quota < 1 MB: new disk quota = current disk quota + 1 MB

                if current_disk_quota < 0:
                    temp_quota = _1GB
                else:
                    temp_quota = current_disk_quota * _1GB - 1

                plesk_site_id = get_plesk_site_id(site.name, plesk_api)

                plesk_api.send(
                    plesk_ops.MailOperator.Update.Set(
                        site_id=plesk_site_id,
                        name=mailbox.name,
                        aliases=None,
                        quota=temp_quota
                    )
                )

                # 3. Return back original mail box disk quota
                #    current_disk_quota in APS - GB, in Plesk - KB
                if current_disk_quota > 0:
                    current_disk_quota = current_disk_quota * _1GB

                plesk_api.send(
                    plesk_ops.MailOperator.Update.Set(
                        site_id=plesk_site_id,
                        name=mailbox.name,
                        aliases=None,
                        quota=current_disk_quota
                    )
                )

    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: parallels.core.global_context.GlobalMigrationContext
        :type subscription: parallels.core.migrated_subscription.MigratedSubscription
        """

        return subscription.service and \
            global_context.migration_type == MIGRATION_TYPE_FULL and \
            global_context.migration_mode != MigrationMode.TEST_MIGRATION and \
            has_any_of_service_types(
                subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]
            ) and \
            not SKIP_MAILBOXES_CONTENT_MIGRATION

    def get_description(self):
        return _T("ACTION_COPY_MAIL")

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: parallels.core.global_context.GlobalMigrationContext
        :type subscription: parallels.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_COPY_MAIL_CONTENT_SUBSCRIPTION_MOST").format(name=subscription.name)

    def is_critical(self):
        """If action is critical or not

        If action is critical and it failed for a subscription, migration tool
        won't run the next operations for the subscription.

        :rtype: bool
        """
        return False

    def get_multithreading_properties(self):
        return MultithreadingProperties(can_use_threads=True)
