# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import posixpath
import ntpath

from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.networking.ip_utils import is_ipv4, is_ipv6
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.web_hosting_shm import \
    AliasModel, ForwardingStdModel, PhysicalHostingModel, NoHostingModel
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.aps.account import get_account_by_subscription
from migrator.odin_automation.platform.lib.data.domain_importer import DomainImporter, DomainImporterPBA
from migrator.odin_automation.services.shm.lib import is_windows_webspace
from migrator.odin_automation.services.shm.lib import get_webspaces_list
from migrator.odin_automation.services.shm.lib.aps.notification import send_notification_migration_in_progress
from migrator.odin_automation.services.shm.lib.aps.webspace import \
    provision_webspace, get_web_configurations, provision_web_configuration, change_webspace_sysuser_password
from migrator.odin_automation.services.shm.lib.aps.website import \
    schedule_provision_website, schedule_provision_website_forwarding, schedule_provision_website_forwarding_frame, \
    get_website_www_root
from migrator.odin_automation.services.shm.lib.aps.website_alias import create_alias
from migrator.odin_automation.services.shm.lib.data.subscription import SHMWebspace, SHMWebsite, SHMWebspaces
from migrator.odin_automation.services.shm.lib.connections.web_server import get_shared_hosting_host_by_webspace
from migrator.products.plesk.lib import get_plesk_webspace_info_by_guid
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.settings import SHARED_HOSTING_SERVICES, MIGRATION_TYPE_FULL, IP_MIGRATION_MODE_SAME_IP_ADDRESS

logger = logging.getLogger(__name__)


class CreateWebspace(SubscriptionAction):

    def __init__(self):
        super(CreateWebspace, self).__init__()
        self._platform_db_reader = PlatformReaderDB()
        global_context = Registry.get_instance().get_context()
        if global_context.target_panel_obj.is_billing_module_installed():
            self._domain_importer = DomainImporterPBA(global_context)
        else:
            self._domain_importer = DomainImporter(global_context)

    def run(self, global_context, subscription):
        """
        :type global_context: migrator.core.session.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """

        subscription_id = subscription.source_info.external_id
        webspaces_list = get_webspaces_list(subscription_id)
        if len(webspaces_list) == 0:
            return

        # TODO: Fixme
        is_windows = is_windows_webspace(webspaces_list[0])
        # Panel migrator should provide web configuration resource it is not provided earlier
        self._provision_web_configuration_if_not_exist(subscription.id, is_windows)

        migrated_webspaces = []

        for webspace in webspaces_list:
            logger.info(_T("CREATING_WEBSPACE").format(name=webspace.name))
            migrated_websites = {}
            keep_ip = global_context.migration_project.ipMappingType == IP_MIGRATION_MODE_SAME_IP_ADDRESS
            webspace_aps = self._create_webspace(subscription, global_context, webspace, subscription_id, keep_ip)
            webspace_id = webspace_aps['aps']['id']
            plesk_target_server = get_shared_hosting_host_by_webspace(webspace_id)
            webspace_plesk = get_plesk_webspace_info_by_guid(plesk_target_server.plesk_api(), webspace_aps['pleskGuid'])

            webspace_www_root = None
            if isinstance(webspace_plesk.hosting, plesk_ops.SubscriptionHostingVirtual):
                if is_windows:
                    webspace_www_root = ntpath.dirname(webspace_plesk.hosting.properties.get('www_root'))
                else:
                    webspace_www_root = posixpath.dirname(webspace_plesk.hosting.properties.get('www_root'))

            # we need to separate aliases from domains because aliases have to be created after all the domains have
            # been created
            aliases = []
            domains = []
            for site in webspace.sites:
                if self._is_alias(site):
                    aliases.append(site)
                else:
                    domains.append(site)

            def _reversed_name(x):
                return x.name[::-1]

            domains.sort(key=_reversed_name)
            aliases.sort(key=_reversed_name)

            sites = {}
            for site in domains + aliases:
                if site.name.startswith('www.') and site.name[4:] in sites:
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription.source_info.external_id,
                        Issue(Problem('www-subdomain', Problem.NOTICE, _T("WWW_SUBDOMAIN_WAS_NOT_IMPORTED").format(
                            no_www_domain_name=site.name[4:], subscription_name=subscription.name
                        )), None)
                    )
                    continue

                # Wildcard subdomains are supported for takeover only and with limitations
                if self._domain_importer.is_wildcard_subdomain(site.name):
                    if self._is_wildcard_subdomains_migration_supported():
                        self._migrate_wildcard_subdomain(site.name, webspace)
                    else:
                        self._report_skipped_wildcard_subdomain(site.name, subscription, global_context)

                    continue

                sites[site.name] = self._create_site_common(
                    site, subscription, webspace, webspace_id, migrated_websites, plesk_target_server
                )

                if global_context.migration_type == MIGRATION_TYPE_FULL and self._is_site_with_physical_hosting(site):
                    www_root = get_website_www_root(subscription.id, sites[site.name][0])
                else:
                    www_root = None

                migrated_websites[site.external_id] = SHMWebsite(
                    id=sites[site.name][0],
                    name=site.name,
                    external_id=site.external_id,
                    status=site.status,
                    is_main=site.is_main,
                    plesk_guid=site.plesk_guid,
                    hosting=site.hosting,
                    www_root=www_root
                )

            migrated_webspaces.append(SHMWebspace(
                id=webspace_id,
                name=webspace.name,
                technical_name=webspace_aps['name'],
                external_id=webspace.external_id,
                type=webspace.type,
                plesk_guid=webspace.plesk_guid,
                orig_system_user=webspace.system_user.name if webspace.system_user else None,
                system_user=webspace_aps['sysUsername'],
                ip_addresses=webspace.ip_addresses,
                web_source_host=global_context.migrator._get_source_web_node(webspace.host_external_id, subscription_id),
                web_target_host=plesk_target_server,
                www_root=webspace_www_root,
                sites=migrated_websites.values(),
                databases=webspace.databases,
                database_users=webspace.database_users,
                ftp_users=webspace.ftp_users,
                web_users=webspace.web_users,
                database_map={},
                database_user_map={},
                same_ssl=not webspace.separate_ssl,
                protected_directories=webspace.protected_directories
            ))

        subscription.service = SHMWebspaces(webspaces=migrated_webspaces)

    def _create_site_common(self, site, subscription, webspace, webspace_id, migrated_websites, plesk_target_server):
        send_notification_migration_in_progress(
            account_aps_id=get_account_by_subscription(subscription.id).aps.id,
            site_name=site.name
        )

        if self._is_alias(site):
            logger.info(_T("CREATING_DOMAIN_ALIAS").format(
                alias_name=site.name, site_name=migrated_websites[site.hosting.site_id].name
            ))
            return self._create_alias(
                subscription, migrated_websites[site.hosting.site_id].id, site.name, site.hosting.seo_redirect
            )
        elif self._is_forwarding_site(site):
            logger.info(_T("CREATING_FORWARDING_DOMAIN").format(
                name=site.name, destination=site.hosting.destination
            ))
            return self._create_website_forwarding(
                subscription, webspace_id, webspace, site, site.hosting.destination,
                site.hosting.http_code
            )
        elif self._is_frame_forwarding_site(site):
            logger.info(_T("CREATING_FRAME_FORWARDING_DOMAIN").format(
                name=site.name, destination=site.hosting.destination
            ))
            return self._create_website_forwarding_frame(
                subscription, webspace_id, webspace, site, site.hosting.destination
            )
        elif self._is_no_hosting_site(site):
            logger.info(_T("CREATING_NO_HOSTING_SITE").format(name=site.name))
            site_id, pa_domain_id = self._create_website(subscription, webspace_id, webspace, site)
            plesk_target_server.plesk_api().send(plesk_ops.SiteOperator.Set(
                filter=plesk_ops.SiteOperator.FilterByName([site.name]),
                hosting=plesk_ops.SubscriptionHostingNone()
            ))
            return site_id, pa_domain_id
        else:
            logger.info(_T("CREATING_SITE").format(name=site.name))
            return self._create_website(subscription, webspace_id, webspace, site)

    def _is_wildcard_subdomains_migration_supported(self):
        return False

    def _report_skipped_wildcard_subdomain(self, wildcard_zone, subscription, global_context):
        """
        :param wildcard_zone: str Example: '*.example.com'
        :param subscription: migrator.odin_automation.platform.lib.data.subscription.PlatformSubscription
        :param global_context: migrator.core.session.global_context.GlobalMigrationContext
        :return: void
        """
        msg = _T("WILDCARD_SUBDOMAIN_WAS_NOT_MIGRATED").format(subdomain=wildcard_zone)
        logger.warning(msg)
        global_context.execution_wrapper.add_subscription_issue(
            subscription.source_info.external_id,
            Issue(Problem('wildcard-subdomain-skipped', Problem.NOTICE, msg), None)
        )

    def _migrate_wildcard_subdomain(self, wildcard_zone, webspace):
        raise NotImplementedError(_T('WILDCARD_SUBDOMAIN_MIGRATION_IS_NOT_IMPLEMENTED'))

    @staticmethod
    def _provision_web_configuration_if_not_exist(subscription_id, is_windows):
        web_configurations = get_web_configurations(subscription_id, is_windows)
        if not web_configurations:
            provision_web_configuration(subscription_id, is_windows)

    def _create_webspace(self, subscription, global_context, webspace, subscription_external_id, keep_ip):
        ipv4 = self._get_webspace_ip_address(webspace, is_ipv4)
        ipv6 = self._get_webspace_ip_address(webspace, is_ipv6)
        if ipv6 and ipv6.type == 'SHARED':
            # We do not support shared IPv6 addresses migration due to OA platform limitation
            ipv6 = None

        webspace_aps = provision_webspace(
            subscription.id,
            webspace.name,
            webspace.system_user.name if webspace.system_user else None,
            is_windows_webspace(webspace),
            ipv4.type if ipv4 else None,
            # If there are multiple WHP nodes, we cannot guarantee that all subscriptions with the same shared IP
            # address will be migrated to the same server.
            # For this reason, we do not keep the shared IPv4 address during migration.
            ipv4.ip if ipv4 and keep_ip and ipv4.type != 'SHARED' else None,
            ipv6.type if ipv6 else None,
            ipv6.ip if ipv6 and keep_ip else None
        )

        if webspace.system_user:
            change_webspace_sysuser_password(subscription.id, webspace_aps['aps']['id'], webspace.system_user.password)

        plesk_target_server = get_shared_hosting_host_by_webspace(webspace_aps['aps']['id'])
        properties = [
            (x, 'true') for x in
            ['asp', 'asp_dot_net', 'php', 'cgi', 'perl', 'python', 'fastcgi'] if x in plesk_target_server.capabilities
        ]
        logger.debug(
            "The following parameters will be set to webspace '{webspace}' at '{host}': '{properties}'".format(
                webspace=webspace_aps['name'], host=plesk_target_server.description(), properties=properties
            )
        )
        plesk_target_server.plesk_api().send(plesk_ops.SubscriptionOperator.Set(
            filter=plesk_ops.SubscriptionOperator.FilterByName([webspace_aps['name']]),
            hosting=plesk_ops.SubscriptionHostingVirtual(properties=properties, ip_addresses=None),
            limits=None
        ))

        return webspace_aps

    def _create_website_forwarding(
        self, subscription, webspace_id, webspace, site, forwarding_destination, forwarding_http_code
    ):
        pa_domain_id = self._domain_importer.import_domain(site.name, target_hosting_subscription=subscription)['aps']['id']
        site_id = schedule_provision_website_forwarding(
            subscription.id, pa_domain_id, webspace_id, site.name, is_windows_webspace(webspace),
            forwarding_destination, forwarding_http_code
        )
        return site_id, pa_domain_id

    def _create_website_forwarding_frame(self, subscription, webspace_id, webspace, site, forwarding_destination):
        pa_domain_id = self._domain_importer.import_domain(site.name, target_hosting_subscription=subscription)['aps']['id']
        site_id = schedule_provision_website_forwarding_frame(
            subscription.id, pa_domain_id, webspace_id, site.name, is_windows_webspace(webspace),
            forwarding_destination
        )
        return site_id, pa_domain_id

    def _create_website(self, subscription, webspace_id, webspace, site):
        pa_domain_id = self._domain_importer.import_domain(site.name, target_hosting_subscription=subscription)['aps']['id']
        site_id = schedule_provision_website(
            subscription.id, pa_domain_id, webspace_id, site.name, is_windows_webspace(webspace),
        )

        return site_id, pa_domain_id

    def _create_alias(self, subscription, site_id, alias_name, seo_redirect):
        pa_domain_id = self._domain_importer.import_domain(alias_name, target_hosting_subscription=subscription)['aps']['id']
        alias_id = create_alias(subscription.id, pa_domain_id, site_id, alias_name, seo_redirect)
        return alias_id, pa_domain_id

    @staticmethod
    def _is_site_with_physical_hosting(site):
        return isinstance(site.hosting, PhysicalHostingModel)

    @staticmethod
    def _is_alias(site):
        return isinstance(site.hosting, AliasModel)

    @staticmethod
    def _is_forwarding_site(site):
        return isinstance(site.hosting, ForwardingStdModel) and site.hosting.http_code

    @staticmethod
    def _is_frame_forwarding_site(site):
        return isinstance(site.hosting, ForwardingStdModel) and not site.hosting.http_code

    @staticmethod
    def _is_no_hosting_site(site):
        return isinstance(site.hosting, NoHostingModel)

    @staticmethod
    def _get_main_site(webspace):
        for site in webspace.sites:
            if site.is_main:
                return site

        return None

    @staticmethod
    def _get_webspace_ip_address(webspace, check_func):
        for ip in webspace.ip_addresses:
            if check_func(ip.ip):
                return ip

        return None

    def get_description(self):
        return _T("CREATE_WEBSPACE_IN_SHM")

    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return has_any_of_service_types(subscription.source_info, SHARED_HOSTING_SERVICES)

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_CREATE_WEBSPACE_IN_SHM").format(name=subscription.name)
