# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.data.source.xml import EntityNotFound
from migrator.core.language import _T
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.source.xml.mail_hosting_shm import ShmMailHostingReaderXml
from migrator.core.utils import urlquote
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.aps.account import get_account_user_by_subscription
from migrator.odin_automation.platform.lib.api.aps.user import get_user_by_login, get_service_user, get_user_account
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError
from migrator.odin_automation.services.shm.lib.aps.mail import get_mail_domain_uid, create_mailbox,\
    configure_antivirus, configure_autoresponder, configure_forwarding, configure_spam_filter,\
    apply_mailbox_configuration, get_mailbox_info_by_name
from migrator.odin_automation.services.shm.lib.migration.restore import import_mailbox_aliases
from migrator.settings import \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass, SKIP_MAILBOXES_ALIASES_MIGRATION


logger = logging.getLogger(__name__)


class RestoreMailHosting(SubscriptionAction):
    @trace
    def run(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """

        webspaces_list = [] if not hasattr(subscription.service, 'webspaces') else subscription.service.webspaces
        if len(webspaces_list) == 0:
            return
        self.poa_api = Registry.get_instance().get_context().conn.target.poa_api()
        subscription_id = subscription.source_info.external_id
        shm_mail_hosting_reader = ShmMailHostingReaderXml(
            global_context.session_files.get_path_to_shm_mail_hosting_xml_file()
        )
        is_windows = global_context.migrator._get_source_mail_node(subscription_id).is_windows()
        for webspace in webspaces_list:
            for site in webspace.sites:
                try:
                    mailboxes = shm_mail_hosting_reader.get_site(
                        subscription_id, webspace.external_id, site.external_id
                    ).mailboxes
                except EntityNotFound:
                    continue

                if not mailboxes:
                    continue

                mail_domain_uid = get_mail_domain_uid(subscription.id, site.name)
                for mailbox in mailboxes:
                    try:
                        user_uid = self._get_user_uid(subscription, site, mailbox, mail_domain_uid)
                    except Exception:
                        logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                        global_context = Registry.get_instance().get_context()
                        global_context.execution_wrapper.add_subscription_issue(
                            subscription_id,
                            Issue(Problem(
                                'restore-hosting', Problem.ERROR,
                                _T("UNABLE_TO_CREATE_A_SERVICE_USER").format(mailbox_name=mailbox.name)
                            ), None)
                        )
                        continue

                    self._import_mailbox(
                        subscription.account_id, subscription.id, mail_domain_uid, user_uid, mailbox, site.name,
                        is_windows, subscription_id, webspace.web_target_host
                    )

    @trace
    def _import_mailbox(
        self, account_id, subscription_id, mail_domain_uid, user_uid, mailbox, site_name, is_windows,
        subscription_external_id, plesk_target_server
    ):
        mailbox_uid = create_mailbox(
            account_id, subscription_id, subscription_external_id, mail_domain_uid, user_uid, mailbox.name, mailbox.password, site_name,
            mailbox.disk_quota
        )

        if not SKIP_MAILBOXES_ALIASES_MIGRATION:
            is_aliases_imported = import_mailbox_aliases(site_name, mailbox, plesk_target_server)
            if not is_aliases_imported:
                warn_message = _T("ERROR_OCCURRED_WHILE_RESTORING_MAILBOX_ALIASES").format(mailbox_name=mailbox.name)
                logger.warning(warn_message)
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.add_subscription_issue(
                    subscription_id,
                    Issue(Problem('unable-to-restore-mailbox-aliases', Problem.WARNING, warn_message), None)
                )

        configuration = {}
        configure_autoresponder(mailbox.autoresponder, configuration)
        configure_forwarding(mailbox.forwarding, configuration)
        configure_antivirus(mailbox.antivirus, configuration)
        if not is_windows:
            configure_spam_filter(mailbox.spam_filter, configuration)

        try:
            apply_mailbox_configuration(subscription_id, mailbox_uid, configuration)
        except APSCTransportError as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)

            if 'Antivirus is not installed' in e.message:
                error_message = _T("UNABLE_TO_CONFIGURE_MAILBOX_ANTIVIRUS_NOT_INSTALLED").format(
                    mailbox_name=mailbox.name,
                    site_name=site_name,
                    ip=plesk_target_server.ip()
                )
            else:
                error_message = _T("UNABLE_TO_CONFIGURE_MAILBOX").format(
                    mailbox_name=mailbox.name,
                    site_name=site_name,

                )

            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_external_id,
                Issue(Problem('restore-hosting', Problem.WARNING, error_message), None)
            )

        return mailbox_uid

    def _get_user_uid(self, subscription, site, mailbox, mail_domain_uid):
        mailbox_info = get_mailbox_info_by_name(subscription.id, mail_domain_uid, mailbox.name)
        if mailbox_info:
            return None

        if not mailbox.is_access_to_customer_panel:
            return get_account_user_by_subscription(subscription.id)['aps']['id']

        login = u'{mailbox_name}@{site_name}'.format(mailbox_name=mailbox.name, site_name=site.name)
        user_info = get_user_by_login(urlquote(login))
        if user_info is not None:
            user_account = get_user_account(user_info['aps']['id'])
            if user_account['id'] == subscription.account_id:
                return user_info['aps']['id']
            else:
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.add_subscription_issue(
                    subscription.source_info.external_id,
                    Issue(Problem(
                        'restore-hosting', Problem.NOTICE,
                        _T("SERVICE_USER_FOR_MAILBOX_CANNOT_BE_CREATED").format(mailbox_name=mailbox.name)
                    ), None)
                )
                return get_account_user_by_subscription(subscription.id)['aps']['id']

        user_id = self.poa_api.account.add_service_user(subscription.account_id, mailbox.name, login, mailbox.password)
        return get_service_user(user_id)['aps']['id']

    def get_description(self):
        return _T("RESTORE_MAIL_HOSTING")

    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return has_any_of_service_types(
            subscription.source_info,
            SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]
        )

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_RESTORE_MAIL_HOSTING").format(name=subscription.name)
