# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.language import _T
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.web_hosting_shm import PhysicalHostingModel
from migrator.core.utils import parse_bool
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.odin_automation.services.shm.lib.migration.restore import \
    import_hosting_settings, import_protected_directories, import_databases, import_ftp_users, import_web_users, \
    import_ssl_certificates, import_application_pools, disable_site_webspace_permissions, import_virtual_directories, \
    restore_site_webspace_permissions_from_web_template
from migrator.products.plesk.lib import get_plesk_site_id, get_plesk_webspace_id_by_website
from migrator.settings import SHARED_HOSTING_SERVICES

logger = logging.getLogger(__name__)


class RestoreHostingSettings(SubscriptionAction):
    @trace
    def run(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """

        errors = []
        is_migration_should_stopped = False
        for webspace in subscription.service.webspaces:
            plesk_target_server = webspace.web_target_host
            plesk_api = plesk_target_server.plesk_api()
            plesk_webspace_id = get_plesk_webspace_id_by_website(webspace.technical_name, plesk_api)
            import_protected_directories(
                webspace.name, webspace.protected_directories, plesk_webspace_id, plesk_target_server, errors,
                subscription.source_info.external_id, webspace.same_ssl
            )
            database_errors = []
            import_databases(subscription, webspace, plesk_webspace_id, plesk_target_server, database_errors)
            if len(database_errors) > 0:
                errors.extend(database_errors)
                is_migration_should_stopped = True

            import_ftp_users(webspace, plesk_webspace_id, plesk_target_server, errors)
            import_web_users(subscription, webspace, plesk_webspace_id, plesk_target_server, errors)
            import_ssl_certificates(webspace, plesk_target_server, errors)
            webspace_hosting_settings_imported = False
            for site in webspace.sites:
                if isinstance(site.hosting, PhysicalHostingModel):
                    if not webspace_hosting_settings_imported:
                        disable_site_webspace_permissions(site, subscription)
                        import_hosting_settings(
                            site, subscription, plesk_target_server, errors, import_settings_to_site_webspace=True
                        )
                        restore_site_webspace_permissions_from_web_template(site, subscription, plesk_target_server)
                        webspace_hosting_settings_imported = True

                    plesk_site_id = get_plesk_site_id(site.name, plesk_api)
                    import_hosting_settings(site, subscription, plesk_target_server, errors)
                    import_protected_directories(
                        site.name, site.hosting.protected_directories, plesk_site_id, plesk_target_server, errors,
                        subscription.source_info.external_id
                    )
                    import_virtual_directories(webspace, site, plesk_site_id, plesk_target_server, errors)
                    if site.hosting.application_pool and parse_bool(
                            site.hosting.hosting_properties.iis_app_pool, {False: [None]}
                    ):
                        import_application_pools(plesk_target_server, site)
                    else:
                        logger.debug(u"There is no Application Pool for site {site}".format(site=site.name))

        if errors:
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem(
                    'restore-hosting', Problem.WARNING,
                    _T("FOLLOWING_ERRORS_OCCURRED_WHILE_RESTORING").format(
                        subscription_name=subscription.name,
                        #  remove dublicates and save order
                        errors='\n'.join([e for i, e in enumerate(errors) if e not in errors[:i]]),
                    )
                ), None)
            )

        if is_migration_should_stopped:
            raise MigrationError(_T("WEB_HOSTING_SETTINGS_NOT_BE_RESTORED"))

    def get_description(self):
        return _T("RESTORE_HOSTING_SETTINGS")

    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return subscription.service and has_any_of_service_types(subscription.source_info, SHARED_HOSTING_SERVICES)

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_RESTORE_HOSTING_SETTINGS").format(name=subscription.name)
