# (C)2017 Ingram Micro Inc. all Rights Reserved
from itertools import chain

from migrator.core.language import _T
from migrator.core.actions.base.subscription_action import SubscriptionAction
from migrator.core.data.model.subscription import has_any_of_service_types
from migrator.core.data.model.migration_project import MigrationMode
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.odin_automation.platform.lib.api.aps.domain import get_pa_domain_by_name
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import APS2
from migrator.odin_automation.platform.lib.api.aps_runtime.core.aps_resource import APSResource
from migrator.odin_automation.platform.lib.api.aps.subscription import get_subscription
from migrator.odin_automation.platform.lib.api.aps.types import \
    PADnsRecord, PADnsARecord, PADnsAAAARecord, PADnsCNAMERecord
from migrator.odin_automation.services.shm.lib.aps.webspace import get_webspace_ips
from migrator.odin_automation.services.shm.lib.openapi.domain import switch_dns_records
from migrator.settings import SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass


class SwitchDomainRecords(SubscriptionAction):
    def run(self, global_context, subscription):
        aps = APS2.get_connector()

        pa_subscription = get_subscription(subscription.id)
        filter_to_enable = 'linkedWith(%s),implementing(%s),select(zone),eq(RRState,inactive)'
        records_to_enable = aps.get_resources(filter_str=filter_to_enable % (pa_subscription['aps']['id'], PADnsRecord))
        records_to_disable = []

        filter_to_disable = 'not(linkedWith(%s)),implementing(%s),select(zone),eq(RRState,active),eq(source,%s),eq(zone.aps.id,%s)'
        for r in records_to_enable:
            for aps_type in set([APSResource(r['aps']['type']), PADnsARecord, PADnsAAAARecord, PADnsCNAMERecord]):
                old_record = aps.get_resources(
                    filter_str=filter_to_disable %
                    (pa_subscription['aps']['id'], aps_type, r['source'] or 'empty()', r['zone']['aps']['id'])
                )
                records_to_disable.extend(old_record)

        switch_dns_records(global_context, records_to_disable, records_to_enable)

        external_domains = {}
        for webspace in subscription.service.webspaces:
            for site in webspace.sites:
                domain = get_pa_domain_by_name(site.name, subscription.id)
                if 'nameServers' not in domain:  # 'hosting' is more correct, but it is absent at subdomains
                    ips = get_webspace_ips(subscription.id, webspace.id)
                    external_domains[domain.name] = [ip['address'] for ip in ips]

        for ext in external_domains:
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem(
                    'external-dns-hosting', Problem.NOTICE,
                    _T("DOMAIN_USES_EXTERNAL_DNS_SWITCH_RECORDS_MANUALLY").format(
                        domain=ext, ip=', '.join(external_domains[ext])
                    )
                    ), None
                )
            )

    def get_description(self):
        return _T("SWITCH_DNS_RECORDS_TO_SHM")

    def filter_subscription(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return subscription.service and \
               global_context.migration_mode == MigrationMode.PRODUCTION and \
               has_any_of_service_types(
                   subscription.source_info,
                   chain(
                       SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.WSH],
                       SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.WHL],
                       SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.LSH],
                   )
               )

    def get_failure_message(self, global_context, subscription):
        """
        :type global_context: migrator.core.global_context.GlobalMigrationContext
        :type subscription: migrator.core.migrated_subscription.MigratedSubscription
        """
        return _T("FAILED_SWITCH_DNS_RECORDS_TO_SHM").format(name=subscription.name)
