# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.checks.base import CheckAbstract
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.filetypes.config_utils import ConfigHostSection
from migrator.core.utils.networking.ip_port import \
    get_unavailable_ports_for_host, PORTS_AVAILABILITY_STATUS_SUCCESS
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.checks.checks import Checks
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import get_aps_connector_with_admin_token
from migrator.odin_automation.services.shm.lib.aps.api import get_shm_application_nodes, is_shm_node_license_valid
from migrator.odin_automation.services.shm.lib.aps.types import WebApplication
from migrator.odin_automation.services.shm.lib.connections.web_server import get_shared_hosting_host_by_ip
from migrator.products.plesk.checks.lib.license import get_plesk_license_status, PLESK_LICENSE_STATUS_VALID
from migrator.settings import \
    MIGRATION_TYPE_FULL, MIGRATION_TYPE_TAKEOVER, PANEL_TYPE_PLESK, OS_LINUX, TargetServiceType, SOURCE_PRODUCT_PLESK, \
    SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, OS_WINDOWS, SSH_DEFAULT_PORT, SMB_DEFAULT_PORT, \
    PLESK_PANEL_DEFAULT_PORT, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH

logger = logging.getLogger(__name__)


class CheckHostsSHM(CheckAbstract):
    SUPPORTED_SOURCE_PRODUCTS = [
        SOURCE_PRODUCT_PLESK, SOURCE_PRODUCT_OBAS, SOURCE_PRODUCT_OA_WSH, SOURCE_PRODUCT_OA_WHL, SOURCE_PRODUCT_OA_LSH
    ]
    SUPPORTED_TARGET_SERVICE_CLASSES = [TargetServiceType.SHM]
    SUPPORTED_MIGRATION_TYPES_ASPECT_MODULE_INSTALLED = [MIGRATION_TYPE_TAKEOVER, MIGRATION_TYPE_FULL]
    SUPPORTED_MIGRATION_TYPES_ASPECT_NODE_READY = [MIGRATION_TYPE_FULL]
    SHM_NODE_COMMUNICATION_PORTS_LINUX = [SSH_DEFAULT_PORT, PLESK_PANEL_DEFAULT_PORT]
    SHM_NODE_COMMUNICATION_PORTS_WINDOWS = [SMB_DEFAULT_PORT, PLESK_PANEL_DEFAULT_PORT]

    def __init__(self):
        super(CheckHostsSHM, self).__init__(Checks.CHECK_HOSTS_SHM, CheckAbstract.PHASE_NO_PHASE)
        self.config = Registry.get_instance().get_context().config

    def get_full_description(self):
        return _T('CHECK_FULL_DESCRIPTION_HOSTS_SHM')

    def _do_check(self):
        try:
            migration_type = self.config.get('GLOBAL', 'migration-type').lower()
            # Error: no SHM module
            if migration_type in self.SUPPORTED_MIGRATION_TYPES_ASPECT_MODULE_INSTALLED:
                aps = get_aps_connector_with_admin_token()
                if aps.get_resources_count(WebApplication) != 1:
                    return self.STATUS_ERROR, _T('CHECK_SHM_HOSTS_ERROR_APP_NOT_FOUND')

            # Try to find suitable hosting node: OS compatible, available, with valid license and "ready" status
            if migration_type in self.SUPPORTED_MIGRATION_TYPES_ASPECT_NODE_READY:
                # Detect required OS types and their availability
                # Example required_operational_systems = {'linux': False, 'windows': False}
                if self.is_remote_source_system():
                    required_operational_systems = self._get_required_operational_systems_for_remote_source_system()
                else:
                    required_operational_systems = self._get_required_operational_systems_for_local_source_system()

                # Detect suitable SHM nodes: host is available, license valid and has "ready" status
                shm_nodes = get_shm_application_nodes()
                for shm_node in shm_nodes:
                    shm_node_os = shm_node.os.lower()
                    if shm_node_os not in required_operational_systems \
                            or required_operational_systems[shm_node_os] is True:
                        continue

                    ports_to_check = self.SHM_NODE_COMMUNICATION_PORTS_LINUX if shm_node_os == OS_LINUX \
                        else self.SHM_NODE_COMMUNICATION_PORTS_WINDOWS
                    shm_node_ip = shm_node['hostname']
                    ports_availability_result = get_unavailable_ports_for_host(shm_node_ip, ports_to_check)
                    if ports_availability_result['status'] == PORTS_AVAILABILITY_STATUS_SUCCESS:
                        plesk_object = get_shared_hosting_host_by_ip(shm_node_ip)
                        if is_shm_node_license_valid(shm_node['uid']) \
                                and get_plesk_license_status(plesk_object) == PLESK_LICENSE_STATUS_VALID:
                            if shm_node['status'] == 'ready':
                                required_operational_systems[shm_node_os] = True
                            else:
                                logger.debug(
                                    _T('CHECK_SHM_HOSTS_DEBUG_UNSUITABLE_NODE_BY_STATUS').format(ip=shm_node_ip)
                                )
                        else:
                            logger.debug(
                                _T('CHECK_SHM_HOSTS_DEBUG_UNSUITABLE_NODE_BY_LICENSE').format(ip=shm_node_ip)
                            )

                for required_operational_system in required_operational_systems:
                    if required_operational_systems[required_operational_system] is False:
                        if required_operational_system == OS_LINUX:
                            node_ports = self.SHM_NODE_COMMUNICATION_PORTS_LINUX
                        else:
                            node_ports = self.SHM_NODE_COMMUNICATION_PORTS_WINDOWS

                        return self.STATUS_ERROR, _T('CHECK_SHM_HOSTS_ERROR_NODE_NOT_FOUND').format(
                            node_os=required_operational_system.capitalize(),
                            node_ports=', '.join(str(port) for port in node_ports)
                        )

            # Success
            return self.STATUS_SUCCESS, _T('CHECK_SHM_HOSTS_SUCCESS')
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return self.STATUS_ERROR, _T('CHECK_SHM_HOSTS_UNEXPECTED')

    def _get_required_operational_systems_for_remote_source_system(self):
        """
        Get OSes required for migration

        :return: dict() The keys show required OSes, the boolean values show if a valid WHP node for a certain OS was
            found in OA. Examples: {'linux': False, 'windows': False} {'linux': False} {'windows': False}
        """
        required_operational_systems = dict()
        source_hosts_to_check = PlatformReaderDB().get_migration_required_source_hosts(PANEL_TYPE_PLESK)
        for plesk_to_check in source_hosts_to_check:
            plesk_config_section = ConfigHostSection(self.config, plesk_to_check.ip)
            plesk_os = plesk_config_section.get('os').lower()
            if plesk_os not in required_operational_systems:
                required_operational_systems[plesk_os] = False

        return required_operational_systems

    def _get_required_operational_systems_for_local_source_system(self):
        return {OS_WINDOWS: False}

    def is_remote_source_system(self):
        return self.source_type not in [SOURCE_PRODUCT_OA_WSH]
