# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.utils import FunctionResultWaiter, urlquote, generate_random_password
from migrator.core.utils.registry import Registry
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import \
    APS2, get_error_message, get_aps_connector, get_aps_connector_with_admin_token, get_app_instance_uid
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError
from migrator.odin_automation.platform.lib.api.aps.domain import get_pa_domain_by_name
from migrator.odin_automation.platform.lib.api.aps.subscription import get_service_template_info
from migrator.odin_automation.platform.lib.api.aps_runtime.utils import subscribe
from migrator.odin_automation.platform.lib.task_manager import wait_for_all_scheduled_tasks_on_resource_completed
from migrator.odin_automation.platform.lib.db import get_application_instances_ids
from migrator.odin_automation.services.shm.lib.aps.types import aps_operations
from migrator.odin_automation.services.shm.lib.data.subscription import MailboxWithoutRestoringPassword
from migrator.settings import SKIP_RESTORE_MAILBOX_PASSWORD

import migrator.odin_automation.platform.lib.api.aps.subscription as subscription
import migrator.odin_automation.services.shm.lib.aps.types as types

logger = logging.getLogger(__name__)


@trace
def get_mail_domain_uid(subscription_id, site_name):
    plan_name = get_service_template_info(subscription_id)['name']
    function = FunctionResultWaiter(
        get_mail_domain,
        MigrationError(_T("UNABLE_TO_CREATE_MAIL_DOMAIN").format(site_name=site_name, plan_name=plan_name)),
        check_function=lambda x: x and x['aps']['status'] == 'aps:ready'
    )

    return function.run(subscription_id, site_name)['aps']['id']


@trace
def get_mail_domain(subscription_id, site_name):
    aps = get_aps_connector(subscription_id)
    pa_domain_uid = get_pa_domain_by_name(site_name, subscription_id)['aps']['id']
    return aps.get_first_resource(filter_str='implementing(%s),eq(domain.aps.id,%s)' % (
        types.MailDomain, pa_domain_uid
    ))


@trace
def get_mailbox_info(subscription_id, mailbox_uid):
    aps = get_aps_connector(subscription_id)
    return aps.get_resource(mailbox_uid)


@trace
def create_mailbox(
    account_id, subscription_id, subscription_external_id, mail_domain_uid, user_uid, mailbox_name, password, site_name,
    diskspace_quota
):
    mailbox_aps = get_mailbox_info_by_name(subscription_id, mail_domain_uid, mailbox_name)
    if not mailbox_aps:
        try:
            aps = APS2.get_connector(account_id, None, subscription_id)
            resource = {
                'aps': {
                    'type': types.Mailbox
                },
                'mailDomain': {
                    'aps': {
                        'id': mail_domain_uid
                    }
                },
                'user': {
                    'aps': {
                        'id': user_uid
                    }
                },
                'name': mailbox_name,
                'password': generate_random_password(),
            }

            aps.post_resource('', '', resource, hidelist_request=['$.password'])
            mailbox_uid = _wait_for_provision_mailbox(subscription_id, mail_domain_uid, mailbox_name, site_name)
        except APSCTransportError as e:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            error_message = get_error_message(e)
            if "Creation of resource with type '{mailbox_type}' is not allowed".format(
                    mailbox_type=types.Mailbox
            ) in error_message:
                plan_name = get_service_template_info(subscription_id)['name']
                raise MigrationError(_T("UNABLE_TO_CREATE_MAIL_ACCOUNT").format(
                    mailbox_name=mailbox_name, site_name=site_name, plan_name=plan_name
                ))
            else:
                raise MigrationError(error_message)
    else:
        mailbox_uid = mailbox_aps['aps']['id']

    try:
        change_mailbox_password(subscription_id, mailbox_uid, password)
    except Exception as e:
        logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
        if SKIP_RESTORE_MAILBOX_PASSWORD:
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_external_id, Issue(Problem(
                    'cannot-restore-mailbox-password', Problem.NOTICE,
                    _T("UNABLE_TO_RESTORE_MAILBOX_PASSWORD").format(mailbox_name=mailbox_name, site_name=site_name)
                ), None)
            )
            global_context.created_mailbox_without_restoring_password.append(MailboxWithoutRestoringPassword(
                site_name=site_name, mailbox_name=mailbox_name)
            )
        else:
            raise e

    update_mailbox_diskspace_quota(
        subscription_id, subscription_external_id, mailbox_uid, mail_domain_uid, diskspace_quota, mailbox_name,
        site_name
    )

    return mailbox_uid


@trace
def register_mailbox(subscription_id, mail_domain_uid, user_uid, mailbox_name, password, site_name, diskspace_quota):
    """
    Register MailBox resource on APS bus w/o provisioning
    """

    mailbox_aps = get_mailbox_info_by_name(subscription_id, mail_domain_uid, mailbox_name)
    if not mailbox_aps:
        pa_subscription_uid = subscription.get_subscription(subscription_id)['aps']['id']
        shm_application_uid = get_app_instance_uid(types.SHMApplication)

        resource = {
            'aps': {
                'type': types.Mailbox
            },
            'mailDomain': {
                'aps': {
                    'id': mail_domain_uid
                }
            },
            'user': {
                'aps': {
                    'id': user_uid
                }
            },
            'primaryEmail': u'{mailbox_name}@{site_name}'.format(mailbox_name=mailbox_name, site_name=site_name),
            'name': mailbox_name,
            'password': password,
        }

        aps = get_aps_connector_with_admin_token()
        aps.setAdditionalHeaders({'APS-Subscription-ID': pa_subscription_uid})
        mailbox_uid = aps.register_resource(
            shm_application_uid, 'MailBox', resource, hidelist_request=['$.password']
        )['aps']['id']
    else:
        mailbox_uid = mailbox_aps['aps']['id']

    shm_application_instance_id = get_application_instances_ids(types.SHMApplication)[0][0]
    aps = APS2.get_connector(app_instance_id=shm_application_instance_id)
    updated_resource = {
        'aps': {'id': mailbox_uid},
        'configuration': get_mailbox_configuration(shm_application_instance_id, mailbox_uid)
    }

    mail_domain = aps.get_resource(mail_domain_uid)
    subscribe(
        aps, mailbox_uid, "http://aps-standard.org/core/events/changed", "eventConfigurationChanged",
        source_uid=mail_domain['service']['aps']['id']
    )

    aps.update_resource(resource=updated_resource, service='MailBox', hidelist_response=["$.password"])

    return mailbox_uid


@trace
def register_mail_domain(website_uid, pa_domain_uid, mail_service_uid, subscription_aps_id):
    shm_application_uid = get_app_instance_uid(types.SHMApplication)

    resource = {
        'aps': {'type': types.MailDomain},
        'isEnabled': True,
        'serviceDescription': 'Email Domain',
        'serviceName': 'SHM Mail Service',
        'domain': {'aps': {'id': pa_domain_uid}},
        'webSite': {'aps': {'id': website_uid}},
        'service': {'aps': {'id': mail_service_uid}}
    }

    aps = get_aps_connector_with_admin_token()
    aps.setAdditionalHeaders({'APS-Subscription-ID': subscription_aps_id})

    mail_domain_aps_id = aps.register_resource(shm_application_uid, 'MailDomain', resource)['aps']['id']

    subscribe(
        aps, mail_domain_aps_id, "http://aps-standard.org/core/events/changed", "eventConfigurationChanged",
        source_uid=mail_service_uid
    )

    return mail_domain_aps_id


@trace
def get_mail_domain_by_website_and_pa_domain_uids(website_uid, pa_domain_uid):
    aps = get_aps_connector_with_admin_token()
    return aps.get_first_resource(
        filter_str='implementing(%s),eq(domain.aps.id,%s),eq(webSite.aps.id,%s)' % (
            types.MailDomain, pa_domain_uid, website_uid)
    )


@trace
def change_mailbox_password(subscription_id, mailbox_uid, password):
    aps = get_aps_connector(subscription_id)
    aps.put_resource(
        mailbox_uid, aps_operations['changePassword'], {'newPassword': password}, hidelist_request=['$.newPassword']
    )
    return wait_for_all_scheduled_tasks_on_resource_completed(mailbox_uid)


@trace
def get_mailbox_info_by_name(subscription_id, mail_domain_uid, mailbox_name):
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(
        resource_path='%s/mailBoxes' % mail_domain_uid,
        filter_str='eq(name,%s)' % urlquote(mailbox_name)
    )


@trace
def _wait_for_provision_mailbox(subscription_id, mail_domain_uid, mailbox_name, site_name):
    function = FunctionResultWaiter(
        get_mailbox_info_by_name,
        MigrationError(_T('UNABLE_TO_CREATE_MAILBOX_FOR_SITE').format(mailbox_name=mailbox_name, site_name=site_name)),
        check_function=lambda x: x and x['aps']['status'] == 'aps:ready'
    )

    return function.run(subscription_id, mail_domain_uid, mailbox_name)['aps']['id']


@trace
def _wait_for_mailbox_update(subscription_id, mailbox_uid, mailbox_name, site_name):
    function = FunctionResultWaiter(
        get_mailbox_info,
        MigrationError(_T('UNABLE_TO_CONFIGURE_MAILBOX').format(mailbox_name=mailbox_name, site_name=site_name)),
        check_function=lambda x: x and x['aps']['status'] == 'aps:ready'
    )

    return function.run(subscription_id, mailbox_uid)['aps']['id']


@trace
def configure_autoresponder(autoresponder, configuration):
    configuration['autoresponder'] = {
        'enabled': False,
        'subject': '',
        'message': '',
        'forwardTo': '',
        'endDate': '',
    }
    if autoresponder is not None:
        configuration['autoresponder']['enabled'] = autoresponder.enabled
        if autoresponder.subject is not None:
            configuration['autoresponder']['subject'] = autoresponder.subject

        if autoresponder.message is not None:
            configuration['autoresponder']['message'] = autoresponder.message

        if autoresponder.forward_to is not None:
            configuration['autoresponder']['forwardTo'] = autoresponder.forward_to

        if autoresponder.end_date is not None:
            configuration['autoresponder']['endDate'] = autoresponder.end_date


@trace
def configure_antivirus(antivirus, configuration):
    configuration['antivirus'] = antivirus


@trace
def configure_spam_filter(spam_filter, configuration):
    configuration['spamFilter'] = {
        'isEnabled': False,
        'sensitivity': 7,
        'action': 'move',
        'rewriteHeader': '',
    }
    if spam_filter is not None:
        configuration['spamFilter']['isEnabled'] = True
        configuration['spamFilter']['patterns'] = {
            'white': [address for address in spam_filter.whitelist],
            'black': [address for address in spam_filter.blacklist],
        }

        if spam_filter.sensitivity is not None:
            configuration['spamFilter']['sensitivity'] = spam_filter.sensitivity

        if spam_filter.action is not None:
            configuration['spamFilter']['action'] = spam_filter.action

        if spam_filter.rewrite_header is not None:
            configuration['spamFilter']['rewriteHeader'] = spam_filter.rewrite_header


@trace
def configure_forwarding(forwarding, configuration):
    configuration['forwarding'] = {
        'isEnabled': False,
    }
    if forwarding is not None:
        configuration['forwarding']['isEnabled'] = True
        configuration['forwarding']['list'] = [address for address in forwarding.addresses]


@trace
def apply_mailbox_configuration(subscription_id, mailbox_uid, configuration):
    aps = get_aps_connector(subscription_id)
    aps.put_resource(mailbox_uid, aps_operations['configuration'], configuration)
    wait_for_all_scheduled_tasks_on_resource_completed(mailbox_uid)


@trace
def get_mailbox_configuration(shm_application_instance_id, mailbox_uid):
    aps = APS2.get_connector(app_instance_id=shm_application_instance_id)
    return aps.get_resources(resource_path='%s/%s' % (mailbox_uid, aps_operations['configuration']))


@trace
def update_mailbox_diskspace_quota(
        subscription_id, subscription_external_id, mailbox_uid, mail_domain_uid, quota, mailbox_name, site_name
):
    aps = get_aps_connector(subscription_id)

    resource_name = {'actualDiskSpace', 'diskSpaceLimit'} & set(types.Mailbox.properties)

    if resource_name:
        resource_name = resource_name.pop()
        # Convert unlimited quota (-1) to maximum allowed value for the subscription.
        actual_disk_space = _get_mailbox_diskspace_limit(subscription_id, mail_domain_uid) if quota < 0 else quota

        # OA 8.4: DEBUG [extension/shm] Incoming request (POST /modules/shm/public/whp-api.php/mail-box/update/be1d418b-6723-4b20-ac9d-756f0dd5a4e4): {"version":"1.1","request":{"mailbox":{"resourceId":"17049c46-a01c-4d81-b745-8bb28786db2a","name":"ani"},"aliases":[],"diskSpaceLimit":-1}}
        # OA 7.4: DEBUG [extension/shm] Incoming request (POST /modules/shm/public/whp-api.php/mail-box/update/f112f1aa-e9dd-430d-87c2-ef261ff042c6): {"version":"1.1","request":{"mailbox":{"resourceId":"965b18bd-e291-41bf-8478-b56f4657249e","name":"ani"},"diskSpaceLimit":-1048576}}
        # WHP in OA 7.4 pass -1 (that is unlimited in OA) as diskSpaceLimit":-1048576 that is incorrect, thus task 'Configuring "MailBox" for APS application WebHosting Plesk' in OA failed with error: 'Unable to set the mailbox size limit for the domain: Icorrect mailbox size specified.'
        # So, we cannot set mailbox size to unlimited in OA 7.4
        if actual_disk_space < 0 and resource_name == 'actualDiskSpace':
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_external_id,
                Issue(Problem(
                    'update-diskspace-quota-is-not-supported', Problem.NOTICE,
                    _T("DISK_SPACE_QUOTA_CANNOT_BE_SET_TO_UNLIMITED").format(mailbox_name=mailbox_name),
                ), None)
            )
        else:
            aps.put_resource(mailbox_uid, '', {resource_name: actual_disk_space})
            _wait_for_mailbox_update(subscription_id, mailbox_uid, mailbox_name, site_name)
            wait_for_all_scheduled_tasks_on_resource_completed(mailbox_uid)


@trace
def _get_mailbox_diskspace_limit(subscription_id, mail_domain_uid):
    aps = get_aps_connector(subscription_id)
    return aps.get_resource(mail_domain_uid)['mailboxDiskSpace']['limit']
