# (C)2017 Ingram Micro Inc. all Rights Reserved

from migrator.core import MigrationNoRepeatError, InternalError
from migrator.core.language import _T
from migrator.core.utils import FunctionResultWaiter
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.registry import Registry

from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import \
    get_aps_connector, get_aps_connector_with_admin_token, get_app_instance_uid, get_error_message, APS2
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError
from migrator.odin_automation.platform.lib.api.aps.domain import get_pa_domain_by_name
from migrator.odin_automation.platform.lib.api.aps.host import get_host_capacities
from migrator.odin_automation.platform.lib.api.aps.types import PADnsARecord, PADnsAAAARecord, PADnsCNAMERecord
from migrator.odin_automation.platform.lib.api.aps_runtime.utils import subscribe
from migrator.odin_automation.platform.lib.db import get_application_instances_ids
from migrator.odin_automation.platform.lib.task_manager import wait_for_all_scheduled_tasks_on_resource_completed
from migrator.odin_automation.services.shm.lib.aps.mail import register_mail_domain, \
    get_mail_domain_by_website_and_pa_domain_uids
from migrator.odin_automation.services.shm.lib.aps.subscription import get_shm_mail_service
from migrator.odin_automation.services.shm.lib.aps.types import aps_operations
from migrator.odin_automation.services.shm.lib.openapi.domain import switch_dns_records

import migrator.odin_automation.platform.lib.api.aps.subscription as subscription

import api
import types
import host
import webspace
import logging

from contextlib import contextmanager

logger = logging.getLogger(__name__)

PROVISION_STATUS_SKIP_MAILDOMAIN_PROVISION = 'takeover:skip_maildomain_provision'
PROVISION_STATUS_MAILDOMAIN_PROVISION_SKIPPED = 'takeover:maildomain_provision_skipped'
PROVISION_STATUS_READY = ''


@contextmanager
@trace
def dns_record_switcher(subscription_id, pa_domain_uid):
    global_context = Registry.get_instance().get_context()
    poa_api = global_context.conn.target.poa_api()

    aps = APS2.get_connector()
    pa_subscription = subscription.get_subscription(subscription_id)
    records = aps.get_resources(
        resource_path='%s/records' % pa_domain_uid,
        filter_str='not(linkedWith(%s)),eq(RRState,active),or(implementing(%s),implementing(%s),implementing(%s))' %
                   (pa_subscription['aps']['id'], PADnsARecord, PADnsAAAARecord, PADnsCNAMERecord)
    )

    _conflicting_sources = ['', 'www', 'ftp', 'www.', 'ftp.']
    conflicting_records = [r for r in records if r['source'] in _conflicting_sources]
    logger.debug(
        'Conflicting records detected and will be temporary disabled: %s ' % [r['sourceFQN'] for r in conflicting_records]
    )

    try:
        for r in conflicting_records:
            poa_api.domain.disable_record(r['recordId'])
        yield
    finally:
        subscription_records = aps.get_resources(
            resource_path='%s/records' % pa_domain_uid,
            filter_str='linkedWith(%s),eq(RRState,active)' % pa_subscription['aps']['id']
        )

        if conflicting_records:
            switch_dns_records(global_context, subscription_records, conflicting_records)


@trace
def _schedule_provision_website(subscription_id, site_name, is_windows, args):
    """
    Create a website and check that it was created successfully.
    If a site with this name already exists, return the site ID for this site.
    """

    service_type = types.WebspaceIIS if is_windows else types.WebspaceApache
    service_aps_id = api.get_service_aps_id(subscription_id, service_type)

    site_type = args['siteType']
    pa_domain_uid = args['domainId']

    try:
        aps = get_aps_connector(subscription_id)
        with dns_record_switcher(subscription_id, pa_domain_uid):
            aps.post_resource(service_aps_id, aps_operations['provisionWebsite'], args)

            # TODO: This is a temporary solution. Provision website should return id in response.
            return get_website_uid_by_pa_domain_uid(subscription_id, pa_domain_uid, site_name.lower(), site_type)
    except APSCTransportError as e:
        raise MigrationNoRepeatError(get_error_message(e))


@trace
def schedule_provision_website(subscription_id, pa_domain_uid, webspace_uid, site_name, is_windows):
    args = {
        'domainId': pa_domain_uid,
        'webSpace': {'id': webspace_uid},
        'siteType': types.SiteWeb
    }
    return _schedule_provision_website(subscription_id, site_name, is_windows, args)


@trace
def _get_destination(destination):
    if destination.lower().startswith('http://') or destination.lower().startswith('https://'):
        return destination
    else:
        return 'http://%s' % destination


@trace
def schedule_provision_website_forwarding(
    subscription_id, pa_domain_uid, webspace_uid, site_name, is_windows, destination, http_code
):
    args = {
        'domainId': pa_domain_uid,
        'webSpace': {'id': webspace_uid},
        'siteType': types.SiteWebForwarding,
        'forwardingDestination': _get_destination(destination),
        'forwardingHttpCode': http_code
    }

    return _schedule_provision_website(subscription_id, site_name, is_windows, args)


@trace
def schedule_provision_website_forwarding_frame(
    subscription_id, pa_domain_uid, webspace_uid, site_name, is_windows, destination
):
    args = {
        'domainId': pa_domain_uid,
        'webSpace': {'id': webspace_uid},
        'siteType': types.SiteWebForwardingFrm,
        'forwardingDestination': _get_destination(destination),
    }

    return _schedule_provision_website(subscription_id, site_name, is_windows, args)


@trace
def register_website(
        target_hosting_subscription, pa_domain_uid, webspace_uid, site_name, plesk_guid, is_mail_domain_need,
        is_main=False
):
    """
    Register SHM Website resource on APS bus w/o provisioning
    """
    resource = {
        'path': site_name,
        'main': is_main,
    }

    return _register_site(
        target_hosting_subscription, pa_domain_uid, webspace_uid, plesk_guid, types.SiteWeb, resource,
        is_mail_domain_need
    )


@trace
def register_website_forwarding(
        target_hosting_subscription, pa_domain_uid, webspace_uid, plesk_guid, destination, http_code,
        is_mail_domain_need
):
    """
    Register SHM Std Forwarding Site resource on APS bus w/o provisioning
    """
    resource = {
        'destination': destination,
        'httpCode': http_code,
    }

    return _register_site(
        target_hosting_subscription, pa_domain_uid, webspace_uid, plesk_guid, types.SiteWebForwarding, resource,
        is_mail_domain_need
    )


@trace
def register_website_forwarding_frame(
        target_hosting_subscription, pa_domain_uid, webspace_uid, plesk_guid, destination, is_mail_domain_need
):
    """
    Register SHM Std Forwarding Site resource on APS bus w/o provisioning
    """
    resource = {
        'destination': destination,
    }

    return _register_site(
        target_hosting_subscription, pa_domain_uid, webspace_uid, plesk_guid, types.SiteWebForwardingFrm, resource,
        is_mail_domain_need
    )


@trace
def _register_site(
        target_hosting_subscription, pa_domain_uid, webspace_uid, plesk_guid, site_type, resource_properties,
        is_mail_domain_need
):
    subscription_id = target_hosting_subscription.id
    pa_subscription_uid = subscription.get_subscription(subscription_id)['aps']['id']
    shm_application_uid = get_app_instance_uid(types.SHMApplication)

    site_aps_resource = get_website_info_by_pa_domain_uid(subscription_id, pa_domain_uid, site_type)
    aps = get_aps_connector_with_admin_token()
    if site_aps_resource:
        site_aps_id = site_aps_resource['aps']['id']
    else:
        resource = {
            'aps': {'type': site_type},
            'pleskGuid': plesk_guid,
            'syncStatus': 'sync:ok',
            'domain': {'aps': {'id': pa_domain_uid}},
            'webSpace': {'aps': {'id': webspace_uid}},
            'provisionStatus': PROVISION_STATUS_SKIP_MAILDOMAIN_PROVISION
        }
        resource.update(resource_properties)

        aps.setAdditionalHeaders({'APS-Subscription-ID': pa_subscription_uid})

        site_aps_resource = aps.register_resource(shm_application_uid, _get_resource_name(site_type), resource)
        site_aps_id = site_aps_resource['aps']['id']
        _increase_websites_consumption_on_host(subscription_id, webspace_uid)

    webspace_info = aps.get_resource(webspace_uid)
    webspace_service_info = aps.get_resource(webspace_info['service']['aps']['id'])
    subscribe(
        aps, site_aps_id, "http://aps-standard.org/core/events/linked", "eventConfigurationLinked",
        relation="configuration", source_uid=webspace_service_info['aps']['id']
    )
    subscribe(
        aps, site_aps_id, "http://aps-standard.org/core/events/unlinked", "eventConfigurationUnlinked",
        relation="configuration", source_uid=webspace_service_info['aps']['id']
    )
    subscribe(
        aps, site_aps_id, "http://aps-standard.org/core/events/changed", "eventConfigurationChanged",
        source_uid=webspace_service_info['configuration']['aps']['id']
    )

    if not get_mail_domain_by_website_and_pa_domain_uids(site_aps_id, pa_domain_uid) and is_mail_domain_need:
        mail_service = get_shm_mail_service(pa_subscription_uid)
        register_mail_domain(site_aps_id, pa_domain_uid, mail_service['aps']['id'], pa_subscription_uid)

    service_name = _get_resource_name(site_type)
    if (service_name == 'SiteWeb' and 'configuration' not in site_aps_resource) or (
                site_aps_resource['provisionStatus'] in [PROVISION_STATUS_SKIP_MAILDOMAIN_PROVISION,
                                                         PROVISION_STATUS_MAILDOMAIN_PROVISION_SKIPPED]):
        shm_application_instance_id = get_application_instances_ids(types.SHMApplication)[0][0]
        aps = APS2.get_connector(app_instance_id=shm_application_instance_id)
        updated_resource = {
            'aps': {'id': site_aps_id},
            'provisionStatus': PROVISION_STATUS_READY
        }

        if service_name == 'SiteWeb':
            updated_resource['configuration'] = get_website_configuration(shm_application_instance_id, site_aps_id)

        FunctionResultWaiter(
            aps.get_resource,
            InternalError(_T("ERROR_TAKEOVER_STATUS_NOT_CHANGED_FOR_WEBSITE").format(site_guid=site_aps_id)),
            check_function=lambda x: x and x['provisionStatus'] == PROVISION_STATUS_MAILDOMAIN_PROVISION_SKIPPED
        ).run(
            resource_uid=site_aps_id,
            hidelist_response=["$.configuration[?(@.name=='webspace')].value.hostingSettings.ftp_password"]
        )

        aps.update_resource(
            resource=updated_resource,
            service=_get_resource_name(site_type),
            hidelist_request=["$.configuration[?(@.name=='webspace')].value.hostingSettings.ftp_password"],
            hidelist_response=["$.configuration[?(@.name=='webspace')].value.hostingSettings.ftp_password"]
        )

    return site_aps_id


@trace
def _increase_websites_consumption_on_host(subscription_id, webspace_uid):
    webspace_resource = webspace.get_webspace_info(subscription_id, webspace_uid)
    host_capacities = get_host_capacities(webspace_resource['paHost']['aps']['id'])
    host.update_host_websites_capacity(host_capacities, host.HOST_INCREASE_CAPACITY_CONSUMPTION)


@trace
def _get_resource_name(site_type):
    try:
        return {
            types.SiteWeb: 'SiteWeb',
            types.SiteWebForwarding: 'SiteForwardingStd',
            types.SiteWebForwardingFrm: 'SiteForwardingFrm',
        }[site_type]
    except KeyError:
        raise InternalError("Unexpected resource type: %s" % site_type)


@trace
def get_website_info_by_pa_domain_uid(subscription_id, pa_domain_uid, site_type=types.SiteWeb):
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(
        filter_str='implementing(%s),eq(domain.aps.id,%s)' % (site_type, pa_domain_uid),
        hidelist_response=["$.*.configuration[?(@.name=='webspace')].value.hostingSettings.ftp_password"]
    )


@trace
def get_website_linked_with_pa_domain_uid(subscription_id, pa_domain_uid):
    for site_type in [types.WebSiteBase, types.WebSiteAlias]:
        res = get_website_info_by_pa_domain_uid(subscription_id, pa_domain_uid, site_type)
        if res:
            return res

    return None


@trace
def get_website_uid_by_pa_domain_uid(subscription_id, pa_domain_uid, site_name, site_type=types.SiteWeb):
    """
    A site is created asynchronously, so check several times (no more than MAX_ATTEMPTS)
    that web hosting was created successfully.
    """

    function = FunctionResultWaiter(
        get_website_info_by_pa_domain_uid,
        MigrationNoRepeatError(_T("FAILED_CREATE_WEBSITE").format(name=site_name)),
        check_function=lambda x: x and x['aps']['status'] == 'aps:ready'
    )

    return function.run(subscription_id, pa_domain_uid, site_type)['aps']['id']


@trace
def get_website_www_root(subscription_id, website_uid):
    """
    :param subscription_id: int Example: 1000139
    :param website_uid: str Example: 'e604f534-56ab-424a-ace6-e99cf3524f59'
    :return: str|None
    """
    aps = get_aps_connector(subscription_id)
    for resource in aps.get_resources(
            resource_path='%s/configuration' % website_uid,
            hidelist_response=["$.[?(@.name=='webspace')].value.hostingSettings.ftp_password"]
    ):
        if resource['name'] == 'www_root':
            return resource['value']

    return None


@trace
def get_website_configuration(shm_application_instance_id, website_uid):
    aps = APS2.get_connector(app_instance_id=shm_application_instance_id)
    return aps.get_resources(
        resource_path='%s/%s' % (website_uid, aps_operations['configuration']),
        hidelist_response=["$.[?(@.name=='webspace')].value.hostingSettings.ftp_password"]
    )


@trace
def update_hosting_settings(subscription_id, hosting_settings, site_name, import_settings_to_site_webspace=False):
    if hosting_settings is not None:
        configuration = {
            'configuration': {
                'properties': hosting_settings
            }
        }
    else:
        configuration = {}

    if import_settings_to_site_webspace:
        resource_uid = webspace.get_webspace_uid_by_site_name(subscription_id, site_name)
        hidelist_response = ['$.sysPassword']
    else:
        pa_domain_uid = get_pa_domain_by_name(site_name, subscription_id)['aps']['id']
        resource_uid = get_website_info_by_pa_domain_uid(subscription_id, pa_domain_uid)['aps']['id']
        hidelist_response = ["$.configuration[?(@.name=='webspace')].value.hostingSettings.ftp_password"]

    aps = get_aps_connector(subscription_id)
    aps.put_resource(resource_uid, aps_operations['customize'], configuration, hidelist_response=hidelist_response)
    wait_for_all_scheduled_tasks_on_resource_completed(resource_uid)

    if not import_settings_to_site_webspace:
        # Update cached settings in SiteWeb via call to /configuration which re-reads data from Plesk
        shm_application_instance_id = get_application_instances_ids(types.SHMApplication)[0][0]
        get_website_configuration(shm_application_instance_id, resource_uid)
