# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import MigrationNoRepeatError
from migrator.core.language import _T
from migrator.core.utils import cached, FunctionResultWaiter, urlquote
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.ip_utils import get_printable_ip
from migrator.core.utils.registry import Registry
from migrator.odin_automation.services.shm.lib.aps import types
from migrator.odin_automation.services.shm.lib.aps.types import aps_operations
from migrator.odin_automation.platform.lib.api.aps import types as platform_types
from migrator.odin_automation.platform.lib.api.aps_runtime.aps2 import \
    get_aps_connector, get_aps_connector_with_admin_token, get_app_instance_uid, get_error_message, get_error, \
    get_message_with_details
from migrator.odin_automation.platform.lib.api.aps_runtime.apsc import APSCTransportError
from migrator.odin_automation.platform.lib.api.aps_runtime.utils import subscribe
from migrator.odin_automation.platform.lib.task_manager import wait_for_all_scheduled_tasks_on_resource_completed

import migrator.odin_automation.platform.lib.api.aps.domain as domain
import migrator.odin_automation.platform.lib.api.aps.ip_address as ip_address
import migrator.odin_automation.platform.lib.api.aps.subscription as subscription

import api
import website


@trace
def provision_webspace(subscription_id, webspace_name, sysuser_name, is_windows,
                       ipv4_type=None, ipv4=None, ipv6_type=None, ipv6=None):
    """
    Create a webspace. If a webspace with this name already exists, return the webspace ID for this webspace.
    If a webspace with sysuser already exists, create webspace with random sysuser name.
    """
    try:
        webspace = _get_webspace_by_webspace_display_name(subscription_id, webspace_name)
        if webspace is not None:
            return webspace

        _provision_webspace(
            subscription_id, is_windows, display_name=webspace_name, sysuser_name=sysuser_name,
            ipv4_type=ipv4_type, ipv4=ipv4, ipv6_type=ipv6_type, ipv6=ipv6
        )
        return _wait_for_provision_webspace(subscription_id, webspace_name)
    except APSCTransportError as e:
        error_message = get_error_message(e)
        if isinstance(e.body, dict) and 'details' in e.body and 'details' in e.body['details'] and \
            'error' in e.body['details']['details']:
            error = e.body['details']['details']['error']
        else:
            error = get_error(e)

        if 'The system user "%s" already exists' % sysuser_name in error_message:
            raise MigrationNoRepeatError(_T('THERE_SYS_USER_ALREADY_EXIST').format(system_user_name=sysuser_name))
        elif error == 'Plesk.ErrorHandling.IPManager.PoolWithSuitableRangeNotFound':
            raise MigrationNoRepeatError(_T('THERE_IS_NO_APPROPRIATE_POOL_FOUND_FOR_IP_ADDRESS').format(
                ip=get_printable_ip(e.body.details.details.details.ip_address)
            ))
        elif error == 'Plesk.ErrorHandling.IPManager.IPNotFound':
            global_context = Registry.get_instance().get_context()
            poa_api = global_context.conn.target.poa_api()
            host_id = e.body.details.details.details.host_id
            raise MigrationNoRepeatError(_T('THERE_IS_NO_APPROPRIATE_POOL_FOUND_FOR_HOST').format(
                host="#%s (%s)" % (host_id, poa_api.host.get_host(int(host_id)).hostname),
                webspace=webspace_name
            ))
        elif error == 'Plesk.ErrorHandling.IPManager.PoolNotFound':
            global_context = Registry.get_instance().get_context()
            poa_api = global_context.conn.target.poa_api()
            host_id = e.body.details.details.details.host_id
            raise MigrationNoRepeatError(_T('THERE_IS_NO_APPROPRIATE_POOL_FOUND_FOR_IP_ADDRESS_ON_HOST').format(
                ip=get_printable_ip(e.body.details.details.details.ip_address),
                host="#%s (%s)" % (host_id, poa_api.host.get_host(int(host_id)).hostname),
            ))
        elif isinstance(e.body, dict) and 'details' in e.body and 'message' in e.body:
            raise MigrationNoRepeatError(get_message_with_details(e))
        else:
            raise MigrationNoRepeatError(error_message)


@trace
def _provision_webspace(
        subscription_id, is_windows, display_name=None,
        ipv4_type=None, ipv4=None, ipv6_type=None, ipv6=None, sysuser_name=None
):

    service_type = types.WebspaceIIS if is_windows else types.WebspaceApache
    service_aps_id = api.get_service_aps_id(subscription_id, service_type)
    pa_subscription_uid = subscription.get_subscription(subscription_id)['aps']['id']
    post_data = {
        'service': {
            'id': service_aps_id,
            'type': service_type
        },
        'paSubscriptionId': pa_subscription_uid
    }
    if display_name is not None:
        post_data['displayName'] = display_name

    if sysuser_name is not None:
        post_data['sysUsername'] = sysuser_name

    if ipv4_type is not None:
        post_data['ipv4AddressType'] = ipv4_type

    if ipv4 is not None:
        post_data['ipv4'] = ipv4

    if ipv6_type is not None:
        post_data['ipv6AddressType'] = ipv6_type

    if ipv6 is not None:
        post_data['ipv6'] = ipv6

    aps = get_aps_connector(subscription_id)

    return aps.post_resource(service_aps_id, aps_operations['provisionWebspace'], post_data)


@trace
def register_webspace(
        subscription_id, webspace_name, sysuser_name, sysuser_password,
        web_configuration, host_id, ipv4_address_uid, ipv6_address_uid, plesk_guid, hosting_type
):
    """
    Register SHM Apache or IIS Webspace resource on APS bus w/o provisioning
    """
    shm_application_uid = get_app_instance_uid(types.SHMApplication)
    pa_subscription = subscription.get_subscription(subscription_id)
    pa_subscription_uid = pa_subscription['aps']['id']
    pa_host_uid = ip_address.get_oa_host_uid(host_id)

    if 'apache' in web_configuration['aps']['type']:
        service_type = types.WebspaceApache
    elif 'iis' in web_configuration['aps']['type']:
        service_type = types.WebspaceIIS
    else:
        raise Exception(
            'Unexpected error: unknown WebHosting Plesk Web configuration type: %s' % web_configuration['aps']['type']
        )

    service_aps_id = api.get_service_aps_id(subscription_id, service_type)
    resource = {
        'aps': {'type': types.AbstractWebspace},
        'service': {'aps': {'id': service_aps_id}},
        'displayName': webspace_name,
        'name': webspace_name,
        'sysUsername': sysuser_name,
        'pleskGuid': plesk_guid,
        'syncStatus': 'sync:ok',
        'paHost': {'aps': {'id': pa_host_uid}},
        'hosts': [pa_host_uid],
        'serviceName': 'Webspace',
        'serviceDescription': '%s (%s)' % (webspace_name, webspace_name),
        'hostingType': hosting_type
    }
    if ipv4_address_uid:
        resource['paIpAddressIPv4'] = {'aps': {'id': ipv4_address_uid}}
    if ipv6_address_uid:
        resource['paIpAddressIPv6'] = {'aps': {'id': ipv6_address_uid}}
    if sysuser_password:
        resource['sysPassword'] = sysuser_password

    aps = get_aps_connector_with_admin_token()
    aps.setAdditionalHeaders({'APS-Subscription-ID': pa_subscription_uid})

    webspace_resource = aps.register_resource(
        shm_application_uid, 'WebSpace', resource, hidelist_request=['$.sysPassword']
    )
    aps.link_resource(webspace_resource['aps']['id'], 'subscription', pa_subscription_uid)

    webspace_service_info = aps.get_resource(service_aps_id)
    subscribe(
        aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/linked", "eventConfigurationLinked",
        relation="configuration", source_uid=service_aps_id
    )
    subscribe(
        aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/unlinked", "eventConfigurationUnlinked",
        relation="configuration", source_uid=service_aps_id
    )
    subscribe(
        aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/changed", "eventConfigurationChanged",
        source_uid=webspace_service_info['configuration']['aps']['id']
    )
    subscribe(
        aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/removed", "onWebSiteRemoved",
        source_type="http://www.odin.com/shm/web/site/1.0"
    )

    if ipv4_address_uid:
        subscribe(
            aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/changed", "onIpAddressChanged",
            source_uid=ipv4_address_uid
        )
    if ipv6_address_uid:
        subscribe(
            aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/changed", "onIpAddressChanged",
            source_uid=ipv6_address_uid
        )

    admin_users = aps.get_resources(
        filter_str='implementing(%s),linkedWith(%s)' % (platform_types.PAAdminUser, pa_subscription_uid)
    )
    for user in admin_users:
        subscribe(
            aps, webspace_resource['aps']['id'], "http://aps-standard.org/core/events/changed", "eventAdminUserChanged",
            source_uid=user['aps']['id']
        )

    return webspace_resource


@trace
def get_webspace_by_webspace_name(subscription_id, webspace_name):
    webspace_type = types.AbstractWebspace
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(filter_str='implementing(%s),eq(name,%s)' % (webspace_type, urlquote(webspace_name)))


@trace
def get_webspace_uid_by_site_name(subscription_id, site_name):
    return _get_webspace_info_by_site_name(subscription_id, site_name)['aps']['id']


@trace
def get_webspace_name_by_site_name(subscription_id, site_name):
    return _get_webspace_info_by_site_name(subscription_id, site_name)['name']


@trace
def get_webspace_info(subscription_id, webspace_uid):
    aps = get_aps_connector(subscription_id)
    return aps.get_resource(webspace_uid)


@trace
def get_sysuser_name_by_site_name(subscription_id, site_name):
    return _get_webspace_info_by_site_name(subscription_id, site_name)['sysUsername']


@trace
def get_web_configurations(subscription_id, is_windows):
    if is_windows:
        webspace_type = types.WebConfigurationIIS
    else:
        webspace_type = types.WebConfigurationApache

    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(filter_str='implementing(%s)' % webspace_type)


@trace
def provision_web_configuration(subscription_id, is_windows):
    if is_windows:
        webspace_type = types.WebConfigurationIIS
    else:
        webspace_type = types.WebConfigurationApache

    # TODO: replace by migrator.odin_automation.platform.lib.api.aps_runtime.apsc.ControllerProxy.provisionResource()
    # call
    aps = get_aps_connector(subscription_id)
    return aps.post_resource('', '', {'aps': {'type': webspace_type}})


@trace
def get_sysuser_name_by_webspace_uid(subscription_id, webspace_uid):
    aps = get_aps_connector(subscription_id)
    return aps.get_resources(webspace_uid)['sysUsername']


@trace
def change_webspace_sysuser_password(subscription_id, webspace_uid, password):
    aps = get_aps_connector(subscription_id)
    aps.put_resource(
        webspace_uid, aps_operations['changePassword'], {'sysPassword': password}, hidelist_request=['$.sysPassword']
    )
    return wait_for_all_scheduled_tasks_on_resource_completed(webspace_uid)


@cached
def _get_web_subscription_uid(subscription_id):
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(filter_str='implementing(%s)' % types.WebSubscription)['aps']['id']


@trace
def _get_webspace_info_by_site_name(subscription_id, site_name):
    pa_domain_uid = domain.get_pa_domain_by_name(site_name, subscription_id)['aps']['id']
    website_uid = website.get_website_uid_by_pa_domain_uid(subscription_id, pa_domain_uid, site_name, types.WebSiteBase)
    return _get_webspace_info_by_website_uid(subscription_id, website_uid)


@trace
def _get_webspace_info_by_website_uid(subscription_id, website_uid):
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(resource_path='%s/webSpace' % website_uid)


@trace
def _get_webspace_by_webspace_display_name(subscription_id, webspace_name):
    webspace_type = types.AbstractWebspace
    aps = get_aps_connector(subscription_id)
    return aps.get_first_resource(
        filter_str='implementing(%s),eq(displayName,%s)' % (webspace_type, urlquote(webspace_name))
    )


@trace
def _wait_for_provision_webspace(subscription_id, webspace_name):
    """
    A webspace is created asynchronously, so check several times (no more than MAX_ATTEMPTS)
    that web hosting was created successfully.
    """

    function = FunctionResultWaiter(
        _get_webspace_by_webspace_display_name,
        MigrationNoRepeatError(_T("FAILED_CREATE_WEBSPACE").format(name=webspace_name)),
        check_function=lambda x: x and x['aps']['status'] == 'aps:ready'
    )

    return function.run(subscription_id, webspace_name)


@trace
def get_webspace_ips(subscription_id, webspace_uid):
    webspace_info = get_webspace_info(subscription_id, webspace_uid)
    ips = []
    aps = get_aps_connector(subscription_id)
    for iptype in ['paIpAddressIPv4', 'paIpAddressIPv6']:
        if iptype in webspace_info:
            ips.append(aps.get_resource(webspace_info[iptype]['aps']['id']))

    return ips

