# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from collections import OrderedDict, namedtuple
from lxml import etree

from migrator import settings
from migrator.core import MigrationError, MigrationNoRepeatError, InternalError, MigrationNotice
from migrator.core.connections.database_server import is_local_database_host
from migrator.core.data.model.web_hosting_shm import PhysicalHostingModel, PHPSettingsModel
from migrator.core.language import _T
from migrator.core.transfer.db import escape_db_passwd, mysql, pgsql
from migrator.core.utils import parse_bool
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.networking.plesk_cli_runner import PleskCLIRunnerCLI
from migrator.core.utils.operating_system.windows_utils import fix_path
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.utils.networking.hcl_runner.hcl_common import HclRunnerException
from migrator.odin_automation.services.shm.lib import \
    get_database_server, get_hash_name_for_database, get_hash_name_for_database2, get_hash_name_for_database_user2, WebspaceMacroDefinition, SiteMacroDefinition
from migrator.odin_automation.services.shm.lib.aps.website import update_hosting_settings
from migrator.odin_automation.services.shm.lib.aps.webspace import get_web_configurations
from migrator.products.plesk.lib import get_plesk_database, get_plesk_database_user, get_protected_directory_id, \
    is_ftp_user_exists_for_webspace, get_plesk_site_id, delete_plesk_database
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import Failure, PleskError

logger = logging.getLogger(__name__)


class TryOtherDatabaseNameError(MigrationError):
    def __init__(self, msg):
        super(MigrationError, self).__init__(msg)


property_name_mapping = {
    'asp_dot_net_version': 'asp_dot_net.version',
    'php_handler_id': 'php.handlerVersion',
}
property_name_mapping.update(PHPSettingsModel.oam_vs_ini_mapping)

# Permissions that should be disabled before settings applied.
# Keys are permissions' names, values are default values that will be restored in case of empty 'properties' field
# of web template WHP resource.
permissions_that_block_settings_applying = {
    'manage_phosting': {
        'name': 'manage_phosting',
        'value': False,
        'class': 'permission',
        'type': 'boolean'
    },
    'manage_php_settings': {
        'name': 'manage_php_settings',
        'value': False,
        'class': 'permission',
        'type': 'boolean'
    },
    'manage_php_version': {
        'name': 'manage_php_version',
        'value': True,
        'class': 'permission',
        'type': 'boolean'
    },
    'manage_webstat': {
        'name': 'manage_webstat',
        'value': True,
        'class': 'permission',
        'type': 'boolean'
    },
    'manage_sh_access': {
        'name': 'manage_sh_access',
        'value': False,
        'class': 'permission',
        'type': 'boolean'
    }
}

# List of OS specific hosting settings
windows_specific_options = ('asp', 'asp_dot_net', 'iis_app_pool', 'write_modify')
linux_specific_options = ()

# List of properties missing from capabilities but supported by webspace/site configuration
configuration_only_options = {
    'webstat_protected': {
        'type': 'boolean',
        'class': 'setting'
    }
}


@trace
def import_hosting_settings(site, subscription, plesk_target_server, errors, import_settings_to_site_webspace=False):
    def _fill_properties_dict(properties_dict, settings, is_windows=False):
        # We do not support cross OS migration (WIN <-> LIN), so remove parameters which are not relevant to
        # specified OS type
        if not is_windows:
            blacklist = windows_specific_options
        else:
            blacklist = linux_specific_options

        for key in settings:
            if key in blacklist:
                continue

            if settings[key] is not None:
                properties_dict[key] = settings[key]
            else:
                logger.debug("Skip property '%s' as undefined" % key)

    def _get_suitable_parameter(param_id, param_name, param_aps, aps_setting, function):
        if param_id in properties_dict:
            if global_context.migration_project.settings[aps_setting] == 'sourceVersion':
                parameter = function(capabilities[param_aps]['enum'], site, properties_dict[param_id])
                if parameter:
                    logger.debug(
                        "The most suitable %s for site '%s' is '%s' at the target %s",
                        param_name, site.name, parameter, plesk_target_server.description()
                    )
                    properties_dict[param_id] = parameter
                else:
                    logger.debug(
                        "The suitable %s for site '%s' is not found at the target %s",
                        param_name, site.name, plesk_target_server.description()
                    )
                    del properties_dict[param_id]
            else:
                logger.debug("Skip migrating %s version", param_name)
                del properties_dict[param_id]

    try:
        properties_dict = {}
        for params in [site.hosting.hosting_properties, site.hosting.scripting, site.hosting.php_settings]:
            _fill_properties_dict(properties_dict, params.as_dictionary(), plesk_target_server.is_windows())

        capabilities = plesk_target_server.capabilities
        capabilities.update(configuration_only_options)

        global_context = Registry.get_instance().get_context()

        _get_suitable_parameter(
            'php_handler_id', 'PHP Handler', 'php.handlerVersion', 'phpVersion', _get_suitable_php_handler
        )
        _get_suitable_parameter(
            'asp_dot_net_version', 'ASP Version', 'asp_dot_net.version', 'aspNetVersion', _get_suitable_asp_version
        )

        hosting_settings = []
        unsupported_settings = []
        for property_name, value in properties_dict.iteritems():
            if property_name in property_name_mapping:
                _property_name = property_name_mapping[property_name]
            else:
                _property_name = property_name

            if _property_name in capabilities:
                if capabilities[_property_name]['type'] == 'boolean':
                    val = 'true' if parse_bool(value) else 'false'
                else:
                    val = value

                hosting_settings.append({
                    'name': _property_name,
                    'value': val,
                    'class': capabilities[_property_name]['class'],
                    'type': capabilities[_property_name]['type'],
                })
            else:
                unsupported_settings.append(_property_name)
                logger.debug(
                    "Property '%s' is not supported at target Plesk %s",
                    _property_name, plesk_target_server.description()
                )

        if unsupported_settings:
            notice_message = _T("UNABLE_TO_RESTORE_UNSUPPORTED_HOSTING_SETTINGS").format(
                sitename=site.name,
                sub_id=subscription.external_id,
                sub_name=subscription.name,
                target=plesk_target_server.description(),
                settings=", ".join(sorted(unsupported_settings))
            )
            logger.warning(notice_message)
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem('unsupported-hosting-settings-were-not-restored', Problem.NOTICE, notice_message), None)
            )

        update_hosting_settings(subscription.id, hosting_settings, site.name, import_settings_to_site_webspace)
    except MigrationError as e:
        logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
        errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_HOSTING_SETTINGS").format(
            site_name=site.name, error=e.message
        ))


@trace
def disable_site_webspace_permissions(site, subscription):
    permissions = []
    for permission_name in list(permissions_that_block_settings_applying.keys()):
        permissions.append({
            'name': permission_name,
            'value': False,
            'class': 'permission',
            'type': 'boolean'
        })

    update_hosting_settings(subscription.id, permissions, site.name, True)


@trace
def restore_site_webspace_permissions_from_web_template(site, subscription, plesk_target_server):
    permissions = []
    web_config = get_web_configurations(subscription.id, plesk_target_server.is_windows())
    names_of_permissions_that_block_settings_applying = list(permissions_that_block_settings_applying.keys())

    if 'properties' in web_config:
        for prop in web_config['properties']:
            if prop['class'] == 'permission' and prop['name'] in names_of_permissions_that_block_settings_applying:
                permissions.append({
                    'name': prop['name'],
                    'value': prop['value'],
                    'class': 'permission',
                    'type': 'boolean'
                })
    else:
        for permission_name, permission_setting in permissions_that_block_settings_applying.iteritems():
            permissions.append(permission_setting)

    update_hosting_settings(subscription.id, permissions, site.name, True)


@trace
def _get_suitable_asp_version(asp_versions, site, source_asp_version):
    available_asp_versions = {}
    for asp_version in asp_versions:
        available_asp_versions[tuple(asp_version.split('.'))] = asp_version

    available_asp_versions = OrderedDict(sorted(available_asp_versions.items(), key=lambda t: t[0]))

    source_asp = tuple(source_asp_version.split('.'))
    logger.debug("Detected ASP version '%s' (%s) for site '%s'", source_asp, source_asp_version, site.name)
    logger.debug("Available ASP versions: '%s'", available_asp_versions)

    for asp_version in available_asp_versions:
        if asp_version >= source_asp:
            return available_asp_versions[asp_version]

    if available_asp_versions:
        return available_asp_versions.popitem()[1]
    else:
        return None


@trace
def _get_suitable_php_handler(php_handlers, site, source_php_handler_id):
    def _parse_php_handler_id(handler_id, version=None):
        tmp = handler_id.replace('_', '-').split('-')
        result = None
        if len(tmp) == 2:  # windows, 'fastcgi-5.4'
            result = {'version': 'php' + tmp[1].replace('.', ''), 'type': tmp[0]}
        elif len(tmp) == 3:  # linux, 'plesk-php54-fastcgi'
            result = {'version': tmp[1], 'type': tmp[2]}
        elif version is not None:
            tmp_v = version.split('.')
            result = {'version': 'php' + tmp_v[0] + tmp_v[1], 'type': tmp[0]}

        return result

    available_php_handlers_version_type_map = {}
    available_php_handlers_version_map = {}
    for handler in php_handlers:
        try:
            _, h_id, _, version = handler.split('/')
            php_handler_id = _parse_php_handler_id(h_id, version)
            if php_handler_id is not None:
                available_php_handlers_version_type_map[(php_handler_id['version'], php_handler_id['type'])] = handler
                available_php_handlers_version_map[php_handler_id['version']] = handler
        except ValueError:
            logger.debug("Could not parse PHP handler: '%s'", handler)

    logger.debug("Available PHP handlers by version and type: '%s'", available_php_handlers_version_type_map)
    logger.debug("Available PHP handlers by version: '%s'", available_php_handlers_version_map)
    php_handler_id = _parse_php_handler_id(source_php_handler_id)
    if php_handler_id is None:
        return

    logger.debug("Detected PHP version '%s' (%s) for site '%s'", php_handler_id, source_php_handler_id, site.name)
    preferred_versions_map = OrderedDict([
        ('php4',    ['php4', 'php5', 'php52', 'php53', 'php54', 'php55', 'php56']),
        ('php52',   ['php5', 'php52', 'php53', 'php54', 'php55', 'php56']),
        ('php53',   ['php53', 'php54', 'php55', 'php56', 'php52']),
        ('php54',   ['php54', 'php55', 'php56', 'php53', 'php52']),
        ('php55',   ['php55', 'php56', 'php54', 'php53', 'php52']),
        ('php56',   ['php56', 'php55', 'php54', 'php53', 'php52']),
        ('php5',    ['php5', 'php52', 'php53', 'php54', 'php55', 'php56']),
        ('php70',   ['php70', 'php71']),
        ('php71',   ['php71']),
    ])

    for source_version in preferred_versions_map:
        if php_handler_id['version'].startswith(source_version):
            for target_version in preferred_versions_map[source_version]:
                if (target_version, php_handler_id['type']) in available_php_handlers_version_type_map:
                    return available_php_handlers_version_type_map[(target_version, php_handler_id['type'])]

            for target_version in preferred_versions_map[source_version]:
                if target_version in available_php_handlers_version_map:
                    return available_php_handlers_version_map[target_version]

    return None


@trace
def import_ftp_users(webspace, plesk_webspace_id, plesk_target_server, errors):
    for ftp_user in webspace.ftp_users:
        if plesk_target_server.is_windows():
            ftp_home_dir = u'%s\%s' % (webspace.www_root, ftp_user.home)
            ftp_home_dir = fix_path(ftp_home_dir)
        else:
            ftp_home_dir = u'%s/%s' % (webspace.www_root, ftp_user.home)
            ftp_home_dir = ftp_home_dir.replace('//', '/')

        with plesk_target_server.runner() as runner:
            runner.mkdir(ftp_home_dir)

        with plesk_target_server.plesk_api().logger.censor({'xml': ['/packet/ftp-user/add/password']}):
            result = plesk_target_server.plesk_api().send(
                plesk_ops.FtpuserOperator.Add(
                    webspace_id=plesk_webspace_id,
                    name=ftp_user.name,
                    password=ftp_user.password,
                    home=ftp_user.home,
                    quota=ftp_user.quota,
                )
            )

        is_ftp_user_exists, is_ftp_user_system = is_ftp_user_exists_for_webspace(
            plesk_api=plesk_target_server.plesk_api(), ftp_user_name=ftp_user.name, webspace_id=plesk_webspace_id
        )

        if isinstance(result, Failure):
            if result.code == 1007:
                if not is_ftp_user_exists:
                    errors.append(_T("ERROR_RESTORING_FTP_USER_ALREADY_EXISTS_FOR_ANOTHER_WEBSPACE").format(
                        ftp_user_name=ftp_user.name, webspace_name=webspace.name
                    ))
                    continue
            else:
                # Plesk returns code 1024 in case when FTP limit is reached, but user can be already created previously
                if result.code != 1024 or not is_ftp_user_exists:
                    errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_FTP_USER").format(
                        ftp_user_name=ftp_user.name, webspace_name=webspace.name, error=result.message.rstrip('.')
                    ))
                    continue

        if not is_ftp_user_system and (ftp_user.read is not None or ftp_user.write is not None):
            result = plesk_target_server.plesk_api().send(plesk_ops.FtpuserOperator.Set(
                name=ftp_user.name, read=ftp_user.read, write=ftp_user.write
            ))
            if isinstance(result, Failure):
                errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_FTP_USER").format(
                    ftp_user_name=ftp_user.name, webspace_name=webspace.name, error=result.message.rstrip('.')
                ))


@trace
def import_web_users(subscription, webspace, plesk_webspace_id, plesk_target_server, errors):
    global_context = Registry.get_instance().get_context()
    for web_user in webspace.web_users:
        with plesk_target_server.plesk_api().logger.censor({'xml': ['/packet/webuser/add/password']}):
            try:
                result = plesk_target_server.plesk_api().send(
                    plesk_ops.WebuserOperator.Add(
                        site_id=plesk_webspace_id,
                        login=web_user.login,
                        password=web_user.password,
                    )
                )
                if isinstance(result, Failure) and result.code != 1007:
                    errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_WEB_USER").format(
                        web_user_login=web_user.login, webspace_name=webspace.name, error=result.message.rstrip('.')
                    ))
                    continue

            except PleskError as e:
                if e.code == 1014:
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription.source_info.external_id,
                        Issue(
                            Problem(
                                'xml-rpc-does-not-support-webusers',
                                Problem.NOTICE,
                                _T('UNABLE_TO_MIGRATE_WEBUSER_DUE_TO_XMLRPC_LIMITATIONS').format(
                                    web_user_login=web_user.login, webspace_name=webspace.name)
                            ),
                            None
                        )
                    )
                    continue
                else:
                    raise

        result = plesk_target_server.plesk_api().send(
            plesk_ops.WebuserOperator.Set(
                login=web_user.login,
                asp=web_user.asp if 'asp' in plesk_target_server.capabilities else None,
                asp_dot_net=web_user.asp_dot_net if 'asp_dot_net' in plesk_target_server.capabilities else None,
                php=web_user.php if 'php' in plesk_target_server.capabilities else None,
                cgi=web_user.cgi if 'cgi' in plesk_target_server.capabilities else None,
                perl=web_user.perl if 'perl' in plesk_target_server.capabilities else None,
                python=web_user.python if 'python' in plesk_target_server.capabilities else None,
                fastcgi=web_user.fastcgi if 'fastcgi' in plesk_target_server.capabilities else None,
                write_modify=web_user.write_modify,
            )
        )
        if isinstance(result, Failure):
            errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_WEB_USER").format(
                web_user_login=web_user.login, webspace_name=webspace.name, error=result.message.rstrip('.')
            ))


@trace
def import_protected_directories(
        site_name, protected_directories, plesk_site_id, plesk_target_server, errors, subscription_id, same_ssl=True
):
    for protected_directory in protected_directories:
        properties = protected_directory.properties
        if same_ssl:
            # filter out both 'ssl' and 'nonssl' pd properties to import all pds into single docroot
            properties = [x for x in protected_directory.properties if 'ssl' not in x.name]

        if protected_directory.name is None:
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_id,
                Issue(Problem(
                    'import-protected-dir', Problem.NOTICE,
                    _T("PROTECTED_DIRECTORY_NOT_RESTORED").format(site_name=site_name),
                ), None)
            )
            continue

        result = plesk_target_server.plesk_api().send(
            plesk_ops.ProtectedDirectoryOperator.Add(
                site_id=plesk_site_id,
                name=protected_directory.name,
                header=protected_directory.header,
                properties=properties,
            )
        )

        if isinstance(result, Failure):
            if result.code == 1007:
                pd_id = get_protected_directory_id(
                    plesk_target_server.plesk_api(), plesk_site_id, protected_directory.name, protected_directory.header
                )

                if pd_id is None:
                    errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_PROTECTED_DIRECTORY").format(
                        pd_name=protected_directory.header, site_name=site_name,
                        error=_T("ERROR_OCCURRED_WHILE_RESTORING_PD_ALREADY_EXISTS").format(
                            pd_name=protected_directory.header
                        )
                    ))
                    continue
            else:
                errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_PROTECTED_DIRECTORY").format(
                    pd_name=protected_directory.header, site_name=site_name, error=result.message.rstrip('.')
                ))
                continue
        else:
            pd_id = result.data

        for pd_user in protected_directory.pd_users:
                result = plesk_target_server.plesk_api().send(
                    plesk_ops.ProtectedDirectoryOperator.AddUser(
                        pd_id=pd_id,
                        login=pd_user.login,
                        password=pd_user.password,
                        password_type=pd_user.password_type
                    )
                )

                if isinstance(result, Failure) and result.code != 1007:
                    errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_PD_USER").format(
                        login=pd_user.login, pd_name=protected_directory.header, site_name=site_name,
                        error=result.message.rstrip('.')
                    ))


@trace
def import_virtual_directories(webspace, site, plesk_site_id, plesk_target_server, errors):
    # Get list of directories to remove from vdirs XML
    # The list is taken from Plesk's RestoreVirtualDirectories function
    # at windows/binary/plesk-utils/PMM/psadumpagent/RestoreEnv.cs

    _reserved_vdirs = [
        '_vti_bin', 'jakarta', 'JRunScripts', 'CFIDE', 'cfdocs', 'cgi-bin', 'plesk-stat', 'sitepreview'
    ]

    vdirs = [vdir for vdir in site.hosting.virtual_directories if vdir not in _reserved_vdirs]
    if not vdirs:
        logger.debug("No virtual directories information for site '%s', won't transfer them" % site.name)
        return

    _serialize_bool = lambda val: None if val is None else 'True' if parse_bool(val) else 'False'

    webspace_macro_definition = WebspaceMacroDefinition(webspace)
    site_macro_definition = SiteMacroDefinition(webspace_macro_definition, site)

    def _add_vdirs(root, vdirs):
        node_vdirs = etree.Element('vdirs')
        root.append(node_vdirs)
        for vdir in vdirs:
            node_vdir = etree.Element('vdir', name=vdir.name)
            attrib = dict(
                path=vdir.path,
                aspEnableClientDebugInherited=_serialize_bool(vdir.asp_enable_client_debug_inherited),
                aspEnableClientDebugDefault=_serialize_bool(vdir.asp_enable_client_debug_default),
                aspEnableClientDebug=_serialize_bool(vdir.asp_enable_client_debug),
                aspEnableServerDebugInherited=_serialize_bool(vdir.asp_enable_server_debug_inherited),
                aspEnableServerDebugDefault=_serialize_bool(vdir.asp_enable_server_debug_default),
                aspEnableServerDebug=_serialize_bool(vdir.asp_enable_server_debug),
                aspBufferingLimitInherited=_serialize_bool(vdir.asp_buffering_limit_inherited),
                aspBufferingLimitDefault=vdir.asp_buffering_limit_default,
                aspBufferingLimit=vdir.asp_buffering_limit,
                aspMaxRequestEntityAllowedInherited=_serialize_bool(vdir.asp_max_request_entity_allowed_inherited),
                aspMaxRequestEntityAllowedDefault=vdir.asp_max_request_entity_allowed_default,
                aspMaxRequestEntityAllowed=vdir.asp_max_request_entity_allowed,
                editable=_serialize_bool(vdir.editable),
                scriptMaps=vdir.script_maps,
                requireSSL=_serialize_bool(vdir.require_ssl),
                authNTLM=_serialize_bool(vdir.auth_ntlm),
                authAnonymous=_serialize_bool(vdir.auth_anonymous),
                aspExecuteInMTA=_serialize_bool(vdir.asp_execute_in_mta),
                aspEnableParentPaths=_serialize_bool(vdir.asp_enable_parent_paths),
                defaultDoc=vdir.default_doc,
                enableDefaultDoc=_serialize_bool(vdir.enable_default_doc),
                appCreate=_serialize_bool(vdir.app_create),
                dontLog=_serialize_bool(vdir.dont_log),
                enableDirBrowsing=_serialize_bool(vdir.enable_dir_browsing),
                accessSource=_serialize_bool(vdir.access_source),
                accessExecute=_serialize_bool(vdir.access_execute),
                accessScript=_serialize_bool(vdir.access_script),
                accessWrite=_serialize_bool(vdir.access_write),
                accessRead=_serialize_bool(vdir.access_read)
            )

            for key in attrib.keys():
                if attrib[key] is None:
                    del attrib[key]

            if 'path' in attrib:
                attrib['path'] = site_macro_definition.substitute(attrib['path'])

            node_vdir.attrib.update(attrib)

            _add_vdirs(node_vdir, vdir.virtual_directories)
            node_vdirs.append(node_vdir)

    root = etree.ElementTree(etree.Element('vdir'))
    _add_vdirs(root.getroot(), vdirs)

    vdir_info = etree.tostring(root, encoding='utf-8', method='xml')

    server = plesk_target_server

    paths = root.xpath('//vdir/@path')
    with server.runner() as runner:
        for path in paths:
            runner.mkdir(path)

    # we need to use this lock because calling "websrvmng" when Plesk restore is running could
    # result in "Access is denied. (Error code 5) at create mutex Global" error
    with server.websrvmng_lock, server.runner() as runner:
        vdir_info_filename = server.get_session_file_path('vdirs.xml')
        runner.upload_file_content(vdir_info_filename, vdir_info)
        runner.sh(
            '{websrvmng_path} --update-vdirs --vhost-name={vhost_name} --vdir-config={vdir_config_file}',
            dict(
                websrvmng_path=server.websrvmng_bin, vhost_name=site.name.encode('idna'),
                vdir_config_file=vdir_info_filename
            )
        )


@trace
def _add_database(plesk_api, database_name, dbtype, plesk_webspace_id, webspace_name, db_server_id):
    plesk_database = get_plesk_database(plesk_api, plesk_webspace_id, database_name, dbtype)
    if plesk_database:
        delete_plesk_database(plesk_api, plesk_database.id)

    result = plesk_api.send(
        plesk_ops.DatabaseOperator.AddDb(
            webspace_id=plesk_webspace_id,
            name=database_name,
            type=dbtype,
            db_server_id=db_server_id
        )
    )

    if isinstance(result, Failure):
        if result.code == 1007:
            plesk_database = get_plesk_database(plesk_api, plesk_webspace_id, database_name, dbtype)

            if plesk_database is None:
                raise TryOtherDatabaseNameError(_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE").format(
                    database_name=database_name, webspace_name=webspace_name,
                    error=_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE_ALREADY_EXISTS").format(
                        database_name=database_name
                    )
                ))
            else:
                return plesk_database.id
        else:
            raise MigrationError(_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE").format(
                database_name=database_name, webspace_name=webspace_name, error=result.message.rstrip('.')
            ))
    else:
        return result.data.id


@trace
def _add_database_user(plesk_api, login, password, db_id, db_server_id, webspace_id, database_name, webspace_name):
    if db_id:
        params = dict(login=login, password=password, db_id=db_id, webspace_id=None, db_server_id=None)
    else:
        params = dict(login=login, password=password, db_id=None, webspace_id=webspace_id, db_server_id=db_server_id)

    result = plesk_api.send(plesk_ops.DatabaseOperator.AddDbUser(**params))
    if isinstance(result, Failure):
        if result.code == 1007:
            plesk_database_user = get_plesk_database_user(plesk_api, db_id, login)

            if not plesk_database_user:
                raise TryOtherDatabaseNameError(_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE_USER").format(
                    database_name=database_name, webspace_name=webspace_name, user_name=login, login=login,
                    error=_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE_USER_ALREADY_EXISTS").format(user_name=login)
                    ))
        else:
            if result.code == 1023 and 'Password validation failed' in result.message:
                error = _T('PASSWORD_DOES_NOT_MEET_WINDOWS_POLICY_REQUIREMENT')
            else:
                error = result.message.rstrip('.')

            raise MigrationError(_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE_USER").format(
                user_name=login, database_name=database_name, webspace_name=webspace_name, error=error
            ))


@trace
def _call_plesk_bin_database(runner, cmd, database, database_server, domain_name, additional_params=None):
    parameters = {
        'domain_name': domain_name,
        'database_type': database_server.dbtype,
        'server_host': database_server.host, 'server_port': database_server.port
    }
    if database:
        parameters['database_name'] = database.name

    if additional_params:
        parameters.update(additional_params)

    try:
        runner.sh(cmd, parameters, valid_codes=[0, 2])  # 2 - Database with such name is already registered.
    except HclRunnerException as e:

        if e.exit_code == 4:  # No such database server registered in Plesk
            raise MigrationNoRepeatError(_T("EXTERNAL_PLESK_DB_IS_NOT_REGISTERED_IN_SHM").format(
                database_name=parameters.get('database_name'),
                subscription_name=parameters['domain_name'],
                database_server=parameters['server_host']
            ))
        else:
            raise


def _get_cmd_database_path(plesk_target_server):
    return '%s%s' % (plesk_target_server.plesk_dir, '\\bin\\database' if plesk_target_server.is_windows() else '/bin/database')


def _get_cmd_database(dbtype, object):
    register_cmd = {
        'database': '{cmd_path} --register {database_name} -domain {domain_name} -type {database_type}',
        'database_user': '{cmd_path} --register-dbuser {database_user} -domain {domain_name} -type {database_type} -database {database_name} -passwd {database_password}'
    }

    register_database_cmd = {
        'mysql': ' -server {server_host}:{server_port}',
        'mssql': ' -server {server_host}',
        'postgresql': ' -server {server_host}:{server_port}',
    }

    if dbtype not in register_database_cmd:
        raise InternalError('Unknown database type: %s' % dbtype)

    return register_cmd[object] + register_database_cmd[dbtype]


@trace
def _register_database(plesk_target_server, database, database_server, domain_name):
    register_database_cmd = _get_cmd_database(database_server.dbtype, 'database')

    params = {
        'cmd_path': _get_cmd_database_path(plesk_target_server)
        }

    with plesk_target_server.runner() as runner:
        _call_plesk_bin_database(
            runner, register_database_cmd, database, database_server, domain_name, params
        )


@trace
def _register_database_user(
    plesk_target_server, user, databases, database_server, dbserver_login, dbserver_password, domain_name, subscription
):
    register_user_cmd = _get_cmd_database(database_server.dbtype, 'database_user')

    for database in databases:
        params = {
            'database_user': user.login, 'database_password': escape_db_passwd(user.password),
            'cmd_path': _get_cmd_database_path(plesk_target_server)
        }

        with plesk_target_server.runner() as runner:
            _call_plesk_bin_database(
                runner, register_user_cmd, database, database_server, domain_name, params
            )

        _allow_user_access = {
            'mysql': mysql.allow_user_access,
            'postgresql': pgsql.allow_user_access,
            'mssql': lambda *x: None,
        }

        try:
            _allow_user_access[database_server.dbtype](
                database_server, dbserver_login, dbserver_password, plesk_target_server, database, user
            )
        except MigrationNotice as e:
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem(None, Problem.NOTICE, e.message), None)
            )


@trace
def import_databases(subscription, webspace, plesk_webspace_id, plesk_target_server, errors):
    def _find_target_db_server_id(source_database_server):
        dbtype = source_database_server.dbtype
        if dbtype in target_db_server:
            return target_db_server[dbtype]
        else:
            errors.append(_T("ERROR_OCCURRED_WHILE_RESTORING_DATABASE").format(
                database_name=database.name, webspace_name=webspace.name,
                error=_T("THERE_IS_NO_DATABASE_SERVER_WITH_TYPE").format(
                    type=dbtype, target=plesk_target_server.description()
                )
            ))
            return None

    def _is_dbtype_supported(dbtype):
        if dbtype not in plesk_target_server.get_supported_db_types():
            global_context.execution_wrapper.add_subscription_issue(
                subscription.source_info.external_id,
                Issue(Problem(
                    'import-databases', Problem.NOTICE,
                    _T('DB_TYPE_NOT_SUPPORTED').format(
                        database_type=dbtype, target_server=plesk_target_server.description()
                    )),
                    None
                )
            )
            return False

        return True

    plesk_api = plesk_target_server.plesk_api()

    result = plesk_api.send(
        plesk_ops.SubscriptionOperator.DbServers.List(plesk_ops.SubscriptionOperator.FilterById([plesk_webspace_id]))
    )
    target_db_server = {x.server_type: x.server_id for x in result[0].data[1] if not isinstance(x, Failure)}

    database_ext_type = namedtuple('database_ext_type', ('target_db_id', 'database'))
    database_ext_map = {}

    global_context = Registry.get_instance().get_context()

    for database in webspace.databases:
        source_database_server = get_database_server(database.database_server_id)
        dbtype = source_database_server.dbtype
        if not _is_dbtype_supported(dbtype):
            continue

        if is_local_database_host(source_database_server.host, [plesk_target_server.ip]):
            db_server_id = _find_target_db_server_id(source_database_server)
            if db_server_id:
                add_database = lambda database_name: \
                    _add_database(plesk_api, database_name, dbtype, plesk_webspace_id, webspace.name, db_server_id)
                try:
                    db_id = add_database(database.name)
                    webspace.database_map[database.name] = database.name
                except TryOtherDatabaseNameError:
                    max_name_len = settings.DatabaseServerTypeInfo[dbtype].database_name_len

                    #  For compatibility with the previous resolving of database name collisions.
                    #  We have to remove the database with non unique name
                    new_database_name = get_hash_name_for_database(source_database_server.host, database.name, max_name_len)
                    plesk_database = get_plesk_database(plesk_api, plesk_webspace_id, new_database_name, dbtype)
                    if plesk_database:
                        delete_plesk_database(plesk_api, plesk_database.id)

                    new_database_name = get_hash_name_for_database2(max_name_len, webspace.web_source_host, source_database_server, database)
                    message = _T('DATABASE_WITH_NAME_ALREADY_EXISTS').format(
                        database_name=database.name, server=plesk_target_server.description(),
                        new_database_name=new_database_name
                    )
                    logger.debug(message)
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription.source_info.external_id,
                        Issue(Problem(None, Problem.NOTICE, message), None)
                    )
                    db_id = add_database(new_database_name)
                    webspace.database_map[database.name] = new_database_name

                database_ext_map[database.external_id] = database_ext_type(db_id, database)
        else:
            try:
                database_server = plesk_target_server.get_database_server(source_database_server)
                if database_server is None:
                    logger.debug('Database server %s (%s) was not found at %s. Add.' % (
                        source_database_server.host, source_database_server.dbtype, plesk_target_server.description()
                    ))
                    plesk_target_server.add_database_server(source_database_server)
                    database_server = source_database_server

                _register_database(plesk_target_server, database, database_server, webspace.technical_name)
                webspace.database_map[database.name] = database.name
            except Exception as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                errors.append(_T('FAILED_REGISTER_DB_ERROR').format(
                    database_name=database.name,
                    database_type=source_database_server.dbtype,
                    subscription_name=subscription.name,
                    error=e.message
                ))

    for user in webspace.database_users:
        source_database_server = get_database_server(user.database_server_id)
        dbtype = source_database_server.dbtype
        if not _is_dbtype_supported(dbtype):
            continue

        if is_local_database_host(source_database_server.host, [plesk_target_server.ip]):
            database_ext = database_ext_map.get(user.database_external_id)
            db_server_id = target_db_server.get(source_database_server.dbtype)
            if db_server_id:
                add_database_user = lambda login: \
                    _add_database_user(
                        plesk_api, login, user.password, database_ext.target_db_id if database_ext else None,
                        db_server_id, plesk_webspace_id, database.name, webspace.name
                    )
                try:
                    add_database_user(user.login)
                    webspace.database_user_map[user.login] = user.login
                except TryOtherDatabaseNameError:
                    max_name_len = settings.DatabaseServerTypeInfo[dbtype].user_name_len
                    new_user_login = get_hash_name_for_database_user2(max_name_len, webspace.web_source_host, source_database_server, user)
                    message = _T('DATABASE_USER_WITH_NAME_ALREADY_EXISTS').format(
                        user_name=user.login, server=plesk_target_server.description(), new_user_name=new_user_login
                    )
                    logger.debug(message)
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription.source_info.external_id,
                        Issue(Problem(None, Problem.NOTICE, message), None)
                    )
                    add_database_user(new_user_login)
                    webspace.database_user_map[user.login] = new_user_login
        else:
            databases = \
                [x for x in webspace.databases if x.external_id == user.database_external_id] or \
                [x for x in webspace.databases if x.database_server_id == source_database_server.id]
            try:
                database_server = plesk_target_server.get_database_server(source_database_server)
                if database_server is None:
                    raise MigrationError(_T("DATABASE_WAS_NOT_REGISTERED"))

                _register_database_user(
                    plesk_target_server, user, databases, database_server, source_database_server.login,
                    source_database_server.password, webspace.technical_name, subscription
                )
                webspace.database_user_map[user.login] = user.login
            except Exception as e:
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                errors.append(_T('FAILED_REGISTER_DB_USER_ERROR').format(
                    database_name=', '.join([x.name for x in databases]),
                    user_name=user.login,
                    database_type=source_database_server.dbtype,
                    subscription_name=subscription.name,
                    error=e.message
                ))


@trace
def import_ssl_certificates(webspace, plesk_target_server, errors):
    for site in webspace.sites:
        if site.hosting is None or not isinstance(site.hosting, PhysicalHostingModel) \
                or site.hosting.ssl_certificate is None:
            continue

        ssl_certificate = site.hosting.ssl_certificate
        certificates_on_site = plesk_target_server.plesk_api().send(
            plesk_ops.SslOperator.Get(
                plesk_ops.SslOperator.FilterByDomainName([site.name])
            )
        )[0].data

        if ssl_certificate.name in certificates_on_site:
            plesk_target_server.plesk_api().send(
                plesk_ops.SslOperator.Remove(
                    filter=plesk_ops.SslOperator.FilterByCertificateName([ssl_certificate.name]),
                    webspace=site.name
                )
            )

        with plesk_target_server.plesk_api().logger.censor({'xml': [
            "/packet/certificate/install/content/csr",
            "/packet/certificate/install/content/pvt",
            "/packet/certificate/install/content/cert",
            "/packet/certificate/install/content/ca",
        ]}):
            result = plesk_target_server.plesk_api().send(
                plesk_ops.SslOperator.Install(
                    name=ssl_certificate.name,
                    webspace=site.name,
                    csr='' if ssl_certificate.csr is None else ssl_certificate.csr,
                    pvt='' if ssl_certificate.private_key is None else ssl_certificate.private_key,
                    cert='' if ssl_certificate.cert is None else ssl_certificate.cert,
                    ca='' if ssl_certificate.ca is None else ssl_certificate.ca,
                )
            )

        if isinstance(result, Failure):
            errors.append(_T("UNABLE_TO_IMPORT_SSL_CERTIFICATE_AND_USE").format(
                certificate_name=ssl_certificate.name,
                site_name=site.name,
            ))
            continue

        result = plesk_target_server.plesk_api().send(
            plesk_ops.SiteOperator.Set(
                filter=plesk_ops.SubscriptionOperator.FilterByName([
                    site.name
                ]),
                hosting=plesk_ops.SubscriptionHostingVirtual(
                    properties=[
                        ('certificate_name', ssl_certificate.name),
                    ],
                    ip_addresses=None
                )
            )
        )
        if isinstance(result[0], Failure):
            errors.append(_T("UNABLE_TO_USE_SSL_CERTIFICATE").format(
                certificate_name=ssl_certificate.name,
                site_name=site.name,
            ))


@trace
def import_application_pools(plesk_target_server, site):
    params = [
        'managed_pipeline_mode', 'max_worker_processes', 'idle_timeout', 'idle_timeout_action',
        'cpu_usage', 'cpu_usage_action', 'cpu_usage_interval',
        'recycling_by_time', 'recycling_by_requests', 'recycling_by_virtual_memory', 'recycling_by_private_memory'
    ]

    param_values = []
    for name in params:
        value = getattr(site.hosting.application_pool, name)
        if value:
            param_values.extend(['-'+name, value])

    PleskCLIRunnerCLI(plesk_target_server).run(
        'site.exe',
        ['--set-iis-app-pool-settings', site.name.encode('idna')] + param_values,
        env=dict(PLESK_RESTORE_MODE='1')
    )


@trace
def import_mailbox_aliases(site_name, mailbox, plesk_target_server):
    site_id = get_plesk_site_id(site_name, plesk_target_server.plesk_api())

    result = plesk_target_server.plesk_api().send(
        plesk_ops.MailOperator.GetMailboxAliases(
            filter=plesk_ops.MailOperator.ComplexFilter(
                plesk_ops.MailOperator.FilterBySiteId([site_id]),
                plesk_ops.MailOperator.FilterByName([mailbox.name])
            )
        )
    )
    if isinstance(result, Failure):
        return False

    if result[0].data:
        result = plesk_target_server.plesk_api().send(
            plesk_ops.MailOperator.Update.Remove(
                site_id=site_id,
                name=mailbox.name,
                aliases=result[0].data
            )
        )
        if isinstance(result, Failure):
            return False

    if mailbox.alias and mailbox.alias.addresses:
        result = plesk_target_server.plesk_api().send(
            plesk_ops.MailOperator.Update.Set(
                site_id=site_id,
                name=mailbox.name,
                aliases=mailbox.alias.addresses,
                quota=None
            )
        )
        if isinstance(result, Failure):
            return False

    return True
