# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import posixpath

from pipes import quote

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.session import config
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.rsync import create_rsync_command

logger = logging.getLogger(__name__)


class CopyMailRsync(object):
    """Copy mail messages with rsync to target Unix server"""

    @trace
    def copy_mail(self, global_context, webspace):
        for site in webspace.sites:
            if not site.mailboxes:
                logger.debug(_T("SKIP_COPYING_MAIL_CONTENT_FOR_DOMAIN_2"), site.name)
                continue

            self._copy_domain_mail(global_context, site, webspace.mail_source_host, webspace.mail_target_host)
            self._repair_file_permissions(site, webspace.mail_target_host)

    @trace
    def _copy_domain_mail(self, global_context, site, source_host, target_host):
        rsync_additional_args = config.read_rsync_additional_args(global_context.config)
        ssh_key_info = global_context.ssh_key_pool.get(source_host, target_host)

        for mailbox in site.mailboxes:
            target_path = posixpath.join(target_host.mail_dir, mailbox.target_path)
            cmd, args = create_rsync_command(
                ssh_key_info.key_pathname, None, source_host.user(), source_host.ip(),
                quote(mailbox.source_path), target_path, [u"maildirsize", u".qmail*", u"@attachments/"],
                rsync_additional_args, '/usr/bin/rsync'
            )
            self._run_rsync(source_host, target_host, cmd, args)

    @staticmethod
    @trace
    def _run_rsync(source_host, target_host, cmd, args):
        try:
            with target_host.runner() as runner_target:
                runner_target.run(cmd, args)
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise MigrationError(
                _T("RSYNC_FAILED_COPY_MAIL_CONTENT_FROM") % (
                    source_host.description(), target_host.description(), str(e)
                )
            )

    @staticmethod
    @trace
    def _repair_file_permissions(site, target_host):
        with target_host.runner() as runner:
            target_path = posixpath.join(
                target_host.mail_dir,
                site.name.encode('idna').lower()
            )
            path_args = dict(path=target_path)
            runner.sh('chown popuser:popuser -R {path}', path_args)
            runner.sh('chmod 750 {path}', path_args)
