# (C)2019 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.checks.base import CheckAbstract
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.transfer.content_migrator import get_content_migrator
from migrator.core.utils.registry import Registry
from migrator.core.utils.loggers.log import trace
from migrator.odin_automation.platform.lib.api.aps.host import get_oa_host_id
from migrator.odin_automation.platform.lib.connections.target_server import AutomationServer
from migrator.odin_automation.platform.lib.db.service_node import get_oa_hostname
from migrator.odin_automation.services.shm.lib.aps.api import get_shm_application_nodes
from migrator.products.oa.lib.api.db.database import get_subscriptions_databases
from migrator.settings import SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, HTML_EOL

logger = logging.getLogger(__name__)


class Check(object):
    MAX_HOSTS_NAMES_TO_SHOW = 5

    @classmethod
    def get_os(cls):
        raise NotImplementedError()

    @classmethod
    def get_source_hosting_class(cls):
        raise NotImplementedError()

    @classmethod
    def get_panel_type(cls):
        raise NotImplementedError()

    @classmethod
    def get_automation_server(cls, conn, id=None, ip=None):
        raise NotImplementedError()

    @classmethod
    def get_content_item_model(cls):
        raise NotImplementedError()

    @classmethod
    def get_rsync_port(cls):
        raise NotImplementedError()

    @classmethod
    def get_could_not_connect_to_source_hosts_msg(cls):
        raise NotImplementedError()

    def run(self):
        global_context = Registry.get_instance().get_context()
        try:
            # Check source nodes and target Plesks availability via HCL
            unavailable_source_hosts = self._get_unavailable_source_hosts()
            unavailable_source_hosts_db = self._get_unavailable_source_hosts_db()
            error_message_parts = []
            if len(unavailable_source_hosts) > 0:
                error_message_parts.append(
                    self._compose_part_of_unavailable_hosts_msg(
                        unavailable_source_hosts,
                        self.get_could_not_connect_to_source_hosts_msg()
                    )
                )

            if len(unavailable_source_hosts_db) > 0:
                error_message_parts.append(
                    self._compose_part_of_unavailable_hosts_msg(
                        unavailable_source_hosts_db,
                        'CHECK_OA_SH_AVAILABILITY_ERRORS_PART_DB_HOSTS'
                    )
                )

            if len(error_message_parts) > 0:
                return CheckAbstract.STATUS_ERROR, _T("CHECK_OA_SH_AVAILABILITY_ERRORS_SUMMARY").format(
                    message_content=HTML_EOL + HTML_EOL.join(error_message_parts) + HTML_EOL
                )

            # Do a fake file transferring between all source nodes and target WHP Plesks
            conn = Registry.get_instance().get_context().conn.target
            source_hosts_to_check = PlatformReaderDB().get_migration_required_source_hosts(self.get_panel_type())
            target_whp_nodes = get_shm_application_nodes()
            error_message_parts = []

            for source_host_to_check in source_hosts_to_check:
                for target_whp_node in target_whp_nodes:
                    if target_whp_node['os'].lower() != self.get_os():
                        continue

                    target_whp_node_id = get_oa_host_id(target_whp_node['uid'])
                    try:
                        source_node = self.get_automation_server(conn, source_host_to_check.external_id, None)
                        target_plesk_node = AutomationServer.get(conn, target_whp_node_id, None)

                        with get_content_migrator(global_context, None, source_node, target_plesk_node, None, True) as content_migrator:
                            content_migrator.copy_content([self.get_content_item_model()])

                    except Exception:
                        logger.error(_T("LOG_EXCEPTION"), exc_info=True)
                        error_message_parts.append(
                            _T('HOSTS_PAIR_WITHOUT_RSYNC_CONNECTION').format(
                                host1=source_host_to_check.name + self.get_rsync_port(),
                                host2=get_oa_hostname(target_whp_node_id)
                            )
                        )

            if len(error_message_parts) > 0:
                hosts_pairs_amount = len(error_message_parts)
                if hosts_pairs_amount <= self.MAX_HOSTS_NAMES_TO_SHOW:
                    message_content = HTML_EOL + HTML_EOL.join(error_message_parts) + HTML_EOL
                else:
                    message_content = HTML_EOL + HTML_EOL.join(error_message_parts[:self.MAX_HOSTS_NAMES_TO_SHOW]) + \
                        HTML_EOL + _T('MSG_PART_FOR_OTHER_ENTITIES_EXCEPT_LISTED').format(
                            amount=hosts_pairs_amount - self.MAX_HOSTS_NAMES_TO_SHOW
                        ) + HTML_EOL

                return CheckAbstract.STATUS_ERROR, _T("CHECK_OA_SH_HOSTS_AVAILABILITY_ERRORS").format(
                    message_content=message_content
                )
            else:
                return CheckAbstract.STATUS_SUCCESS, _T("CHECK_OA_PANEL_AVAILABILITY_SUCCESS")
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return CheckAbstract.STATUS_ERROR, _T("CHECK_OA_PANEL_AVAILABILITY_UNEXPECTED_RESULT")
        finally:
            global_context.ssh_key_pool.remove_all()
            global_context.rsync_pool.stop_all()

    def _compose_part_of_unavailable_hosts_msg(self, unavailable_hosts_names_list, msg_part_id):
        """
        :param unavailable_hosts_names_list: list[str] Example: [u'HOST-1.AQA.INT.ZONE', u'HOST-2.AQA.INT.ZONE']
        :param msg_part_id: str Example: "CHECK_OA_WHL_AVAILABILITY_ERRORS_PART_APACHE_HOSTS"
        :return: string
        """
        unavailable_hosts_amount = len(unavailable_hosts_names_list)
        if unavailable_hosts_amount <= self.MAX_HOSTS_NAMES_TO_SHOW:
            msg = ', '.join(unavailable_hosts_names_list)
        else:
            msg = '%s %s' % (
                ', '.join(unavailable_hosts_names_list[:self.MAX_HOSTS_NAMES_TO_SHOW]),
                _T('MSG_PART_FOR_OTHER_ENTITIES_EXCEPT_LISTED').format(
                    amount=unavailable_hosts_amount - self.MAX_HOSTS_NAMES_TO_SHOW
                )
            )

        return _T(msg_part_id).format(hosts=msg)

    @classmethod
    @trace
    def _get_unavailable_source_hosts(cls):
        """
        :return: list[]
        """
        unavailable_source_hosts = []
        conn = Registry.get_instance().get_context().conn.target
        source_hosts_to_check = PlatformReaderDB().get_migration_required_source_hosts(cls.get_panel_type())
        for host_to_check in source_hosts_to_check:
            try:
                cls.get_automation_server(conn=conn, id=host_to_check.external_id, ip=None)
            except Exception:
                unavailable_source_hosts.append(host_to_check.name)

        return unavailable_source_hosts

    @classmethod
    @trace
    def _get_unavailable_source_hosts_db(cls):
        """
        :return: list[]
        """
        unavailable_source_hosts_db = []
        conn = Registry.get_instance().get_context().conn.target
        platform_reader = PlatformReaderDB()
        migration_project_items = platform_reader.get_migration_project_items(
            SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[cls.get_source_hosting_class()]
        )

        subscriptions_to_migrate_external_ids = list()
        for migration_project_item in migration_project_items:
            subscription_to_migrate = platform_reader.get_subscription(
                primary_id=migration_project_item.subscription_id
            )
            subscriptions_to_migrate_external_ids.append(
                subscription_to_migrate.external_id
            )

        db_hosts_ids = set()
        for sub_db_info in get_subscriptions_databases(subscriptions_to_migrate_external_ids):
            db_hosts_ids.add(sub_db_info['database_host_id'])

        for db_host in db_hosts_ids:
            try:
                AutomationServer.get(conn=conn, id=db_host, ip=None)
            except Exception:
                unavailable_source_hosts_db.append(get_oa_hostname(db_host))

        return unavailable_source_hosts_db
