# (C)2019 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.checks.base import CheckAbstract, get_subscriptions_as_limited_string
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.openapi.service import \
    R_EMAIL_DISK, R_EMAIL_USERS, R_MAIL_TEMPL, R_PLESK_MAIL, R_EMAIL_DOMAIN, RL_QMAIL, RW_EXCHANGE, R_O365_DOMAINS, \
    RES_AVAILABILITY_RESULT_SOME_RESOURCES_NOT_FOUND, RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND, \
    RW_EXCHANGE_MAILBOXES, RW_EXCHANGE_RES_MAILBOXES, RW_EXCHANGE_MAILBOX_TEMPL
from migrator.settings import HTML_EOL, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP

logger = logging.getLogger(__name__)


class Check(object):
    PRE_CHECK_DOC_URL = '94138.htm'
    SH_POSSIBLE_MAIL_SERVICES = [
        RL_QMAIL, RW_EXCHANGE, RW_EXCHANGE_MAILBOXES, RW_EXCHANGE_RES_MAILBOXES, RW_EXCHANGE_MAILBOX_TEMPL,
        R_O365_DOMAINS
    ]
    WHP_MAIL_RESOURCES = [R_PLESK_MAIL, R_MAIL_TEMPL, R_EMAIL_USERS, R_EMAIL_DOMAIN, R_EMAIL_DISK]
    WHP_MAIL_RESOURCE_MAIN = R_PLESK_MAIL

    @classmethod
    def get_source_hosting_class(cls):
        raise NotImplementedError()

    def run(self):
        try:
            source_subscriptions = PlatformReaderDB().get_migration_project_subscriptions()
            source_subscriptions_with_mail = {}
            for source_subscription in source_subscriptions:
                source_service_template = self.get_service_template(source_subscription.service_plan_external_id)
                if source_service_template.are_resources_exist(self.SH_POSSIBLE_MAIL_SERVICES)['status'] == RES_AVAILABILITY_RESULT_ALL_RESOURCES_FOUND:
                    source_subscriptions_with_mail[source_subscription.id] = source_subscription

            conflicts_data = {}
            wsh_hosting_class = SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[self.get_source_hosting_class()]
            for migration_project_item in PlatformReaderDB().get_migration_project_items(wsh_hosting_class):
                if migration_project_item.subscription_id in source_subscriptions_with_mail:
                    target_service_template = self.get_service_template(migration_project_item.target_service_plan_id)
                    if target_service_template.are_resources_exist([self.WHP_MAIL_RESOURCE_MAIN])['status'] == RES_AVAILABILITY_RESULT_SOME_RESOURCES_NOT_FOUND:
                        continue

                    if target_service_template.st_id not in conflicts_data:
                        conflicts_data[target_service_template.st_id] = {
                            'name': target_service_template.get_identificators(),
                            'subs': []
                        }

                    conflicts_data[target_service_template.st_id]['subs'].append(
                        source_subscriptions_with_mail[migration_project_item.subscription_id].name
                    )

            if not conflicts_data:
                return CheckAbstract.STATUS_SUCCESS, _T('CHECK_ST_CONFLICTS_SUCCESS')

            errors = []
            for conflict in conflicts_data.itervalues():
                errors.append(
                    _T('CHECK_ST_CONFLICTS_ERROR_MSG_PART').format(
                        st_description=conflict['name'],
                        conflicting_services=u', '.join(
                            [u"'%s'" % res.description for res in self.SH_POSSIBLE_MAIL_SERVICES]
                        ),
                        conflicting_subs=get_subscriptions_as_limited_string(conflict['subs'])
                    )
                )

            return CheckAbstract.STATUS_ERROR, _T('CHECK_ST_CONFLICTS_ERROR_MAIN_MSG').format(
                whp_mail_main_resource=self.WHP_MAIL_RESOURCE_MAIN.description,
                msg_content=HTML_EOL + HTML_EOL.join(errors) + HTML_EOL,
                whp_mail_resources=u', '.join(u"'%s'" % res.description for res in self.WHP_MAIL_RESOURCES),
                precheck_doc_link=self.PRE_CHECK_DOC_URL
            )
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return CheckAbstract.STATUS_ERROR, _T('CHECK_ST_CONFLICTS_ERROR_UNEXPECTED')

    def get_service_template(self, service_plan_external_id):
        """
        :param service_plan_external_id: int|str
        :return: ServiceTemplateDetails
        """
        global_context = Registry.get_instance().get_context()
        if not global_context.target_panel_obj.is_billing_module_installed():
            st_id = service_plan_external_id
        else:
            target_panel_billing_api = global_context.target_panel_obj.get_billing_api(global_context)
            st_id = target_panel_billing_api.plan_details_get(service_plan_external_id).st_id

        return global_context.conn.target.poa_api().service.get_service_template(st_id, True)
