# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import InternalError
from migrator.core.utils.registry import Registry
from migrator.products.oa.lib.api.db.site import get_site_names
from migrator.products.oa.lib.api.db.webspace import get_webspace_prot_dirs

_site_docroots_map = {
    '': (
        ('httpdocs', ''),
    ),
    'httpDocRoot': (
        ('httpdocs', ''),
        ('httpsdocs', 'migrated_httpsdocs/')
    ),
    'httpsDocRoot': (
        ('httpsdocs', ''),
        ('httpdocs', 'migrated_httpdocs/')
    )
}


def get_site_source_webroot_location(separate_ssl):
    default_src_docroot = _get_default_src_docroot()
    if not separate_ssl or default_src_docroot == 'httpDocRoot':
        return 'httpdocs'
    else:
        return 'httpsdocs'


def get_site_content_webroot_map(separate_ssl):
    if not separate_ssl:
        return _site_docroots_map['']

    return _site_docroots_map[_get_default_src_docroot()]


def _get_default_src_docroot():
    global_context = Registry.get_instance().get_context()
    default_src_docroot = global_context.migration_project.settings['defaultDocRoot']
    if default_src_docroot not in _site_docroots_map:
        raise InternalError('Unsupported migration project defaultDocRoot property value: %s' % default_src_docroot)

    return default_src_docroot


def get_webspace_protected_dirs(webspace_id):
    """
    Retrieve webspace protected dirs and group them by site name.

    In WHL all protected dirs are belong to a webspace.
    Site protected dirs are just webspace ones with site name in the dirs name:
      ws_prot_dir/ - webspace protected dir
      site.tld/site_prot_dir/ - site 'site.tld' protected dir

    So, we need to analyze all webspace protected dirs and group them by site name.
    NOTE: webspace own dirs will be grouped by '' (empty string) site name.

    Return value example:
    {
        ('', u'qwe', u'ssl'): {
            'header': u'qweqwe',
            'users': []
        },
        (u'http-https-domain.tld', u'321', u'nossl'): {
            'header': u'http-https-domain.tld 321',
            'users': [
                (u'someuser123', u'$1$YSrocQ8R$JLGDIndzbyXwLAEBqSU8d1'),
            ]
        },
        (u'http-https-domain.tld', u'toremove', u'nossl'): {
            'header': u'toremove',
            'users': []
        },
        ('', u'qwe', u'nossl'): {
            'header': u'sdfdfdf',
            'users': []}
        }
    }

    :param webspace_id: int
    :return: Dict
    """
    sites = get_site_names(webspace_id)
    pds_raw = get_webspace_prot_dirs(webspace_id)

    prot_dirs = dict()
    for pd_raw in pds_raw:
        pd_name = pd_raw['name'].strip('/')
        pd_name_parts = pd_name.split('/', 1) + ['']
        if pd_name_parts[0] in sites:
            # actually, it's a site protected dir
            pd_key, pd_info = _get_pd(pd_name_parts[0], pd_name_parts[1], pd_raw)
        else:
            # webspace own protected dir
            pd_key, pd_info = _get_pd('', pd_name, pd_raw)

        if pd_key in prot_dirs:
            prot_dirs[pd_key]['users'] += pd_info['users']
        else:
            prot_dirs[pd_key] = pd_info

    return prot_dirs


def _get_pd(site_name, pd_name, pd_raw):
    users = []
    if pd_raw['username']:
        users.append((pd_raw['username'], pd_raw['password']))

    pd_key = (site_name, pd_name, pd_raw['dir_type'].lower())
    pd_info = {
        'header': pd_raw['header'],
        'users': users
    }
    return pd_key, pd_info
