# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import os

from migrator.core.data.format.web_content_shm import ShmWebContentReader
from migrator.core.data.model.web_content_shm import \
    SubscriptionModel, WebspaceModel, SiteModel, ContentItemModel, OwnerModel, PermissionsModel
from migrator.core.utils import cached
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.operating_system.unix_utils import getent
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.db import OaServiceName
from migrator.products.oa.whl.lib.api.db.host import get_apache_host
from migrator.products.oa.lib.api.db.site import get_site_ids, get_site, get_applications
from migrator.products.oa.lib.api.db.webspace import get_webspace_ids, get_webspace_hosting_settings
from migrator.products.oa.whl.lib.api.db.webspace import get_webspace_root
from migrator.products.oa.whl.lib.export import get_site_source_webroot_location, get_site_content_webroot_map
from migrator.settings import DEFAULT_PLESK_SITE_PERMISSION

logger = logging.getLogger(__name__)

_sysgroups_map = {
    'pemsrv': 'psaserv',
    'pemcln': 'psacln'
}


class WHLContentExporter(ShmWebContentReader):
    @trace
    def get_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: SubscriptionModel
        """
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id)
        )

    @trace
    def get_webspaces_list(self, subscription_id):
        """
        :param subscription_id: int
        :return: list[WebspaceModel]
        """
        webspaces = []
        for webspace_id in get_webspace_ids(subscription_id, OaServiceName.APACHE_NG):
            webspaces.append(self.get_webspace(subscription_id, webspace_id))

        return webspaces

    @trace
    def get_webspace(self, subscription_id, webspace_id):
        """
        :param subscription_id: int
        :param webspace_id: int
        :return: WebspaceModel
        """
        # 1. Determine webspace system user and pem groups names and ids
        owners = self._get_content_owners(subscription_id, webspace_id)

        # 2. Generate list of content items to migrate
        sites = self.get_sites_list(subscription_id, webspace_id)

        # Sites are subdirs in the webspace httpdocs/httpsdocs folder. We need to exclude them from the webspace content
        exclude_paths = []
        is_webspace_content = True
        for site in sites:
            site_info = get_site(site.external_id)
            site_wwwroot = site_info['path_postfix']
            exclude_paths.append(site_wwwroot)
            if site_wwwroot == '':
                is_webspace_content = False

        if is_webspace_content:
            content = [
                ContentItemModel(
                    source_path=os.path.join(self._get_webspace_root(webspace_id), 'httpdocs'),
                    target_path='${WEBSPACEROOT}/',
                    exclude=exclude_paths,
                    permissions=PermissionsModel(owners)
                ),
            ]

            separate_ssl = bool(cached(get_webspace_hosting_settings)(webspace_id)['separate_http_dirs'])
            if separate_ssl:
                content.append(
                    ContentItemModel(
                        source_path=os.path.join(self._get_webspace_root(webspace_id), 'httpsdocs'),
                        target_path='${WEBSPACEROOT}/',
                        exclude=exclude_paths,
                        permissions=PermissionsModel(owners)
                    )
                )
        else:
            content = []

        return WebspaceModel(
            external_id=webspace_id,
            sites=sites,
            content=content
        )

    @trace
    def get_sites_list(self, subscription_id, webspace_id):
        """
        :param subscription_id: int
        :param webspace_id: int
        :return: list[SiteModel]
        """
        sites = []

        for site_id in get_site_ids(webspace_id):
            sites.append(self.get_site(subscription_id, webspace_id, site_id))

        return sites

    @trace
    def get_site(self, subscription_id, webspace_id, site_id):
        content = []
        exclude_paths = ['.protect']

        # 1. Determine webspace system user and pem groups names and ids
        owners = self._get_content_owners(subscription_id, webspace_id)

        # 2. Generate list of content items to migrate
        site_info = get_site(site_id)
        separate_ssl = bool(cached(get_webspace_hosting_settings)(webspace_id)['separate_http_dirs'])

        webspace_root = self._get_webspace_root(webspace_id)

        for source_webroot, target_webroot in get_site_content_webroot_map(separate_ssl):
            webroot = os.path.join(webspace_root, source_webroot, site_info['path_postfix'])
            content.append(
                ContentItemModel(
                    source_path=os.path.join(webroot, ''),
                    target_path=os.path.join('${SITEROOT}', target_webroot),
                    is_wwwroot=True,
                    exclude=exclude_paths,
                    permissions=PermissionsModel(owners, DEFAULT_PLESK_SITE_PERMISSION)
                )
            )

        for site_app in get_applications(site_id):
            content.append(
                ContentItemModel(
                    source_path=os.path.join(webspace_root, site_app['app_path'], 'htdocs/'),
                    target_path=os.path.join('${SITEROOT}', site_app['app_url'])
                )
            )

        site_webroot = os.path.join(
            webspace_root,
            get_site_source_webroot_location(separate_ssl), site_info['path_postfix']
        )
        return SiteModel(
            external_id=site_id,
            name=site_info['name'],
            source_webroot=site_webroot,
            content=content
        )

    @cached
    @trace
    def _get_content_owners(self, subscription_id, webspace_id):
        global_context = Registry.get_instance().get_context()
        webspace_host_id = get_apache_host(webspace_id)['host_id']
        webspace_httpdocs_path = os.path.join(self._get_webspace_root(webspace_id), 'httpdocs')
        source_web_node = global_context.migrator._get_source_web_node(webspace_host_id, subscription_id)
        owners = []
        groups = []
        with source_web_node.runner() as runner:
            for raw_group in getent(runner, 'group', _sysgroups_map.keys()):
                group_info = raw_group.split(':')
                group_name = group_info[0]
                group_id = group_info[2]
                groups.append((group_name, group_id))

            stat_info = runner.stat(webspace_httpdocs_path)
            for group in groups:
                # Map WHL system groups names to Plesk ones but leaving original WHL ids
                owners.append(
                    OwnerModel('${SYSUSER}', stat_info.user_id, _sysgroups_map[group[0]], group[1])
                )
                owners.append(
                    OwnerModel('${SYSUSER}', stat_info.user, _sysgroups_map[group[0]], group[1])
                )

        return owners

    @staticmethod
    @cached
    @trace
    def _get_webspace_root(webspace_id):
        return get_webspace_root(webspace_id)
