# (C)2017 Ingram Micro Inc. all Rights Reserved
import enum
import inspect
import logging
import os

from itertools import chain

from migrator.core import InternalError
from migrator.core.data.format.web_hosting_shm import ShmWebHostingReader
from migrator.core.data.model.web_hosting_shm import \
    SubscriptionModel, WebspaceModel, SiteModel, PhysicalHostingModel, HostingPropertiesModel, ScriptingModel, \
    PHPSettingsModel, FtpuserModel, SystemUser, SslCertificateModel, ApplicationModel, IpAddressModel, \
    DatabaseServerModel, DatabaseModel, DatabaseUserModel, ProtectedDirectoryModel, PropertyModel, \
    ProtectedDirectoryUserModel
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.db import OaServiceName
from migrator.odin_automation.platform.lib.api.aps.domain import is_pa_domain_owned_by_provider
from migrator.odin_automation.platform.lib.utils.decrypt import decrypt_data

from migrator.products.oa.whl.lib.api.db.host import get_apache_host
from migrator.products.oa.lib.api.db.site import \
    get_site_name, get_site_ids, get_applications, is_exists_wildcard_subdomain
from migrator.products.oa.lib.api.db.user import get_sys_user, get_ftp_users
from migrator.products.oa.lib.api.db.webspace import \
    get_webspace_ids, get_webspace_name, get_webspace_ips, get_webspace_ssl_certificates, \
    get_webspace_hosting_settings
from migrator.products.oa.whl.lib.api.db.webspace import get_webspace_root
from migrator.products.oa.whl.lib.export import \
    get_webspace_protected_dirs, get_site_content_webroot_map, get_site_source_webroot_location
from migrator.products.oa.whl.lib.source_host import \
    get_source_host_actual_php_fcgi_version, get_source_host_custom_php_settings
from migrator.products.oa.lib.api.db.database import \
    get_database_servers, get_subscriptions_databases, get_subscription_database_users
from migrator.settings import XML_IP_SHARED, XML_IP_EXCLUSIVE

logger = logging.getLogger(__name__)


class ShIPType(enum.Enum):
    IP_TYPE_SHARED = 0
    IP_TYPE_EXCLUSIVE = 1
    IP_TYPE_EXCLUSIVE_AND_V6 = 2
    IP_TYPE_V6_ONLY = 3
    IP_TYPE_SHARED_V4_EXCLUSIVE_V6 = 4
    IP_TYPE_RENEW = 5
    IP_TYPE_MAX = 6


class WHLHostingExporter(ShmWebHostingReader):
    WHL_LIMITATIONS_URL = '94138.htm'

    @trace
    def get_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: SubscriptionModel
        """
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id)
        )

    @trace
    def get_webspaces_list(self, subscription_id):
        """
        :param subscription_id: int
        :return: list[WebspaceModel]
        """
        webspaces = []

        if is_exists_wildcard_subdomain(subscription_id, OaServiceName.APACHE_NG):
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_id,
                Issue(Problem(
                    'skipped-wildcard-subdomains',
                    Problem.NOTICE,
                    _T("SUBSCRIPTION_CONTAINS_WILDCARD_SUBDOMAINS")), None)
            )

        # Databases and DB users of subscription will be exported as part of the first webspace
        first_webspace_in_subscription = False

        for webspace_id in get_webspace_ids(subscription_id, OaServiceName.APACHE_NG):
            webspaces.append(self.get_webspace(subscription_id, webspace_id, not first_webspace_in_subscription))
            first_webspace_in_subscription = True

        return webspaces

    @trace
    def get_webspace(self, subscription_id, webspace_id, first_webspace_in_subscription=False):
        """
        :param subscription_id: int
        :param webspace_id: int
        :param first_webspace_in_subscription: boolean
        :return: WebspaceModel
        """
        databases = []
        database_users = []
        ftp_users = []

        global_context = Registry.get_instance().get_context()

        if first_webspace_in_subscription:
            for db_data in get_subscriptions_databases([subscription_id]):
                databases.append(
                    DatabaseModel(
                        external_id=db_data['external_id'],
                        database_server_id='{host_id}_{service_id}'.format(
                            host_id=db_data['database_host_id'], service_id=db_data['database_service_id']
                        ),
                        name=db_data['name'],
                    )
                )

            for db_user_data in get_subscription_database_users(subscription_id):
                database_users.append(
                    DatabaseUserModel(
                        login=db_user_data['login'],
                        password=decrypt_data(db_user_data['password']),
                        database_external_id=db_user_data['database_external_id'],
                        database_server_id='{host_id}_{service_id}'.format(
                            host_id=db_user_data['database_host_id'], service_id=db_user_data['database_service_id']
                        ),
                    )
                )

        ftp_users_with_truncated_logins = {}
        for ftp_user_data in get_ftp_users(webspace_id):
            if len(ftp_user_data['name']) > 15:
                ftp_user_name = ftp_user_data['name'][:15]
                ftp_users_with_truncated_logins[ftp_user_data['name']] = ftp_user_name
            else:
                ftp_user_name = ftp_user_data['name']

            ftp_user_data['name'] = ftp_user_name
            ftp_user_data['password'] = decrypt_data(ftp_user_data['password'])
            ftp_user_data.update({'read': True, 'write': True, 'home': '/', 'quota': -1})
            ftp_users.append(FtpuserModel(**ftp_user_data))

        if len(ftp_users_with_truncated_logins) > 0:
            ftp_users_truncation_details = ''
            for key, value in ftp_users_with_truncated_logins.iteritems():
                ftp_users_truncation_details += '\n%s => %s' % (key, value)

            logger.warning(
                _T("SH_LONG_FTP_USERNAME_WILL_BE_SHORTENED_DEBUG").format(logins_list=ftp_users_truncation_details)
            )
            logger.warning(ftp_users_truncation_details)
            global_context.execution_wrapper.add_subscription_issue(
                subscription_id,
                Issue(Problem(
                    'sh-long-ftp-usernames-were-shortened',
                    Problem.NOTICE,
                    _T("SH_LONG_FTP_USERNAME_WILL_BE_SHORTENED")), None)
            )

        sysuser_data = get_sys_user(webspace_id)
        return WebspaceModel(
            name='whp_%s' % get_webspace_name(webspace_id, OaServiceName.APACHE_NG),
            external_id=webspace_id,
            type='apache',
            system_user=SystemUser(**sysuser_data),
            ftp_users=ftp_users,
            web_users=[],
            ip_addresses=self._get_webspace_ips(webspace_id),
            protected_directories=self._get_webspace_protected_dirs(webspace_id),
            separate_ssl=bool(cached(get_webspace_hosting_settings)(webspace_id)['separate_http_dirs']),
            sites=self.get_sites_list(subscription_id, webspace_id),
            databases=databases,
            database_users=database_users,
            host_external_id=get_apache_host(webspace_id)['host_id'],
            hosting_type='virtual_hosting'
        )

    @trace
    def get_sites_list(self, subscription_id, webspace_id):
        """
        :param subscription_id: int
        :param webspace_id: int
        :return: list[SiteModel]
        """
        sites = []
        for site_id in get_site_ids(webspace_id):
            exported_site = self.get_site(subscription_id, webspace_id, site_id)
            if exported_site:
                sites.append(exported_site)

        return sites

    @trace
    def get_site(self, subscription_id, webspace_id, site_id):
        """
        :param subscription_id: int
        :param webspace_id: int
        :param site_id: int
        :return: SiteModel
        """
        site_name = get_site_name(site_id)
        if is_pa_domain_owned_by_provider(site_name.lower()):
            Registry.get_instance().get_context().execution_wrapper.add_subscription_issue(
                subscription_id,
                Issue(Problem(
                    'sh-site-with-provider-dns',
                    Problem.NOTICE,
                    _T('SH_SITE_WITH_APP_ON_ADMIN_DOMAIN').format(sitename=site_name, doc_link=self.WHL_LIMITATIONS_URL)
                ), None)
            )
            return None

        return SiteModel(
            name=site_name,
            external_id=site_id,
            status='active',
            is_main=False,
            hosting=self._get_site_hosting(subscription_id, webspace_id, site_id)
        )

    @trace
    def get_database_servers(self):
        database_servers = []
        data = get_database_servers()
        for db_server_data in data:
            database_servers.append(
                DatabaseServerModel(
                    id='{host_id}_{service_id}'.format(
                        host_id=db_server_data['database_host_id'], service_id=db_server_data['database_service_id']
                    ),
                    host=db_server_data['host'],
                    alt_hosts=[db_server_data['hostname'], db_server_data['service_ip']],
                    port=db_server_data['port'],
                    dbtype=db_server_data['dbtype'],
                    login=db_server_data['login'],
                    password=decrypt_data(db_server_data['password']),
                    default=0,
                    local=0
                )
            )

        return database_servers

    @trace
    def _get_site_hosting(self, subscription_id, webspace_id, site_id):
        site_name = get_site_name(site_id)
        host_id = get_apache_host(webspace_id)['host_id']
        ws_hsettings = cached(get_webspace_hosting_settings)(webspace_id)
        ws_protdirs = cached(get_webspace_protected_dirs)(webspace_id)

        # WHL webspace uses single SSL certificate for all its sites. Thus we will copy it to every site.
        ssl_certificate = self._get_webspace_ssl_certificate(webspace_id)
        if ssl_certificate:
            ssl_certificate.name = _T('SSL_CERTIFICATE_FOR_SITE').format(site_name=site_name)

        return PhysicalHostingModel(
            hosting_properties=self._get_hosting_properties(ws_hsettings),
            scripting=self._get_scripting(ws_hsettings, host_id),
            php_settings=self._get_php_settings(ws_hsettings, host_id),
            ssl_certificate=ssl_certificate,
            protected_directories=self._get_site_protected_dirs(site_name, ws_protdirs, ws_hsettings),
            virtual_directories=[],
            application_pool=None,
            applications=self._get_applications(webspace_id, site_id)
        )

    @trace
    def _get_webspace_ssl_certificate(self, webspace_id):
        """
        Return SSL Certificate w/o name

        :param webspace_id: int
        :return: SslCertificateModel or None
        """
        ssl_certs = get_webspace_ssl_certificates(webspace_id)
        if not ssl_certs:
            return None

        ssl_cert = SslCertificateModel('', '', '', '', '')
        for cert_raw in ssl_certs:
            if cert_raw['cert_type'] == 1:
                # Process SSL Certificate
                # SSL Certificate is combined with private key. We need to split them.
                cert_data = bytearray(cert_raw['cert_data']).decode('ascii')
                pkey_pos = cert_data.find('-----BEGIN RSA PRIVATE KEY-----')
                ssl_cert.cert = cert_data[:pkey_pos]
                ssl_cert.private_key = cert_data[pkey_pos:]
            else:
                # Process CA
                ssl_cert.ca = bytearray(cert_raw['cert_data']).decode('ascii')

        return ssl_cert

    @staticmethod
    @trace
    def _get_hosting_properties(ws_hsettings):
        """
        :param ws_hsettings: dict
        :return: HostingPropertiesModel
        """
        return HostingPropertiesModel(
            iis_app_pool=None,
            ftp_quota=None,
            errdocs=bool(ws_hsettings['erd_state']),
            wuscripts=None,
            at_domains=None,
            seo_redirect=None,
            webstat='webalizer' if bool(ws_hsettings['webstat_state']) else 'none',
            webstat_protected=True if bool(ws_hsettings['webstat_state']) else False,
            ssl=bool(ws_hsettings['ssl_state']),
            write_modify=None
        )

    @trace
    def _get_scripting(self, ws_hsettings, host_id):
        """
        :param ws_hsettings: dict
        :return: ScriptingModel
        """
        php_version = ws_hsettings['php_version']
        php_type = 'cgi'
        if ws_hsettings['path_to_php'] == 'fastcgi':
            php_version = get_source_host_actual_php_fcgi_version(host_id)
            php_type = 'fastcgi'

        php_handler_id = None
        if php_version:
            php_version_parts = php_version.split('.')
            php_handler_id = '%s-%s.%s' % (php_type, php_version_parts[0], php_version_parts[1])

        return ScriptingModel(
            php=True if php_version else False,
            perl=None,
            asp=False,
            asp_dot_net=False,
            asp_dot_net_version=None,
            php_handler_id=php_handler_id,
            cgi=bool(ws_hsettings['cgi_state']),
            cgi_mode=None,
            fastcgi=None,
            python=None,
            shell=None
        )

    @trace
    def _get_php_settings(self, ws_hsettings, host_id):
        """
        :param ws_hsettings: dict
        :param host_id: int
        :return: PHPSettingsModel
        """
        if not ws_hsettings['path_to_php'] or ws_hsettings['path_to_php'] == 'fastcgi':
            return PHPSettingsModel()

        webspace_root = get_webspace_root(ws_hsettings['webspace_id'])
        custom_php_ini = '/'.join((webspace_root.rstrip('/'), '../php/php.ini'))

        php_settings = inspect.getargspec(PHPSettingsModel.__init__).args
        php_settings.remove('self')
        # 'open_basedir' and 'include_path' could not be correctly migrated -> skip
        php_settings.remove('open_basedir')
        php_settings.remove('include_path')

        for oam_key, ini_key in PHPSettingsModel.oam_vs_ini_mapping.iteritems():
            if oam_key in php_settings:
                # Apply mapping to convert oam-compatible settings' names to ini-compatible settings' names
                php_settings.append(ini_key.replace('.', '\\.'))
                php_settings.remove(oam_key)

        result_php_settings = get_source_host_custom_php_settings(host_id, custom_php_ini, tuple(php_settings))
        ini_vs_oam_mapping = {v: k for k, v in PHPSettingsModel.oam_vs_ini_mapping.iteritems()}
        for ini_key, oam_key in ini_vs_oam_mapping.iteritems():
            if ini_key in result_php_settings:
                # Apply mapping to convert settings' names to oam-compatible settings' names for model initialization
                result_php_settings[oam_key] = result_php_settings.pop(ini_key)

        return PHPSettingsModel(**result_php_settings)

    @trace
    def _get_applications(self, webspace_id, site_id):
        """
        :param site_id: int
        :return: list[ApplicationModel]|None
        """
        applications = get_applications(site_id)
        if not applications:
            return None

        webspace_root = get_webspace_root(webspace_id)
        return [
            ApplicationModel(
               name=app['app_name'],
               source_path=os.path.join(webspace_root, app['app_path'], 'htdocs'),
               target_path=os.path.join('${SITEROOT}', app['app_url'])
            ) for app in applications
        ]

    @staticmethod
    @trace
    def _get_webspace_ips(webspace_id):
        ipv4, ipv6, ip_type_raw = get_webspace_ips(webspace_id)[0]
        if ip_type_raw < ShIPType.IP_TYPE_SHARED.value or ip_type_raw >= ShIPType.IP_TYPE_RENEW.value:
            raise InternalError('Unsupported WHL IP address type for webspace %s: %s' % (webspace_id, ip_type_raw))

        ws_ips = []

        ip_type = ShIPType(ip_type_raw)
        if ip_type == ShIPType.IP_TYPE_SHARED:
            ws_ips.append(IpAddressModel(XML_IP_SHARED, ipv4, 'any'))
        elif ip_type == ShIPType.IP_TYPE_EXCLUSIVE:
            ws_ips.append(IpAddressModel(XML_IP_EXCLUSIVE, ipv4, 'any'))
        elif ip_type == ShIPType.IP_TYPE_EXCLUSIVE_AND_V6:
            ws_ips.append(IpAddressModel(XML_IP_EXCLUSIVE, ipv4, 'any'))
            ws_ips.append(IpAddressModel(XML_IP_EXCLUSIVE, ipv6, 'any'))
        elif ip_type == ShIPType.IP_TYPE_SHARED_V4_EXCLUSIVE_V6:
            ws_ips.append(IpAddressModel(XML_IP_SHARED, ipv4, 'any'))
            ws_ips.append(IpAddressModel(XML_IP_EXCLUSIVE, ipv6, 'any'))
        elif ip_type == ShIPType.IP_TYPE_V6_ONLY:
            ws_ips.append(IpAddressModel(XML_IP_EXCLUSIVE, ipv6, 'any'))

        return ws_ips

    @trace
    def _get_webspace_protected_dirs(self, webspace_id):
        def _merge_pd_usesrs(pd_users1, pd_users2):
            res = []
            uniq_logins = set()
            for pd_user in chain(pd_users1, pd_users2):
                if pd_user.login not in uniq_logins:
                    res.append(pd_user)
                    uniq_logins.add(pd_user.login)

            return res

        result = []
        just_ws_protdirs = dict()

        ws_protdirs = cached(get_webspace_protected_dirs)(webspace_id)
        for pd_key in ws_protdirs:
            # Look for webspace own prot dirs. See get_webspace_protected_dirs() description for details.
            if pd_key[0] != '':
                continue

            # Group pd_keys by prot dir name to find protected dirs which are present on both https & httpsdocs.
            # In WHL such the dirs are separate dirs with different dir_type (ssl & nossl)
            dir_name = pd_key[1]
            if dir_name in just_ws_protdirs:
                just_ws_protdirs[dir_name].append(pd_key)
            else:
                just_ws_protdirs[dir_name] = [pd_key]

        for pd_keys in just_ws_protdirs.values():
            if len(pd_keys) > 1:
                # Merge two protected dirs with the same name but different dir_type into single protected dir
                # with both 'ssl' and 'nonssl' properties set to 'true'. This is needed to create the dir on both
                # httpdocs & httpsdocs. Consecutive creation of the same dir with different properties will not work.
                pd_key1, pd_key2 = pd_keys
                pd_model1 = self._get_pd_model(pd_key1, ws_protdirs)
                pd_model2 = self._get_pd_model(pd_key2, ws_protdirs)

                pd_model1.properties = [PropertyModel('ssl', 'true'), PropertyModel('nonssl', 'true')]
                pd_model1.pd_users = _merge_pd_usesrs(pd_model1.pd_users, pd_model2.pd_users)

                result.append(pd_model1)
            else:
                result.append(self._get_pd_model(pd_keys[0], ws_protdirs))

        return result

    @trace
    def _get_site_protected_dirs(self, site_name, ws_protdirs, ws_hsettings):
        result = []

        separate_ssl = bool(ws_hsettings['separate_http_dirs'])
        if separate_ssl:
            source_docroot = get_site_source_webroot_location(separate_ssl)
            docroots_map = get_site_content_webroot_map(separate_ssl)
            path_prefix = [x[1] for x in docroots_map if x[1]].pop()

        for pd_key in ws_protdirs:
            if pd_key[0] != site_name:
                continue

            pd_model = self._get_pd_model(pd_key, ws_protdirs)
            if separate_ssl:
                # Non-primary docroot protected dirs will be renamed and exported as 'migrated_http(s)docs/<dir_name>'
                if (source_docroot == 'httpdocs' and pd_key[1] == 'ssl') \
                        or (source_docroot == 'httpsdocs' and pd_key[2] != 'ssl'):
                    pd_model.name = ''.join((path_prefix, pd_model.name))

            result.append(pd_model)

        return result

    @staticmethod
    @trace
    def _get_pd_model(pd_key, ws_protdirs):
        pd = ws_protdirs[pd_key]

        pd_users = []
        if pd['users']:
            pd_users = [ProtectedDirectoryUserModel(x[0], x[1], 'crypt') for x in pd['users']]

        return ProtectedDirectoryModel(pd_key[1], pd['header'], pd_users)
