# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import InternalError
from migrator.core.data.model.subscription import get_service_property
from migrator.core.data.model.web_hosting_shm import IpAddressModel
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.db import OaServiceName
from migrator.settings import \
    XML_IP_SHARED, XML_IP_EXCLUSIVE, SOURCE_SERVICE_WEB_WSH, IP_MIGRATION_STRATEGY_ONE, IP_MIGRATION_STRATEGY_ALL, \
    SERVICE_PROPERTY_IP_MIGRATION_STRATEGY
from migrator.odin_automation.platform.lib.db.ip import get_ip_address_id, IPType
from migrator.products.oa.wsh.lib.api.db.site import \
    get_sites_info_by_ip_address, get_site_name, get_technical_site_name
from migrator.products.oa.lib.api.db.webspace import get_webspace_ids

logger = logging.getLogger(__name__)


class WSHBaseExporter(object):
    def __init__(self):
        self.export_ip_addresses = False

    def get_webspaces_list(self, subscription_id):
        """
        :param subscription_id: int
        :return: list[WebspaceModel]
        """
        # Databases, DB users, FTP and web users of subscription will be exported as part of the first webspace
        closure = {'first_webspace_in_subscription': False}

        def _append_webspace_to_list(sub_id, ws_id, ipaddr, ipaddr_type, filter_sites_by_ip=True):
            ws_id = self._compose_webspace_id(ws_id, get_ip_address_id(ipaddr), filter_sites_by_ip)
            ws = self.get_webspace(sub_id, ws_id, not closure['first_webspace_in_subscription'])
            if self.export_ip_addresses:
                ws.ip_addresses = [IpAddressModel(type=ipaddr_type, ip=ipaddr, interface='any')]

            webspaces.append(ws)
            closure['first_webspace_in_subscription'] = True

        webspaces = []
        selected_ip_migration_strategy = self._choose_ip_migration_strategy(subscription_id)

        for webspace_id in get_webspace_ids(subscription_id, OaServiceName.WSH):
            sites_with_ips_list, is_technical_site_in_list = get_sites_info_by_ip_address(webspace_id)
            exclusive_ips = set()
            shared_ip_address = None
            if is_technical_site_in_list:
                # Webspace wwwroot should be migrated as website
                technical_site_name = get_technical_site_name(webspace_id)
                technical_site_name_for_migration = self._get_technical_site_name_for_migration(technical_site_name)

                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.add_subscription_issue(
                    subscription_id,
                    Issue(Problem(
                        'domain_name_will_be_generated_for_webspace_wwwroot_migration',
                        Problem.NOTICE,
                        _T("WSH_DOMAIN_WILL_BE_CHANGED_DURING_MIGRATION").format(
                            source_name=technical_site_name, webspace_id=webspace_id,
                            target_name=technical_site_name_for_migration
                        )), None)
                )

            for site in sites_with_ips_list:
                if site['ip_type'] == IPType.IPT_EXCLUSIVE_USED:
                    exclusive_ips.add(site['ip'])
                elif site['ip_type'] == IPType.IPT_SHARED:
                    shared_ip_address = site['ip']

            def _do_ip_migration_strategy_one():
                ip, ip_type = [shared_ip_address, XML_IP_SHARED] if len(exclusive_ips) == 0 \
                    else [min(exclusive_ips), XML_IP_EXCLUSIVE]
                _append_webspace_to_list(subscription_id, webspace_id, ip, ip_type, False)

            if selected_ip_migration_strategy == IP_MIGRATION_STRATEGY_ONE:
                _do_ip_migration_strategy_one()
            elif selected_ip_migration_strategy == IP_MIGRATION_STRATEGY_ALL:
                if len(exclusive_ips) <= 1:
                    _do_ip_migration_strategy_one()
                else:
                    if shared_ip_address:
                        _append_webspace_to_list(subscription_id, webspace_id, shared_ip_address, XML_IP_SHARED)

                    for ip in exclusive_ips:
                        _append_webspace_to_list(subscription_id, webspace_id, ip, XML_IP_EXCLUSIVE)

            else:
                raise InternalError("Unknown IP migration strategy '%s'" % selected_ip_migration_strategy)

        return webspaces

    def get_webspace(self, subscription_id, webspace_id, first_webspace_in_subscription=False):
        pass

    def get_sites_list(self, subscription_id, webspace_id):
        """
        :param subscription_id: int
        :param webspace_id: int
        :return: list[SiteModel]
        """
        real_webspace_id, ip_address_oa_id, filter_by_ip = self._read_webspace_id(webspace_id)

        sites = []
        sites_with_ips_list, is_technical_site_in_list = \
            get_sites_info_by_ip_address(real_webspace_id, ip_address_oa_id if filter_by_ip else None)
        for site in sites_with_ips_list:
            exported_site = self.get_site(subscription_id, real_webspace_id, site['site_id'])
            if exported_site:
                sites.append(exported_site)

        return sites

    def get_site(self, subscription_id, webspace_id, site_id):
        pass

    @classmethod
    def _get_site_name_for_migration(cls, site_id):
        site_name, is_technical_site = get_site_name(site_id)
        if is_technical_site:
            name = cls._get_technical_site_name_for_migration(site_name)
        else:
            name = site_name

        return name.decode('utf-8')

    @staticmethod
    def _get_technical_site_name_for_migration(site_name):
        return 'whp-%s' % site_name

    @staticmethod
    def _compose_webspace_id(webspace_id, ip_address_oa_id, filter_by_ip):
        return '%s__%s__%s' % (webspace_id, ip_address_oa_id, '1' if filter_by_ip else '0')

    @staticmethod
    def _read_webspace_id(webspace_id):
        if '__' not in str(webspace_id):
            webspace_id, modifier, filter_by_ip = webspace_id, None, False
        else:
            webspace_id, modifier, filter_by_ip = webspace_id.split('__')
            filter_by_ip = bool(int(filter_by_ip))

        return webspace_id, modifier, filter_by_ip

    @staticmethod
    def _choose_ip_migration_strategy(subscription_id):
        global_context = Registry.get_instance().get_context()
        platform_reader = PlatformReaderDB()
        db_subscription = platform_reader.get_subscription(subscription_id)

        project_ip_migration_strategy = global_context.migration_project.ipMigrationStrategy
        subscription_ip_migration_strategy = get_service_property(
            db_subscription, SOURCE_SERVICE_WEB_WSH, SERVICE_PROPERTY_IP_MIGRATION_STRATEGY
        )

        # Changing of IP migration strategy for migrated subscription that already created in OA is not allowed
        if db_subscription.oa_id:
            if not subscription_ip_migration_strategy:
                raise InternalError(("Database inconsistency found. IP Migration strategy for subscription '%s' "
                                     "is empty but it should not.") % subscription_id)

            selected_ip_migration_strategy = subscription_ip_migration_strategy
            if selected_ip_migration_strategy != project_ip_migration_strategy:
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.add_subscription_issue(
                    subscription_id,
                    Issue(Problem(
                        'ip-migration-strategy-cannot-be-changed-for-subscription-created-in-oa',
                        Problem.NOTICE,
                        _T("IP_MIGRATION_STRATEGY_CANNOT_BE_CHANGED_FOR_MIGRATED_SUBSCRIPTION")), None)
                )
        else:
            selected_ip_migration_strategy = project_ip_migration_strategy
            platform_reader.update_subscription_service_property(
                db_subscription.id,
                SOURCE_SERVICE_WEB_WSH,
                SERVICE_PROPERTY_IP_MIGRATION_STRATEGY,
                selected_ip_migration_strategy
            )

        return selected_ip_migration_strategy
