# (C)2017 Ingram Micro Inc. all Rights Reserved
import inspect
import logging
import ntpath

from migrator.core.data.format.web_hosting_shm import ShmWebHostingReader
from migrator.core.data.model.web_hosting_shm import \
    SubscriptionModel, WebspaceModel, SiteModel, PhysicalHostingModel, HostingPropertiesModel, ScriptingModel, \
    PHPSettingsModel, DatabaseServerModel, DatabaseModel, DatabaseUserModel, FtpuserModel, SystemUser, \
    SslCertificateModel, ApplicationModel, VirtualDirectoryModel
from migrator.core.language import _T
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.operating_system.application_pool import get_application_pool_settings, get_application_pool
from migrator.core.utils.registry import Registry
from migrator.odin_automation.platform.lib.api.aps.domain import is_pa_domain_owned_by_provider
from migrator.odin_automation.platform.lib.db import OaServiceName
from migrator.odin_automation.platform.lib.utils.decrypt import decrypt_data
from migrator.products.oa.wsh.lib.api.db.application import get_applications
from migrator.products.oa.lib.api.db.database import \
    get_subscription_database_users, get_database_servers, get_subscriptions_databases
from migrator.products.oa.wsh.lib.api.db.host import get_iis_host, get_iis_host_php_version, get_host_php_path
from migrator.products.oa.wsh.lib.api.db.service_template import get_subscription_php_engine
from migrator.products.oa.wsh.lib.api.db.site import \
    get_site_ssl_certificate, is_ssl_enabled_for_site, get_site_path, get_site_read_write_permissions, \
    get_virtual_dirs, get_virtual_dir_default_docs, get_site_postfix, get_sites_postfixes
from migrator.products.oa.wsh.lib.api.db.user import get_sys_user, get_web_users, get_ftp_users
from migrator.products.oa.lib.api.db.webspace import get_webspace_name
from migrator.products.oa.wsh.lib.api.db.webspace import \
    get_webspace_scripting_params, is_webspace_application_pool_exclusive, get_webspace_root
from migrator.products.oa.wsh.lib.export.web_base_shm import WSHBaseExporter
from migrator.products.oa.wsh.lib.source_host import AutomationServerIIS, get_source_host_php_settings

logger = logging.getLogger(__name__)


class WSHHostingExporter(WSHBaseExporter, ShmWebHostingReader):
    WSH_LIMITATIONS_URL = '94138.htm'

    def __init__(self):
        super(WSHHostingExporter, self).__init__()
        self.export_ip_addresses = True
        self.application_pool_settings = {}

    def get_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: SubscriptionModel
        """
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id)
        )

    def get_webspace(self, subscription_id, webspace_id, first_webspace_in_subscription=False):
        """
        :param subscription_id: int
        :param webspace_id: int
        :param first_webspace_in_subscription: boolean
        :return: WebspaceModel
        """
        databases = []
        database_users = []
        ftp_users = []

        global_context = Registry.get_instance().get_context()
        real_webspace_id, webspace_id_modifier, filter_by_ip = self._read_webspace_id(webspace_id)

        if first_webspace_in_subscription:
            for db_data in get_subscriptions_databases([subscription_id]):
                databases.append(
                    DatabaseModel(
                        external_id=db_data['external_id'],
                        database_server_id='{host_id}_{service_id}'.format(
                            host_id=db_data['database_host_id'], service_id=db_data['database_service_id']
                        ),
                        name=db_data['name'],
                    )
                )

            for db_user_data in get_subscription_database_users(subscription_id):
                database_users.append(
                    DatabaseUserModel(
                        login=db_user_data['login'],
                        password=decrypt_data(db_user_data['password']),
                        database_external_id=db_user_data['database_external_id'],
                        database_server_id='{host_id}_{service_id}'.format(
                            host_id=db_user_data['database_host_id'], service_id=db_user_data['database_service_id']
                        ),
                    )
                )

            web_users = get_web_users(subscription_id)

            if len(web_users) > 0:
                web_users_logins = [x['login'] for x in web_users]
                logger.warning(
                    _T("WSH_WEB_USERS_WILL_NOT_BE_RESTORED_LOG").format(
                        sub_id=subscription_id, list_with_logins=','.join(web_users_logins))
                )
                global_context.execution_wrapper.add_subscription_issue(
                    subscription_id,
                    Issue(Problem(
                        'wsh-web-users-will-not-be-migrated',
                        Problem.NOTICE,
                        _T("WSH_WEB_USERS_WILL_NOT_BE_RESTORED").format(id=subscription_id)), None)
                )

        ftp_users_with_truncated_logins = {}
        for ftp_user_data in get_ftp_users(real_webspace_id):
            if len(ftp_user_data['name']) > 15:
                ftp_user_name = ftp_user_data['name'][:15]
                ftp_users_with_truncated_logins[ftp_user_data['name']] = ftp_user_name
            else:
                ftp_user_name = ftp_user_data['name']

            ftp_user_data['name'] = ftp_user_name
            ftp_user_data['password'] = decrypt_data(ftp_user_data['password'])
            ftp_user_data.update({'read': True, 'write': True, 'home': '/', 'quota': -1})
            ftp_users.append(FtpuserModel(**ftp_user_data))

        if len(ftp_users_with_truncated_logins) > 0:
            ftp_users_truncation_details = ''
            for key, value in ftp_users_with_truncated_logins.iteritems():
                ftp_users_truncation_details += '\n%s => %s' % (key, value)

            logger.warning(
                _T("SH_LONG_FTP_USERNAME_WILL_BE_SHORTENED_DEBUG").format(logins_list=ftp_users_truncation_details)
            )
            logger.warning(ftp_users_truncation_details)
            global_context.execution_wrapper.add_subscription_issue(
                subscription_id,
                Issue(Problem(
                    'sh-long-ftp-usernames-were-shortened',
                    Problem.NOTICE,
                    _T("SH_LONG_FTP_USERNAME_WILL_BE_SHORTENED")), None)
            )

        sysuser_data = get_sys_user(real_webspace_id)
        sysuser_data['password'] = decrypt_data(sysuser_data['password'])
        if not first_webspace_in_subscription:
            sysuser_data['name'] += '__%s' % webspace_id_modifier

        return WebspaceModel(
            name='whp_%s_%s' % (webspace_id_modifier, get_webspace_name(real_webspace_id, OaServiceName.WSH)),
            external_id=webspace_id,
            type='iis',
            system_user=SystemUser(**sysuser_data),
            ftp_users=ftp_users,
            web_users=[],
            ip_addresses=[],
            sites=self.get_sites_list(subscription_id, webspace_id),
            databases=databases,
            database_users=database_users,
            host_external_id=get_iis_host(real_webspace_id)['host_id'],
            hosting_type='virtual_hosting'
        )

    def get_site(self, subscription_id, webspace_id, site_id):
        """
        :param subscription_id: int
        :param webspace_id: int
        :param site_id: int
        :return: SiteModel
        """

        def convert_buffer_to_str(buf):
            return (b"%s" % buf).decode("utf-8").replace('\r\n', '\n') if buf else ''

        site_name = self._get_site_name_for_migration(site_id)

        if is_pa_domain_owned_by_provider(site_name.lower()):
            Registry.get_instance().get_context().execution_wrapper.add_subscription_issue(
                subscription_id,
                Issue(Problem(
                    'sh-site-with-provider-dns',
                    Problem.NOTICE,
                    _T('SH_SITE_WITH_APP_ON_ADMIN_DOMAIN').format(sitename=site_name, doc_link=self.WSH_LIMITATIONS_URL)
                ), None)
            )
            return None

        ssl_cert_data = get_site_ssl_certificate(site_id)
        if ssl_cert_data:
            ssl_cert_data['private_key'] = convert_buffer_to_str(ssl_cert_data['private_key'])
            ssl_cert_data['cert'] = convert_buffer_to_str(ssl_cert_data['cert'])
            ssl_cert_data['ca'] = convert_buffer_to_str(ssl_cert_data['ca'])
            ssl_cert_data['csr'] = ''
            ssl_cert_data['name'] = _T("SSL_CERTIFICATE_FOR_SITE").format(site_name=site_name)
            ssl_certificate = SslCertificateModel(**ssl_cert_data)
        else:
            ssl_certificate = None

        host_id = get_iis_host(webspace_id)['host_id']

        return SiteModel(
            name=site_name,
            external_id=site_id,
            status='active',
            is_main=False,
            hosting=PhysicalHostingModel(
                hosting_properties=self._get_hosting_properties(site_id, webspace_id),
                scripting=self._get_scripting(subscription_id, webspace_id, host_id),
                php_settings=self._get_php_settings(site_id, host_id),
                ssl_certificate=ssl_certificate,
                protected_directories=[],
                virtual_directories=self._get_virtual_directories(subscription_id, webspace_id, site_id),
                application_pool=self._get_application_pool(webspace_id, host_id),
                applications=self._get_applications(webspace_id, site_id)
            )
        )

    def _get_virtual_directories(self, subscription_id, webspace_id, site_id):
        wwwroot = get_site_postfix(site_id)
        vdirs = get_virtual_dirs(site_id)
        other_site_prefixes = [
            x['site_postfix'] for x in get_sites_postfixes(webspace_id) if not x['site_id'] == site_id
        ]
        res = []
        for vdir in vdirs:
            default_docs = ','.join(get_virtual_dir_default_docs(vdir['dir_id']))
            location = vdir['desire_location']
            if location == wwwroot:
                location = '${SITEROOT}'
            elif location.startswith(wwwroot) and ntpath.dirname(location) not in other_site_prefixes:
                location = '${SITEROOT}\\' + location[len(wwwroot):]
            elif location.startswith('siteapps') and ntpath.dirname(location) not in other_site_prefixes:
                location = '${WEBSPACEROOT}\\' + location
            else:
                # We cannot migrate virtual dirs physically placed outside of site's wwwroot
                global_context = Registry.get_instance().get_context()
                global_context.execution_wrapper.add_subscription_issue(
                    subscription_id, Issue(Problem(
                        'virtual-directory',
                        Problem.NOTICE,
                        _T('UNABLE_TO_MIGRATE_THE_VIRTUAL_DIRECTORY').format(
                            site_name=self._get_site_name_for_migration(site_id),
                            vdir_name=vdir['desire_dir_name'],
                            vdir_path=vdir['desire_location'],
                        )
                    ), None)
                )
                continue

            res.append(VirtualDirectoryModel(
                name=vdir['dir_name'],
                path=location,
                virtual_directories=[],
                asp_enable_client_debug_inherited=None,
                asp_enable_client_debug_default=None,
                asp_enable_client_debug=False,  # Taken from IIS,
                asp_enable_server_debug_inherited=None,
                asp_enable_server_debug_default=None,
                asp_enable_server_debug=False,  # Taken from IIS,
                asp_buffering_limit_inherited=None,
                asp_buffering_limit_default=None,
                asp_buffering_limit=4194304,  # Taken from IIS
                asp_max_request_entity_allowed_inherited=None,
                asp_max_request_entity_allowed_default=None,
                asp_max_request_entity_allowed=200000,  # Taken from IIS
                editable=None,
                script_maps=None,  # php and asp.net settings are migrated separately
                require_ssl=False,  # Taken from IIS
                auth_ntlm=vdir['auth_windows'] > 0,
                auth_anonymous=vdir['auth_anonymous'] > 0,
                asp_execute_in_mta=False,  # Taken from IIS
                asp_enable_parent_paths=True,  # Taken from IIS
                default_doc=default_docs,
                enable_default_doc=True,  # Taken from IIS
                app_create=True,  # Taken from IIS
                dont_log=vdir['sec_log_visits'] == 0,
                enable_dir_browsing=vdir['sec_browse'] > 0,
                access_source=vdir['sec_script_src'] > 0,
                access_execute=vdir['sec_exec'] > 0,
                access_script=vdir['sec_script'] > 0,
                access_write=vdir['sec_write'] > 0,
                access_read=vdir['sec_read'] > 0
            ))

        return res

    def _get_applications(self, webspace_id, site_id):
        """
        :param site_id: int
        :return: list[ApplicationModel]|None
        """
        applications = get_applications(site_id)
        if not applications:
            return None

        webspace_root = get_webspace_root(webspace_id)['path']
        return [
            ApplicationModel(
                name=app['app_name'],
                source_path=ntpath.join(webspace_root, app['app_path']),
                target_path=ntpath.join('${WEBSPACEROOT}', app['app_path'])
            ) for app in applications
        ]

    def _get_hosting_properties(self, site_id, webspace_id):
        """
        :param site_id: int
        :param webspace_id: int
        :return: HostingPropertiesModel
        """
        read, write = get_site_read_write_permissions(site_id)

        return HostingPropertiesModel(
            iis_app_pool=is_webspace_application_pool_exclusive(webspace_id),
            ftp_quota=None,
            errdocs=True,
            wuscripts=None,
            at_domains=None,
            seo_redirect=None,
            webstat='none',
            webstat_protected=None,
            ssl=is_ssl_enabled_for_site(site_id),
            write_modify=read and write
        )

    def _get_scripting(self, subscription_id, webspace_id, host_id):
        """
        :param subscription_id: string Example: '1000001'
        :param webspace_id: int Example: 100000
        :param host_id: int Example: 4
        :return: ScriptingModel
        """

        handler_mode_status_map = {
            0: False,  # Off
            1: True,  # ISAPI
            2: True  # CGI/FastCGI
        }

        asp_dot_net_status_map = {
            0: False,
            11: True,
            20: True,
            40: True,
            45: True
        }

        asp_dot_net_version_map = {
            0: None,
            11: '1.1',
            20: '2.0',
            40: '4.0',
            45: '4.5'
        }

        site_scripting_main_params = get_webspace_scripting_params(webspace_id)[0]
        wsh_host_php_full_version = get_iis_host_php_version(host_id)[0]['php_version']
        php_version_parts = wsh_host_php_full_version.split('.')
        if len(php_version_parts) > 1:
            whp_php_version = "%s.%s" % (php_version_parts[0], php_version_parts[1])
        else:
            whp_php_version = "%s" % php_version_parts[0]

        php_handler_id = '%s-%s' % (
            get_subscription_php_engine(subscription_id)[0]['php_handler'].lower(),
            whp_php_version
        )
        return ScriptingModel(
            php=handler_mode_status_map[site_scripting_main_params['php_status']],
            perl=None,
            asp=handler_mode_status_map[site_scripting_main_params['asp_status']],
            asp_dot_net=asp_dot_net_status_map[site_scripting_main_params['aspdotnet_status']],
            asp_dot_net_version=asp_dot_net_version_map[site_scripting_main_params['aspdotnet_status']],
            php_handler_id=php_handler_id,
            cgi=False,
            cgi_mode=None,
            fastcgi=None,
            python=None,
            shell=None
        )

    def _get_php_settings(self, site_id, host_id):
        """
        :param site_id: int
        :param host_id: int
        :return: PHPSettingsModel
        """
        php_settings = inspect.getargspec(PHPSettingsModel.__init__).args
        php_settings.remove('self')
        for oam_key, ini_key in PHPSettingsModel.oam_vs_ini_mapping.iteritems():
            if oam_key in php_settings:
                # Apply mapping to convert oam-compatible settings' names to ini-compatible settings' names
                php_settings.append(ini_key.replace('.', '\\.'))
                php_settings.remove(oam_key)

        php_settings.append('user_ini.filename')

        iis_host_php_ini_path = '%s\\php.ini' % get_host_php_path(host_id)
        iis_host_global_php_settings = get_source_host_php_settings(host_id, iis_host_php_ini_path, tuple(php_settings))

        user_ini_filename = iis_host_global_php_settings.get('user_ini.filename', '.user.ini')
        website_php_ini_path = '%s\\%s' % (get_site_path(site_id), user_ini_filename)
        website_php_settings = get_source_host_php_settings(host_id, website_php_ini_path, tuple(php_settings),
                                                            throw_exception_if_file_absent=False)

        result_php_settings = iis_host_global_php_settings.copy()
        result_php_settings.update(website_php_settings)

        # 'open_basedir' could not be correctly migrated - skip
        # 'user_ini.filename' is not presented in PHPSettingsModel
        for name in ['user_ini.filename', 'open_basedir']:
            if name in result_php_settings:
                del result_php_settings[name]

        ini_vs_oam_mapping = {v: k for k, v in PHPSettingsModel.oam_vs_ini_mapping.iteritems()}
        for ini_key, oam_key in ini_vs_oam_mapping.iteritems():
            if ini_key in result_php_settings:
                # Apply mapping to convert settings' names to oam-compatible settings' names for model initialization
                result_php_settings[oam_key] = result_php_settings.pop(ini_key)

        return PHPSettingsModel(**result_php_settings)

    def get_database_servers(self):
        database_servers = []
        data = get_database_servers()
        for db_server_data in data:
            database_servers.append(
                DatabaseServerModel(
                    id='{host_id}_{service_id}'.format(
                        host_id=db_server_data['database_host_id'], service_id=db_server_data['database_service_id']
                    ),
                    host=db_server_data['host'],
                    alt_hosts=[db_server_data['hostname'], db_server_data['service_ip']],
                    port=db_server_data['port'],
                    dbtype=db_server_data['dbtype'],
                    login=db_server_data['login'],
                    password=decrypt_data(db_server_data['password']),
                    default=0,
                    local=0
                )
            )

        return database_servers

    def _get_application_pool(self, webspace_id, host_id):
        webspace_name = "webspace_%08d - root domain" % int(webspace_id)
        server = self._get_server(host_id)
        win_application_pools, win_sites = self._get_application_pool_settings(host_id, server)
        return get_application_pool(webspace_name, win_application_pools, win_sites, server)

    def _get_application_pool_settings(self, host_id, server):
        if host_id not in self.application_pool_settings:
            self.application_pool_settings[host_id] = get_application_pool_settings(server)

        return self.application_pool_settings[host_id]

    @staticmethod
    def _get_server(host_id):
        global_context = Registry.get_instance().get_context()
        conn = global_context.target_panel_obj.get_connections(global_context)
        return AutomationServerIIS(conn, id=host_id)
