# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""Tools to generate XML node describing account"""
import logging
import re

from DataSources import \
    AccountDataSource, ServiceTemplateDataSource, SubscriptionDataSource, DBDataSource, PaymentToolsSource, \
    XmlSubscriptionDataSource, DataSource
from AccountGeneratorBase import \
    AccountGeneratorBase, setDisableStateAttr, formatStrDatetime
from xml.dom.minidom import Element
from XMLExtensions import TextElement


logger = logging.getLogger(__name__)


class AccountGenerator(AccountGeneratorBase):
    def __init__(self, account_id_list=None, subscription_id_list=None, export_inactive_subscriptions=False):
        # data sources

        if isinstance(account_id_list, (list, tuple)) and '-1' in account_id_list:
            # We need to skip account data export for performance reasons if account with id '-1' in account_id_list
            # This is mostly for AgentImpl.getSubscriptionList() method
            self.accSource = None
        else:
            self.accSource = AccountDataSource(account_id_list)

        if subscription_id_list:
            self.subscriptionSource = SubscriptionDataSource(subscription_id_list, export_inactive_subscriptions)
        else:
            self.subscriptionSource = None

        if account_id_list and len(account_id_list) == 1 and '1' in account_id_list:
            # we must export all STs when called from PlatformGenerator.getNodes()
            serviceTemplateIdList = self._getServiceTemplateIds(None, None, export_inactive_subscriptions)
        else:
            serviceTemplateIdList = self._getServiceTemplateIds(
                account_id_list, subscription_id_list, export_inactive_subscriptions
            )

        self.serviceTemplateDataSource = ServiceTemplateDataSource(serviceTemplateIdList)
        self.paymentToolDataSource = PaymentToolsSource()
        self.hwNodes = {}

    def _getServiceTemplateIds(self, account_id_list, subscription_id_list, export_inactive_subscriptions):
        res = []
        stds = ServiceTemplateDataSource([])

        sts_types = [
            stds.TYPE_DOMAIN_REGISTRATION, stds.TYPE_PLESK_DOMAIN, stds.TYPE_PLESK_CLIENT,
            stds.TYPE_VIRTUOZZO_CONTAINER, stds.TYPE_VIRTUAL_MACHINE
        ]

        query = "SELECT DISTINCT(plan_id) hp_id FROM subscr_base sb " \
                "  LEFT JOIN bm_hp ON sb.plan_id = bm_hp.hp_id " \
                "WHERE bm_hp.type IN (%s) AND sb.status IN (%s)" % (
                    ','.join(sts_types),
                    ','.join(XmlSubscriptionDataSource.getSubscriptionStatusIdsToExport(export_inactive_subscriptions))
                )

        if subscription_id_list:
            # get hp ids by subscription ids
            query += " AND subscr_id IN (%s)" % ','.join(subscription_id_list)
        elif account_id_list:
            # get hp ids by owner ids
            query += " AND account_no IN (%s)" % ','.join(account_id_list)

        document = DBDataSource.getDomTreeData(query, stds.HOST_PLANS_FIELD)
        hostingPlans = document.getElementsByTagName(stds.HOST_PLANS_FIELD)
        for i in range(hostingPlans.length):
            curPlan = hostingPlans.item(i)
            res.append(stds._getChildNodeText(curPlan, stds.HP_ID_FIELD))

        return res

    def __repr__(self):
        return "AccountGenerator:\n- accSource: {accSource}\n- subscriptionSource: {subscriptionSource}\n- serviceTemplateDataSource: {serviceTemplateDataSource}\n- paymentToolDataSource: {paymentToolDataSource}\n- hwNodes: {hwNodes}".format(
            accSource=str(self.accSource),
            subscriptionSource=str(self.subscriptionSource),
            serviceTemplateDataSource=str(self.serviceTemplateDataSource),
            paymentToolDataSource=str(self.paymentToolDataSource),
            hwNodes=str(self.hwNodes)
        )

    def _makeSubscriptionNode(self, account, subscription):
        """Generate DOM node describing single subscription.

           account: SimpleTypes.AccountInfo describing subscription owner.
           subscription: SimpleTyps.SubscriptionInfo subscription info"""
        logger.debug("Generate DOM node for subscription '{subscription_name}'.".format(
            subscription_name=subscription.name.encode('utf-8')
        ))

        isDomainSubscription = subscription.type == XmlSubscriptionDataSource.TYPE_DM

        subscrNode = Element('subscription')
        subscription_name = subscription.name
        if isDomainSubscription or subscription.type == XmlSubscriptionDataSource.TYPE_PD:
            subscription_name = subscription.name.encode('idna').decode('idna')

        subscrNode.setAttribute("name", subscription_name)
        subscrNode.setAttribute("id", subscription.id)
        subscrNode.setAttribute("serviceTemplateId", subscription.serviceTemplateId)
        subscrNode.setAttribute("isDomainSubscription", "true" if isDomainSubscription else "false")
        setDisableStateAttr(not subscription.active, subscrNode)
        # add creation date
        subscrNode.appendChild(TextElement("creationDate", formatStrDatetime(subscription.startDate)))
        # add expiration date
        subscrNode.appendChild(TextElement("expirationDate", formatStrDatetime(subscription.endDate)))
        subscrNode.appendChild(TextElement("status", subscription.status))

        hwNode = self._makeHwNode(subscription)
        if hwNode is not None:
            subscrNode.appendChild(TextElement("hwId", str(hwNode.getAttribute("id"))))

        serviceTemplate = self.serviceTemplateDataSource.getEntityByID(subscription.serviceTemplateId)
        # VPS: Container (CT)
        if subscription.type == XmlSubscriptionDataSource.TYPE_VE:
            vpsSrc = DBDataSource.getData(DBDataSource.TYPE_VPS_CT + ' WHERE subscr_ve.subscr_id=%s' % subscription.id)
            vpsInfo = vpsSrc.getElements("subscr_id", subscription.id)[0]
            subscrNode.appendChild(TextElement("veId", str(vpsInfo["uuid"])))
            subscrNode.appendChild(TextElement("ctShortId", str(vpsInfo["ve_id"])))
            subscrNode.appendChild(TextElement("veName", unicode(vpsInfo["utf_ve_name"])))
            subscrNode.appendChild(TextElement("veHostname", unicode(vpsInfo["ve_name"])))
            subscrNode.appendChild(TextElement("veType", 'ct'))
            subscrNode.appendChild(TextElement("hwId", str(vpsInfo["hw_id"])))
            subscrNode.setAttribute("type", 'vps')
        # VPS: Virtual Machine (VM)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_VM:
            vpsSrc = DBDataSource.getData(DBDataSource.TYPE_VPS_VM + ' WHERE pserver_vm.subscr_id=%s' % subscription.id)
            vpsInfo = vpsSrc.getElements("subscr_id", subscription.id)[0]
            veName = unicode(vpsInfo["name"])
            veId = str(vpsInfo["vm_eid"])
            veHostname = veName if self._isVmHostnameValid(veName) else veId
            subscrNode.appendChild(TextElement("veId", veId))
            subscrNode.appendChild(TextElement("veName", veName))
            subscrNode.appendChild(TextElement("veHostname", veHostname))
            subscrNode.appendChild(TextElement("veType", 'vm'))
            subscrNode.appendChild(TextElement("hwId", str(vpsInfo["hw_id"])))
            subscrNode.setAttribute("type", 'vps')
        elif subscription.type == XmlSubscriptionDataSource.TYPE_DM:
            dmBaseSrc = DBDataSource.getData(DBDataSource.TYPE_DOMAIN_REG + ' WHERE subscr_domain.subscr_id=%s' % subscription.id)
            domainInfo = dmBaseSrc.getElements("subscr_id", subscription.id)[0]
            domainName = domainInfo["domain"]
            domainId = domainInfo["id"]
            dmUseSrc = DBDataSource.getData(DBDataSource.TYPE_DOMAIN_USE + ' WHERE id=%s' % domainId)
            dmNsSrc = DBDataSource.getData(DBDataSource.TYPE_DOMAIN_NS + ' WHERE domain_id = %s' % domainId)
            domainLinkToSubscription = dmUseSrc.getElements("domain", domainName)
            isDnsManaged = dmBaseSrc.getElements("domain", domainName)[0]["dns_enabled"]
            if len(domainLinkToSubscription) > 0:
                subscrNode.appendChild(TextElement("targetSubscriptionId", domainLinkToSubscription[0]["subscr_id"]))

            subscrNode.appendChild(TextElement("isDnsManaged", isDnsManaged))
            subscrNode.appendChild(TextElement("domainId", domainId))
            for i in range(len(serviceTemplate.tld)):
                curTdl = serviceTemplate.tld[i]
                if domainName[-len(curTdl):] == curTdl:
                    subscrNode.setAttribute("serviceTemplateId", subscription.serviceTemplateId + "_" + curTdl)
                    break

            nameServers = dmNsSrc.getElements("domain_id", domainId)
            subscrNode.appendChild(self._getNameServersNode(nameServers))

        self._createTimePeriod(subscrNode, "subscriptionPeriod", subscription.subscriptionPeriod)
        self._createTimePeriod(subscrNode, "billingPeriod", subscription.billingPeriod)

        # Multiplier for translation years into minutes.
        # Actual only for Domain Registration subscriptions.
        k = 1
        if isDomainSubscription:
            fees = serviceTemplate.tld_fees[curTdl]
            k = 12 * 30 * 24 * 3600
        else:
            fees = serviceTemplate.fees
        for i in range(len(fees)):
            curFee = fees[i]
            if int(subscription.billingPeriod) == int(curFee.period) * k:
                subscrNode.appendChild(TextElement("subscrFee", curFee.subscrFee))
                subscrNode.appendChild(TextElement("setupFee", curFee.setupFee))
                break

        return subscrNode

    def _isVmHostnameValid(self, vm_hostname):
        """
        Check if VM hostname is valid (https://tools.ietf.org/html/rfc953)
        OACI uses the following regular expressions:
        - Linux: "^[a-zA-Z0-9][a-zA-Z0-9\\-]{0,61}[a-zA-Z0-9](\\.[a-zA-Z0-9][a-zA-Z0-9\\-]{0,61}[a-zA-Z0-9])*$"
        - Windows: "^[a-zA-Z][a-zA-Z0-9\\-]{0,13}[a-zA-Z0-9](\\.[a-zA-Z0-9][a-zA-Z0-9\\-]{0,13}[a-zA-Z0-9])*$"
        :param vm_hostname: string The hostname to check
        :return: bool
        """
        allowed_hostname_regexp = '^[a-zA-Z]([a-zA-Z0-9\.\-]+[0-9a-zA-Z])*$'
        return re.match(allowed_hostname_regexp, vm_hostname) is not None

    def _makeHwNode(self, subscription):
        logger.debug("Generate hardware node for subscription '{subscription_name}'.".format(
            subscription_name=subscription.name.encode('utf-8')
        ))
        if subscription.type == XmlSubscriptionDataSource.TYPE_PD:
            plesk_info = DataSource().api("HSPC/MT/Plesk").find_plesk_domain_cache("subscr_id", subscription.id)
            if plesk_info is not None:
                return self.makeHwNodeById(plesk_info.hw_id)

            logger.warning('No data in Plesk domain cache for subscription: %s', subscription.id)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_PC:
            plesk_info = DataSource().api("HSPC/MT/Plesk").find_plesk_client_cache("subscr_id", subscription.id)
            if plesk_info is not None:
                return self.makeHwNodeById(plesk_info.hw_id)

            logger.warning('No data in Plesk client cache for subscription: %s', subscription.id)

        return None

    @staticmethod
    def makeHwNodeById(hw_id):
        logger.debug("Generate hardware node by ID '{id}'.".format(id=hw_id))
        hwInfo = DBDataSource.getHwInfoByIdXml(hw_id)
        hwNode = Element("hw")
        hwId = hwInfo.getElementsByTagName("hw_id")
        if not hwId.length:
            logger.warning('No such hardware node with ID: %s', hw_id)
            return None

        hwNode.setAttribute("id", hwId[0].firstChild.nodeValue)
        hwNode.appendChild(TextElement("hostname", hwInfo.getElementsByTagName("hostname")[0].firstChild.nodeValue))
        hwPlatformId = int(hwInfo.getElementsByTagName("platform_id")[0].firstChild.nodeValue)
        hwNode.appendChild(TextElement("platformId", str(hwPlatformId)))
        tagIdWithHwIp = 'hostname' if hwPlatformId == DataSource.VPS_PLATFORM_SW_PLATFORM_LINUX_PSBM else 'ip_address'
        hwNode.appendChild(TextElement("ip", hwInfo.getElementsByTagName(tagIdWithHwIp)[0].firstChild.nodeValue))

        query_client = \
            "select GROUP_CONCAT(subscr_id) AS subscription_id_list from plesk_client_cache where hw_id = %s" % \
                       hwId[0].firstChild.nodeValue
        query_domain = \
            "select GROUP_CONCAT(subscr_id) AS subscription_id_list from plesk_domain_cache where hw_id = %s" % \
                       hwId[0].firstChild.nodeValue

        for s in DBDataSource().getData(query_client).elements:
            hwNode.appendChild(TextElement("plesk_client_subscriptions", s['subscription_id_list']))

        for s in DBDataSource().getData(query_domain).elements:
            hwNode.appendChild(TextElement("plesk_domain_subscriptions", s['subscription_id_list']))

        return hwNode
