# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""Tools to generate XML node describing account"""
import logging

from DataSources import XmlServiceTemplateDataSource, XmlSubscriptionDataSource, DBDataSource

from Exceptions import InvalidAccountType
from SimpleTypes import AccountInfo
from XMLExtensions import TextElement
from datetime import datetime
from xml.dom.minidom import Element

logger = logging.getLogger(__name__)


ACCOUNT_TYPES = {
    AccountInfo.TYPE_PROVIDER: 'provider',
    AccountInfo.TYPE_RESELLER: 'reseller',
    AccountInfo.TYPE_CUSTOMER: 'customer',
}  # this dictionary defines mapping between internal account type and XML
# account type


def formatDatetime(dt):
    """Format datetime to satisfy XSD schema.

       dt: datetime in UTC"""

    return dt.replace(microsecond=0).isoformat() + "Z"


def formatStrDatetime(dt):
    """Convert string to datetime and then to propper string format.

       dtStr: datetime"""

    return formatDatetime(datetime(*dt[:6]))


def makeCompanyName(account):
    """Generate company name for specified account.

       account: AccountInfo describing account to generate company for."""
    return account.companyName


def makeNotificationEmail(account):
    """Generate notification email address for specified account.

       account: AccountInfo describing account to generate email for."""
    return account.notificationEmail


def makeContactEmail(account, prefix):
    """Generate contact email address for scpecified account using specified
       prefix.

       account: AccountInfo describing account to generate email for.
       prefix: email address prefix (string)."""
    return account.contactEmail


def makeAddress(account):
    """Generate address string."""
    return account.address


def setDisableStateAttr(disableState, node):
    """Set 'DisableState' attribute group on specified node.

       disableState: state to set (Boolean)
       node: DOM node to affect"""

    if disableState:
        node.setAttribute("disabled", "true")
        return None


def setAccountAttributes(account, accountNode):
    """Set all apropriate attributes for account node.

       account: SimpleType.AccountInfo describing account to affect
       accountNode: DOM node to set attributes for"""

    accountNode.setAttribute("id", account.id)
    accountNode.setAttribute("login", account.persons[0].contactEmail)
    accountNode.setAttribute("type", ACCOUNT_TYPES[account.type])
    setDisableStateAttr(not account.active, accountNode)


class AccountGeneratorBase:

    """Account XML node generator"""
    PROVIDER_ACCOUNT = '1'

    settings = {  # generator settings
        "account": {  # account-related settings
            "probs": {  # probabilities settings
                "payment-tools": 1,  # payment tools presence prob
            },
            "payment-tools": {  # payment tools settings
                "credit-cards": {  # credit cards generation settings
                    # NOTE: credit card is validated by billing software so we
                    # need some real cards here:
                    "cards": [("4111111111111111", "Visa"), ],  # valid cards
                    "expiration-offset": 300,  # offset (in days) from now to the
                    # future to generate expiration date
                    "expiration-offset-rand": 100,  # expiration date random
                    # offset
                    "probs": {  # probabilities settings
                        "auto-payments": 1,  # probability of
                        # 'useForAutoPayments' = true
                        "holder": 1,  # probability of 'cardHolder' presence
                        "email": 1,  # probability of 'email' node presence
                        "address": 1,  # probability of 'address' presence
                        "phone": 1,  # probability of 'phone' node presence
                        "fax": 1,  # probability of 'fax' node presence
                    }
                },
                "credit-cards-min": 1,  # min count of credit cards to generate
                # NOTE: we have only one real credit card, so we can't generate
                # more than one card
                "credit-cards-max": 1,  # max count of credit cards to generate
            },
        },
    }

    def _setTesting(self, curDict):
        """Set all probs in curDict to 1. Could be useful for testing agent.

           curDict: dict to process"""
        for key in curDict:
            curEntry = curDict[key]
            if isinstance(curEntry, dict):
                if "probs" == key:
                    for k in curEntry:
                        curEntry[k] = 1
                else:
                    self._setTesting(curEntry)

    def setTestMode(self):
        """Set all probabilities to 1. This method could be useful for testing
           agent."""
        self._setTesting(self.settings)

    def getPlatformAccount(self, accountId):
        """Generate XML node describing specified account as needed in platform
           document.

           accountId: id of account to process.
        """
        logger.debug("Generate XML node for account '{account_id}'.".format(account_id=accountId))
        try:
            account = self.accSource.getEntityByID(accountId)
        except Exception:
            logger.warning("Unable to get account information for account '{account_id}'.".format(account_id=accountId))
            return None

        if not account.persons:
            logger.warning('Skipping account %s because it has no staff members', accountId)
            return None

        isParent = AccountInfo.TYPE_PROVIDER == account.type or AccountInfo.TYPE_RESELLER == account.type
        # create 'account' node and fill attributes
        rootNode = Element("account")
        setAccountAttributes(account, rootNode)
        # attach company name and notification email nodes
        rootNode.appendChild(self._makeCompanyNameNode(account))
        rootNode.appendChild(TextElement("email", makeNotificationEmail(account)))
        rootNode.appendChild(self._makeContactInfoNode(account.admin, "adminContactInfo", "admin"))
        rootNode.appendChild(self._makeContactInfoNode(account.billing, "billingContactInfo", "billing"))
        rootNode.appendChild(self._makeContactInfoNode(account.technical, "techContactInfo", "tech"))
        rootNode.appendChild(self._makeAddressNode(account, "companyAddress"))
        rootNode.appendChild(self._makeMembersNode(account))
        # attach service templates in case of provider or reseller
        if isParent:
            serviceTemplatesNode = self._makeServiceTemplatesNode(account)
            if serviceTemplatesNode.hasChildNodes():
                rootNode.appendChild(serviceTemplatesNode)
        # attach subscriptions node
        subscriptionsNode = self._makeSubscriptionsNode(account)
        if subscriptionsNode.hasChildNodes():
            rootNode.appendChild(subscriptionsNode)
        # attach child accounts in case of provider or reseller
        if isParent:
            accountsNode = self.getPlatformAccounts(account.id)
            if accountsNode.hasChildNodes():
                rootNode.appendChild(accountsNode)

        if accountId == AccountGeneratorBase.PROVIDER_ACCOUNT:
            hwNodesElement = Element('hwNodes')
            for hwNode in self.hwNodes.itervalues():
                hwNodesElement.appendChild(hwNode)
            rootNode.appendChild(hwNodesElement)

        return rootNode

    def getPlatformAccounts(self, ownerId):
        """Get 'accounts' DOM node describing accounts owned by
           argument-specified account.

           ownerId: provider or reseller ID to get accouts owned by.
           maxChildCount: limit of child accounts listed.

           Note: if maxChildCount is None then no limit will be used"""
        logger.debug("Get platform accounts for account '{owner_id}'.".format(owner_id=ownerId))
        accountsNode = Element('accounts')
        childIDs = self.accSource.getEntityIDsByOwner(ownerId)
        count = len(childIDs)
        for id in childIDs[:count]:
            accountNode = self.getPlatformAccount(id)
            if accountNode is not None:
                accountsNode.appendChild(accountNode)

        return accountsNode

    def getFullAccount(self, accountId):
        """Get 'account' DOM node for Account XML document.

           accountId: ID of account to process."""
        logger.debug("Get full account information for account '{account_id}'.".format(account_id=accountId))
        try:
            account = self.accSource.getEntityByID(accountId)
        except Exception:
            logger.warning("Unable to get account information for account '{account_id}'.".format(account_id=accountId))
            return None

        if not account.persons:
            logger.warning('Skipping account %s because it has no staff members', accountId)
            return None

        accSettings = self.settings["account"]
        probs = accSettings["probs"]
        isParent = AccountInfo.TYPE_PROVIDER == account.type or AccountInfo.TYPE_RESELLER == account.type
        # create 'account' node and fill it's attributes
        rootNode = Element("account")
        setAccountAttributes(account, rootNode)
        # attach company and contact info
        rootNode.appendChild(self._makeCompanyNameNode(account))
        rootNode.appendChild(self._makeContactInfoNode(account.admin, "adminContactInfo", "admin"))
        rootNode.appendChild(self._makeContactInfoNode(account.billing, "billingContactInfo", "billing"))
        rootNode.appendChild(self._makeContactInfoNode(account.technical, "techContactInfo", "tech"))
        rootNode.appendChild(self._makeAddressNode(account, "companyAddress"))
        # attach billing info
        rootNode.appendChild(self._makeBillingNode(account))
        # if random() < probs["payment-tools"]:
        self._makePaymentToolsNode(rootNode, account)
        # attach staff members node
        rootNode.appendChild(self._makeMembersNode(account))
        # attach service templates (in case of provider/reseller) and
        # subscriptions info
        if isParent:
            serviceTemplatesNode = self._makeServiceTemplatesNode(account)
            if serviceTemplatesNode.hasChildNodes():
                rootNode.appendChild(serviceTemplatesNode)

        subscrNode = self._makeSubscriptionsNode(account)

        if subscrNode.hasChildNodes():
            rootNode.appendChild(subscrNode)

        rootNode.appendChild(self._makeDomainsWithoutSubscriptionNode(account.id))

        return rootNode

    def _makeCompanyNameNode(self, account):
        """Generate DOM node containing account's company name.

           account: AccountInfo describing account to generate company for"""
        return TextElement("companyName", makeCompanyName(account))

    def _makeServiceTemplatesNode(self, account):
        """Generate DOM node enumerating account's service templates.

           account: provider or reseller to enumerate service templates owned
           by.

           Note: if specified account is not provider or reseller
           Exceptions.InvalidAccountType is raised"""

        logger.debug("Generate node enumerating account's service templates for account '{account_id}'.".format(
            account_id=account.id
        ))
        if AccountInfo.TYPE_PROVIDER != account.type and\
           AccountInfo.TYPE_RESELLER != account.type:
            raise InvalidAccountType(account.id, account.type)

        stsNode = Element("serviceTemplates")
        if account.id == '1':
            sts = self.serviceTemplateDataSource.data.keys()
        else:
            sts = self.serviceTemplateDataSource.getEntityIDsByOwner(account.id)

        for stId in sts:
            st = self.serviceTemplateDataSource.getEntityByID(stId)
            self._makeServiceTemplateNode(st, stsNode)

        return stsNode

    def _makeServiceTemplateNode(self, st, parentNode):
        """Generate DOM node describing single service template.

           st: SimpleTypes.ServiceTemplateInfo describing service template."""
        logger.debug("Generate DOM node for service template '{st_id}'.".format(st_id=st.id))
        stNode = Element("serviceTemplate")
        stNode.setAttribute("id", st.id)
        stNode.appendChild(TextElement("name", st.name))
        stNode.appendChild(TextElement("description", st.description))
        stNode.appendChild(TextElement("category", st.category))
        stNode.appendChild(TextElement("type", st.typeName))
        stNode.appendChild(TextElement("hostingType", st.hostingType))
        stNode.appendChild(TextElement("seriesKey", st.seriesKey))
        parentNode.appendChild(stNode)

    def _makeSubscriptionsNode(self, account):
        """Generate DOM node describing account's subscriptions.

           account: SimpleTypes.AccountInfo with info about subscriptions owner"""
        logger.debug("Generate DOM node describing account's subscriptions for account '{account_id}'.".format(
            account_id=account.id
        ))
        subscrNode = Element("subscriptions")
        if self.subscriptionSource:
            subscrIDs = self.subscriptionSource.getEntityIDsByOwner(account.id)
            for id in subscrIDs:
                subscr = self.subscriptionSource.getEntityByID(id)
                try:
                    curNode = self._makeSubscriptionNode(account, subscr)
                except Exception, e:
                    logger.warning('Skipping subscription %s due to error: %s', id, str(e), exc_info=True)
                    continue

                subscrNode.appendChild(curNode)

        return subscrNode

    def _createTimePeriod(self, subscrNode, nodeName, secPeriod):
        datePeriods = [
            ["hour", 3600],
            ["day", 24],
            ["month", 30],
            ["year", 12],
        ]

        curDiv = 1
        for i in range(len(datePeriods)):
            curDiv *= datePeriods[i][1]
        numSecPeriod = float(secPeriod)
        datePeriods.reverse()
        nextTaken = False
        for i in range(len(datePeriods)):
            curPeriod = datePeriods[i]
            period = numSecPeriod / curDiv
            iPeriod = int(period)
            if not nextTaken and period != iPeriod and ((numSecPeriod - curDiv) * curPeriod[1] / curDiv) >= 1:
                nextTaken = True
            else:
                if iPeriod > 0:
                    retNode = TextElement(nodeName, str(iPeriod))
                    retNode.setAttribute("type", curPeriod[0])
                    subscrNode.appendChild(retNode)
                    break
            curDiv /= curPeriod[1]

    def _makeContactInfoNode(self, account, nodeName, prefix):
        """Generate contact info and store it in DOM node.

           account: account to generate info for
           nodeName: name of generated node
           prefix: contact info prefix, e.g. 'admin',
           settings: dict with generation settings"""

        ciNode = Element(nodeName)
        ciNode.appendChild(TextElement("title", account.title))
        ciNode.appendChild(TextElement("firstName", account.firstName))
        ciNode.appendChild(TextElement("middleName", account.middleName))
        ciNode.appendChild(TextElement("lastName", account.lastName))
        ciNode.appendChild(TextElement("email", makeContactEmail(account, prefix)))
        ciNode.appendChild(self._makePhoneNode(account.phone, "phone"))
        ciNode.appendChild(self._makePhoneNode(account.fax, "fax"))
        ciNode.appendChild(self._makePhoneNode(account.mobile, "mobilePhone"))

        return ciNode

    def _makePhoneNode(self, number, nodeName):
        """Make DOM node with specified node name containing random phone
           number.

           nodeName: name of node to generate
           settings: dict with generation settings.

           Note: if setting is None then self.settings["phone"] is used"""
        phoneNode = Element(nodeName)
        if number.__class__.__name__ == 'structType':
            numberParts = [number.country_code, number.area_code,
                           number.number, number.extension]
        else:
            numberParts = number.split("|")
        partNames = ["countryCode", "areaCode", "phoneNumber", "extNumber"]
        for i in range(len(partNames)):
            value = str(numberParts[i]) if i < len(numberParts) else ''
            node = TextElement(partNames[i], value)
            phoneNode.appendChild(node)
        return phoneNode

    def _makeAddressNode(self, account, nodeName):
        """Make 'Address' DOM node with specified name, containing random
           address.

           nodeName: name of node to generate"""
        addressNode = Element(nodeName)
        addressNode.appendChild(TextElement("address1", account.address1))
        addressNode.appendChild(TextElement("address2", account.address2))
        addressNode.appendChild(TextElement("zipcode", account.zip))
        addressNode.appendChild(TextElement("city", account.city))
        addressNode.appendChild(TextElement("country", account.country))
        addressNode.appendChild(TextElement("state", account.state))
        if hasattr(account, "state_alt"):
            addressNode.appendChild(TextElement("state_alt", account.state_alt))

        return addressNode

    def _makeBillingNode(self, account):
        """Make 'billingInfo' node with random balance"""
        logger.debug("Generate node with billing information for account '{account_id}'.".format(account_id=account.id))
        biNode = Element("billingInfo")
        biNode.appendChild(TextElement("balance", account.balance))
        biNode.appendChild(TextElement("tax_reg_id", account.tax_reg_id))
        return biNode

    def dynamicObjectToXML(self, dObj, curNode=None):
        newNode = None
        for attr_name in [arg for arg in dir(dObj) if not arg.startswith('_')]:
            attr = getattr(dObj, attr_name)
            if attr_name == 'address':
                curNode.appendChild(self._makeAddressNode(attr, "address"))
            elif attr_name == 'phone_src':
                curNode.appendChild(self._makePhoneNode(attr, "phone"))
            elif attr.__class__.__name__ == "DynamicObject":
                newNode = Element(attr_name)
                self.dynamicObjectToXML(attr, newNode)
                if curNode:
                    curNode.appendChild(newNode)
            elif attr.__class__.__name__ == "list":
                newNode = Element(attr_name)
                for item in attr:
                    self.dynamicObjectToXML(item, newNode)
                if curNode:
                    curNode.appendChild(newNode)
            else:
                curNode.appendChild(TextElement(attr_name, str(attr).decode('utf-8')))
        return newNode

    def _makePaymentToolsNode(self, parentNode, account):
        """Make 'paymentTools' node.

           account: SimpleTypes.AccountInfo describing account to create node
                    for.
           settings: dict with generation settings."""
        tools = self.paymentToolDataSource.getCardsByOwner(account.id)
        ptNode = Element("paymentTools")
        ptNode.appendChild(self.dynamicObjectToXML(tools, ptNode))
        parentNode.appendChild(ptNode)
        return ptNode

    def _makeMemberNode(self, account):
        """Make DOM node describing one staff member.

           account: SimpleType.AccountInfo describing account to generate member
                    for.
        """
        memberNode = Element("member")

        ciNode = self._makeContactInfoNode(account, "contactInfo", "staff")
        memberNode.appendChild(ciNode)

        addressNode = self._makeAddressNode(account, "address")
        memberNode.appendChild(addressNode)

        loginNode = TextElement("login", account.contactEmail)
        memberNode.appendChild(loginNode)

        pwdNode = TextElement("password", account.password)
        pwdNode.setAttribute("type", "plain")
        memberNode.appendChild(pwdNode)

        langNode = TextElement("lang", account.lang)
        memberNode.appendChild(langNode)

        return memberNode

    def _makeMembersNode(self, account):
        """Make 'staffMembers' DOM node.

           account: SimpleTypes.AccountInfo describing account to generate
                    members for.
           settings: dict with generation settings.
           ciSettings: dict with contact info generation settings."""
        membersNode = Element("staffMembers")
        for i in range(len(account.persons)):
            membersNode.appendChild(self._makeMemberNode(account.persons[i]))
        return membersNode

    def _makeDomainsWithoutSubscriptionNode(self, accountId):
        dmBaseSrc = DBDataSource.getData(DBDataSource.TYPE_DOMAIN_REG + ' WHERE dm_basedomain.account_no=%s' % accountId)
        dmUseSrc = DBDataSource.getData(DBDataSource.TYPE_DOMAIN_USE + ' WHERE b.account_no=%s' % accountId)

        domainsWithoutSubscription = []
        allDomains = set()
        for domain in dmBaseSrc.getElements('account_no', accountId):
            allDomains.add(domain['domain'])

        for domainName in allDomains:
            if not self._domain_registration_subscription_for_domain_exists(dmBaseSrc, domainName) or \
                    (not self._active_domain_registration_subscription_for_domain_exists(dmBaseSrc, domainName) and
                     self._is_domain_linked_to_active_hosting_subscription(dmUseSrc, domainName)):
                domainsWithoutSubscription.append(domainName)

        domainsWithoutSubscriptionNode = Element("domainsWithoutSubscription")
        for domain in domainsWithoutSubscription:
            isDnsManaged = dmBaseSrc.getElements("domain", domain)[0]["dns_enabled"]
            domainId = dmBaseSrc.getElements("domain", domain)[0]["id"]
            domainTargetSubscription = dmUseSrc.getElements("domain", domain)
            dmNsSrc = DBDataSource.getData(DBDataSource.TYPE_DOMAIN_NS + ' WHERE domain_id = %s' % domainId)

            domainNode = Element("domain")
            domainNode.appendChild(TextElement("id", domainId))
            domainNode.appendChild(TextElement("name", domain))
            domainNode.appendChild(TextElement("isDnsManaged", isDnsManaged))
            if len(domainTargetSubscription) > 0:
                domainNode.appendChild(TextElement("targetSubscriptionId", domainTargetSubscription[0]["subscr_id"]))

            nameServers = dmNsSrc.getElements("domain_id", domainId)
            domainNode.appendChild(self._getNameServersNode(nameServers))

            domainsWithoutSubscriptionNode.appendChild(domainNode)

        return domainsWithoutSubscriptionNode

    def _getNameServersNode(self, nameServers):
        nameServersNode = Element("nameservers")
        for ns in nameServers:
            nsNode = Element("ns")
            nsNode.appendChild(TextElement("id", ns["ns_id"]))
            nsNode.appendChild(TextElement("registrar_syncronized", ns["registrar_syncronized"]))
            nameServersNode.appendChild(nsNode)

        return nameServersNode

    def _domain_registration_subscription_for_domain_exists(self, dmBaseSrc, domain):
        records = dmBaseSrc.getElements("domain", domain)
        for record in records:
            if record["subscr_id"]:
                return True

        return False

    def _active_domain_registration_subscription_for_domain_exists(self, dmBaseSrc, domain):
        records = dmBaseSrc.getElements("domain", domain)
        for record in records:
            if record["subscr_id"] and \
                            record["subscription_status"] in XmlSubscriptionDataSource.getActiveSubscriptionStatusIds():
                return True

        return False

    def _is_domain_linked_to_active_hosting_subscription(self, dmUseSrc, domain):
        domainTargetSubscription = dmUseSrc.getElements("domain", domain)
        return len(domainTargetSubscription) > 0 \
               and domainTargetSubscription[0]["status"] in XmlSubscriptionDataSource.getActiveSubscriptionStatusIds()
