# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""OBAS MigrationAgent main class"""
import logging
from time import sleep

from xml.dom.minidom import Document, Element

from AccountGenerator import AccountGenerator
from Exceptions import APIversionError
from PlatformGenerator import PlatformGenerator
from ServicesGenerator import ServicesGenerator
from DataSources import DataSource, DBDataSource
from DNSGenerator import DNSGenerator

logger = logging.getLogger(__name__)

OBAS_VERSION_API = 7
OBAS_API_VERSION_ERROR = "OBAS API version is not suitable for migration. Please upgrade to OBAS 4.5.5-15 or latter."
OBAS_API_DETERMINATION_ERROR = "Cannot determine OBAS API version: %s"
OBAS_API_ATTEMPT_TO_GET_API_VERSION = "Trying to detect an OBAS API version, attempt %d of %d"
OBAS_API_VERSION_WAIT_BETWEEN_ATTEMPTS = "Failed to detect an OBAS API version. Sleeping %d seconds before the next attempt."
OBAS_API_VERSION_FOUND = "Detected the following OBAS API version: %d"
OBAS_AGENT_INIT_MAX_ATTEMPTS = 5
OBAS_AGENT_INIT_WAIT_INTERVAL = 60


class AgentImpl:
    """OBAS migration agent main class"""

    def __init__(self):
        self.ds = DataSource()
        for i in xrange(OBAS_AGENT_INIT_MAX_ATTEMPTS):
            try:
                logger.debug(OBAS_API_ATTEMPT_TO_GET_API_VERSION % (i + 1, OBAS_AGENT_INIT_MAX_ATTEMPTS))
                version = self.ds.api().get_api_version()
                logger.debug(OBAS_API_VERSION_FOUND % version)
                break
            except Exception as e:
                if i >= OBAS_AGENT_INIT_MAX_ATTEMPTS - 1:
                    raise APIversionError(OBAS_API_DETERMINATION_ERROR % str(e))
                else:
                    logger.warning(OBAS_API_VERSION_WAIT_BETWEEN_ATTEMPTS % OBAS_AGENT_INIT_WAIT_INTERVAL)
                    sleep(OBAS_AGENT_INIT_WAIT_INTERVAL)

        if OBAS_VERSION_API > version:
            raise APIversionError(OBAS_API_VERSION_ERROR)

    def getPlatform(self, account_id_list=None):
        """Get platform XML document. This method returns DOM Document with platform info."""

        logger.debug("Get platform XML document for accounts '{account_id_list}'.".format(
            account_id_list=None if account_id_list is None else ','.join(account_id_list)
        ))
        generator = PlatformGenerator()
        doc = Document()
        doc.appendChild(generator.getPlatform(account_id_list))
        return doc

    def getNodes(self, account_id_list=None, export_inactive=False):
        """
        Get minimalistic platform XML document with OBAS nodes info.
        This method returns DOM Document with platform info.

        NOTE: Only Plesk hosts will be returned for now
        """
        logger.debug("Get minimalistic platform XML document with OBAS nodes info.")
        generator = PlatformGenerator(export_inactive)
        doc = Document()
        doc.appendChild(generator.getNodes())
        return doc

    def getOwnershipTree(self):
        """Get platform XML document.
           This method returns DOM Document with platform info."""

        generator = PlatformGenerator()
        return generator.getOwnershipTree()

    def getPleskClientLogins(self, subscriptionIdList):
        """Get platform XML document. This method returns DOM Document with plesk client logins info."""

        generator = PlatformGenerator()
        return generator.getPleskClientLogins(subscriptionIdList)

    def getAccountList(self, accountIdList, subscriptionIdList):
        generator = AccountGenerator(accountIdList, subscriptionIdList)
        doc = Document()
        accountsNode = Element('accounts')
        for accountId in accountIdList:
            accountNode = generator.getFullAccount(accountId)
            if accountNode is not None:
                accountsNode.appendChild(accountNode)

        doc.appendChild(accountsNode)
        return doc

    def getSubscriptionList(self, subscriptionIdList, exportInactive=False):
        logger.debug(
            "Getting subscriptions XML document for the following subscriptions ({filter_inactive} inactive): {sub_list}".format(
                filter_inactive='including' if exportInactive else 'excluding',
                sub_list=', '.join(subscriptionIdList)
            )
        )
        generator = AccountGenerator(('-1',), subscriptionIdList, exportInactive)
        doc = Document()
        subscriptionsNode = Element('subscriptions')
        for subscriptionId in subscriptionIdList:
            subscr = generator.subscriptionSource.getEntityByID(subscriptionId)
            curNode = generator._makeSubscriptionNode(None, subscr)
            subscriptionsNode.appendChild(curNode)

        doc.appendChild(subscriptionsNode)
        return doc

    def getAccount(self, accountId):
        """ Get account XML document for specific account.
            This method returns DOM Document with account info.
            Params: accountId  -  ID of account to get info about.
        """
        logger.debug("Get account XML document for account '{account_id}'.".format(account_id=accountId))

        if accountId is not None:
            account_id_list = [accountId]
        else:
            account_id_list = None
        generator = AccountGenerator(account_id_list=account_id_list)
        doc = Document()
        accountNode = generator.getFullAccount(accountId)
        if accountNode is not None:
            doc.appendChild(accountNode)

        return doc

    def getServices(self, subscriptionId, mnIp=None):
        """Get XML document describing specified subscription's services.
           This method returns DOM Document with services info.
           Params: subscriptionId  -  ID of subscription to get services for."""

        logger.debug("Get XML document describing services for subscription'{subscriptionId}'.".format(
            subscriptionId=subscriptionId
        ))
        generator = ServicesGenerator()
        doc = Document()
        doc.appendChild(generator.getServices(subscriptionId, mnIp))
        return doc

    def getDnsRecords(self, domainIdList):
        generator = DNSGenerator(domainIdList)
        doc = Document()
        domainsRecords = generator.getDomainsRecords()
        if domainsRecords:
            doc.appendChild(domainsRecords)

        return doc

    def disableServices(self, subscriptionId):
        self.ds.api().set_subscr_on_hold('subscr_id', subscriptionId)

    def disableAccount(self, accountId):
        self.ds.api().set_account_on_hold('account_no', accountId)

    def disableZone(self, zoneName):
        zoneToDisable = self.ds.api(namespace='HSPC/MT/DM/DNSGate').find_zone('zone', zoneName)
        doc = Document()
        tasksNode = Element("tasks")

        if zoneToDisable.is_active != 0:
            maxTaskId = DBDataSource.getData(DBDataSource.TYPE_MAX_TASK_ID).elements[0]["id"]
            self.ds.api(namespace='HSPC/MT/DM/DNSGate').switch_activity('zone', zoneName, 'is_active', 0)
            newTasks = DBDataSource.getData(DBDataSource.TYPE_NS_SYNC_TASKS % (maxTaskId, zoneName))

            for task in newTasks.elements:
                taskNode = Element("task")
                taskNode.setAttribute("id", task["task_id"])
                tasksNode.appendChild(taskNode)

        else:
            logger.debug(u'Skipped DNS zone "%s" disabling because it is not active' % zoneName)

        doc.appendChild(tasksNode)
        return doc

    def configurePleskNodeForTakeover(self, hostId):
        configurable_host = self.ds.api(namespace='HSPC/MT/OM').find_hwnode('id', hostId)
        if configurable_host['id'] is None:
            raise Exception("Unable to configure the Plesk node with ID '%s' for takeover: the host not found" % hostId)

        if configurable_host['sync_plesk_objects'] == 0 and configurable_host['is_disabled'] == 1:
            logger.debug("The Plesk node with ID '%s' is ready for takeover, configuring was skipped" % hostId)
            return

        result = DBDataSource.postData(DBDataSource.UPDATE_HW_FOR_TAKEOVER % int(hostId))
        if result == 0:
            logger.debug("Successfully configured the Plesk node with ID '%s' for takeover" % hostId)
        else:
            raise Exception(
                "Failed to configure the Plesk node with ID '%s' for takeover. For details, see the log file "
                "'/opt/oam/agent/lib/obas-agent/debug.log' on the Odin Business Automation Standard management node."
                % hostId
            )

    def waitForTasks(self, taskIds, attempts=30, timeout=5):
        if not taskIds:
            logger.debug("No tasks to wait for")
            return

        logger.debug("Waiting for the tasks, ids = %s" % taskIds)
        thereAreTasksInProgress = True
        while attempts > 0 and thereAreTasksInProgress:
            tasks = DBDataSource.getData(DBDataSource.TYPE_TASK_QUEUE + " WHERE task_id in (%s)" % ','.join(taskIds))
            logger.debug(tasks.elements)
            thereAreTasksInProgress = False
            for task in tasks.elements:
                # None == task has not been ran yet, -255 == pending, -1 == task is in progress
                if task["status"] in ["-1", "-255", None]:
                    thereAreTasksInProgress = True
                    logger.debug("Task #%s is still in progress, waiting for 5 seconds" % task["task_id"])
                    sleep(timeout)
                    break

                if int(task["status"]) < 0:
                    raise Exception("The task #%s '%s' has failed" % (task["task_id"], task["description"]))

        logger.debug("All tasks have been completed")
