# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""Tools to generate custom dns settings info"""
import logging

from DataSources import DBDataSource
from XMLExtensions import TextElement
from xml.dom.minidom import Element, DocumentFragment


logger = logging.getLogger(__name__)


class DNSGenerator:

    """DNS service info generator"""

    def __init__(self, domainIdList):
        logger.debug("Creating DNS Generator for the domains with IDs: %s" % ",".join(domainIdList))
        self.domainIdList = domainIdList
        self.zoneSrc = DBDataSource().getData(
            DBDataSource.TYPE_DM_ZONE + " where dom.id in (%s)" % ",".join(domainIdList)
        )

        zone_id_list = []
        for zone_node in self.zoneSrc.elements:
            if zone_node['dns_enabled'] == '1':
                zone_id_list.append(zone_node["zone_id"])
            else:
                logger.debug("Skipped the retrieving of DNS resource records for the domain %s without DNS hosting." % zone_node['domain_id'])

        if zone_id_list:
            self.dnsSrc = DBDataSource().getData(
                DBDataSource.TYPE_DNS + " where zone.zone_id in (%s)" % ",".join(zone_id_list)
            )
        else:
            logger.warning("None of the domains have any resource records to export and import to Odin Automation.")
            self.dnsSrc = None

    def getDomainsRecords(self):
        logger.debug("Generate DNS DOM node.")
        if not len(self.domainIdList):
            logger.debug("Generating a DNS DOM node was skipped because DNS zones were not found")
            return None

        rootNode = Element("dns")
        rootNode.appendChild(self._makeDomainNodes())
        return rootNode

    def _makeDomainNode(self, domainId):
        logger.debug("Generate records node for domain id '{id}'".format(id=domainId))
        zone = self.zoneSrc.getElements("domain_id", domainId)
        if len(zone) < 1:
            logger.warning("DNS zone for domain '{id}' was not found".format(id=domainId))
            return None

        zone = zone[0]
        if zone["dns_enabled"] == "0":
            logger.warning(u"DNS is disabled for DNS zone '{name}' although it was implied that it is enabled".format(
                name=unicode(zone["domain"]))
            )

        is_dns_managed = zone["dns_enabled"]
        domainNode = Element("domain")
        domainNode.setAttribute("id", domainId)
        domainNode.setAttribute("name", zone["domain"])
        domainNode.setAttribute("dns_managed", is_dns_managed)
        domainNode.setAttribute("enabled", zone["is_active"] if is_dns_managed == '1' else '0')
        records = self.dnsSrc.getElements("zone_id", zone["zone_id"]) if is_dns_managed == '1' else []
        domainNode.appendChild(self._makeRecordsNode(records))
        return domainNode

    def _makeDomainNodes(self):
        domainNodes = DocumentFragment()
        for domainId in self.domainIdList:
            domainNode = self._makeDomainNode(domainId)
            if domainNode is not None:
                domainNodes.appendChild(domainNode)

        return domainNodes

    def _makeRecordsNode(self, records):
        recordsNode = Element("records")
        supportedTypes = set([
            "mx",
            "cname",
            "a",
            "txt",
            "aaaa",
            "srv",
            "soa"
        ])

        for record in records:
            recordType = record["type_txt"].lower()
            if recordType in supportedTypes:
                node = self._makeCommonRecordNode(
                    recordType, record["name"], record["value"], str(record["ttl"]), record["priority"], record["is_edit_allowed"]
                )
                recordsNode.appendChild(node)

        return recordsNode

    def _makeCommonRecordNode(self, recordType, name, value, ttl, priority, is_edit_allowed):
        recordNode = Element("record")
        recordNode.appendChild(TextElement("type", recordType))
        recordNode.appendChild(TextElement("name", name))
        recordNode.appendChild(TextElement("value", value))
        recordNode.appendChild(TextElement("ttl", ttl))
        recordNode.appendChild(TextElement("priority", priority or ''))
        recordNode.appendChild(TextElement("is_edit_allowed", is_edit_allowed))
        return recordNode
