# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""Data sources for SampleAgent"""
import ConfigParser
import logging
import os
import re
import sys
import subprocess
import tempfile
import time
import xml.dom.minidom

from Exceptions import DuplicateID, NoSuchElement
from SimpleTypes import AccountInfo, DynamicObject, DBData
from SOAPpy import SOAPProxy
from commands import *
from pipes import quote
from string import Template
from xml.sax import saxutils


logger = logging.getLogger(__name__)

class DataSource:

    """ Base class for DataSources.
        All data is stored in dictionary 'self.data' in the following format:
        { id: (entity_data, child_ids) }
    """
    USE_LOCAL_SOURCES = False

    TYPE_VE = 'VE_GENERIC'
    TYPE_VM = 'VM_GENERIC'
    TYPE_DM = 'DM_GENERIC'
    TYPE_MISC = 'MISC'
    TYPE_PC = 'PLESK_CLIENT'
    TYPE_PD = 'PLESK_DOMAIN'
    TYPE_DEDICATED_SERVER = 'HN_GENERIC'
    TYPE_POA = 'POA_GENERIC'
    TYPE_SSL = 'SSL_SINGLE'
    TYPE_PLESK_HN = 'VE_PLESK_HN'
    TYPE_PLESK_SERVER = 'PLESK_SERVER'
    TYPE_HN_VIRTUOZZO = 'HN_VIRTUOZZO'

    # Keep this list synchronized with SUPPORTED_PLESK_PLATFORMS in migrator\products\obas\lib\export\platform.py
    SUPPORTED_PLESK_PLATFORMS = [
        201,  # SW_PLATFORM_LINUX_PLESK7
        202,  # SW_PLATFORM_WINDOWS_PLESK7
        203   # SW_PLATFORM_PLESK_UNIVERSAL
    ]

    # SW_PLATFORM_LINUX (Virtuozzo 2.0)
    VPS_PLATFORM_SW_PLATFORM_LINUX = 1
    # SW_PLATFORM_LINUX_VZ30
    VPS_PLATFORM_SW_PLATFORM_LINUX_VZ30 = 3
    # SW_PLATFORM_WINDOWS_VZ30
    VPS_PLATFORM_SW_PLATFORM_WINDOWS_VZ30 = 4
    # SW_PLATFORM_LINUX_X64 (Parallels Virtuozzo Containers for Linux (x86_64))
    VPS_PLATFORM_SW_PLATFORM_LINUX_X64 = 5
    # SW_PLATFORM_LINUX_IA64
    VPS_PLATFORM_SW_PLATFORM_LINUX_IA64 = 6
    # SW_PLATFORM_LINUX_PSBM (PCS6)
    VPS_PLATFORM_SW_PLATFORM_LINUX_PSBM = 400

    # Keep this list synchronized with SUPPORTED_VPS_PLATFORMS in migrator\products\obas\lib\export\platform.py
    SUPPORTED_VPS_PLATFORMS = [
        VPS_PLATFORM_SW_PLATFORM_LINUX_X64,
        VPS_PLATFORM_SW_PLATFORM_LINUX_PSBM
    ]

    def init(self):
        self.data = {}
        self.children = {}

    def api(self, namespace='HSPC/IEM/Migrator'):
        self.servername = "http://127.0.0.1:8080/hspc/xml-api"
        p = SOAPProxy(self.servername, namespace)
        p.config.namespaceStyle = '2001'
        return p

    def _addEntity(self, id, entity):
        """Add new entity to pool"""

        if id in self.data:
            ex = DuplicateID(self.ENTITY_TYPE, id)
            raise DuplicateID, ex
        self.data[id] = entity

    def _addChild(self, ownerId, childId):
        """Add child ID to account's child list"""
        if not ownerId in self.children:
            self.children[ownerId] = []
        self.children[ownerId].append(childId)

    def _getXmlNodeText(self, node):
        rc = []
        if node != None and node.nodeType == node.TEXT_NODE:
            rc.append(node.data)
        text = ''.join(rc)
        return self.fix_double_xml_entity_escaping(text)

    def fix_double_xml_entity_escaping(self, double_escaped_text):
        """Fix XML entity escaping of hspc-export.pl output

        'hspc-export.pl' returns double escaped XML entities like this:
        <email>eino.sporer&amp;amp;foobar@nonexist.com</email>

        To fix this, the method makes another round of XML entity unescaping.
        """
        xml_predefined_entities = {'&apos;': "'", '&quot;': '"'}
        return saxutils.unescape(double_escaped_text, xml_predefined_entities)

    def _getChildNodeText(self, parentNode, childName, strictMode=False):
        nodesList = parentNode.getElementsByTagName(childName)
        for i in range(nodesList.length):
            if not strictMode or nodesList.item(i).parentNode == parentNode:
                return self._getXmlNodeText(nodesList.item(i).firstChild)
        return ""

    def _export(self, objName, propName, inSet):
        """ Export accounts data from OBAS to XML file.
        """
        logger.debug("Export data from OBAS to XML file for object '{objName}' with filter by '{propName}' for {inSet}.".format(
            objName=objName, propName=propName, inSet=inSet
        ))
        if DataSource.USE_LOCAL_SOURCES:
            return self.DB_NAME

        # Generate XML filter for hspc-export script: /tmp/hspc-filter.xml
        # NOTE. Though the documentation of the script says it can read data STDIN if
        #       no input file specified, there is a bug allowing to do that.
        (fd, fname) = tempfile.mkstemp(text=True)
        f = os.fdopen(fd, "w+b")
        xmldata = """<?xml version="1.0" encoding="UTF-8"?>
            <filters>
                <objects_name>""" + objName + """</objects_name>
                <filter>
                  <property_name>""" + propName + """</property_name>
                  <where><in>""" + inSet + """</in></where>
                </filter>
            </filters>
            """
        f.write(xmldata)
        f.close()

        sys.stderr.write(xmldata)

        # Run patched hspc-export.pl from agent dir and pass the filter.

        cmd = "perl ./hspc-export.pl -f " + fname
        errcode, stdout, stderr = local_command_exec(cmd)

        # Parse output and extract xml.gz.
        # Output example:
        #
        # Collecting data....
        #                         OK
        # Exporting data....
        #                         OK
        #
        # Your file: export-XXXXXXXX.xml.gz
        #
        file_name_match = re.search(ur"Your file:\s.*", stdout)
        if file_name_match is None:
            error_message = "Failed to export data from OBAS: unable to detect a filename of XML data report.\nCalled command: {cmd}\nExit code: {code}\nCommand output:\n{stdout}\nCommand errors:\n{stderr}".format(
                cmd=cmd, code=str(errcode), stdout=stdout, stderr=stderr
            )
            logger.error(error_message)
            raise Exception(error_message)

        file_name = file_name_match.group()
        file_name = re.search(ur"export-.*\.xml\.gz", file_name).group()

        # Unzip the output file
        getoutput("gunzip -f " + file_name)
        file_name = file_name[:-3]

        dbName = file_name

        os.remove(fname)
        return dbName

    def getEntityByID(self, id):
        """Get entity with argument-specified ID.
           Params:  id  -  ID to get entity with"""
        if not id in self.data:
            ex = NoSuchElement(self.ENTITY_TYPE, id)
            raise NoSuchElement, ex
        return self.data[id]

    def getEntityIDsByOwner(self, ownerId):
        """Get IDs of all entities which have specified owner.
           Params: ownerId  -  id of owner to get entities for"""
        if not ownerId in self.children:
            return []
        return self.children[ownerId]

    def getXmlElementDictionary(self, xmlData, containerElementName, keyElementName):
        """Return a dictionary of XmlElements 'containerElementName' with the keys taken from 'keyElementName'"""
        nodes = xmlData.getElementsByTagName(containerElementName)
        elementList = dict()
        for nodeIndex in range(nodes.length):
            node = nodes.item(nodeIndex)
            key_data = self._getChildNodeText(node, keyElementName).strip()
            elementList[key_data] = node
        return elementList

    def __repr__(self):
        return "\n{objectType}:\n- self.data: {data}\n- self.children: {children}\n".format(
            objectType=self.__class__.__name__,
            data='; '.join('{k}: {v}'.format(k=str(key), v=str(value)) for key, value in self.data.iteritems()),
            children='; '.join('{k}: {v}'.format(k=str(key), v=str(value)) for key, value in self.children.iteritems())
        )


class XMLDataWithAddress(DataSource):
    ADDRESS_ZIP_FIELD = "zip"
    ADDRESS_A1_FIELD = "address1"
    ADDRESS_A2_FIELD = "address2"
    ADDRESS_CITY_FIELD = "city"
    ADDRESS_COUNTRY_FIELD = "country"
    ADDRESS_STATE_FIELD = "state"
    ADDRESS_STATE_ALT_FIELD = "state_alt"

    def parseAddress(self, addressNode, resObj):
        resObj.zip = self._getChildNodeText(addressNode, self.ADDRESS_ZIP_FIELD)
        resObj.address1 = self._getChildNodeText(addressNode, self.ADDRESS_A1_FIELD)
        resObj.address2 = self._getChildNodeText(addressNode, self.ADDRESS_A2_FIELD)
        resObj.city = self._getChildNodeText(addressNode, self.ADDRESS_CITY_FIELD)
        resObj.country = self._getChildNodeText(addressNode, self.ADDRESS_COUNTRY_FIELD).lower()
        resObj.state = self._getChildNodeText(addressNode, self.ADDRESS_STATE_FIELD)
        resObj.state_alt = self._getChildNodeText(addressNode, self.ADDRESS_STATE_ALT_FIELD)


class XmlAccountDataSource(XMLDataWithAddress):

    """ Account info data source. This class gets info from XML file.
    """
    # Account entity name:
    ENTITY_TYPE = "Account"
    # Filename of accounts CSV DB:
    DB_NAME = "accounts6.xml"
    if DataSource.USE_LOCAL_SOURCES:
        KEY_CONFIG_NAME = "hspc.conf"
    else:
        KEY_CONFIG_NAME = "/etc/hspc/hspc.conf"
    # Column names of CSV DB:
    HOST_PLANS_FIELD = "account"
    NAME_FIELD = "id"
    TAX_REG_ID = "tax_ex_number"
    OWNER_FIELD = "vendor_id"
    OWNER_ID_FIELD = "type"

    ADMIN_PREFIX = "admin_"
    BILLING_PREFIX = "billing_"
    TECHNICAL_PREFIX = "technical_"
    PERSON_PRFIX = ""
    # fields that will be used together with some prefixes: admin, bulling, technical, or without any
    TITLE_FIELD = "prefix"
    FIRST_NAME_FIELD = "fname"
    MIDDLE_NAME_FIELD = "mname"
    LAST_NAME_FIELD = "lname"
    MAIL_FIELD = "email"
    ADDRESS_PHONE_FIELD = "phone_src"
    ADDRESS_FAX_FIELD = "fax_src"
    ADDRESS_MOBILE_FIELD = "mobile_src"
    PASSWORD_HASH = "password_hash"

    LANGUAGE = "lang"

    COMPANY_NAME_FIELD = "name"
    STATUS_FIELD = "status"

    ADDRESS_FIELD = "address"

    # Mapping between account types in CSV and AccountInfo
    ACCOUNT_TYPES = {
        'HSP': AccountInfo.TYPE_PROVIDER,
        'RESELLER': AccountInfo.TYPE_RESELLER,
        'CUSTOMER': AccountInfo.TYPE_CUSTOMER,
    }

    def __init__(self, account_id_list):
        self.init()

        config = open(self.KEY_CONFIG_NAME, "r").read()
        self.hmacKey = re.search(ur"(?<=PASSPORT_HMAC_KEY=).*", config).group().replace("\\#", "#") + "\n"
        if not account_id_list:
            account_filter = '*'
        else:
            if not '1' in account_id_list:
                account_id_list.append('1')
            account_filter = ','.join(account_id_list)
        XmlAccountDataSource.DB_NAME = self._export("account", "id", account_filter)
        self.parseDoc(xml.dom.minidom.parse(XmlAccountDataSource.DB_NAME))

    def fillContactData(self, xmlNode, prefix, objToFill):
        objToFill.title = self._getChildNodeText(xmlNode, prefix + self.TITLE_FIELD)
        objToFill.firstName = self._getChildNodeText(xmlNode, prefix + self.FIRST_NAME_FIELD)
        objToFill.middleName = self._getChildNodeText(xmlNode, prefix + self.MIDDLE_NAME_FIELD)
        objToFill.lastName = self._getChildNodeText(xmlNode, prefix + self.LAST_NAME_FIELD)
        objToFill.contactEmail = self._getChildNodeText(xmlNode, prefix + self.MAIL_FIELD)
        objToFill.phone = self._getChildNodeText(xmlNode, prefix + self.ADDRESS_PHONE_FIELD)
        objToFill.fax = self._getChildNodeText(xmlNode, prefix + self.ADDRESS_FAX_FIELD)
        objToFill.mobile = self._getChildNodeText(xmlNode, prefix + self.ADDRESS_MOBILE_FIELD)

    def parseDoc(self, document):
        accounts = document.getElementsByTagName(self.HOST_PLANS_FIELD)
        balanceData = self._getAccountBalanceData()
        for i in range(accounts.length):
            accountXml = accounts.item(i)
            account = AccountInfo()
            account.id = accountXml.getAttribute(self.NAME_FIELD)
            account.tax_reg_id = self._getChildNodeText(accountXml, self.TAX_REG_ID)
            if account.id in balanceData:
                account.balance = self._getChildNodeText(balanceData[account.id], 'balance').strip()
            else:
                account.balance = str(0)
            account.type = XmlAccountDataSource.ACCOUNT_TYPES[self._getChildNodeText(accountXml, self.OWNER_ID_FIELD)]
            account.companyName = self._getChildNodeText(accountXml, self.COMPANY_NAME_FIELD)
            account.active = self._getChildNodeText(accountXml, self.STATUS_FIELD) == "active"

            account.ownerId = self._getChildNodeText(accountXml, self.OWNER_FIELD)

            # address parsing goes here
            self.parseAddress(accountXml.getElementsByTagName(self.ADDRESS_FIELD).item(0), account)

            account.admin = DynamicObject()
            self.fillContactData(accountXml, self.ADMIN_PREFIX, account.admin)
            account.billing = DynamicObject()
            self.fillContactData(accountXml, self.BILLING_PREFIX, account.billing)
            account.technical = DynamicObject()
            self.fillContactData(accountXml, self.TECHNICAL_PREFIX, account.technical)

            account.notificationEmail = account.admin.contactEmail

            account.persons = []
            persons = accountXml.getElementsByTagName("person")
            personsCount = persons.length
            for personId in range(personsCount):
                curPerson = persons[personId]
                thisPersonObj = DynamicObject()
                self.fillContactData(curPerson, self.PERSON_PRFIX, thisPersonObj)
                self.parseAddress(curPerson, thisPersonObj)
                thisPersonObj.password = self.hmacKey + self._getChildNodeText(curPerson, self.PASSWORD_HASH)
                thisPersonObj.lang = self._getChildNodeText(curPerson, self.LANGUAGE)
                account.persons.append(thisPersonObj)

            self._addEntity(account.id, account)
            # add ID of current account to parent's account child list
            if AccountInfo.TYPE_PROVIDER != account.type:
                self._addChild(account.ownerId, account.id)

    def _getAccountBalanceData(self):
        """Get all account balance records; return a dictionary of XmlElements"""
        xmlData = DBDataSource.getDomTreeData(DBDataSource.TYPE_BALANCE, 'balance_record')
        return self.getXmlElementDictionary(xmlData, 'balance_record', 'account_no')

AccountDataSource = XmlAccountDataSource


class XmlServiceTemplateDataSource(DataSource):

    """Service template info data source. This class gets info from csv file"""

    # Service template entity name:
    ENTITY_TYPE = "Service template"
    # Filename of service templates CSV DB:
    DB_NAME = ""

    # Node names of XML DB:
    HOST_PLANS_FIELD = "hosting_plan"
    OWNER_TYPE_FIELD = "type"
    OWNER_FIELD = "vendor_id"
    CATEGORY_FIELD = "cat_id"
    SERIES_KEY = "series_key"

    HP_ID_FIELD = "hp_id"
    NAME_FIELD = "name"
    DESCRIPTION_FIELD = "description"

    FEE_FIELD = "fee"
    FEE_PERIOD_FIELD = "period"
    SETUP_FEE_FIELD = "setup_fee"
    SUBSCR_FEE_FIELD = "subscr_fee"

    TYPE_VIRTUOZZO_CONTAINER = '1'
    TYPE_DEDICATED_SERVER = '3'
    TYPE_VIRTUOZZO_DEDICATED_NODE = '4'
    TYPE_DOMAIN_REGISTRATION = '6'
    TYPE_MISCELLANEOUS = '7'
    TYPE_PLESK_DEDICATED_SERVER = '8'
    TYPE_PLESK_DOMAIN = '9'
    TYPE_PLESK_CLIENT = '10'
    TYPE_PLESK_VIRTUAL_NODE = '11'
    TYPE_VIRTUAL_MACHINE = '14'

    def __init__(self, ids):
        self.init()
        if len(ids) > 0:
            self.exportIds = ids
            # old by f*EPAM
            self.load(DBDataSource.getHostingPlansXml(self.exportIds))
            #self.parseResponse(self.api().get_hosting_plans('plans_ids', ids))

    def load(self, document):
        hostingPlans = document.getElementsByTagName(self.HOST_PLANS_FIELD)
        plansCount = hostingPlans.length
        # next if statement is for debug purposes
        if plansCount > len(self.exportIds):
            self.exportIds = [str(i + 1) for i in range(plansCount)]
        tldData = DBDataSource.getData(DBDataSource.TYPE_HPTLD % ",".join(self.exportIds))
        planCategories = self._getPlanCategories()
        hostingPlansTypes = self._getPlansTypesNames()

        for i in range(plansCount):
            curPlan = hostingPlans.item(i)
            hostingPlan = DynamicObject()
            hostingPlan.id = self._getChildNodeText(curPlan, self.HP_ID_FIELD)
            hostingPlan.type = self._getChildNodeText(curPlan, self.OWNER_TYPE_FIELD)
            hostingPlan.seriesKey = self._getChildNodeText(curPlan, self.SERIES_KEY)

            if hostingPlan.type in [self.TYPE_DEDICATED_SERVER, self.TYPE_PLESK_DEDICATED_SERVER]:
                hostingPlan.hostingType = 'dedicated'
            elif hostingPlan.type in [self.TYPE_DOMAIN_REGISTRATION, self.TYPE_PLESK_DOMAIN, self.TYPE_PLESK_CLIENT]:
                hostingPlan.hostingType = 'shared'
            elif hostingPlan.type in [self.TYPE_VIRTUOZZO_CONTAINER, self.TYPE_VIRTUOZZO_DEDICATED_NODE, self.TYPE_PLESK_VIRTUAL_NODE, self.TYPE_VIRTUAL_MACHINE]:
                hostingPlan.hostingType = 'vps'
            else:
                hostingPlan.hostingType = ''

            if hostingPlan.type == self.TYPE_DOMAIN_REGISTRATION:
                hostingPlan.type = self.TYPE_DM
            elif hostingPlan.type == self.TYPE_PLESK_DOMAIN:
                hostingPlan.type = self.TYPE_PD
            elif hostingPlan.type == self.TYPE_PLESK_CLIENT:
                hostingPlan.type = self.TYPE_PC
            elif hostingPlan.type == self.TYPE_MISCELLANEOUS:
                hostingPlan.type = self.TYPE_MISC
            else:  # "1"
                hostingPlan.type = self.TYPE_VE

            stTldData = tldData.getElements(self.HP_ID_FIELD, hostingPlan.id)
            hostingPlan.tld = []
            hostingPlan.tld_fees = {}
            for j in range(len(stTldData)):
                hostingPlan.tld.append(stTldData[j]["tld"])
            hostingPlan.ownerId = self._getChildNodeText(curPlan, self.OWNER_FIELD)

            # Name is a xml inside node.
            hostingPlan.name = self._get_def_lang(self._getChildNodeText(curPlan, self.NAME_FIELD).strip())

            # Description is a xml inside node.
            hostingPlan.description = self._get_def_lang(
                self._getChildNodeText(curPlan, self.DESCRIPTION_FIELD).strip()
            )

            # Add hosting plan category
            categoryId = self._getChildNodeText(curPlan, self.CATEGORY_FIELD)
            if categoryId in planCategories:
                hostingPlan.category = planCategories[categoryId]
            else:
                hostingPlan.category = ''

            # Add hosting plan type name
            if hostingPlan.type == self.TYPE_PD:
                hostingPlan.typeName = 'Plesk Domain'
            elif hostingPlan.type == self.TYPE_PC:
                hostingPlan.typeName = 'Plesk Client'
            else:
                typeId = self._getChildNodeText(curPlan, self.OWNER_TYPE_FIELD)
                hostingPlan.typeName = self._get_def_lang(hostingPlansTypes[typeId])

            if hostingPlan.type == self.TYPE_DM:
                for k in range(len(stTldData)):
                    tld = stTldData[k]["tld"]
                    # feeNodes = DBDataSource.getDomainRegistrationFeesXml(hostingPlan.id, tld)
                    hostingPlan.tld_fees[tld] = []
                    # feeNodes = feeNodes.getElementsByTagName(self.FEE_FIELD)
                    continue
                    '''
                    for j in range(feeNodes.length):
                        curFeeNode = feeNodes[j]
                        feeObj = DynamicObject()
                        feeObj.period = self._getChildNodeText(curFeeNode, "period")
                        feeObj.setupFee = self._getChildNodeText(curFeeNode, "registration_fee")
                        feeObj.subscrFee = self._getChildNodeText(curFeeNode, "renew_fee")
                        hostingPlan.tld_fees[tld].append(feeObj)
                    '''
            else:
                # feeNodes = DBDataSource.getHostingPlanFeesXml(hostingPlan.id)
                # feeNodes = feeNodes.getElementsByTagName(self.FEE_FIELD)
                hostingPlan.fees = []
                '''
                for j in range(feeNodes.length):
                    curFeeNode = feeNodes[j]
                    feeObj = DynamicObject()
                    feeObj.period = self._getChildNodeText(curFeeNode, self.FEE_PERIOD_FIELD)
                    feeObj.setupFee = self._getChildNodeText(curFeeNode, self.SETUP_FEE_FIELD)
                    feeObj.subscrFee = self._getChildNodeText(curFeeNode, self.SUBSCR_FEE_FIELD)
                    hostingPlan.fees.append(feeObj)
                '''

            self._addEntity(hostingPlan.id, hostingPlan)
            # add ID of current service template to parent's account child list
            self._addChild(hostingPlan.ownerId, hostingPlan.id)

    def _getPlanCategories(self):
        """Get all available plan categories"""
        categories = dict()
        categoryListNode = DBDataSource.getHostingPlanCategoryXml()
        categoryNodes = categoryListNode.getElementsByTagName('category')
        for nodeIndex in range(categoryNodes.length):
            categoryNode = categoryNodes.item(nodeIndex)
            # Parse category content, which is an XML string itself
            # retrieve the element 'lang_def', which is the default category name
            categoryName = self._get_def_lang(self._getChildNodeText(categoryNode, "name").strip())
            categoryId = self._getChildNodeText(categoryNode, "cat_id").strip()
            categories[categoryId] = categoryName

        return categories

    def _getPlansTypesNames(self):
        """Get all hosting plans types names"""
        res = {}
        tree = DBDataSource.getHostingPlansTypesXml()
        nodes = tree.getElementsByTagName('hosting_plan')
        for i in xrange(nodes.length):
            curr_type = nodes.item(i)
            type_id = self._getChildNodeText(curr_type, 'type_id')
            res[type_id] = self._getChildNodeText(curr_type, 'name').strip()

        return res

    def parseResponse(self, plan_list):
        for plan in plan_list:
            hostingPlan = DynamicObject()
            hostingPlan.id = str(plan.id)
            hostingPlan.type = str(plan.type)
            hostingPlan.description = str(plan.name).decode('utf-8')
            # hostingPlan.description=str(plan.description)
            hostingPlan.ownerId = str(plan.vendor_id)
            hostingPlan.tld = []
            hostingPlan.tld_fees = {}
            hostingPlan.fees = []

            if hasattr(plan, 'tld_info'):
                for tld in plan.tld_info:
                    hostingPlan.tld.append(tld.tld)
                    hostingPlan.tld_fees[tld.tld] = []
                    for fee in tld.fee_list:
                        feeObj = DynamicObject()
                        feeObj.period = str(fee.period)
                        feeObj.setupFee = str(fee.registration_fee)
                        feeObj.subscrFee = str(fee.renew_fee)
                        hostingPlan.tld_fees[tld.tld].append(feeObj)
            if hasattr(plan, 'fee_info'):
                for fee in plan.fee_info:
                    feeObj = DynamicObject()
                    feeObj.period = str(fee.period)
                    feeObj.setupFee = str(fee.setup_fee)
                    feeObj.subscrFee = str(fee.subscr_fee)
                    hostingPlan.fees.append(feeObj)

            self._addEntity(hostingPlan.id, hostingPlan)
            # add ID of current service template to parent's account child list
            self._addChild(hostingPlan.ownerId, hostingPlan.id)

    @staticmethod
    def _get_def_lang(data):
        lang_def = filter(lambda x: x.startswith(u'def>'), map(unicode.strip, data.split(u'<lang_')))
        if lang_def:
            lang_def = lang_def[0].strip()
            return lang_def[len('def>'): -len('</lang_def>')].strip()

        return data


ServiceTemplateDataSource = XmlServiceTemplateDataSource


class XmlSubscriptionDataSource(DataSource):

    """Subscription info data source. This class gets info from csv file"""

    TIME_FORMAT = "%Y-%m-%d %H:%M:%S"
    # Service template entity name:
    ENTITY_TYPE = "Subscription"
    # Filename of service templates CSV DB:
    DB_NAME = "subscription1.xml"
    # Column names of CSV DB:
    HOST_PLANS_FIELD = "account"
    NAME_FIELD = "id"

    SUBSCRIPTION_FIELD = "subscription"
    SUBSCRIPTION_ID_FIELD = "id"
    SUBSCRIPTION_NAME_FIELD = "name"

    OWNER_ID_FIELD = "plan_id"
    OWNER_TYPE_FIELD = "plan_type"
    PERIOD_FIELD = "period"
    STATUS_FIELD = "status"
    START_DATE_FIELD = "start_date"
    END_DATE_FIELD = "end_date"

    STATUS_ACTIVE = "ACTIVE"
    STATUS_DELETED = "DELETED"
    STATUS_PENDING = "PENDING"
    STATUS_ON_HOLD = "ON_HOLD"
    STATUS_GRACE = "GRACE"
    STATUS_EXPIRED = "EXPIRED"
    STATUS_TERMINATING = "TERMINATING"

    STATUS_ID_ACTIVE = '1'
    STATUS_ID_DELETED = '2'
    STATUS_ID_PENDING = '3'
    STATUS_ID_ON_HOLD = '4'
    STATUS_ID_GRACE = '10'
    STATUS_ID_EXPIRED = '11'
    STATUS_ID_TERMINATING = '12'

    def __init__(self, subscription_id_list, export_inactive=False):
        self.export_inactive = export_inactive
        self.init()
        if subscription_id_list:
            id_list = ','.join(subscription_id_list)
        else:
            id_list = '*'
        XmlSubscriptionDataSource.DB_NAME = self._export("subscription", "id", id_list)
        self.load(xml.dom.minidom.parse(XmlSubscriptionDataSource.DB_NAME))

    def getServiceTemplateIds(self):
        resObj = {}
        resList = []
        for i in self.data:
            curId = self.getEntityByID(i).serviceTemplateId
            resObj[curId] = curId

        for i in resObj:
            resList.append(int(i))

        resList.sort()
        for i in range(len(resList)):
            resList[i] = str(resList[i])

        return resList

    def _parseSubscription(self, ownerId, subscriptionNode):
        subscription = DynamicObject()
        subscription.status = self._getChildNodeText(subscriptionNode, self.STATUS_FIELD)
        subscription.need_to_export = subscription.status in self.getSubscriptionStatusNamesToExport(
            self.export_inactive
        )
        if not subscription.need_to_export:
            return

        subscription.active = subscription.status in self.getActiveSubscriptionStatusNames()
        subscription.id = subscriptionNode.getAttribute(self.SUBSCRIPTION_ID_FIELD)
        subscription.name = self._getChildNodeText(subscriptionNode, self.SUBSCRIPTION_NAME_FIELD)
        subscription.ownerId = ownerId
        subscription.serviceTemplateId = self._getChildNodeText(subscriptionNode, self.OWNER_ID_FIELD)
        subscription.type = self._getChildNodeText(subscriptionNode, self.OWNER_TYPE_FIELD)
        subscription.subscriptionPeriod = self._getChildNodeText(subscriptionNode, self.PERIOD_FIELD)
        # FIXME: billingPeriod is not used by OAM right now. Research how to get its actual value if needed.
        subscription.billingPeriod = subscription.subscriptionPeriod
        subscription.startDate = time.strptime(
            self._getChildNodeText(subscriptionNode, self.START_DATE_FIELD), XmlSubscriptionDataSource.TIME_FORMAT)
        subscription.endDate = time.strptime(
            self._getChildNodeText(subscriptionNode, self.END_DATE_FIELD), XmlSubscriptionDataSource.TIME_FORMAT)

        self._addEntity(subscription.id, subscription)
        # add ID of current subscription to parent's account child list
        self._addChild(subscription.ownerId, subscription.id)

    def load(self, document):
        accounts = document.getElementsByTagName(self.HOST_PLANS_FIELD)
        accountsCount = accounts.length
        for i in range(accountsCount):
            curAccount = accounts.item(i)
            accountId = curAccount.getAttribute(self.NAME_FIELD)

            subscriptions = curAccount.getElementsByTagName(self.SUBSCRIPTION_FIELD)
            subscriptionsCount = subscriptions.length

            for j in range(subscriptionsCount):
                self._parseSubscription(accountId, subscriptions.item(j))

    @staticmethod
    def getActiveSubscriptionStatusNames():
        return [
            XmlSubscriptionDataSource.STATUS_ACTIVE,
            XmlSubscriptionDataSource.STATUS_GRACE,
        ]

    @staticmethod
    def getSupportedSubscriptionStatusNames():
        return [
            XmlSubscriptionDataSource.STATUS_ACTIVE,
            XmlSubscriptionDataSource.STATUS_GRACE,
            XmlSubscriptionDataSource.STATUS_DELETED,
            XmlSubscriptionDataSource.STATUS_ON_HOLD,
            XmlSubscriptionDataSource.STATUS_EXPIRED,
            XmlSubscriptionDataSource.STATUS_TERMINATING,
        ]

    @staticmethod
    def getSubscriptionStatusNamesToExport(export_inactive):
        if not export_inactive:
            return XmlSubscriptionDataSource.getActiveSubscriptionStatusNames()
        else:
            return XmlSubscriptionDataSource.getSupportedSubscriptionStatusNames()

    @staticmethod
    def getActiveSubscriptionStatusIds():
        return [
            XmlSubscriptionDataSource.STATUS_ID_ACTIVE,
            XmlSubscriptionDataSource.STATUS_ID_GRACE,
        ]

    @staticmethod
    def getSupportedSubscriptionStatusIds():
        return [
            XmlSubscriptionDataSource.STATUS_ID_ACTIVE,
            XmlSubscriptionDataSource.STATUS_ID_GRACE,
            XmlSubscriptionDataSource.STATUS_ID_DELETED,
            XmlSubscriptionDataSource.STATUS_ID_ON_HOLD,
            XmlSubscriptionDataSource.STATUS_ID_EXPIRED,
            XmlSubscriptionDataSource.STATUS_ID_TERMINATING,
        ]

    @staticmethod
    def getSubscriptionStatusIdsToExport(export_inactive):
        if not export_inactive:
            return XmlSubscriptionDataSource.getActiveSubscriptionStatusIds()
        else:
            return XmlSubscriptionDataSource.getSupportedSubscriptionStatusIds()


SubscriptionDataSource = XmlSubscriptionDataSource


class XMLPaymentToolsSource(XMLDataWithAddress):
    # mapping between OBAS and POA card types
    CARD_TYPES = {
        'A': "American Express",
        'B': "Diners Club",
        'C': "Diners Club",
        'D': "Discover",
        'E': "Diners Club",
        'J': "JCB",
        'L': "Visa Electron",
        'M': "MasterCard",
        'S': 'Maestro',
        'T': "Visa",
        'V': "Visa",
        'W': 'Maestro',
        'BT_SEPA': 'SEPA DirectDebit',
        'BT_Dummy': 'BTDummy',
        'BT_DTAUS': 'DTAUS',
        'BT_Clieop03': 'Clieop03',
        'BT_Norma19': 'Norma19',
        'OP_Anet': 'AnetToken',
        'OP_Protx': 'SagePayToken',
        'OP_PayPal': 'PayPalAdaptiveToken',
        'OP_WebMoney': 'WebMoney',
    }

    def getCardsByOwner(self, ownerId):
        resObj = DynamicObject()
        resObj.CreditCards = []
        resObj.BankAccounts = []
        resObj.OnlineAccounts = []
        resObj.CustomTools = []

        try:
            api = self.api()
            response = api.get_decrypted_pay_methods_for_account('account_id', ownerId)
        except Exception as e:
            logger.error(
                'Skipping the export of payment methods for the account %s because of the error: %s',
                ownerId, str(e), exc_info=True
            )
            return resObj

        for card in response:
            try:
                pmObj = DynamicObject()
                cardObj = DynamicObject()
                cardObj.use_for_recurring = card.recurring

                if card.template_id == 'OP_CCard':
                    cardObj.cardNumber = card.secure_data.card_number
                    cardObj.cvv = card.secure_data.cvv
                    cardObj.cardHolder = str(card.secure_data.holder_name) + ' ' \
                                         + str(card.secure_data.holder_name2)
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.issueNumber = '' if card.secure_data.issue_number is None else  card.secure_data.issue_number

                    if hasattr(card.public_data, 'valid_from') and card.public_data.valid_from is not None:
                        recTime = time.strptime(str(card.public_data.valid_from), XmlSubscriptionDataSource.TIME_FORMAT)
                        cardObj.validFrom = DynamicObject()
                        cardObj.validFrom.year = recTime.tm_year
                        cardObj.validFrom.month = recTime.tm_mon

                    cardObj.address = DynamicObject()
                    cardObj.address.zip = str(card.secure_data.zip).decode('utf-8')
                    cardObj.address.address1 = str(card.secure_data.address1).decode('utf-8')
                    cardObj.address.address2 = str(card.secure_data.address2).decode('utf-8')
                    cardObj.address.city = str(card.secure_data.city).decode('utf-8')
                    cardObj.address.country = card.secure_data.country.lower()
                    cardObj.address.state = str(card.secure_data.state_uscanada).decode('utf-8') or \
                                            str(card.secure_data.state).decode('utf-8')

                    recTime = time.strptime(str(card.expire_date), XmlSubscriptionDataSource.TIME_FORMAT)
                    cardObj.expirationDate = DynamicObject()
                    cardObj.expirationDate.year = recTime.tm_year
                    cardObj.expirationDate.month = recTime.tm_mon

                    cardObj.phone_src = card.secure_data.phone_src

                    pmObj.creditCard = cardObj
                    resObj.CreditCards.append(pmObj)
                elif card.template_id == 'BT_SEPA':
                    cardObj.holder_name = str(card.account_name)
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.mandate_id = card.public_data.mandate_id
                    cardObj.is_verified = card.public_data.is_verified
                    cardObj.mandate_sign_date = card.public_data.mandate_sign_date
                    cardObj.account_number = card.secure_data.account_number
                    cardObj.bank_code = card.secure_data.bank_code
                    pmObj.bankAccount = cardObj
                    resObj.BankAccounts.append(pmObj)
                elif card.template_id == 'BT_Dummy':  # or card.template_id == 'BT_DTAUS':
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.holder_name = card.public_data.holder_name
                    cardObj.account_number = card.secure_data.account_number
                    cardObj.bank_code = card.secure_data.bank_code
                    pmObj.bankAccount = cardObj
                    resObj.BankAccounts.append(pmObj)
              #  elif card.template_id == 'BT_Clieop03':
              #      cardObj.holder_name = card.public_data.holder_name
              #      cardObj.agree = card.public_data.agree
              #      cardObj.no_agreement = card.public_data.no_agreement
              #      cardObj.account_number = card.secure_data.account_number
              #      pmObj.bankAccount = cardObj
              #      resObj.BankAccounts.append(pmObj)
              #  elif card.template_id == 'BT_Norma19':
              #      cardObj.holder_name = card.public_data.holder_name
              #      cardObj.account_number=card.public_data.account_number
              #      cardObj.bank_agency_number=card.public_data.bank_agency_number
              #      cardObj.bank_branch_number=card.public_data.bank_branch_number
              #      cardObj.stored_control_dig=card.public_data.stored_control_dig
              #      cardObj.cif=card.public_data.cif
              #      pmObj.bankAccount = cardObj
              #      resObj.BankAccounts.append(pmObj)
                elif card.template_id == 'OP_Anet':
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.profile_id = card.public_data.profile_id
                    cardObj.paymethod_id = card.public_data.paymethod_id
                    cardObj.method_name = card.public_data.method_name
                    cardObj.pp_plugin_id = card.public_data.pp_plugin_id
                    pmObj.onlineAccount = cardObj
                    resObj.OnlineAccounts.append(pmObj)
                elif card.template_id == 'OP_Protx':
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.vendor_tx_code = card.public_data.vendor_tx_code
                    cardObj.vps_tx_id = card.public_data.vps_tx_id
                    cardObj.status3ds = card.public_data.status3ds
                    cardObj.card_type = card.public_data.card_type
                    cardObj.last4digits = card.public_data.last4digits
                    cardObj.address_result = card.public_data.address_result
                    cardObj.post_code_result = card.public_data.post_code_result
                    cardObj.avscv2 = card.public_data.avscv2
                    cardObj.cv2_result = card.public_data.cv2_result
                    cardObj.cavv = card.public_data.cavv
                    cardObj.expire_date = card.public_data.expire_date
                    cardObj.tx_auth_no = card.secure_data.tx_auth_no
                    cardObj.secure_key = card.secure_data.secure_key
                    pmObj.onlineAccount = cardObj
                    resObj.OnlineAccounts.append(pmObj)
                elif card.template_id == 'OP_PayPal':
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.pkey = card.secure_data.pkey
                    cardObj.approved = card.public_data.approved
                    cardObj.number_of_payments = card.public_data.number_of_payments
                    cardObj.total_amount = card.public_data.total_amount
                    cardObj.currency_code = card.public_data.currency_code
                    cardObj.max_total_amount = card.public_data.max_total_amount
                    cardObj.sender_email = card.public_data.sender_email
                    cardObj.expire_date = card.public_data.expire_date
                    cardObj.pp_plugin_id = card.public_data.pp_plugin_id
                    pmObj.onlineAccount = cardObj
                    resObj.OnlineAccounts.append(pmObj)
                elif card.template_id == 'OP_WebMoney':
                    cardObj.paymentSystem = self.CARD_TYPES[card.type]
                    cardObj.holder_name = card.public_data.holder_name
                    cardObj.wm_id = card.secure_data.wm_id
                    pmObj.onlineAccount = cardObj
                    resObj.OnlineAccounts.append(pmObj)
            except Exception as e:
                logger.error(
                    'Skipping the export of credit card %s for the account %s because of the error: %s',
                    card.template_id, ownerId, str(e), exc_info=True
                )
                    
         #   else:
         #       custTool=self._parseCustomTool(card)
         #       custTool.paymentSystem = card.type
         #       pmObj.customPayTool = custTool
         #       resObj.CustomTools.append(pmObj)
        return resObj

    def _parseCustomTool(self, paymethod):
        custTool = DynamicObject()
        for attr_name in [arg for arg in dir(paymethod.public_data) if not arg.startswith('_')]:
            attr = getattr(paymethod.public_data, attr_name)
            setattr(custTool, attr_name, attr)
        for attr_name in [arg for arg in dir(paymethod.secure_data) if not arg.startswith('_')]:
            attr = getattr(paymethod.secure_data, attr_name)
            setattr(custTool, attr_name, attr)
        return custTool

    def __init__(self):
        self.init()

PaymentToolsSource = XMLPaymentToolsSource


class DBDataSource:
    UPDATE_HW_FOR_TAKEOVER = "UPDATE hw SET sync_plesk_objects = 0, is_disabled = 1 WHERE hw_id = %d"

    if DataSource.USE_LOCAL_SOURCES:
        TYPE_DOMAIN_REG = "domains_reg.xml"
        TYPE_DOMAIN_USE = "domains_use.xml"
        TYPE_DOMAIN_NS = "domains_ns.xml"
        TYPE_BALANCE = "balance.xml"
        TYPE_HPTLD = "hptld.xml"
        TYPE_DNS = "dns.xml"
        TYPE_DM_ZONE = "zone.xml"
        TYPE_SSH_NAMESERVERS = "ssh_nameservers.xml"
        TYPE_MAX_TASK_ID = "max_task_id.xml"
        TYPE_NS_SYNC_TASKS = "ns_sync_tasks.xml"
        TYPE_TASK_QUEUE = "task_queue.xml"
        TYPE_VPS_CT = 'vps_ct.xml'
        TYPE_VPS_VM = 'vps_vm.xml'
    else:
        TYPE_DOMAIN_REG = \
            "SELECT dm_basedomain.id, dm_basedomain.account_no, utf_domain AS domain, subscr_domain.subscr_id, " \
            "dns_enabled, status AS subscription_status " \
            "FROM dm_basedomain " \
            "  LEFT JOIN subscr_domain ON dm_basedomain.domain = subscr_domain.domain_name " \
            "  JOIN dm_domain ON dm_domain.id = dm_basedomain.id " \
            "  LEFT JOIN subscr_base ON subscr_base.subscr_id = subscr_domain.subscr_id"
        TYPE_DOMAIN_USE = \
            "SELECT id, utf_domain AS domain, l.subscr_id, s.status " \
            "FROM dm_basedomain b " \
            "  JOIN subscr_linked_domain l ON b.id = l.domain_id " \
            "  JOIN subscr_base s ON l.subscr_id = s.subscr_id"
        TYPE_DOMAIN_NS = "SELECT * FROM dm_domain_ns"
        TYPE_BALANCE = "SELECT account_no, balance FROM acc_balance"
        TYPE_HPTLD = \
            "SELECT tld, hp_id " \
            "FROM bm_hp " \
            "  JOIN bm_hp_tld ON bm_hp.series_key = bm_hp_tld.hp_series_key " \
            "WHERE bm_hp.hp_id IN (%s)"
        TYPE_DNS = \
            "SELECT zone.zone_id, rec.name AS name, ttl, value, type_txt, priority, is_edit_allowed " \
            "FROM dm_zone zone JOIN dm_record rec ON zone.zone_id = rec.zone_id"
        TYPE_DM_ZONE = \
            "SELECT dom.id AS domain_id, utf_domain AS domain, zone_id, is_active, dns_enabled " \
            "FROM dm_basedomain dom " \
            "  LEFT JOIN dm_zone zone ON dom.domain = zone.zone"
        TYPE_SSH_NAMESERVERS = \
            "SELECT dm_ns.ns_id as ns_id, ns_type_id, ip_address, utf8_hostname, named_conf " \
            "FROM dm_ns LEFT OUTER JOIN dm_ns_ssh ON dm_ns_ssh.ns_id = dm_ns.ns_id"
        TYPE_MAX_TASK_ID = \
            "SELECT MAX(task_id) as id FROM task_queue"
        TYPE_NS_SYNC_TASKS = \
            "SELECT * FROM task_queue " \
            "WHERE task_id > %s AND class = 'HSPC::MT::DM::DNS::Synchronizator' AND method = 'sync_zone' AND description LIKE '%%(%s)%%'"
        TYPE_TASK_QUEUE = \
            "SELECT * FROM task_queue"
        TYPE_VPS_CT = \
            "SELECT subscr_ve.subscr_id, subscr_ve.ve_id, ve_status.hw_id, ve.utf_ve_name, ve.ve_name, ve.uuid " \
            "FROM subscr_ve " \
            "  LEFT JOIN ve_status ON subscr_ve.ve_id = ve_status.ve_id " \
            "  LEFT JOIN ve ON subscr_ve.ve_id = ve.ve_id "
        TYPE_VPS_VM = \
            "SELECT pserver_vm.subscr_id, pserver_vm.vm_eid, pserver_vm.hw_id, pserver_vm.name FROM pserver_vm "
    ROW_FIELD = "row"
    ELEMENT_FIELD = "field"
    MY_CNF = 'aspc.cnf'
    OBAS_CONFIG = '/etc/hspc/hspc.conf'

    @staticmethod
    def getData(dataType):
        # for local debug filename passed instead request
        if DataSource.USE_LOCAL_SOURCES:
            return DBDataSource._loadDocument(xml.dom.minidom.parse(dataType))

        dom = DBDataSource.getDomTreeData(dataType)
        return DBDataSource._loadDocument(dom)

    @staticmethod
    def postData(dataType):
        return DBDataSource.makeMysqlRequest(dataType, True)

    @staticmethod
    def getDomTreeData(dataType, elementName=None):
        # for local debug filename passed instead request
        if DataSource.USE_LOCAL_SOURCES:
            return DBDataSource._loadDocument(xml.dom.minidom.parse(dataType))

        output = DBDataSource.makeMysqlRequest(dataType)
        dom = xml.dom.minidom.parseString(output)

        if elementName is None:
            return dom

        return DBDataSource.renameMysqlXmlOutputTags(dom, elementName)

    @staticmethod
    def makeMysqlRequest(sql_request, return_exit_code_only=False):
        logger.debug("Run query '{sql_request}'".format(sql_request=sql_request))
        file_is_newer = lambda filename1, filename2: os.stat(filename1).st_mtime > os.stat(filename2).st_mtime
        if not os.path.isfile(DBDataSource.MY_CNF) or file_is_newer(DBDataSource.OBAS_CONFIG, DBDataSource.MY_CNF):
            DBDataSource._configureMysqlConnection()

        exit_code, stdout, stderr = local_command_exec(
            'mysql --defaults-file=%s aspc --xml -e "%s"' % (DBDataSource.MY_CNF, sql_request), log_only_error=True
        )

        return exit_code if return_exit_code_only else stdout

    @staticmethod
    def _configureMysqlConnection():
        """Create a .cnf file with MySQL client connection options"""
        obas_config_options = DBDataSource.readObasConfig()

        mycnf = ConfigParser.ConfigParser()
        mycnf.add_section('client')
        mycnf.set('client', 'user', obas_config_options['MAIN_DB_USER'])
        mycnf.set('client', 'password', obas_config_options['MAIN_DB_PASS'])
        mycnf.set('client', 'host', obas_config_options['MAIN_DB_HOST'])
        mycnf.set('client', 'default-character-set', 'utf8')
        cnf = open(DBDataSource.MY_CNF, 'w')
        mycnf.write(cnf)
        cnf.close()

    @staticmethod
    def readObasConfig():
        obas_config = open(DBDataSource.OBAS_CONFIG)

        obas_config_dict = dict()
        for line in obas_config.readlines():
            db_setting = re.search(r'\s*([\w]+\S+)\s*=\s*(\S+)', line)
            if db_setting is not None:
                setting_name, setting_value = db_setting.groups()
                obas_config_dict[setting_name] = setting_value

        return obas_config_dict

    @staticmethod
    def renameMysqlXmlOutputTags(dom, newTagName):
        """Convert mysql -e 'SELECT ...' --xml output to agent internal format"""
        renames = []

        resultset = dom.getElementsByTagName('resultset')
        renames.extend((resultset.item(n), 'root') for n in xrange(resultset.length))

        rowNodeList = dom.getElementsByTagName('row')
        renames.extend((rowNodeList.item(n), newTagName) for n in xrange(rowNodeList.length))

        fields = dom.getElementsByTagName('field')
        renames.extend((fields.item(n), fields.item(n).getAttribute('name')) for n in xrange(fields.length))

        for node, newName in renames:
            dom.renameNode(node, xml.dom.minidom.EMPTY_NAMESPACE, newName)

        return dom

    @staticmethod
    def getHostingPlansXml(ids):
        query = Template("SELECT hp_id, name, description, type, vendor_id, cat_id, series_key "
                         "FROM bm_hp "
                         "WHERE hp_id IN ($ids)")
        return DBDataSource.getDomTreeData(query.substitute(ids=",".join(ids)), "hosting_plan")

    @staticmethod
    def getHwInfoByIdXml(hw_id):
        query = Template("SELECT * "
                         "FROM hw "
                         "WHERE hw_id = '$id'")
        return DBDataSource.getDomTreeData(query.substitute(id=hw_id), "hw")

    @staticmethod
    def getHostingPlanFeesXml(hp_id):
        query = Template("SELECT period, setup_fee, subscr_fee "
                         "FROM bm_hp_fees "
                         "WHERE hp_id = $hp_id")
        return DBDataSource.getDomTreeData(query.substitute(hp_id=hp_id), "plan_fee")

    @staticmethod
    def getHostingPlanCategoryXml():
        return DBDataSource.getDomTreeData("SELECT cat_id, name FROM bm_hp_category;", "category")

    @staticmethod
    def getHostingPlansTypesXml():
        return DBDataSource.getDomTreeData("SELECT type_id, name FROM bm_hp_type", "hosting_plan")

    @staticmethod
    def getHostingPlanTypeXml(type_id):
        query = Template("SELECT name "
                         "FROM bm_hp_type "
                         "WHERE type_id = $type_id")
        return DBDataSource.getDomTreeData(query.substitute(type_id=type_id), "hosting_plan")

    @staticmethod
    def getDomainRegistrationFeesXml(hp_id, tld):
        query = Template("SELECT period, registration_fee, renew_fee "
                         "  FROM bm_hp_tld_regopt "
                         " WHERE tld_res_id = ("
                         "       SELECT tld_res_id "
                         "         FROM bm_hp_tld "
                         "        WHERE active = 1 AND tld = '$tld' AND hp_series_key = ("
                         "              SELECT series_key "
                         "                FROM bm_hp "
                         "               WHERE hp_id = $hp_id"
                         "       )"
                         " )")
        return DBDataSource.getDomTreeData(query.substitute(hp_id=hp_id, tld=tld), "registration_fee")

    @staticmethod
    def _loadDocument(document):
        dbDataList = DBData()
        for row in document.getElementsByTagName(DBDataSource.ROW_FIELD):
            attributesObj = {}
            for element in row.getElementsByTagName(DBDataSource.ELEMENT_FIELD):
                value = None
                if element.firstChild is not None:
                      value = element.firstChild.nodeValue

                attributesObj[element.attributes['name'].value] = value

            dbDataList.addElement(attributesObj)

        return dbDataList


def local_command_exec(command, stdin_content=None, env=None, log_only_error=False):
    if type(command) is list or type(command) is tuple:
        command_str = u' '.join([quote(c) for c in command])
        command_encoded = [c.encode('utf-8') for c in command]
    else:
        command_str = command
        command_encoded = command.encode('utf-8')
    logger.debug("Call local command: {command_str}".format(command_str=command_str))

    command_env = os.environ.copy()
    command_env['LANG'] = 'en_US.utf-8'
    if env is not None:
        command_env.update(env)

    try:
        p = subprocess.Popen(
            command_encoded,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=command_env,
            shell=True,
        )
        stdout_str, stderr_str = [s for s in p.communicate(stdin_content)]
        if not log_only_error:
            logger.debug("Command stdout: %s", stdout_str)

        logger.debug("Command stderr: %s", stderr_str)
        logger.debug("Command exit code: %s", p.returncode)
        return p.returncode, stdout_str, stderr_str
    except Exception as e:
        logger.debug("Exception:", exc_info=True)
        raise Exception("Failed to execute local command '%s': %s" % (command_str, unicode(e)))
