# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""Tools to generate XML node describing platform"""
import logging

from xml.dom.minidom import Element
from DataSources import DBDataSource, DataSource, XmlServiceTemplateDataSource, XmlSubscriptionDataSource
from AccountGeneratorBase import AccountGeneratorBase
from AccountGenerator import AccountGenerator
from XMLExtensions import TextElement


logger = logging.getLogger(__name__)


class PlatformGenerator:

    """Platform XML node generator"""

    PLATFORM_NAME = 'OBAS Source Server'
    PLATFORM_VERSION = '1.0'
    DB_ACCOUNT_ACTIVE = '0'
    DB_ACCOUNT_ON_HOLD = '1'
    DB_ACCOUNT_FOR_APPROVAL = '2'
    DB_ACCOUNT_DELETED = '255'

    def __init__(self, export_inactive_subscriptions=False):
        self.export_inactive_subscriptions = export_inactive_subscriptions

    def getPlatform(self, account_id_list=None):
        """Generate platform document"""

        # root node for platform document
        rootNode = Element("platform")
        # add common elements
        nameNode = TextElement("name", self.PLATFORM_NAME)
        versionNode = TextElement("version", self.PLATFORM_VERSION)
        rootNode.appendChild(nameNode)
        rootNode.appendChild(versionNode)
        # add accounts hierarchy to doc
        accountGenerator = AccountGenerator(account_id_list)

        accountNode = accountGenerator.getPlatformAccount(AccountGeneratorBase.PROVIDER_ACCOUNT)
        if accountNode is not None:
            rootNode.appendChild(accountNode)

        return rootNode

    def getNodes(self):
        """Generate minimalistic platform document with OBAS nodes list"""
        logger.debug("Generate minimalistic platform document with OBAS nodes list.")
        # root node for platform document
        rootNode = Element("platform")
        # add common elements
        nameNode = TextElement("name", self.PLATFORM_NAME)
        versionNode = TextElement("version", self.PLATFORM_VERSION)
        rootNode.appendChild(nameNode)
        rootNode.appendChild(versionNode)
        # add accounts hierarchy to doc
        accountGenerator = AccountGenerator(
            [AccountGeneratorBase.PROVIDER_ACCOUNT], export_inactive_subscriptions=self.export_inactive_subscriptions
        )

        accountGenerator.hwNodes = self._get_nodes_info()

        accountNode = accountGenerator.getPlatformAccount(AccountGeneratorBase.PROVIDER_ACCOUNT)
        if accountNode is not None:
            accountNode.appendChild(self.getNameserversNode())
            rootNode.appendChild(accountNode)

        return rootNode

    def getOwnershipTree(self):
        """Generate account, subscription ownership document"""

        plans = (
            XmlServiceTemplateDataSource.TYPE_PLESK_DOMAIN,
            XmlServiceTemplateDataSource.TYPE_PLESK_CLIENT,
            XmlServiceTemplateDataSource.TYPE_DOMAIN_REGISTRATION,
            XmlServiceTemplateDataSource.TYPE_VIRTUOZZO_CONTAINER,
            XmlServiceTemplateDataSource.TYPE_VIRTUAL_MACHINE
        )
        account_statuses = (
            self.DB_ACCOUNT_ACTIVE,
            self.DB_ACCOUNT_ON_HOLD,
            self.DB_ACCOUNT_FOR_APPROVAL
        )
        # By default, the result of the execution of GROUP_CONCAT has the maximum allowed length of 1024.
        # If an account has multiple subscriptions, not all of them can fit in. For this reason, we are expanding the
        # maximum length of a specific response to 100000 by using the group_concat_max_len constant.
        query = """
            SET SESSION group_concat_max_len = 100000;
            SELECT account.account_no, account.vendor, GROUP_CONCAT(subscr_id) AS subscription_id_list
            FROM (account account), (subscr_base sb)
            WHERE (account.account_no <> 0) AND sb.account_no = account.account_no
              AND sb.plan_type IN (%s)
              AND sb.status IN (%s)
              AND account.status IN (%s)
            GROUP BY account.account_no
            UNION
            SELECT account.account_no, account.vendor, '' AS subscription_id_list
            FROM account
            WHERE account_type = 2
              AND account.status IN (%s);
        """ % (
            ','.join(plans), ','.join(XmlSubscriptionDataSource.getActiveSubscriptionStatusIds()),
            ','.join(account_statuses), ','.join(account_statuses)
        )

        ownership_info = DBDataSource.makeMysqlRequest(query)
        # TODO error processing
        return ownership_info

    def getPleskClientLogins(self, subscriptionIdList):
        """Generate plesk client logins document"""

        query = """
            SELECT subscr_id, plesk_login
            FROM plesk_client_cache
            WHERE subscr_id IN (%s);
        """ % (','.join(subscriptionIdList))

        plesk_client_logins_info = DBDataSource.makeMysqlRequest(query)
        # TODO error processing
        return plesk_client_logins_info

    @staticmethod
    def _get_nodes_info():
        hw_nodes = dict()

        active_nodes = DataSource().api('HSPC/MT/OM/HN').list(
            'attributes', ['id'],
            'filter', {
                'platform_id': DataSource.SUPPORTED_PLESK_PLATFORMS + DataSource.SUPPORTED_VPS_PLATFORMS
            }
        )
        for node_info in active_nodes:
            hw_id = node_info[0]
            hw_node = AccountGenerator.makeHwNodeById(hw_id)
            hw_nodes[int(hw_id)] = hw_node

        return hw_nodes

    @staticmethod
    def getNameserversNode():
        sshNsSrc = DBDataSource.getData(DBDataSource.TYPE_SSH_NAMESERVERS)

        nameServersNode = Element("nameservers")
        for ns in sshNsSrc.elements:
            nsNode = Element("ns")
            nsNode.setAttribute("id", ns["ns_id"])
            nsNode.setAttribute("type", ns["ns_type_id"])
            nsNode.appendChild(TextElement("ip", ns["ip_address"]))
            nsNode.appendChild(TextElement("hostname", ns["utf8_hostname"]))
            nsNode.appendChild(TextElement("named_conf", ns["named_conf"]))
            nameServersNode.appendChild(nsNode)

        return nameServersNode
