# (C)2017 Ingram Micro Inc. all Rights Reserved
"""Tools to generate XML node describing services"""
import logging

from AccountGeneratorBase import AccountGeneratorBase
from BasicGenerator import BasicGenerator
from SimpleTypes import DynamicObject
from DNSGenerator import DNSGenerator
from DataSources import DataSource, XmlSubscriptionDataSource
from XMLExtensions import TextElement, EnhElement
from Exceptions import ServiceError


logger = logging.getLogger(__name__)


class ServicesGenerator(BasicGenerator):

    def __init__(self):
        # data sources
        self.api = DataSource()

    def getServices(self, subscriptionId, mnIp=None):
        """Generate services node.

           subscriptionId: subscription ID to get services for"""

        logger.debug("Generate services node for '{subscriptionId}' and management node '{mnIp}'.".format(
            subscriptionId=subscriptionId, mnIp=mnIp
        ))
        subscription = self.api.api().get_subscr_info('subscr_id', subscriptionId, 'mn_ip', mnIp)
        if hasattr(subscription, 'error'):
            raise ServiceError(subscription.error)

        servicesNode = EnhElement("services")
        # services
        servNode = None
        if subscription.type == XmlSubscriptionDataSource.TYPE_VE:
            servNode = self._getContainer(subscription.ve)
        if subscription.type == XmlSubscriptionDataSource.TYPE_PLESK_HN:
            servNode = self._getContainer(subscription.ve)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_VM:
            servNode = self._getVirtMachine(subscription.vm)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_PD:
            servNode = self._getWebspaces(subscription.plesk_info)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_PC:
            servNode = self._getWebspaces(subscription.plesk_info)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_MISC:
            servNode = EnhElement("misc")
        elif subscription.type == XmlSubscriptionDataSource.TYPE_DM:
            servNode = self._getDomain(subscription.dom_info)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_POA:
            servNode = self._getPAsubscr(subscription.pa_subscr)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_DEDICATED_SERVER:
            servNode = self._getDSsubscr(subscription.ds)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_SSL:
            servNode = self._getSSL(subscription.ssl_certificate)
        elif subscription.type == XmlSubscriptionDataSource.TYPE_HN_VIRTUOZZO:
            servNode = EnhElement("virtuozzo_node")
            servNode.appendChild(self.__HwNode(subscription.hw))
        elif subscription.type == XmlSubscriptionDataSource.TYPE_PLESK_SERVER:
            servNode = EnhElement("plesk_node")
            servNode.appendChild(self.__HwNode(subscription.hw))

        if servNode:
            servicesNode.appendChild(servNode)
            servNode.insertBefore(TextElement("src_subscr_id", str(subscriptionId)), servNode.firstChild)
        else:
            return servicesNode

        # licenses
        if hasattr(subscription, 'licenses'):
            servNode.appendChild(self._getLicenses(subscription.licenses))
        # resources
        if hasattr(subscription, 'resources'):
            resourcesNode = servNode.addElement("resources")
            for res in subscription.resources:
                resNode = resourcesNode.addElement("resource")
                value = str(res.value)
                if res.is_unlim:
                    value = "-1"
                resNode.appendChild(TextElement("name", str(res.short_name)))
                if hasattr(res, 'poa_restype_id'):
                    resNode.appendChild(TextElement("pa_restype_id", str(res.poa_restype_id)))

                resNode.appendChild(TextElement("value", str(value)))
                if res.units:
                    resNode.appendChild(TextElement("units", str(res.units)))

        if hasattr(subscription, 'resource_usage'):
            resusageNode = servNode.addElement("resource_usage")
            resusageNode.appendChild(TextElement("last_bill_date", str(subscription.resource_usage.last_b_date)))
            if hasattr(subscription.resource_usage, 'resource_stats'):
                rsNode = resusageNode.addElement("resources")
                for res in subscription.resource_usage.resource_stats:
                    for usage in res:
                        rNode = rsNode.addElement("resource")
                        rNode.attributes["pa_restype_id"] = str(usage.key)
                        for us in usage.value:
                            uNode = TextElement("usage", str(us.total))
                            uNode.attributes["time_stamp"] = str(us.time_stamp)
                            rNode.appendChild(uNode)

            if hasattr(subscription.resource_usage, 'traffic'):
                trafNode = resusageNode.addElement("traffic")
                for res in subscription.resource_usage.traffic:
                    usageNode = trafNode.addElement("usage")
                    usageNode.attributes["units"] = str(res.units)
                    usageNode.appendChild(TextElement("class_id", str(res.class_id)))
                    if hasattr(res, 'in_used'):
                        usageNode.appendChild(TextElement("in_used", str(res.in_used)))
                    if hasattr(res, 'out_used'):
                        usageNode.appendChild(TextElement("out_used", str(res.out_used)))
                    if hasattr(res, 'all_used'):
                        usageNode.appendChild(TextElement("all_used", str(res.all_used)))
            if hasattr(subscription.resource_usage, 'disk_usage'):
                diskNode = resusageNode.addElement("disk_usage")
                for res in subscription.resource_usage.disk_usage:
                    node = TextElement("usage", str(res.usage))
                    node.attributes["time_stamp"] = str(res.time_stamp)
                    node.attributes["units"] = str(res.units)
                    diskNode.appendChild(node)

        # attributes
        if hasattr(subscription, 'attributes'):
            servNode.appendChild(self._getAttributes(subscription))

        # Refactored DNSGenerator to work with xml data returned by mysql
        # # dns
        # if hasattr(subscription, 'used_domains'):
        #     dnsGenerator = DNSGenerator(subscription.used_domains, subscription.ownerId, subscription.reverse_dns)
        #     dnsNode = dnsGenerator.generate()
        #     if dnsNode != None:
        #         servNode.appendChild(dnsNode)
        # return servicesNode

    def _getSSL(self, ssl):
        servNode = EnhElement("ssl")

        if hasattr(ssl, 'csr_data'):
            csrdNode = servNode.addElement("csr_data")
            for attr_name in [arg for arg in dir(ssl.csr_data) if not arg.startswith('_')]:
                csrdNode.appendChild(TextElement(attr_name, str(getattr(ssl.csr_data, attr_name))))

        if hasattr(ssl, 'csr'):
            servNode.appendChild(TextElement("value", str(ssl.csr)))

        # if hasattr(ssl,'contact_data'):

        if hasattr(ssl, 'ext_attr'):
            extNode = servNode.addElement("ext_attr")
            for attr_name in [arg for arg in dir(ssl.ext_attr) if not arg.startswith('_')]:
                extNode.appendChild(TextElement(attr_name, str(getattr(ssl.ext_attr, attr_name))))

        if hasattr(ssl, 'contact_data'):
            contactsNode = servNode.addElement("contacts")
            for cnt_type in [arg for arg in dir(ssl.contact_data) if not arg.startswith('_')]:
                cntNode = contactsNode.addElement(cnt_type)
                contact = getattr(ssl.contact_data, cnt_type)
                for attr_name in [arg for arg in dir(contact) if not arg.startswith('_')]:
                    value = getattr(contact, attr_name)
                    if value == None:
                        value = ""
                    cntNode.appendChild(TextElement(attr_name, str(value).decode('utf-8')))

        servNode.appendChild(TextElement("product_uc", str(ssl.product_uc)))
        servNode.appendChild(TextElement("product", str(ssl.product)))
        servNode.appendChild(TextElement("software_type", str(ssl.software_type)))
        servNode.appendChild(TextElement("software_type_uc", str(ssl.software_type_uc)))
        servNode.appendChild(TextElement("server_count", str(ssl.server_count)))
        servNode.appendChild(TextElement("plugin_name", str(ssl.plugin_name)))
        servNode.appendChild(TextElement("plugin_id", str(ssl.plugin_id)))
        servNode.appendChild(TextElement("plugin_template", str(ssl.plugin_template)))
        servNode.appendChild(TextElement("status", str(ssl.status)))
        servNode.appendChild(TextElement("private_key", str(ssl.private_key)))
        servNode.appendChild(TextElement("certificate_body", str(ssl.certificate_body)))
        servNode.appendChild(TextElement("domain", str(ssl.domain)))
        servNode.appendChild(TextElement("date_expired", str(ssl.date_expired)))
        servNode.appendChild(TextElement("approver_email", str(ssl.approver_email)))
        servNode.appendChild(TextElement("date_issued", str(ssl.date_issued)))

        return servNode

    def _getAttributes(self, obj):
        attrsNode = EnhElement("attributes")
        if hasattr(obj, 'attributes'):
            for gr in obj.attributes:
                grNode = attrsNode.addElement("group")
                grNode.attributes["name"] = str(gr.name)
                grNode.attributes["id"] = str(gr.gr_id)
                for at in gr.attrs:
                    atNode = TextElement("attr", at.name)
                    atNode.attributes["id"] = str(at.attr_id)
                    grNode.appendChild(atNode)
        return attrsNode

    def _getDSsubscr(self, ds):
        servNode = EnhElement("dedicated_server")
        if hasattr(ds, 'datacenter'):
            dcNode = servNode.addElement("datacenter")
            dcNode.appendChild(TextElement("id", str(ds.datacenter.id)))
            dcNode.appendChild(TextElement("identification", str(ds.datacenter.identification)))
            dcNode.appendChild(TextElement("location", str(ds.datacenter.location)))
        if hasattr(ds, 'rack'):
            rNode = servNode.addElement("rack")
            rNode.appendChild(TextElement("id", str(ds.rack.id)))
            rNode.appendChild(TextElement("identification", str(ds.rack.identification)))
            rNode.appendChild(TextElement("capacity", str(ds.rack.capacity)))
        if hasattr(ds, 'switch'):
            sNode = servNode.addElement("switch")
            sNode.appendChild(TextElement("id", str(ds.switch.id)))
            sNode.appendChild(TextElement("identification", str(ds.switch.identification)))
            sNode.appendChild(TextElement("model", str(ds.switch.model)))
            sNode.appendChild(TextElement("brand", str(ds.switch.brand)))
            sNode.appendChild(TextElement("ports_num", str(ds.switch.ports_num)))
            sNode.appendChild(TextElement("snmp_ip", str(ds.switch.snmp_ip)))
            sNode.appendChild(TextElement("read_community", str(ds.switch.read_community)))

        dsNode = servNode.addElement("ds")
        dsNode.appendChild(self.__IPaddresesNode(ds.ip_addresses))

        dsNode.appendChild(self._getAttributes(ds))
        if ds.identification:
            dsNode.appendChild(TextElement("identification", str(ds.identification)))
        if ds.form_factor:
            dsNode.appendChild(TextElement("form_factor", str(ds.form_factor)))
        if ds.admin_login:
            dsNode.appendChild(TextElement("admin_login", str(ds.admin_login)))
        if ds.admin_password:
            dsNode.appendChild(TextElement("admin_password", str(ds.admin_password)))
        if ds.port:
            dsNode.appendChild(TextElement("port", str(ds.port)))
        if ds.ifname:
            dsNode.appendChild(TextElement("ifname", str(ds.ifname)))

        return servNode

    def _getPAsubscr(self, pa_subscr):
        servNode = EnhElement("paSubscr")
        servNode.appendChild(self.__HwNode(pa_subscr.hw))

        sNode = servNode.addElement("subscr_info")
        sNode.appendChild(TextElement("pa_subscr_id", str(pa_subscr.pa_subscr_id)))
        sNode.appendChild(TextElement("pa_owner_id", str(pa_subscr.owner_id)))
        return servNode

    def _getDomain(self, domain):
        servNode = EnhElement("domain")
        operation = 'registration'
        if domain.action == 3:
            operation = 'transfer'

        servNode.appendChild(TextElement("operation", str(operation)))
        servNode.appendChild(TextElement("name", str(domain.domain).decode('utf-8')))
        servNode.appendChild(TextElement("status", str(domain.reg_status_txt).decode('utf-8')))
        servNode.appendChild(TextElement("plugin_name", str(domain.plugin_name).decode('utf-8')))
        servNode.appendChild(TextElement("plugin_type", domain.plugin_type))
        servNode.appendChild(TextElement("id_protect", str(domain.id_protect)))
        if domain.registrar_lock.is_success:
            servNode.appendChild(TextElement("registrar_lock", str(domain.registrar_lock.value)))

        if domain.login:
            servNode.appendChild(TextElement("login", str(domain.login)))
        if domain.password:
            servNode.appendChild(TextElement("password", domain.password))
        servNode.appendChild(TextElement("real_expire_time", str(domain.real_expire_time)))

        nsesNode = servNode.addElement("nameservers")
        for ns in domain.ns_info:
            nsNode = nsesNode.addElement("ns")
            nsNode.appendChild(TextElement("host", str(ns.hostname)))
            nsNode.appendChild(TextElement("ip", str(ns.ip)))

        domextNode = servNode.addElement("domainExtData")
        for attr_name in [arg for arg in dir(domain.domain_extdata) if not arg.startswith('_')]:
            attr = getattr(domain.domain_extdata, attr_name)
            domextNode.appendChild(TextElement(attr_name, str(attr).decode('utf-8')))

        cntctsNode = servNode.addElement("contacts")
        for attr_name in [arg for arg in dir(domain.contacts) if not arg.startswith('_')]:
            cntNode = cntctsNode.addElement(attr_name)
            cntNode.appendChild(TextElement('id', str(getattr(domain.contacts_ids, attr_name))))
            contact = getattr(domain.contacts, attr_name)
            contactInfoFields = {
                'fname': 'firstName',
                'lname': 'lastName',
                'email': 'contactEmail',
                'phone': 'phone',
                'fax': 'fax'
            }
            addressFields = {
                'address': 'address1',
                'address2': 'address2',
                'zip': 'zip',
                'city': 'city',
                'state': 'state',
                'state_alt': 'state_alt',
                'country': 'country'
            }
            account = DynamicObject()
            account.title = ''
            account.mobile = ''
            ac = AccountGeneratorBase()

            for ci_attr in contactInfoFields:
                if hasattr(contact, ci_attr):
                    value = getattr(contact, ci_attr)
                    if not value:
                        value = ''
                    setattr(account, contactInfoFields[ci_attr], str(value).decode('utf-8'))
            cntNode.appendChild(ac._makeContactInfoNode(account, "contactInfo", ""))

            for ci_attr in addressFields:
                if hasattr(contact, ci_attr):
                    value = getattr(contact, ci_attr)
                    if not value:
                        value = ''
                    setattr(account, addressFields[ci_attr], str(value).decode('utf-8'))
            cntNode.appendChild(ac._makeAddressNode(account, "address"))

            otherNode = cntNode.addElement('other')

            for cattr in [arg for arg in dir(contact) if not arg.startswith('_')]:
                value = getattr(contact, cattr)
                if value:
                    value = str(value).decode('utf-8')
                else:
                    value = ""

                if not contactInfoFields.has_key(cattr) and not addressFields.has_key(cattr) and value:
                    otherNode.appendChild(TextElement(cattr, value))

            if hasattr(domain.contacts_extdata, attr_name):
                cntextNode = otherNode.addElement("contactExtData")
                ext_data = getattr(domain.contacts_extdata, attr_name)
                for ext_attr in [arg for arg in dir(ext_data) if not arg.startswith('_')]:
                    value = getattr(ext_data, ext_attr)
                    if value:
                        if value.__class__.__name__ == "structType":
                            intNode = cntextNode.addElement(ext_attr)
                            for int_ext_attr in [arg for arg in dir(value) if not arg.startswith('_')]:
                                int_value = getattr(value, int_ext_attr)
                                intNode.appendChild(TextElement(int_ext_attr, str(int_value).decode('utf-8')))
                        else:
                            cntextNode.appendChild(TextElement(ext_attr, str(value).decode('utf-8')))
        return servNode

    def _getWebspaces(self, plesk_info):
        plskNode = EnhElement("plesk_hosting")
        if hasattr(plesk_info, 'aux_accounts'):
            auxNode = plskNode.addElement("aux_accounts")
            for acc in plesk_info.aux_accounts:
                memNode = auxNode.addElement("member")

                memNode.attributes["type"] = 'customer'
                memNode.attributes["plesk_id"] = str(acc.id)
                memNode.appendChild(TextElement("companyName", str(acc.cname).decode('utf-8')))

                accNode = memNode.addElement("contactInfo")
                accNode.appendChild(TextElement("personName", str(acc.pname).decode('utf-8')))
                accNode.appendChild(TextElement("email", str(acc.email)))
                accNode.appendChild(TextElement("phone", str(acc.phone)))
                accNode.appendChild(TextElement("fax", str(acc.fax)))

                addNode = memNode.addElement("address")
                addNode.appendChild(TextElement("address1", str(acc.address).decode('utf-8')))
                addNode.appendChild(TextElement("zipcode", str(acc.pcode)))
                addNode.appendChild(TextElement("city", str(acc.city).decode('utf-8')))
                addNode.appendChild(TextElement("state", str(acc.state).decode('utf-8')))
                addNode.appendChild(TextElement("state_alt", str(acc.state_alt).decode('utf-8')))
                addNode.appendChild(TextElement("country", str(acc.country)))

                memNode.appendChild(TextElement("login", str(acc.login).decode('utf-8')))
                passwdNode = TextElement("password", str(acc.password))
                memNode.appendChild(passwdNode)
                passwdNode.attributes["type"] = 'plesk_' + str(acc.password_type)
                memNode.appendChild(TextElement("locale", str(acc.locale)))

                otherNode = memNode.addElement("other")
                otherNode.appendChild(TextElement("guid", str(acc.guid)))
                otherNode.appendChild(TextElement("creationDate", acc.cr_date))
                otherNode.appendChild(TextElement("description", str(acc.description).decode('utf-8')))
                otherNode.appendChild(TextElement("plesk-id", str(acc.id)))
                otherNode.appendChild(TextElement("plesk-owner-id", str(getattr(acc, "owner-id"))))
                otherNode.appendChild(TextElement("vendor-guid", str(getattr(acc, "vendor-guid"))))

        webspacesNode = plskNode.addElement("webspaces")
        for webs in plesk_info.webspaces:
            wsNode = webspacesNode.addElement("webspace")
            wsNode.appendChild(TextElement("plesk_id", str(webs.plesk_id)))
            wsNode.appendChild(TextElement("plesk_name", str(webs.plesk_name)))
            wsNode.appendChild(TextElement("owner_id", str(webs.owner_id)))

        if plesk_info.hw:
            plskNode.appendChild(self.__HwNode(plesk_info.hw))

        return plskNode

    def _getLicenses(self, licenses):
        licensesNode = EnhElement("licenses")
        for licset in licenses:
            licsetNode = licensesNode.addElement("licenses_set")

            licsetNode.attributes["api_login"] = licset.config.ka_api_login
            licsetNode.attributes["client_login"] = licset.config.ka_client_login
            licsetNode.attributes["server_host"] = licset.config.ka_srv_host
            # licsetNode.attributes["ka_api_password"]=licset.config.ka_api_password
            licsetNode.attributes["server_port"] = str(licset.config.ka_srv_port)

            for lic in licset.licenses:
                licNode = licsetNode.addElement("license")
                licNode.appendChild(TextElement("lic_key", str(lic.id)))
                licNode.appendChild(TextElement("lic_class", str(lic.lic_class)))
                licNode.appendChild(TextElement("content", str(lic.content)))

                featNode = licNode.addElement("features")
                for feature in lic.features:
                    featNode.appendChild(TextElement("feature", feature))

                if hasattr(lic, 'add_keys'):
                    addNode = licNode.addElement("add_keys")
                    for add_key in lic.add_keys:
                        keyNode = addNode.addElement("key")
                        keyNode.appendChild(TextElement("lic_key", str(add_key.id)))
                        keyNode.appendChild(TextElement("lic_class", str(add_key.lic_class)))
                        keyNode.appendChild(TextElement("content", str(add_key.content)))
                        featNode = keyNode.addElement("features")
                        for feature in add_key.features:
                            featNode.appendChild(TextElement("feature", feature))
        return licensesNode

    def _getContainer(self, ve):
        vpsNode = EnhElement("vps")

        vpsNode.appendChild(self.__HwNode(ve.hw))

        veNode = vpsNode.addElement("ve")
        veNode.appendChild(TextElement("virtualization", 'container'))
        veNode.appendChild(TextElement("ve_id", str(ve.id)))
        veNode.appendChild(TextElement("ve_eid", str(ve.eid)))
        veNode.appendChild(TextElement("ve_uuid", str(ve.uuid)))
        veNode.appendChild(TextElement("ve_name", str(ve.name)))
        veNode.appendChild(TextElement("std_template", str(ve.standard_template)))
        veNode.appendChild(TextElement("platform", str(ve.platform)))
        templatePrefix = ""
        if not ve.standard_template:
            templatePrefix = "."
        veNode.appendChild(TextElement("os", templatePrefix + ve.os_package))
        veNode.appendChild(self.__IPaddresesNode(ve.ip_addresses))

        appsNode = veNode.addElement("apps")
        for app in ve.applications:
            templateNode = TextElement("template", templatePrefix + app.package)
            appsNode.appendChild(templateNode)
            templateNode.attributes["os_package"] = templatePrefix + ve.os_package

        if hasattr(ve, 'backups'):
            backupsNode = vpsNode.addElement("backups")
            for backup in ve.backups:
                bNode = TextElement("backup", backup.backup_id)
                backupsNode.appendChild(bNode)
                if hasattr(backup, 'hw'):
                    bNode.attributes["hostname"] = backup.hw.ip_address
                    bNode.attributes["service_ve_ip"] = backup.hw.service_ve_ip
                    bNode.attributes["vza_passwd"] = backup.hw.root_passwd
                    bNode.attributes["vza_login"] = backup.hw.vza_login
                    bNode.attributes["vza_version"] = str(backup.hw.vza_version)

        if ve.is_plesk_vps:
            pleskNode = vpsNode.addElement("plesk")
            pleskNode.appendChild(TextElement("admin_passwd", ve.plesk_vps_pwd))

        if hasattr(ve, 'plesk_node'):
            pleskNode = vpsNode.addElement("plesk_node_in_vps")
            pleskNode.appendChild(self.__HwNode(ve.plesk_node))
        return vpsNode

    def _getVirtMachine(self, vm):
        vpsNode = EnhElement("vps")

        vpsNode.appendChild(self.__HwNode(vm.hw))

        veNode = vpsNode.addElement("ve")
        veNode.appendChild(TextElement("virtualization", 'hyperv'))
        veNode.appendChild(TextElement("name", str(vm.name)))
        veNode.appendChild(TextElement("eid", str(vm.eid)))

        veNode.appendChild(self.__IPaddresesNode(vm.ip_addresses))

        return vpsNode

    def __IPaddresesNode(self, list):
        addressesNode = EnhElement("ip_addresses")
        for ip in list:
            ipNode = addressesNode.addElement("ip")
            ipNode.appendChild(TextElement("ip_address", ip.ip_address))
            if ip.subnet_prefix:
                ipNode.appendChild(TextElement("subnet_prefix", ip.subnet_prefix))
            if ip.netmask:
                ipNode.appendChild(TextElement("netmask", ip.netmask))
            if ip.gateway:
                ipNode.appendChild(TextElement("gateway", ip.gateway))
        return addressesNode

    def __HwNode(self, hw):
        hwNode = EnhElement("hw")
        if hasattr(hw, "hostname"):
            hwNode.appendChild(TextElement("hostname", str(hw.hostname)))
        if hasattr(hw, "vza_login"):
            hwNode.appendChild(TextElement("vza_login", str(hw.vza_login)))
        if hasattr(hw, "root_passwd"):
            hwNode.appendChild(TextElement("vza_passwd", str(hw.root_passwd)))
        if hasattr(hw, "os"):
            hwNode.appendChild(TextElement("os", str(hw.os)))
        if hasattr(hw, "vza_version"):
            hwNode.appendChild(TextElement("vza_version", str(hw.vza_version)))
        if hasattr(hw, "ip_address"):
            hwNode.appendChild(TextElement("hostname", str(hw.ip_address)))
        if hasattr(hw, "service_ve_ip"):
            hwNode.appendChild(TextElement("service_ve_ip", str(hw.service_ve_ip)))

        if hasattr(hw, "username"):
            hwNode.appendChild(TextElement("username", str(hw.username)))
        if hasattr(hw, "passwd") and hw.passwd:
            hwNode.appendChild(TextElement("passwd", str(hw.passwd)))

        if hasattr(hw, "plesk_port"):
            hwNode.appendChild(TextElement("plesk_port", str(hw.plesk_port)))
        if hasattr(hw, "key") and hw.key:
            hwNode.appendChild(TextElement("api_key_for_mn", str(hw.key)))

        return hwNode
