#!/usr/bin/env python
# (C)2017 Ingram Micro Inc. all Rights Reserved
# -*- coding: utf-8 -*-
"""OBAS Agent entry file"""

import logging
import os
import sys
import xml.dom.minidom

from optparse import OptionParser

from Exceptions import APIversionError


logging.basicConfig(
    format='[%(asctime)s] [%(levelname)s]: %(message)s',
    level=logging.DEBUG,
    filename='debug.log',
)
logger = logging.getLogger(__name__)
logger.info("Agent was started: %s %s" % (sys.argv[0], " ".join("'%s'" % arg for arg in sys.argv[1:])))

try:
    from AgentImpl import AgentImpl
except ImportError, e:
    logger.error("Exception:", exc_info=True)
    if str(e).find('SOAPpy') != -1:
        print >> sys.stderr, "ERROR: SOAPpy is not installed, for installation run on OBAS node: \n # yum install SOAPpy"
        sys.exit(1)
    else:
        raise

OPT_EXTRACT = "extract"
OPT_ACCOUNT = "accountId"
OPT_SUBSCRIPTION = "subscriptionId"
OPT_MN_IP = "mnIp"
OPT_CONTENT = "contentId"
OPT_FILE = "file"
OPT_DISABLE = "disable"
OPT_ID = "Id"
OPT_DOMAIN = "domainId"
OPT_NAME = "Name"
OPT_HOST = "hostId"
OPT_EXPORT_INACTIVE = "exportInactive"

EXT_PLATFORM = "platform"
EXT_NODES = "nodes"
EXT_OWNERSHIP_TREE = "ownership"
EXT_PLESK_CLIENT_LOGINS = "plesk-client-logins"
EXT_ACCOUNT = "account"
EXT_SERVICES = "services"
EXT_SUBSCRIPTION = "subscription"
EXT_DNS_RECORDS = "dns-records"
EXT_ZONE = "dns-zone"
EXT_HOST = 'host'
EXT_TASK = 'task'


class BadArgs(Exception):

    """ Indicates invalid arguments
    """

    def __init__(self, msg):
        self.msg = msg

    def what(self):
        """ Get error message """
        return self.msg


def defined(args, arg):
    """ Checks if specified argument is present in argument list.

        args: optparse.Values containing various arguments.
        arg: argument (string) to check presence.
    """
    if not hasattr(args, arg) or getattr(args, arg) is None:
        return False
    return True


def getPlatform(args):
    """ Collect and return platform XML document.
        Params: args - ignored
    """
    logger.info("Collect and return platform XML document.")
    agent = AgentImpl()
    account_id_list = _get_id_list(args, OPT_ACCOUNT)
    return agent.getPlatform(account_id_list)


def getNodes(args):
    """ Collect and return minimalistic platform XML document with OBAS nodes info
        Params: args - ignored
    """
    logger.info("Collect and return minimalistic platform XML document with OBAS nodes info.")
    agent = AgentImpl()
    export_inactive = _get_bool_flag_value(args, OPT_EXPORT_INACTIVE)
    return agent.getNodes(export_inactive=export_inactive)


def getOwnershipTree(args):
    """ Return a JSON structure that holds IDs of accounts and subscriptions owned by them
        Params: args - ignored
    """
    logger.info("Return a JSON structure that holds IDs of accounts and subscriptions owned by them.")
    agent = AgentImpl()
    return agent.getOwnershipTree()


def getPleskClientLogins(args):
    """ Return a JSON structure that holds plesk client logins by subscription ids
        Params: args - ignored
    """
    logger.info("Return a JSON structure that holds plesk client logins by subscription ids.")
    agent = AgentImpl()
    subscription_id_list = _get_id_list(args, OPT_SUBSCRIPTION)
    return agent.getPleskClientLogins(subscription_id_list)


def getAccount(args):
    """Collect and return account XML document"""
    logger.info("Collect and return account XML document.")
    if not defined(args, OPT_ACCOUNT):
        raise BadArgs("specifying account ID is mandatory when extracting accounts")

    agent = AgentImpl()
    account_id_list = _get_id_list(args, OPT_ACCOUNT)
    subscription_id_list = _get_id_list(args, OPT_SUBSCRIPTION)
    return agent.getAccountList(account_id_list, subscription_id_list)


def getSubscription(args):
    """Collect and return account XML document"""
    logger.info("Collect and return account XML document.")
    if not defined(args, OPT_SUBSCRIPTION):
        raise BadArgs("specifying subscription ID is mandatory when extracting subscription")

    agent = AgentImpl()
    subscription_id_list = _get_id_list(args, OPT_SUBSCRIPTION)
    export_inactive = _get_bool_flag_value(args, OPT_EXPORT_INACTIVE)
    return agent.getSubscriptionList(subscription_id_list, export_inactive)


def getServices(args):
    """ Collect and return services XML document.
        Params: args - optparse.Values with various arguments. Only
                OPT_SUBSCRIPTION is recognized (and mandatory).
    """
    logger.info("Collect and return services XML document.")
    if not defined(args, OPT_SUBSCRIPTION):
        raise BadArgs("specifying subscription ID is mandatory when extracting services")

    agent = AgentImpl()
    return agent.getServices(getattr(args, OPT_SUBSCRIPTION), getattr(args, OPT_MN_IP))


def getDnsRecords(args):
    if not defined(args, OPT_DOMAIN):
        raise BadArgs("specifying domain ID is mandatory when extracting DNS records")

    domain_id_list = _get_id_list(args, OPT_DOMAIN)
    agent = AgentImpl()
    return agent.getDnsRecords(domain_id_list)


def disableAccount(args):
    if not defined(args, OPT_ID):
        raise BadArgs("specifying account ID is mandatory when disabling account")

    agent = AgentImpl()
    return agent.disableAccount(getattr(args, OPT_ID))


def disableServices(args):
    if not defined(args, OPT_ID):
        raise BadArgs("specifying subscription ID is mandatory when disabling subscription")

    agent = AgentImpl()
    return agent.disableServices(getattr(args, OPT_ID))


def disableZone(args):
    if not defined(args, OPT_NAME):
        raise BadArgs("specifying zone name is mandatory when disabling zone")

    agent = AgentImpl()
    return agent.disableZone(getattr(args, OPT_NAME))


def configurePleskNodeForTakeover(args):
    if not defined(args, OPT_HOST):
        raise BadArgs("specifying host id is mandatory when configuring takeoverable host")

    agent = AgentImpl()
    return agent.configurePleskNodeForTakeover(getattr(args, OPT_HOST))


def invalidExtract(args):
    """ Handle invalid extract parameter. Actually it means some kind of
        internal logic error, because OptionParser should accept only known
        values for this field.
    """
    error_message = "ERROR: unknown --extract parameter value: " + getattr(args, OPT_EXTRACT) + \
                    ". It seems like internal logic error."
    logger.error(error_message)
    print >>sys.stderr, error_message
    sys.exit(1)


def invalidDisable(args):
    """ Handle invalid disable parameter. Actually it means some kind of
        internal logic error, because OptionParser should accept only known
        values for this field.
    """
    error_message = "ERROR: unknown --disable parameter value: " + getattr(args, OPT_DISABLE) + \
                    ". It seems like internal logic error."
    logger.error(error_message)
    print >>sys.stderr, error_message
    sys.exit(1)


def waitForTasks(args):
    if not defined(args, OPT_ID):
        raise BadArgs("specifying tasks IDs is mandatory when waiting tasks")

    agent = AgentImpl()
    task_id_list = _get_id_list(args, OPT_ID)
    return agent.waitForTasks(task_id_list, attempts=30, timeout=10)


def main():
    parser = OptionParser()
    parser.add_option("--extract", dest=OPT_EXTRACT, type="choice",
                      choices=[
                          EXT_PLATFORM, EXT_OWNERSHIP_TREE, EXT_ACCOUNT, EXT_SERVICES, EXT_NODES,
                          EXT_SUBSCRIPTION, EXT_DNS_RECORDS, EXT_PLESK_CLIENT_LOGINS
                      ],
                      help="Extract info about entity WHAT", metavar="WHAT")
    parser.add_option("--account-id", dest=OPT_ACCOUNT,
                      help="Specify account ID", metavar="ID")
    parser.add_option("--subscription-id", dest=OPT_SUBSCRIPTION,
                      help="Specify subscription ID", metavar="ID")
    parser.add_option("--domain-id", dest=OPT_DOMAIN,
                      help="Specify domain ID", metavar="ID")
    parser.add_option("--mn-ip", dest=OPT_MN_IP,
                      help="Specify IP of management node")
    parser.add_option("--id", dest=OPT_ID,
                      help="Specify entity ID to disable", metavar="ID")
    parser.add_option("--name", dest=OPT_NAME,
                      help="Specify entity name to disable", metavar="NAME")
    parser.add_option("--host-id", dest=OPT_HOST,
                      help="Specify host id to configure", metavar="ID")
    parser.add_option("--file", dest=OPT_FILE,
                      help="Save output to FILE", metavar="FILE")
    parser.add_option("--disable", dest=OPT_DISABLE, type="choice",
                      choices=[EXT_ACCOUNT, EXT_SUBSCRIPTION, EXT_ZONE, EXT_HOST, EXT_TASK],
                      help="Disables some entity ID on the source system", metavar="ID")
    parser.add_option("--export-inactive", action="store_true", dest=OPT_EXPORT_INACTIVE,
                      help="Export inactive subscriptions")

    options, args = parser.parse_args()

    if len(args):  # we don't support positional arguments
        parser.error("Unrecognized option: `" + args[0] + "'")

    if not defined(options, OPT_EXTRACT) and not defined(options, OPT_DISABLE):  # --extract arg is mandatory
        parser.error("You should specify --extract or --disable option")

    switch = {
        EXT_PLATFORM: getPlatform,
        EXT_NODES: getNodes,
        EXT_OWNERSHIP_TREE: getOwnershipTree,
        EXT_PLESK_CLIENT_LOGINS: getPleskClientLogins,
        EXT_ACCOUNT: getAccount,
        EXT_SUBSCRIPTION: getSubscription,
        EXT_SERVICES: getServices,
        EXT_DNS_RECORDS: getDnsRecords,
    }

    disable_switch = {
        EXT_ACCOUNT: disableAccount,
        EXT_SUBSCRIPTION: disableServices,
        EXT_ZONE: disableZone,
        EXT_HOST: configurePleskNodeForTakeover,
        EXT_TASK: waitForTasks,
    }

    ret = None
    try:
        os.umask(077)
        if defined(options, OPT_EXTRACT):
            ret = switch.get(getattr(options, OPT_EXTRACT),
                             invalidExtract)(options)

        elif defined(options, OPT_DISABLE):
            ret = disable_switch.get(getattr(options, OPT_DISABLE),
                                     invalidDisable)(options)
    except BadArgs, e:
        logger.error("Exception:", exc_info=True)
        parser.error(e.what())
    except APIversionError, e:
        logger.error("Exception:", exc_info=True)
        print >>sys.stderr, "ERROR: " + e.what()
        sys.exit(1)

    if isinstance(ret, xml.dom.minidom.Document):
        if defined(options, OPT_FILE):
            out = open(getattr(options, OPT_FILE), "w")
        else:
            out = sys.stdout
        print >>out, ret.toprettyxml(indent="  ", encoding="UTF-8")
        out.close()
    else:
        print ret
    return 0


def _get_id_list(args, option):
    if defined(args, option):
        option_str = getattr(args, option)
        account_id_list = option_str.split(',') if option_str else None
    else:
        account_id_list = None
    return account_id_list


def _get_bool_flag_value(args, option):
    if defined(args, option):
        return getattr(args, option)

    return False


if __name__ == "__main__":
    sys.exit(main())
