# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core.checks.base import CheckAbstract
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.networking.ip_port import \
    get_formatted_host_with_unavailable_ports, get_unavailable_ports_for_host, PORTS_AVAILABILITY_STATUS_ERROR
from migrator.core.utils.registry import Registry
from migrator.products.plesk.checks.hosts_availability import \
    PLESK_COMMUNICATION_PORTS_LINUX, PLESK_COMMUNICATION_PORTS_WINDOWS
from migrator.settings import HTML_EOL, PANEL_TYPE_PLESK, SUPPORTED_OS_TYPES, OS_LINUX

logger = logging.getLogger(__name__)

# At the moment, there is no difference in check logic between OBAS MN and Plesk server
OBAS_MN_COMMUNICATION_PORTS_LINUX = PLESK_COMMUNICATION_PORTS_LINUX
OBAS_MN_COMMUNICATION_PORTS_WINDOWS = PLESK_COMMUNICATION_PORTS_WINDOWS


class Check(object):
    def run(self):
        global_context = Registry.get_instance().get_context()
        management_node_os = global_context.config.get('obas', 'os').lower()
        management_node_ip = global_context.config.get('obas', 'ip')
        unavailable_hosts = []
        unchecked_hosts = []
        error_message_parts = []
        try:
            if management_node_os not in SUPPORTED_OS_TYPES:
                raise NotImplemented()

            ports_to_check = OBAS_MN_COMMUNICATION_PORTS_LINUX if management_node_os == OS_LINUX else OBAS_MN_COMMUNICATION_PORTS_WINDOWS
            mn_test_result = get_unavailable_ports_for_host(management_node_ip, ports_to_check)
            if mn_test_result['status'] == PORTS_AVAILABILITY_STATUS_ERROR:
                unavailable_hosts.append(mn_test_result)

            plesks_to_check = PlatformReaderDB().get_migration_required_source_hosts(PANEL_TYPE_PLESK)
            for plesk_to_check in plesks_to_check:
                if not global_context.config.has_section(plesk_to_check.ip):
                    unchecked_hosts.append(plesk_to_check.ip)
                    continue

                plesk_node_os = global_context.config.get(plesk_to_check.ip, 'os').lower()
                plesk_node_ip = global_context.config.get(plesk_to_check.ip, 'ip')
                if plesk_node_os not in SUPPORTED_OS_TYPES:
                    raise NotImplemented()

                ports_to_check = PLESK_COMMUNICATION_PORTS_LINUX if plesk_node_os == OS_LINUX else PLESK_COMMUNICATION_PORTS_WINDOWS
                plesk_test_result = get_unavailable_ports_for_host(plesk_node_ip, ports_to_check)
                if plesk_test_result['status'] == PORTS_AVAILABILITY_STATUS_ERROR:
                    unavailable_hosts.append(plesk_test_result)

            # success
            if not unavailable_hosts and not unchecked_hosts:
                return CheckAbstract.STATUS_SUCCESS, _T('CHECK_OBAS_PANEL_AVAILABILITY_SUCCESS').format(
                    hostname=management_node_ip
                )

            # errors
            if unavailable_hosts:
                hosts_with_errors = ', '.join(
                    get_formatted_host_with_unavailable_ports(
                        unavailable_host['ip'],
                        unavailable_host['unavailable_ports']
                    ) for unavailable_host in unavailable_hosts
                )
                error_message_parts.append(_T('CHECK_OBAS_HOSTS_AVAILABILITY_ERRORS').format(hosts=hosts_with_errors))

            if unchecked_hosts:
                error_message_parts.append(_T('CHECK_OBAS_PANEL_AVAILABILITY_UNCHECKED_HOSTS').format(
                    hosts=', '.join(unchecked_hosts)
                ))

            error_message = _T('CHECK_OBAS_PANEL_AVAILABILITY_ERRORS_SUMMARY').format(
                hostname=management_node_ip,
                message_content=HTML_EOL + HTML_EOL.join(error_message_parts)
            )
            return CheckAbstract.STATUS_ERROR, error_message
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return CheckAbstract.STATUS_ERROR, _T("CHECK_OBAS_PANEL_AVAILABILITY_UNEXPECTED_RESULT").format(
                hostname=management_node_ip
            )
