# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import InternalError
from migrator.core.checks.base import CheckAbstract
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils.registry import Registry
from migrator.products.plesk.checks.lib.license import \
    get_plesk_license_status, PLESK_LICENSE_STATUS_VALID, PLESK_LICENSE_STATUS_INVALID, \
    PLESK_LICENSE_STATUS_UNEXPECTED, PLESK_LICENSE_STATUS_FORBIDDEN
from migrator.settings import HTML_EOL, PANEL_TYPE_PLESK

logger = logging.getLogger(__name__)


class Check(object):
    def run(self):
        # OBAS license does not affect a migration process, so let's check licenses of connected Plesks
        global_context = Registry.get_instance().get_context()
        management_node_ip = global_context.config.get('obas', 'ip')
        try:
            plesks_with_license_invalid = []
            plesks_with_license_unchecked = []
            plesks_with_license_restricted = []
            hosts = PlatformReaderDB().get_migration_required_source_hosts(PANEL_TYPE_PLESK)
            logger.debug(_T("CHECK_OBAS_PANEL_FOUND_HOSTS_TO_CHECK").format(
                hostname=management_node_ip, plesks=', '.join(host.name for host in hosts))
            )
            for source_plesk_model in hosts:
                plesk_license_status = get_plesk_license_status(
                    global_context.conn.get_source_server_object(source_plesk_model.ip)
                )
                if plesk_license_status == PLESK_LICENSE_STATUS_INVALID:
                    plesks_with_license_invalid.append(source_plesk_model.name)
                elif plesk_license_status == PLESK_LICENSE_STATUS_UNEXPECTED:
                    plesks_with_license_unchecked.append(source_plesk_model.name)
                elif plesk_license_status == PLESK_LICENSE_STATUS_FORBIDDEN:
                    plesks_with_license_restricted.append(source_plesk_model.name)
                elif plesk_license_status != PLESK_LICENSE_STATUS_VALID:
                    raise InternalError("Unable to check Plesk server's license: an internal error occured. Contact technical support to resolve the issue.")

            if not plesks_with_license_invalid and not plesks_with_license_unchecked \
                    and not plesks_with_license_restricted:
                return CheckAbstract.STATUS_SUCCESS, _T('CHECK_OBAS_PANEL_LICENSE_SUCCESS').format(
                    hostname=management_node_ip
                )

            error_message_parts = []
            if plesks_with_license_invalid:
                error_message_parts.append(
                    _T('CHECK_OBAS_PANEL_LICENSE_ERRORS').format(
                        hostname=management_node_ip,
                        plesks=', '.join(plesks_with_license_invalid)
                    )
                )

            if plesks_with_license_restricted:
                error_message_parts.append(
                    _T('CHECK_OBAS_PANEL_LICENSE_ACCESS_DENIED').format(
                        hostname=management_node_ip,
                        plesks=', '.join(plesks_with_license_restricted)
                    )
                )

            if plesks_with_license_unchecked:
                error_message_parts.append(
                    _T('CHECK_OBAS_PANEL_FOUND_PLESKS_LICENSES_WITH_UNEXPECTED_RESULT').format(
                        plesks=', '.join(plesks_with_license_unchecked),
                        hostname=management_node_ip,
                    )
                )

            return CheckAbstract.STATUS_ERROR, HTML_EOL.join(error_message_parts)
        except Exception:
            logger.error(_T("LOG_EXCEPTION"), exc_info=True)
            return CheckAbstract.STATUS_ERROR, _T('CHECK_OBAS_PANEL_LICENSE_UNEXPECTED').format(
                hostname=management_node_ip,
            )
