# (C)2017 Ingram Micro Inc. all Rights Reserved
import grp
import logging
import os
import pwd
import codecs

import migrator.products.obas

from migrator.core.product.lib.agent import ProductAgentService, get_agent_file_path
from migrator.settings import OA_APS_APP_SYSGROUP

logger = logging.getLogger(__name__)


class ObasMigrationAgent(ProductAgentService):
    def __init__(self, source_server):
        super(ObasMigrationAgent, self).__init__(source_server)

        self._agent_product_lib['src_dir'] = get_agent_file_path(migrator.products.obas, '')
        self._legacy_agent_dir = os.path.join(self._agent_product_lib['dst_dir'], 'obas-agent')
        self._legacy_agent_bin = os.path.join(self._legacy_agent_dir, 'agent')

    def create_dump(self, filename):
        return self._extract(filename, 'platform')

    def dump_plesk_nodes(self, filename):
        return self._extract(filename, 'nodes', export_inactive_subscriptions=True)

    def dump_account_list(self, account_id_list, subscription_id_list, filename):
        return self._extract(filename, 'account', account_id_list, subscription_id_list)

    def dump_subscription_list(self, subscription_id_list, filename, export_inactive=False):
        return self._extract(
            filename, 'subscription', subscription_id_list=subscription_id_list,
            export_inactive_subscriptions=export_inactive
        )

    def dump_records_list(self, domain_id_list, filename):
        return self._extract(filename, 'dns-records', domain_id_list=domain_id_list)

    def get_subscription_ownership_data(self, filename):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --extract=ownership'
            data = runner.sh(command, options, log_output=False)

        with codecs.open(filename, 'w', 'utf-8') as fp:
            fp.write(data)

    def get_plesk_client_logins(self, subscription_id_list, filename):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --extract=plesk-client-logins --subscription-id=' + ','.join(
                map(str, subscription_id_list)
            )
            data = runner.sh(command, options, log_output=False)

        with codecs.open(filename, 'w', 'utf-8') as fp:
            fp.write(data)

    def dump_account_info(self, account_id, filename):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
            'account_id': account_id,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --extract=account --account-id={account_id}'
            data = runner.sh(command, options, log_output=False)

        with codecs.open(filename, 'w', 'utf-8') as fp:
            fp.write(data)

    def disable_subscription(self, subscription_id):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
            'subscription_id': subscription_id,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --disable=subscription --id={subscription_id}'
            runner.sh(command, options, log_output=False)

    def disable_dns_zone(self, zone_name, filename):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
            'zone_name': zone_name,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --disable=dns-zone --name={zone_name}'
            data = runner.sh(command, options, log_output=False)

        with codecs.open(filename, 'w', 'utf-8') as fp:
            fp.write(data)

    def configure_plesk_node_for_takeover(self, host_id):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
            'host_id': host_id,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --disable=host --host-id={host_id}'
            runner.sh(command, options, log_output=False)

    def _extract(self, filename, object_name, account_id_list=None, subscription_id_list=None, domain_id_list=None,
                 export_inactive_subscriptions=False):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
            'object_name': object_name,
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --extract={object_name}'
            if account_id_list is not None:
                command += ' --account-id=' + ','.join(map(str, account_id_list))

            if subscription_id_list is not None:
                command += ' --subscription-id=' + ','.join(map(str, subscription_id_list))

            if domain_id_list is not None:
                command += ' --domain-id=' + ','.join(map(str, domain_id_list))

            if export_inactive_subscriptions:
                command += ' --export-inactive'

            data = runner.sh(command, options, log_output=False)
            with codecs.open(filename, 'w', 'utf-8') as fp:
                fp.write(data)

                # Need to give permission to read to the group 'oam'
                uid = pwd.getpwnam('root').pw_uid
                gid = grp.getgrnam(OA_APS_APP_SYSGROUP).gr_gid
                os.chown(filename, uid, gid)
                os.chmod(filename, 0640)

    def _upload_agent_product_lib(self):
        super(ObasMigrationAgent, self)._upload_agent_product_lib()
        if not self._server.is_windows():
            with self._server.runner() as runner:
                runner.sh('chmod u+x {agent}'.format(agent=self._legacy_agent_bin))

    def wait_for_tasks(self, task_id_list):
        options = {
            'agent': self._legacy_agent_bin,
            'agent_dir': self._legacy_agent_dir,
            'task_id_list': ','.join(map(str, task_id_list)),
        }
        with self._server.runner() as runner:
            command = 'cd {agent_dir} && {agent} --disable=task --id={task_id_list}'
            runner.sh(command, options, log_output=False)

