# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging


from migrator.core import CommandExecutionError
from migrator.core.hosting.dns.dns_forwarding import LinuxBindForwarding
from migrator.core.language import _T
from migrator.core.utils.filetypes.config_utils import get_source_type_section
from migrator.core.utils.registry import Registry
from migrator.products.obas.config import read_obas_settings

logger = logging.getLogger(__name__)


class ObasBindForwarding(LinuxBindForwarding):
    def __init__(self, migrator_server, runner, source_host_id, named_conf, tasks):
        super(ObasBindForwarding, self).__init__(migrator_server, runner, source_host_id, named_conf)
        self.tasks = tasks

    def _include_index_to_named_conf(self):
        named_cf_remote = self._named_conf
        named_cf_local = self._migrator_server.get_session_file_path("named.conf.%s" % (self._source_host_id,))
        self._runner.get_file(named_cf_remote, named_cf_local)

        with open(named_cf_local, 'r') as f:
            named_cf_content = f.readlines()

        include_str = """include "%s";\n""" % self._index_filename_remote_chrooted
        if include_str not in named_cf_content:
            try:
                index_of_obas_view_declaration = named_cf_content.index('view pbas_resolver {\n')
            except ValueError:
                index_of_obas_view_declaration = len(named_cf_content)
                named_cf_content.extend(['view pbas_resolver {\n', '\tmatch-clients      { any; };\n', '\tmatch-destinations { any; };\n', '\trecursion yes;\n', '};\n'])

            named_cf_content.insert(index_of_obas_view_declaration + 1, include_str)
            with open(named_cf_local, 'w') as f:
                f.writelines(named_cf_content)

            self._runner.upload_file(named_cf_local, named_cf_remote)

    def _restart_named(self):
        global_context = Registry.get_instance().get_context()
        if self.tasks:
            global_context.migrator.agent.wait_for_tasks([task.pid for task in self.tasks])

        try:
            super(ObasBindForwarding, self)._restart_named()
        except CommandExecutionError:
            # OBAS may not immediately update records on the name server. In this case, it is necessary to force
            # synchronization of records on the server. As this operation may take some time, it is better to use it if
            # inconsistency was detected.
            self._sync_obas_ns()
            super(ObasBindForwarding, self)._restart_named()

    def _sync_obas_ns(self):
        ns_id = self._source_host_id
        # An ID should look like 'ns_123'.
        if not ns_id.startswith('ns_'):
            logger.debug(_T("INVALID_NS_SERVER_ID").format(ns_id=ns_id))
            return

        ns_id = ns_id[3:]
        cmd = '/usr/sbin/hspc-dns-sync.pl -ns %s' % ns_id
        global_context = Registry.get_instance().get_context()
        section = get_source_type_section(global_context.config)
        source_server, _ = read_obas_settings(section)
        with source_server.runner() as runner:
            exit_code, stdout, stderr = runner.sh_unchecked(cmd)
            if exit_code != 0:
                raise CommandExecutionError(
                    message=_T("FAILED_TO_SYNC_NS"),
                    stdout=stdout,
                    stderr=stderr,
                    command_str=cmd,
                    exit_code=exit_code
                )
            else:
                logger.debug(_T("NS_SYNC_FINISHED_SUCCESSFULLY"))
                return
