# (C)2017 Ingram Micro Inc. all Rights Reserved
from lxml import etree

from migrator.core import _T, MigrationError
from migrator.core.data.model.subscription import has_any_of_service_types, get_service_property
from migrator.core.data.model.task import TaskModel
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.product.lib.dns_zone_manager import DnsZone
from migrator.core.utils.filetypes.config_utils import ConfigHostSection
from migrator.core.utils.networking.runner import get_runner
from migrator.products.obas.lib.dns_forwarding import ObasBindForwarding
from migrator.products.plesk.lib.dns_zone_manager import PleskDnsZoneManager
from migrator.settings import \
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass, PANEL_TYPE_BIND_OBAS, SOURCE_SERVICE_DNS


class ObasDnsZoneManager(PleskDnsZoneManager):
    def __init__(self):
        super(ObasDnsZoneManager, self).__init__()
        self._plesks_api = {}
        self.tasks = []

    def disable_zone_on_source(self, subscription, zone):
        if has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            self._disable_obas_zone(zone.name)
        elif has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]):
            plesk_api = self._get_source_plesk_api(subscription.source_info.external_id)
            self._disable_plesk_zone(zone.id, zone.is_alias, plesk_api)

    def get_subscription_zones(self, subscription):
        if has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):
            return self._get_obas_zones(subscription)
        elif has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.SHARED]):
            return super(ObasDnsZoneManager, self).get_subscription_zones(subscription)

        return []

    def get_dns_forwarding_impl(self, service):
        source_host = PlatformReaderDB().get_source_host(service.source_host_external_id)
        config_section = ConfigHostSection(self.global_context.config, source_host.ip)
        if source_host.panel == PANEL_TYPE_BIND_OBAS:
            named_conf = service.get_property('named_conf')
            runner = get_runner(config_section)
            return ObasBindForwarding(self.global_context.migrator_server, runner, source_host.external_id, named_conf, self.tasks)

        return super(ObasDnsZoneManager, self).get_dns_forwarding_impl(service)

    @staticmethod
    def _get_obas_zones(subscription):
        return [
            DnsZone(
                id=subscription.source_info.external_id,
                name=get_service_property(subscription.source_info, SOURCE_SERVICE_DNS, 'domain')
            )
        ]

    def _disable_obas_zone(self, zone_name):
        data_file = self.global_context.migrator_server.get_session_file_path(
            'tasks_{zone_name}.xml'.format(zone_name=zone_name.encode('idna'))
        )

        self.global_context.migrator.agent.disable_dns_zone(zone_name.encode('idna'), data_file)

        with open(data_file) as source:
            root = etree.ElementTree(file=source)

        self.tasks.extend([TaskModel(pid=x.attrib['id'], aps_id=None, operation_name=None) for x in root.findall('task')])

    def _get_source_plesk_api(self, subscription_external_id):
        source_host = PlatformReaderDB().get_source_host_by_subscription_service(
            subscription_id=subscription_external_id,
            service_type='web'
        )
        if source_host.external_id not in self._plesks_api:
            for plesk_id in self.global_context.conn.get_source_plesks():
                plesk_server = self.global_context.conn.get_source_server_object(plesk_id)
                ip = plesk_server.ip()
                if ip == source_host.ip:
                    self._plesks_api[source_host.external_id] = plesk_server.plesk_api()
                    break

        if source_host.external_id not in self._plesks_api:
            raise MigrationError(
                _T('FAILED_TO_GET_OBAS_PLESK_SERVER').format(hw_id=source_host.external_id, plesk_ip=source_host.ip)
            )

        return self._plesks_api[source_host.external_id]

