# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import xml.sax

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils import cached
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import PleskError
from migrator.products.plesk.lib.panel import split_version_string
from migrator.products.plesk.target.utils.plesk_api_utils import get_plesk_version
from migrator.settings import PANEL_TYPE_PLESK, SOURCE_SERVICE_DNS

logger = logging.getLogger(__name__)


class ServicePlanType(object):
    PLESK_CLIENT = 'Plesk Client'
    PLESK_DOMAIN = 'Plesk Domain'


class ObasSubscriptionExternalId(object):
    TYPE_HOSTING = 1
    TYPE_DNS = 2

    def __init__(self, subscription_type, subscription_id, domain_id):
        self.type = subscription_type
        self.subscription_id = subscription_id
        self.domain_id = domain_id

    def get_external_id(self):
        if self.type == self.TYPE_HOSTING:
            return self.subscription_id

        if self.subscription_id is not None:
            return 'dns_%s_%s' % (self.domain_id, self.subscription_id)
        else:
            return 'dns_%s' % self.domain_id

    @classmethod
    def from_string(cls, external_id):
        if external_id.isdigit():
            return cls(ObasSubscriptionExternalId.TYPE_HOSTING, external_id, None)
        else:
            id_parts = external_id.rsplit('_', 2)
            if len(id_parts) < 2 or id_parts[0] != SOURCE_SERVICE_DNS:
                raise Exception(
                    _T("INVALID_OBAS_SUBSCRIPTION_EXTERNAL_ID_FORMAT").format(external_id=external_id)
                )

            if len(id_parts) == 3:
                subscription_id = id_parts[2]
            else:
                subscription_id = None

            return cls(ObasSubscriptionExternalId.TYPE_DNS, subscription_id, id_parts[1])

    @classmethod
    def from_data(cls, subscription_type, subscription_id=None, domain_id=None):
        return cls(subscription_type, subscription_id, domain_id)


class ObasExporterBase(object):
    def __init__(self):
        self.global_context = Registry.get_instance().get_context()
        self.agent = self.global_context.migrator.agent

        self._plesk_subscription_logins = None
        self._db = PlatformReaderDB()

        self._source_hosts = self._db.get_migration_required_source_hosts(PANEL_TYPE_PLESK)
        self._source_plesks = {}
        self._plesks_api = {}
        self._exporters = {}

        source_by_ip = dict((h.ip, h) for h in self._source_hosts)
        for plesk_id in self.global_context.conn.get_source_plesks():
            plesk_server = self.global_context.conn.get_source_server_object(plesk_id)
            ip = plesk_server.ip()
            if ip in source_by_ip:
                hw_id = source_by_ip[ip].external_id
                self._source_plesks[hw_id] = plesk_server

    def _init_plesk_exporter(self, hw_id, get_plesk_exporter_class_func, target_service_type):
        if hw_id not in self._source_plesks:
            source_by_hw_id = dict((h.external_id, h) for h in self._source_hosts)
            raise MigrationError(
                _T('FAILED_TO_GET_OBAS_PLESK_SERVER').format(hw_id=hw_id, plesk_ip=source_by_hw_id[hw_id]))

        if hw_id not in self._plesks_api:
            self._plesks_api[hw_id] = self._source_plesks[hw_id].plesk_api()

        if hw_id not in self._exporters:
            plesk = self._source_plesks[hw_id]
            self._exporters[hw_id] = get_plesk_exporter_class_func(target_service_type, plesk)(plesk)

    def _get_plesk_id_by_subscription(self, subscription):
        hw_id = subscription.services[0].source_host_external_id
        # TODO: Move to plesk exporter
        version = self._get_plesk_version(hw_id)

        if version < split_version_string('9.0'):
            operator = plesk_ops.DomainOperator.Get(
                filter=plesk_ops.DomainOperator.FilterByName8([subscription.name]),
                dataset=[plesk_ops.DomainOperator.Dataset.GEN_INFO]
            )
        elif version < split_version_string('10.0'):
            operator = plesk_ops.DomainOperator.Get(
                filter=plesk_ops.DomainOperator.FilterByDomainName([subscription.name]),
                dataset=[plesk_ops.DomainOperator.Dataset.GEN_INFO]
            )
        else:
            operator = plesk_ops.SubscriptionOperator.Get(
                filter=plesk_ops.SubscriptionOperator.FilterByName([subscription.name]),
                dataset=[plesk_ops.SubscriptionOperator.Dataset.GEN_INFO]
            )

        return self._plesks_api[hw_id].send(operator)[0].data[0]

    def _get_webspaces_list(self, hw_id, subscription):
        webspaces_list = []
        service_plan_type = PlatformReaderDB().get_subscription_service_plan(subscription.external_id).type
        if service_plan_type == ServicePlanType.PLESK_DOMAIN:
            plesk_id = self._get_plesk_id_by_subscription(subscription)
            webspaces_list = self._get_webspaces_list_for_plesk_subscription(hw_id, plesk_id, subscription.external_id)
        elif service_plan_type == ServicePlanType.PLESK_CLIENT:
            plesk_ids = self._get_plesk_client_subscription_ids(subscription)
            for plesk_id in plesk_ids:
                webspaces_list.extend(
                    self._get_webspaces_list_for_plesk_subscription(hw_id, plesk_id, subscription.external_id)
                )

        return webspaces_list

    def _get_webspaces_list_for_plesk_subscription(self, hw_id, plesk_subscription_id, obas_subscription_id):
        return self._exporters[hw_id].get_webspaces_list(plesk_subscription_id)

    def _get_plesk_client_subscription_ids(self, subscription):
        hw_id = subscription.services[0].source_host_external_id
        version = self._get_plesk_version(hw_id)
        if self._plesk_subscription_logins is None:
            self._plesk_subscription_logins = self._get_plesk_subscription_logins()

        account_login = self._plesk_subscription_logins[subscription.external_id]
        subscription_ids = self._get_account_subscriptions(version, hw_id, account_login)

        for customer_login in self._get_reseller_customers(version, hw_id, account_login):
            subscription_ids.extend(self._get_account_subscriptions(version, hw_id, customer_login))

        return subscription_ids

    def _get_reseller_customers(self, plesk_version, hw_id, account_login):
        # TODO: Move to plesk exporter
        if plesk_version < split_version_string('9.0'):
            return []

        if plesk_version < split_version_string('10.0'):
            operator = plesk_ops.ClientOperator.Get(
                filter=plesk_ops.ClientOperator.FilterOwnerLogin([account_login]),
                dataset=[plesk_ops.ClientOperator.Dataset.GEN_INFO]
            )
        else:
            operator = plesk_ops.CustomerOperator.Get(
                filter=plesk_ops.CustomerOperator.FilterByOwnerLogin([account_login]),
                dataset=[plesk_ops.CustomerOperator.Dataset.GEN_INFO]
            )

        try:
            result = self._plesks_api[hw_id].send(operator)
            return [x.data[1].login for x in result]
        except PleskError as e:
            if e.code == 1060:
                # Reseller has no clients
                return []

            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise e

    def _get_account_subscriptions(self, plesk_version, hw_id, account_login):
        # TODO: Move to plesk exporter
        if plesk_version < split_version_string('9.0'):
            operator = plesk_ops.DomainOperator.Get(
                filter=plesk_ops.DomainOperator.FilterByClientLogin([account_login]),
                dataset=[plesk_ops.DomainOperator.Dataset.GEN_INFO]
            )
        elif plesk_version < split_version_string('10.0'):
            operator = plesk_ops.DomainOperator.Get(
                filter=plesk_ops.DomainOperator.FilterByOwnerLogin([account_login]),
                dataset=[plesk_ops.DomainOperator.Dataset.GEN_INFO]
            )
        else:
            operator = plesk_ops.SubscriptionOperator.Get(
                filter=plesk_ops.SubscriptionOperator.FilterByOwnerLogin([account_login]),
                dataset=[plesk_ops.SubscriptionOperator.Dataset.GEN_INFO]
            )

        return [x.data[0] for x in self._plesks_api[hw_id].send(operator)]

    def _get_plesk_subscription_logins(self):
        subscription_id_list = [
            x.external_id for x in PlatformReaderDB().get_subscriptions_by_service_plan_type(
                ServicePlanType.PLESK_CLIENT
            )
        ]

        data_filename = self.global_context.migrator_server.get_session_file_path('plesk_subscription_logins.xml')
        self.agent.get_plesk_client_logins(subscription_id_list, data_filename)

        parser = xml.sax.make_parser()
        handler = PleskClientLoginsHandler()
        with open(data_filename) as xmldata:
            parser.setContentHandler(handler)
            parser.parse(xmldata)

        return handler.subscription_logins

    @staticmethod
    def _get_node_value(node, path, default=''):
        value = node.find(path)
        if value is None or value.text is None:
            return default

        return value.text

    @staticmethod
    def _get_xpath_result(elem, xpath):
        xpath_result = elem.xpath(xpath)
        if len(xpath_result) < 1:
            return None
        else:
            return xpath_result[0]

    @cached
    def _get_plesk_version(self, hw_id):
        return split_version_string(get_plesk_version(self._plesks_api[hw_id]))


class PleskClientLoginsHandler(xml.sax.ContentHandler):
    """
    Process MySQL query response that contains list of subscription logins

    Returns a dictionary of subscription_id => plesk_client_login
    """
    subscription_logins = {}
    current_subscription = None
    current_element_name = None

    def startElement(self, name, attrs):
        if name == "field" and attrs['name'] == 'subscr_id':
            self.current_element_name = 'subscription'
        elif name == "field" and attrs['name'] == 'plesk_login':
            self.current_element_name = 'login'

    def endElement(self, name):
        if name == "field":
            if self.current_element_name == 'subscription':
                if not self.current_text.isdigit():
                    raise ValueError("Unable to read subscription ID from string '%s'" % self.current_text)

                self.current_subscription = self.current_text
                self.subscription_logins[self.current_subscription] = []
            elif self.current_element_name == 'login':
                self.subscription_logins[self.current_subscription] = self.current_text

    def characters(self, content):
        self.current_text = content
