# (C)2017 Ingram Micro Inc. all Rights Reserved
import time

from lxml import etree

from migrator.core.language import _T
from migrator.core import MigrationError, InternalError
from migrator.core.data.format.billing import BillingReader
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.data.model.billing import AccountBillingModel, BankAccountModel, CreditCardModel, \
    CreditCardInfoModel, SubscriptionBillingModel, billing_type_period, SourceSubscriptionStatus
from migrator.core.data.model.account import AddressModel, PhoneModel
from migrator.products.obas.lib.export.base import ObasSubscriptionExternalId, ObasExporterBase


class ObasBillingExporter(ObasExporterBase, BillingReader):
    def __init__(self):
        super(ObasBillingExporter, self).__init__()

        self.export_data = []
        self._set_page_size(self.DEFAULT_PAGE_SIZE)
        self.source_system_reader = PlatformReaderDB()

    def get_next_account_page(self, subscription_id_list):
        account_objects = [self.source_system_reader.get_subscription(item_id).account for item_id in subscription_id_list]
        account_id_list = list({account.external_id for account in account_objects})

        subscription_id_list = [
            ObasSubscriptionExternalId.from_string(subscription_id).subscription_id
            for subscription_id in subscription_id_list
        ]

        for id_page in self._paginate(account_id_list):
            account_id_range = [str(aggregate(id_page)) for aggregate in min, max]
            filename = 'account_info_{file_id}.xml'.format(file_id='_'.join(account_id_range))
            self.agent.dump_account_list(id_page, filter(None, subscription_id_list), filename)
            self.export_data.append(filename)

            with open(filename) as source:
                root = etree.ElementTree(file=source)
                yield [self._create_account_billing_object(account_id, root) for account_id in id_page]

    def get_next_subscription_page(self, subscription_id_list):
        for id_list in self._paginate(list(subscription_id_list)):
            yield [self.get_subscription(s) for s in id_list]

    def get_account(self, account_id):
        """

        :param account_id: str
        :return: migrator.core.data.model.billing.AccountModel
        """
        data_file = self.global_context.migrator_server.get_session_file_path(
            'account_info_{account_id}.xml'.format(account_id=account_id)
        )
        if data_file not in self.export_data:
            self.agent.dump_account_info(account_id, data_file)
            self.export_data.append(data_file)

        with open(data_file) as source:
            root = etree.ElementTree(file=source)

        return self._create_account_billing_object(account_id, root)

    def _create_account_billing_object(self, account_id, root):
        """Create a AccountBilingModel using XML export data"""
        accountElement = root.find('/account[@id="%s"]' % account_id)
        if accountElement is None:
            raise MigrationError(_T("UNABLE_TO_FIND_ACCOUNT_XMLDATA").format(account_id=account_id))

        bank_accounts = []
        for bank_account in accountElement.findall('./paymentTools/BankAccounts/bankAccount'):
            bank_accounts.append(
                BankAccountModel(
                    account_number=self._get_node_value(bank_account, './account_number'),
                    payment_system=self._get_node_value(bank_account, './paymentSystem'),
                    bank_number=self._get_node_value(bank_account, './bank_code'),
                    holder_name=self._get_node_value(bank_account, './holder_name'),
                    use_for_recurring=int(self._get_node_value(bank_account, './use_for_recurring')),
                )
            )

        credit_cards = []
        for credit_card in accountElement.findall('./paymentTools/CreditCards/creditCard'):
            if credit_card.find('./validFrom') is None:
                start_date = ''
            else:
                start_date = self._get_card_date(credit_card.find('./validFrom'))

            credit_cards.append(
                CreditCardModel(
                    credit_card_info=CreditCardInfoModel(
                        payment_system=self._get_node_value(credit_card, './paymentSystem'),
                        card_number=self._get_node_value(credit_card, './cardNumber'),
                        card_holder_name=self._get_node_value(credit_card, './cardHolder'),
                        cvv=self._get_node_value(credit_card, './cvv'),
                        expiration_date=self._get_card_date(credit_card.find('./expirationDate')),
                        start_date=start_date,
                        issue_number=self._get_node_value(credit_card, './issueNumber'),
                    ),
                    address=self._get_address_model(credit_card.find('./address')),
                    phone=self._get_phone_model(credit_card.find('./phone')),
                    fax=PhoneModel(
                        country_code='',
                        area_code='',
                        phone_num='',
                        ext_num='',
                    ),
                    email=self._get_node_value(accountElement, './adminContactInfo/email'),
                    use_for_auto_payments=int(self._get_node_value(credit_card, './use_for_recurring')),
                )
            )

        # In OBAS, balance - how much money customers have, in OA - how many customers should pay.
        # So, due to this we invert exported balance value
        return AccountBillingModel(
            external_id=account_id,
            balance=-float(self._get_node_value(accountElement, './billingInfo/balance')),
            tax_reg_id=self._get_node_value(accountElement, './billingInfo/tax_reg_id'),
            credit_cards=credit_cards,
            bank_accounts=bank_accounts,
        )

    def get_subscription(self, subscription_id):
        """

        :param subscription_id: str
        :return: migrator.core.data.model.billing.SubscriptionModel
        """

        external_subscription_id = ObasSubscriptionExternalId.from_string(subscription_id).subscription_id

        if external_subscription_id is not None:
            subscription_node = self._dump_subscription(external_subscription_id)
            return self._create_subscription_model(subscription_id, subscription_node)

        subscription = self.source_system_reader.get_subscription(subscription_id)
        if subscription.parent_external_id:
            parent_subscription_node = self._dump_subscription(subscription.parent_external_id)
            model = self._create_subscription_model(subscription_id, parent_subscription_node)
            # set subscription period to None to pick suitable period from service plan when importing the subscription
            model.subscription_period = None
            model.subscription_period_type = None
            return model
        else:
            return SubscriptionBillingModel(subscription_id, SourceSubscriptionStatus.STATUS_ACTIVE)

    def _dump_subscription(self, subscription_id):
        data_file = self.global_context.migrator_server.get_session_file_path(
            'subscription_info_{subscription_id}.xml'.format(subscription_id=subscription_id)
        )
        if data_file not in self.export_data:
            # export the subscriptions that have need even if they are inactive
            self.agent.dump_subscription_list([subscription_id], data_file, True)
            self.export_data.append(data_file)

        with open(data_file) as source:
            root = etree.ElementTree(file=source)

        return self._get_subscription_node(root, subscription_id)

    def _create_subscription_model(self, subscription_id, subscription_node):
        if subscription_node is None:
            raise MigrationError(_T("SUBSCRIPTION_IS_NOT_PRESENT_ON_SOURCE").format(subscription=subscription_id))

        statuses_map = {
            "ACTIVE": SourceSubscriptionStatus.STATUS_ACTIVE,
            "DELETED": SourceSubscriptionStatus.STATUS_DELETED,
            "ON_HOLD": SourceSubscriptionStatus.STATUS_ON_HOLD,
            "GRACE": SourceSubscriptionStatus.STATUS_GRACED,
            "EXPIRED": SourceSubscriptionStatus.STATUS_EXPIRED,
            "TERMINATING": SourceSubscriptionStatus.STATUS_TERMINATING,
        }
        subscription_period_type = self._get_node_attribute(
            subscription_node, './subscriptionPeriod', 'type'
        ).lower()
        start_date = int(time.mktime(time.strptime(self._get_node_value(
            subscription_node, './creationDate'
        ), '%Y-%m-%dT%H:%M:%SZ')))
        expiration_date = int(time.mktime(time.strptime(self._get_node_value(
            subscription_node, './expirationDate'
        ), '%Y-%m-%dT%H:%M:%SZ')))
        status = self._get_node_value(
            subscription_node, './status'
        )
        if status not in statuses_map:
            raise InternalError("Unknown OBAS subscription status: '%s'" % status)

        return SubscriptionBillingModel(
            external_id=subscription_id,
            status=statuses_map[status],
            creation_date=start_date,
            expiration_date=expiration_date,
            subscription_period=int(self._get_node_value(subscription_node, './subscriptionPeriod')),
            subscription_period_type=billing_type_period[subscription_period_type],
            last_billing_date=start_date,
            next_billing_date=expiration_date,
        )

    def _get_phone_model(self, phone_node):
        return PhoneModel(
            country_code=self._get_node_value(phone_node, './countryCode'),
            area_code=self._get_node_value(phone_node, './areaCode'),
            phone_num=self._get_node_value(phone_node, './phoneNumber'),
            ext_num=self._get_node_value(phone_node, './extNumber'),
        )

    def _get_address_model(self, address_node):
        return AddressModel(
            zipcode=self._get_node_value(address_node, './zipcode'),
            city=self._get_node_value(address_node, './city'),
            country=self._get_node_value(address_node, './country'),
            state=self._get_node_value(address_node, './state'),
            address1=self._get_node_value(address_node, './address1'),
            address2=self._get_node_value(address_node, './address2'),
        )

    def _get_card_date(self, date_node):
        month = '0' + self._get_node_value(date_node, './month')
        year = self._get_node_value(date_node, './year')
        return "{month}/{year}".format(
            month=month[-2:],
            year=year[-2:],
        )

    @staticmethod
    def _get_node_attribute(node, path, attribute):
        value = node.find(path)
        if value is None or attribute not in value.attrib or value.attrib[attribute] is None:
            return ''

        return value.attrib[attribute]

    def _get_subscription_node(self, root, external_id):
        obas_subscription_external_id = ObasSubscriptionExternalId.from_string(external_id)
        if obas_subscription_external_id.type == ObasSubscriptionExternalId.TYPE_HOSTING:
            # Hosting subscription id is exported as is
            return self._get_xpath_result(
                root, './subscription[@id="{id}"]'.format(id=obas_subscription_external_id.subscription_id)
            )
        else:
            return self._get_xpath_result(
                root,
                './subscription/domainId[text()="{domain_id}"]/..'.format(
                    domain_id=obas_subscription_external_id.domain_id
                )
            )
