# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from lxml import etree

from migrator.core import _T, MigrationError
from migrator.core.data.format.dns import DnsReader
from migrator.core.data.model.dns import DnsZoneModel, DnsSoaRecordModel, DnsRecordModel
from migrator.products.obas.lib.export.base import ObasExporterBase, ObasSubscriptionExternalId

logger = logging.getLogger(__name__)


class ObasDnsExporter(ObasExporterBase, DnsReader):
    def __init__(self):
        super(ObasDnsExporter, self).__init__()

        self.export_data = []
        self._set_page_size(self.DEFAULT_PAGE_SIZE)

    def get_next_zone_page(self, subscription_id_list):
        for subscription_id_page in self._paginate(list(subscription_id_list)):
            domain_id_page = [ObasSubscriptionExternalId.from_string(id).domain_id for id in subscription_id_page]
            domain_id_range = [str(aggregate(domain_id_page)) for aggregate in min, max]
            filename = self.global_context.migrator_server.get_session_file_path(
                'dns_info_{file_id}.xml'.format(file_id='_'.join(domain_id_range))
            )
            self.agent.dump_records_list(domain_id_page, filename)
            self.export_data.append(filename)
            logger.debug(_T("READING_DNS_ZONES_INFORMATION_FROM_FILENAME").format(filename=filename))
            with open(filename) as source:
                root = etree.ElementTree(file=source)
                yield [self._create_dns_zone_object(domain_id, root) for domain_id in subscription_id_page]

    def _create_dns_zone_object(self, subscription_id, root):
        """
        :param subscription_id: str Example: u'dns_104'
        :param root: lxml.etree._ElementTree
        :return: DnsZoneModel
        """
        domain_id = ObasSubscriptionExternalId.from_string(subscription_id).domain_id
        domain_node = root.find('./domain[@id="%s"]' % domain_id)
        if domain_node is None:
            raise MigrationError(_T("DNS_ZONE_FOR_DOMAIN_WITH_ID_WAS_NOT_FOUND").format(domain_id=domain_id))

        is_dns_managed = domain_node.attrib['dns_managed']
        if is_dns_managed == '0':
            logger.debug(_T("FOUND_EMPTY_DNS_ZONE_WITHIN_EXPORT").format(domain_id=domain_id))

        domain_name = domain_node.attrib['name']
        records = []
        soa_record = None
        for record_node in domain_node.findall('./records/record'):
            if self._get_node_value(record_node, 'type') == 'soa':
                soa_record = self._create_soa_record_object(record_node)
            else:
                records.append(self._create_record_object(record_node))

        return DnsZoneModel(
            external_id=subscription_id,
            name=domain_name,
            records=records,
            soa=soa_record,
            enabled=domain_node.attrib['enabled'] == '1',
            dns_managed=domain_node.attrib['dns_managed'] == '1'
        )

    def _create_soa_record_object(self, soa_record_node):
        soa_record_content = self._get_node_value(soa_record_node, './value')
        # https://tools.ietf.org/html/rfc1035
        server, email, serial, refresh, retry, expire, minimum_ttl = soa_record_content.split(' ')
        return DnsSoaRecordModel(
            ttl=self._get_node_value(soa_record_node, './ttl'),
            refresh=refresh,
            retry=retry,
            expire=expire,
            minimum=minimum_ttl,
            serial=serial,
            email=email
        )

    def _create_record_object(self, record_node):
        record_type = self._get_node_value(record_node, './type')
        return DnsRecordModel(
            host=self._get_node_value(record_node, './name'),
            type=record_type,
            # add a dot at the end of value for CNAME records
            value=self._get_node_value(record_node, './value') + ('.' if record_type == 'cname' else ''),
            ttl=self._get_node_value(record_node, './ttl', None),
            priority=self._get_node_value(record_node, './priority', None),
            is_edit_allowed=self._get_node_value(record_node, './is_edit_allowed', None),
        )

    def get_zone(self, dns_subscription_id):
        raise NotImplementedError()
