# (C)2017 Ingram Micro Inc. all Rights Reserved
from contextlib import contextmanager
import itertools
import logging
import xml.sax

from lxml import etree

from migrator.core import _T, MigrationError
from migrator.core.data.model.billing import billing_type_period
from migrator.core.utils import is_subdomain_of_any, parse_bool
from migrator.core.utils.filetypes.config_utils import ConfigSection
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.dns import resolve
from migrator.core.data.format.platform import PagedPlatformReader
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils.registry import Registry
from migrator.core.data.model.account import \
    AccountModel, AuthModel, PersonalInfoModel, ContactModel, AddressModel, StaffMemberModel, PhoneModel
from migrator.core.data.model.reseller import ResellerModel, ProviderModel
from migrator.core.data.model.service import ServiceModel
from migrator.core.data.model.service_plan import ServicePlanModel
from migrator.core.data.model.service_property import ServicePropertyModel
from migrator.core.data.model.source_host import SourceHostModel
from migrator.core.data.model.subscription import SubscriptionModel
from migrator.odin_automation.platform.lib.db import get_system_locales
from migrator.products.obas.lib.export.base import ObasSubscriptionExternalId
from migrator.settings import \
    PANEL_TYPE_OBAS, PANEL_TYPE_PLESK, PANEL_TYPE_BIND_OBAS, SOURCE_SERVICE_VPS, PANEL_TYPE_VIRTUOZZO, \
    SOURCE_SERVICE_DNS_NS, SOURCE_SERVICE_DNS_CUSTOM_NS, SOURCE_SERVICE_DNS_NON_MANAGEABLE_NS, SOURCE_SERVICE_WEB, \
    SOURCE_SERVICE_DB, SOURCE_SERVICE_FTP, SOURCE_SERVICE_MAIL, SOURCE_SERVICE_DNS

logger = logging.getLogger(__name__)

OBAS_ADMIN_ID = u'1'

# Keep this list synchronized with SUPPORTED_PLESK_PLATFORMS in migrator\products\obas\agent\obas-agent\DataSources.py
SUPPORTED_PLESK_PLATFORMS = [
    '201',  # SW_PLATFORM_LINUX_PLESK7
    '202',  # SW_PLATFORM_WINDOWS_PLESK7
    '203'   # SW_PLATFORM_PLESK_UNIVERSAL
]

# Keep this list synchronized with SUPPORTED_VPS_PLATFORMS in migrator\products\obas\agent\obas-agent\DataSources.py
SUPPORTED_VPS_PLATFORMS = [
    '5',    # SW_PLATFORM_LINUX_X64 (Parallels Virtuozzo Containers for Linux (x86_64))
    '400'   # SW_PLATFORM_LINUX_PSBM (PCS6)
]


class ObasExporter(PagedPlatformReader):
    """Exporter object that retrieves objects in batches instead of exporting them all at once."""
    OBAS_MN_SOURCE_HOST_EXTERNAL_ID = 'obas_mn'

    def __init__(self):
        super(ObasExporter, self).__init__()

        self.global_context = Registry.get_instance().get_context()
        self.agent = self.global_context.migrator.agent
        self.export_data = []

        self._account_subscription_map = dict()
        self._accounts_and_owners = dict()  # This is a client => owner map
        self._parse_all_account_subscription_ids()

        self._hw_source_host_nodes = {}
        self._ns_source_host_nodes = set()
        self._parse_hw_node_source_list()

        self._is_partial = False

    @contextmanager
    @trace
    def filter_by_accounts(self, account_ids=None):
        old_accounts_and_owners = self._accounts_and_owners
        old_account_subscription_map = self._account_subscription_map

        if account_ids is not None:
            logger.debug('Filter by accounts %s' % account_ids)
            self._is_partial = True

            self._accounts_and_owners = dict()

            for acc, par in old_accounts_and_owners.iteritems():
                if acc in account_ids:
                    self._accounts_and_owners[acc] = par

            for acc, par in PlatformReaderDB().get_accounts_and_owner():
                if acc.external_id in account_ids:
                    self._accounts_and_owners[acc.external_id] = par.external_id

            while True:
                repeat = False
                for acc, par in old_accounts_and_owners.iteritems():
                    if par in self._accounts_and_owners.keys():
                        self._accounts_and_owners[acc] = par
                        repeat = True
                for acc, par in PlatformReaderDB().get_accounts_and_owner():
                    if par.external_id in self._accounts_and_owners.keys():
                        self._accounts_and_owners[acc.external_id] = par.external_id
                        repeat = True

                if not repeat:
                    break

            self._account_subscription_map = {
                acc: subscrs for acc, subscrs in old_account_subscription_map.iteritems() if acc in self._accounts_and_owners
            }
            for acc, subscrs in PlatformReaderDB().get_account_subscription_map().iteritems():
                if acc in self._accounts_and_owners:
                    if acc in self._account_subscription_map:
                        self._account_subscription_map[acc].update(subscrs)
                    else:
                        self._account_subscription_map[acc] = subscrs

            err_accounts = account_ids - set(self._account_subscription_map.keys())
            if err_accounts:
                raise MigrationError(_T('THERE_ARE_NO_SOURCE_ACCOUNTS').format(hosts=', '.join(err_accounts)))

        yield

        self._accounts_and_owners = old_accounts_and_owners
        self._account_subscription_map = old_account_subscription_map
        self._is_partial = False

    @contextmanager
    @trace
    def filter_by_hosts(self, host_ids=None):
        account_ids = None
        old_hw_source_host_nodes = self._hw_source_host_nodes

        if host_ids is not None:
            logger.debug('Filter by hosts %s' % host_ids)
            err_host = host_ids - set(old_hw_source_host_nodes.keys())
            if err_host:
                raise MigrationError(_T('THERE_ARE_NO_SOURCE_HOSTS').format(hosts=', '.join(err_host)))

            self._is_partial = True
            self._hw_source_host_nodes = {node: old_hw_source_host_nodes[node] for node in host_ids if node in old_hw_source_host_nodes}

            subscription_ids = set()
            for host_id in self._hw_source_host_nodes:
                source_host_node = self._hw_source_host_nodes[host_id]
                subscription_ids |= source_host_node['plesk_client_subscriptions']
                subscription_ids |= source_host_node['plesk_domain_subscriptions']

            account_ids = {acc for acc in self._account_subscription_map if subscription_ids & self._account_subscription_map[acc]}

            # We have to get account from database, also.
            # Because if these accounts/subscriptions were removed from source system,
            # they will not be removed from database, else
            account_ids |= set([x.external_id for x in PlatformReaderDB().get_accounts_by_hosts(host_ids)])

        with self.filter_by_accounts(account_ids=account_ids):
            yield

        self._hw_source_host_nodes = old_hw_source_host_nodes
        self._is_partial = False

    @property
    def is_partial(self):
        return self._is_partial

    @trace
    def export_next_account_page(self):
        """Retrieve source panel accounts in batches and return a batch upon each call"""

        # TODO use implementation from the base class
        # TODO _paginate(admins) + _paginate(resellers)
        # TODO: implement self._get_admin_account_list()

        # we have to skip owners of accounts/resellers in case of partial refresh
        # it not lead to skip update of resellers in case of partial refresh
        if not self.is_partial:
            resellers_and_admins = self.get_reseller_list()

            # export and commit admins first so we can refer to them from resellers
            admins = [x for x in resellers_and_admins if isinstance(x, ProviderModel)]
            if len(admins) > 0:
                yield admins

            # export and commit resellers so we can refer to them from accounts
            resellers = [
                x for x in resellers_and_admins if isinstance(x, ResellerModel) and not isinstance(x, ProviderModel)
            ]
            if len(resellers) > 0:
                for res in resellers:
                    res.parent_external_id = OBAS_ADMIN_ID

                yield resellers

        account_ids = list(set(self._accounts_and_owners.keys()) - set(self._accounts_and_owners.values()))
        account_ids.sort()
        account_batches = self._paginate(account_ids)

        for account_id_batch in account_batches:
            logger.debug('Joined System Bach exporting the account IDs ' + repr(account_id_batch))
            new_accounts = self.get_account_list(account_id_batch)
            for acc in new_accounts:
                acc.parent_external_id = self._accounts_and_owners[acc.external_id]

            yield new_accounts

    @trace
    def get_account_ownership_list(self):
        """Return a list of source panel accounts in AccountOwnership format"""
        return self._accounts_and_owners

    @trace
    def get_account_list(self, account_id_list):
        """Retrieve account information from source system and create an AccountModel for each one."""
        subscription_id_list = self._get_subscription_id_list(account_id_list)
        return self._get_account_list(account_id_list, subscription_id_list)

    @trace
    def get_service_plans_list(self, owner_id=OBAS_ADMIN_ID):
        service_plans = []
        # _get_nodes_data() dumps nodes and full provider's info with STs
        for service_template_node in self._get_nodes_data('./account/serviceTemplates/serviceTemplate'):
            # Subscription service domain id for DNS look like '1_com'
            if not service_template_node.attrib['id'].isdigit():
                continue

            service_plans.append(ServicePlanModel(
                external_id=service_template_node.attrib['id'],
                name=self._get_node_text(service_template_node, './name'),
                type=self._get_node_text(service_template_node, './type'),
            ))

        return service_plans

    @trace
    def get_reseller_list(self):
        """Return info about all resellers + admin"""
        account_id_list = set(self._accounts_and_owners.values())
        subscription_id_list = self._get_subscription_id_list(account_id_list)
        return self._get_account_list(account_id_list, subscription_id_list)

    @trace
    def get_source_hosts_list(self):
        source_hosts = []
        source_hosts.extend(self._get_hw_nodes_source_host_list())
        if not self.is_partial:
            source_hosts.extend(self._get_nameservers_source_host_list())
            source_hosts.append(self._get_obas_mn_source_host())

        return source_hosts

    @trace
    def _get_hw_nodes_source_host_list(self):
        source_hosts = []
        for source_host_node_id in self._hw_source_host_nodes:
            source_host_node = self._get_nodes_data('./account/hwNodes/hw[@id="%s"]' % source_host_node_id)[0]
            platform_id = self._get_node_text(source_host_node, './platformId')
            if platform_id in SUPPORTED_PLESK_PLATFORMS:
                panel = PANEL_TYPE_PLESK
            elif platform_id in SUPPORTED_VPS_PLATFORMS:
                panel = PANEL_TYPE_VIRTUOZZO
            else:
                raise Exception(_T("FAILED_TO_DETECT_PANEL_FOR_SOURCE_HOST").format(platform_id=platform_id))

            source_hosts.append(SourceHostModel(
                external_id=unicode(source_host_node.attrib['id']),
                name=self._get_node_text(source_host_node, './hostname'),
                ip=self._get_node_text(source_host_node, './ip'),
                panel=panel,
            ))

        return source_hosts

    @trace
    def _get_obas_mn_source_host(self):
        obas_section = ConfigSection(self.global_context.config, 'obas')
        obas_hostname = obas_section.get('ip')
        obas_ip = resolve(obas_hostname)
        if obas_ip is None:
            raise Exception(_T("FAILED_TO_RESOLVE_HOSTNAME_TO_AN_IP_ADDRESS").format(hostname=obas_hostname))

        return SourceHostModel(
            external_id=self.OBAS_MN_SOURCE_HOST_EXTERNAL_ID,
            name=obas_hostname,
            ip=obas_ip,
            panel=PANEL_TYPE_OBAS
        )

    @trace
    def _get_nameservers_source_host_list(self):
        source_hosts = []
        for source_host_node_id in self._ns_source_host_nodes:
            source_host_node = self._get_nodes_data('./account/nameservers/ns[@id="%s"]' % source_host_node_id)[0]

            source_host_model = None

            if source_host_node.attrib['type'] == 'SSHNS':
                source_host_model = SourceHostModel(
                    external_id="ns_" + source_host_node.attrib['id'],
                    name=self._get_node_text(source_host_node, './hostname'),
                    ip=self._get_node_text(source_host_node, './ip'),
                    panel=PANEL_TYPE_BIND_OBAS,
                )
            elif source_host_node.attrib['type'] == 'NMCNS':
                source_host_model = SourceHostModel(
                    external_id="cns_" + source_host_node.attrib['id'],
                    name=self._get_node_text(source_host_node, './hostname'),
                    ip=self._get_node_text(source_host_node, './ip'),
                )
            elif source_host_node.attrib['type'] == 'NMNS':
                source_host_model = SourceHostModel(
                    external_id="nmns_" + source_host_node.attrib['id'],
                    name=self._get_node_text(source_host_node, './hostname'),
                    ip=self._get_node_text(source_host_node, './ip'),
                )

            if source_host_model:
                source_hosts.append(source_host_model)

        return source_hosts

    @trace
    def _get_subscription_id_list(self, account_ownership_list):
        """Return a list of subscription IDs that belong to the child accounts of the specified account list"""
        ids = [
            self._account_subscription_map[x] for x in account_ownership_list if x in self._account_subscription_map
        ]
        subscription_id_list = list(itertools.chain(*ids))

        return subscription_id_list

    @trace
    def _parse_all_account_subscription_ids(self):
        """
        Read the account IDs and subscription IDs and store it for further reference.
        This is a lower-level method specific to OBAS agent implementation.

        Returns: a dictionary of account_id => (subscription1, ..., subscriptionN)
        """
        # read data from source panel only if it is not already cached
        data_filename = self.global_context.migrator_server.get_session_file_path('ownership_info.xml')
        self.agent.get_subscription_ownership_data(data_filename)

        parser = xml.sax.make_parser()
        handler = AccountSubscriptionsRequestHandler()
        with open(data_filename) as xmldata:
            parser.setContentHandler(handler)
            parser.parse(xmldata)

            self._account_subscription_map = {k: v for k, v in handler.account_subscriptions.items()}
            self._accounts_and_owners = handler.accounts_and_owners

        self._account_subscription_map[OBAS_ADMIN_ID] = set()

    @trace
    def _parse_hw_node_source_list(self):
        for source_host_node in self._get_nodes_data('./account/hwNodes/hw'):
            self._hw_source_host_nodes[source_host_node.attrib['id']] = {
                'plesk_client_subscriptions': set(self._get_node_text(source_host_node, './plesk_client_subscriptions').split(',')),
                'plesk_domain_subscriptions': set(self._get_node_text(source_host_node, './plesk_domain_subscriptions').split(',')),
            }

        for source_host_node in self._get_nodes_data('./account/nameservers/ns'):
            self._ns_source_host_nodes.add(source_host_node.attrib['id'])

    @trace
    def _get_account_list(self, account_id_list, subscription_id_list):
        accounts = []

        if account_id_list:
            account_data = self._get_account_list_xmltree(account_id_list, subscription_id_list or None)
            for account_id in account_id_list:
                xmldata = account_data.find('./account[@id="%s"]' % account_id)
                if xmldata is None:
                    # raise MigrationError("Account XML data is not found for account ID %s", account)
                    continue

                accounts.append(self._get_account(account_id, xmldata))

        return accounts

    def _get_account(self, account_id, account_xmldata):
        data_model = AccountModel
        if account_xmldata.attrib['type'] == u'reseller':
            data_model = ResellerModel
        elif account_xmldata.attrib['type'] == u'provider':
            data_model = ProviderModel

        return data_model(
            external_id=account_id,
            info=PersonalInfoModel(
                company_name=self._get_node_text(account_xmldata, './companyName'),
                admin_contact=self._get_contact_model(account_xmldata.find('./adminContactInfo')),
                tech_contact=self._get_contact_model(account_xmldata.find('./techContactInfo')),
                billing_contact=self._get_contact_model(account_xmldata.find('./billingContactInfo')),
                company_address=self._get_address_model(account_xmldata.find('./companyAddress')),
                # Get account language from first found OBAS account's staff member
                language=self._lang_to_locale(self._get_node_text(account_xmldata, './staffMembers/member/lang')),
            ),
            staff_members=self._get_staff_members_list(account_xmldata),
            subscriptions=self._get_subscriptions_list(account_xmldata),
        )

    def _get_subscriptions_list(self, account_xmldata):
        account_id = unicode(account_xmldata.attrib['id'])
        subscription_ids = self._account_subscription_map[account_id]

        subscription_domain_names = account_xmldata.xpath(
            './subscriptions/subscription/@name', smart_strings=False
        )
        standalone_domain_names = account_xmldata.xpath(
            './domainsWithoutSubscription/domain/name/text()', smart_strings=False
        )
        domain_names = set(map(unicode, subscription_domain_names + standalone_domain_names))

        subscriptions = []
        for subscription_node in account_xmldata.iterfind('./subscriptions/subscription'):
            if subscription_node.attrib['isDomainSubscription'] and \
               parse_bool(subscription_node.attrib['isDomainSubscription'].lower()):
                if self._need_to_export_dns_subscription(
                        domain=unicode(subscription_node.attrib['name']),
                        target_subscription_id=self._get_node_text(subscription_node, './targetSubscriptionId', None),
                        domain_names=domain_names,
                        subscription_ids=subscription_ids
                ):
                    subscriptions.append(self._get_dns_subscription(subscription_node))
            elif self._get_node_attribute(subscription_node, '.', 'type') == SOURCE_SERVICE_VPS:
                subscriptions.append(self._get_vps_subscription(subscription_node))
            else:
                subscriptions.append(self._get_web_hosting_subscription(subscription_node))

        # Create subscription models for domains which are not bound to any subscription in OBAS
        # (that are 'registered externally')
        for domain_node in account_xmldata.iterfind('./domainsWithoutSubscription/domain'):
            if self._need_to_export_dns_subscription(
                    domain=self._get_node_text(domain_node, './name'),
                    target_subscription_id=self._get_node_text(domain_node, './targetSubscriptionId', None),
                    domain_names=domain_names,
                    subscription_ids=subscription_ids
            ):
                subscriptions.append(self._get_standalone_dns_subscription(domain_node))

        return subscriptions

    def _get_dns_subscription(self, subscription_node):
        is_dns_managed = self._get_node_text(subscription_node, './isDnsManaged')
        domain_id = self._get_node_text(subscription_node, './domainId')
        subscription_period_type = self._get_node_attribute(subscription_node, './subscriptionPeriod', 'type').lower()
        subscription_period = self._get_node_text(subscription_node, './subscriptionPeriod')
        dns_service_properties = [
            ServicePropertyModel('domain', unicode(subscription_node.attrib['name'])),
            ServicePropertyModel('is_domain_billed', '1'),
            ServicePropertyModel('is_dns_managed', is_dns_managed),
        ]
        nameservers = map(unicode, subscription_node.xpath('./nameservers/ns/id/text()', smart_strings=False))
        ns_services = filter(None, [self._get_dns_ns_service(ns_id) for ns_id in nameservers])
        return SubscriptionModel(
            external_id=ObasSubscriptionExternalId.from_data(
                ObasSubscriptionExternalId.TYPE_DNS, unicode(subscription_node.attrib['id']), domain_id
            ).get_external_id(),
            name=unicode(subscription_node.attrib['name']),
            parent_external_id=self._get_node_text(subscription_node, './targetSubscriptionId', None),
            # Subscription service template id for DNS looks like '1_com'
            service_plan_external_id=subscription_node.attrib['serviceTemplateId'].split('_')[0],
            services=ns_services + [
                ServiceModel(
                    type=SOURCE_SERVICE_DNS,
                    source_host_external_id=self.OBAS_MN_SOURCE_HOST_EXTERNAL_ID,
                    properties=dns_service_properties
                ),
            ],
            period_type=billing_type_period[subscription_period_type],
            period=subscription_period
        )

    def _get_web_hosting_subscription(self, subscription_node):
        source_host_external_id = self._get_node_text(subscription_node, './hwId')
        subscription_id = unicode(subscription_node.attrib['id'])
        subscription_period_type = self._get_node_attribute(subscription_node, './subscriptionPeriod', 'type').lower()
        subscription_period = self._get_node_text(subscription_node, './subscriptionPeriod')
        return SubscriptionModel(
            external_id=ObasSubscriptionExternalId.from_data(
                ObasSubscriptionExternalId.TYPE_HOSTING, subscription_id=subscription_id
            ).get_external_id(),
            name=unicode(subscription_node.attrib['name']),
            service_plan_external_id=subscription_node.attrib['serviceTemplateId'],
            services=[
                ServiceModel(type=SOURCE_SERVICE_WEB, source_host_external_id=source_host_external_id),
                ServiceModel(type=SOURCE_SERVICE_MAIL, source_host_external_id=source_host_external_id),
                ServiceModel(type=SOURCE_SERVICE_FTP, source_host_external_id=source_host_external_id),
                ServiceModel(type=SOURCE_SERVICE_DB, source_host_external_id=source_host_external_id),
            ],
            period_type=billing_type_period[subscription_period_type],
            period=subscription_period
        )

    def _get_vps_subscription(self, subscription_node):
        source_host_external_id = self._get_node_text(subscription_node, './hwId')
        subscription_id = unicode(subscription_node.attrib['id'])
        subscription_period_type = self._get_node_attribute(subscription_node, './subscriptionPeriod', 'type').lower()
        subscription_period = self._get_node_text(subscription_node, './subscriptionPeriod')
        vps_service_properties = [
            ServicePropertyModel('ve_id', self._get_node_text(subscription_node, './veId')),
            ServicePropertyModel('ve_name', self._get_node_text(subscription_node, './veName')),
            ServicePropertyModel('ve_hostname', self._get_node_text(subscription_node, './veHostname')),
            ServicePropertyModel('ve_type', self._get_node_text(subscription_node, './veType'))
        ]
        return SubscriptionModel(
            external_id=ObasSubscriptionExternalId.from_data(
                ObasSubscriptionExternalId.TYPE_HOSTING, subscription_id=subscription_id
            ).get_external_id(),
            name=unicode(subscription_node.attrib['name']),
            service_plan_external_id=subscription_node.attrib['serviceTemplateId'],
            services=[
                ServiceModel(
                    type=SOURCE_SERVICE_VPS,
                    source_host_external_id=source_host_external_id,
                    properties=vps_service_properties
                )
            ],
            period_type=billing_type_period[subscription_period_type],
            period=subscription_period
        )

    def _get_standalone_dns_subscription(self, domain_node):
        # 'standalone' means that the domain has no Domain Registration subscription
        domain_name = self._get_node_text(domain_node, './name')
        domain_id = self._get_node_text(domain_node, './id')
        is_dns_managed = self._get_node_text(domain_node, './isDnsManaged')
        target_subscription = self._get_node_text(domain_node, './targetSubscriptionId', None)
        dns_service_properties = [
            ServicePropertyModel('domain', domain_name),
            ServicePropertyModel('is_domain_billed', '0'),
            ServicePropertyModel('is_dns_managed', is_dns_managed),
        ]
        nameservers = map(unicode, domain_node.xpath('./nameservers/ns/id/text()', smart_strings=False))
        ns_services = filter(None, [self._get_dns_ns_service(ns_id) for ns_id in nameservers])
        return SubscriptionModel(
            external_id=ObasSubscriptionExternalId.from_data(
                ObasSubscriptionExternalId.TYPE_DNS, domain_id=domain_id
            ).get_external_id(),
            name=domain_name,
            service_plan_external_id=0,
            parent_external_id=target_subscription,
            services=ns_services + [
                ServiceModel(
                    type=SOURCE_SERVICE_DNS,
                    source_host_external_id=self.OBAS_MN_SOURCE_HOST_EXTERNAL_ID,
                    properties=dns_service_properties
                ),
            ],
        )

    def _get_dns_ns_service(self, ns_id):
        ns = self._get_nodes_data('./account/nameservers/ns[@id="{ns_id}"]'.format(ns_id=ns_id))[0]
        if ns:
            if ns.attrib['type'] == 'SSHNS':
                service_properties = [ServicePropertyModel('named_conf', self._get_node_text(ns, './named_conf'))]
                return ServiceModel(SOURCE_SERVICE_DNS_NS, 'ns_' + ns_id, service_properties)

            elif ns.attrib['type'] == 'NMCNS':
                return ServiceModel(SOURCE_SERVICE_DNS_CUSTOM_NS, 'cns_' + ns_id)

            elif ns.attrib['type'] == 'NMNS':
                return ServiceModel(SOURCE_SERVICE_DNS_NON_MANAGEABLE_NS, 'nmns_' + ns_id)

    def _need_to_export_dns_subscription(self, domain, target_subscription_id, domain_names, subscription_ids):
        return not is_subdomain_of_any(domain, domain_names) and \
               not self._is_domain_linked_to_invalid_hosting_subscription(target_subscription_id, subscription_ids)

    @staticmethod
    def _is_domain_linked_to_invalid_hosting_subscription(target_subscription_id, subscription_ids):
        return target_subscription_id is not None and target_subscription_id not in subscription_ids

    @staticmethod
    def _get_node_text(node, path, default=''):
        value = node.find(path)
        if value is None or value.text is None:
            return default

        return unicode(value.text)

    def _get_contact_model(self, contact_node):
        return ContactModel(
            title=self._get_node_text(contact_node, './title'),
            first_name=self._get_node_text(contact_node, './firstName'),
            middle_name=self._get_node_text(contact_node, './middleName'),
            last_name=self._get_node_text(contact_node, './lastName'),
            phone=self._get_phone_model(contact_node.find('./phone')),
            fax=self._get_phone_model(contact_node.find('./fax')),
            email=self._get_node_text(contact_node, './email'),
        )

    def _get_phone_model(self, phone_node):
        return PhoneModel(
            country_code=self._get_node_text(phone_node, './countryCode'),
            area_code=self._get_node_text(phone_node, './areaCode'),
            phone_num=self._get_node_text(phone_node, './phoneNumber'),
            ext_num=self._get_node_text(phone_node, './extNumber'),
        )

    def _get_address_model(self, address_node):
        country = self._get_node_text(address_node, './country')
        if country.lower() in ['us', 'ca']:
            state = self._get_node_text(address_node, './state')
        else:
            state = self._get_node_text(address_node, './state_alt')

        return AddressModel(
            zipcode=self._get_node_text(address_node, './zipcode'),
            city=self._get_node_text(address_node, './city'),
            country=country,
            state=state,
            address1=self._get_node_text(address_node, './address1'),
            address2=self._get_node_text(address_node, './address2'),
        )

    def _get_staff_members_list(self, account_xmldata):
        staff_members = []
        for staff_member_node in account_xmldata.iterfind('./staffMembers/member'):
            staff_members.append(
                StaffMemberModel(
                    authentication=AuthModel(
                        login=self._get_node_text(staff_member_node, './contactInfo/email'),
                        password='',
                    ),
                    contact=self._get_contact_model(staff_member_node.find('./contactInfo')),
                    address=self._get_address_model(staff_member_node.find('./address')),
                    language=self._lang_to_locale(self._get_node_text(staff_member_node, './lang'))
                )
            )

        return staff_members

    def _get_nodes_data(self, xpath):
        """Read the whole source system XML data file, yield the XML nodes that match the XPath"""
        data_file = self.global_context.migrator_server.get_session_file_path('export_data.xml')
        if data_file not in self.export_data:
            self.agent.dump_plesk_nodes(data_file)
            self.export_data.append(data_file)

        with open(data_file) as source:
            parser = etree.XMLParser(collect_ids=False, resolve_entities=False)
            root = etree.ElementTree(parser=parser, file=source)

        return root.xpath(xpath, smart_string=False)

    def _get_account_list_xmltree(self, account_id_list, subscription_id_list):
        account_id_range = [aggregate(account_id_list) for aggregate in min, max]
        filename = 'account_info_{file_id}.xml'.format(file_id='_'.join(account_id_range))

        data_file = self.global_context.migrator_server.get_session_file_path(filename)
        if data_file not in self.export_data:
            self.agent.dump_account_list(account_id_list, subscription_id_list, data_file)
            self.export_data.append(data_file)

        logger.debug("Opening account data file '%s'", data_file)
        with open(data_file) as source:
            parser = etree.XMLParser(collect_ids=False, resolve_entities=False)
            root = etree.ElementTree(parser=parser, file=source)

        return root

    @staticmethod
    def _get_node_attribute(node, path, attribute):
        value = node.find(path)
        if value is None or attribute not in value.attrib or value.attrib[attribute] is None:
            return ''

        return unicode(value.attrib[attribute])

    @staticmethod
    def _lang_to_locale(language):
        """Map OBAS language code to OA locale. Unsupported language will be replaced with en_US"""
        return ([x for x in get_system_locales() if x.startswith(language)] + ['en_US'])[0]


class AccountSubscriptionsRequestHandler(xml.sax.ContentHandler):
    """
    Process MySQL query response that contains list of accounts and subscriptions

    Returns a dictionary of account_id => (subscription_id{0,n})
    """
    account_subscriptions = {}
    accounts_and_owners = {}
    current_account = None
    current_element_name = None
    current_text = ''

    def startElement(self, name, attrs):
        self.current_text = ''
        if name == 'field' and attrs['name'] == 'account_no':
            self.current_element_name = 'account'
        elif name == 'field' and attrs['name'] == 'vendor':
            self.current_element_name = 'vendor'
        elif name == 'field' and attrs['name'] == 'subscription_id_list':
            self.current_element_name = 'subscription'

    def endElement(self, name):
        if name == 'field':
            if self.current_element_name == 'account':
                if not self.current_text.isdigit():
                    raise ValueError("Unable to read account ID from string '%s'" % self.current_text)

                self.current_account = self.current_text
                if self.current_account not in self.account_subscriptions:
                    self.account_subscriptions[self.current_account] = set([])
            elif self.current_element_name == 'vendor':
                self.accounts_and_owners[self.current_account] = self.current_text
            elif self.current_element_name == 'subscription':
                if self.current_text.strip():
                    self.account_subscriptions[self.current_account] = set(self.current_text.split(','))

    def characters(self, content):
        self.current_text += content
