# (C)2017 Ingram Micro Inc. all Rights Reserved
import re

from collections import OrderedDict

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.product.lib.host import get_service_status_linux
from migrator.core.product.panel import PanelInfoBase, ProductBase
from migrator.core.utils import cached
from migrator.products.obas.export import supported_services_exporters_map
from migrator.products.obas.host import AgentInstaller
from migrator.products.obas.lib.dns_zone_manager import ObasDnsZoneManager
from migrator.products.obas.lib.suspend_subscription import suspend_subscription
from migrator.settings import SOURCE_PRODUCT_OBAS


class Product(ProductBase):
    dns_zone_manager = ObasDnsZoneManager()

    @classmethod
    def get_agent_instance(cls):
        return AgentInstaller.install()

    @classmethod
    def get_supported_services_exporters_map(cls):
        return supported_services_exporters_map

    @classmethod
    def get_dns_zone_manager(cls):
        return cls.dns_zone_manager

    @classmethod
    def suspend_subscription(cls, subscription):
        suspend_subscription(subscription)

    @classmethod
    def get_id(cls):
        return SOURCE_PRODUCT_OBAS


class PanelInfo(PanelInfoBase):
    @property
    def name(self):
        return 'Odin Business Automation Standard'

    @property
    def short_name(self):
        return 'obas'

    @staticmethod
    def is_installed(runner):
        return runner.file_exists('/etc/hspc-release')

    @property
    def url(self):
        return 'https://{ip}/hspc/pcc'.format(ip=self.runner.ssh.ip)

    @property
    def version(self):
        version_str = self._get_version()
        return version_str.split('-')[0]

    @property
    def patch(self):
        version_str = self._get_version()
        return version_str.split('-')[1]

    def get_service_list(self):
        return OrderedDict([
            ('Control Panel', dict(name_list=('hspcd',))),
            ('PowerDNS', dict(name_list=('pdns',))),
            ('VZAgent Collector', dict(name_list=('vzcoll',))),
            ('Task Manager', dict(name_list=('taskmand',))),
            ('Service Monitor', dict(name_list=('servicemon',))),
        ])

    def get_service_status(self, service_name):
        status_string, exit_code, stdout = get_service_status_linux(self.runner, service_name)

        # OBAS initscripts 'servicemon' and 'pdns' misbehave in a way that their exit code
        # does not match service status; we parse stdout instead
        obas_service_is_stopped = 'not running' in stdout or 'service disabled' in stdout
        if obas_service_is_stopped and exit_code == 0:
            # Fix exit code and status string for the stopped services
            status_string = _T("SERVICE_STATUS_STOPPED")
            exit_code = 1

        return status_string, exit_code

    @cached
    def _get_version(self):
        hspc_release = self.runner.sh(u'cat /etc/hspc-release')
        groups = re.search(r'.*\s(.*)\s.*\s', hspc_release).groups()
        if len(groups) < 1:
            raise MigrationError(_T("UNABLE_TO_DETERMINE_OBAS_VERSION"))

        version_str = groups[0]
        return version_str
