# (C)2017 Ingram Micro Inc. all Rights Reserved
import copy

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.product.export import get_product_exporter, get_product_content_exporter
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib.api.xmlrpc import operator as pleskapi
from migrator.products.plesk.lib.export.billing import PleskBillingExporter, Plesk9BillingExporter
from migrator.products.plesk.lib.export.dns import PleskDnsExporter
from migrator.products.plesk.lib.export.platform import \
    PleskExporter, Plesk10Exporter, Plesk9Exporter, Plesk8Exporter, Plesk178Exporter
from migrator.products.plesk.lib.export.web_content_shm import \
    PleskHostingContentExporter, Plesk10HostingContentExporter, Plesk9HostingContentExporter
from migrator.products.plesk.lib.export.web_hosting_shm import \
    PleskHostingExporter, Plesk115HostingExporter, Plesk10HostingExporter, Plesk9HostingExporter, Plesk178HostingExporter
from migrator.products.plesk.lib.export.mail_hosting_shm import \
    PleskMailExporter, Plesk9MailExporter, Plesk10MailExporter
from migrator.products.plesk.lib.export.mail_content_shm import \
    PleskMailContentExporter, Plesk9MailContentExporter, Plesk10MailContentExporter
from migrator.products.plesk.lib.panel import split_version_string
from migrator.settings import TargetServiceType

supported_services_exporters_map = {
    TargetServiceType.PLATFORM: {
        'settings': PleskExporter,
    },
    TargetServiceType.BILLING: {
        'settings': PleskBillingExporter,
    },
    TargetServiceType.SHM: {
        'settings': PleskHostingExporter,
        'content': PleskHostingContentExporter,
    },
    TargetServiceType.MAIL: {
        'settings': PleskMailExporter,
        'content': PleskMailContentExporter,
    },
    TargetServiceType.DNS: {
        'content': PleskDnsExporter
    }
}


def _get_plesk_version(plesk_server=None):
    if plesk_server is None:
        context = Registry.get_instance().get_context()
        plesk_server = context.conn.get_panel_server(context, 'plesk')

    plesk_api = plesk_server.plesk_api()

    dataset = [pleskapi.server.ServerOperator.Dataset.STAT]
    result = plesk_api.send(pleskapi.server.ServerOperator.Get(dataset))
    plesk_version_string = result.data.statistics.versions.plesk_version

    return split_version_string(plesk_version_string)


def _check_supported_plesk_version(version):
    if version >= split_version_string('17.9') or version < split_version_string('8.6'):
        raise MigrationError(_T("SOURCE_PLESK_VERSION_IS_NOT_SUPPORTED") % '.'.join(map(str, version)))


def get_plesk_exporter_class(service, plesk_server=None):
    version = _get_plesk_version(plesk_server)
    _check_supported_plesk_version(version)

    exporters_map = copy.deepcopy(supported_services_exporters_map)
    if version < split_version_string('9.0'):
        exporters_map[TargetServiceType.PLATFORM]['settings'] = Plesk8Exporter
        exporters_map[TargetServiceType.BILLING]['settings'] = Plesk9BillingExporter
        exporters_map[TargetServiceType.SHM]['settings'] = Plesk9HostingExporter
        exporters_map[TargetServiceType.MAIL]['settings'] = Plesk9MailExporter
    elif version < split_version_string('10.0'):
        exporters_map[TargetServiceType.PLATFORM]['settings'] = Plesk9Exporter
        exporters_map[TargetServiceType.BILLING]['settings'] = Plesk9BillingExporter
        exporters_map[TargetServiceType.SHM]['settings'] = Plesk9HostingExporter
        exporters_map[TargetServiceType.MAIL]['settings'] = Plesk9MailExporter
    elif version < split_version_string('11.0.10'):
        exporters_map[TargetServiceType.PLATFORM]['settings'] = Plesk10Exporter
        exporters_map[TargetServiceType.SHM]['settings'] = Plesk10HostingExporter
        exporters_map[TargetServiceType.MAIL]['settings'] = Plesk10MailExporter
    elif version < split_version_string('12.0'):
        exporters_map[TargetServiceType.SHM]['settings'] = Plesk115HostingExporter

    elif version > split_version_string('17.8'):
        exporters_map[TargetServiceType.PLATFORM]['settings'] = Plesk178Exporter
        exporters_map[TargetServiceType.SHM]['settings'] = Plesk178HostingExporter

    return get_product_exporter(service, exporters_map)


def get_plesk_content_exporter_class(service, plesk_server=None):
    version = _get_plesk_version(plesk_server)
    _check_supported_plesk_version(version)

    exporters_map = copy.deepcopy(supported_services_exporters_map)
    if version < split_version_string('10.0'):
        exporters_map[TargetServiceType.SHM]['content'] = Plesk9HostingContentExporter
        exporters_map[TargetServiceType.MAIL]['content'] = Plesk9MailContentExporter
    elif version < split_version_string('11.0.10'):
        exporters_map[TargetServiceType.SHM]['content'] = Plesk10HostingContentExporter
        exporters_map[TargetServiceType.MAIL]['content'] = Plesk10MailContentExporter

    return get_product_content_exporter(service, exporters_map)
