# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.utils.loggers.log import trace
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import Failure


@trace
def get_plesk_webspace_id_by_website(website_name, plesk_api):
    result = plesk_api.send(
        plesk_ops.SiteOperator.Get(
            dataset=[plesk_ops.SiteOperator.Dataset.GEN_INFO],
            filter=plesk_ops.SiteOperator.FilterByName([website_name])
        )
    )
    if not result or isinstance(result[0], Failure):
        return None

    return result[0].data[1].gen_info.webspace_id


@trace
def get_plesk_site_id(site_name, plesk_api):
    result = plesk_api.send(
        plesk_ops.SiteOperator.Get(
            dataset=[plesk_ops.SiteOperator.Dataset.GEN_INFO],
            filter=plesk_ops.SiteOperator.FilterByName([site_name])
        )
    )
    if not result or isinstance(result[0], Failure):
        return None

    return result[0].data[0]


@trace
def get_plesk_database(plesk_api, plesk_webspace_id, db_name, db_type):
    for result in plesk_api.send(
        plesk_ops.DatabaseOperator.GetDb(
            filter=plesk_ops.DatabaseOperator.GetDb.FilterByWebspaceId([plesk_webspace_id])
        )
    ):
        if isinstance(result, Failure):
            continue

        if result.data.name == db_name and result.data.type == db_type:
            return result.data

    return None


@trace
def delete_plesk_database(plesk_api, plesk_database_id):
    plesk_api.send(plesk_ops.DatabaseOperator.DelDb(
        filter=plesk_ops.DatabaseOperator.DelDb.FilterByDbIds([plesk_database_id]))
    )


@trace
def get_plesk_database_user(plesk_api, db_id, user_name):
    for result in plesk_api.send(
        plesk_ops.DatabaseOperator.GetDbUsers(
                filter=plesk_ops.DatabaseOperator.GetDbUsers.FilterByDbIds([db_id])
        )
    ):
        if isinstance(result, Failure):
            continue

        if result.data.login == user_name:
            return result.data

    return None


@trace
def get_protected_directory_id(plesk_api, plesk_site_id, name, header):
    for result in plesk_api.send(
        plesk_ops.ProtectedDirectoryOperator.Get(
            filter=plesk_ops.ProtectedDirectoryOperator.FilterBySiteId([plesk_site_id])
        )
    ):
        if isinstance(result, Failure):
            continue

        if result.data.name == name:
            return result.data.id

    return None


@trace
def get_plesk_webspace_info(plesk_api, webspace_id):
    operator = plesk_ops.SubscriptionOperator.Get(
        filter=plesk_ops.SubscriptionOperator.FilterById([webspace_id]),
        dataset=[plesk_ops.SubscriptionOperator.Dataset.HOSTING, plesk_ops.SubscriptionOperator.Dataset.GEN_INFO]
    )
    with plesk_api.logger.censor({'xml': [
        "/packet/webspace/get/result/data/hosting/vrt_hst/property/name[text()='ftp_password']/../value",
        "/packet/webspace/get/result/data/hosting/vrt_hst/property/name[text()='fp_admin_password']/../value",
        # exactly fp_admin_password NOT ftp_admin_password
    ]}):
        return plesk_api.send(operator)[0].data[1]

@trace
def get_plesk_webspace_info_by_guid(plesk_api, webspace_guid):
    operator = plesk_ops.SubscriptionOperator.Get(
        filter=plesk_ops.SubscriptionOperator.FilterByGuid([webspace_guid]),
        dataset=[plesk_ops.SubscriptionOperator.Dataset.HOSTING]
    )
    with plesk_api.logger.censor({'xml': [
        "/packet/webspace/get/result/data/hosting/vrt_hst/property/name[text()='ftp_password']/../value",
        "/packet/webspace/get/result/data/hosting/vrt_hst/property/name[text()='fp_admin_password']/../value",
        # exactly fp_admin_password NOT ftp_admin_password
    ]}):
        return plesk_api.send(operator)[0].data[1]


@trace
def get_plesk_sites_info(plesk_api, webspace_id=None, site_id=None, dataset=None):
    if webspace_id is None and site_id is not None:
        filter = plesk_ops.SiteOperator.FilterById([site_id])
    elif webspace_id is not None and site_id is None:
        filter = plesk_ops.SiteOperator.FilterByParentId([webspace_id])
    elif webspace_id is None and site_id is None:
        filter = None
    else:
        assert False

    with plesk_api.logger.censor({'xml': ["/packet/site/get/result/data/hosting/vrt_hst/property/name[text()='ftp_password']/../value",]}):
        return plesk_api.send(plesk_ops.SiteOperator.Get(filter=filter, dataset=dataset))


@trace
def get_plesk_domains_info(plesk_api, webspace_id):
    operator = plesk_ops.DomainOperator.Get(
        filter=plesk_ops.DomainOperator.FilterById([webspace_id]),
        dataset=[plesk_ops.DomainOperator.Dataset.HOSTING]
    )
    with plesk_api.logger.censor({'xml': ["/packet/domain/get/result/data/hosting/vrt_hst/property/name[text()='ftp_password']/../value"]}):
        return plesk_api.send(operator)[0].data[1]


@trace
def is_ftp_user_exists_for_webspace(plesk_api, ftp_user_name, webspace_id):
    operator = plesk_ops.FtpuserOperator.Get(
        filter=plesk_ops.FtpuserOperator.FilterByWebspaceId([webspace_id])
    )

    for result in plesk_api.send(operator):
        if result.data.name == ftp_user_name:
            return True, False

    webspace_properties = get_plesk_webspace_info(plesk_api, webspace_id).hosting.properties
    key = 'ftp_login'
    if key in webspace_properties and webspace_properties[key] and webspace_properties[key] == ftp_user_name:
        return True, True

    return False, False
