# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from itertools import groupby
from lxml import etree

from migrator.core import version_tuple
from migrator.core.utils import if_not_none, partition, default
from migrator.core.utils.filetypes.xml import elem, text_elem
from migrator.core.utils.lang.profiler import get_default_steps_profiler
from migrator.core.utils.networking.http_xml_client import HttpXmlClient
from migrator.core.language import _T
from migrator.products.plesk.lib.api.xmlrpc.core import PleskError

profiler = get_default_steps_profiler()


class UnsupportedOperationException(Exception):
    pass


class Client(HttpXmlClient):
    logger = logging.getLogger(__name__)

    def __init__(self, url, api_version, **kwargs):
        """
        Create Plesk API-RPC client

        url - url to Plesk API agent
        api_version - Plesk API version to use

        Additional keywords parameters are:
            Auth parameters:
                username - username to use when sending Plesk API requests
                password - password to use when sending Plesk API requests

            Alternate auth parameters:
                secret_key - existing Plesk API secret key registered for source_addr IP address
                source_addr - bind connection socket to the local address when sending Plesk API requests

            Misc:
                pretty_print - whether Plesk should format response packets or not (False by default)

        Example:
            c1 = Client(<url>, '1.6.7.0', username='admin', password=<password>)
            c2 = Client(<url>, '1.6.7.0', secret_key=<key>, source_addr='10.37.37.37')

        Note:
            Secret key must be registered in Plesk before usage.
            You can use secret_keys.PleskAPISecretKeysManager() for this task.
        """
        super(Client, self).__init__(url, kwargs.get('source_addr'))
        self.api_version = api_version

        self.username = kwargs.get('username')
        self.password = kwargs.get('password')
        self.secret_key = kwargs.get('secret_key')
        self.pretty_print = kwargs.get('pretty_print', False)

    def send(self, operation, **request_settings):
        return self.send_many(operation, **request_settings)[0]

    def send_many(self, *operations, **request_settings):
        # check all operators are supported in the client
        unsupported_operations = [
            u"%s.%s" % (op.operator_name, op.operation_name)
            for op in operations if not self._is_operation_supported(op)
        ]
        if len(unsupported_operations) > 0:
            raise UnsupportedOperationException(
                _T("OPERATIONS_ARE_NOT_SUPPORTED_BY_CLIENT") %
                (unsupported_operations, self.api_version)
            )

        # Many Plesk operations accept filter - set of identifiers to limit operation to.
        # However, when filter is empty Plesk threats it as "all objects". It is very easy
        # to get empty IDs list in filter and it is very dangerous when it is used with
        # destructive operations like 'del' or 'update'. So this API wrapper tries to fix
        # it: you need to pass special `FilterAll` filter class to really apply operation
        # to all objects.
        # Since it is not possible to call operation with empty (i.e. meaning "none") filter,
        # client must not send such operations to API. Operation which uses filter must
        # provide fake request to use instead of real one.

        # Split all operations to ones that should be skipped and ones that shouldn't.
        # But remember their original position in operation list.
        skipped, left = partition(enumerate(operations), lambda (i, op): op.should_be_skipped())

        operation_responses = []
        if left:
            # Group operations by operator.
            ops = [op for (i, op) in left]
            grouped = [(n, list(g)) for n, g in groupby(ops, lambda op: op.operator_name)]
            request = self._create_packet(grouped, request_settings)

            with profiler.measure_api_call(
                    self.url, 'Plesk API', "\n".join(
                        ["%s/%s" % (op.operator_name, op.operation_name) for op in operations])
                ), \
                    self.logger.censor({'xml': ['.//passwd', './/password']}):
                response = self.send_xml(request)

            root = response.getroot()

            # Handle system-level error
            status = root.find('system/status')
            if status is not None and status.text == 'error':
                raise PleskError(
                    code=if_not_none(root.findtext('system/errcode'), int),
                    message=root.findtext('system/errtext')
                )

            for (operator_name, operations), operator_response in zip(grouped, root):
                assert operator_name == operator_response.tag

                # In some cases and Plesk versions, system level errors like expired license
                # are under operators in response XML. If any operator have such response - raise exception.
                if operator_response.findtext('status') == 'error':
                    raise PleskError(
                        code=if_not_none(operator_response.findtext('errcode'), int),
                        message=operator_response.findtext('errtext')
                    )

                assert len(operations) == len(operator_response)
                for operation, operation_response in zip(operations, operator_response):
                    assert operation.operation_name == operation_response.tag
                    operation_responses.append(operation.parse(operation_response))

        fake_responses = [op.fake_response() for i, op in skipped]

        # Now merge real and fake responses matching operations order.
        responses_with_index = [
            (i, resp) for ((i, op), resp) in zip(left, operation_responses) + zip(skipped, fake_responses)
        ]
        return [resp for i, resp in sorted(responses_with_index, key=lambda (i, r): i)]

    def _create_packet(self, grouped, request_settings=None):
        request_settings = default(request_settings, {})
        request = etree.ElementTree(etree.Element('packet', dict(version=self.api_version)))

        if request_settings:
            request.getroot().append(elem('request-settings', [
                elem('setting', [text_elem('name', name), text_elem('value', value)])
                for name, value in request_settings.iteritems()
            ]))

        for operator_name, operations in grouped:
            operator_node = etree.Element(operator_name)
            for op in operations:
                operation_node = etree.Element(op.operation_name)
                operation_node.extend(op.inner_xml())
                operator_node.append(operation_node)

            request.getroot().append(operator_node)

        return request

    def _get_headers(self):
        headers = super(Client, self)._get_headers()
        if self.secret_key:
            headers['KEY'] = self.secret_key
        else:
            headers['HTTP_AUTH_LOGIN'] = self.username
            headers['HTTP_AUTH_PASSWD'] = self.password

        if self.pretty_print:
            headers['HTTP_PRETTY_PRINT'] = 'true'

        return headers

    def _is_operation_supported(self, operation):
        """Check if API versions supported by the operator matches server API version"""
        api_version_tuple = version_tuple(self.api_version)

        min_api_version = version_tuple(operation.min_api_version) if operation.min_api_version is not None else None
        max_api_version = version_tuple(operation.max_api_version) if operation.max_api_version is not None else None
        lower_bound_is_ok = min_api_version is None or min_api_version <= api_version_tuple
        upper_bound_is_ok = max_api_version is None or max_api_version >= api_version_tuple

        return lower_bound_is_ok and upper_bound_is_ok
