# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple
from datetime import datetime
from itertools import chain

from migrator.core.utils import if_not_none
from migrator.core.utils.filetypes.xml import elem, seq, text_elem
from .. import core
from .subscription import list_ips, parse_plesk_ips


DomainInfo = namedtuple('DomainInfo', (
    'gen_info', 'hosting'
))

DomainGenInfo = namedtuple('DomainGenInfo', (
    'creation_date', 'name', 'ascii_name', 'status', 'real_size', 'dns_ip_address',
    'htype', 'guid',
))


class DomainHostingNone(namedtuple('DomainHostingNone', ())):
    def outer_xml(self):
        return elem('none')


class DomainHostingStandardForwarding(
    namedtuple('DomainHostingStandardForwarding', ('dest_url', 'ip_addresses'))
):
    def outer_xml(self):
        return elem('std_fwd', chain(
            [text_elem('dest_url', self.dest_url)],
            [text_elem('ip_address', ip_address) for ip_address in list_ips(self.ip_addresses)]
        ))


class DomainHostingFrameForwarding(
    namedtuple('DomainHostingFrameForwarding', ('dest_url', 'ip_addresses'))
):
    def outer_xml(self):
        return elem('frm_fwd', chain(
            [text_elem('dest_url', self.dest_url)],
            [text_elem('ip_address', ip_address) for ip_address in list_ips(self.ip_addresses)]
        ))


class DomainHostingVirtual(namedtuple('DomainHostingVirtual', ('properties', 'ip_addresses'))):
    def outer_xml(self):
        subelements = [
            elem('property', [
                text_elem('name', name),
                text_elem('value', value),
            ])
            for name, value in self.properties
        ]
        if self.ip_addresses is not None:
            subelements.extend([text_elem('ip_address', ip_address) for ip_address in list_ips(self.ip_addresses)])

        return elem('vrt_hst', subelements)


class DomainOperator(object):
    ERR_DOMAIN_DOES_NOT_EXIST = 1013

    class Dataset(object):
        GEN_INFO = 'gen_info'
        HOSTING = 'hosting'
        # Not implemented: 'limits', 'stat', 'prefs', 'user', 'disk_usage', 'performance'

        values = [GEN_INFO, HOSTING]

    FilterAll = core.FilterAll
    FilterByName = core.declare_filter('FilterByName', 'name')
    FilterByDomainName = core.declare_filter('FilterByName', 'domain-name')
    FilterByName8 = core.declare_filter('FilterByName8', 'domain_name')
    FilterById = core.declare_filter('FilterById', 'id')
    FilterByClientLogin = core.declare_filter('FilterByClientLogin', 'client_login')
    FilterByOwnerLogin = core.declare_filter('FilterByClientLogin', 'owner-login')
    FilterByOwnerId = core.declare_filter('FilterByOwnerId', 'owner-id')

    class Get(core.operation_with_filter('Get', ('dataset',))):
        operator_name = 'domain'
        operation_name = 'get'
        min_api_version = '1.4.1.2'
        max_api_version = None

        def data_xml(self):
            return [
                elem('dataset', [elem(value) for value in DomainOperator.Dataset.values if value in self.dataset])
            ]

        @classmethod
        def parse(cls, element):
            results = [core.Result.parse(r, cls._parse_data) for r in element.findall('result')]
            if len(results) == 1 and results[0].ok and results[0].data is None:
                results = []

            return results

        @classmethod
        def _parse_data(cls, element):
            domain_id = if_not_none(element.findtext('id'), int)
            if domain_id is None:
                return None

            data = element.find('data')
            gen_info = if_not_none(data.find('gen_info'), cls._parse_gen_info)
            hosting = if_not_none(data.find('hosting'), cls._parse_hosting)

            return domain_id, DomainInfo(gen_info, hosting)

        @classmethod
        def _parse_gen_info(cls, element):
            return DomainGenInfo(
                creation_date=datetime.strptime(element.findtext('cr_date'), '%Y-%m-%d'),
                name=element.findtext('name'), ascii_name=element.findtext('ascii-name'),
                status=int(element.findtext('status')), real_size=int(element.findtext('real_size')),
                dns_ip_address=element.findtext('dns_ip_address'),
                htype=element.findtext('htype'), guid=element.findtext('guid'),
            )

        @classmethod
        def _parse_hosting(cls, element):
            if element.find('none') is not None:
                return DomainHostingNone()
            elif element.find('std_fwd') is not None:
                return DomainHostingStandardForwarding(
                    dest_url=element.findtext('std_fwd/dest_url'),
                    ip_addresses=parse_plesk_ips([ip_node.text for ip_node in element.findall('std_fwd/ip_address')])
                )
            elif element.find('frm_fwd') is not None:
                return DomainHostingFrameForwarding(
                    dest_url=element.findtext('frm_fwd/dest_url'),
                    ip_addresses=parse_plesk_ips([ip_node.text for ip_node in element.findall('frm_fwd/ip_address')])
                )
            elif element.find('vrt_hst') is not None:
                properties = dict()
                for property_node in element.findall('vrt_hst/property'):
                    properties[property_node.findtext('name')] = property_node.findtext('value')

                ip_addresses = parse_plesk_ips([ip_node.text for ip_node in element.findall('vrt_hst/ip_address')])

                return DomainHostingVirtual(properties=properties, ip_addresses=ip_addresses)

    DomainSetInfo = namedtuple('DomainSetInfo', ('user',))
    DomainSetUserInfo = namedtuple('DomainSetUserInfo', ('password',))

    class Set(core.operation_with_filter('Set', ('values',))):
        operator_name = 'domain'
        operation_name = 'set'
        min_api_version = '1.4.1.2'
        max_api_version = None

        def data_xml(self):
            return seq(
                elem('values', seq(
                    if_not_none(self.values.user, lambda ui: elem('user', [
                        if_not_none(ui.password, lambda p: text_elem('password', ui.password))
                    ]))
                ))
            )

        @classmethod
        def parse(cls, element):
            return core.Result.parse(element.find('result'))
