# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple

from migrator.core.utils.filetypes.xml import elem, text_elem
from .. import core

MailUserInfo = namedtuple('MailUserInfo', ('id', 'name', 'quota',))
AutorespondersInfo = namedtuple('AutorespondersInfo', ('enabled', 'autoresponders'))
AutoresponderInfo = namedtuple('AutoresponderInfo', ('enabled', 'subject', 'message', 'forward_to', 'end_date'))
AntivirusInfo = namedtuple('AntivirusInfo', ('virusfilter'))


class MailOperator(object):
    class Dataset(object):
        GEN_INFO = 'gen_info'
        values = [GEN_INFO]

    FilterBySiteId = core.declare_filter('FilterBySiteId', 'site-id')
    FilterByDomainId = core.declare_filter('FilterByDomainId', 'domain_id')
    FilterByName = core.declare_filter('FilterByName', 'name')
    ComplexFilter = core.declare_complex_filter()

    class Update(object):
        @staticmethod
        def get_inner_xml(operator, name, site_id, aliases, quota=None):
            xml = []
            if not name or not site_id:
                return xml

            xml.append(
                elem(operator, [
                    elem('filter', [
                        text_elem('site-id', site_id),
                        elem(
                            'mailname',
                            [text_elem('name', name)] +
                                ([text_elem('alias', v) for v in aliases] if aliases else []) +
                                ([elem('mailbox', [text_elem('quota', quota)])] if quota is not None else [])
                        )
                    ])
                ])
            )

            return xml

        class Set(core.operation('Set', ('name', 'site_id', 'aliases', 'quota'))):
            operator_name = 'mail'
            operation_name = 'update'
            min_api_version = '1.3.5.0'
            max_api_version = None

            def inner_xml(self):
                return MailOperator.Update.get_inner_xml('set', self.name, self.site_id, self.aliases, self.quota)

            @classmethod
            def parse(cls, elem):
                return [core.Result.parse(r) for r in elem.findall('result')]

        class Remove(core.operation('Remove', ('name', 'site_id', 'aliases'))):
            operator_name = 'mail'
            operation_name = 'update'
            min_api_version = '1.3.5.0'
            max_api_version = None

            def inner_xml(self):
                return MailOperator.Update.get_inner_xml('remove', self.name, self.site_id, self.aliases)

            @classmethod
            def parse(cls, elem):
                return [core.Result.parse(r) for r in elem.findall('result')]

    class SetPrefs(core.operation_with_filter('Get', ('nonexistent_user',))):
        operator_name = 'mail'
        operation_name = 'set_prefs'
        min_api_version = '1.3.5.0'
        max_api_version = None

        class NonexistentUserForward(namedtuple('NonexistentUserForward', ('forward',))):
            def inner_xml(self):
                return text_elem('forward', self.forward)

        class NonexistentUserReject(object):
            def inner_xml(self):
                return elem('reject')

        def data_xml(self):
            xml = []
            if self.nonexistent_user is not None:
                xml.append(
                    elem(
                        'prefs',
                        [
                            elem('nonexistent-user', [self.nonexistent_user.inner_xml()])
                        ]
                    )
                )
            return xml
                    
        @classmethod
        def parse(cls, elem):
            return [core.Result.parse(r) for r in elem.findall('result')]

    class Get(core.operation_with_filter('Get', ('mailbox',))):
        operator_name = 'mail'
        operation_name = 'get_info'
        min_api_version = '1.3.5.0'
        max_api_version = None

        def data_xml(self):
            return [
                elem('mailbox', [])
            ]

        @classmethod
        def parse(cls, elem):
            return core.filter_owner_not_found_error(core.parse_result_set(elem,  cls._parse_data, 'mailname'))

        @classmethod
        def _parse_data(cls, elem):
            return MailUserInfo(
                id=elem.findtext('mailname/id'),
                name=elem.findtext('mailname/name'),
                quota=elem.findtext('mailname/mailbox/quota'),
            )

    class GetMailboxAliases(core.operation_with_filter('Get')):
        operator_name = 'mail'
        operation_name = 'get_info'
        min_api_version = '1.3.5.0'
        max_api_version = None

        def data_xml(self):
            return [
                elem('aliases', [])
            ]

        @classmethod
        def parse(cls, elem):
            return core.parse_result_set(elem,  cls._parse_data, 'mailname')

        @classmethod
        def _parse_data(cls, elem):
            return [prop_node.text for prop_node in elem.findall('mailname/alias')]

    class GetAutoresponders(core.operation_with_filter('Get', ())):
        operator_name = 'mail'
        operation_name = 'get_info'
        min_api_version = '1.3.5.0'
        max_api_version = '1.6.2.0'

        def data_xml(self):
            return [
                elem('mailbox', []),
                elem('autoresponders', [])
            ]

        @classmethod
        def parse(cls, elem):
            return core.parse_result_set(elem,  cls._parse_data, 'mailname/autoresponders')

        @classmethod
        def _parse_data(cls, elem):
            return (
                MailUserInfo(
                    id=elem.findtext('mailname/id'),
                    name=elem.findtext('mailname/name'),
                    quota=elem.findtext('mailname/mailbox/quota'),
                ),
                AutorespondersInfo(
                    enabled=elem.findtext('mailname/autoresponders/enabled'),
                    autoresponders=cls._get_autoresponders(elem)
                )
            )

        @classmethod
        def _get_autoresponders(cls, elem):
            autoresponders = []
            autoresponder_elems = elem.findall('mailname/autoresponders/autoresponder')
            for autoresponder_elem in autoresponder_elems:
                # in plesk 8 autoresponder elements may be nested into each other instead of being on one level
                while autoresponder_elem is not None:
                    autoresponders.append(AutoresponderInfo(
                        enabled=autoresponder_elem.findtext('enabled'),
                        subject=autoresponder_elem.findtext('subject'),
                        message=autoresponder_elem.findtext('text'),
                        forward_to=autoresponder_elem.findtext('forward'),
                        end_date=None,
                    ))
                    autoresponder_elem = autoresponder_elem.find('autoresponder')

            return autoresponders

    class GetAutoresponder(core.operation_with_filter('Get', ())):
        operator_name = 'mail'
        operation_name = 'get_info'
        min_api_version = '1.6.3.0'
        max_api_version = None

        def data_xml(self):
            return [
                elem('mailbox', []),
                elem('autoresponder', [])
            ]

        @classmethod
        def parse(cls, elem):
            return core.filter_owner_not_found_error(core.parse_result_set(elem,  cls._parse_data, 'mailname/autoresponder'))

        @classmethod
        def _parse_data(cls, elem):
            return (
                MailUserInfo(
                    id=elem.findtext('mailname/id'),
                    name=elem.findtext('mailname/name'),
                    quota=elem.findtext('mailname/mailbox/quota'),
                ),
                AutoresponderInfo(
                    enabled=elem.findtext('mailname/autoresponder/enabled'),
                    subject=elem.findtext('mailname/autoresponder/subject'),
                    message=elem.findtext('mailname/autoresponder/text'),
                    forward_to=elem.findtext('mailname/autoresponder/forward'),
                    end_date=elem.findtext('mailname/autoresponder/end_date'),
                )
            )

    class GetAntivirus(core.operation_with_filter('Get', ())):
        operator_name = 'mail'
        operation_name = 'get_info'
        min_api_version = '1.3.5.0'
        max_api_version = None

        def data_xml(self):
            return [
                elem('mailbox', []),
            ]

        @classmethod
        def parse(cls, elem):
            return core.filter_owner_not_found_error(core.parse_result_set(elem,  cls._parse_data, 'mailname/antivir'))

        @classmethod
        def _parse_data(cls, elem):
            return (
                MailUserInfo(
                    id=elem.findtext('mailname/id'),
                    name=elem.findtext('mailname/name'),
                    quota=elem.findtext('mailname/mailbox/quota'),
                ),
                AntivirusInfo(
                    virusfilter=elem.findtext('mailname/antivir'),
                )
            )
