# (C)2017 Ingram Micro Inc. all Rights Reserved
from collections import namedtuple
from datetime import datetime
from itertools import chain

from migrator.core.utils import if_not_none
from migrator.core.utils.filetypes.xml import elem, text_elem, seq
from .. import core
from .subscription import list_ips, parse_plesk_ips


SiteInfo = namedtuple('SiteInfo', (
    'gen_info', 'hosting'
))

SiteGenInfo = namedtuple('SiteGenInfo', (
    'creation_date', 'name', 'ascii_name', 'status', 'real_size', 'dns_ip_address',
    'htype', 'guid', 'webspace_guid', 'webspace_id',
))


class SiteHostingNone(namedtuple('SiteHostingNone', ())):
    def outer_xml(self):
        return elem('none')


class SiteHostingStandardForwarding(
    namedtuple('SiteHostingStandardForwarding', ('dest_url', 'ip_addresses', 'http_code'))
):
    def outer_xml(self):
        return elem('std_fwd', chain(
            [text_elem('dest_url', self.dest_url)],
            [text_elem('ip_address', ip_address) for ip_address in list_ips(self.ip_addresses)],
            [text_elem('http_code', self.http_code)],
        ))


class SiteHostingFrameForwarding(
    namedtuple('SiteHostingFrameForwarding', ('dest_url', 'ip_addresses'))
):
    def outer_xml(self):
        return elem('frm_fwd', chain(
            [text_elem('dest_url', self.dest_url)],
            [text_elem('ip_address', ip_address) for ip_address in list_ips(self.ip_addresses)]
        ))


class SiteHostingVirtual(namedtuple('SiteHostingVirtual', ('properties', 'ip_addresses'))):
    def outer_xml(self):
        subelements = [
            elem('property', [
                text_elem('name', name),
                text_elem('value', value),
            ])
            for name, value in self.properties
        ]
        if self.ip_addresses is not None:
            subelements.extend([text_elem('ip_address', ip_address) for ip_address in list_ips(self.ip_addresses)])

        return elem('vrt_hst', subelements)


class SiteOperator(object):
    class Dataset(object):
        GEN_INFO = 'gen_info'
        HOSTING = 'hosting'
        # Not implemented: 'stat', 'prefs', 'disk_usage'

        values = [GEN_INFO, HOSTING]

    FilterAll = core.FilterAll
    FilterByName = core.declare_filter('FilterByName', 'name')
    FilterById = core.declare_filter('FilterById', 'id')
    FilterByParentId = core.declare_filter('FilterByParentId', 'parent-id')
    FilterByParentName = core.declare_filter('FilterByParentName', 'parent-name')
    FilterByGuid = core.declare_filter('FilterByGuid', 'guid')
    FilterByParentGuid = core.declare_filter('FilterByParentGuid', 'parent-guid')

    class Get(core.operation_with_filter('Get', ('dataset',))):
        operator_name = 'site'
        operation_name = 'get'
        min_api_version = '1.6.3.0'
        max_api_version = None

        def data_xml(self):
            return [
                elem('dataset', [elem(value) for value in SiteOperator.Dataset.values if value in self.dataset])
            ]

        @classmethod
        def parse(cls, element):
            results = core.filter_owner_not_found_error(core.parse_result_set(element, cls._parse_data, 'id'))
            if len(results) == 1 and results[0].ok and results[0].data is None:
                results = []

            return results

        @classmethod
        def _parse_data(cls, element):
            site_id = int(element.findtext('id'))

            data = element.find('data')
            gen_info = if_not_none(data.find('gen_info'), cls._parse_gen_info)
            hosting = if_not_none(data.find('hosting'), cls._parse_hosting)

            return site_id, SiteInfo(gen_info, hosting)

        @classmethod
        def _parse_gen_info(cls, element):
            return SiteGenInfo(
                creation_date=datetime.strptime(element.findtext('cr_date'), '%Y-%m-%d'),
                name=element.findtext('name'), ascii_name=element.findtext('ascii-name'),
                status=int(element.findtext('status')), real_size=int(element.findtext('real_size')),
                dns_ip_address=element.findtext('dns_ip_address'),
                htype=element.findtext('htype'), guid=element.findtext('guid'),
                webspace_guid=element.findtext('webspace-guid'),
                webspace_id=element.findtext('webspace-id'),
            )

        @classmethod
        def _parse_hosting(cls, element):
            if element.find('none') is not None:
                return SiteHostingNone()
            elif element.find('std_fwd') is not None:
                return SiteHostingStandardForwarding(
                    dest_url=element.findtext('std_fwd/dest_url'),
                    ip_addresses=parse_plesk_ips([ip_node.text for ip_node in element.findall('std_fwd/ip_address')]),
                    http_code=element.findtext('std_fwd/http_code'),
                )
            elif element.find('frm_fwd') is not None:
                return SiteHostingFrameForwarding(
                    dest_url=element.findtext('frm_fwd/dest_url'),
                    ip_addresses=parse_plesk_ips([ip_node.text for ip_node in element.findall('frm_fwd/ip_address')])
                )
            elif element.find('vrt_hst') is not None:
                properties = dict()
                for property_node in element.findall('vrt_hst/property'):
                    properties[property_node.findtext('name')] = property_node.findtext('value')

                ip_addresses = parse_plesk_ips([ip_node.text for ip_node in element.findall('vrt_hst/ip_address')])

                return SiteHostingVirtual(properties=properties, ip_addresses=ip_addresses)

    class Set(core.operation_with_filter('Set', ('hosting', ))):
        operator_name = 'site'
        operation_name = 'set'
        min_api_version = '1.6.3.0'
        max_api_version = None

        def data_xml(self):
            return [
                elem(
                    'values',
                    seq(
                        if_not_none(self.hosting, lambda h: elem('hosting', [h.outer_xml()]))
                    )
                ),
            ]

        @classmethod
        def parse(cls, element):
            return [core.Result.parse(r) for r in element.findall('result')]
