# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import InternalError, CommandExecutionError, MigrationNoRepeatError
from migrator.core.language import _T


class PleskAPISecretKeysManager(object):
    def __init__(self, runner, is_windows, plesk_dir):
        self._keys = {}
        self._ips = {}

        self._is_windows = is_windows
        self._runner = runner
        self._plesk_dir = plesk_dir

        try:
            self._refresh_keys()
        except CommandExecutionError as e:
            if 'The license key is invalid.' in e.stderr:
                raise MigrationNoRepeatError(
                    _T("FAILED_TO_CHECK_PLESK_API_EXPIRED_LICENSE").format(server=e.host_description)
                )
            raise

    @property
    def keys(self):
        return self._keys

    @property
    def ips(self):
        return self._ips

    def add_key(self, ip, description):
        cmd = '{cmd} -c -ip-address {ip} -description "{description}"'.format(
            cmd=self._get_skey_command(), ip=ip, description=description
        )
        self._runner.sh(cmd)
        self._refresh_keys()

    def del_key(self, key):
        cmd = '{cmd} -d -key {key}'.format(cmd=self._get_skey_command(), key=key)
        self._runner.sh(cmd)
        self._refresh_keys()

    def _refresh_keys(self):
        cmd = '%s -l' % self._get_skey_command()
        out = self._runner.sh(cmd, log_output=False)
        self._keys, self._ips = self._parse_skeys_list(out)

    @staticmethod
    def _parse_skeys_list(data):
        keys = {}
        ips = {}

        for key_data in data.split('\n\n'):
            if not key_data.strip():
                continue

            single_key = {}

            for line in key_data.splitlines():
                result = line.strip().split(':', 1)
                if len(result) > 1:
                    single_key[result[0].strip().lower()] = result[1].strip()

            if len(single_key.keys()) != 3:
                raise InternalError(
                    "Unable to parse secret key details from Plesk 'secret_key' command output. "
                    "An internal error occurred. Contact Odin technical support for assistance."
                )

            key = single_key['key']
            ip = single_key['ip']
            if ip not in ips:
                ips[ip] = [key]
            else:
                ips[ip].append(key)

            keys[key] = {'ip': ip, 'description': single_key['description']}

        return keys, ips

    def _get_skey_command(self):
        if self._is_windows:
            cmd = ur'"%s\bin\secret_key"' % self._plesk_dir
        else:
            cmd = u'%s/bin/secret_key' % self._plesk_dir

        return cmd
