# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core import _T
from migrator.core.data.model.subscription import get_service, has_any_of_service_types, get_service_property
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.hosting.dns.dns_forwarding import \
    LinuxBindForwarding, WindowsBindForwarding, MsDnsForwarding, NoneDnsForwarding
from migrator.core.product.lib.dns_zone_manager import DnsZoneManager, DnsZone
from migrator.core.session.config import is_windows_host
from migrator.core.utils.filetypes.config_utils import ConfigHostSection
from migrator.core.utils.networking.runner import get_runner
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops, api_versions
from migrator.products.plesk.lib.export.platform import PleskExporter
from migrator.products.plesk.lib.panel import split_version_string
from migrator.products.plesk.target.utils import plesk_utils
from migrator.products.plesk.target.utils.plesk_utils import get_windows_plesk_dir
from migrator.settings import PANEL_TYPE_PLESK, SOURCE_SERVICE_DNS_NS,\
    SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP, SourceHostingClass, SOURCE_SERVICE_DNS


class PleskDnsDnsZone(DnsZone):
    def __init__(self, id, name, is_alias):
        super(PleskDnsDnsZone, self).__init__(id, name)
        self.is_alias = is_alias


class PleskDnsZoneManager(DnsZoneManager):
    def __init__(self):
        super(PleskDnsZoneManager, self).__init__()
        self.global_context = Registry.get_instance().get_context()
        self.plesk_dns_server = None

    def disable_zone_on_source(self, subscription, zone):
        plesk_server = self.global_context.conn.get_panel_server(self.global_context, 'plesk')
        plesk_api = plesk_server.plesk_api()
        self._disable_plesk_zone(zone.id, zone.is_alias, plesk_api)

    def get_subscription_zones(self, subscription):
        if self._is_plesk_dns_enabled(subscription) and \
           has_any_of_service_types(subscription.source_info, SOURCE_HOSTING_CLASSES_TO_SERVICES_MAP[SourceHostingClass.DNS]):

            external_id = subscription.source_info.external_id
            if not external_id.startswith('dns_alias_'):
                # alias is not a site and it does not have site id, moreover alias id enabled/disabled together with his domain

                return [
                    PleskDnsDnsZone(
                        id=PleskExporter.extract_domain_id(external_id),
                        name=get_service_property(subscription.source_info, SOURCE_SERVICE_DNS, 'domain'),
                        is_alias=False
                    )
                ]
        return []

    def get_dns_forwarding_impl(self, service):
        source_host = PlatformReaderDB().get_source_host(service.source_host_external_id)
        config_section = ConfigHostSection(self.global_context.config, source_host.ip)
        is_windows = is_windows_host(config_section)
        if source_host.panel == PANEL_TYPE_PLESK and not is_windows:
            runner = get_runner(config_section)
            named_chroot = plesk_utils.get_unix_conf_var(runner, 'NAMED_RUN_ROOT_D')
            named_conf = named_chroot.rstrip('/') + '/etc/named.conf'
            return LinuxBindForwarding(self.global_context.migrator_server, runner, source_host.external_id, named_conf)
        elif source_host.panel == PANEL_TYPE_PLESK and is_windows:
            runner = get_runner(config_section)
            dns_server_type = self._get_plesk_windows_dns_type(runner)
            plesk_dir = get_windows_plesk_dir(runner)
            if dns_server_type == 'bind':
                named_conf = plesk_dir.rstrip('\\') + '\\dns\\etc\\named.conf'
                return WindowsBindForwarding(self.global_context.migrator_server, runner, source_host.external_id, named_conf)
            elif dns_server_type == 'msdns':
                plesk_version = self._get_plesk_windows_version(runner)
                if split_version_string(plesk_version) < split_version_string("11.5"):
                    dnscmd_path = plesk_dir.rstrip('\\') + "\\admin\\bin\\dnscmd"
                else:
                    dnscmd_path = "C:\\Windows\\System32\\dnscmd.exe"

                return MsDnsForwarding(self.global_context.migrator_server, runner, source_host.external_id, dnscmd_path)
            elif dns_server_type == 'none':
                return NoneDnsForwarding()

        raise Exception(
            _T("INTERNAL_ERROR_UNKNOWN_DNS_SERVICE_TYPE").format(panel=source_host.panel, ip=source_host.ip)
        )

    @staticmethod
    def _get_plesk_windows_version(runner):
        version_str = runner.sh(
            ur'cmd /C type {versionfile}', {'versionfile': ur"%plesk_dir%version"})
        return version_str.split()[0]

    def _get_plesk_windows_dns_type(self, runner):
        if self.plesk_dns_server is None:
            script = r'''
@echo off

if "%processor_architecture%"=="AMD64" set infix=Wow6432Node\
if "%processor_architecture%"=="IA64" set infix=Wow6432Node\

for /f "tokens=3" %%s in ('reg query "HKEY_LOCAL_MACHINE\SOFTWARE\%infix%Plesk\PSA Config\Config\Packages\dnsserver" /ve') do echo %%s'''
            script_path = 'c:\\plesk_dns_type.bat'

            runner.upload_file_content(script_path, script)

            stdout = runner.sh(
                ur'cmd.exe /C "{script_filename_source}"', dict(script_filename_source=script_path)
            )
            self.plesk_dns_server = stdout.strip().split("\n")[0]

        return self.plesk_dns_server

    def _disable_plesk_zone(self, zone_id, is_alias, plesk_api):
        # Aliases exported as "alias_ID". Need to cut "alias_" prefix for sending correct request to Plesk
        if is_alias:
            zone_id = zone_id.replace("alias_", "")

        operator = self._get_plesk_dns_disable_operator(plesk_api.api_version)(
            filter=self._get_plesk_dns_disable_operator_filter(plesk_api.api_version, is_alias)(
                [zone_id]
            )
        )
        plesk_api.send(operator)

    @staticmethod
    def _get_plesk_dns_disable_operator_filter(api_version, is_alias):
        if split_version_string(api_version) < split_version_string(api_versions['10.0']):
            if is_alias:
                return plesk_ops.DnsOperator9.FilterByDomainAliasId
            else:
                return plesk_ops.DnsOperator9.FilterByDomainId
        else:
            if is_alias:
                return plesk_ops.DnsOperator.FilterBySiteAliasId
            else:
                return plesk_ops.DnsOperator.FilterBySiteId

    @staticmethod
    def _get_plesk_dns_disable_operator(api_version):
        if split_version_string(api_version) < split_version_string(api_versions['10.0']):
            return plesk_ops.DnsOperator9.Disable
        else:
            return plesk_ops.DnsOperator.Disable

    def _is_plesk_dns_enabled(self, subscription):
        ns_service = get_service(subscription.source_info, SOURCE_SERVICE_DNS_NS)
        source_host = PlatformReaderDB().get_source_host(ns_service.source_host_external_id)
        config_section = ConfigHostSection(self.global_context.config, source_host.ip)
        if is_windows_host(config_section):
            runner = get_runner(config_section)
            return self._get_plesk_windows_dns_type(runner) != 'none'
        else:
            return True
