# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import _T
from migrator.core.data.format.billing import BillingReader
from migrator.core.data.model.billing import \
    AccountBillingModel, SubscriptionBillingModel, SourceSubscriptionStatus
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import PleskError
from migrator.products.plesk.lib.export.platform import PLESK_ADMIN_ID

logger = logging.getLogger(__name__)


class PleskBillingExporter(BillingReader):
    """This is the exporter of mostly fake billing info from Plesk to make it possible to migrate to OA with Billing"""

    def __init__(self):
        super(PleskBillingExporter, self).__init__()
        context = Registry.get_instance().get_context()
        self.plesk_api = context.conn.get_panel_server(context, 'plesk').plesk_api()

    def get_next_account_page(self, subscription_id_list):
        """Divide account ID list to pages; each call, return a next page of account billing objects"""
        account_id_list = self._get_account_id_list(subscription_id_list)
        for id_list in self._paginate(account_id_list):
            yield [self.get_account(account_id) for account_id in id_list]

    def get_next_subscription_page(self, subscription_id_list):
        for id_list in self._paginate(list(subscription_id_list)):
            yield [self.get_subscription(s) for s in id_list]

    def _get_account_id_list(self, subscription_id_list):
        """Return account billing objects that own the subscriptions from subscription_id_list"""
        source_system_reader = PlatformReaderDB()
        account_objects = [source_system_reader.get_subscription(item_id).account for item_id in subscription_id_list]
        return list({account.external_id for account in account_objects})

    def get_account(self, account_id):
        """

        :param account_id: str
        :return: migrator.core.data.model.billing.AccountModel
        """

        if account_id == PLESK_ADMIN_ID:
            return self._get_plesk_admin_account()

        return AccountBillingModel(
            external_id=account_id,
            balance=0,
        )

    def get_reseller(self, reseller_id):
        """

        :param reseller_id: str
        :return: migrator.core.data.model.billing.ResellerModel
        """
        return AccountBillingModel(
            external_id=reseller_id,
            balance=0,
        )

    def get_subscription(self, subscription_id):
        source_system_reader = PlatformReaderDB()
        source_subscription = source_system_reader.get_subscription(subscription_id)
        if source_subscription.parent_external_id is None:
            status = self._get_subscription_status(subscription_id)
        else:
            status = self._get_subscription_status(source_subscription.parent_external_id)

        return SubscriptionBillingModel(subscription_id, status)

    def _get_plesk_admin_account(self):
        admin_info = self.plesk_api.send(
            plesk_ops.ServerOperator.Get(
                dataset=[plesk_ops.ServerOperator.Dataset.ADMIN]
            )
        ).data.admin
        login = admin_info.email

        return AccountBillingModel(
            external_id=login,
            balance=0,
        )

    def _get_subscription_status(self, subscription_id):
        source_subscription = self._get_plesk_subscription_api(subscription_id)
        try:
            subscription_status = int(source_subscription[0].data[1].gen_info.status)
        except PleskError as e:
            if e.code == 1013:
                return SourceSubscriptionStatus.STATUS_DELETED

            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            raise Exception(_T("FAILED_TO_RETRIEVE_STATUS_OF_PLESK_SUBSCRIPTION").format(
                id=subscription_id, reason=unicode(e)
            ))

        if subscription_status == 0:
            return SourceSubscriptionStatus.STATUS_ACTIVE
        elif subscription_status == 256:
            return SourceSubscriptionStatus.STATUS_EXPIRED
        else:
            return SourceSubscriptionStatus.STATUS_ON_HOLD

    def _get_plesk_subscription_api(self, subscription_id):
        dataset = [
            plesk_ops.SubscriptionOperator.Dataset.GEN_INFO,
        ]
        api_filter = plesk_ops.SubscriptionOperator.FilterById([subscription_id])
        plesk_operator = plesk_ops.SubscriptionOperator.Get(dataset=dataset, filter=api_filter)
        return self.plesk_api.send(plesk_operator)


class Plesk9BillingExporter(PleskBillingExporter):
    def _get_plesk_subscription_api(self, subscription_id):
        dataset = [plesk_ops.DomainOperator.Dataset.GEN_INFO]
        api_filter = plesk_ops.DomainOperator.FilterById([subscription_id])
        request = plesk_ops.DomainOperator.Get(dataset=dataset, filter=api_filter)
        return self.plesk_api.send(request)
