# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from migrator.core import MigrationError, _T, InternalError
from migrator.core.data.format.dns import DnsReader
from migrator.core.data.model.dns import \
    DNS_RESOURCE_RECORD_A, DNS_RESOURCE_RECORD_AAAA, DNS_RESOURCE_RECORD_CNAME, DNS_RESOURCE_RECORD_MX, \
    DNS_RESOURCE_RECORD_SRV, DNS_RESOURCE_RECORD_TXT, DnsZoneModel, DnsSoaRecordModel, DnsRecordModel
from migrator.core.data.model.subscription import get_service_property
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import Failure
from migrator.products.plesk.lib.api.xmlrpc.operator import ZONE_STATUS_ENABLED, ZONE_TYPE_MASTER
from migrator.products.plesk.lib.export.platform import PleskExporter
from migrator.settings import SOURCE_SERVICE_DNS, PLESK_TAKEOVER_DNS_RECORD_IS_EDIT_ALLOWED

logger = logging.getLogger(__name__)


class PleskDnsExporter(PleskExporter, DnsReader):
    SUPPORTED_DNS_RECORDS_TYPES = [
        DNS_RESOURCE_RECORD_A, DNS_RESOURCE_RECORD_AAAA, DNS_RESOURCE_RECORD_CNAME, DNS_RESOURCE_RECORD_MX,
        DNS_RESOURCE_RECORD_SRV, DNS_RESOURCE_RECORD_TXT
    ]

    def get_next_zone_page(self, dns_subscription_id_list):
        """
        :param dns_subscription_id_list: set(str)
        :return: [migrator.core.data.model.dns.DnsZoneModel]
        """
        for subscription_id_page in self._paginate(list(dns_subscription_id_list)):
            yield [self._create_dns_zone_object(dns_subscription_id) for dns_subscription_id in subscription_id_page]

    def _create_dns_zone_object(self, dns_subscription_id):
        """
        :param dns_subscription_id: str Examples: 'dns_5', 'dns_alias_6'
        :return: migrator.core.data.model.dns.DnsZoneModel
        """
        is_domain_alias = self.is_domain_alias(dns_subscription_id)
        if is_domain_alias:
            dns_zone_id = self.extract_domain_alias_id(dns_subscription_id)
            zone_filter = plesk_ops.DnsOperator.FilterBySiteAliasId([dns_zone_id])
        else:
            dns_zone_id = self.extract_domain_id(dns_subscription_id)
            zone_filter = plesk_ops.DnsOperator.FilterBySiteId([dns_zone_id])

        domain_general_info = self.get_domain_general_info(dns_zone_id, is_domain_alias, zone_filter)
        subscription = PlatformReaderDB().get_subscription(subscription_external_id=dns_subscription_id)
        return DnsZoneModel(
            external_id=dns_subscription_id,
            name=get_service_property(subscription, SOURCE_SERVICE_DNS, 'domain'),
            records=self.create_dns_records(dns_zone_id, is_domain_alias, zone_filter),
            soa=self.create_soa_record_object(domain_general_info),
            enabled=self.is_dns_zone_enabled(domain_general_info),
            dns_managed=self.is_dns_zone_managed(domain_general_info)
        )

    def is_dns_zone_enabled(self, domain_general_info):
        """
        :param domain_general_info: migrator.products.plesk.lib.api.xmlrpc.core.Success Example:
            Success(data=DnsZone(
                object_id=15,
                object_type='site',
                soa=Soa(ttl=86400, refresh=10800, retry=3600, expire=604800, minimum=10800),
                zone_status='enabled',
                zone_type='master'
            ))
        :return: bool
        """
        return domain_general_info.data.zone_status == ZONE_STATUS_ENABLED

    def is_dns_zone_managed(self, domain_general_info):
        """
        :param domain_general_info: migrator.products.plesk.lib.api.xmlrpc.core.Success Example:
            Success(data=DnsZone(
                object_id=15,
                object_type='site',
                soa=Soa(ttl=86400, refresh=10800, retry=3600, expire=604800, minimum=10800),
                zone_status='enabled',
                zone_type='master'
            ))
        :return: bool
        """
        return domain_general_info.data.zone_type == ZONE_TYPE_MASTER

    def get_domain_general_info(self, dns_zone_id, is_domain_alias, zone_filter):
        """
        Get a general info with SOA record for the DNS zone

        :param dns_zone_id: str Examples: '2', '5'
        :param is_domain_alias: bool
        :param zone_filter: instance Examples: FilterBySiteId([u'18']), FilterBySiteAliasId([u'2'])
        :return: migrator.products.plesk.lib.api.xmlrpc.core.Success Example:
            Success(data=DnsZone(
                object_id=15,
                object_type='site',
                soa=Soa(ttl=86400, refresh=10800, retry=3600, expire=604800, minimum=10800),
                zone_status='enabled',
                zone_type='master'
            ))
        """
        plesk_operator = plesk_ops.DnsOperator.Get(filter=zone_filter, soa=True)
        domain_general_info = self.plesk_api.send(plesk_operator)[0]
        if isinstance(domain_general_info, Failure):
            self._raise_exception_if_dns_zone_not_found(is_domain_alias, dns_zone_id)

        return domain_general_info

    def create_soa_record_object(self, domain_general_info):
        """
        :param domain_general_info: migrator.products.plesk.lib.api.xmlrpc.core.Success Example:
            Success(data=DnsZone(
                object_id=15,
                object_type='site',
                soa=Soa(ttl=86400, refresh=10800, retry=3600, expire=604800, minimum=10800),
                zone_status='enabled',
                zone_type='master'
            ))
        :return: migrator.core.data.model.dns.DnsSoaRecordModel Example:
            DnsSoaRecordModel(retry=3600, refresh=10800, minimum=10800, expire=604800, ttl=86400, serial=None, email=None)
        """
        return DnsSoaRecordModel(
            ttl=domain_general_info.data.soa.ttl,
            refresh=domain_general_info.data.soa.refresh,
            retry=domain_general_info.data.soa.retry,
            expire=domain_general_info.data.soa.expire,
            minimum=domain_general_info.data.soa.minimum
        )

    def _raise_exception_if_dns_zone_not_found(self, is_domain_alias, dns_zone_id):
        """
        :param is_domain_alias: bool
        :param dns_zone_id: str Examples: '2', '5'
        :return: void
        """
        if is_domain_alias:
            error_message = _T('DNS_ZONE_FOR_DOMAIN_ALIAS_WITH_ID_NOT_FOUND').format(alias_id=dns_zone_id)
        else:
            error_message = _T('DNS_ZONE_FOR_DOMAIN_WITH_ID_WAS_NOT_FOUND').format(domain_id=dns_zone_id)

        raise MigrationError(error_message)

    def create_dns_records(self, dns_zone_id, is_domain_alias, zone_filter):
        """
        :param dns_zone_id: str '2', '5'
        :param is_domain_alias: bool
        :param zone_filter: instance Examples: FilterBySiteId([u'18']), FilterBySiteAliasId([u'2'])
        :return: [migrator.core.data.model.dns.DnsRecordModel]
        """
        plesk_operator = plesk_ops.DnsOperator.GetRec(filter=zone_filter, template=False)
        zone_records = self.plesk_api.send(plesk_operator)
        # Plesk API always returns DNS records in any zone's states: enabled (master), enabled (slave), disabled
        if not zone_records:
            self._raise_exception_if_dns_zone_not_found(is_domain_alias, dns_zone_id)

        dns_records_to_export = []
        for dns_record in zone_records:
            if dns_record.data.rec_type.lower() not in self.SUPPORTED_DNS_RECORDS_TYPES:
                continue

            dns_records_to_export.append(self.create_dns_record_object(dns_record))

        return dns_records_to_export

    def create_dns_record_object(self, dns_record):
        """
        :param dns_record: migrator.products.plesk.lib.api.xmlrpc.core.Success
        Examples:
            - Success(data=DnsRecord(id='1', src=u'_SIP._tcp.sip.ex.com.', dst='ex.com.', opt='0 10 566', rec_type='SRV'))
            - Success(data=DnsRecord(id='2', src=u'ex.com.', dst='v=spf1 +a +mx +a:ex.com -all', opt='', rec_type='TXT'))
            - Success(data=DnsRecord(id='3', src=u'ipv6.ex.com.', dst='2002:7b7b:7b7b::1', opt='', rec_type='AAAA'))
            - Success(data=DnsRecord(id='4', src=u'www.ex.com.', dst=u'ex.com.', opt='', rec_type='CNAME'))
            - Success(data=DnsRecord(id='5', src=u'ex.com.', dst=u'mail.ex.com.', opt='10', rec_type='MX'))
            - Success(data=DnsRecord(id='6', src=u'ns1.ex.com.', dst='10.37.47.80', opt='', rec_type='A'))
            - Success(data=DnsRecord(id='7', src=u'ex.com.', dst=u'ns1.ex.com.', opt='', rec_type='NS'))
        :return: migrator.core.data.model.dns.DnsRecordModel
        """
        record_type = dns_record.data.rec_type.lower()
        record_dst = dns_record.data.dst
        record_opt = dns_record.data.opt
        record_src = dns_record.data.src

        if record_type == DNS_RESOURCE_RECORD_MX:
            value = self.get_fqdn_without_last_dot(record_dst)
            priority = record_opt
        elif record_type == DNS_RESOURCE_RECORD_A:
            value = record_dst
            priority = None
        elif record_type == DNS_RESOURCE_RECORD_AAAA:
            value = record_dst
            priority = None
        elif record_type == DNS_RESOURCE_RECORD_CNAME:
            value = record_dst
            priority = None
        elif record_type == DNS_RESOURCE_RECORD_SRV:
            record_opt_parts = record_opt.split()
            priority = record_opt_parts[0]
            value = u'%s %s %s' % (record_opt_parts[1], record_opt_parts[2], self.get_fqdn_without_last_dot(record_dst))
        elif record_type == DNS_RESOURCE_RECORD_TXT:
            value = self.get_fqdn_without_last_dot(record_dst)
            priority = None
        else:
            raise InternalError(_T('UNEXPECTED_DNS_RR_TYPE_TO_EXPORT').format(rr_type=record_type))

        return DnsRecordModel(
            host=self.get_fqdn_without_last_dot(record_src), type=record_type, value=value, ttl=None, priority=priority,
            is_edit_allowed=PLESK_TAKEOVER_DNS_RECORD_IS_EDIT_ALLOWED
        )

    def get_zone(self, dns_subscription_id):
        raise NotImplementedError()
