# (C)2017 Ingram Micro Inc. all Rights Reserved
import posixpath

from migrator.core.data.format.mail_content_shm import ShmMailContentReader
from migrator.core.data.model.mail_content_shm import SubscriptionModel, WebspaceModel, SiteModel, MailboxModel
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.networking.dns import normalize_domain_name
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib import get_plesk_webspace_info, get_plesk_sites_info, get_plesk_domains_info
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import PleskError
from migrator.products.plesk.lib.export.utils import fix_plesk9_subdomains_ids


class PleskMailContentExporter(ShmMailContentReader):
    def __init__(self, server=None):
        if server is None:
            context = Registry.get_instance().get_context()
            server = context.conn.get_panel_server(context, 'plesk')

        self._is_windows = server.is_windows()
        self._plesk_api = server.plesk_api()
        self._server = server

    def get_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: SubscriptionModel
        """
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id, subscription_external_id=subscription_id)
        )

    def get_webspaces_list(self, subscription_id, subscription_external_id=None):
        """
        :param subscription_id: int
        :return: list[WebspaceModel]
        """

        return [self.get_webspace(subscription_id, subscription_id, subscription_external_id)]

    def get_webspace(self, subscription_id, webspace_id, subscription_external_id=None):
        """
        :param webspace_id: int
        :return: WebspaceModel
        """
        webspace_info = self._get_plesk_webspace_info(webspace_id)
        return WebspaceModel(
            external_id=webspace_id,
            name=webspace_info.gen_info.name,
            sites=self.get_sites_list(subscription_id, webspace_id, subscription_external_id),
        )

    def get_sites_list(self, subscription_id, webspace_id, subscription_external_id=None):
        """
        :param webspace_id: int
        :return: list[SiteModel]
        """
        sites = dict()

        # Make Plesk webspace to be a main site due to SHM requirements
        webspace_info = self._get_plesk_webspace_info(webspace_id)
        webspace_name = webspace_info.gen_info.name
        sites[webspace_info.gen_info.name] = SiteModel(
            name=webspace_name,
            external_id=webspace_id,
            mailboxes=self._get_site_mailboxes(webspace_id, webspace_name),
        )

        # Get all webspace sites
        sites_info = self._get_plesk_sites(webspace_id)
        for result in sites_info:
            site_id = result.data[0]
            site_info = result.data[1]
            site_name = site_info.gen_info.name

            sites[site_info.gen_info.name] = SiteModel(
                name=site_name,
                external_id=site_id,
                mailboxes=self._get_site_mailboxes(site_id, site_name),
            )

        # Get all subdomains as sites
        subdomains_info = self._get_plesk_subdomains(sites)
        for result in subdomains_info:
            subdomain_id = result.data[0]
            subdomain_info = result.data[1]
            subdomain_name = subdomain_info.name

            sites[subdomain_info.name] = SiteModel(
                name=subdomain_name,
                external_id=subdomain_id,
                mailboxes=self._get_subdomain_mailboxes(subdomain_id, subdomain_name),
            )

        is_mailbox_exists = False
        for site in sites.values():
            if site.mailboxes:
                is_mailbox_exists = True
                break

        if is_mailbox_exists and self._is_windows and self._is_installed_unsupported_mailenable():
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_external_id, Issue(Problem(
                    'old-mail-enable',
                    Problem.NOTICE,
                    _T("MAILBOX_CONTENT_FOR_ALL_MAILBOXES").format(webspace_name=webspace_name)
                ), None)
            )
            return []

        return sites.values()

    def _get_site_mailboxes(self, site_id, site_name):
        """
        :param site_id: int
        :return: list[MailboxModel]
        """

        mailboxes = []
        try:
            for mailbox in self._get_mailboxes_info(site_id):
                if not self._is_windows:
                    source_path = self._get_source_mailbox_directory(mailbox.data.name, site_name)
                    if source_path is None:
                        continue

                    mailboxes.append(
                        MailboxModel(
                            name=mailbox.data.name,
                            source_path=source_path + '/',
                            target_path='{site_name}/{mailbox_name}/'.format(
                                site_name=site_name.encode('idna'),
                                mailbox_name=mailbox.data.name.encode('idna')
                            ).lower(),
                        )
                    )
                else:
                    mailboxes.append(
                        MailboxModel(name=mailbox.data.name)
                    )
        except PleskError as e:
            if e.code == 1026:
                # It's ok if mail server component is not available in Plesk.
                pass
            else:
                raise e

        return mailboxes

    @cached
    def _is_installed_unsupported_mailenable(self):
        server_info = self._plesk_api.send(
            plesk_ops.ServerOperator.Get(
                dataset=[plesk_ops.ServerOperator.Dataset.COMPONENTS],
            )
        )
        # Search for unsupported email services: MailEnable 1.986.
        for component in server_info.data.components.values():
            if component.name.lower() == 'mailenable' and component.version == '1.986':
                return True

        return False

    def _get_subdomain_mailboxes(self, site_id, site_name):
        return self._get_site_mailboxes(site_id, site_name)

    def _get_source_mailbox_directory(self, mailbox_name, site_name):
        _site_name = site_name.encode('idna')
        for _site_name in [_site_name, _site_name.lower()]:
            for _mailbox_name in [mailbox_name, mailbox_name.lower()]:
                source_mailbox_directory = posixpath.join(self._server.mail_dir, _site_name, _mailbox_name)

                with self._server.runner() as runner:
                    if runner.sh(
                        'test -e {directory} && echo "yes" || echo "no"', dict(directory=source_mailbox_directory)
                    ).strip() == 'yes':
                        return source_mailbox_directory

        return None

    def _get_mailboxes_info(self, site_id):
        with self._plesk_api.logger.censor({'xml': ['/packet/mail/get_info/result/mailname/password/value']}):
            return self._plesk_api.send(
                plesk_ops.MailOperator.Get(
                    mailbox=[plesk_ops.MailOperator.Dataset.GEN_INFO],
                    filter=plesk_ops.MailOperator.FilterBySiteId([site_id])
                )
            )

    def _get_plesk_subdomains(self, sites):
        # Plesk 11 & 12 report subdomains via Site operator.
        return []

    @cached
    def _get_plesk_sites(self, webspace_id):
        return get_plesk_sites_info(
            self._plesk_api,
            webspace_id=webspace_id,
            dataset=[plesk_ops.SiteOperator.Dataset.GEN_INFO, plesk_ops.SiteOperator.Dataset.HOSTING]
        )

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        return get_plesk_webspace_info(self._plesk_api, webspace_id)


class Plesk10MailContentExporter(PleskMailContentExporter):
    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([x for x in sites.iterkeys()])
        )
        return self._plesk_api.send(operator)


class Plesk9MailContentExporter(Plesk10MailContentExporter):
    def _get_plesk_sites(self, webspace_id):
        # Plesk 8 & 9 don't support addon domains
        return []

    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([normalize_domain_name(x) for x in sites.iterkeys()]))
        subdomains_info = self._plesk_api.send(operator)

        return fix_plesk9_subdomains_ids(subdomains_info)

    def _get_subdomain_mailboxes(self, site_id, site_name):
        # Plesk 8 & 9 don't support mail on subdomains
        return []

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        # Plesk 8 & 9 don't have subscriptions and webspaces. We use domains for this role
        return get_plesk_domains_info(self._plesk_api, webspace_id)

    def _get_mailboxes_info(self, site_id):
        with self._plesk_api.logger.censor({'xml': ['/packet/mail/get_info/result/mailname/password/value']}):
            return self._plesk_api.send(
                plesk_ops.MailOperator.Get(
                    mailbox=[plesk_ops.MailOperator.Dataset.GEN_INFO],
                    filter=plesk_ops.MailOperator.FilterByDomainId([site_id])
                )
            )
