# (C)2017 Ingram Micro Inc. all Rights Reserved
from migrator.core.data.format.mail_hosting_shm import ShmMailHostingReader
from migrator.core.data.model.mail_hosting_shm import \
    SubscriptionModel, WebspaceModel, SiteModel, MailboxModel, AutoresponderModel, ForwardingModel, \
    AliasModel, SpamFilterModel
from migrator.core.language import _T
from migrator.core.utils import cached, generate_random_password, parse_bool
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.networking.dns import normalize_domain_name
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib import get_plesk_webspace_info, get_plesk_sites_info, get_plesk_domains_info
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import PleskError
from migrator.products.plesk.lib.decrypt_password import decrypt_account_passwords
from migrator.products.plesk.lib.export.utils import fix_plesk9_subdomains_ids
from migrator.products.plesk.target.utils.plesk_utils import query_plesk_db
from migrator.settings import ONE_MB

# Constant for maximum batch size
MAXIMUM_BATCH_SIZE = 100


class PleskMailExporter(ShmMailHostingReader):
    def __init__(self, server=None):
        if server is None:
            context = Registry.get_instance().get_context()
            server = context.conn.get_panel_server(context, 'plesk')

        self._is_windows = server.is_windows()
        self._plesk_api = server.plesk_api()
        self._server = server

    def get_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: SubscriptionModel
        """
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id, subscription_external_id=subscription_id)
        )

    def get_webspaces_list(self, subscription_id, subscription_external_id=None):
        """
        :param subscription_id: int
        :return: list[WebspaceModel]
        """
        return [self.get_webspace(subscription_id, subscription_id, subscription_external_id)]

    def get_webspace(self, subscription_id, webspace_id, subscription_external_id=None):
        """
        :param webspace_id: int
        :return: WebspaceModel
        """
        return WebspaceModel(
            external_id=webspace_id,
            sites=self.get_sites_list(subscription_id, webspace_id, subscription_external_id),
        )

    def get_sites_list(self, subscription_id, webspace_id, subscription_external_id=None):
        """
        :param webspace_id: int
        :return: list[SiteModel]
        """
        sites = dict()

        # Make Plesk webspace to be a main site due to SHM requirements
        webspace_info = self._get_plesk_webspace_info(webspace_id)
        webspace_name = webspace_info.gen_info.name
        sites[webspace_name] = SiteModel(
            name=webspace_name,
            external_id=webspace_id,
            mailboxes=self._get_site_mailboxes(webspace_id, webspace_info.gen_info.ascii_name),
        )

        # Get all webspace sites
        sites_info = self._get_plesk_sites(webspace_id)
        for result in sites_info:
            site_id = result.data[0]
            site_info = result.data[1]
            site_name = site_info.gen_info.name

            sites[site_name] = SiteModel(
                name=site_name,
                external_id=site_id,
                mailboxes=self._get_site_mailboxes(site_id, site_info.gen_info.ascii_name),
            )

        # Get all subdomains as sites
        subdomains_info = self._get_plesk_subdomains(sites)
        for result in subdomains_info:
            subdomain_id = result.data[0]
            subdomain_info = result.data[1]
            subdomain_name = subdomain_info.name

            sites[subdomain_name] = SiteModel(
                name=subdomain_name,
                external_id=subdomain_id,
                mailboxes=self._get_subdomain_mailboxes(subdomain_id, subdomain_name),
            )

        self._set_mailbox_passwords(sites, webspace_info.gen_info.name, subscription_external_id)
        return sites.values()

    def _get_site_mailboxes(self, site_id, site_name):
        """
        :param site_id: int
        :param site_name: str
        :return: list[MailboxModel]
        """

        mailboxes = []
        autoresponders = self._get_autoresponders(site_id)
        antiviruses = self._get_antiviruses(site_id)
        query = "SELECT id FROM mail WHERE dom_id = {site_id}".format(site_id=site_id)
        mailbox_ids = query_plesk_db(self._server, query)
        for batch_mailboxes in self._batch([x[0] for x in mailbox_ids]):
            mailboxes_info = self._get_mailboxes_info(batch_mailboxes, antiviruses)
            forwardings = self._get_forwardings(batch_mailboxes)
            aliases = self._get_aliases(batch_mailboxes)

            for id, is_access_to_customer_panel, mail_name, mbox_quota, antivirus, is_spamfilter in mailboxes_info:
                mailboxes.append(
                    MailboxModel(
                        external_id=id,
                        name=mail_name,
                        password='',
                        is_password_generated=True,
                        is_access_to_customer_panel=is_access_to_customer_panel,
                        disk_quota=self._get_disk_quota(mbox_quota),
                        autoresponder=self._get_autoresponder(id, autoresponders),
                        forwarding=self._get_forwarding(id, forwardings),
                        alias=self._get_alias(id, aliases),
                        antivirus=self._get_antivirus(antivirus),
                        spam_filter=self._get_spam_filter(is_spamfilter, mail_name, site_name),
                    )
                )

        return mailboxes

    def _get_subdomain_mailboxes(self, site_id, site_name):
        return self._get_site_mailboxes(site_id, site_name)

    def _get_mailboxes_info(self, batch_mailboxes, antiviruses=None):
        def _is_access_to_customer_panel(user_id, is_locked):
            return str(user_id) != '0' and str(is_locked) == '0'

        mailboxes_info = []
        if self._is_windows:
            query = "SELECT m.id, m.userId, s.isLocked, m.mail_name, m.mbox_quota " \
                    "FROM mail as m " \
                    "LEFT JOIN smb_users as s ON m.userId = s.id " \
                    "WHERE m.id in ({mailbox_ids})".format(
                        mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])
                    )
            result = query_plesk_db(self._server, query)
            for id, user_id, is_locked, mail_name, mbox_quota in result:
                if id in antiviruses:
                    antivirus = antiviruses[id]
                else:
                    antivirus = 'off'

                is_access_to_customer_panel = _is_access_to_customer_panel(user_id, is_locked)
                mailboxes_info.append((id, is_access_to_customer_panel, mail_name, mbox_quota, antivirus, 'false'))
        else:
            query = "SELECT m.id, m.userId, s.isLocked, m.mail_name, m.mbox_quota, m.virusfilter, m.spamfilter " \
                    "FROM mail as m " \
                    "LEFT JOIN smb_users as s ON m.userId = s.id " \
                    "WHERE m.id in ({mailbox_ids})".format(
                        mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])
                    )

            for id, user_id, is_locked, mail_name, mbox_quota, virusfilter, spamfilter in query_plesk_db(self._server, query):
                is_access_to_customer_panel = _is_access_to_customer_panel(user_id, is_locked)
                mailboxes_info.append((id, is_access_to_customer_panel, mail_name, mbox_quota, virusfilter, spamfilter))

        return mailboxes_info

    def _get_autoresponders(self, site_id):
        autoresponders = {}
        try:
            api_filter = plesk_ops.MailOperator.FilterBySiteId([site_id])
            plesk_operator = plesk_ops.MailOperator.GetAutoresponder(filter=api_filter)
            with self._plesk_api.logger.censor({'xml': ['/packet/mail/get_info/result/mailname/password/value']}):
                for result in self._plesk_api.send(plesk_operator):
                    mailbox_info, responder = result.data
                    autoresponders[mailbox_info.id] = (
                        responder.subject, responder.message, responder.forward_to, responder.end_date,
                        parse_bool(responder.enabled)
                    )
        except PleskError as e:
            if e.code == 1026:
                # It's ok if mail server component is not available in Plesk.
                pass
            else:
                raise e

        return autoresponders

    def _get_antiviruses(self, site_id):
        # For linux get antivirus from database
        if not self._is_windows:
            return None

        antiviruses = {}
        try:
            api_filter = self._get_antivirus_filter(site_id)
            plesk_operator = plesk_ops.MailOperator.GetAntivirus(filter=api_filter)
            with self._plesk_api.logger.censor({'xml': ['/packet/mail/get_info/result/mailname/password/value']}):
                for result in self._plesk_api.send(plesk_operator):
                    mailbox_info, antivirus = result.data
                    antiviruses[mailbox_info.id] = antivirus.virusfilter
        except PleskError as e:
            if e.code == 1026:
                # It's ok if mail server component is not available in Plesk.
                pass
            else:
                raise e

        return antiviruses

    def _get_antivirus_filter(self, site_id):
        return plesk_ops.MailOperator.FilterBySiteId([site_id])

    def _get_autoresponder(self, id, autoresponders):
        if id not in autoresponders:
            return None

        (subject, message, forward_to, end_date, enabled) = autoresponders[id]
        return AutoresponderModel(
            subject=subject,
            message=message,
            forward_to=forward_to,
            end_date=end_date,
            enabled=enabled,
        )

    def _get_antivirus(self, antivirus):
        if antivirus in ['inout', 'in', 'out']:
            return antivirus
        elif antivirus == 'any':
            return 'inout'
        elif antivirus == 'incoming':
            return 'in'
        elif antivirus == 'outgoing':
            return 'out'
        else:
            return 'off'

    def _get_forwardings(self, batch_mailboxes):
        forwarding = {}
        query = "SELECT mn_id, address " \
                "FROM mail_redir " \
                "WHERE mn_id in ({mailbox_ids})".format(
                    mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])
                )
        for id, address in query_plesk_db(self._server, query):
            if id not in forwarding:
                forwarding[id] = []

            forwarding[id].append(address)

        return forwarding

    def _get_forwarding(self, id, forwardings):
        if id not in forwardings:
            return None

        addresses = [x for x in forwardings[id] if self._is_valid_email(x)]
        if not addresses:
            return None

        return ForwardingModel(addresses=addresses)

    def _is_valid_email(self, email):
        # For Plesk for Windows, the migration tool cannot properly retrieve IDN addresses for mail forwarding.
        # Non-UTF symbols contained in email addresses are rendered as "?".
        # Such email addresses are considered invalid and are not migrated.
        if '?' in email:
            return False

        return True

    def _get_aliases(self, batch_mailboxes):
        aliases = {}
        query = "SELECT mn_id, alias " \
                "FROM mail_aliases " \
                "WHERE mn_id in ({mailbox_ids})".format(
                    mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])
                )
        for id, alias in query_plesk_db(self._server, query):
            if id not in aliases:
                aliases[id] = []

            aliases[id].append(alias)

        return aliases

    def _get_alias(self, id, aliases):
        if id not in aliases:
            return None

        return AliasModel(addresses=[x for x in aliases[id]])

    def _get_spam_filter(self, is_spamfilter, mailbox_name, site_name):
        if not parse_bool(is_spamfilter.lower()):
            return None

        full_mailbox_name = '{mail_name}@{site_name}'.format(mail_name=mailbox_name, site_name=site_name)
        query = "SELECT preference, value " \
                "FROM spamfilter " \
                "INNER JOIN spamfilter_preferences ON spamfilter.id = spamfilter_preferences.spamfilter_id " \
                "WHERE username = '{mailbox_name}'".format(mailbox_name=full_mailbox_name)
        params = {
            'required_score': None,
            'rewrite_header': None,
            'action': None,
        }
        list_form = {
            'blacklist_from': [],
            'whitelist_from': []
        }

        for preference, value in query_plesk_db(self._server, query, values_count=2):
            if preference in params:
                params[preference] = value
            elif preference in list_form:
                list_form[preference].append(value)

        return SpamFilterModel(
            sensitivity=float(params['required_score']) if params['required_score'] is not None else None,
            rewrite_header=params['rewrite_header'],
            action=params['action'],
            blacklist=[x for x in list_form['blacklist_from']],
            whitelist=[x for x in list_form['whitelist_from']],
        )

    def _get_plesk_subdomains(self, sites):
        # Plesk 11 & 12 report subdomains via Site operator.
        return []

    def _get_plesk_sites(self, webspace_id):
        return get_plesk_sites_info(
            self._plesk_api,
            webspace_id=webspace_id,
            dataset=[plesk_ops.SiteOperator.Dataset.GEN_INFO, plesk_ops.SiteOperator.Dataset.HOSTING]
        )

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        return get_plesk_webspace_info(self._plesk_api, webspace_id)

    def _set_mailbox_passwords(self, sites, webspace_name, subscription_external_id):
        passwords = {}
        for site in sites.values():
            if not site.mailboxes:
                continue

            for batch_mailboxes in self._batch(site.mailboxes):
                query = "SELECT mail.account_id, mail_name, type, password " \
                        "FROM accounts " \
                        "INNER JOIN mail ON mail.account_id = accounts.id " \
                        "WHERE mail.id in ({mailbox_ids})".format(
                            mailbox_ids=','.join(
                                ["{id}".format(id=mailbox.external_id) for mailbox in batch_mailboxes]
                            )
                        )

                for result in query_plesk_db(self._server, query, values_count=4, log_output=False):
                    account_id = result[0]
                    mailbox_name = result[1]
                    if len(result) == 4 and result[3] != '':
                        password_type = result[2]
                        password = result[3]
                    else:
                        password_type = 'plain'
                        password = None

                    full_mailbox_name = u'{mailbox_name}@{site_name}'.format(
                        mailbox_name=mailbox_name, site_name=site.name
                    )
                    passwords[full_mailbox_name] = (account_id, password, password_type)

        if not passwords:
            return

        plain_passwords = decrypt_account_passwords(self._server, passwords)
        mailboxes_with_warnings = []
        for site in sites.values():
            for mailbox in site.mailboxes:
                full_mailbox_name = u'{mailbox_name}@{site_name}'.format(mailbox_name=mailbox.name, site_name=site.name)
                password = plain_passwords[full_mailbox_name]
                is_password_generated = False
                if password is None:
                    password = generate_random_password()
                    is_password_generated = True
                    mailboxes_with_warnings.append(full_mailbox_name)

                mailbox.password = password
                mailbox.is_password_generated = is_password_generated

        if mailboxes_with_warnings:
            global_context = Registry.get_instance().get_context()
            global_context.execution_wrapper.add_subscription_issue(
                subscription_external_id, Issue(Problem(
                    'decrypt-password',
                    Problem.NOTICE,
                    _T("UNABLE_TO_RETRIEVE_THE_PASSWORDS_FOR_MAILBOXES").format(
                        webspace_name=webspace_name,
                        mailboxes_with_warnings=', '.join(mailboxes_with_warnings)
                    )
                ), None)
            )

    @staticmethod
    def _batch(iterable):
        l = len(iterable)
        for ndx in range(0, l, MAXIMUM_BATCH_SIZE):
            yield iterable[ndx:min(ndx + MAXIMUM_BATCH_SIZE, l)]

    @staticmethod
    def _get_disk_quota(plesk_quota):
        """
        Convert Plesk mail box disk quota value from bytes to megabytes or return -1 in case of unlimited quota

        :type plesk_quota: unicode | str | int
        :rtype: int
        """
        in_bytes = int(plesk_quota)
        return in_bytes / ONE_MB if in_bytes >= 0 else -1


class Plesk10MailExporter(PleskMailExporter):
    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([x for x in sites.iterkeys()])
        )
        return self._plesk_api.send(operator)


class Plesk9MailExporter(Plesk10MailExporter):
    def _get_plesk_sites(self, webspace_id):
        # Plesk 8 & 9 don't support addon domains
        return []

    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([normalize_domain_name(x) for x in sites.iterkeys()]))
        subdomains_info = self._plesk_api.send(operator)

        return fix_plesk9_subdomains_ids(subdomains_info)

    def _get_subdomain_mailboxes(self, site_id, site_name):
        # Plesk 8 & 9 don't support mail on subdomains
        return []

    def _get_mailboxes_info(self, batch_mailboxes, antiviruses=None):
        mailboxes_info = []
        if self._is_windows:
            query = "SELECT id, mail_name, mbox_quota " \
                    "FROM mail " \
                    "WHERE mail.id in ({mailbox_ids})".format(
                        mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])

                    )
            result = query_plesk_db(self._server, query)
            for id, mail_name, mbox_quota in result:
                if id in antiviruses:
                    antivirus = antiviruses[id]
                else:
                    antivirus = 'off'

                mailboxes_info.append((id, 1, mail_name, mbox_quota, antivirus, 'false'))
        else:
            query = "SELECT m.id, m.mail_name, m.mbox_quota, m.virusfilter, m.spamfilter " \
                    "FROM mail as m " \
                    "WHERE m.id in ({mailbox_ids})".format(
                        mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])
                    )

            for id, mail_name, mbox_quota, virusfilter, spamfilter in query_plesk_db(self._server, query):
                mailboxes_info.append((id, 1, mail_name, mbox_quota, virusfilter, spamfilter))

        return mailboxes_info

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        # Plesk 8 & 9 don't have subscriptions and webspaces. We use domains for this role
        return get_plesk_domains_info(self._plesk_api, webspace_id)

    def _get_forwardings(self, batch_mailboxes):
        forwarding = {}
        query = "SELECT id, redir_addr " \
                "FROM mail " \
                "WHERE id in ({mailbox_ids}) AND redirect = 'true'".format(
                    mailbox_ids=','.join(["{id}".format(id=id) for id in batch_mailboxes])
                )
        for id, address in query_plesk_db(self._server, query):
            if id not in forwarding:
                forwarding[id] = []

            forwarding[id].append(address)

        return forwarding

    def _get_autoresponders(self, site_id):
        autoresponders = {}
        try:
            api_filter = plesk_ops.MailOperator.FilterByDomainId([site_id])
            plesk_operator = plesk_ops.MailOperator.GetAutoresponders(filter=api_filter)
            for result in self._plesk_api.send(plesk_operator):
                # Plesks 8 and 9 allow to configure multiple autoresponders per mailbox.
                # Each can be disabled individually and there is a mailbox-wide option
                # for the mailbox to disable/enable autoresponders. An autoresponder
                # works only if autoresponding is enabled for the mail box and the autoresponder itself is enabled
                mailbox_info, autoresponding = result.data
                autoresponders_enabled = parse_bool(autoresponding.enabled)
                for responder in autoresponding.autoresponders:
                    # add first enabled responder that we encounter
                    if parse_bool(responder.enabled):
                        autoresponders[mailbox_info.id] = (
                            responder.subject, responder.message, responder.forward_to, responder.end_date,
                            autoresponders_enabled
                        )
                        break
                else:
                    # if there are no enabled responders, but there are disabled ones, add the first disabled responder
                    if len(autoresponding.autoresponders) > 0:
                        responder = autoresponding.autoresponders[0]
                        autoresponders[mailbox_info.id] = (
                            responder.subject, responder.message, responder.forward_to, responder.end_date, False
                        )
        except PleskError as e:
            if e.code == 1026:
                # It's ok if mail server component is not available in Plesk.
                pass
            else:
                raise e

        return autoresponders

    def _get_antivirus_filter(self, site_id):
        return plesk_ops.MailOperator.FilterByDomainId([site_id])
