# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging

from collections import namedtuple

from migrator.core import _T, InternalError
from migrator.core.data.format.platform import PagedPlatformReader
from migrator.core.data.model.account import \
    AccountModel, AuthModel, PersonalInfoModel, ContactModel, AddressModel, StaffMemberModel, PhoneModel
from migrator.core.data.model.reseller import ResellerModel, ProviderModel
from migrator.core.data.model.service import ServiceModel
from migrator.core.data.model.service_plan import ServicePlanModel
from migrator.core.data.model.service_property import ServicePropertyModel
from migrator.core.data.model.source_host import SourceHostModel
from migrator.core.data.model.subscription import SubscriptionModel
from migrator.core.utils import is_subdomain
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.dns import resolve
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib import get_plesk_sites_info
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.api.xmlrpc.core import PleskErrorCode
from migrator.products.plesk.lib.api.xmlrpc.operator import ZONE_STATUS_ENABLED, ZONE_TYPE_MASTER
from migrator.settings import \
    PANEL_TYPE_PLESK, SOURCE_SERVICE_DNS_NS, SOURCE_SERVICE_FTP, SOURCE_SERVICE_MAIL, \
    SOURCE_SERVICE_WEB, SOURCE_SERVICE_DB, SOURCE_SERVICE_DNS
from migrator.odin_automation.platform.lib.db import get_system_locales

logger = logging.getLogger(__name__)

PLESK_ADMIN_ID = u'1'
WebspaceDomainsInfo = namedtuple('WebspaceDomainsInfo', ('webspace', 'addon_domains', 'aliases'))


class PleskExporter(PagedPlatformReader):
    """This is the exporter for the most up-to-date Plesk version."""
    ADMIN_ID = PLESK_ADMIN_ID

    def __init__(self):
        super(PleskExporter, self).__init__()
        context = Registry.get_instance().get_context()
        self.plesk_server = context.conn.get_panel_server(context, 'plesk')
        self.plesk_api = self.plesk_server.plesk_api()

    @trace
    def export_next_account_page(self):
        """Retrieve source panel accounts in batches and return a batch upon each call"""

        # TODO use implementation from the base class
        # TODO _paginate(admins) + _paginate(resellers)
        # export and commit admins first so we can refer to them from resellers
        admins = self._get_admin_account_list()
        if len(admins) > 0:
            yield admins

        # export and commit resellers so we can refer to them from accounts
        resellers = self.get_reseller_list(with_accounts=False)
        if len(resellers) > 0:
            for res in resellers:
                res.parent_external_id = self.ADMIN_ID

            yield resellers

        account_ownership = self.get_account_ownership_list()
        account_ids = account_ownership.keys()
        account_batches = self._paginate(account_ids)

        for account_id_batch in account_batches:
            logger.debug('Joined System Bach exporting the account IDs ' + repr(account_id_batch))
            new_accounts = self._export_accounts(account_id_batch)
            for acc in new_accounts:
                acc.parent_external_id = account_ownership[acc.external_id]

            yield new_accounts

    @trace
    def get_account_ownership_list(self):
        # XXX This is a client => owner map that we further need internally but the interface would probably
        # has to change to 'owner' => (client1..clientN)
        resellers = self.get_reseller_list(with_accounts=False)
        owner_id_list = [self.ADMIN_ID] + [r.external_id for r in resellers]

        accounts_and_owners = []
        for account_id in owner_id_list:
            xml_api_filter = self._api_filter_owner(account_id)
            client_accounts = self._get_client_ids(xml_api_filter)
            accounts_and_owners.extend((a, account_id) for a in client_accounts)

        return dict(accounts_and_owners)

    @trace
    def _export_accounts(self, account_id_list):
        """Retrieve from the source system the account objects with the specified IDs"""
        account_list = []
        admin_account_id = self._find_admin_account_id(account_id_list)
        if admin_account_id:
            # Retrieve admin's own subscriptions
            account_id_list.remove(admin_account_id)
            account_list.extend(self._get_admin_account_list())

        xml_api_filter = self._api_operator_customer.FilterById(account_id_list)
        account_list.extend(self._get_client_account_objects(xml_api_filter))

        return account_list

    @trace
    def _find_admin_account_id(self, account_id_list):
        """Find the OA_ADMIN_ID value in the given list of account IDs"""
        return self.ADMIN_ID if self.ADMIN_ID in account_id_list else None

    @trace
    def get_accounts_owned_by(self, owner_id):
        """
        Return accounts of the clients that belong to reseller with ID 'owner_id'.

        If owner_id points to Plesk admin account, return all admin's clients and a newly created additional
        account, that contains subscriptions owned by admin himself
        """
        accounts = self._get_admin_account_list() if owner_id == self.ADMIN_ID else []

        filter = self._api_filter_owner(owner_id)
        account_list = self._get_client_account_objects(filter)
        for acc in account_list:
            acc.parent_external_id = owner_id

        accounts.extend(account_list)

        return accounts

    @trace
    def get_reseller_list(self, with_accounts=True):
        # TODO cache the result to avoid repetitive API requests
        source_resellers = self.plesk_api.send(
            plesk_ops.ResellerOperator.Get(
                dataset=[plesk_ops.ResellerOperator.Dataset.GEN_INFO],
                filter=plesk_ops.ResellerOperator.FilterAll()
            )
        )

        resellers = []
        for reseller in source_resellers:
            reseller_id = reseller.data[0]
            reseller_data = reseller.data[1]

            contact = self._create_contact_object(reseller_data.gen_info)
            address = self._create_address_object(reseller_data.gen_info)

            staff_members = self._get_staff_members_list(reseller_data.gen_info)
            staff_members.extend(
                self._create_account_admin_staff_member(reseller_data.gen_info.email, address, contact)
            )
            accounts = self.get_accounts_owned_by(reseller_id) if with_accounts else []
            resellers.append(ResellerModel(
                external_id=str(reseller_id),
                info=PersonalInfoModel(
                    company_name=reseller_data.gen_info.cname,
                    admin_contact=contact,
                    tech_contact=contact,
                    billing_contact=contact,
                    company_address=address,
                    language=self._lang_to_locale(reseller_data.gen_info.locale),
                ),
                accounts=accounts,
                staff_members=staff_members,
                subscriptions=self.get_subscriptions_list(reseller_id),
                service_plans=self.get_service_plans_list(reseller_id),
            ))

        return resellers

    @trace
    def _get_staff_members_list(self, owner_data):
        plesk_operator = self._create_staff_members_api_request(owner_data.guid)
        source_staff_members = self.plesk_api.send(plesk_operator)

        staff_members = []
        for staff_member in source_staff_members:
            staff_member_data = staff_member.data.gen_info
            if staff_member_data.status == 'disabled' or staff_member_data.mailname_id is not None:
                continue

            staff_members.append(self._get_staff_member_model(staff_member_data))

        return staff_members

    @trace
    def _get_staff_member_model(self, staff_member_data):
        return StaffMemberModel(
            authentication=AuthModel(
                login=staff_member_data.email or '',
                password='',
            ),
            address=AddressModel(
                zipcode=staff_member_data.contact_info.zip,
                city=staff_member_data.contact_info.city,
                country=staff_member_data.contact_info.country,
                state=staff_member_data.contact_info.state,
                address1=staff_member_data.contact_info.address,
                address2='',
            ),
            contact=ContactModel(
                title='',
                first_name=staff_member_data.name,
                middle_name='',
                last_name='',
                phone=PhoneModel('', '', staff_member_data.contact_info.phone, ''),
                fax=PhoneModel('', '', staff_member_data.contact_info.fax, ''),
                email=staff_member_data.email,
            ),
            language=self._lang_to_locale(staff_member_data.contact_info.locale)
        )

    @trace
    def get_subscriptions_list(self, owner_id=None):
        subscriptions = []
        dataset = [
            plesk_ops.SubscriptionOperator.Dataset.GEN_INFO,
            plesk_ops.SubscriptionOperator.Dataset.SUBSCRIPTIONS
        ]
        api_filter = plesk_ops.SubscriptionOperator.FilterByOwnerId([owner_id])
        plesk_operator = plesk_ops.SubscriptionOperator.Get(dataset=dataset, filter=api_filter)
        source_subscriptions = self.plesk_api.send(plesk_operator)

        for subscription in source_subscriptions:
            subscription_id, subscription_data = subscription.data
            service_plan_external_id = self._get_service_plan_external_id(subscription_data)

            subscriptions.append(SubscriptionModel(
                external_id=subscription_id,
                name=subscription_data.gen_info.name,
                service_plan_external_id=service_plan_external_id,
                services=self._get_subscription_services(),
            ))

            webspace_domains = self._get_webspace_domains(subscription)
            subscriptions.extend(self._get_addon_domain_dns_subscriptions_list(webspace_domains))
            subscriptions.extend(self._get_domain_alias_dns_subscriptions_list(webspace_domains))

        subscriptions.extend(self._get_main_domains_dns_subscriptions_list(source_subscriptions))

        return subscriptions

    @trace
    def _get_dns_subscriptions_list(self, sites_data, parent_external_id=None, is_alias=False):
        dns_zones = self._get_dns_zones([site_id for site_id, domain_name, service_plan in sites_data], is_alias)
        subscriptions = []
        for site_id, domain_name, service_plan in sites_data:
            if site_id not in dns_zones:
                # TODO: report as a subscription export error
                logger.warning(_T("DNS_ZONE_FOR_SITE_WAS_NOT_FOUND").format(
                    site_name=domain_name
                ))
                continue

            dns_zone = dns_zones[site_id]
            dns_service_properties = [
                ServicePropertyModel('domain', domain_name),
                ServicePropertyModel('is_domain_billed', '0'),
                ServicePropertyModel('is_dns_managed', '1' if self._is_zone_managed(dns_zone) else '0')
            ]
            subscriptions.append(SubscriptionModel(
                external_id=self._get_external_id_for_dns_subscription(site_id, is_alias),
                name=domain_name,
                service_plan_external_id=service_plan,
                services=[ServiceModel(SOURCE_SERVICE_DNS_NS, '1', []), ServiceModel(SOURCE_SERVICE_DNS, '1', dns_service_properties)],
                parent_external_id=parent_external_id or site_id,
            ))

        return subscriptions

    @trace
    def _get_main_domains_dns_subscriptions_list(self, domains):
        sites = []
        for domain in domains:
            domain_name = domain.data[1].gen_info.name
            site_id = domain.data[0]
            service_plan_external_id = self._get_service_plan_external_id(domain.data[1])
            sites.append(
                (site_id, domain_name, service_plan_external_id)
            )

        return self._get_dns_subscriptions_list(sites)

    @trace
    def _get_addon_domain_dns_subscriptions_list(self, webspace_domains):
        sites = []
        subscription_id = webspace_domains.webspace.data[0]
        service_plan_external_id = self._get_service_plan_external_id(webspace_domains.webspace.data[1])
        for domain in webspace_domains.addon_domains:
            domain_name = domain.data[1].gen_info.name
            site_id = domain.data[0]
            if not self._is_subdomain(domain_name, webspace_domains):
                sites.append(
                    (site_id, domain_name, service_plan_external_id)
                )

        return self._get_dns_subscriptions_list(sites, subscription_id)

    @trace
    def _get_domain_alias_dns_subscriptions_list(self, webspace_domains):
        sites = []
        subscription_id = webspace_domains.webspace.data[0]
        service_plan_external_id = self._get_service_plan_external_id(webspace_domains.webspace.data[1])
        for domain in webspace_domains.aliases:
            domain_name = domain.data[1].name
            site_id = domain.data[0]
            if not self._is_subdomain(domain_name, webspace_domains):
                sites.append(
                    (site_id, domain_name, service_plan_external_id)
                )

        return self._get_dns_subscriptions_list(sites, subscription_id, True)

    @trace
    def _get_webspace_domains(self, webspace):
        addon_domains = self._get_addon_domains(webspace.data[0])
        site_ids = [domain.data[0] for domain in addon_domains] + [webspace.data[0]]
        aliases = self._get_domain_aliases(site_ids)
        return WebspaceDomainsInfo(
            webspace=webspace,
            addon_domains=addon_domains,
            aliases=aliases
        )

    @staticmethod
    def _is_subdomain(domain, webspace_domains):
        if is_subdomain(domain, webspace_domains.webspace.data[1].gen_info.name):
            return True

        for addon_domain in webspace_domains.addon_domains:
            if is_subdomain(domain, addon_domain.data[1].gen_info.name):
                return True

        for alias in webspace_domains.aliases:
            if is_subdomain(domain, alias.data[1].name):
                return True

        return False

    @trace
    def _get_dns_zones(self, site_ids, is_alias):
        if is_alias:
            entities_filter = self._api_filter_dns_domain_alias_ids(site_ids)
        else:
            entities_filter = self._api_filter_dns_site_ids(site_ids)

        plesk_operator = plesk_ops.DnsOperator.Get(filter=entities_filter, soa=None)
        dns_zones = self.plesk_api.send(plesk_operator)
        dns_zones_dict = {}
        for dns_zone in dns_zones:
            if not hasattr(dns_zone, 'code'):
                dns_zones_dict[dns_zone.data.object_id] = dns_zone.data

        return dns_zones_dict

    @trace
    def _get_addon_domains(self, subscription_id):
        return get_plesk_sites_info(
            self.plesk_api,
            webspace_id=subscription_id,
            dataset=[plesk_ops.SiteOperator.Dataset.GEN_INFO]
        )

    @trace
    def _get_domain_aliases(self, site_ids):
        operator = self._api_operator_alias.Get(filter=self._api_filter_alias_site_ids(site_ids))
        return self.plesk_api.send(operator)

    @staticmethod
    def _is_zone_managed(zone):
        return zone.zone_status == ZONE_STATUS_ENABLED and zone.zone_type == ZONE_TYPE_MASTER

    @classmethod
    def _get_external_id_for_dns_subscription(cls, site_id, is_alias):
        new_id = site_id
        if is_alias:
            # Aliases and domains are different kinds of objects in Plesk and their ids may intersect, so prefix
            # alias id with 'alias-'
            new_id = 'alias_%s' % site_id

        return 'dns_%s' % new_id

    @classmethod
    def is_domain_alias(cls, dns_subscription_external_id):
        """
        :param dns_subscription_external_id: str Examples: 'dns_5', 'dns_alias_2'
        :return: bool
        """
        return 'dns_alias_' in dns_subscription_external_id

    @classmethod
    def extract_domain_alias_id(cls, dns_subscription_external_id):
        """
        Get domain-alias ID

        :param dns_subscription_external_id: str Examples: 'dns_5', 'dns_alias_2'
        :return: str Example: '2'
        """
        if not cls.is_domain_alias(dns_subscription_external_id):
            raise InternalError(_T('UNEXPECTED_DNS_SUBSCRIPTION_ID'))

        return dns_subscription_external_id.split('dns_alias_')[1]

    @classmethod
    def extract_domain_id(cls, dns_subscription_external_id):
        """
        Get domain ID

        :param dns_subscription_external_id: str Examples: 'dns_5', 'dns_alias_2'
        :return: str Example: '5'
        """
        if cls.is_domain_alias(dns_subscription_external_id) or 'dns_' not in dns_subscription_external_id:
            raise InternalError(_T('UNEXPECTED_DNS_SUBSCRIPTION_ID'))

        return dns_subscription_external_id.split('dns_')[1]

    @staticmethod
    def get_fqdn_without_last_dot(dns_zone):
        """
        Get FQDN of dns_zone without last dot (.) of 0-level domain

        :param dns_zone: str Examples: u'mail.example.com.', u'mail.example.com'
        :return: str u'mail.example.com'
        """
        return dns_zone[:-1] if dns_zone.endswith(u'.') else dns_zone

    @property
    def _api_operator_alias(self):
        return plesk_ops.AliasOperator

    @staticmethod
    def _api_filter_dns_site_ids(site_ids):
        return plesk_ops.DnsOperator.FilterBySiteId(site_ids)

    @staticmethod
    def _api_filter_dns_domain_alias_ids(site_ids):
        return plesk_ops.DnsOperator.FilterBySiteAliasId(site_ids)

    @staticmethod
    def _api_filter_alias_site_ids(site_ids):
        return plesk_ops.AliasOperator.FilterBySiteId(site_ids)

    @staticmethod
    def _get_service_plan_external_id(site_data):
        if len(site_data.subscriptions.subscriptions) > 0:
            return site_data.subscriptions.subscriptions[0].plan.guid

        return None

    @trace
    def get_service_plans_list(self, owner_id=PLESK_ADMIN_ID):
        service_plans = []
        api_filter = plesk_ops.ServicePlanOperator.FilterAll()
        plesk_operator = plesk_ops.ServicePlanOperator.Get(owner_id=owner_id, owner_all=None, filter=api_filter)
        source_service_templates = self.plesk_api.send(plesk_operator)

        for service_template in source_service_templates:
            plan_guid = service_template.data.plan_guid
            model_object = ServicePlanModel(external_id=plan_guid, name=service_template.data.name)
            service_plans.append(model_object)

        return service_plans

    @trace
    def get_source_hosts_list(self):
        plesk_hostname = self.plesk_server.ip()
        plesk_ip = resolve(plesk_hostname)
        if plesk_ip is None:
            raise Exception(_T("FAILED_TO_RESOLVE_HOSTNAME_TO_AN_IP_ADDRESS").format(hostname=plesk_hostname))

        return [SourceHostModel(
            external_id='1',
            name=plesk_hostname,
            ip=plesk_ip,
            panel=PANEL_TYPE_PLESK,
        )]

    @trace
    def _get_admin_account_list(self):
        """Create an AccountModel object for Plesk admin. Returns a list."""
        subscriptions = self.get_subscriptions_list(self.ADMIN_ID)

        dataset = [plesk_ops.ServerOperator.Dataset.ADMIN]
        admin_info = self.plesk_api.send(plesk_ops.ServerOperator.Get(dataset=dataset)).data.admin
        login = admin_info.email

        contact = self._create_contact_object(admin_info)
        address = self._create_address_object(admin_info)

        company_name = admin_info.cname if admin_info.cname.strip() != '' else admin_info.email

        # Explicitly mark Plesk admin account as such for better readability in UI
        admin_name_suffix = 'Plesk admin'
        if contact.first_name.strip() == '':
            contact.first_name = u'{0}'.format(admin_name_suffix)
        else:
            contact.first_name = u'{0} ({1})'.format(contact.first_name, admin_name_suffix)

        admin_account = ProviderModel(
            external_id=self.ADMIN_ID,
            info=PersonalInfoModel(
                company_name=company_name,
                admin_contact=contact,
                tech_contact=contact,
                billing_contact=contact,
                company_address=address,
                language=self._lang_to_locale(''),
            ),
            staff_members=[
                StaffMemberModel(
                    authentication=AuthModel(
                        login=login,
                        password='',
                    ),
                    address=address,
                    contact=contact,
                    language=self._lang_to_locale(''),
                )],
            subscriptions=subscriptions,
        )
        return [admin_account]

    @property
    def _api_operator_customer(self):
        """Could be ClientOperator or CustomerOperator, depending on Plesk version."""
        return plesk_ops.CustomerOperator

    def _api_filter_owner(self, owner_id):
        """Could be FilterOwnerId or FilterByOwnerId, or FilterAll, depending on Plesk version"""
        return self._api_operator_customer.FilterByOwnerId([owner_id])

    def _get_reseller_accounts(self):
        """Return accounts that contain subscriptions owned by resellers"""
        reseller_accounts = []
        # TODO
        return reseller_accounts

    @trace
    def _get_client_ids(self, xml_api_filter):
        """Make a 'customer.get' XML API request and return a list of existing client IDs"""
        accounts = []
        dataset = [self._api_operator_customer.Dataset.GEN_INFO]
        request = self._api_operator_customer.Get(dataset=dataset, filter=xml_api_filter)
        source_accounts = self.plesk_api.send(request)

        for account in source_accounts:
            if not account.ok and account.code == PleskErrorCode.OBJECT_CAN_NOT_BE_PARENT:
                continue

            account_data = account.data[1]
            accounts.append(str(account_data.id))

        return accounts

    @trace
    def _get_client_account_objects(self, xml_api_filter):
        """Make a 'client.get' Plesk XML API request; return a list of Plesk account objects"""
        request = self._api_operator_customer.Get(
            dataset=[
                self._api_operator_customer.Dataset.GEN_INFO
            ],
            filter=xml_api_filter
        )
        source_accounts = self.plesk_api.send(request)

        accounts = []
        for account in source_accounts:
            if not account.ok:
                if account.code == PleskErrorCode.OBJECT_CAN_NOT_BE_PARENT:
                    continue

                account.check()

            account_data = account.data[1]
            contact = self._create_contact_object(account_data)
            address = self._create_address_object(account_data)
            staff_members = self._get_staff_members_list(account_data)
            staff_members.extend(self._create_account_admin_staff_member(account_data.email, address, contact))
            accounts.append(AccountModel(
                external_id=str(account_data.id),
                info=PersonalInfoModel(
                    company_name=account_data.cname,
                    admin_contact=contact,
                    tech_contact=contact,
                    billing_contact=contact,
                    company_address=address,
                    language=self._lang_to_locale(account_data.locale),
                ),
                staff_members=staff_members,
                subscriptions=self.get_subscriptions_list(account_data.id),
            ))

        return accounts

    @staticmethod
    def _get_subscription_services():
        """Return a list of subscription service model objects (hard-coded for now)"""
        service_names = [
            SOURCE_SERVICE_WEB,
            SOURCE_SERVICE_MAIL,
            SOURCE_SERVICE_FTP,
            SOURCE_SERVICE_DB,
        ]

        return [ServiceModel(type=s, source_host_external_id='1') for s in service_names]

    @staticmethod
    def _create_staff_members_api_request(owner_guid):
        dataset = [plesk_ops.UserOperator.Dataset.GEN_INFO]
        api_filter = plesk_ops.UserOperator.FilterOwnerGuid([owner_guid])

        return plesk_ops.UserOperator.Get(dataset=dataset, filter=api_filter)

    @staticmethod
    def _create_contact_object(account_data):
        return ContactModel(
            title='',
            first_name=account_data.pname,
            middle_name='',
            last_name='',
            phone=PhoneModel('', '', account_data.phone, ''),
            fax=PhoneModel('', '', account_data.fax, ''),
            email=account_data.email,
        )

    @staticmethod
    def _create_address_object(account_data):
        return AddressModel(
            zipcode=account_data.pcode,
            city=account_data.city,
            country=account_data.country,
            state=account_data.state,
            address1=account_data.address,
            address2='',
        )

    def _create_account_admin_staff_member(self, email, address, contact):
        """Create a new staff member based on main account information"""
        return []

    @staticmethod
    def _lang_to_locale(language):
        """Map Plesk locale to OA locale. Unsupported locale will be replaced with en_US"""
        language = language.replace('-', '_')
        return language if language in get_system_locales() else 'en_US'


class Plesk178Exporter(PleskExporter):
    def _get_staff_members_list(self, owner_data):
        return []

    def _create_account_admin_staff_member(self, email, address, contact):
        """Create a default staff member based on existing account information"""
        authentication = AuthModel(login=email, password='')
        return [
            StaffMemberModel(
                authentication=authentication,
                address=address,
                contact=contact,
                language=self._lang_to_locale(''),
            )
        ]


class Plesk10Exporter(PleskExporter):
    @trace
    def _get_staff_members_list(self, owner_data):
        plesk_operator = self._create_staff_members_api_request(owner_data.guid)
        source_staff_members = self.plesk_api.send(plesk_operator)

        staff_members = []
        for staff_member in source_staff_members:
            staff_member_data = staff_member.data.gen_info
            if staff_member_data.status == 'disabled' or staff_member_data.mailname_id is not None:
                continue

            staff_member_model = self._get_staff_member_model(staff_member_data)
            if staff_member_data.is_built_in:
                # We need to use owner's account email for built in staff member login
                staff_member_model.authentication.login = owner_data.email
                staff_member_model.contact.email = owner_data.email
            else:
                staff_member_model.authentication.login = staff_member_data.login
                staff_member_model.contact.email = staff_member_data.login

            staff_members.append(staff_member_model)

        return staff_members


class Plesk9Exporter(Plesk10Exporter):
    @trace
    def get_subscriptions_list(self, owner_id=None):
        """Retrieve a list of domains for Plesk versions that has no subscription feature"""

        domains = []
        domain_objects = self._get_domains(owner_id)
        for domain in domain_objects:
            domain_id = domain.data[0]
            domain_info = domain.data[1]
            domain_model = SubscriptionModel(
                external_id=domain_id,
                name=domain_info.gen_info.name,
                service_plan_external_id=None,
                services=self._get_subscription_services(),
            )
            domains.append(domain_model)

            webspace_domains = self._get_webspace_domains(domain)
            domains.extend(self._get_domain_alias_dns_subscriptions_list(webspace_domains))

        domains.extend(self._get_main_domains_dns_subscriptions_list(domain_objects))
        return domains

    def get_service_plans_list(self, owner_id=PLESK_ADMIN_ID):
        """Plesk version 9 or earlier has no service plan feature"""
        return []

    def _get_staff_members_list(self, owner_data):
        return []

    @trace
    def _get_domains(self, owner_id):
        dataset = [self._api_operator_customer.Dataset.GEN_INFO]
        api_filter = self._api_filter_domain_owner(owner_id)
        request = plesk_ops.DomainOperator.Get(dataset=dataset, filter=api_filter)
        source_domains = self.plesk_api.send(request)

        return source_domains

    def _create_account_admin_staff_member(self, email, address, contact):
        """Create a default staff member based on existing account information"""
        authentication = AuthModel(login=email, password='')
        return [
            StaffMemberModel(
                authentication=authentication,
                address=address,
                contact=contact,
                language=self._lang_to_locale(''),
            )
        ]

    @staticmethod
    def _get_service_plan_external_id(site_data):
        return None

    def _get_addon_domains(self, subscription_id):
        return []

    @property
    def _api_operator_customer(self):
        return plesk_ops.ClientOperator

    @property
    def _api_operator_alias(self):
        return plesk_ops.DomainAliasOperator

    def _api_filter_owner(self, owner_id):
        """Could be FilterOwnerId of FilterByOwnerId, depending on Plesk version"""
        return self._api_operator_customer.FilterOwnerId([owner_id])

    def _api_filter_domain_owner(self, owner_id):
        """Could be FilterByClientLogin of FilterByOwnerId, depending on Plesk version"""
        return plesk_ops.DomainOperator.FilterByOwnerId([owner_id])

    @staticmethod
    def _api_filter_dns_site_ids(site_ids):
        return plesk_ops.DnsOperator.FilterByDomainId(site_ids)

    @staticmethod
    def _api_filter_alias_site_ids(site_ids):
        return plesk_ops.DomainAliasOperator.FilterByDomainId(site_ids)

    @staticmethod
    def _api_filter_dns_domain_alias_ids(site_ids):
        return plesk_ops.DnsOperator.FilterByDomainAliasId(site_ids)


class Plesk8Exporter(Plesk9Exporter):
    # Plesk 8.x has no admin account in DB at all.
    # We need to simulate it by using other than PLESK_ADMIN_ID unique ID.
    ADMIN_ID = u'-1'

    def get_reseller_list(self, with_accounts=False):
        """Old versions of Plesk (<9.0.0) do not have reseller feature. We return an empty list"""
        return []

    @trace
    def get_subscriptions_list(self, owner_id=None):
        """Plesk 8 API requires client login for subscription.get(), instead of client ID"""
        if owner_id == self.ADMIN_ID:
            # Plesk 8 admin cannot own subscriptions
            return []

        dataset = [self._api_operator_customer.Dataset.GEN_INFO]
        api_filter = self._api_operator_customer.FilterById([owner_id])
        request = self._api_operator_customer.Get(dataset=dataset, filter=api_filter)
        client_info = self.plesk_api.send(request)
        client_login = client_info[0].data[1].login

        return super(Plesk8Exporter, self).get_subscriptions_list(client_login)

    def _find_admin_account_id(self, account_id_list):
        """Plesk 8 admin cannot own subscriptions, hence we do not migrate admin account"""
        return None

    def _api_filter_domain_owner(self, owner_id):
        """Could be FilterByClientLogin of FilterByOwnerId, depending on Plesk version"""
        return plesk_ops.DomainOperator.FilterByClientLogin([owner_id])

    def _api_filter_owner(self, owner_id):
        """In Plesk 8, all clients are owned by admin"""
        return self._api_operator_customer.FilterAll()
