# (C)2017 Ingram Micro Inc. all Rights Reserved
import ntpath
import posixpath

from lxml import etree

from migrator.core.data.model.web_hosting_shm import VirtualDirectoryModel
from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.loggers.report import Issue, Problem
from migrator.core.utils.operating_system.windows_utils import fix_path
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib import get_plesk_webspace_info, get_plesk_sites_info
from migrator.products.plesk.lib.export import is_outside
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops


class PleskHostingBaseExporter(object):

    _physical_hosting = [
        plesk_ops.SiteHostingVirtual, plesk_ops.SubscriptionHostingVirtual, plesk_ops.SubdomainHostingVirtual,
        plesk_ops.DomainHostingVirtual
    ]

    _std_forwarding = [plesk_ops.SiteHostingStandardForwarding]

    _frame_forwarding = [
        plesk_ops.SiteHostingFrameForwarding, plesk_ops.SubscriptionHostingFrameForwarding,
        plesk_ops.SubscriptionHostingStandardForwarding, plesk_ops.DomainHostingFrameForwarding,
        plesk_ops.DomainHostingStandardForwarding
    ]

    _alias = [plesk_ops.AliasInfo]

    _nohosting = [plesk_ops.SubscriptionHostingNone, plesk_ops.SiteHostingNone]

    @staticmethod
    def _is_hosting_type(hosting, types):
        for x in types:
            if isinstance(hosting, x):
                return True
        return False

    @classmethod
    def _is_physical_hosting(cls, hosting):
        return cls._is_hosting_type(hosting, cls._physical_hosting)

    @classmethod
    def _is_std_forwarding(cls, hosting):
        return cls._is_hosting_type(hosting, cls._std_forwarding)

    @classmethod
    def _is_frame_forwarding(cls, hosting):
        return cls._is_hosting_type(hosting, cls._frame_forwarding)

    @classmethod
    def _is_alias(cls, hosting):
        return cls._is_hosting_type(hosting, cls._alias)

    @classmethod
    def _is_nohosting(cls, hosting):
        return cls._is_hosting_type(hosting, cls._nohosting)

    def __init__(self, server=None):
        if server is None:
            context = Registry.get_instance().get_context()
            server = context.conn.get_panel_server(context, 'plesk')

        self._is_windows = server.is_windows()
        self._plesk_api = server.plesk_api()
        self._server = server
        self._pathlib = ntpath if self._is_windows else posixpath

    def _get_vdirs(
            self, node, subscription_id, site_name, www_root, webspace_www_root, other_site_prefixes, template_location
    ):

        # Get list of directories to remove from vdirs XML
        # The list is taken from Plesk's RestoreVirtualDirectories function
        # at windows/binary/plesk-utils/PMM/psadumpagent/RestoreEnv.cs

        _reserved_vdirs = [
            '_vti_bin', 'jakarta', 'JRunScripts', 'CFIDE', 'cfdocs', 'cgi-bin', 'plesk-stat', 'sitepreview'
        ]

        nodes = node.xpath(u'./vdirs/vdir')
        vdirs = []
        for node in nodes:
            attrib = dict(node.attrib)
            if attrib['name'] in _reserved_vdirs:
                continue

            location = attrib['path']
            if self._is_windows:
                location = fix_path(attrib['path'])

            if template_location:
                if location == www_root:
                    location = '${SITEROOT}'
                elif location.startswith(www_root) and is_outside(location, self._is_windows, *other_site_prefixes):
                    location = '${SITEROOT}\\' + location[len(www_root):]
                elif location.startswith(webspace_www_root) and is_outside(location, self._is_windows, *other_site_prefixes):
                    location = '${WEBSPACEROOT}\\' + location[len(webspace_www_root):]
                else:
                    # We cannot migrate virtual dirs phisically placed outside of site's wwwroot
                    global_context = Registry.get_instance().get_context()
                    global_context.execution_wrapper.add_subscription_issue(
                        subscription_id, Issue(Problem(
                            'virtual-directory',
                            Problem.NOTICE,
                            _T('UNABLE_TO_MIGRATE_THE_VIRTUAL_DIRECTORY').format(
                                site_name=site_name,
                                vdir_name=attrib['name'],
                                vdir_path=attrib['path'],
                            )
                        ), None)
                    )
                    continue

            vdirs.append(VirtualDirectoryModel(
                name=attrib['name'],
                path=location,
                virtual_directories=self._get_vdirs(
                    node, subscription_id, site_name, www_root, webspace_www_root, other_site_prefixes,
                    template_location
                ),

                asp_enable_client_debug_inherited=attrib.get('aspEnableClientDebugInherited'),
                asp_enable_client_debug_default=attrib.get('aspEnableClientDebugDefault'),
                asp_enable_client_debug=attrib.get('aspEnableClientDebug'),
                asp_enable_server_debug_inherited=attrib.get('aspEnableServerDebugInherited'),
                asp_enable_server_debug_default=attrib.get('aspEnableServerDebugDefault'),
                asp_enable_server_debug=attrib.get('aspEnableServerDebug'),
                asp_buffering_limit_inherited=attrib.get('aspBufferingLimitInherited'),
                asp_buffering_limit_default=attrib.get('aspBufferingLimitDefault'),
                asp_buffering_limit=attrib.get('aspBufferingLimit'),
                asp_max_request_entity_allowed_inherited=attrib.get('aspMaxRequestEntityAllowedInherited'),
                asp_max_request_entity_allowed_default=attrib.get('aspMaxRequestEntityAllowedDefault'),
                asp_max_request_entity_allowed=attrib.get('aspMaxRequestEntityAllowed'),
                editable=attrib.get('editable'),
                script_maps=attrib.get('scriptMaps'),
                require_ssl=attrib.get('requireSSL'),
                auth_ntlm=attrib.get('authNTLM'),
                auth_anonymous=attrib.get('authAnonymous'),
                asp_execute_in_mta=attrib.get('aspExecuteInMTA'),
                asp_enable_parent_paths=attrib.get('aspEnableParentPaths'),
                default_doc=attrib.get('defaultDoc'),
                enable_default_doc=attrib.get('enableDefaultDoc'),
                app_create=attrib.get('appCreate'),
                dont_log=attrib.get('dontLog'),
                enable_dir_browsing=attrib.get('enableDirBrowsing'),
                access_source=attrib.get('accessSource'),
                access_execute=attrib.get('accessExecute'),
                access_script=attrib.get('accessScript'),
                access_write=attrib.get('accessWrite'),
                access_read=attrib.get('accessRead')
            ))

        return vdirs

    @cached
    def _get_virtual_directories(self, subscription_id, www_root, site_name, template_location):
        www_root = fix_path(www_root)
        webspace_info = self._get_plesk_webspace_info(subscription_id)
        webspace_www_root = fix_path(self._get_plesk_site_webroot(webspace_info))

        other_site_prefixes = set([
            fix_path(self._get_plesk_site_webroot(x.data[1]))
                for x in self._get_plesk_sites(subscription_id)
                    if self._is_physical_hosting(x.data[1].hosting)
        ])
        other_site_prefixes.add(webspace_www_root)
        if www_root in other_site_prefixes:
            other_site_prefixes.remove(www_root)

        webspace_www_root = self._pathlib.dirname(webspace_www_root)
        if self._is_windows:
            with self._server.websrvmng_lock, self._server.runner() as runner:
                raw_vdir_info = runner.sh(
                    '{websrvmng_path} --list-vdirs --vhost-name={vhost_name}',
                    dict(websrvmng_path=self._server.websrvmng_bin, vhost_name=site_name.encode('idna'))
                ).encode('utf-8')

            vdir_xml = etree.XML(raw_vdir_info)
            nodes = vdir_xml.xpath(u'/vhost')
            return self._get_vdirs(
                nodes[0], subscription_id, site_name, www_root, webspace_www_root, other_site_prefixes,
                template_location
            )
        else:
            return []

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        return get_plesk_webspace_info(self._plesk_api, webspace_id)

    def _get_plesk_site_webroot(self, site_info):
        www_root = site_info.hosting.properties.get('www_root')
        if isinstance(www_root, basestring):
            www_root = www_root.rstrip('/\\')

        return www_root

    @cached
    def _get_plesk_sites(self, webspace_id):
        return get_plesk_sites_info(
            self._plesk_api,
            webspace_id=webspace_id,
            dataset=[plesk_ops.SiteOperator.Dataset.GEN_INFO, plesk_ops.SiteOperator.Dataset.HOSTING]
        )

    def _get_plesk_subdomains(self, sites):
        # Plesk 11 & 12 report subdomains via Site operator.
        return []
