# (C)2017 Ingram Micro Inc. all Rights Reserved
import os
import ntpath
import logging
import posixpath

from migrator.core.data.format.web_content_shm import ShmWebContentReader
from migrator.core.data.model.web_content_shm import \
    SubscriptionModel, WebspaceModel, SiteModel, ContentItemModel, PermissionsModel, OwnerModel
from migrator.core.utils import cached
from migrator.core.utils.networking.dns import normalize_domain_name
from migrator.core.utils.operating_system.windows_utils import fix_path
from migrator.products.plesk.lib import get_plesk_domains_info
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.export import is_outside
from migrator.products.plesk.lib.export.utils import fix_plesk9_subdomains_ids
from migrator.products.plesk.lib.export.web_base_shm import PleskHostingBaseExporter
from migrator.products.plesk.target.utils.plesk_utils import get_unix_vhosts_dir, get_windows_vhosts_dir

logger = logging.getLogger(__name__)


class PleskHostingContentExporter(PleskHostingBaseExporter, ShmWebContentReader):
    _vrt_hst_type = plesk_ops.SubscriptionHostingVirtual

    def __init__(self, server=None):
        super(PleskHostingContentExporter, self).__init__(server)

    def get_subscription(self, subscription_id):
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id)
        )

    def get_webspaces_list(self, subscription_id):
        res = list()
        webspace = self.get_webspace(subscription_id, subscription_id)
        if webspace:
            res.append(webspace)

        return res

    def get_webspace(self, subscription_id, webspace_id):
        webspace_info = self._get_plesk_webspace_info(webspace_id)
        sites = self.get_sites_list(subscription_id, webspace_id)
        webspace = WebspaceModel(
            external_id=webspace_id,
            sites=sites,
            content=self._get_plesk_webspace_content(webspace_id, webspace_info, sites)
        )
        if not self._is_windows and webspace.sites:
            # Add permissions model to every site content item to make migrated content have proper permissions and
            # ownership on target WHP node(s) in case of system user change during webspace creation.
            # We are on Linux and webspace has sites -> webspace_info.hosting is of virtual hosting type already
            with self._server.runner() as runner:
                stat_info = runner.stat(webspace.sites[0].source_webroot)

            for site in webspace.sites:
                for content_item in site.content:
                    content_item.permissions = self._get_permissions(stat_info)

        return webspace

    def _get_permissions(self, stat_info):
        return PermissionsModel(
            [
                OwnerModel('${SYSUSER}', stat_info.user_id),
                OwnerModel('${SYSUSER}', stat_info.user)
            ]
        )

    def get_sites_list(self, subscription_id, webspace_id):
        sites = dict()

        # Make Plesk webspace to be a site due to SHM requirements
        webspace_info = self._get_plesk_webspace_info(webspace_id)
        if not isinstance(webspace_info.hosting, self._vrt_hst_type):
            # Return empty list since you can't have addon sites or subdomains on webspace w/o physical hosting
            return []

        sites[webspace_info.gen_info.name] = SiteModel(
            external_id=webspace_id,
            name=webspace_info.gen_info.name,
            source_webroot=self._get_plesk_site_webroot(webspace_info),
            content=self._get_plesk_site_content(webspace_info)
        )

        # Get all webspace sites
        sites_info = self._get_plesk_sites(webspace_id)
        for result in sites_info:
            site_id = result.data[0]
            site_info = result.data[1]

            if not isinstance(site_info.hosting, plesk_ops.SiteHostingVirtual):
                # Skip sites w/o physical hosting
                continue

            sites[site_info.gen_info.name] = SiteModel(
                external_id=site_id,
                name=site_info.gen_info.name,
                source_webroot=self._get_plesk_site_webroot(site_info),
                content=self._get_plesk_site_content(site_info)
            )

        # Get all subdomains as sites
        subdomains_info = self._get_plesk_subdomains(sites)
        for result in subdomains_info:
            subdomain_id = result.data[0]
            subdomain_info = result.data[1]

            subdomain_content = self._get_plesk_subdomain_content(subdomain_info)
            if not subdomain_content:
                # skip subdomain w/o physical hosting
                continue

            sites[subdomain_info.name] = SiteModel(
                external_id=subdomain_id,
                name=subdomain_info.name,
                source_webroot=self._get_plesk_site_webroot(subdomain_info),
                content=subdomain_content
            )

        return sites.values()

    def _get_plesk_site_content(self, site_info):
        return self._get_plesk_site_content_items(self._get_plesk_site_webroot(site_info))

    def _get_plesk_site_content_items(self, site_webroot):
        return [
            ContentItemModel(
                source_path=os.path.join(site_webroot, ''),
                target_path='${SITEROOT}/',
                is_wwwroot=True,
                exclude=['.user.ini'] if self._is_windows else None,
            ),
        ]

    def _get_plesk_webspace_content(self, webspace_id, webspace_info, sites):
        if not self._is_windows or not sites:
            return []

        pathlib = ntpath if self._is_windows else posixpath
        webspace_www_root = pathlib.dirname(fix_path(self._get_plesk_site_webroot(webspace_info)))

        other_site_prefixes = [fix_path(x.source_webroot) for x in sites]
        vdir_pathes = set([])
        for site in sites:
            for vdir in self._get_virtual_directories(webspace_id, site.source_webroot, site.name, False):
                vdir_path = fix_path(vdir.path)
                if is_outside(vdir_path, self._is_windows, *other_site_prefixes) and vdir_path.startswith(webspace_www_root):
                    pos = vdir_path.find('\\', len(webspace_www_root) + 1)
                    vdir_pathes.add(vdir_path if pos == -1 else vdir_path[:pos])

        content = []
        for path in vdir_pathes:
            content.append(ContentItemModel(
                source_path=path,
                target_path='${WEBSPACEROOT}/'
            ))

        return content

    def _get_plesk_subdomain_content(self, subdomain_info):
        return []


class Plesk10HostingContentExporter(PleskHostingContentExporter):
    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([x for x in sites.iterkeys()])
        )
        return self._plesk_api.send(operator)

    def _get_plesk_subdomain_content(self, subdomain_info):
        site_webroot = self._get_plesk_site_webroot(subdomain_info)
        if not site_webroot:
            # subdomain w/o physical hosting have empty www_root value in Plesk 10.4
            return []

        return self._get_plesk_site_content_items(site_webroot)


class Plesk9HostingContentExporter(Plesk10HostingContentExporter):
    _vrt_hst_type = plesk_ops.DomainHostingVirtual

    def __init__(self, server=None):
        super(Plesk9HostingContentExporter, self).__init__(server)

        agent_api = self._server.agent_api()
        if self._is_windows:
            self._vhosts_dir = get_windows_vhosts_dir(agent_api)
        else:
            self._vhosts_dir = get_unix_vhosts_dir(agent_api)

        logger.debug('Detected vhosts_dir: %s', self._vhosts_dir)

    def _get_plesk_sites(self, webspace_id):
        # Plesk 8 & 9 don't support addon domains
        return []

    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([normalize_domain_name(x) for x in sites.iterkeys()]))
        subdomains_info = self._plesk_api.send(operator)

        return fix_plesk9_subdomains_ids(subdomains_info)

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        # Plesk 8 & 9 don't have subscriptions and webspaces. We use domains for this role.
        return get_plesk_domains_info(self._plesk_api, webspace_id)

    def _get_plesk_subdomain_content(self, subdomain_info):
        return self._get_plesk_site_content_items(self._get_plesk_site_webroot(subdomain_info))

    def _get_plesk_site_webroot(self, site_info):
        # Plesk 8 & 9 don't return 'www_root' hosting property. We need to make it ourselves.
        path_cls = ntpath if self._is_windows else os.path

        if isinstance(site_info, plesk_ops.SubdomainInfo):
            subdomain = site_info.name[:len(site_info.name)-len(site_info.parent)-1]
            site_webroot_components = [
                site_info.parent.encode('idna'), 'subdomains', subdomain.encode('idna'), 'httpdocs'
            ]
        else:
            site_webroot_components = [site_info.gen_info.ascii_name, 'httpdocs']

        return path_cls.join(self._vhosts_dir, *site_webroot_components)

    def _get_permissions(self, stat_info):
        return PermissionsModel(
            [
                OwnerModel('${SYSUSER}', stat_info.user_id)
            ]
        )
