# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import ntpath
import os
import urllib

from lxml import etree

from migrator.core.data.format.web_hosting_shm import ShmWebHostingReader
from migrator.core.data.model.web_hosting_shm import \
    SubscriptionModel, WebspaceModel, SystemUser, IpAddressModel, SiteModel, PhysicalHostingModel, \
    HostingPropertiesModel, ScriptingModel, PHPSettingsModel, DatabaseModel, DatabaseUserModel, WebuserModel, \
    ProtectedDirectoryModel, PropertyModel, DatabaseServerModel, FtpuserModel, ForwardingStdModel, AliasModel, \
    ProtectedDirectoryUserModel, NoHostingModel, SslCertificateModel, ApplicationModel
from migrator.core.data.source.db.platform import PlatformReaderDB
from migrator.core.language import _T
from migrator.core.utils import cached, generate_random_password
from migrator.core.utils.networking.dns import normalize_domain_name
from migrator.core.utils.operating_system.application_pool import get_application_pool_settings, get_application_pool
from migrator.core.utils.registry import Registry
from migrator.products.plesk.lib import get_plesk_domains_info
from migrator.products.plesk.lib.api.xmlrpc import operator as plesk_ops
from migrator.products.plesk.lib.decrypt_password import decrypt_account_passwords
from migrator.products.plesk.lib.export.utils import fix_plesk9_subdomains_ids, subdomain_prefix_for_plesk_9
from migrator.products.plesk.lib.export.web_base_shm import PleskHostingBaseExporter
from migrator.products.plesk.target.utils.plesk_utils import \
    query_plesk_db_one_value, query_plesk_db, get_unix_vhosts_dir, get_windows_vhosts_dir

logger = logging.getLogger(__name__)


class PleskHostingExporter(PleskHostingBaseExporter, ShmWebHostingReader):
    def __init__(self, server=None):
        super(PleskHostingExporter, self).__init__(server)

        self._win_application_pools = None
        self._win_sites = None

        if self._is_windows:
            self._win_application_pools, self._win_sites = get_application_pool_settings(self._server)

        self._applications_info = {}

    def get_subscription(self, subscription_id):
        """
        :param subscription_id: int
        :return: SubscriptionModel
        """
        return SubscriptionModel(
            external_id=subscription_id,
            webspaces=self.get_webspaces_list(subscription_id, subscription_external_id=str(subscription_id))
        )

    def get_webspaces_list(self, subscription_id, subscription_external_id=None):
        """
        :param subscription_id: int
        :param subscription_external_id: str
        :return: list[WebspaceModel]
        """
        return [self.get_webspace(subscription_id, subscription_id, subscription_external_id)]

    def get_webspace(self, subscription_id, webspace_id, subscription_external_id=None):
        """
        :param subscription_id: int
        :param webspace_id: int
        :param subscription_external_id: str
        :return: WebspaceModel
        """

        webspace_info = self._get_plesk_webspace_info(webspace_id)
        return WebspaceModel(
            name=webspace_info.gen_info.name,
            external_id=webspace_id,
            type='iis' if self._is_windows else 'apache',
            system_user=self._get_plesk_webspace_sysuser(webspace_info.hosting),
            ftp_users=self._get_ftp_users(webspace_id, webspace_info.gen_info.name, webspace_info.hosting),
            web_users=self._get_webusers(webspace_id),
            ip_addresses=self._get_plesk_webspace_ips(webspace_info.hosting),
            sites=self.get_sites_list(subscription_id, webspace_id),
            databases=self._get_databases(webspace_id),
            database_users=self._get_database_users(webspace_id),
            plesk_guid=webspace_info.gen_info.guid,
            host_external_id=PlatformReaderDB().get_subscription(
                subscription_external_id=subscription_external_id
            ).services[0].source_host_external_id,
            hosting_type=self._get_plesk_webspace_hosting_type(webspace_info.hosting)
        )

    def get_sites_list(self, subscription_id, webspace_id):
        """
        :param webspace_id: int
        :return: list[SiteModel]
        """
        sites = dict()

        # Make Plesk webspace to be a main site due to SHM requirements
        webspace_info = self._get_plesk_webspace_info(webspace_id)
        sites[webspace_info.gen_info.name] = SiteModel(
            name=webspace_info.gen_info.name,
            external_id=webspace_id,
            status='active',
            is_main=True,
            hosting=self._get_hosting(webspace_id, webspace_id, webspace_info, webspace_info.gen_info.name, webspace_info.hosting),
            plesk_guid=webspace_info.gen_info.guid,
        )

        # Get all webspace sites
        sites_info = self._get_plesk_sites(webspace_id)
        for result in sites_info:
            site_id = result.data[0]
            site_info = result.data[1]

            sites[site_info.gen_info.name] = SiteModel(
                name=site_info.gen_info.name,
                external_id=site_id,
                status='active',
                is_main=False,
                hosting=self._get_hosting(webspace_id, site_id, site_info, site_info.gen_info.name, site_info.hosting),
                plesk_guid=site_info.gen_info.guid,
            )

        alias_info = self._get_alias(sites)
        for result in alias_info:
            site_id = result.data[0]
            site_info = result.data[1]
            sites[site_info.name] = SiteModel(
                name=site_info.name,
                external_id='alias_%s' % site_id,  # to avoid the same external_id with real sites
                status='active',
                is_main=False,
                hosting=self._get_hosting(webspace_id, site_id, site_info, site_info.name, site_info),
                plesk_guid=None,
            )

        # Get all subdomains as sites
        subdomains_info = self._get_plesk_subdomains(sites)
        for result in subdomains_info:
            subdomain_id = result.data[0]
            subdomain_info = result.data[1]
            parent_site = sites[subdomain_info.parent]

            sites[subdomain_info.name] = SiteModel(
                name=subdomain_info.name,
                external_id=subdomain_id,
                status=parent_site.status,
                is_main=False,
                hosting=self._get_hosting(
                    webspace_id, subdomain_id, subdomain_info, subdomain_info.name, subdomain_info.hosting,
                    sites[subdomain_info.parent].external_id
                ),
                plesk_guid=None,
            )

        return sites.values()

    def _get_alias(self, sites):
        return self._plesk_api.send(
            plesk_ops.AliasOperator.Get(filter=plesk_ops.AliasOperator.FilterBySiteName([x for x in sites.iterkeys()]))
        )

    def _get_hosting(self, webspace_id, site_id, site_info, site_name, hosting, parent_site_id=None):
        """
        :param site_id: int
        :return: PhysicalHostingModel
        """

        if self._is_physical_hosting(hosting):
            hosting_properties = hosting.properties

            www_root = self._get_plesk_site_webroot(site_info)
            # Workaround for Plesk 10.4: if you create subdomain and convert it to a domain forwarder, then Plesk API
            # return incorrect hosting type.
            if www_root == '':
                return None

            return PhysicalHostingModel(
                self._get_hosting_properties(hosting_properties),
                self._get_scripting(hosting_properties),
                self._get_php_settings(hosting_properties),
                self._get_ssl_certificate(site_id, site_name),
                self._get_protected_directories(site_id),
                self._get_virtual_directories(webspace_id, www_root, site_name, True),
                self._get_application_pool(site_name),
                self._get_applications(webspace_id, site_id, www_root),
            )
        elif self._is_std_forwarding(hosting):
            return ForwardingStdModel(hosting.dest_url, hosting.http_code)
        elif self._is_frame_forwarding(hosting):
            return ForwardingStdModel(hosting.dest_url, None)
        elif self._is_alias(hosting):
            seo_redirect = False if hosting.seo_redirect is None else hosting.seo_redirect
            return AliasModel(hosting.site_id, seo_redirect)
        elif self._is_nohosting(hosting):
            return NoHostingModel()
        else:
            return None

    def _get_hosting_properties(self, hosting_properties):
        """
        :param hosting_properties: dict
        :return: HostingPropertiesModel
        """
        return HostingPropertiesModel(
            hosting_properties.get('iis_app_pool'),
            hosting_properties.get('ftp_quota'),
            hosting_properties.get('errdocs'),
            hosting_properties.get('wuscripts'),
            hosting_properties.get('at_domains'),
            hosting_properties.get('seo-redirect'),
            hosting_properties.get('webstat'),
            self._fix_property_value(hosting_properties.get('webstat_protected')),
            hosting_properties.get('ssl'),
            self._fix_property_value(hosting_properties.get('write_modify')),
        )

    def _get_scripting(self, hosting_properties):
        """
        :param hosting_properties: dict
        :return: ScriptingModel
        """
        return ScriptingModel(
            hosting_properties.get('shell'),
            hosting_properties.get('php'),
            self._get_php_handler_id(hosting_properties),
            hosting_properties.get('cgi'),
            hosting_properties.get('perl'),
            hosting_properties.get('python'),
            hosting_properties.get('asp'),
            hosting_properties.get('asp_dot_net'),
            hosting_properties.get('managed_runtime_version'),
            hosting_properties.get('fastcgi'),
            hosting_properties.get('cgi_mode'),
        )

    def _get_php_handler_id(self, hosting_properties):
        php_handler_id = hosting_properties.get('php_handler_id')
        # if php_handler_id like 'plesk-php54-fastcgi' or fastcgi-5.4' return php handler
        if php_handler_id is not None and len(php_handler_id.replace('_', '-').split('-')) > 1:
            return php_handler_id

        if php_handler_id is not None:
            handler_type = php_handler_id
        else:
            handler_type = hosting_properties.get('php_handler_type')

        if handler_type == 'module' or handler_type is None:
            handler_type = 'fpm'

        try:
            tmp = self._server.plesk_components['php'].version.split('-')[0].split('.')
            return '%s-%s.%s' % (handler_type, tmp[0], tmp[1])
        except (KeyError, AttributeError, IndexError):
            return None

    def _get_php_settings(self, hosting_properties):
        """
        :param hosting_properties: dict
        :return: PHPSettingsModel
        """
        php_settings = {}
        property_name_mapping = {
            'mail_force_extra_parameters': 'mail.force_extra_parameters',
        }
        for name in [
            'memory_limit', 'max_execution_time', 'max_input_time', 'post_max_size', 'upload_max_filesize', 'safe_mode',
            'safe_mode_include_dir', 'safe_mode_exec_dir', 'include_path', 'mail_force_extra_parameters',
            'register_globals', 'error_reporting', 'display_errors', 'log_errors', 'allow_url_fopen', 'file_uploads',
            'short_open_tag', 'magic_quotes_gpc'
        ]:
            php_settings[name] = hosting_properties.get(
                property_name_mapping[name] if name in property_name_mapping else name
            )

        return PHPSettingsModel(**php_settings)

    def _get_application_pool(self, site_name):
        return get_application_pool(site_name, self._win_application_pools, self._win_sites, self._server)

    def _get_db_users_query(self):
        return """
            select u.login, u.db_id, u.id, a.type, a.password from accounts a, db_users u
            where a.id=u.account_id and u.login in ({logins})
        """

    def _get_database_users(self, webspace_id):
        """
        :param webspace_id: int
        :return: list[DatabaseModel]
        """
        databases = self._get_databases_info(webspace_id)

        db_users = {}
        for database in databases:
            db_users[database.data.db_server_id] = set()

        for database in databases:
            for user in self._plesk_api.send(
                    plesk_ops.DatabaseOperator.GetDbUsers(
                            filter=plesk_ops.DatabaseOperator.GetDbUsers.FilterByDbIds([database.data.id])
                    )
            ):
                db_users[database.data.db_server_id].add(user.data)

        logins = ','.join(["'%s'" % el.login for lst in db_users.values() for el in lst])
        passwords = {}
        if logins:
            query = self._get_db_users_query()
            for result in query_plesk_db(self._server, query.format(logins=logins), values_count=5, log_output=False):
                passwords[result[0], int(result[1]) or None] = (result[2], result[4], result[3])

        plain_passwords = decrypt_account_passwords(self._server, passwords) if passwords else {}

        database_user_models = []
        for db_server_id in db_users:
            for db_user in db_users[db_server_id]:
                database_user_models.append(
                    DatabaseUserModel(
                        login=db_user.login,
                        password=plain_passwords[db_user.login, db_user.db_id],
                        database_external_id=str(db_user.db_id),
                        database_server_id=str(db_server_id)
                    )
                )
        return database_user_models

    def _get_databases(self, webspace_id):
        """
        :param webspace_id: int
        :return: list[DatabaseModel]
        """
        databases = self._get_databases_info(webspace_id)

        database_models = []
        for database in databases:
            database_models.append(
                DatabaseModel(
                    external_id=str(database.data.id),
                    database_server_id=str(database.data.db_server_id),
                    name=database.data.name,
                )
            )

        return database_models

    def _get_databases_info(self, webspace_id):
        return self._plesk_api.send(
            plesk_ops.DatabaseOperator.GetDb(
                filter=plesk_ops.DatabaseOperator.GetDb.FilterByWebspaceId([webspace_id])
            )
        )

    def get_database_server_password(self, dbhost, dbtype, dbport):
        return self._server.get_database_server_password(dbtype, dbhost, dbport)

    def get_database_servers(self):
        """
        :return: list[DatabaseServerModel]
        """
        database_servers = []
        for database_server in self._plesk_api.send(
            plesk_ops.DbServerOperator.Get(
                filter=plesk_ops.DbServerOperator.FilterAll()
            )
        ):
            data = database_server.data
            if data.db_num:
                password = data.password or self.get_database_server_password(data.host, data.dbtype, data.port) or ''
                database_servers.append(
                    DatabaseServerModel(
                        id=str(data.id),
                        host=data.host,
                        port=data.port,
                        dbtype=data.dbtype,
                        login=data.admin,
                        password=password,
                        default=data.default,
                        local=data.local
                    )
                )

        return database_servers

    def _get_ssl_certificate_info(self, site_name, ssl_certifiacte_id):
        """
        :return: SslCertificateModel
        """
        if ssl_certifiacte_id == '' or ssl_certifiacte_id == '0':
            return None

        return SslCertificateModel(
            name=_T('SSL_CERTIFICATE_FOR_SITE').format(site_name=site_name),
            csr=urllib.unquote_plus(query_plesk_db_one_value(
                self._server, 'SELECT csr FROM certificates WHERE id = {id}'.format(id=ssl_certifiacte_id),
                log_output=False
            ).strip()),
            private_key=urllib.unquote_plus(query_plesk_db_one_value(
                self._server, 'SELECT pvt_key FROM certificates WHERE id = {id}'.format(id=ssl_certifiacte_id),
                log_output=False
            ).strip()),
            cert=urllib.unquote_plus(query_plesk_db_one_value(
                self._server, 'SELECT cert FROM certificates WHERE id = {id}'.format(id=ssl_certifiacte_id),
                log_output=False
            ).strip()),
            ca=urllib.unquote_plus(query_plesk_db_one_value(
                self._server, 'SELECT ca_cert FROM certificates WHERE id = {id}'.format(id=ssl_certifiacte_id),
                log_output=False
            ).strip()),
        )

    def _get_ftp_users(self, webspace_id, webspace_name, hosting_info):
        """
        :param webspace_id: int
        :return: list[FtpuserModel]
        """
        ftp_users = []
        for ftp_user in self._plesk_api.send(
            plesk_ops.FtpuserOperator.Get(
                filter=plesk_ops.FtpuserOperator.FilterByWebspaceId([webspace_id])
            )
        ):
            home = ftp_user.data.home
            www_root = hosting_info.properties.get('www_root').rstrip('/\\')
            vhost_dir = self._server.get_vhost_dir(webspace_name)
            if www_root.startswith(vhost_dir):
                www_root = www_root[len(vhost_dir):]
                if home.startswith(www_root):
                    home = u'/%s%s' % (webspace_name, home[len(www_root):])

            ftp_users.append(
                FtpuserModel(
                    name=ftp_user.data.name,
                    password=generate_random_password(),
                    home=home,
                    read=ftp_user.data.read,
                    write=ftp_user.data.write,
                    quota=ftp_user.data.quota
                )
            )

        return ftp_users

    def _get_webusers(self, webspace_id):
        """
        :param webspace_id: int
        :return: list[WebuserModel]
        """

        webusers = []
        for webuser in self._get_webusers_info(webspace_id):
            webusers.append(
                WebuserModel(
                    login=webuser.data.login,
                    password=generate_random_password(),
                    ftp_quota=webuser.data.ftp_quota,
                    asp=webuser.data.services.get('fastcgi'),
                    asp_dot_net=webuser.data.services.get('asp_dot_net'),
                    php=webuser.data.services.get('php'),
                    cgi=webuser.data.services.get('cgi'),
                    perl=webuser.data.services.get('perl'),
                    python=webuser.data.services.get('python'),
                    fastcgi=webuser.data.services.get('fastcgi'),
                    write_modify=webuser.data.services.get('write_modify'),
                )
            )

        return webusers

    def _get_webusers_info(self, site_id):
        return self._plesk_api.send(
            plesk_ops.WebuserOperator.Get(filter=plesk_ops.WebuserOperator.FilterBySiteId([site_id]))
        )

    def _get_protected_directories(self, site_id):
        """
        :param site_id: int
        :return: list[ProtectedDirectoryModel]
        """
        protected_directories = []
        for pdir in self._get_protected_directories_info(site_id):
            properties = [PropertyModel(x.name, x.value) for x in pdir.data.properties]
            pd_users = []
            for pd_user in self._get_protected_directory_users_info(pdir.data.id):
                pd_users.append(
                    ProtectedDirectoryUserModel(
                        login=pd_user.data.login,
                        password=generate_random_password(),
                    )
                )

            protected_directories.append(
                ProtectedDirectoryModel(
                    name=pdir.data.name,
                    header=pdir.data.header,
                    pd_users=pd_users,
                    properties=properties,
                )
            )

        return protected_directories

    def _get_ssl_certificate(self, site_id, site_name):
        if site_id is None:
            return None

        query = "SELECT certificate_id FROM hosting "\
                "JOIN domains ON domains.id = hosting.dom_id "\
                "JOIN Repository ON domains.cert_rep_id = Repository.rep_id "\
                "WHERE Repository.component_id = hosting.certificate_id "\
                "AND domains.id = {domain_id}".format(domain_id=site_id)
        certificate_id = query_plesk_db_one_value(self._server, query).strip()
        return self._get_ssl_certificate_info(site_name, certificate_id)

    def _get_applications(self, webspace_id, site_id, www_root):
        applications = []
        applications_info_xml = self._get_applications_info(webspace_id)
        if applications_info_xml is None:
            return applications

        applications_info = etree.fromstring(applications_info_xml.encode('utf-8'))
        for application in applications_info.findall('sapp-installed'):
            id = application.find('vhost/id').text
            if str(site_id) != id:
                continue

            name = application.find('sapp-spec/sapp-name').text
            database_node = application.find('database')
            if database_node is not None:
                database_external_id = database_node.attrib['id']
            else:
                database_external_id = None

            sapp_prefix = application.find('sapp-installdir/sapp-prefix').text

            applications.append(ApplicationModel(
                name=name,
                source_path=self._pathlib.join(www_root, sapp_prefix or ''),
                target_path=self._pathlib.join('${SITEROOT}', sapp_prefix or ''),
                database_external_id=database_external_id
            ))

        return applications

    def _get_applications_info(self, webspace_id):
        if webspace_id not in self._applications_info:
            global_context = Registry.get_instance().get_context()
            local_script_path = global_context.session_files.get_path_to_aps_backup_tool()
            if not os.path.exists(local_script_path):
                return None

            remote_script_path = self._server.get_session_file_path("aps_backup.php")
            webspaces_ids_path = self._server.get_session_file_path("webspaces_ids.txt")
            with self._server.runner() as runner:
                runner.upload_file(local_script_path, remote_script_path)
                runner.upload_file_content(webspaces_ids_path, str(webspace_id))
                if self._is_windows:
                    applications = runner.sh(
                        ur'cmd.exe /C "{plesk_dir}\admin\bin\php.exe -dauto_prepend_file="" {aps_backup} --get-instances -webspaces-ids-from {webspaces_ids_path}',
                        dict(
                            plesk_dir=self._server.plesk_dir, aps_backup=remote_script_path,
                            webspaces_ids_path=webspaces_ids_path
                        ),
                        log_output=False
                    )
                else:
                    applications = runner.sh(
                        ur'{plesk_dir}/admin/bin/php -dauto_prepend_file="" {aps_backup} --get-instances -webspaces-ids-from {webspaces_ids_path}',
                        dict(
                            plesk_dir=self._server.plesk_dir, aps_backup=remote_script_path,
                            webspaces_ids_path=webspaces_ids_path
                        ),
                        log_output=False
                    )

                runner.remove_file(remote_script_path)
                runner.remove_file(webspaces_ids_path)

            self._applications_info[webspace_id] = applications

        return self._applications_info[webspace_id]

    def _get_protected_directories_info(self, site_id):
        """
        :param site_id: int
        :return: list[ProtectedDirectoryModel]
        """
        return self._plesk_api.send(
            plesk_ops.ProtectedDirectoryOperator.Get(
                filter=plesk_ops.ProtectedDirectoryOperator.FilterBySiteId([site_id])
            )
        )

    def _get_protected_directory_users_info(self, pd_id):
        """
        :param pd_id: int
        :return: list[ProtectedDirectoryModel]
        """
        return self._plesk_api.send(
            plesk_ops.ProtectedDirectoryOperator.GetUser(
                filter=plesk_ops.ProtectedDirectoryOperator.FilterByPDId([pd_id])
            )
        )

    @staticmethod
    def _get_plesk_webspace_sysuser(hosting):
        if isinstance(hosting, plesk_ops.SubscriptionHostingVirtual):
            return SystemUser(hosting.properties.get('ftp_login'), hosting.properties.get('ftp_password'))
        else:
            return None

    def _get_plesk_webspace_ips(self, hosting):
        result = []
        if isinstance(hosting, plesk_ops.SubscriptionHostingNone) or isinstance(hosting, plesk_ops.SiteHostingNone):
            return result

        plesk_ips = self._get_plesk_ips()
        webspace_ips = hosting.ip_addresses

        for ip in webspace_ips.v4, webspace_ips.v6:
            if ip is not None:
                result.append(plesk_ips[ip])

        return result

    @staticmethod
    def _get_plesk_webspace_hosting_type(hosting):
        if isinstance(hosting, plesk_ops.SubscriptionHostingVirtual):
            return 'virtual_hosting'
        elif isinstance(hosting, plesk_ops.SubscriptionHostingStandardForwarding):
            return 'standard_forwarding'
        elif isinstance(hosting, plesk_ops.SubscriptionHostingFrameForwarding):
            return 'frame_forwarding'
        elif isinstance(hosting, plesk_ops.SubscriptionHostingNone):
            return 'none'

    @cached
    def _get_plesk_ips(self):
        ips = self._plesk_api.send(plesk_ops.IpOperator.Get()).data
        ip_addresses = dict()

        for ip in ips:
            ip_addresses[ip.ip_address] = IpAddressModel(ip.ip_type.upper(), ip.ip_address, ip.interface)

        return ip_addresses

    def _fix_property_value(self, value):
        if value is None or value == '':
            return None
        else:
            return value


class Plesk178HostingExporter(PleskHostingExporter):
    def _get_webusers_info(self, site_id):
        return []


class Plesk115HostingExporter(PleskHostingExporter):

    def _get_app_resource_map(self):
        query_apsc_app_resource_map = """
            select aps_registry_object_application.uid as app_uid, aps_registry_object_resource.uid as resource_uid
            from
                apsc.aps_registry_object as aps_registry_object_application
                JOIN apsc.aps_application
                    on apsc.aps_application.registry_object_id = aps_registry_object_application.id
                JOIN apsc.aps_resource as aps_dependent_resource
                    on aps_dependent_resource.app_id = apsc.aps_application.id
                JOIN apsc.aps_resource_adjacency_list
                    on apsc.aps_resource_adjacency_list.dependent_resource_id = aps_dependent_resource.id
                JOIN apsc.aps_resource as aps_base_resource
                    on aps_base_resource.id = apsc.aps_resource_adjacency_list.base_resource_id
                JOIN apsc.aps_registry_object as aps_registry_object_resource
                    on aps_registry_object_resource.id = aps_base_resource.registry_object_id;
        """
        apsc_app_resource_map = {}
        for (app_uid, resource_uid) in query_plesk_db(self._server, query_apsc_app_resource_map):
            apsc_app_resource_map[resource_uid] = app_uid

        return apsc_app_resource_map

    def _get_applications(self, webspace_id, site_id, www_root):
        apsc_app_resource_map = self._get_app_resource_map()

        query_resource_database_map = """
            select apsResources.registryId, data_bases.id
            from apsResources, data_bases
            where apsResources.pleskId = data_bases.id and apsResources.pleskType='db';
        """
        resource_database_map = {}
        for (resource_uid, database_external_id) in query_plesk_db(self._server, query_resource_database_map):
            if resource_uid in apsc_app_resource_map:
                resource_database_map[apsc_app_resource_map[resource_uid]] = database_external_id
            else:
                logger.info(
                    "Resource '%s' (database) absent at APS Controller, looks like inconsistency" % resource_uid
                )

        query_resource_prefix_map = """
            select apsResources.registryId, domains.name as domain, value as url_prefix
            from apsResources, domains, apsResourcesParameters
            where apsResources.pleskId = domains.id and apsResources.pleskType='hosting' and
                  apsResourcesParameters.apsResourceId = apsResources.id and apsResourcesParameters.name='urlPrefix';
        """
        resource_prefix_map = {}
        for (resource_uid, domain, url_prefix) in query_plesk_db(self._server, query_resource_prefix_map):
            if resource_uid in apsc_app_resource_map:
                resource_prefix_map[apsc_app_resource_map[resource_uid]] = url_prefix
            else:
                logger.info(
                    "Resource '%s' (urlPrefix) absent at APS Controller, looks like inconsistency" % resource_uid
                )

        query = """
            select smb_apsMetas.name, apsContextsApplications.registryApplicationId
            FROM
                smb_apsMetas
                JOIN smb_apsInstances on smb_apsInstances.metaId = smb_apsMetas.id
                JOIN apsContextsApplications on apsContextsApplications.id = smb_apsInstances.id
                JOIN apsContexts ON apsContexts.id = apsContextsApplications.apsContextId
                WHERE pleskId = {domain_id}
        """

        applications = []
        for (application_name, registryApplicationId) in query_plesk_db(self._server, query.format(domain_id=site_id)):
            applications.append(ApplicationModel(
                name=application_name,
                source_path=self._pathlib.join(www_root, resource_prefix_map.get(registryApplicationId) or ''),
                target_path=self._pathlib.join('${SITEROOT}', resource_prefix_map.get(registryApplicationId) or ''),
                database_external_id=resource_database_map.get(registryApplicationId)
            ))

        return applications


class Plesk10HostingExporter(Plesk115HostingExporter):
    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([x for x in sites.iterkeys()])
        )
        return self._plesk_api.send(operator)

    def _get_ssl_certificate(self, site_id, site_name):
        if not self._is_windows:
            return super(Plesk10HostingExporter, self)._get_ssl_certificate(site_id, site_name)

        if site_id is None:
            return None

        query = "SELECT ssl_certificate_id FROM IP_Addresses "\
                "JOIN dom_param on dom_param.val = IP_Addresses.id "\
                "WHERE dom_param.param = 'ip_addr_id'"\
                "AND dom_param.dom_id = {domain_id}".format(domain_id=site_id)
        certificate_id = query_plesk_db_one_value(self._server, query).strip()
        return self._get_ssl_certificate_info(site_name, certificate_id)

    def _get_app_resource_map(self):
        query_apsc_app_resource_map = """
            select aps_registry_object_application.uid as app_uid, aps_resource_requirement.uid as resource_uid
            from
                apsc.aps_registry_object as aps_registry_object_application
                JOIN apsc.aps_application
                    on apsc.aps_application.registry_object_id = aps_registry_object_application.id
                LEFT OUTER JOIN apsc.aps_application_resource
                    on aps_application_resource.app_id = aps_application.id
                LEFT OUTER JOIN apsc.aps_resource_requirement
                    on aps_resource_requirement.resource_id = apsc.aps_application_resource.resource_id
        """
        apsc_app_resource_map = {}
        for (app_uid, resource_uid) in query_plesk_db(self._server, query_apsc_app_resource_map):
            apsc_app_resource_map[resource_uid] = app_uid

        return apsc_app_resource_map

    def _get_db_users_query(self):
        if self._is_windows:
            return """
                select login, db_id, id, 'plain', passwd from db_users where login in ({logins})
            """
        else:
            return super(Plesk10HostingExporter, self)._get_db_users_query()


class Plesk9HostingExporter(Plesk10HostingExporter):
    def __init__(self, server=None):
        super(Plesk9HostingExporter, self).__init__(server)

        agent_api = self._server.agent_api()
        if self._is_windows:
            self._vhosts_dir = get_windows_vhosts_dir(agent_api)
        else:
            self._vhosts_dir = get_unix_vhosts_dir(agent_api)

        logger.debug('Detected vhosts_dir: %s', self._vhosts_dir)

    def _get_plesk_sites(self, webspace_id):
        # Plesk 8 & 9 don't support addon domains
        return []

    def _get_plesk_subdomains(self, sites):
        operator = plesk_ops.SubdomainOperator.Get(
            filter=plesk_ops.SubdomainOperator.FilterByParentName([normalize_domain_name(x) for x in sites.iterkeys()]))
        subdomains_info = self._plesk_api.send(operator)

        return fix_plesk9_subdomains_ids(subdomains_info)

    def _get_alias(self, sites):
        return self._plesk_api.send(
            plesk_ops.DomainAliasOperator.Get(
                filter=plesk_ops.DomainAliasOperator.FilterByDomainName([x for x in sites.iterkeys()])
            )
        )

    def _get_hosting(self, webspace_id, site_id, site_info, site_name, hosting, parent_site_id=None):
        if isinstance(hosting, plesk_ops.SubdomainHostingVirtual):
            # Plesk 8 & 9 don't support protected directories for subdomains.
            # Return PhysicalHostingModel w/o protected_directories.
            hosting_properties = hosting.properties
            www_root = self._get_plesk_site_webroot(site_info)
            return PhysicalHostingModel(
                self._get_hosting_properties(hosting_properties),
                self._get_scripting(hosting_properties),
                self._get_php_settings(hosting_properties),
                self._get_ssl_certificate(parent_site_id, site_name),
                applications=self._get_applications(webspace_id, site_id, www_root),
            )
        elif isinstance(hosting, plesk_ops.DomainHostingNone):
            return NoHostingModel()
        else:
            return super(Plesk9HostingExporter, self)._get_hosting(
                webspace_id, site_id, site_info, site_name, hosting, parent_site_id
            )

    @cached
    def _get_plesk_webspace_info(self, webspace_id):
        # Plesk 8 & 9 don't have subscriptions and webspaces. We use domains for this role.
        return get_plesk_domains_info(self._plesk_api, webspace_id)

    @staticmethod
    def _get_plesk_webspace_sysuser(hosting):
        if isinstance(hosting, plesk_ops.DomainHostingVirtual):
            return SystemUser(hosting.properties.get('ftp_login'), hosting.properties.get('ftp_password'))
        else:
            return None

    def _get_ftp_users(self, webspace_id, webspace_name, hosting_info):
        return []

    def _get_databases_info(self, webspace_id):
        return self._plesk_api.send(
            plesk_ops.DatabaseOperator.GetDb(
                filter=plesk_ops.DatabaseOperator.GetDb.FilterByDomainId([webspace_id])
            )
        )

    def _get_webusers_info(self, site_id):
        return self._plesk_api.send(
            plesk_ops.WebuserOperator.Get(
                filter=plesk_ops.WebuserOperator.FilterByDomainId([site_id])
            )
        )

    def _get_protected_directories_info(self, site_id):
        """
        :param site_id: int
        :return: list[ProtectedDirectoryModel]
        """
        return self._plesk_api.send(
            plesk_ops.ProtectedDirectoryOperator.Get(
                filter=plesk_ops.ProtectedDirectoryOperator.FilterByDomainId([site_id])
            )
        )

    def _get_hosting_properties(self, hosting_properties):
        model = super(Plesk9HostingExporter, self)._get_hosting_properties(hosting_properties)
        # PleskWin 8 & 9 may return '' as a value for webstat_protected.
        # We need to fix its value to avoid xsd validation error during import
        if isinstance(model.webstat_protected, basestring) and not model.webstat_protected:
            model.webstat_protected = False

        return model

    def _get_plesk_webspace_ips(self, hosting):
        if isinstance(hosting, plesk_ops.DomainHostingNone):
            return []

        return super(Plesk9HostingExporter, self)._get_plesk_webspace_ips(hosting)

    def _get_ssl_certificate(self, site_id, site_name):
        if site_id is None:
            return None

        query = "SELECT ssl_certificate_id FROM (((IP_Addresses "\
                "INNER JOIN hosting on hosting.ip_address_id = IP_Addresses.id) "\
                "INNER JOIN domains ON domains.id = hosting.dom_id) "\
                "INNER JOIN Repository ON domains.cert_rep_id = Repository.rep_id) "\
                "WHERE Repository.component_id = IP_Addresses.ssl_certificate_id "\
                "AND hosting.dom_id = {domain_id}".format(domain_id=site_id)
        certificate_id = query_plesk_db_one_value(self._server, query).strip()
        return self._get_ssl_certificate_info(site_name, certificate_id)

    def _get_applications(self, webspace_id, site_id, www_root):
        applications = []

        query = """
        SELECT SiteAppPackages.name, SiteApps.install_prefix, SiteAppResources.res_id as database_external_id
        FROM (((SiteApps
            INNER JOIN APSClientApplicationItems on APSClientApplicationItems.id=SiteApps.capp_item_id)
            INNER JOIN SiteAppPackages on SiteAppPackages.id=APSClientApplicationItems.app_item_id)
            LEFT OUTER JOIN SiteAppResources on (SiteAppResources.app_id=SiteApps.id and SiteAppResources.type='database'))
            WHERE dom_id = {domain_id} and dom_type = '{dom_type}'
        """

        if str(site_id).startswith(subdomain_prefix_for_plesk_9):
            domain_id = int(site_id[len(subdomain_prefix_for_plesk_9):])
            dom_type = 'subdomain'
        else:
            domain_id = site_id
            dom_type = 'domain'

        for (application_name, url_prefix, database_external_id) in query_plesk_db(
                self._server, query.format(domain_id=domain_id, dom_type=dom_type)
        ):
            applications.append(ApplicationModel(
                name=application_name,
                source_path=self._pathlib.join(www_root, url_prefix or ''),
                target_path=self._pathlib.join('${SITEROOT}', url_prefix or ''),
                database_external_id=database_external_id
            ))

        return applications

    def get_database_server_password(self, dbhost, dbtype, dbport):
        if not self._is_windows and dbhost == 'localhost' and dbtype == 'mysql':
            with self._server.runner() as runner:
                return runner.get_file_contents(u'/etc/psa/.psa.shadow').rstrip('\n')
        else:
            query = "select admin_password from DatabaseServers where type='{dbtype}' and host='{dbhost}' and port='{dbport}'"
            return query_plesk_db_one_value(
                self._server, query.format(dbtype=dbtype, dbhost=dbhost, dbport=dbport), log_output=False
            ).strip()

    def _get_plesk_site_webroot(self, site_info):
        # Plesk 8 & 9 don't return 'www_root' hosting property. We need to make it ourselves.
        path_cls = ntpath if self._is_windows else os.path

        if isinstance(site_info, plesk_ops.SubdomainInfo):
            subdomain = site_info.name[:len(site_info.name)-len(site_info.parent)-1]
            site_webroot_components = [
                site_info.parent.encode('idna'), 'subdomains', subdomain.encode('idna'), 'httpdocs'
            ]
        else:
            site_webroot_components = [site_info.gen_info.ascii_name, 'httpdocs']

        return path_cls.join(self._vhosts_dir, *site_webroot_components)

