# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import posixpath
import ntpath

from collections import OrderedDict
from datetime import datetime
from urllib2 import URLError
from lxml import etree
from lxml.etree import ParseError

import migrator.core.product.panel
import migrator.products.plesk.host

from migrator.core import MigrationError
from migrator.core.language import _T
from migrator.core.session.config import is_windows_host, SSHAuthPassword
from migrator.core.product.lib.jsonrpc.client import JsonRpcClient
from migrator.core.product.lib.host import get_service_status_linux, get_service_status_windows
from migrator.core.product.panel import PanelInfoBase, ProductBase
from migrator.core.utils import cached
from migrator.core.utils.filetypes.config_utils import ConfigSection, read_auth
from migrator.core.utils.filetypes.ssl_keys import get_oam_ssl_certificate
from migrator.core.utils.loggers.log import trace
from migrator.core.utils.networking.ip_port import is_network_port_open
from migrator.core.utils.networking.http_xml_client import HttpClientError
from migrator.core.utils.networking.run_command import SSHRunner, WinexeRunner, WinexeRunnerException
from migrator.core.utils.operating_system import windows_utils
from migrator.core.utils.registry import Registry
from migrator.products.plesk.export import get_plesk_exporter_class, get_plesk_content_exporter_class
from migrator.products.plesk.lib.agent import install_agent_on_server
from migrator.products.plesk.lib.api import xmlrpc
from migrator.products.plesk.lib.dns_zone_manager import PleskDnsZoneManager
from migrator.products.plesk.lib.panel import split_version_string
from migrator.products.plesk.source.plesk.server import PleskSourceServer
from migrator.products.plesk.target.config import SourcePleskUnixConfig, SourcePleskWindowsConfig
from migrator.products.plesk.target.utils.plesk_utils import get_unix_product_root_dir
from migrator.settings import \
    SOURCE_PRODUCT_PLESK, AGENT_PORT, WINDOWS_DEFAULT_SESSION_DIR, UNIX_DEFAULT_SESSION_DIR, SSH_DEFAULT_PORT, \
    PLESK_PANEL_DEFAULT_PORT

logger = logging.getLogger(__name__)


class Product(ProductBase):
    @classmethod
    def get_agent_instance(cls):
        agent_installer = migrator.products.plesk.host.AgentInstaller()
        return agent_installer.install()

    @classmethod
    def get_exporter_instance(cls, service):
        return get_plesk_exporter_class(service)()

    @classmethod
    def get_content_exporter_instance(cls, service):
        return get_plesk_content_exporter_class(service)()

    @classmethod
    def get_dns_zone_manager(cls):
        return PleskDnsZoneManager()

    @classmethod
    def suspend_subscription(cls, subscription):
        logger.debug("Not disabling Plesk subscription ID '%s'" % subscription.external_id)

    @classmethod
    def get_id(cls):
        return SOURCE_PRODUCT_PLESK


class PanelInfo(PanelInfoBase):
    """PanelInfo object that is common for Windows and Linux installations of Plesk Panel"""
    def __init__(self, runner, config_section=None):
        super(PanelInfo, self).__init__(runner, config_section)
        global_context = Registry.get_instance().get_context()
        settings = self.read_source_plesk_settings(global_context.config, 'generic')
        self.source_server = PleskSourceServer(self.ip(), settings, global_context.migrator_server)

    def read_source_plesk_settings(self, config, section_name):
        """Read Plesk API and SSH connection settings.
        """
        section = ConfigSection(config, section_name)

        ip = section['ip']
        is_windows = is_windows_host(section)
        session_dir = section.get(
            'session-dir',
            WINDOWS_DEFAULT_SESSION_DIR if is_windows else UNIX_DEFAULT_SESSION_DIR
        )

        auth = read_auth(section)
        if is_windows:
            windows_auth = auth
            return SourcePleskWindowsConfig(section_name, ip, windows_auth, None, session_dir, None, None)
        else:
            ssh_auth = SSHAuthPassword(
                port=int(section.get('ssh-port', SSH_DEFAULT_PORT)),
                username=auth.username,
                password=auth.password
            )
            return SourcePleskUnixConfig(section_name, ip, ssh_auth, None, session_dir, None)

    @property
    def name(self):
        return 'Plesk'

    @property
    def short_name(self):
        return 'plesk'

    @staticmethod
    def is_installed(runner):
        for info_class in PleskUnixInfo, PleskWindowsInfo:
            if info_class.is_installed(runner):
                return True
        return False

    @classmethod
    @trace
    def factory(cls, runner, config_section):
        """Create an OS-specific instance of PanelInfo subclass"""
        if isinstance(runner, SSHRunner):
            return PleskUnixInfo(runner, config_section)
        elif isinstance(runner, WinexeRunner):
            return PleskWindowsInfo(runner, config_section)
        else:
            MigrationError(_T("UNABLE_TO_CONSTRUCT_PANEL_INFO_OBJECT").format(runner=runner))

    @trace
    def get_service_list(self):
        return OrderedDict([
            ('Plesk XML API', dict(func=self._probe_xmlrpc)),
            ('Plesk license', dict(func=self._check_plesk_license)),
        ])

    @trace
    def _check_plesk_license(self, service_status):
        """Get license info via XML RPC and check that it valid and active"""
        service = ''

        if service_status['Plesk XML API']['exit_code'] != 0:
            return service, _T("SERVICE_STATUS_UNAVAILABLE"), 1, _T("UNABLE_TO_CHECK_THE_LICENSE")

        xmlrpc_client = self.source_server.plesk_api()
        license_resource = self.get_plesk_license_info(xmlrpc_client)
        license_is_valid = self.is_plesk_license_valid(license_resource)

        if license_is_valid:
            status = _T("SERVICE_STATUS_VALID")
            error_message = None
            exit_code = 0
        else:
            status = _T("SERVICE_STATUS_INVALID")
            error_message = _T("THE_LICENSE_KEY_HAS_EXPIRED")
            exit_code = 1

        return service, status, exit_code, error_message

    @trace
    def _probe_xmlrpc(self, service_status):
        """Make a sample XML RPC request and verify that Plesk XML API is accessible"""
        service = ''
        error_message = None
        exit_code = 1
        status = _T("SERVICE_STATUS_UNAVAILABLE")

        plesk_api_url = 'https://{ip}:{port}/enterprise/control/agent.php'.format(ip=self.ip(), port=PLESK_PANEL_DEFAULT_PORT)
        if not is_network_port_open(self.ip(), PLESK_PANEL_DEFAULT_PORT):
            return service, status, exit_code, _T("CHECK_THE_FIREWALL_SETTINGS").format(
                plesk_api_url=plesk_api_url,
                host_description=self.runner.host_description,
            )

        try:
            xmlrpc_client = self.source_server.plesk_api()
            self.get_plesk_license_info(xmlrpc_client)
        except (URLError, WinexeRunnerException):
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            error_message = self._get_xmlrpc_down_diagnostics(plesk_api_url, service_status)
        except ParseError:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            error_message = _T("CHECK_WHETHER_THE_CONTROL_PANEL_IS_OPERATING").format(
                plesk_api_url=plesk_api_url,
                host_description=self.runner.host_description,
            )
        except MigrationError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            if 'MySQL' in service_status and service_status['MySQL']['exit_code'] != 0:
                error_message = _T("CHECK_WHETHER_THE_CONTROL_PANEL_IS_OPERATING").format(
                    plesk_api_url=plesk_api_url,
                    host_description=self.runner.host_description,
                )
            else:
                error_message = _T("UNABLE_TO_CHECK_THE_CONNECTION_TO_THE_PLESK_API").format(
                    plesk_api_url=plesk_api_url,
                    host_description=self.runner.host_description,
                    error_message=e.message,
                )
        except HttpClientError as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            if 'Control Panel Engine' in service_status and service_status['Control Panel Engine']['exit_code'] != 0:
                error_message = _T("CHECK_WHETHER_THE_CONTROL_PANEL_IS_OPERATING").format(
                    plesk_api_url=plesk_api_url,
                    host_description=self.runner.host_description,
                )
            else:
                error_message = _T("UNABLE_TO_CHECK_THE_CONNECTION_TO_THE_PLESK_API").format(
                    plesk_api_url=plesk_api_url,
                    host_description=self.runner.host_description,
                    error_message=e.cause,
                )
        except Exception as e:
            logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
            error_message = _T("UNABLE_TO_CHECK_THE_CONNECTION_TO_THE_PLESK_API").format(
                plesk_api_url=plesk_api_url,
                host_description=self.runner.host_description,
                error_message=e.message,
            )
        else:
            status = _T("SERVICE_STATUS_AVAILABLE")
            exit_code = 0

        return service, status, exit_code, error_message

    @trace
    def _get_xmlrpc_down_diagnostics(self, plesk_api_url, service_status):
        """Return appropriate error message in case if migrator cannot connect to XML API"""
        system_service_is_stopped = (
            service_status['Control Panel Web Server']['exit_code'] > 0 or
            service_status['Control Panel Database']['exit_code'] > 0
        )
        if system_service_is_stopped:
            error_message = _T("CHECK_WHETHER_THE_CONTROL_PANEL_IS_OPERATING").format(
                plesk_api_url=plesk_api_url,
                host_description=self.runner.host_description,
            )
        else:
            error_message = _T("CHECK_THE_FIREWALL_SETTINGS").format(
                plesk_api_url=plesk_api_url,
                host_description=self.runner.host_description,
            )

        return error_message

    @staticmethod
    @trace
    def get_plesk_license_info(xmlrpc_client):
        """Make server/get/key XML API request"""
        with xmlrpc_client.logger.censor({'xml': ["/packet/server/get/result/key/content"]}):
            operations = [xmlrpc.operator.server.ServerOperator.Dataset.KEY]
            return xmlrpc_client.send(xmlrpc.operator.server.ServerOperator.Get(operations))

    @staticmethod
    @trace
    def is_plesk_server_license_valid(source_server):
        xmlrpc_client = source_server.plesk_api()
        license_resource = PanelInfo.get_plesk_license_info(xmlrpc_client)
        return PanelInfo.is_plesk_license_valid(license_resource)

    @staticmethod
    @trace
    def is_plesk_license_valid(license_resource):
        if isinstance(license_resource, xmlrpc.core.Failure):
            return False
        else:
            try:
                logger.debug(_T("CHECK_PLESK_LICENSE_EXPIRATION_DATE"))
                expiry_date = datetime.strptime(license_resource.data.key.properties['lim_date'], '%Y-%m-%d')
                return expiry_date > datetime.now()
            except ValueError:
                # We consider license as valid, if 'lim_date' is not containing a date
                # For example, '0' is a sign of permanent license
                logger.debug(_T("LOG_EXCEPTION"), exc_info=True)
                return True

    @trace
    def _get_patch(self, mu_xml_path, product_id):
        if not self.runner.file_exists(mu_xml_path):
            return None

        xml = self.runner.get_file_contents(mu_xml_path)
        patches = etree.fromstring(str(xml))
        return patches.xpath(
            "product[@id = '{0}' and @version = '{1}']/patch/@version".format(product_id, self.version)
        )[0]

    @trace
    def ip(self):
        """Return main IP address of the server"""
        return self._config_section.get('ip')

    @cached
    @trace
    def agent_api(self):
        install_agent_on_server(self)

        url = 'https://{host}:{port}'.format(host=self.ip(), port=AGENT_PORT)
        oam_cert = get_oam_ssl_certificate()
        ssl_info = {
            'cert_file': oam_cert.cert_filename,
            'key_file': oam_cert.pkey_filename,
        }

        return JsonRpcClient(url, ssl_info)


class PleskUnixInfo(PanelInfo):
    """PanelInfo object for Linux installations of Plesk Panel"""
    def __init__(self, runner, config_section=None):
        super(PleskUnixInfo, self).__init__(runner, config_section)
        self._skeymng = None

    @staticmethod
    @trace
    def is_installed(runner):
        return runner.file_exists('/etc/psa/psa.conf')

    @property
    @trace
    def url(self):
        return 'https://{ip}:{port}'.format(ip=self.runner.ssh.ip, port=PLESK_PANEL_DEFAULT_PORT)

    @property
    @trace
    def version(self):
        plesk_dir = get_unix_product_root_dir(self.runner)
        version_str = self.runner.get_file_contents(posixpath.join(plesk_dir, 'version'))
        return version_str.split()[0]

    @property
    @trace
    def patch(self):
        mu_xml = '/root/.autoinstaller/microupdates.xml'
        return self._get_patch(mu_xml, 'plesk')

    @trace
    def get_service_list(self):
        if split_version_string(self.version) >= split_version_string('11.5'):
            linux_services_list = OrderedDict([
                ('Control Panel Web Server', dict(name_list=('sw-cp-server',))),
                ('Control Panel Engine', dict(name_list=('sw-engine',))),
                ('MySQL', dict(name_list=('mysql', 'mysqld', 'mariadb'))),
            ])
        else:
            linux_services_list = OrderedDict([
                ('Control Panel Web Server', dict(name_list=('psa',))),
                ('MySQL', dict(name_list=('mysql', 'mysqld', 'mariadb'))),
            ])

        linux_services_list.update(super(PleskUnixInfo, self).get_service_list())
        return linux_services_list

    @trace
    def get_service_status(self, service_name):
        status_string, exit_code, _ = get_service_status_linux(self.runner, service_name)
        return status_string, exit_code


class PleskWindowsInfo(PanelInfo):
    """PanelInfo object for Windows installations of Plesk Panel"""
    def __init__(self, runner, config_section=None):
        super(PleskWindowsInfo, self).__init__(runner, config_section)
        self._skeymng = None

    @staticmethod
    @trace
    def is_installed(runner):
        return runner.file_exists(ur"%plesk_dir%version")

    @property
    @trace
    def url(self):
        return 'https://{ip}:{port}'.format(ip=self.runner.settings.ip, port=PLESK_PANEL_DEFAULT_PORT)

    @property
    @trace
    def version(self):
        version_str = self.runner.sh(
            ur'cmd /C type {versionfile}', {'versionfile': ur"%plesk_dir%version"})
        return version_str.split()[0]

    @property
    @trace
    def patch(self):
        ai_keys = [
            'HKLM\SOFTWARE\Wow6432Node\Parallels\Installer',
            'HKLM\SOFTWARE\Parallels\Installer'
        ]
        ai_dir = windows_utils.get_from_registry(self.runner, ai_keys, 'HOMEDIR')
        if ai_dir is None:
            return None

        mu_xml = ntpath.join(ai_dir, 'microupdates.xml')
        return self._get_patch(mu_xml, 'panel')

    @trace
    def get_service_list(self):
        """Return a dictionary of services to be verified"""
        windows_services_list = OrderedDict([
            ('Control Panel Web Server', dict(name_list=('plesksrv',))),
            ('Control Panel Database', dict(name_list=('PleskSqlServer',))),
            ('IIS Web Server', dict(name_list=('W3SVC',))),
            ('Microsoft .NET 4 Framework', dict(func=self._get_dotnet_4_framework_status))
        ])

        windows_services_list.update(super(PleskWindowsInfo, self).get_service_list())
        return windows_services_list

    @trace
    def get_service_status(self, service_name):
        """Get status of a system service"""
        return get_service_status_windows(self.runner, service_name)

    @trace
    def _get_dotnet_4_framework_status(self, service_status):
        """
        Make sure .net framework version 4 is available.

        This is a prerequisite for Plesk for Windows migration agent.
        """
        service = ''
        dotnet_v4_key = 'HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\.NETFramework'
        if windows_utils.find_in_registry(self.runner, [dotnet_v4_key], 'v4.*'):
            status = _T("SERVICE_STATUS_AVAILABLE")
            error_message = None
            exit_code = 0
        else:
            status = _T("SERVICE_STATUS_NOT_INSTALLED")
            error_message = _T("NO_DOTNET_V4_FRAMEWORK_INSTALLED")
            exit_code = 1

        return service, status, exit_code, error_message
