# (C)2017 Ingram Micro Inc. all Rights Reserved
import logging
import ntpath
import posixpath
import re
from xml.etree import ElementTree

from migrator.core.language import _T
from migrator.core.utils import cached
from migrator.core.utils.networking.plesk_cli_runner import PleskCLIRunnerCLI
from migrator.core.utils.operating_system.windows_utils import get_from_registry, cmd_command

logger = logging.getLogger(__name__)


def get_windows_vhosts_dir(runner):
    vhosts_dir = get_from_registry(
        runner,
        ['HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Session Manager\Environment'],
        'plesk_vhosts'
    )

    if vhosts_dir is None:
        raise Exception(_T("FAILED_DETECT_PLESK_VIRTUAL_HOSTS_DIRECTORY"))

    vhosts_dir = vhosts_dir.strip('\\')
    logger.debug(_T("PLESK_VIRTUAL_HOSTS_DIRECTORY_IS_S") % vhosts_dir)
    return vhosts_dir


def get_windows_plesk_dir(runner):
    plesk_dir = get_from_registry(
        runner,
        ['HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\Session Manager\Environment'],
        'plesk_dir'
    )

    if plesk_dir is None:
        raise Exception(_T("FAILED_DETECT_PLESK_BASE_DIRECTORY"))

    plesk_dir = plesk_dir.strip('\\')
    logger.debug(_T("DEBUG_PLESK_BASE_DIRECTORY"), plesk_dir)
    return plesk_dir


def get_unix_vhosts_dir(runner):
    """Get directory with virtual hosts on Plesk for Unix server"""
    return get_unix_conf_var(
        runner, 'HTTPD_VHOSTS_D', _T("DEBUG_UNIX_VHOSTS_DIRECTORY")
    )


def get_unix_mailnames_dir(runner):
    """Get directory with mail messages on Plesk for Unix server"""
    return get_unix_conf_var(
        runner, '(PLESK|QMAIL)_MAILNAMES_D', _T("DEBUG_MAIL_MESSAGES_DIRECTORY")
    )


def get_unix_product_root_dir(runner):
    """Get directory where Plesk for Unix is installed"""
    return get_unix_conf_var(
        runner, 'PRODUCT_ROOT_D', _T("DEBUG_UNIX_PRODUCT_ROOT_DIRECTORY")
    )


def get_unix_conf_var(runner, var_name, description=None):
    """Read variable from psa.conf config of Plesk for Unix

    Arguments:
    - runner - runner for Plesk for Unix server (from run_command module)
    - var_name - regular expression to match variable name
    - description - description of variable, just for debug log

    Returns variable value.
    """
    stdout = runner.run(
        '/bin/grep', ['-m1', '-E', '^\s*%s' % var_name, '/etc/psa/psa.conf']
    )
    var_value = stdout.strip().replace('\t', ' ').partition(' ')[2].strip().rstrip('/')
    if description is not None:
        logger.debug('%s: %s', description, var_value)
    else:
        logger.debug(
            _T("PLESK_FOR_UNIX_CONFIGURATION_VARIABLE_S"),
            var_name, var_value
        )
    return var_value


def get_windows_db_server_credentials(server, db_type, db_host, db_port):
    def _get_credentials(_server, _db_type, _db_host, _db_port):
        exit_code, stdout, _ = _call_php_cli_unchecked(
            _server, 'cu\database-registrar.php', ['--get-credentials', '%s:%s' % (_db_host, _db_port), '-type', _db_type], log_output=False
        )
        if exit_code != 0:
            return None
        return stdout.strip()

    output = None
    if db_type == 'mssql' and db_host.startswith('.\\') and ':' not in db_host:
        # workaround for bug PPP-15779: port 0 specified in Plesk database for local MSSQL server,
        # installer automatically, so first try to check this case
        output = _get_credentials(server, db_type, '%s:0' % db_host, db_port)
    if output is None:
        output = _get_credentials(server, db_type, db_host, db_port)

    if output is None:
        raise Exception(_T("UNABLE_RETRIVE_CREDENTIALS_FOR_S_DATABASE") % (
            db_type, '%s:%s' % (db_host, db_port), server.description()
        ))

    xml = ElementTree.XML(output)
    return xml.findtext('username'), xml.findtext('password')


def get_unix_db_server_credentials(server, db_type, db_host, db_port):
    output = _call_php_cli(
        server, 'api-cli/database-registrar.php', ['--get-credentials', '%s:%s' % (db_host, db_port), '-type', db_type], log_output=False
    ).strip()
    xml = ElementTree.XML(output)
    return xml.findtext('username'), xml.findtext('password')


def get_plesk_for_unix_version(runner, plesk_dir):
    """Get Plesk version of Plesk for Unix

    Returns two first numbers of the version as tuple of integers, for example (12, 5) or (9, 0).

    :rtype: tuple[int]
    """
    version_str = runner.get_file_contents(posixpath.join(plesk_dir, 'version'))
    return tuple(map(int, re.split('[\. ]', version_str)[:2]))


def get_plesk_for_windows_version(runner, plesk_dir):
    """Get Plesk version for Plesk for Windows

    Returns two first numbers of the version as tuple of integers, for example (12, 5) or (9, 0).

    :rtype: tuple[int]
    """
    version_str = runner.get_file_contents(ntpath.join(plesk_dir, 'version'))
    return tuple(map(int, re.split('[\. ]', version_str)[:2]))


def suspend_subscription(server, subscription_idn_name):
    """
    :type server: parallels.core.connections.plesk_server.PleskServer |
    migrator.core.connections.source_server.SourceServer
    """
    runner = PleskCLIRunnerCLI(server)
    runner.run('subscription', ['--webspace-off', subscription_idn_name])


def query_plesk_db(panel_server, query, values_count=None, log_output=True):
    """
    The function runs an SQL query on the server with Plesk and parses data.
    The values_count parameter shows the number of columns in a query.
    If it is not specified, it will be divided automatically.

    :type panel_server: parallels.core.connections.plesk_server.PleskServer
    :type query: string
    :type values_count: int
    """
    command = _get_command_for_query(panel_server, query)
    with panel_server.runner() as runner:
        lines = runner.sh(command, log_output=log_output).splitlines()

    if len(lines) == 0:
        return []

    if panel_server.is_windows():
        # the first line is a header line
        data_lines = lines[1:]
    else:
        # column names are skipped with an option, so data starts from the 1st line
        data_lines = lines

    results = []
    for line in data_lines:
        if line.strip() != '':
            if panel_server.is_windows():
                # windows --direct-sql returns excess tab at the end of string
                line = line[:line.rfind('\t')]

            if values_count is None:
                line_data = line.split('\t')
            else:
                line_data = line.split('\t', values_count-1)

            results.append([d.strip() for d in line_data])

    return results


def query_plesk_db_one_value(panel_server, query, log_output=True):
    """
    The function runs an SQL query on the server with Plesk so as to retrieve a specific value.
    It is required for the cases when text data are stored in a field, for example, with carriage return symbols.

    :type panel_server: parallels.core.connections.plesk_server.PleskServer
    :type query: string
    """
    command = _get_command_for_query(panel_server, query)
    with panel_server.runner() as runner:
        result = runner.sh(command, log_output=log_output)

    if panel_server.is_windows():
        # the first line is a header line
        result = ''.join(result.splitlines(True)[1:])

    return result


def _get_command_for_query(panel_server, query):
    if panel_server.is_windows():
        command = ur'%s\admin\bin\dbclient --direct-sql --sql="{query}"' % panel_server.plesk_dir
        query = " ".join([line.strip() for line in query.split()])
    else:
        command = u"mysql --default-character-set=utf8 --silent --skip-column-names -u admin -p`cat /etc/psa/.psa.shadow` psa -e \"{query}\""

    return command.format(query=query)


def _call_php_cli(plesk_server, path, args, log_output=True):
    with plesk_server.runner() as runner:
        return runner.sh(_get_php_cli_command(plesk_server, path, args), log_output=log_output)


def _call_php_cli_unchecked(plesk_server, path, args, log_output=True):
    with plesk_server.runner() as runner:
        return runner.sh_unchecked(_get_php_cli_command(plesk_server, path, args), log_output=log_output)


def _get_php_cli_command(plesk_server, path, args):
    if plesk_server.is_windows():
        command_pattern = cmd_command(
            ur'"{plesk_dir}\admin\bin\php" -dauto_prepend_file="" '
            ur'"{plesk_dir}\admin\plib\{cli_path}" {args_string}'
        )
    else:
        command_pattern = ur'{plesk_dir}/admin/bin/php {plesk_dir}/admin/plib/{cli_path} {args_string}'
    return command_pattern.format(
        plesk_dir=plesk_server.plesk_dir,
        cli_path=path,
        args_string=' '.join(args)
    )


class PleskDbCredentialsFetcher(object):
    def __init__(self):
        self._db_credentials_cache = dict()

    @staticmethod
    @cached
    def get_instance():
        return PleskDbCredentialsFetcher()

    def get(self, plesk_server, db_type, db_params):
        """Get login and password for specified database server

        :rtype tuple
        """

        # Make cache_key more unique because ('mysql', 'localhost') case is a bad one
        server_id = getattr(plesk_server, 'plesk_host_id', plesk_server.ip())
        cache_key = (db_type, server_id, db_params.host)

        login, password = self._db_credentials_cache.get(
            cache_key,
            (None, None)  # default
        )

        if login is None or password is None:
            if plesk_server.is_windows():
                login, password = get_windows_db_server_credentials(
                    plesk_server, db_type, db_params.host, db_params.port
                )
            else:
                login = db_params.admin
                password = db_params.password
                if login is None or password is None:
                    login, password = get_unix_db_server_credentials(
                        plesk_server, db_type, db_params.host, db_params.port
                    )

            # put to cache
            cache_value = (login, password)
            self._db_credentials_cache[cache_key] = cache_value

        return login, password
